/* SoundEvents.c- sound event configurations
 *
 *  WSoundPrefs - WMSound Server Preferences Program
 *
 *  Copyright (c) 1998 Pascal G. Hofstee
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#include "WSoundPrefs.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <assert.h>

#ifndef PATH_MAX                                                     
#define PATH_MAX 1024                          
#endif

extern proplist_t WMSoundDomain;

typedef struct _Panel {
    WMFrame *frame;
    char *sectionName;

    CallbackRec callbacks;

    WMWindow *win;

    WMFrame *sndevntF;
    WMPopUpButton *sndevntP;
    WMLabel *sndevntL;

    WMFrame *sndfileF;
    WMTextField *sndfileT;
    WMButton *sndplayB;
    WMButton *sndclearB;
    WMButton *sndbrowseB;

    WMFrame *sndsetF;
    WMButton *sndsetloadB;
    WMButton *sndsetsaveB;
    WMLabel *sndsetL;

    WMFrame *devF;
    WMLabel *devL;
    WMTextField *devT;

    char lastBrowseDir[PATH_MAX];
    char lastLoadDir[PATH_MAX];
    char lastSaveDir[PATH_MAX];
} _Panel;

#define ICON_FILE	"events"
#define LOAD_SOUNDSET_ICON_FILE	"loadset"
#define SAVE_SOUNDSET_ICON_FILE	"saveset"

static void changeSoundEvent(WMWidget *w, void *data);

static void playClick(WMWidget *w, void *data);
static void clearClick(WMWidget *w, void *data);
static void browseClick(WMWidget *w, void *data);

static void loadClick(WMWidget *w, void *data);
static void saveClick(WMWidget *w, void *data);

static char* get_path_from_name(char *path);
static char* checkSoundPath(char* file);
static void changedSound(void *observerData, WMNotification *notification);
static void changedDevice(void *observerData, WMNotification *notification);

static void showData(_Panel *panel);

static char* expandSoundPath(WMWidget *w, void *data, char* file);
int getEventID(char *eventkey);


static char *eventkey[] = {
        "Startup",
        "AppStart",
        "AppExit",
        "Maximize",
        "UnMaximize",
        "Iconify",
        "DeIconify",
        "Hide",
        "UnHide",
        "Shade",
        "UnShade",
        "UserDefined",
        NULL
};


static void                                                             
changedDevice(void *observerData, WMNotification *notification)      
{
    WMTextField *t = (WMTextField*)WMGetNotificationObject(notification);
 
    SetStringForKey(WMGetTextFieldText(t), "Device");             
}


static void
changedSound(void *observerData, WMNotification *notification)
{
    _Panel *panel = (_Panel*)observerData;
    WMTextField *t = (WMTextField*)WMGetNotificationObject(notification);      

    SetStringForKey(WMGetTextFieldText(t), eventkey[WMGetPopUpButtonSelectedItem(panel->sndevntP)]);
}


static void                                                             
showData(_Panel *panel)                                              
{
    WMSetTextFieldText(panel->sndfileT, GetStringForKey("Startup"));    
}


static void
createPanel(Panel *p)
{
    _Panel *panel = (_Panel*)p;
    WMScreen *scr = WMWidgetScreen(panel->win);
    WMFont *font = WMSystemFontOfSize(scr, 10);
    WMPixmap *icon;
    RColor color;
    char *path;

    color.red = 0xaa;
    color.green = 0xae;
    color.blue = 0xaa;

    panel->frame = WMCreateFrame(panel->win);
    WMResizeWidget(panel->frame, FRAME_WIDTH, FRAME_HEIGHT);
    WMMoveWidget(panel->frame, FRAME_LEFT, FRAME_TOP);

    /*********** Sound Event **********/

    panel->sndevntF = WMCreateFrame(panel->frame);
    WMResizeWidget(panel->sndevntF, 230, 125);
    WMMoveWidget(panel->sndevntF, 15, 10);
    WMSetFrameTitle(panel->sndevntF, _("Sound Event"));

    panel->sndevntP = WMCreatePopUpButton(panel->sndevntF);
    WMResizeWidget(panel->sndevntP, 200, 20);
    WMMoveWidget(panel->sndevntP, 15, 30);
    WMAddPopUpButtonItem(panel->sndevntP, _("WMSound Startup"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Application Startup"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Application Exit"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Maximize"));
    WMAddPopUpButtonItem(panel->sndevntP, _("UnMaximize"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Iconify"));
    WMAddPopUpButtonItem(panel->sndevntP, _("DeIconify"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Hide"));
    WMAddPopUpButtonItem(panel->sndevntP, _("UnHide"));
    WMAddPopUpButtonItem(panel->sndevntP, _("Shade"));
    WMAddPopUpButtonItem(panel->sndevntP, _("UnShade"));
    WMSetPopUpButtonAction(panel->sndevntP, changeSoundEvent, panel);
    WMSetPopUpButtonSelectedItem(panel->sndevntP, 0);

    panel->sndevntL = WMCreateLabel(panel->sndevntF);
    WMResizeWidget(panel->sndevntL, 211, 62);
    WMMoveWidget(panel->sndevntL, 15, 55);
    WMSetLabelTextAlignment(panel->sndevntL, WACenter);
    WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                     "the WMSound Server is started.\n"));



    WMMapSubwidgets(panel->sndevntF);


    /*************** Sound File *****************/

    panel->sndfileF = WMCreateFrame(panel->frame);
    WMResizeWidget(panel->sndfileF, 230, 85);
    WMMoveWidget(panel->sndfileF, 15, 140);
    WMSetFrameTitle(panel->sndfileF, _("Sound File"));

    panel->sndfileT = WMCreateTextField(panel->sndfileF);
    WMResizeWidget(panel->sndfileT, 200, 20);
    WMMoveWidget(panel->sndfileT, 15, 20);
    WMAddNotificationObserver(changedSound, panel, WMTextDidChangeNotification,
                              panel->sndfileT);

    panel->sndplayB = WMCreateCommandButton(panel->sndfileF);
    WMResizeWidget(panel->sndplayB, 60, 24);
    WMMoveWidget(panel->sndplayB, 15, 50);
    WMSetButtonText(panel->sndplayB, _("Play"));
    WMSetButtonAction(panel->sndplayB, (WMAction *)playClick, panel);

    panel->sndclearB = WMCreateCommandButton(panel->sndfileF);
    WMResizeWidget(panel->sndclearB, 60, 24);
    WMMoveWidget(panel->sndclearB, 85, 50);
    WMSetButtonText(panel->sndclearB, _("Clear"));
    WMSetButtonAction(panel->sndclearB, (WMAction *)clearClick, panel);

    panel->sndbrowseB = WMCreateCommandButton(panel->sndfileF);
    WMResizeWidget(panel->sndbrowseB, 60, 24);
    WMMoveWidget(panel->sndbrowseB, 155, 50);
    WMSetButtonText(panel->sndbrowseB, _("Browse"));
    WMSetButtonAction(panel->sndbrowseB, (WMAction *)browseClick, panel);

    WMMapSubwidgets(panel->sndfileF);


    /*************** Sound Sets *****************/

    panel->sndsetF = WMCreateFrame(panel->frame);
    WMResizeWidget(panel->sndsetF, 249, 76);
    WMMoveWidget(panel->sndsetF, 260, 10);
    WMSetFrameTitle(panel->sndsetF, _("Sound Sets"));

    panel->sndsetloadB = WMCreateCommandButton(panel->sndsetF);
    WMResizeWidget(panel->sndsetloadB, 48, 48);
    WMMoveWidget(panel->sndsetloadB, 12, 18);
    WMSetButtonImagePosition(panel->sndsetloadB, WIPImageOnly);
    WMSetButtonAction(panel->sndsetloadB, (WMAction *)loadClick, panel);
    path = LocateImage(LOAD_SOUNDSET_ICON_FILE);
    if (path) {
        icon = WMCreateBlendedPixmapFromFile(scr, path, &color);
        if (icon) {
            WMSetButtonImage(panel->sndsetloadB, icon);
            WMReleasePixmap(icon);
        } else {
            wwarning(_("could not load icon %s"), path);
        }
        free(path);
    }
    
    panel->sndsetsaveB = WMCreateCommandButton(panel->sndsetF);
    WMResizeWidget(panel->sndsetsaveB, 48, 48);
    WMMoveWidget(panel->sndsetsaveB, 60, 18);
    WMSetButtonImagePosition(panel->sndsetsaveB, WIPImageOnly);
    WMSetButtonAction(panel->sndsetsaveB, (WMAction *)saveClick, panel);
    path = LocateImage(SAVE_SOUNDSET_ICON_FILE);
    if (path) {
        icon = WMCreateBlendedPixmapFromFile(scr, path, &color);
        if (icon) {
            WMSetButtonImage(panel->sndsetsaveB, icon);
            WMReleasePixmap(icon);
        } else {
            wwarning(_("could not load icon %s"), path);
        }
        free(path);
    }

    panel->sndsetL = WMCreateLabel(panel->sndsetF);
    WMResizeWidget(panel->sndsetL, 129, 50);
    WMMoveWidget(panel->sndsetL, 116, 18);
    WMSetLabelTextAlignment(panel->sndsetL, WALeft);
    WMSetLabelText(panel->sndsetL, _("Load and save entire soundsets."));

    WMMapSubwidgets(panel->sndsetF);


    /********** sound device ***********/

    panel->devF = WMCreateFrame(panel->frame);
    WMResizeWidget(panel->devF, 249, 92);
    WMMoveWidget(panel->devF, 260, 93);
    WMSetFrameTitle(panel->devF, _("Sound Device"));

    panel->devT = WMCreateTextField(panel->devF);
    WMResizeWidget(panel->devT, 219, 20);
    WMMoveWidget(panel->devT, 15, 20);
    WMSetTextFieldText(panel->devT, GetStringForKey("Device"));
    WMAddNotificationObserver(changedDevice, panel, WMTextDidChangeNotification,
                              panel->devT);

    panel->devL = WMCreateLabel(panel->devF);
    WMResizeWidget(panel->devL, 219, 30);
    WMMoveWidget(panel->devL, 15, 50);
    WMSetLabelTextAlignment(panel->devL, WACenter);
    WMSetLabelText(panel->devL, _("Specify the sound device that should be "\
                   "used to play the sounds"));

    WMMapSubwidgets(panel->devF);



    WMRealizeWidget(panel->frame);
    WMMapSubwidgets(panel->frame);

    WMReleaseFont(font);

    showData(panel);
    strcpy(panel->lastBrowseDir, "\0");
}


static void
storeData(_Panel *panel)
{
/*
//    SetSpeedForKey(i, "IconSlideSpeed");


//    SetSpeedForKey(i, "ShadeSpeed");

//    SetBoolForKey(WMGetButtonSelected(panel->newsB), "NewStyle");

//    SetBoolForKey(!WMGetButtonSelected(panel->animB), "DisableAnimations");
//    SetBoolForKey(WMGetButtonSelected(panel->supB), "Superfluous");
//    SetBoolForKey(!WMGetButtonSelected(panel->sfxB), "DisableSound");

//    SetBoolForKey(WMGetButtonSelected(panel->dithB), "DisableDithering");
//    SetIntegerForKey(WMGetSliderValue(panel->dithS), "ColormapSize");
*/
}


static void                               
changeSoundEvent(WMWidget *w, void *data)
{
    _Panel *panel = (_Panel*)data;

    WMSetTextFieldText(panel->sndfileT, GetStringForKey(eventkey[WMGetPopUpButtonSelectedItem(w)]));

    switch (WMGetPopUpButtonSelectedItem(w)) {
     case 0:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
			"the WMSound Server is started.\n"));
        break;
     case 1:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a new application is started."));
        break;
     case 2:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "an application is shut down."));
        break;
     case 3:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is maximized."));
        break;
     case 4:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is restored to it's\n"\
			"original window size again."));
        break;
     case 5:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is iconified."));
        break;
     case 6:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is restored to a\n"\
			"non-iconified state."));
        break;
     case 7:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "an application is hidden behind it's\n"\
			"application icon."));
        break;
     case 8:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "an application is restored from\n"\
			"underneath it's application icon."));
        break;
     case 9:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is reduced to a\n"\
			"mere titlebar."));
        break;
     case 10:
        WMSetLabelText(panel->sndevntL, _("This sound will be played when\n"\
                        "a window is restored from a\n"\
			"shaded state."));
        break;
    }
}


Panel*
InitSoundEvents(WMScreen *scr, WMWindow *win)
{
    _Panel *panel;

    panel = wmalloc(sizeof(_Panel));
    memset(panel, 0, sizeof(_Panel));

    panel->sectionName = _("Sound Events / Device Configuration");

    panel->win = win;

    panel->callbacks.createWidgets = createPanel;
    panel->callbacks.updateDomain = storeData;

    AddSection(panel, ICON_FILE);

    return panel;
}


static void
browseClick(WMWidget *w, void *data)
{
    _Panel *panel = (_Panel*)data;
    WMOpenPanel *browseP;
    char path[PATH_MAX] = "\0";
    char *spath = NULL;

    WMSetButtonEnabled(panel->sndbrowseB, NO);
    WMSetButtonEnabled(panel->sndsetloadB, NO);
    WMSetButtonEnabled(panel->sndsetsaveB, NO);

    if ((strcmp(panel->lastBrowseDir,"\0") == 0)) {
      strcpy(path, wexpandpath("~/GNUstep/Library/WindowMaker/Sounds"));
      spath = wmalloc(strlen(path)+2);
      strcpy(spath, path);
    }
    else {
      spath = wmalloc(strlen(panel->lastBrowseDir)+2);
      strcpy(spath, panel->lastBrowseDir);
    }


    browseP = WMGetOpenPanel(WMWidgetScreen(w));
    WMSetFilePanelCanChooseDirectories(browseP, NO);
    WMSetFilePanelCanChooseFiles(browseP, YES);
    if (WMRunModalOpenPanelForDirectory(browseP, panel->win, spath , NULL, NULL)) {
      char *filename;
      filename = WMGetFilePanelFileName(browseP);
      
      strcpy(panel->lastBrowseDir,(get_path_from_name(filename)));
      WMSetTextFieldText(panel->sndfileT,checkSoundPath(filename));
      SetStringForKey(WMGetTextFieldText(panel->sndfileT), eventkey[WMGetPopUpButtonSelectedItem(panel->sndevntP)]);
      free(filename);
    }
    WMFreeFilePanel(browseP);

    WMSetButtonEnabled(panel->sndbrowseB, YES);
    WMSetButtonEnabled(panel->sndsetloadB, YES);
    WMSetButtonEnabled(panel->sndsetsaveB, YES);
    free(spath);
}

static void
loadClick(WMWidget *w, void *data)
{
    _Panel *panel = (_Panel*)data;
    WMOpenPanel *loadP;
    proplist_t newset;
    char path[PATH_MAX] = "\0";
    char *spath = NULL;

    WMSetButtonEnabled(panel->sndsetloadB, NO);
    WMSetButtonEnabled(panel->sndbrowseB, NO);
    WMSetButtonEnabled(panel->sndsetsaveB, NO);

    if ((strcmp(panel->lastLoadDir,"\0") == 0)) {
      strcpy(path, wexpandpath("~/GNUstep/Library/WindowMaker/SoundSets"));
      spath = wmalloc(strlen(path)+2);
      strcpy(spath, path);
    }
    else {
      spath = wmalloc(strlen(panel->lastLoadDir)+2);
      strcpy(spath, panel->lastLoadDir);
    }


    loadP = WMGetOpenPanel(WMWidgetScreen(w));
    WMSetFilePanelCanChooseDirectories(loadP, NO);
    WMSetFilePanelCanChooseFiles(loadP, YES);
    if (WMRunModalOpenPanelForDirectory(loadP, panel->win, spath , NULL, NULL)) {
      char *filename;
      filename = WMGetFilePanelFileName(loadP);

      strcpy(panel->lastLoadDir,(get_path_from_name(filename)));
      newset = PLGetProplistWithPath(filename);

      if(!newset) {
        WMRunAlertPanel(WMWidgetScreen(w), panel->win, _("Error"), _("Could not load specified SoundSet"), _("OK"), NULL, NULL);
      }
      else {
        PLMergeDictionaries(WMSoundDomain,newset);
        WMSetTextFieldText(panel->sndfileT, GetStringForKey(eventkey[WMGetPopUpButtonSelectedItem(panel->sndevntP)]));
      }
      PLRelease(newset);
      free(filename);
    }
    WMFreeFilePanel(loadP);

    WMSetButtonEnabled(panel->sndsetloadB, YES);
    WMSetButtonEnabled(panel->sndbrowseB, YES);
    WMSetButtonEnabled(panel->sndsetsaveB, YES);

    free(spath);
}


static void
saveClick(WMWidget *w, void *data)
{
    _Panel *panel = (_Panel*)data;
    WMSavePanel *saveP;
    proplist_t newset, key, val;
    char path[PATH_MAX] = "\0";
    char *spath = NULL;
    FILE *file;
    int i;

    WMSetButtonEnabled(panel->sndbrowseB, NO);
    WMSetButtonEnabled(panel->sndsetloadB, NO);
    WMSetButtonEnabled(panel->sndsetsaveB, NO);

    if ((strcmp(panel->lastSaveDir,"\0") == 0)) {
      strcpy(path, wexpandpath("~/GNUstep/Library/WindowMaker/SoundSets"));
      spath = wmalloc(strlen(path)+2);
      strcpy(spath, path);
    }
    else {
      spath = wmalloc(strlen(panel->lastSaveDir)+2);
      strcpy(spath, panel->lastSaveDir);
    }


    saveP = WMGetSavePanel(WMWidgetScreen(w));
    WMSetFilePanelCanChooseDirectories(saveP, NO);
    WMSetFilePanelCanChooseFiles(saveP, YES);
    if (WMRunModalSavePanelForDirectory(saveP, panel->win, spath , NULL)) {
      char *filename = NULL;
      filename = WMGetFilePanelFileName(saveP);
      strcpy(panel->lastSaveDir,(get_path_from_name(filename)));

      file = fopen (filename, "w+");
      if (file) {
        newset = PLMakeDictionaryFromEntries(NULL,NULL,NULL);

        for(i=0;eventkey[i]!="UserDefined";i++) {
          key = PLMakeString(eventkey[i]);
          val = PLGetDictionaryEntry(WMSoundDomain,key);
          if(val) PLInsertDictionaryEntry(newset,key,val);
        }

        PLRelease(key);

        fprintf (file, PLGetDescriptionIndent(newset,0));
        PLRelease(newset);
        fclose (file);
      }
      else {
        WMRunAlertPanel(WMWidgetScreen(w), panel->win, _("Error"), _("Could not save SoundSet"), _("OK"), NULL, NULL);
      }

      free(filename);
    }
    WMFreeFilePanel(saveP);

    WMSetButtonEnabled(panel->sndbrowseB, YES);
    WMSetButtonEnabled(panel->sndsetloadB, YES);
    WMSetButtonEnabled(panel->sndsetsaveB, YES);

    free(spath);
}


static void                                  
clearClick(WMWidget *w, void *data)                
{
    _Panel *panel = (_Panel*)data;
    
   WMSetTextFieldText(panel->sndfileT, "None");
}


static void
playClick(WMWidget *w, void *data)
{
    _Panel 	*panel = (_Panel*)data;
    char 	*text, *soundfile;

    text = WMGetTextFieldText(panel->sndfileT);

    soundfile = expandSoundPath(w, data, text);
    wmsndPlaySound(soundfile);
    free(soundfile);
}


static char*
get_path_from_name(char *path)
{
    int size;
    char result[PATH_MAX];

    assert(path!=NULL);

    size = strlen(path);

    while (size > 1 && path[size-1] != '/') {
        size--;
    }

    return strncpy(result, path, size);
}

static char*
checkSoundPath(char* file)
{
    proplist_t array, val;
    int nr_elem;
    int i = 0;
    char *path;

    assert(file!=NULL);

    array = GetObjectForKey("SoundPath");
    nr_elem = PLGetNumberOfElements(array);

    while (i < nr_elem) {
      val = PLGetArrayElement(array, i);
      path = wexpandpath(PLGetString(val));

      if ((path+strlen(path)-1) != "/") {
        strcat(path, "/");
      }

      if ((strncmp(path, file, strlen(path)-1)) == 0) {
        strcpy(file, file+strlen(path));
        break;
      }
      i++;
    }
    return file;
}


static char*
expandSoundPath (WMWidget *w, void *data, char* file)
{
    _Panel *panel = (_Panel*)data;
    proplist_t array, val;
    int nr_elem;
    int i = 0;
    char *path;
    char *tmp;

    assert(file!=NULL);

    if (strncmp(file, "/", 1) == 0) {
      return wstrdup(file);
    }

    array = GetObjectForKey("SoundPath");
    nr_elem = PLGetNumberOfElements(array);

    while (i < nr_elem) {
      val = PLGetArrayElement(array, i);
      path = wexpandpath(PLGetString(val));

      if ((path+strlen(path)-1) != "/") {
        tmp = wstrappend(path, "/");
        if (path)
          free(path);
        path = tmp;
      }

      tmp = wstrappend(path, file);
      if (path)
        free(path);
      path = tmp;

      if (access(path, R_OK) == 0) {
#ifdef DEBUG
	fprintf(stderr, "%s \t Found !\n", path);
#endif
        return path;
      }
#ifdef DEBUG
      fprintf(stderr, "%s \t Not Found!\n", path);
#endif
      free(path);
      i++;
    }
#ifdef DEBUG
    fprintf(stderr,"Tough Luck ... couldn't find specified sound.\n");
#endif
    WMRunAlertPanel(WMWidgetScreen(w), panel->win, _("Error"), _("Unable to locate specified soundfile.\n It either does not exist or is not accessible for the current user."), _("OK"), NULL, NULL);
    return wstrdup(file);
}


int
getEventID(char *eventkey)
{
    if (strcmp (eventkey, WMSND_STARTUP) == 0)
	return MIN_SOUND_ID;
    if (strcmp (eventkey, WMSND_SHADE) == 0)
	return MIN_SOUND_ID + 1;
    if (strcmp (eventkey, WMSND_UNSHADE) == 0)
        return MIN_SOUND_ID + 2;
    if (strcmp (eventkey, WMSND_MAXIMIZE) == 0)
	return MIN_SOUND_ID + 3;
    if (strcmp (eventkey, WMSND_UNMAXIMIZE) == 0)
	return MIN_SOUND_ID + 4;
    if (strcmp (eventkey, WMSND_ICONIFY) == 0)
	return MIN_SOUND_ID + 5;
    if (strcmp (eventkey, WMSND_DEICONIFY) == 0)
	return MIN_SOUND_ID + 6;
    if (strcmp (eventkey, WMSND_HIDE) == 0)
	return MIN_SOUND_ID + 7;
    if (strcmp (eventkey, WMSND_UNHIDE) == 0)
	return MIN_SOUND_ID + 8;
    if (strcmp (eventkey, WMSND_APPSTART) == 0)
	return MIN_SOUND_ID + 9;
    if (strcmp (eventkey, WMSND_APPEXIT) == 0)
	return MIN_SOUND_ID + 10;
    return 0;
}
