
#include "structures.h"
#include "moduleserver.h"
#include "synthmodule.h"

#define mico_throw(foo) ;

#define dname(dir) ((dir)==Arts::input?"input":"output")
#define pstat \
	printf("port name %s, direction %s, id %d\n",_Name.c_str(),dname(_Type.Direction),_ID);

PortDesc_impl::PortDesc_impl(Arts::ModuleDesc_ptr parent, string name,
															Arts::PortType type)
{
	char * pname = parent->Name();
	describe("PortDesc."+string(pname)+string(".")+name);
	_Name = name;
	_Type = type;
	_Parent = parent;
	_isConnected = false;
	_hasValue = false;
	_Connections = new Arts::PortDescSeq;
	_FloatValue = 0.0;
	_StringValue = "";
	Arts::StructureDesc_var sd = parent->Parent();
	_ID = sd->obtainID();
}

void PortDesc_impl::cleanUp()
{
	disconnectAll();
	delete _Connections;
}

void PortDesc_impl::disconnectAll()
{
	// disconnect all connected ports
	while(_Connections->length())
		(*_Connections)[0]->disconnectFrom(this);
}

PortDesc_impl::~PortDesc_impl()
{
}

// Implementation for interface PortDesc
CORBA::Long PortDesc_impl::ID()
{
	return _ID; 
}

Arts::ModuleDesc_ptr PortDesc_impl::Parent()
{
	return Arts::ModuleDesc::_duplicate(_Parent); 
}

char* PortDesc_impl::Name()
{
	return CORBA::string_dup(_Name.c_str()); 
}

Arts::PortType PortDesc_impl::Type()
{
	return _Type; 
}

CORBA::Boolean PortDesc_impl::isConnected()
{
	return _isConnected; 
}

CORBA::Boolean PortDesc_impl::hasValue()
{
	return _hasValue; 
}

void PortDesc_impl::hasValue(CORBA::Boolean newvalue)
{
	if(_hasValue != newvalue)
	{
		assert(newvalue == false);
		_hasValue = newvalue;
	}
}

Arts::PortDescSeq* PortDesc_impl::Connections()
{
	Arts::PortDescSeq* retval = new Arts::PortDescSeq(*_Connections);
	return retval; 
}

CORBA::Float PortDesc_impl::FloatValue()
{
	assert(_hasValue);
	assert(_Type.DataType == Arts::audio_data);
	return _FloatValue; 
}

void PortDesc_impl::FloatValue( CORBA::Float _new_value )
{
	assert(!_isConnected);
	assert(_Type.Direction == Arts::input);
	assert(_Type.DataType == Arts::audio_data);
	_FloatValue = _new_value;
	_hasValue = true;
}

char* PortDesc_impl::StringValue()
{
	assert(_hasValue);
	assert(_Type.DataType == Arts::string_data);
	return CORBA::string_dup(_StringValue.c_str());
}

void PortDesc_impl::StringValue( const char* _new_value )
{
	assert(!_isConnected);	// shouldn't happen, but check anyway
	printf("StringValue(%s) of ",_new_value);
	pstat;
	assert(_Type.Direction == Arts::input);
	assert(_Type.DataType == Arts::string_data);
	_StringValue = _new_value;
	_hasValue = true;
}

CORBA::Boolean
PortDesc_impl::connectTo( Arts::PortDesc_ptr port )
{
	// check if we are already connected to that port:

	unsigned long i;
	for(i=0;i<_Connections->length();i++)
		if((*_Connections)[i]->ID() == port->ID()) return true;

	Arts::PortType rType = port->Type();

	// only stream or event channels may be connected
	if( _Type.ConnType != rType.ConnType || _Type.ConnType == Arts::property)
		return false;

	// only same data type connections allowed
	if( _Type.DataType != rType.DataType ) return false;

	// only opposite directions
	if( _Type.Direction == rType.Direction ) return false;

	// always first connect the input port to the output port and
	// then the other direction

	if( _Type.Direction == Arts::input )
	{
		if(!_isConnected)
		{
			assert(_Connections->length() == 0);
			_Connections->length(1);
			(*_Connections)[0] = Arts::PortDesc::_duplicate(port);

			port->internalConnectInput(this);

			_isConnected = true;
			return true;
		}
	}
	if( _Type.Direction == Arts::output )
	{
		return port->connectTo(this);
	}
	return false;
}

void PortDesc_impl::internalConnectInput( Arts::PortDesc_ptr port )
{
	unsigned long l = _Connections->length();

	_Connections->length(l+1);
	(*_Connections)[l] = Arts::PortDesc::_duplicate(port);
	_isConnected = true;
}

void
PortDesc_impl::disconnectFrom( Arts::PortDesc_ptr port )
{
	unsigned long i,found = 0,newlen = 0;
	Arts::PortDescSeq *_NConnections = new Arts::PortDescSeq;

	printf("port %ld disconnecting from port %ld\n",ID(),port->ID());

	for(i=0;i<_Connections->length();i++)
	{
		Arts::PortDesc_ptr other = (*_Connections)[i];
		if(other->ID() != port->ID())
		{
			_NConnections->length(newlen+1);
			(*_NConnections)[newlen] =
				Arts::PortDesc::_duplicate((*_Connections)[i]);
			newlen++;
		}
		else found++;
	}

	delete _Connections;
	_Connections = _NConnections;

	_isConnected = (_Connections->length()>0);
	printf("_Parent = %s, isConnected = %d\n",_Parent->Name(),_isConnected);
	if(found) port->disconnectFrom(this);
}

void ModuleDesc_impl::cleanUp()
{
	unsigned long i;
	for(i=0;i<_Ports->length();i++)
		(*_Ports)[i]->decRef();
	delete _Ports;
}

// Implementation for interface ModuleDesc
CORBA::Long ModuleDesc_impl::ID()
{
	return _ID; 
}

Arts::StructureDesc_ptr ModuleDesc_impl::Parent()
{
	return Arts::StructureDesc::_duplicate(_Parent); 
}

char* ModuleDesc_impl::Name()
{
	return CORBA::string_dup(_Name.c_str()); 
}

Arts::PortDescSeq* ModuleDesc_impl::Ports()
{
	Arts::PortDescSeq* retval = new Arts::PortDescSeq(*_Ports);
	return retval; 
}

CORBA::Long ModuleDesc_impl::X()
{
	return _X;
}

CORBA::Long ModuleDesc_impl::Y()
{
	return _Y;
}

CORBA::Long ModuleDesc_impl::Width()
{
  CORBA::Long retval;

  // add your implementation here
    // REMOVE  
    mico_throw(::CORBA::NO_IMPLEMENT());
    // REMOVE 

  return retval; 
}

CORBA::Long ModuleDesc_impl::Height()
{
  CORBA::Long retval;

  // add your implementation here
    // REMOVE  
    mico_throw(::CORBA::NO_IMPLEMENT());
    // REMOVE 

  return retval; 
}


CORBA::Boolean
ModuleDesc_impl::moveTo( CORBA::Long X, CORBA::Long Y )
{
	// fixme: collision checking!
	_X = X;
	_Y = Y;

	return(true);
}


// Implementation for interface StructureDesc
CORBA::Long StructureDesc_impl::Width()
{
  CORBA::Long retval;

  // add your implementation here
    // REMOVE  
    mico_throw(::CORBA::NO_IMPLEMENT());
    // REMOVE 

  return retval; 
}

CORBA::Long StructureDesc_impl::Height()
{
  CORBA::Long retval;

  // add your implementation here
    // REMOVE  
    mico_throw(::CORBA::NO_IMPLEMENT());
    // REMOVE 

  return retval; 
}

/*
 * Query the module for it's paramenters
 */

ModuleDesc_impl::ModuleDesc_impl( Arts::StructureDesc_ptr parent,
										SynthModule *module, string name )
{
	describe(string("ModuleDesc ")+name);
	_Name = name;
	_X = -1;		// no position assigned
	_Y = -1;
	_ID = parent->obtainID();
	_Parent = parent;
	_Ports = new Arts::PortDescSeq;
	_Ports->length(0);

	collectPorts(module, SynthModule::arg_in, Arts::input);
	collectPorts(module, SynthModule::arg_out,Arts::output);
}

ModuleDesc_impl::~ModuleDesc_impl()
{
}

long ModuleDesc_impl::collectPorts( SynthModule *module, SynthModule::ArgType mdir, Arts::PortDirection pdir )
{
	string portname;
	char *pn;
	long portcount = 0;
	long len;

	Arts::PortType porttype;
	Arts::PortDesc_ptr port;

	while((portname = module->getArgument(portcount,mdir)) != "")
	{
		porttype.Direction = pdir;
		porttype.DataType = Arts::audio_data;
		porttype.ConnType = Arts::stream;

		if(portname[0] == '_')		// is there a better way?
		{
			pn = strdup(portname.c_str());
			pn ++;					// cut the first byte
			portname = pn;
			pn --;					// but free the right thing
			free(pn);

			porttype.DataType = Arts::string_data;
			porttype.ConnType = Arts::property;
		}

		printf("#%d: %s\n",portcount,portname.c_str());

		port = new PortDesc_impl(this, portname, porttype);
		port->incRef();			// this is my port

		len = _Ports->length();
		_Ports->length(len+1);
		(*_Ports)[len] = Arts::PortDesc::_duplicate(port);

		portcount++;
	};
	return(portcount);
}

Arts::ModuleDesc_ptr
StructureDesc_impl::createModuleDesc( const Arts::ModuleInfo& info )
{
	return createModuleDesc(info.name);
}

Arts::ModuleDesc_ptr StructureDesc_impl::createModuleDesc( const char *name )
{
	/*
	 * to create a representation of a specified synth module, we
	 *
	 * - create an instance of this synth module by contacting the
	 *   module server and telling him to do so (result is a C++ class)
	 *
	 * - create an instance of a ModuleDesc, and tell it to query the
	 *   module for it's parameters (result is a CORBA object)
	 *
	 * - destroy the synth module (C++ class) again and return a reference
	 *   to the CORBA object
	 */

	ModuleServer<SynthModule> *MS_SynthModule;
	MS_SynthModule = (ModuleServer<SynthModule> *)SynthModule::get_MS();

	SynthModule *m = (SynthModule *)MS_SynthModule->getModule(name);

  	Arts::ModuleDesc_ptr moduledesc =
		Arts::ModuleDesc::_duplicate(new ModuleDesc_impl(this, m, name));

	delete m;

	long mcount = _Modules->length();
	_Modules->length(mcount+1);
	(*_Modules)[mcount++] = Arts::ModuleDesc::_duplicate(moduledesc);
	moduledesc->incRef();	// this is my! module now ;)

	return moduledesc;
}

void StructureDesc_impl::freeModuleDesc(Arts::ModuleDesc_ptr moduledesc)
{
	Arts::ModuleDescSeq *new_Modules = new Arts::ModuleDescSeq;
	int i,newcount = 0;

	for(i=0;i<_Modules->length();i++)
	{
		if((*_Modules)[i]->ID() != moduledesc->ID())
		{
			// ok, this should not be deleted

			new_Modules->length(newcount+1);
			(*new_Modules)[newcount] =
				Arts::ModuleDesc::_duplicate((*_Modules)[i]);
			newcount++;
		}
		else
		{
			(*_Modules)[i]->decRef();		// will get freed automagically
		}
	}
	delete _Modules;
	_Modules = new_Modules;
}

Arts::ModuleDescSeq *StructureDesc_impl::Modules()
{
	Arts::ModuleDescSeq *retval = new Arts::ModuleDescSeq(*_Modules);
	return(retval);
}

StructureDesc_impl::StructureDesc_impl()
{
	describe("StructureDesc");
	nextID = 0;
	_Modules = new Arts::ModuleDescSeq;
}

StructureDesc_impl::~StructureDesc_impl()
{
	printf("StructureDesc released...\n");
}

CORBA::Long StructureDesc_impl::obtainID()
{
	return(nextID++);
}

void StructureDesc_impl::cleanUp()
{	
	printf("cleanUp...\n");
	clear();							// frees all modules
	delete _Modules;
}

void StructureDesc_impl::clear()
{
	unsigned long i;
	for(i=0;i<_Modules->length();i++)
		(*_Modules)[i]->decRef();		// I don't need that module any longer
	_Modules->length(0);				// should remove every module
}

// "file" management

void sqprintf(Arts::StringSeq *list, const char *fmt, ...)
{
	char p[1024];
	va_list ap;
	va_start(ap, fmt);
	(void) vsnprintf(p, 1024, fmt, ap);
	va_end(ap);

	unsigned long len = list->length();
	list->length(len+1);
	(*list)[len] = CORBA::string_dup(p);
}

void addSubStringSeq(Arts::StringSeq *target, Arts::StringSeq *source)
{
	unsigned long i;

	sqprintf(target,"{");
	for(i=0;i<source->length();i++)
	{
		unsigned long len = target->length();
		target->length(len+1);
		string srcstring = string("  ") + string((*source)[i]);
		(*target)[len] = CORBA::string_dup(srcstring.c_str());
	}
	sqprintf(target,"}");
}

int parse_line(char *in, char *& cmd, char *& param)
{
	int i,cmdlen=0,paramlen=0;
	static char static_cmd[1000], static_param[1000];

	cmd = static_cmd;
	param = static_param;
	i = 0;

	while(in[i] == ' ' || in[i] == '\t') i++;

	if(in[i] == 0) return(0);

	while(in[i] != '=' && in[i] != 0) cmd[cmdlen++] = in[i++];
	if(in[i] != 0) i++;
	while(in[i] != 0) param[paramlen++] = in[i++];

	cmd[cmdlen] = 0;
	param[paramlen] = 0;

	if(paramlen) return(2);
	if(cmdlen) return(1);
	return(0);
}

Arts::StringSeq *getSubStringSeq(Arts::StringSeq *seq,unsigned long& i)
{
	Arts::StringSeq *result = new Arts::StringSeq;
	char *cmd = "",*param;


	while(strcmp(cmd,"{") && i<seq->length())
		parse_line((*seq)[i++],cmd,param);

	int brackets = 1;
	
	while(i<seq->length())
	{
		parse_line((*seq)[i],cmd,param);
		if(strcmp(cmd,"{") == 0) brackets++;
		if(strcmp(cmd,"}") == 0) brackets--;
		if(brackets == 0) return(result);

		unsigned long len = result->length();
		result->length(len+1);
		(*result)[len] = CORBA::string_dup((*seq)[i]);
		i++;
	}
	return(result);
}

Arts::StringSeq *PortDesc_impl::saveToList()
{
	Arts::StringSeq *list = new Arts::StringSeq;

	sqprintf(list,"id=%ld",_ID);
	if(_hasValue)
	{
		if(_Type.DataType == Arts::string_data)
		{
			sqprintf(list,"string_data=%s",_StringValue.c_str());
		}
		if(_Type.DataType == Arts::audio_data)
		{
			sqprintf(list,"audio_data=%2.5f",_FloatValue);
		}
	}

	if(_isConnected)
	{
		unsigned long i;
		for(i=0;i<_Connections->length();i++)
		{
			Arts::PortDesc_ptr port = (*_Connections)[i];
			
			sqprintf(list,"connect_to=%ld",port->ID());
		}
	}
	return list;
}

Arts::StringSeq *ModuleDesc_impl::saveToList()
{
	Arts::StringSeq *list = new Arts::StringSeq;
	Arts::PortDesc *pd;

	unsigned long i;

	sqprintf(list,"id=%ld",_ID);
	sqprintf(list,"x=%ld",_X);
	sqprintf(list,"y=%ld",_Y);
	for(i=0;i<_Ports->length();i++)
	{
		pd = (*_Ports)[i];
		sqprintf(list,"port=%s",pd->Name());

		Arts::StringSeq_var portlist = pd->saveToList();
		addSubStringSeq(list,portlist);
	}
	return list;
}

Arts::StringSeq *StructureDesc_impl::saveToList()
{
	Arts::StringSeq *list = new Arts::StringSeq;
	Arts::ModuleDesc *md;

	unsigned long i;

	for(i=0;i<_Modules->length();i++)
	{
		md = (*_Modules)[i];
		sqprintf(list,"module=%s",md->Name());

		Arts::StringSeq_var modulelist = md->saveToList();
		addSubStringSeq(list,modulelist);
	}
	return list;
}

void PortDesc_impl::internalReConnect( const Arts::PortDescSeq& allports )
{
	unsigned long i;
	for(i=0;i<allports.length();i++)
	{
		long oid = allports[i]->internalOldID();

		if(find(oldConnections.begin(),oldConnections.end(),oid)
													!= oldConnections.end())
		{
			connectTo(allports[i]);
		}
	}
}

CORBA::Long PortDesc_impl::internalOldID()
{
	return _OldID;
}

void PortDesc_impl::loadFromList(const Arts::StringSeq& list)
{
	unsigned long i;
	char *cmd,*param;
	for(i=0;i<list.length();i++)
	{
		if(parse_line(list[i],cmd,param))	// otherwise: empty or comment
		{
			if(strcmp(cmd,"audio_data") == 0) {
				FloatValue(atof(param));
			} else if(strcmp(cmd,"string_data") == 0) {
				StringValue(param);
			} else if(strcmp(cmd,"id") == 0) {
				_OldID = atol(param);
			} else if(strcmp(cmd,"connect_to") == 0) {
				oldConnections.push_back(atol(param));
			}
		}
	}
}

void ModuleDesc_impl::loadFromList(const Arts::StringSeq& list)
{
	printf("mlist-----------\n");
	unsigned long i;
	char *cmd,*param;
	for(i=0;i<list.length();i++)
	{
		if(parse_line(list[i],cmd,param))	// otherwise: empty or comment
		{
			printf("MD: load-> cmd was %s\n",cmd);
			if(strcmp(cmd,"port") == 0)
			{
				Arts::PortDesc_ptr pd = 0,pdi;

				unsigned long pi;
				for(pi=0;pi<_Ports->length();pi++)
				{
					pdi = (*_Ports)[pi];
					printf("pdi = %s, param = %s\n",pdi->Name(),param);
					if(strcmp(pdi->Name(),param) == 0) pd = pdi;
				}
				assert(pd);

				Arts::StringSeq_var plist = getSubStringSeq(&list,i);

				pd->loadFromList(*plist);
			} else if(strcmp(cmd,"x") == 0) {
				_X = atol(param);
				printf("X set to %ld (param was %s)\n",_X,param);
			} else if(strcmp(cmd,"y") == 0) {
				_Y = atol(param);
				printf("Y set to %ld (param was %s)\n",_Y,param);
			}
		}
	}
	printf("-----------mlist\n");
}

void StructureDesc_impl::loadFromList(const Arts::StringSeq& list)
{
	char *cmd,*param;
	unsigned long pi,i,portcount = 0;
	Arts::PortDescSeq allports;

	printf("loadFromList; listlen = %d\n",list.length());
	for(i=0;i<list.length();i++)
	{
		if(parse_line(list[i],cmd,param))	// otherwise: empty or comment
		{
			printf("SD: load-> cmd was %s\n",cmd);
			if(strcmp(cmd,"module") == 0)
			{
				Arts::ModuleDesc_var md = createModuleDesc(param);
				Arts::PortDescSeq_var pd = md->Ports();
				Arts::StringSeq_var mlist = getSubStringSeq(&list,i);

				md->loadFromList(*mlist);

				for(pi = 0; pi < pd->length();pi++) 
				{
					allports.length(portcount+1);
					allports[portcount++]=Arts::PortDesc::_duplicate((*pd)[pi]);
				}
			}
		}
	}

	for(i=0;i<allports.length();i++)
		allports[i]->internalReConnect(allports);
}
