	/*

	Copyright (C) 1998 Stefan Westerfeld
                       stefan@space.twc.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */

#include <sys/types.h>
#include <sys/time.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include "synth_impl.h"
#include "structures.h"
#include "ioevent.h"
 
extern "C" {
#include "sound.h"
#include "utils.h"
}

#define Die(a) { fprintf(stderr,"ERROR: %s\n",a); exit(1); }

//****************** Synthesizer *******************

Synthesizer_impl::Synthesizer_impl(CORBA::ORB_var *my_orb)
{
	describe("Synthesizer");

	this->my_orb = my_orb;
	printf("... Synthesizer_impl created\n");
	SynthModules = NULL;
	MCount = 0;
	MS_SynthModule = (ModuleServer<SynthModule> *)SynthModule::get_MS();
	nextID = 1;
}

Synthesizer_impl::~Synthesizer_impl()
{
	printf("... Synthesizer_impl destroyed\n");
}

void Synthesizer_impl::Run(long cycles)
{
	long i,n;

	//fprintf(stderr,"Using %ld modules for synthesis\n",MCount);
	for (n=0;n<cycles;n++)
		for(i=0;i<MCount;i++) SynthModules[i]->Calculate();

	// Calculate cpu usage.
	status_cycles += cycles;
	if(status_cycles > 20000)
	{
	    float cpu_time = (clock()-status_clocks)/(float)CLOCKS_PER_SEC;
		float real_time = benchmark(BENCH_END);

		if(cpu_time > 0 && real_time > 0) // there may be wraparounds
		{
			status_cpu_usage = cpu_time / real_time;
			if(status_cpu_usage > 0.9)	  // more than 90%  -> not good!
			{
				status_stalled ++;
			}
			else status_stalled = 0;

			if(status_stalled > 5)
			{
				// ok, cancel synthesis due to cpu overload!
				status_halted = true;
				Stop();
			}
		}

		// prepare for next checkpoint
		status_clocks = clock(); 
		status_cycles -= 20000;
		benchmark(BENCH_BEGIN);
	}
}

Arts::Status Synthesizer_impl::getStatus()
{
	Arts::Status status;

	status.cpu_usage = status_cpu_usage;
	status.halted = status_halted;
	return(status);
}

void Synthesizer_impl::Start(long cycles)
{
	int i;

	status_halted = false;
	status_stalled = 0;
	status_cpu_usage = 0;
	status_clocks = clock();
	status_cycles = 0;	
	benchmark(BENCH_BEGIN);

	printf("beginning execution...\n");
	for(i=0;i<MCount;i++)
	{
		int infd,outfd;
		ArtsIOEvent *ev;

		SynthModules[i]->getfds(infd,outfd);

		if(infd != -1)
		{
			ev = new ArtsIOEvent(this,cycles);
			(*my_orb)->dispatcher()->rd_event(ev,infd);
			IOEvents.push_back(ev);
		}
		if(outfd != -1)
		{
			ev = new ArtsIOEvent(this,cycles);
			(*my_orb)->dispatcher()->wr_event(ev,outfd);
			IOEvents.push_back(ev);
		}
	}
	printf("ready for the sound? ;)\n");
}

void Synthesizer_impl::Stop()
{
	list<ArtsIOEvent *>::iterator i;
	for (i=IOEvents.begin(); i != IOEvents.end(); i++)
	{
		(*my_orb)->dispatcher()->remove((*i),CORBA::Dispatcher::All);
		delete (*i);
		printf("removed io event\n");
	}

	IOEvents.erase(IOEvents.begin(), IOEvents.end());
}

CORBA::Long Synthesizer_impl::getModuleCount()
{
	return(MS_SynthModule->getModuleCount());
}

void collectPorts(Arts::StringSeq &ports,SynthModule *m,SynthModule::ArgType argtype)
{
	int i = 0;
	string argname;

	do
	{
		argname = m->getArgument(i,argtype);
		if(argname != "")
		{
			i++;
			ports.length(i);
			//printf("%s/inarg[%d] = %s\n",name,i-1,argname.data());
			ports[i-1] = CORBA::string_dup(argname.c_str());
		}
	} while(argname != "");
}

CORBA::Boolean Synthesizer_impl::getModuleInfo( CORBA::Long nr, Arts::ModuleInfo*& info )
{
	char *name=MS_SynthModule->getModuleName(nr);
	if(name)
	{
		SynthModule *m = (SynthModule *)MS_SynthModule->getModule(name);

		if(m)
		{
			info = new Arts::ModuleInfo;
			info->name = CORBA::string_dup(name);

			// collect inports & outports
			collectPorts(info->inports,m,SynthModule::arg_in);
			collectPorts(info->outports,m,SynthModule::arg_out);
			return(true);
		}
	}
	return(false);
}

Arts::StructureDesc *Synthesizer_impl::createStructureDesc()
{
	printf("creating new structure desc\n");

	Arts::StructureDesc_ptr sd = new StructureDesc_impl;
	return Arts::StructureDesc::_duplicate(sd);
}

CORBA::Long Synthesizer_impl::createStructure(Arts::StructureDesc
														*structuredesc)
{
	Arts::ModuleDescSeq_var modules = structuredesc->Modules();
	Arts::ModuleDesc *md,*md2;
	Arts::PortDesc *pd;
	SynthModule *NewModules[modules->length()];

	printf("Execute I <Creation>\n");
	// first step: create modules, allocate output buffers
	unsigned long m,m2,p,p2;

	for(m=0;m<modules->length();m++)
	{
		md = (*modules)[m];
		Arts::PortDescSeq_var ports = md->Ports();
		printf("creating module %d %s\n",m,md->Name());

		SynthModule *NewModule =
			(SynthModule *)MS_SynthModule->getModule(md->Name());
		assert(NewModule);
		NewModule->Synthesizer = this;

		NewModules[m] = NewModule;
		NewModule->out_count = 0; NewModule->out = 0; NewModule->outID = 0;
		NewModule->in_count = 0; NewModule->in = 0; NewModule->inID = 0;
		NewModule->ID = md->ID();
		NewModule->executionID = nextID;

		for(p=0;p<ports->length();p++)
		{
			pd = (*ports)[p];

			if(pd->Type().Direction == Arts::output)
			{
				NewModule->out = (float **)realloc(NewModule->out,
									sizeof(float*)*(NewModule->out_count+1));
				NewModule->out[NewModule->out_count] =
									(float *)calloc(1,sizeof(float));
				NewModule->outID = (long *)realloc(NewModule->outID,
									sizeof(long)*(NewModule->out_count+1));
				NewModule->outID[NewModule->out_count] = pd->ID();
				printf("created port id %d\n",pd->ID());
				NewModule->out_count++;
			}
		}
	}

	// second step: finialize modules ->  connect or allocate input buffers

	printf("Execute II <Connection>\n");
	for(m=0;m<modules->length();m++)
	{
		string textargs = "";
		SynthModule *NewModule = NewModules[m];
		md = (*modules)[m];
		Arts::PortDescSeq_var ports = md->Ports();

		for(p=0;p<ports->length();p++)
		{
			Arts::PortType pt;

			pd = (*ports)[p];
			pt = pd->Type();

			if(pt.Direction == Arts::input)
			{
				if(pt.ConnType == Arts::stream)
				{
					assert(pt.DataType == Arts::audio_data);
					
					NewModule->in = (float **)realloc(NewModule->in,
									sizeof(float*)*(NewModule->in_count+1));
					NewModule->inID = (long *)realloc(NewModule->inID,
									sizeof(long)*(NewModule->in_count+1));
					NewModule->inID[NewModule->in_count] = pd->ID();

					if(pd->isConnected())
					{
						Arts::PortDescSeq_var conns = pd->Connections();
						assert(conns->length() == 1);
						Arts::PortDesc *pd2 = (*conns)[0];

						long wantedID = pd2->ID();
						long foundnr = 0;
						SynthModule *other,*found=0;
						// find other module
				
						for(m2=0;m2<modules->length();m2++)
						{
							other = NewModules[m2];
							for(p2=0;p2<other->out_count;p2++)
							{
								if(other->outID[p2] == wantedID)
								{
									found = other;
									foundnr=p2;
								}
							}
						}
						if(!found)
						{
							printf("fatal: didn't find port %ld, which should provide input for module %s\n", wantedID,md->Name());
						}
						assert(found);
						NewModule->in[NewModule->in_count]=found->out[foundnr];
					}
					else
					{
						float *port = NewModule->in[NewModule->in_count] =
									(float *)calloc(1,sizeof(float));

						if(pd->hasValue())
							*port = pd->FloatValue();
						else
							*port = 0.0;
					}
					NewModule->in_count++;
				}
				if(pt.ConnType == Arts::property)
				{
					assert(pt.DataType == Arts::string_data);
					assert(textargs == "");

					// only one textarg may be passed in that primitive form
					textargs = pd->StringValue();
					printf("got stringproperty; %s\n",textargs.c_str());
				}
				assert(pt.ConnType != Arts::event);   // no event channels yet
			}
		}
		NewModule->Initialize(textargs);
		printf("Initialize: textargs = %s\n",textargs.c_str());
		SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*(MCount+1));
		SynthModules[MCount++] = NewModule;
		for(p=0;p<NewModule->out_count;p++) assert(NewModule->out[p]);
	}
	printf("Execute ok\n");
	return nextID++;
}

CORBA::Boolean Synthesizer_impl::freeStructure(CORBA::Long id)
{
	long NCount = 0,freed = 0,i;
	printf("freeing structure id = %ld\n",id);
	if(SynthModules)
	{
		// kill those that match the id
		for(i=0;i<MCount;i++)
		{
			if(SynthModules[i]->executionID == id)
			{
				SynthModules[i]->DeInitialize();
				delete SynthModules[i];
				freed++;
			}
			else SynthModules[NCount++] = SynthModules[i]; // "compress" modules
		}
		MCount = NCount;
		SynthModules = (SynthModule **)realloc(SynthModules,
									sizeof(SynthModule *)*MCount);
		if(MCount == 0) SynthModules = 0;

		if(freed)
		{
			printf("freed %d modules\n",freed);
			return(true);
		}
	}
	printf("no modules freed\n");
	return false;
}
