// Copyright 1998 by Kevin Atkinson under the terms of the LGPL

#include "approx.hh"
#include "approx-t.hh"
#include <vector>

//
// approximate_match return a distance or number between 0 and 1 which
// depends on how well two short strings (say less than 32 chars)
// match.
//

// It is TOTALLY independent of any particular language

//
// See the header file approx-t.hh for a description of
// approx_match_fast
//

double approx_match_score_fast(const char *s1, const char *s2) {
  return ap_fast<DoNothing,DoNothing>(s1,s2,DoNothing(),DoNothing()).score();
}

int approx_match_dist_fast(const char *s1, const char *s2) {
  return ap_fast<DoNothing,DoNothing>(s1,s2,DoNothing(),DoNothing()).distance();
}

void approx_match_fast(const char *s1, const char *s2, 
		       double &score, int &dist) {
  ap_fast<DoNothing,DoNothing> s(s1,s2,DoNothing(),DoNothing());
  score = s.score();
  dist = s.distance();
}

//
// weighed_score - like simple score except that it gives higher
// scores to charters matching near the beginning of the word.
//
// for example aaa vs aab would have a higher score than aaa vs aba
// and that would have a higher score than aaa bv baa.  The scoring
// method is used by approximate_match2
//

static double weighed_score (const char *s1, const char *s2) {
  double score = 0;
  double p = 1;
  for (; *s1 && *s2; ++s1, ++s2) {
    if (*s1 == *s2)
      score += p;
    p /= 1.3333;
  }
  return score;
}

//
// Highest Score is a simple data structure that holds the higest
// matching pair of strings bases on the given score
//

class HighestScore {
private:
  double score_;
  int    s1_;
  int    s2_;
public:
  HighestScore() : score_(0) {}
  HighestScore(double score, int s1, int s2) 
    : score_(score), s1_(s1), s2_(s2) {}
  void clear() {score_ = 0;}
  void add_first(double score, int s1, const int s2) {
    score_ = score; s1_ = s1; s2_ = s2;
  }    
  void add(double score, int s1, const int s2) {
    if (score > score_) {score_ = score; s1_ = s1; s2_ = s2;}
  }
  int get_first() {return s1_;}
  int get_second() {return s2_;}
};

//
// approx_match_simple tries to get the two strings to match as best
// as possable by deleting characters.  For example, "redo" vs "doing"
// would become "do" vs "do".
//
// It determines when it is appropriate to delete a charter based on
// if it would lead to a higher weighted score.  It *never* goes back
// to see if it can match better by deleting something else.
//
// The distance then is:
//   max size - final matched charters
// The score then is:
//   final matched charters / max size
//

class ap_normal_simple {
  int max_size;
  int final_score;
public:
  ap_normal_simple(const char *s1, const char *s2);
  int distance() {return max_size - final_score;}
  double score() {return static_cast<double>(final_score)/max_size;}
};

ap_normal_simple::ap_normal_simple (const char *s1, const char *s2) {

  final_score=0;
  int size1=0;
  int size2=0;
  int i;
  HighestScore higest;

  for (;;) {
    while (*s1 && *s2 && *s1 == *s2) {
      ++s1; ++s2;
      ++size1; ++size2;
      ++final_score;
    }
    
    if (!*s1) {
      while (*s2) ++s2, ++size2;
      break;
    }
    if (!*s2) {
      while (*s1) ++s1, ++size1;
      break;
    }

    // try to match better by deleting
    
    higest.add_first(weighed_score(s1, s2), 0, 0);
    
    for (i = 1; s1[i]; ++i)
      higest.add(weighed_score(s1+i, s2), i, 0);
    
    for (i = 1; s2[i]; ++i)
      higest.add(weighed_score(s1, s2+i), 0, i);

    s1 += higest.get_first();
    size1 += higest.get_first();
    s2 += higest.get_second();
    size2 += higest.get_second();
    
    if (*s1 == *s2) ++final_score;
    
    ++s1; ++size1;
    ++s2; ++size2;
  }

  max_size = size1 > size2 ? size1 : size2;
}

double approx_match_score_simple(const char *s1, const char *s2) {
  return ap_normal_simple(s1,s2).score();
}

int approx_match_dist_simple(const char *s1, const char *s2) {
  return ap_normal_simple(s1,s2).distance();
}

void approx_match_simple(const char *s1, const char *s2,
		  double &score, int &dist) {
  ap_normal_simple s(s1,s2);
  score = s.score();
  dist = s.distance();
}

//
// approx_match is like approx_match_simple except that it then tries
// to analysize the deletions to classify the replasments as
// delitions, insertions (or deletions of the other element), swaps
// (exchanging two adjacent letters), or substitutions (replacing one
// letter with another) and returns the sum of the four with optional
// weights attached to each for the distance.
//
// For example redos vs odz would have:
//   2 deletions from the first (redos -> dos)
//   1 swap (dos -> ods)
//   1 substitution (ods -> odz)
//
// If the total is larger than the size of the maximum size than the
// maximun size is returned because the same thing can be accived with
// nothing but substations and deletions from ONE of the words
// 
// For example: abcdefghij vs ijklmnopq would normally lead to:
//   8 deletions from the first
//   7 deletions from the second
// Which would normally have a distance of 15 (with equal weights).
// However the same results could of been acceved by
//   9 substitutions
//   1 deletions form the first
// Which would lead to a score of 10.  However by algorithm will
// not no this thus I will just return the maximum size.
//
// The score is calculated as follows:
//   (max_size - dist)/max_size
//

class ap_normal {
  int max_size;
  int del1;
  int del2;
  int swaps;
  int subs;
public:
  static double min(double a, double b) {return a < b ? a : b;}
  ap_normal(const char *s1, const char *s2);
  double distance() {return min(del1+del2+swaps+subs,max_size);}
  double score() {return (max_size - distance())/max_size;}
  double distance(const dist_weights& w) 
    {return min(w.del1*del1 + w.del2*del2 + w.swaps*swaps + w.subs*subs,
		max_size);}
  double score(const dist_weights& w) 
    {return (max_size - distance(w))/max_size;}
  void distance(dist_components& c) {c.set(del1,del2,swaps,subs);}
};

ap_normal::ap_normal (const char *b1, const char *b2) 
  : del1(0), del2(0), swaps(0), subs(0)
{ 
  const char *s1 = b1;
  const char *s2 = b2;
  
  int i;
  HighestScore higest;

  vector<bool> track1;
  vector<bool> track2;

  for (;;) {
    while (*s1 && *s2 && *s1 == *s2) {
      track1.push_back(1); track2.push_back(1);
      ++s1; ++s2;
    }
    
    if (!*s1) {
      while (*s2) track2.push_back(0), ++s2;
      break;
    }
    if (!*s2) {
      while (*s1) track1.push_back(0), ++s1;
      break;
    }

    // try to match better by deleting
    
    higest.add_first(weighed_score(s1, s2), 0, 0);
    
    for (i = 1; s1[i]; ++i)
      higest.add(weighed_score(s1+i, s2), i, 0);
    
    for (i = 1; s2[i]; ++i)
      higest.add(weighed_score(s1, s2+i), 0, i);

    for (i = 0; i != higest.get_first(); ++i)
      track1.push_back(0);
    s1 += higest.get_first();

    for (i = 0; i != higest.get_second(); ++i)
      track2.push_back(0);
    s2 += higest.get_second();

    if (*s1 == *s2) {
      track1.push_back(1);
      track2.push_back(1);
    } else {
      track1.push_back(0);
      track2.push_back(0);
    }
    ++s1;
    ++s2;
  }
  
  vector<bool>::size_type i1 = 0;
  vector<bool>::size_type i2 = 0;
  vector<bool>::size_type size1 = track1.size();
  vector<bool>::size_type size2 = track2.size();
  
  for(;;) {
    if (i1 == size1) {
      del2 += size2 - i2;
      break;
    }
    if (i2 == size2) {
      del1 += size1 - i1;
      break;
    }
    if (i1+1 != size1 && i2+1 != size2 &&
	track1[i1] != track1[i1+1] && track2[i2] != track2[i2+1] && 
	track1[i1] != track2[i2] &&
	b1[i1] == b2[i2+1] && b1[i1+1] == b2[i2]) {
      ++swaps;
      ++i1; ++i2;
      ++i1; ++i2;
    } else if (track1[i1] == 0 && track2[i2] == 0) {
      ++subs;
      ++i1; ++i2;
    } else if (track1[i1] == track2[i2]) {
      ++i1; ++i2;
    } else if (track1[i1] == 0) {
      ++del1;
      ++i1;
    } else if (track2[i2] == 0) {
      ++del2;
      ++i2;
    }
  }

  max_size = size1 > size2 ? size1 : size2;
}

double approx_match_score(const char *s1, const char *s2) {
  return ap_normal(s1,s2).score();
}

double approx_match_dist(const char *s1, const char *s2) {
  return ap_normal(s1,s2).distance();
}

void approx_match(const char *s1, const char *s2,
		  double &score, double &dist) {
  ap_normal s(s1,s2);
  score = s.score();
  dist = s.distance();
}

double approx_match_score(const char *s1, const char *s2, 
			  const dist_weights& w) {
  return ap_normal(s1,s2).score(w);
}

double approx_match_dist(const char *s1, const char *s2,
			 const dist_weights& w) {
  return ap_normal(s1,s2).distance(w);
}

void approx_match(const char *s1, const char *s2,
		  double &score, double &dist,
		  const dist_weights& w) {
  ap_normal s(s1,s2);
  score = s.score(w);
  dist = s.distance(w);
}

void approx_match(const char *s1, const char *s2,
		  dist_components& c) {
  ap_normal(s1,s2).distance(c);
}

