#include <iterator>

template <class Return> 
class VirtualForwardIterator {
private:
  virtual bool is_equal (const VirtualForwardIterator&) = 0;
public:
  friend bool operator== (const VirtualForwardIterator& rhs,
			  const VirtualForwardIterator& lhs) 
    {return rhs.is_equal(lhs);}
  friend bool operator!= (const VirtualForwardIterator& rhs,
			  const VirtualForwardIterator& lhs)
    {return !rhs.is_equal(lhs);}
public:
  virtual Return operator*() = 0;
  virtual const Return operator*() const = 0;
  virtual VirtualForwardIterator & operator++() = 0;

  virtual VirtualForwardIterator * clone() const = 0; 
                            // return a copy of itself allocated with new

  virtual ~VirtualForwardIterator() {}
};

template <class Iterator, class Return>
class MakeVirtualForwardIterator : public VirtualForwardIterator<Return> {
private:
  Iterator itr;
  typedef MakeVirtualForwardIterator     Self;
  typedef VirtualForwardIterator<Return> Base;
  bool is_equal (const Base &other) {
    if (const Self* p = dynamic_cast<const Self*>(&other))
      return itr == p->itr;
    else
      return false;
  }
public:
  MakeVirtualForwardIterator() {}
  MakeVirtualForwardIterator(Iterator i) itr(i) {}
  Return operator*() {return *itr;}
  const Return operator*() const {return *itr;}
  Self& operator++() {++itr; return *this;}
  Self * clone() const {return new Self(this);}
};

template <class Value, class Return=Value&>
class VirtualForwardIteratorProxy : forward_iterator<Value, long>
{
private:
  VirtualForwardIterator *itr;
  typedef typename VirtualForwardIterator<Return> RealItr;
  typedef VirtualForwardIteratorProxy             Self;
public:
  VirtualForwardIteratorProxy() {itr = 0;}
  VirtualForwardIteratorProxy(const RealItr& i) {itr = i.clone();}
  VirtualForwardIteratorProxy(Self& other) {itr = other.itr.clone();}
  Self& operator= (const Self& other) {
    if (this != &other) {
      delete itr;
      itr = other.itr.clone();
    }
    return *this;
  }
  ~VirtualForwardIteratorProxy() {delete itr;}

  Return operator*() {return itr->operator*();}
  const Return operator*() const {return itr->operator*();}

  Self& operator++() {itr->operator++(); return *this;}
  Self operator++(int) {Self temp=*this; itr->operator++(); return temp;}

  friend bool operator==(Self &lhs, Self &rhs) {
    return *lhs.itr == *rhs.itr;
  }
  friend bool operator!=(Self &lhs, Self &rhs) {
    return *lhs.itr != *rhs.itr;
  }
};

template <class Value, class Return, class Container, class MultiItr>
class MultiForwardIterator {
private:
  typedef MultiForwardIterator                       Self;
  typedef VirtualForwardIteratorProxy<Value, Return> RealItr;

  MultiItr   cur;
  MultiItr   end;
  RealItr    itr;
public:
  MultiForwardIterator() {}
  MultiForwardIterator(Container *c, MultiItr h, RealItr i)
    : container(c), cur(h), itr(i) {}
  Return operator*() {return *itr;}
  const Return operator*() const {return *itr;}

  Self& operator++();
  Self operator++(int) {Self temp=*this; itr->operator++(); return temp;}

  friend bool operator==(Self &lhs, Self &rhs) {
    return lhs.cur == rhs.cur && lhs.itr == rhs.itr;
  }
  friend bool operator!=(Self &lhs, Self &rhs) {
    return !(lhs == rhs);
  }
};
  
template <class Value, class Return, class Container, class MultiItr>
MultiForwardIterator<Value, Return, Container, MultiItr>&
MultiForwardIterator<Value, Return, Container, MultiItr>::operator++() {
  ++itr;
  if (itr == cur->end) {
    ++cur;
    if (cur != end)
      itr = cur->begin;
  }
  return *this;
}


template <class Value, class Return = Value&>
class MultiForwardContainer {
public:
  typedef unsigned long size_type;
  typedef Value         value_type;

  static const size_type npos = -1;

private:
  typedef VirtualForwardIteratorProxy<Value, Return> RealItr;
  struct BeginEndSize {
    Itr begin;
    Itr end;
    size_type size;
  };
  vector<BeginEndSize> itrs;
  mutable size_type    size;
public:
  
  // Note: begin() and end() are undefined on an empty Container

  iterator begin()
    {return iterator(itrs.begin(), itrs.end(), itrs.begin()->begin);}
  iterator end() 
    {return iterator(itrs.end(), itrs.end(), itrs.begin()->begin);}
  const_iterator begin() const;
    {return iterator(itrs.begin(), itrs.end(), itrs.begin()->begin);}
  const_iterator end() const 
    {return iterator(itrs.end(), itrs.end(), itrs.begin()->begin);}

  size_type size() const;
  bool empty() const
    {return !itrs.size();}
};


template <class Value, class Return = Value&>
unsigned long MultiForwardContainer::size() const {
  if (size == npos) {
    size = 0;
    vector<BeginEndSize> i = itrs.begin();
    vector<BeginEndSize> e = itrs.end();
    for (; i != e; ++i)
      size += i->size;
  }
  return size;
}
