// Aspell test program
// Copyright 1998 by Kevin Atkinson under the terms of the GPL

#include <string>
#include <rope>
#include <algo.h>
#include <iterator.h>
#include <deque>
#include <hash_map>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <strstream.h>
#include <time.h>
#include "aspell.hh"
#include "as_check.hh"
#include "as_convert.hh"
#include "gen_lang_loader.hh"
#include "const_string.hh"
#include "hash_string.hh"

#include "as_config.h"

typedef deque<const_string> Args;
typedef hash_map<const_string, const_string, hash_string<const_string>, equal_to<const_string> > Options;

void pipe(Args&, Options&);
void print_ver(Args&, Options&);
void check(Args&, Options&);
void master(Args&, Options&);
void personal(Args&, Options&);
void test(Args&, Options&);
void lang(Args&, Options&);
void suggest(Args&, Options&);

void print_help() {
  cout << endl
       << "Aspell " <<VERSION<< " alpha. An Ispell look alike." << endl
       << "Usage: " << endl
       << endl
       << "  aspell <command> [file] [options]" << endl
       << endl
       << "  <command> is one of: " << endl
       << "    check - to check a file" << endl 
       << "    pipe - ispell '-a' compatible mode." << endl
       << "    master - to create a master word list from stdin." << endl
       << "      The file is the base name of the word list to create" << endl
       << "    personal - to create a personal word list from stdin." << endl
       << "      The file is the base name of the word list to create" << endl
       << "    test - to go into a special test mode. " << endl
       << endl
       << "  [options] is any or all of the following" << endl
       << "    -master <base> - the base name of the master dictionary to use" << endl
       << "    -personal <base> - the base name of the personal dictionary to use" << endl
       << "    -replacement <base> -  the base name of the replacement dict. to use" << endl
       << "    -ignore-repl - do not store replacements in the repl. dict." << endl
       << "    -reverse - reverse the order of the suggestions list." << endl 
       << endl
       << "  In addition Aspell with try to make seance out of ispell's command" << endl
       << "  line options so that it can function as a drop in replacement for ispell." << endl
       << endl;
  exit(-1);
}

int main (int argc, char *argv[]) {
  Options options;
  Args arg;
  copy(argv, argv+argc, back_inserter(arg) );
  arg.pop_front();

  for(unsigned int i = 0; i != arg.size(); ++i) {
    if (arg[i][0] == '-') {
      if (arg[i] == "-d") arg[i] = "-master";
      if (arg[i] == "-p") arg[i] = "-personal";
      if ((arg[i] == "-master" || arg[i] == "-personal" || arg[i] == "-log"
	   || arg[i] == "-lang" || arg[i] == "-w" || arg[i] == "-W" || arg[i] == "-T") 
	  && i+1 != arg.size()) {
	options[arg[i]] = arg[i+1];
	arg.erase(arg.begin()+i, arg.begin()+i+2);
	--i;
      } else {
	options[arg[i]] = "";
	arg.erase(arg.begin()+i, arg.begin()+i+1);
	--i;
      }
    }
  }

  if (options.count("-log")) {
    ofstream LOG(options["-log"], ios::app);
    LOG << "Command: ";
    for(int i = 0; i != argc; ++i)
      LOG << argv[i] << "~";
    LOG << endl;
  }

  if (!options.count("-master"))
    options["-master"] = "english";
  if (!options.count("-personal"))
    options["-personal"] = "";
  if (!options.count("-replacement"))
    options["-replacement"] = "";
  if (!options.count("-lang"))
    options["-lang"] = "english";

  const_string command;
  if (arg.size()) {
    command = arg[0];
    arg.pop_front();
  } else {
    command = "";
  }

  if (command == "pipe" || options.count("-a"))
    pipe(arg, options);
  else if (command == "version" || options.count("-v"))
    print_ver(arg, options);
  else if (command == "check")
    check(arg, options);
  else if (command == "master")
    master(arg, options);
  else if (command == "personal")
    personal(arg, options);
  else if (command == "test")
    test(arg, options);
  else if (command == "lang")
    lang(arg, options);
  else if (command.size()) {
    arg.push_front(command);
    check(arg, options);
  } else
    print_help();
  return 0;
}

struct PrintStatus {
  bool print_star;
  PrintStatus() : print_star (true) {}
  void operator() () const {if (print_star) cout << "*" << endl;}
};

void print_ver (Args& arg, Options &options) {
  cout << "@(#) International Ispell Version 3.1.20 (but really Aspell " <<VERSION<< " alpha)" << endl;
}

string trim_space(const string &s) {
  string::size_type beg = s.find_first_not_of(" \t\n");
  string::size_type end = s.find_last_not_of(" \t\n");
  if (beg == string::npos)
    return string();
  else
    return s.substr(beg,end-beg+1);
}

void get_word_pair(string &w1, string &w2) {
  getline(cin, w1, ',');
  w1 = trim_space(w1);
  getline(cin, w2);
  w2 = trim_space(w2);
}

void ignore_rest() {
  cin.ignore(32767,'\n');
}

template<class Con>
void print_container(const Con& con) {
  typename Con::const_iterator i   = con.begin();
  typename Con::const_iterator end = con.end();
  cout << con.size() << ": ";
  if (i != end) {
    for (;;) {
      cout << *i;
      ++i;
      if (i == end) break;
      cout << ", ";
    }
  }
  cout << endl;
}

void pipe (Args& arg, Options &options) {
  ofstream     LOG;
  bool         keep_log = false;
  bool         terse_mode = true;
  if (options.count("-log")) {
    keep_log = true;
    LOG.open(options["-log"], ios::app);
  }
  aspell sc(options["-master"].c_str(), options["-personal"].c_str(),
	    options["-replacement"].c_str(), options.count("-ignore-repl"));
  if (sc.problem()) {
    cout << sc.problem().message() << endl;
    exit (-1);
  }
  CheckResults<const_string::iterator> res;
  aspell::Suggestions suggestions;
  aspell::WordList    words;
  unsigned int s;
  const_string::iterator i,j;
  const_string::iterator end;
  string line;
  string word;
  string word2;
  int    ignore;
  PrintStatus print_status;

  print_ver(arg, options);
  
  // NOTE: some how fix the reads from cin so it will break out of the
  //   loop at any time upon failer
  char c;
  while (c = cin.get(), cin) {
    ignore = 0;
    switch (c) {
    case '*':
      cin >> word;
      sc.add_personal(word);
      ignore_rest();
      break;
    case '&':
      cin >> word;
      sc.add_personal(sc.to_lower(word));
      ignore_rest();
      break;
    case '@':
      cin >> word;
      sc.add_session(word);
      ignore_rest();
      break;
    case '#':
      sc.save_dicts();
      ignore_rest();
      break;
    case '+':
    case '-':
      ignore_rest();
      break;
    case '!':
      terse_mode = true;
      print_status.print_star = false;
      ignore_rest();
      break;
    case '%':
      terse_mode = false;
      print_status.print_star = true;
      ignore_rest();
      break;
    case '^':
      ignore = 1;
    case '$':
      if (cin.get() == '$') {
	switch(cin.get()) {
	case 's':
	  get_word_pair(word,word2);
	  cout << sc.score(word.c_str(),word2.c_str()) << endl;
	  break;
	case 'S':
	  switch(cin.get()) {
	  case 'W':
	  case 'w':
	    cin >> word;
	    cout << sc.to_soundslike(word) << endl;
	    ignore_rest();
	    break;
	  case 'L':
	  case 'l':
	    cin >> word;
	    print_container(sc.soundslike(word));
	    ignore_rest();
	    break;
	  default:
	    ignore_rest();
	  }
	  break;
	case 'P':
	  switch(cin.get()) {
	  case 'W':
	  case 'w':
	    cin >> word;
	    cout << sc.to_phoneme(word) << endl;
	    ignore_rest();
	    break;
	  default:
	    ignore_rest();
	  }
	  break;
	case 'r':
	  switch(cin.get()) {
	  case 'a':
	    get_word_pair(word,word2);
	    sc.store_replacement(word,word2);
	    break;
	  case 'i':
	    switch(cin.get()) {
	    case 'c':
	      cout << sc.ignore_replacements() << endl;
	      break;
	    case '0':
	      sc.ignore_replacements(0);
	      break;
	    case '1':
	      sc.ignore_replacements(1);
	      break;
	    }
	    ignore_rest();
	    break;
	  default:
	    ignore_rest();
	  }
	  break;
	case 'p':
	  switch (cin.get()) {
	  case 'p':
	    print_container(sc.personal_dict());
	    break;
	  case 's':
	    print_container(sc.session_dict());
	    break;
	  }
	  ignore_rest();
	  break;
	case 'l':
	  cout << sc.lang_name() << endl;
	  ignore_rest();
	  break;
	default:
	  ignore_rest();
	}
	break;
      } else {
	cin.unget();
	// continue on (no break)
      }
    default:
      cin.unget();
      getline(cin, line);
      i = line.begin() + ignore;
      end = line.end();
      while (res = check(sc, i, end, print_status)) {
	word.resize(0);
	for(j = res.begin; j != res.end; ++j) word +=*j;
	suggestions = sc.suggest(word);
	if (suggestions.size()) {
	  cout << "& " << word 
	       << " " << suggestions.size() 
	       << " " << res.begin - line.begin()
	       << ":";
	  if (!options.count("-reverse")) {
	    for (s = 0; s != suggestions.size() - 1; ++s)
	      cout << " " << suggestions[s] << ",";
	    cout << " " << suggestions[suggestions.size() - 1] << endl;
	  } else {
	    for (s = suggestions.size()-1; s != 0; --s)
	      cout << " " << suggestions[s] << ",";
	    cout << " " << suggestions[0] << endl;
	  }
	} else {
	  cout << "# " << word << " " << res.begin - line.begin() << endl;
	}
	i = res.end;
      }
      cout << endl;
    }
  }
}

void check (Args& arg, Options &options) {
  crope file;
  string file_name = arg[0].c_str();
  {
    ifstream FILE(file_name.c_str());
    char c;
    while ((c = FILE.get()) != EOF) {
      file += c;
    }
  }

  aspell sc(options["-master"].c_str(), options["-personal"].c_str(),
	    options["-replacement"].c_str(), options.count("-ignore-repl"));

  CheckResults<crope::const_iterator> res;
  crope::const_iterator               i,k;
  crope::size_type                    ii;
  crope::const_iterator  line_begin, line_end;
  string word, new_word;
  string choice;
  aspell::Suggestions suggestions;
  aspell::Suggestions::size_type suggestions_mid, suggestions_size;
  aspell::Suggestions::size_type j;

  hash_map<string,string,hash_string<string>,equal_to<string> > replace_list;
  cout.setf(ios::left);

  i = file.begin();
  while (res = check(sc, i, file.end())) {
    word.resize(0);
    for (k = res.begin; k != res.end; ++k)
      word += *k;

    //
    // check if it is in the replace list
    //

    if (replace_list.count(word)) {
      ii = res.begin - file.begin();
      file.replace(res.begin - file.begin(), res.end - res.begin, replace_list[word].c_str());
      i = file.begin() + ii;
      continue;
    }

    //
    // print the line with the misspelled word highlighted;
    //

    for (line_begin = res.begin; 
	 line_begin != file.begin() && *line_begin != '\n';
	 --line_begin);
    if (line_begin != file.begin())
      ++line_begin;
    for (line_end = res.end;
	 line_end != file.end() && *line_end != '\n';
	 ++line_end);
    for(i = line_begin; i != res.begin; ++i)
      cout << *i;
    cout << '*' << word << '*';
    i = res.end;
    for(; i != line_end; ++i) 
      cout << *i;
    cout << endl;
    
    //
    // print the suggestions and menu choices
    //

    suggestions = sc.suggest(word);
    suggestions_size = suggestions.size();
    if (suggestions_size > 9)
      suggestions_size = 9;
    suggestions_mid = suggestions_size / 2;
    if (suggestions_size % 2) suggestions_mid++; // if odd
    for (j = 0; j != suggestions_mid; ++j) {
      cout << setw(1) << j+1 << ") " << setw(37) << suggestions[j].c_str();
      if (j + suggestions_mid != suggestions_size) 
	cout << setw(1) << j+suggestions_mid+1 << ") " << suggestions[j+suggestions_mid];
      cout << endl;
    }
    cout << setw(40) << "i) Ignore"
	 << "I) Ignore all" << endl;
    cout << setw(40) << "r) Replace" 
	 << "R) Replace all" << endl;
    cout << setw(40) << "a) Add"
	 << setw(40) << "o) change Options" << endl;
    cout << "? ";

    //
    // Handle the users choice
    //

    cin >> choice;
    
    switch (choice[0]) {
    case 'i':
      i = res.end;
      break;
    case 'I':
      sc.add_session(word);
      i = res.end;
      break;
    case 'a':
      sc.add_personal(word);
      i = res.end;
      break;
    case 'R':
    case 'r':
      cout << "With: ";
      cin >> new_word;
      if (new_word[0] >= '1' && new_word[0] < suggestions_size + '1')
	new_word = suggestions[new_word[0]-'1'].c_str();
      ii = res.begin - file.begin();
      file.replace(res.begin - file.begin(), res.end - res.begin, new_word.c_str());
      sc.store_replacement(word,new_word.c_str());
      i = file.begin() + ii;
      if (choice[0] == 'R')
	replace_list[word] = new_word;
      break;
    default:
      if (choice[0] >= '1' && choice[0] < suggestions_size + '1') { 
	ii = res.begin - file.begin();
	file.replace(res.begin - file.begin(), res.end - res.begin, suggestions[choice[0]-'1'].c_str());
	sc.store_replacement(word,suggestions[choice[0]-'1'].c_str());
	i = file.begin() + ii;
      } else {
	cout << "Sorry that is an invalid choice!" << endl;
	i = res.begin;
      }
    }
  }
  
  {
    ofstream FILE((file_name+".checked").c_str());
    FILE << file;
  }
  sc.save_dicts();
}

void master (Args& arg, Options &options) {
  if (arg.size() == 0) {
    cerr << "No output file base specified.\n";
    exit(-1);
  }
  
  string base = arg[0].c_str();

  SC_Error err = create_master(base, istream_iterator<string>(cin),
			       istream_iterator<string>(),
			       static_cast<string>(options["-lang"].c_str()) );
 
  if (err) {
    cout << err.message() << endl;
    exit(-1);
  }
}

void personal (Args& arg, Options &options) {
  if (arg.size() == 0) {
    cerr << "No output file base specified.\n";
    exit(-1);
  }
  
  string base = arg[0].c_str();

  SC_Error err = create_personal(base, istream_iterator<string>(cin),
				 istream_iterator<string>(),
				 static_cast<string>(options["-lang"].c_str()));

  if (err) {
    cout << err.message() << endl;
    exit(-1);
  }
}

void test (Args& arg, Options &options) {
  clock_t start, end;
  cout << "Loading word list, please wait...\n";
  start = clock();
  aspell sc(options["-master"].c_str(), options["-personal"].c_str(),
	    options["-replacement"].c_str(), options.count("-ignore-repl"));

  end = clock();
  if (sc.problem()) {
    cout << sc.problem().message() << endl;
    exit (-1);
  }
  cout << "Word list loaded.  Load time: "  
       << double(end - start)/ CLOCKS_PER_SEC << " seconds" << endl;

  string choice;
  string word;

  aspell::Suggestions suggestions;
  aspell::WordList    list;
  for (;;) {
    cout << "? ";
    start = 0;
    if (!cin) break;
    cin >> choice;
    if (!cin) break;
    switch(choice[0]) {
    case 'h':
      cout << "h        help" << endl
	   << "ap word  add the word to the personal word list" << endl
	   << "as word  add the word to the session word list" << endl
	   << "c  word  check a word" << endl
	   << "cs       clear the session word list" << endl
	   << "cp base  change personal dictionaries" << endl
	   << "l  word  lookup word" << endl
	   << "L  word  lookup word in ultra mode" << endl
	   << "lp       list the words in the personal word list" << endl
	   << "ls       list the words in the session word list" << endl
	   << "sp       save the personal word list to disk" << endl
	   << "x        exit" << endl;
      break; 
    case 'a':
      cin >> word;
      switch(choice[1]) {
      case 'p':
	sc.add_personal(word);
	break;
      case 's':
	sc.add_session(word);
	break;
      }
      break;
    case 's':
      sc.save_dicts();
      break;
    case 'c':
      if (choice.size() == 1) {
	cin >> word;
	cout << (sc.check(word) ? "true" : "false") << endl;
      } else {
	switch(choice[1]) {
	case 's':
	  sc.clear_session();
	  break;
	case 'p':
	  cin >> word;
	  sc.personal_dict().load(word);
	  break;
	}
      }
      break;
    case 'L':
    case 'l':
      if (choice.size() == 1) {
	cin >> word;
	start = clock();
	if (choice[0] == 'l')
	  suggestions = sc.suggest(word);
	else
	  suggestions = sc.suggest_ultra(word);
	end = clock();
	if (!options.count("-reverse")) {
	  for (aspell::Suggestions::iterator i = suggestions.begin();
	       i != suggestions.end();
	       ++i) 
	    cout << *i << endl;
	} else {
	  for (aspell::Suggestions::reverse_iterator i = suggestions.rbegin();
	       i != suggestions.rend();
	       ++i) 
	    cout << *i << endl;
	}
	cout << "Time: " << double(end - start)/ CLOCKS_PER_SEC << endl;
      } else {
	switch(choice[1]) {
	case 'p':
	  for (WritableWL::iterator i = sc.personal_dict().begin();
	       i != sc.personal_dict().end();
	       ++i) 
	    cout << *i << endl;
	  break;
	case 's':
	  for (WritableWL::iterator i = sc.session_dict().begin();
	       i != sc.session_dict().end();
	       ++i) 
	    cout << *i << endl;
	}
	break;
      }
      break;
    case 'x':
      exit(0);
      break;
    default:
      cout << "Sorry invalid choice.  Type 'h' for help" << endl;
    }
  }
}

void lang(Args& arg, Options&) {
  string file = arg[0].c_str();
  string path;
  string::size_type pos = file.rfind('/');
  if (pos == string::npos) {
    path = "";
  } else {
    path = file.substr(0,pos+1);
    file = file.substr(pos+1);
  }
  cout << path << " / " << file << endl;
  SC_GenericLangLoader l(file, path);
  l.print_code_to_file();
}
