// Aspell Data Objects Implementations 
// Copyright 1998 Under the terms of the LGPL

//
// Aspell's main word list data is stored in 4 large blocks of memory
//
// * The Word Hash Table
// * The Word List
// * The Soundslike Hash Table
// * The Soundslike List
//
// 1a) The Word Hash Table
// This consists of an open address hash table which contains pointers
// to the actual words in the word list
//
// 1b) The Word List
// This consists of the actual word list and is layed out as follows:
//   <Word1><null char><Word2><null char>...
//
// 2a) The Soundslike Hash Table
// This consists of an open address hash table which contains pointers
// to a soundslike object.
//
// 2b) The Soundslike Object
// The soundslike object is layed out as follow:
//  <Word1 pointer><Word2 pointer>...<Num of Words><Soundslike><null char>
//  <const char *><const char *>...<unsigned short int><char[]><char>
// The pointer to the object points to the beginning of the Soundslike string
// The Word pointers consists of the the words which have the same 
//   soundslike pattern
//
// 2c) The Soundslike List
// This consists of Soundslike Objects back to back:
//  <Soundslike object 1><Soundslike object 2> ...
// There is no delimiter between the objects
//
//
//                          Format of the *.wrd files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Wordlist>
// <Word Hash Table>
//
// The word hash table is a vector of unsigned its which contains an offset
// of where they can be found in the word list.
//
//                          Format of the *.sl files
//
// (This part is in ascii format)
// <"master_wl"><ws><lang name><ws><# words><ws>
//     <hash size><ws><size of list block><\n>
// (The rest is in binary format>
// <Soundslike object list>
// <Soundslike Hash Table>
//
// Soundslike oject is laid out as follows:
//   <Num of Words><Word 1 offset>...<Soundslike><\0>
//   <unsigned short int><unsigned int>...<char[]><char>
// And like the .wrd file the hash table contains offsets not pointers.
//
//
// FIXME: There is a now unnessary empty list in the soundslikes lookup
//        table which needs to be removed.  It was orignally there becuase
//        the words_w_soundslike returned a refrence instead of a pointer.
//        Something needed to be returned when no matches was found so I 
//        returned an empty list.  Now that I am returning pointers this
//        empty list is no longer needed as NULL will do.

#include <string>
#include <fstream.h>
#include <strstream.h>

#include <time.h>
#include <sys/stat.h>

#include "aspell.hh"
#include "data.hh"
#include "mwordlist.hh"
#include "hash-t.hh"

#define try_file(name) \
  { path = (name); \
    FILE.open( path.c_str() ); \
    if ( !(FILE.rdstate() & ios::badbit)) \
      return path;}

string find_file(const string &file, const string &suffix_, 
		 const string &try_dir_, bool suffix_must = false) {
  ifstream FILE;
  
  string suffix;
  if (!suffix_must && 
      file.substr(file.size()-suffix_.size(),suffix_.size()) == suffix_)
    suffix = "";
  else 
    suffix = suffix_;
  string try_dir;
  if (file[0] == '/')
    try_dir = "";
  else
    try_dir = try_dir_ + '/';

  string path;
  
  if (!suffix_must) 		 
    try_file(file);
  try_file(file + suffix);
  if (!suffix_must)
    try_file(try_dir + file);
  try_file(try_dir + file + suffix);

  return path;
} 

string find_data_file(const string &file, const string &suffix) {
  const char *temp = getenv("ASPELL_DATA_DIR");
  if (temp)
    return find_file(file, suffix, temp, true);
  else
    return find_file(file, suffix, LIBDIR, true);
}

string find_per_file(const string &file, const string &suffix) {
 const char *temp = getenv("HOME");
  if (temp) 
    return find_file(file, suffix, temp);
  else 
    return find_file(file, suffix, "");
}

const SC_Error SC_ReadOnly_WordList::load(const string &file_name) {
  
  ifstream WORDS((file_name).c_str(), ios::in | ios::bin);
  if ( WORDS.rdstate() & ios::badbit )
    return SC_Error(file_name, SC_Error::nonexistent);

  string check_word;
  WORDS >> check_word;
  if (check_word != "master_wl")
    return SC_Error(file_name, SC_Error::bad_format);

  WORDS >> check_word;
  {
    string lang_dir;
    string::size_type pos = file_name.rfind('/');
    if (pos == string::npos) {
      lang_dir = "";
    } else {
      lang_dir = file_name.substr(0,pos+1);
    }

    SC_Error error = _spell_checker->set_check_lang(check_word,lang_dir);
    if (error) {
      error.file = file_name;
      return error;
    }
  }

  int count, bucket, size;
  WORDS >> count >> bucket >> size;
  WORDS.ignore();
  _master = new char[size];

  WORDS.read(_master, size);

  word_list.table().resize(bucket);
  WordList::table_type::iterator i   = word_list.table().begin();

  unsigned int pos;
  while (!WORDS.eof()) {
    WORDS.read(&pos, sizeof(unsigned int));
    *i = _master + pos;
    ++i;
  }

  word_list.set_size(count);

  return SC_Error();
}

const SC_Error SC_ReadOnly_Soundslikes::
  load(const string &file_name, const char *d)
{
  ifstream SOUNDS((file_name).c_str(), ios::in, ios::bin);
  if ( SOUNDS.rdstate() & ios::badbit )
    return SC_Error(file_name, SC_Error::nonexistent);
  
  string check_word;
  SOUNDS >> check_word;
  if (check_word != "master_ml") 
    return SC_Error(file_name, SC_Error::bad_format);

  SOUNDS >> check_word;
  {
    string lang_dir;
    string::size_type pos = file_name.rfind('/');
    if (pos == string::npos) {
      lang_dir = "";
    } else {
      lang_dir = file_name.substr(0,pos+1);
    }

    SC_Error error = _spell_checker->set_check_lang(check_word,lang_dir);
    if (error) {
      error.file = file_name;
      return error;
    }
  }

  int count, bucket, size;
  SOUNDS >> count >> bucket >> size;
  main_data = new char[size];
  char *data = main_data;
  
  SOUNDS.ignore();

  *reinterpret_cast<unsigned short int *>(data) = 0;
  *(data + sizeof(unsigned short int *)) = '\0';
  data = main_data + sizeof(const char *);
  
  unsigned short int list_count;
  unsigned int       pos;
  unsigned int       m;
  while (SOUNDS.read(reinterpret_cast<void*>(&list_count),sizeof(unsigned short int)),
	 list_count) {
    for(unsigned short int i=0; i != list_count; ++i) {
      SOUNDS.read(&pos, sizeof(unsigned int));
      *reinterpret_cast<const char **>(data) = d + pos;
      data += sizeof(const char *);
    }
    *reinterpret_cast<unsigned short int *>(data) = list_count;
    data += sizeof(unsigned short int);
    while ((*data = SOUNDS.get()) != '\0') ++data;
    ++data;
    m = (data - main_data) % sizeof(const char *);
    if (m != 0)
      data += sizeof(const char *) - m;
  }

  lookup_table.table().resize(bucket);

  LookupTable::table_type::iterator i   = lookup_table.table().begin();

  while (!SOUNDS.eof()) {
    SOUNDS.read(&pos, sizeof(unsigned int));
    *i = main_data + pos;
    ++i;
  }

  lookup_table.set_size(count);

  return SC_Error();
}

const SC_ReadOnly_Soundslikes::WordList * SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word) {
  LookupTable::iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) { 
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return NULL;                                 
  } else {                                       
    return &(*i);
  }
}

const SC_ReadOnly_Soundslikes::WordList * SC_ReadOnly_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) {
  LookupTable::iterator i = lookup_table.find(word);
  if (i == lookup_table.end() || !(*i).size()) {
    //                                  ^^ FIXME
    // There should not be a need for the size check as there should never be
    // any empty lists in the lookup_table.
    return NULL;
  } else {
    M = (*i).key();
    return &(*i);
  }
}

void SC_Writable_Soundslikes::add(const char* w, const string &s) {
  lookup_table[s.c_str()].push_back(w);
}

void SC_Writable_Soundslikes::erase(const char* w, const string &s) {
  WordList& item = lookup_table.find(s.c_str())->second;
  WordList::iterator i = item.begin();
  WordList::iterator end = item.end();
  while(i != end && *i != w) ++i;
  if (i == end) return;
  item.erase(i);
  if (!item.size()) {
    lookup_table.erase(s.c_str());
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word) {
  LookupTable::iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    return &(i->second);
  }
}

const SC_Writable_Soundslikes::WordList * SC_Writable_Soundslikes
::words_w_soundslike(const char *word, SC_Sound &M) {
  LookupTable::iterator i = lookup_table.find(word);
  if (i == lookup_table.end()) {
    return NULL;
  } else {
    M = i->first.c_str();
    return &(i->second);
  }
}

//
//
//



struct ReadOnlyWL::Internal {
  string wl_file_name;
  string sl_file_name;
  aspell *_spell_checker;
  SC_ReadOnly_WordList    words;
  SC_ReadOnly_Soundslikes soundslikes;
  Internal(aspell *sc) : _spell_checker(sc), words(sc), soundslikes(sc) {}
};

#define real_itr SC_ReadOnly_WordList::const_iterator
#define base     ReadOnlyWL
#define value_func(obj) (obj).c_str()

#include "proxy_iterator.cc"

#undef value_func
#undef base

ReadOnlyWL::const_iterator ReadOnlyWL::begin() const {
  real_itr i = internal->words.begin();
  return const_iterator(&i);
}

ReadOnlyWL::const_iterator ReadOnlyWL::end() const {
  real_itr i = internal->words.end();
  return const_iterator(&i);
}

ReadOnlyWL::size_type ReadOnlyWL::size() const {
  return internal->words.size();
}

bool ReadOnlyWL::empty() const {
  return internal->words.empty();
}

#undef real_itr

ReadOnlyWL::ReadOnlyWL(aspell *sc) {
  internal = new Internal(sc);
}

ReadOnlyWL::~ReadOnlyWL() {
  delete internal;
}

ReadOnlyWL::WordList::WordList() {
  info = 0;
}

ReadOnlyWL::WordList::WordList(const SC_ReadOnly_Soundslikes::WordList *i)
  : info (i) {}

ReadOnlyWL::WordList::const_iterator ReadOnlyWL::WordList::begin() const {
  return info->begin();
}

ReadOnlyWL::WordList::const_iterator ReadOnlyWL::WordList::end() const {
  return info->end();
}

ReadOnlyWL::WordList::size_type ReadOnlyWL::WordList::size() const {
  if (!info) return 0;
  return info->size();
}

bool ReadOnlyWL::WordList::empty() const {
  if (!info) return true;
  return info->empty();
}

const SC_Error ReadOnlyWL::load(const string &b) {
  internal->wl_file_name = find_data_file(b, ".wrd");
  internal->sl_file_name = find_data_file(b, ".sl");
  SC_Error error = internal->words.load(internal->wl_file_name);
  if (error) return error;
  return internal->soundslikes.load(internal->sl_file_name, 
				    internal->words.get_word_block());
}

bool ReadOnlyWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool ReadOnlyWL::operator [] (const string &word) const {
  return internal->words[word];
}

const ReadOnlyWL::WordList ReadOnlyWL::words_w_soundslike(const char *word) {
  return internal->soundslikes.words_w_soundslike(word);
}

//
//
//

static time_t modification_date(const string &file) {
  struct stat file_stat;
  if (stat(file.c_str(), &file_stat) == 0)
    return file_stat.st_mtime;
  else 
    return 0;
}

//
//
//

#define real_itr SC_Writable_Soundslikes::WordList::const_iterator
#define base     WritableWL::WordList 
#define include_operator_plus

#include "proxy_iterator.cc"
#include "proxy_container.cc"

#undef include_operator_plus
#undef real_itr
#undef base

struct WritableWL::Internal {
  string file_name;
  time_t file_date;
  aspell *_spell_checker;
  SC_Writable_WordList    words;
  SC_Writable_Soundslikes soundslikes;
  Internal(aspell *sc) : _spell_checker(sc), words(sc), soundslikes(sc) {}
  const SC_Error save_as(const string &);
  const SC_Error merge(const string &);
};


#define real_itr SC_Writable_WordList::const_iterator
#define base     WritableWL
#define value_func(obj) (obj).c_str()

#include "proxy_iterator.cc"

#undef value_func
#undef base

WritableWL::const_iterator WritableWL::begin() const {
  real_itr i = internal->words.begin();
  return const_iterator(&i);
}

WritableWL::const_iterator WritableWL::end() const {
  real_itr i = internal->words.end();
  return const_iterator(&i);
}

WritableWL::size_type WritableWL::size() const {
  return internal->words.size();
}

bool WritableWL::empty() const {
  return internal->words.empty();
}

#undef real_itr

const SC_Error WritableWL::Internal::merge(const string &b) {
  string fn = find_per_file(b,".per");
  ifstream WORDS( fn.c_str() );
  if ( WORDS.rdstate() & ios::badbit ) return SC_Error();

  int c;
  string word, sound;
  const char *word2;
  
  WORDS >> word;
  if (word != "personal_wl")
    return SC_Error(fn , SC_Error::bad_format);

  WORDS >> word;
  {
    SC_Error error = _spell_checker->set_check_lang(word);
    if (error) {
      error.file = fn;
      return error;
    }
  }

  WORDS >> c; // not used at the moment
  
  for (;;) {
    WORDS >> word >> sound;
    if (!WORDS) break;
    word2 = words.add(word);
    soundslikes.add(word2, sound);
  }

  return SC_Error();
}

const SC_Error WritableWL::Internal::save_as(const string &b) {
  file_name = find_per_file(b, ".per");
  
  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    return SC_Error(file_name, SC_Error::cant_write);

  WORDS << "personal_wl" << ' ' << _spell_checker->lang_name() <<  endl;
  WORDS << words.size() << endl;

  SC_Writable_Soundslikes::const_iterator i = soundslikes.begin();
  SC_Writable_Soundslikes::const_iterator e = soundslikes.end();

  SC_Writable_Soundslikes::WordList::const_iterator j;
  
  for (;i != e; ++i) {
    for (j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << *j << ' ' << i->first << endl;
    }
  }

  WORDS.close();
  file_date = modification_date(file_name);
  
  return SC_Error();
}

WritableWL::WritableWL(aspell *sc) {
  internal = new Internal(sc);
}

const SC_Error WritableWL::load(const string &b) {
  internal->file_name = find_per_file(b, ".per");
  internal->file_date = modification_date(internal->file_name);
  SC_Error error = internal->merge(internal->file_name);
}

const SC_Error WritableWL::merge(const string &b) {
  return internal->merge(b);
}

const SC_Error WritableWL::update() {
  internal->file_date = modification_date(internal->file_name);
  return internal->merge(internal->file_name);
}

const SC_Error WritableWL::save_as(const string &b) {
  return internal->save_as(b);
}

const SC_Error WritableWL::save() {
  time_t new_date = modification_date(internal->file_name);
  if (new_date > internal->file_date)
    update();
  SC_Error error = internal->save_as(internal->file_name);
  if (!error)
    internal->file_date = modification_date(internal->file_name);
  return error;
}

const char * WritableWL::file_name() {
  return internal->file_name.c_str();
}

void WritableWL::add(const string &w, const string &s) {
  if (!operator[](w))
    internal->soundslikes.add(internal->words.add(w),s);
}

void WritableWL::clear() {
  internal->words.clear(); 
  internal->soundslikes.clear();
}

bool WritableWL::operator [] (const char *word) const {
  return internal->words[word];
}

bool WritableWL::operator [] (const string &word) const {
  return internal->words[word];
}

const WritableWL::WordList WritableWL::words_w_soundslike(const char *word) {
  return internal->soundslikes.words_w_soundslike(word);
}

//
//
//

#define real_itr vector<WritableReplList::Replacements>::const_iterator
#define base     WritableReplList::WordList

#include "proxy_iterator.cc"
#include "proxy_container.cc"

#undef real_itr
#undef base

bool WritableReplList::Replacements::exists(const string &word) {
  iterator i = begin();
  iterator e = end();
  while (i != e) {
    if (*i == word) return true;
    ++i;
  }
  return false;
}

bool WritableReplList::Replacements::add(const string &word) {
  if (exists(word)) return false;
  info.push_back(word);
  return true;
}

bool WritableReplList::Replacements::erase(const string &word) {
  vector<string>::iterator i = info.begin() + 1;
  vector<string>::iterator e = info.end();
  while (i != e) {
    if (*i == word) {
      info.erase(i);
      return true;
    }
    ++i;
  }
  return false;
}

struct WritableReplList::Internal {
  string file_name;
  time_t file_date;
  aspell *_spell_checker;
  typedef hash_map<string, WordList::Real, hash_string<string> > LookupTable;
  LookupTable lookup_table;
  Internal(aspell *sc) : _spell_checker(sc) {}
  const SC_Error save_as(const string &);
  const SC_Error load(const string &);
  const SC_Error merge(const string &);
};

WritableReplList::WritableReplList(aspell *sc) {
  internal = new Internal(sc);
}

WritableReplList::~WritableReplList() {
  delete internal;
}

void WritableReplList
::add(const string &mis, const string &cor, const string &s) {
  Internal::LookupTable::iterator i = internal->lookup_table.find(s);
  if (i == internal->lookup_table.end())
    i = internal->lookup_table.insert
          (Internal::LookupTable::value_type(s, WordList::Real())).first;
  
  WordList::Real::iterator    j = i->second.begin();
  WordList::Real::iterator    e = i->second.end();
  for (; j != e; ++j) {
    if ((*j).misspelled_word() == mis) {
      (*j).add(cor);
      return;
    }
  }
  i->second.push_back(Replacements(mis,cor));
}

const WritableReplList::WordList WritableReplList
::words_w_soundslike(const char *word) const {
  Internal::LookupTable::const_iterator i = internal->lookup_table.find(word);
  if (i == internal->lookup_table.end()) {
    return NULL;
  } else {
    return &(i->second);
  }
}

const SC_Error WritableReplList::Internal::save_as(const string &b) {
  file_name = find_per_file(b,".rpl");
  ofstream WORDS((file_name).c_str());

  if (WORDS.rdstate() & ios::badbit) 
    return SC_Error(file_name, SC_Error::cant_write);

  WORDS << "personal_repl" << ' ' << _spell_checker->lang_name() <<  endl;
  WORDS << lookup_table.size() << ' ' << lookup_table.bucket_count() << endl;
  
  LookupTable::iterator i = lookup_table.begin();
  LookupTable::iterator e = lookup_table.end();
  
  for (;i != e; ++i) {
    WORDS << i->first << ' ' << i->second.size() << endl;
    for (WordList::Real::iterator j = i->second.begin(); j != i->second.end(); ++j) {
      WORDS << "  " << (*j).misspelled_word() << ' ' << (*j).size() << ' ';
      for (Replacements::iterator k = j->begin(); k != j->end(); ++k) {
	WORDS << *k << ',';
      }
      WORDS << endl;
    }
  }

  WORDS.close();
  
  file_date = modification_date(file_name);

  return SC_Error();
}

const SC_Error WritableReplList::save_as(const string &b) {
  return internal->save_as(b);
}

const SC_Error WritableReplList::save() {
  time_t new_date = modification_date(internal->file_name);
  if (new_date > internal->file_date)
    update();
  SC_Error error = internal->save_as(internal->file_name);
  if (!error)
    internal->file_date = modification_date(internal->file_name);
  return error;
}

const SC_Error WritableReplList::Internal::load(const string &b) {
  file_name = find_per_file(b, ".rpl");
  file_date = modification_date(file_name);
  ifstream WORDS( (file_name).c_str() );
  if ( WORDS.rdstate() & ios::badbit ) return SC_Error();

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    return SC_Error(file_name, SC_Error::bad_format);

  WORDS >> word;
  {
    SC_Error error = _spell_checker->set_check_lang(word);
    if (error) {
      error.file = file_name;
      return error;
    }
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    WordList::Real &wl = lookup_table.insert(
      LookupTable::value_type(sound,WordList::Real())).first->second;
    wl.reserve(num_words);
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      wl.push_back(Replacements(mis,num_repls));
      Replacements &rp = *(wl.end()-1);
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS, repl, ',');
	rp.add_nocheck(repl);
      }
    }
  }
  return SC_Error();
}

const SC_Error WritableReplList::Internal::merge(const string &b) {

  string fn = find_per_file(b, ".rpl");

  ifstream WORDS( (fn).c_str() );
  if ( WORDS.rdstate() & ios::badbit ) return SC_Error();

  int c;
  string word, mis, sound, repl;
  int num_words, num_repls;

  WORDS >> word;
  if (word != "personal_repl")
    return SC_Error(fn, SC_Error::bad_format);

  WORDS >> word;
  {
    SC_Error error = _spell_checker->set_check_lang(word);
    if (error) {
      error.file = fn;
      return error;
    }
  }

  int num_soundslikes;
  WORDS >> num_soundslikes;
  WORDS >> c;  // not used at the moment

  int h,i,j;
  WordList::Real::iterator k;
  bool found;
  Replacements * rp;
  for (h=0; h != num_soundslikes; ++h) {
    WORDS >> sound >> num_words;
    WordList::Real &wl = lookup_table.insert(
      LookupTable::value_type(sound,WordList::Real())).first->second;
    for (i = 0; i != num_words; ++i) {
      WORDS >> mis >> num_repls;
      WORDS.ignore(); // ignore space
      found = false;
      for (k = wl.begin(); k != wl.end(); ++k) {
	if (mis == (*k).misspelled_word()) {
	  found = true;
	  break;
	}
      }
      if (found) {
	rp = &(*k);
      } else {
	wl.push_back(Replacements(mis,num_repls));
        rp = &(*(wl.end()-1));
      }
      for (j = 0; j != num_repls; ++j) {
	getline(WORDS,repl,';');
	rp->add(repl);
      }
    }
  }
  return SC_Error();
}

const SC_Error WritableReplList::load(const string &b) {
  return internal->load(b);
}

const SC_Error WritableReplList::merge(const string &b) {
  return internal->merge(b);
}

const SC_Error WritableReplList::update() {
  internal->file_date = modification_date(internal->file_name);
  return internal->merge(internal->file_name);
}

const char * WritableReplList::file_name() {
  return internal->file_name.c_str();
}
