#ifndef data_hh_
#define data_hh_

#include "as_config.h"
#include <string>
#include <hash_set>
#include <hash_map>
#include "hash_string.hh"
#include "hash.hh"
#include "as_error.hh"
#include "aspell.hh"
#include "as_data.hh"

class SC_ReadOnly_WordList {
private:
  typedef 
    vector_hash_set
     < SC_Word, hash_string<const_string>, 
       is_nonexistent_container<const_string>, 
       make_nonexistent_container<const_string>,
       equal_to<SC_Word> 
     >
     WordList;

  aspell      *_spell_checker;
  WordList    word_list;
  char*      _master;

  SC_ReadOnly_WordList(const SC_ReadOnly_WordList&);
  SC_ReadOnly_WordList& operator = (const SC_ReadOnly_WordList&);

public:
  typedef WordList::size_type size_type;
  typedef WordList::const_iterator const_iterator;
  typedef WordList::const_iterator iterator;

  size_type size() const {return word_list.size();}
  bool empty() const {return word_list.empty();}
  const_iterator begin() const {return word_list.begin();}
  const_iterator end() const {return word_list.end();}
  
  const SC_Error load(const string &);

  SC_ReadOnly_WordList(SpellChecker *sc) : _spell_checker(sc), _master(0) {}
  ~SC_ReadOnly_WordList() {delete[] _master;}

  const char * get_word_block() {return _master;}

  bool operator [] (const char *word) const 
    {return word_list.count(word);}
  bool operator [] (const string &word) const 
    {return operator[] (word.c_str());}
  const SC_Word& get(const char *word) const 
    {return *word_list.find(word);}
  const SC_Word& get(const string &word) const 
    {return get(word.c_str());}
};

class SC_Writable_WordList {
private:
  typedef
    hash_set<string, hash_string<string>, equal_to<string> >
    WordList;
  WordList word_list;
  aspell  *_spell_checker;
  string  name;

public:
  SC_Writable_WordList(SpellChecker *sc) : _spell_checker(sc) {}

  typedef WordList::size_type size_type;
  typedef WordList::const_iterator const_iterator;
  typedef WordList::const_iterator iterator;

  size_type size() const {return word_list.size();}
  bool empty() const {return word_list.empty();}
  const_iterator begin() const {return word_list.begin();}
  const_iterator end() const {return word_list.end();}

  SC_Word add(const string &word) {return word_list.insert(word).first->c_str();}
  SC_Word add(const char *word) {return word_list.insert(word).first->c_str();}
  void erase(const string &word) {word_list.erase(word);}
  void erase(const char *word) {word_list.erase(word);}

  void clear() {word_list.erase(word_list.begin(),word_list.end());}

  bool operator [] (const char *word) const 
    {return word_list.count(word);}
  bool operator [] (const string &word) const 
    {return word_list.count(word);}
  const SC_Word get(const char *word) const 
    {return word_list.find(word)->c_str();}
  const SC_Word get(const string &word) const 
    {return word_list.find(word)->c_str();}
};


class  SC_ReadOnly_Soundslikes_WordList {
private:
  const char *data_;
public:
  SC_ReadOnly_Soundslikes_WordList() {}
  SC_ReadOnly_Soundslikes_WordList(const char *d) : data_(d) {}
  
  SC_ReadOnly_Soundslikes_WordList& operator= (const char *d) 
    {data_=d; return *this;}

  const_string key() const {return data_;}

  typedef const char * *      iterator;
  typedef unsigned short int  size_type;

  size_type size() const 
    {return *reinterpret_cast<const size_type *>(data_-sizeof(size_type));}
  bool empty() const {return !size();}
  iterator begin() const
    {return end() - size();}
  iterator end() const
    {return reinterpret_cast<iterator>(data_-sizeof(size_type));}
};

typedef SC_ReadOnly_Soundslikes_WordList SoundslikeList;

struct SL_Hash {
  hash_string<const_string>  hasher;
  size_t operator() (const SoundslikeList& i) const {return hasher(i.key());}
  size_t operator() (const_string i) const {return hasher(i);}
};

struct SL_NonExistent {
  bool operator() (const SoundslikeList &i) const {return i.key().empty();}
};

struct SL_MakeNonExistent {
  void operator() (SoundslikeList &i) const {i = "";}
};

struct SL_EqualTo {
  bool operator() (const SoundslikeList& i, const SoundslikeList& j) const
    {return i.key() == j.key();}
  bool operator() (const SoundslikeList &i, const_string j) const
    {return i.key() == j;}
};

class SC_ReadOnly_Soundslikes {
private:
  typedef SoundslikeList   WordList;
  aspell *       _spell_checker;

  typedef  
    vector_hash_table
    <
     vector<SoundslikeList>, const_string, SoundslikeList,
     SL_Hash, SL_EqualTo, SL_NonExistent, SL_MakeNonExistent
    >
    LookupTable;
  LookupTable   lookup_table;
  char*         main_data;

  SC_ReadOnly_Soundslikes(const SC_Soundslikes&);
  SC_ReadOnly_Soundslikes& operator = (const SC_ReadOnly_Soundslikes&);

public:
  const WordList * words_w_soundslike(const char *word);
  const WordList * words_w_soundslike(const char *word, SC_Sound &M);

  const SC_Error load(const string &, const char *);

  SC_ReadOnly_Soundslikes(SpellChecker *sc) : _spell_checker(sc), main_data(0) {}
  ~SC_ReadOnly_Soundslikes() {delete[] main_data;}
};

typedef vector<SC_Word> SC_Writable_Soundslikes_WordList;

class SC_Writable_Soundslikes {
public:
  typedef vector<SC_Word> WordList;

private:
  static const WordList empty_list;
  aspell *       _spell_checker;

  typedef
    hash_map<string, vector<SC_Word>, hash_string<string>, equal_to<string> >
    LookupTable;
  LookupTable lookup_table;


public:
  typedef LookupTable::const_iterator const_iterator;
  typedef const_iterator                    iterator;

  const_iterator begin() {return lookup_table.begin();}
  const_iterator end()   {return lookup_table.end();}
  
  void add(const char* word, const string&);
  void erase(const char* word, const string&);
  void clear() {lookup_table.erase(lookup_table.begin(),lookup_table.end());}

  const WordList * words_w_soundslike(const char *word);
  const WordList * words_w_soundslike(const char *word, SC_Sound &M);

  SC_Writable_Soundslikes(SpellChecker *sc) : _spell_checker(sc) {}
};

#endif



