#include <ctype.h>
#include <string>

#define FALSE (0)
#define TRUE (!0)

// English Spelling to Phoneme translator.  Extracted from:

/*
                        Final Version of
                ENGLISH TO PHONEME TRANSLATION
                           4/15/85

        Here it is one last time.  I have fixed all of the bugs I
        heard about and added a new feature or two (it now talks
        money as well as numbers).  I think that this version is
        good enough for most purposes.  I have proof-read the
        phoneme rules (found one bug) and made the program more
        "robust".  I added protection against the "toupper()"
        problem some people had with earlier versions.

        If you make a major addition (like better abbreviation
        handling or an exception dictionary) please send me a
        copy.  As before, this is all public domain and I make
        no copyright claims on it.  The part derived from the
        Naval Research Lab should be public anyway.  Sell it
        if you can!

                -John A. Wasser
Work address:
ARPAnet:        WASSER%VIKING.DEC@decwrl.ARPA
Usenet:         {allegra,Shasta,decvax}!decwrl!dec-rhea!dec-viking!wasser
Easynet:        VIKING::WASSER
Telephone:      (617)486-2505
USPS:           Digital Equipment Corp.
                Mail stop: LJO2/E4
                30 Porter Rd
                Littleton, MA  01460
*/

// I isolated the spelling to pheneme conversion parts and extracted
// them from his program and changed to ANSI C so that it can compile
// ok under C++ and then a added a tiny bit of C++ code.

//
// As with the orignal version this file is in the Public Domain
//

//
// The orignal can be found at:
//    ftp://svr-ftp.eng.cam.ac.uk/pub/comp.speech/synthesis/english2phoneme.tar.gz
// or ftp://ftp.doc.ic.ac.uk/packages/unix-c/utils/phoneme.c.gz
// I found it off of the comp.speech.faq which can be found at
//   http://robotweb.ri.cmu.edu/comp.speech/

/* ** English to Phoneme rules.
**
**      Derived from: 
**
**           AUTOMATIC TRANSLATION OF ENGLISH TEXT TO PHONETICS
**                  BY MEANS OF LETTER-TO-SOUND RULES
**
**                      NRL Report 7948
**
**                    January 21st, 1976
**          Naval Research Laboratory, Washington, D.C.
**
**
**      Published by the National Technical Information Service as
**      document "AD/A021 929".
**
**
**
**      The Phoneme codes:
**
**              IY (E)  bEEt            IH (i)  bIt
**              EY (A)  gAte            EH (e)  gEt
**              AE (a)  fAt             AA (4)  fAther
**              AO (1)  lAWn            OW (o)  lOne
**              UH (U)  fUll            UW (O)  fOOl
**              ER (2)  mURdER          AX (5)  About
**              AH (u)  bUt             AY (I)  hIde
**              AW (3)  hOW             OY (Y)  tOY
**      
**              p       Pack            b       Back
**              t       Time            d       Dime
**              k       Coat            g       Goat
**              f       Fault           v       Vault
**              TH (T)  eTHer           DH (D)  eiTHer
**              s       Sue             z       Zoo
**              SH (S)  leaSH           ZH (Z)  leiSure
**              HH (h)  How             m       suM
**              n       suN             NG (N)  suNG
**              l       Laugh           w       Wear
**              y       Young           r       Rate
**              CH (C)  CHar            j       Jar
**              WH (W)  WHere
**
**  The charters in parenthesis are the charters actually used in the
**  conversion
**
**      Rules are made up of four parts:
**      
**              The left context.
**              The text to match.
**              The right context.
**              The phonemes to substitute for the matched text.
**
**      Procedure:
**
**              Seperate each block of letters (apostrophes included) 
**              and add a space on each side.  For each unmatched 
**              letter in the word, look through the rules where the 
**              text to match starts with the letter in the word.  If 
**              the text to match is found and the right and left 
**              context patterns also match, output the phonemes for 
**              that rule and skip to the next unmatched letter.
**
**
**      Special Context Symbols:
**
**              #       One or more vowels
**              :       Zero or more consonants
**              ^       One consonant.
**              .       One of B, D, V, G, J, L, M, N, R, W or Z (voiced 
**                      consonants)
**              %       One of ER, E, ES, ED, ING, ELY (a suffix)
**                      (Found in right context only)
**              +       One of E, I or Y (a "front" vowel)
** */


/* Context definitions */
static char Anything[] = "";    /* No context requirement */
static char Nothing[] = " ";    /* Context is beginning or end of word */

/* Phoneme definitions */
static char Pause[] = " ";      /* Short silence */
static char Silent[] = "";      /* No phonemes */

#define LEFT_PART       0
#define MATCH_PART      1
#define RIGHT_PART      2
#define OUT_PART        3

typedef char *Rule[4];  /* Rule is an array of 4 character pointers */

/*0 = Punctuation */
/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule punct_rules[] =
        {
        {Anything,      " ",            Anything,       Pause   },
        {Anything,      "-",            Anything,       Silent  },
        {".",           "'S",           Anything,       "z"     },
        {"#:.E",        "'S",           Anything,       "z"     },
        {"#",           "'S",           Anything,       "z"     },
        {Anything,      "'",            Anything,       Silent  },
        {Anything,      ",",            Anything,       Pause   },
        {Anything,      ".",            Anything,       Pause   },
        {Anything,      "?",            Anything,       Pause   },
        {Anything,      "!",            Anything,       Pause   },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule A_rules[] =
        {
        {Anything,      "A",            Nothing,        "5"    },
        {Nothing,       "ARE",          Nothing,        "4r"   },
        {Nothing,       "AR",           "O",            "5r"   },
        {Anything,      "AR",           "#",            "er"   },
        {"^",           "AS",           "#",            "As"   },
        {Anything,      "A",            "WA",           "5"    },
        {Anything,      "AW",           Anything,       "1"    },
        {" :",          "ANY",          Anything,       "enE" },
        {Anything,      "A",            "^+#",          "A"    },
        {"#:",          "ALLY",         Anything,       "5lE" },
        {Nothing,       "AL",           "#",            "5l"   },
        {Anything,      "AGAIN",        Anything,       "5gen"},
        {"#:",          "AG",           "E",            "ij"   },
        {Anything,      "A",            "^+:#",         "a"    },
        {" :",          "A",            "^+ ",          "A"    },
        {Anything,      "A",            "^%",           "A"    },
        {Nothing,       "ARR",          Anything,       "5r"   },
        {Anything,      "ARR",          Anything,       "ar"   },
        {" :",          "AR",           Nothing,        "4r"   },
        {Anything,      "AR",           Nothing,        "2"    },
        {Anything,      "AR",           Anything,       "4r"   },
        {Anything,      "AIR",          Anything,       "er"   },
        {Anything,      "AI",           Anything,       "A"    },
        {Anything,      "AY",           Anything,       "A"    },
        {Anything,      "AU",           Anything,       "1"    },
        {"#:",          "AL",           Nothing,        "5l"   },
        {"#:",          "ALS",          Nothing,        "5lz"  },
        {Anything,      "ALK",          Anything,       "1k"   },
        {Anything,      "AL",           "^",            "1l"   },
        {" :",          "ABLE",         Anything,       "Ab5l"},
        {Anything,      "ABLE",         Anything,       "5b5l"},
        {Anything,      "ANG",          "+",            "Anj"  },
        {Anything,      "A",            Anything,       "a"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule B_rules[] =
        {
        {Nothing,       "BE",           "^#",           "bi"   },
        {Anything,      "BEING",        Anything,       "bEiN"},
        {Nothing,       "BOTH",         Nothing,        "boT" },
        {Nothing,       "BUS",          "#",            "biz"  },
        {Anything,      "BUIL",         Anything,       "bil"  },
        {Anything,      "B",            Anything,       "b"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule C_rules[] =
        {
        {Nothing,       "CH",           "^",            "k"     },
        {"^E",          "CH",           Anything,       "k"     },
        {Anything,      "CH",           Anything,       "C"    },
        {" S",          "CI",           "#",            "sI"   },
        {Anything,      "CI",           "A",            "S"    },
        {Anything,      "CI",           "O",            "S"    },
        {Anything,      "CI",           "EN",           "S"    },
        {Anything,      "C",            "+",            "s"     },
        {Anything,      "CK",           Anything,       "k"     },
        {Anything,      "COM",          "%",            "kum"  },
        {Anything,      "C",            Anything,       "k"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule D_rules[] =
        {
        {"#:",          "DED",          Nothing,        "did"  },
        {".E",          "D",            Nothing,        "d"     },
        {"#:^E",        "D",            Nothing,        "t"     },
        {Nothing,       "DE",           "^#",           "di"   },
        {Nothing,       "DO",           Nothing,        "dO"   },
        {Nothing,       "DOES",         Anything,       "duz"  },
        {Nothing,       "DOING",        Anything,       "dOiN"},
        {Nothing,       "DOW",          Anything,       "d3"   },
        {Anything,      "DU",           "A",            "jO"   },
        {Anything,      "D",            Anything,       "d"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule E_rules[] =
        {
        {"#:",          "E",            Nothing,        Silent  },
        {"':^",         "E",            Nothing,        Silent  },
        {" :",          "E",            Nothing,        "E"    },
        {"#",           "ED",           Nothing,        "d"     },
        {"#:",          "E",            "D ",           Silent  },
        {Anything,      "EV",           "ER",           "ev"   },
        {Anything,      "E",            "^%",           "E"    },
        {Anything,      "ERI",          "#",            "ErE" },
        {Anything,      "ERI",          Anything,       "eri" },
        {"#:",          "ER",           "#",            "2"    },
        {Anything,      "ER",           "#",            "er"   },
        {Anything,      "ER",           Anything,       "2"    },
        {Nothing,       "EVEN",         Anything,       "Even"},
        {"#:",          "E",            "W",            Silent  },
        {"T",           "EW",           Anything,       "O"    },
        {"S",           "EW",           Anything,       "O"    },
        {"R",           "EW",           Anything,       "O"    },
        {"D",           "EW",           Anything,       "O"    },
        {"L",           "EW",           Anything,       "O"    },
        {"Z",           "EW",           Anything,       "O"    },
        {"N",           "EW",           Anything,       "O"    },
        {"J",           "EW",           Anything,       "O"    },
        {"TH",          "EW",           Anything,       "O"    },
        {"CH",          "EW",           Anything,       "O"    },
        {"SH",          "EW",           Anything,       "O"    },
        {Anything,      "EW",           Anything,       "yO"   },
        {Anything,      "E",            "O",            "E"    },
        {"#:S",         "ES",           Nothing,        "iz"   },
        {"#:C",         "ES",           Nothing,        "iz"   },
        {"#:G",         "ES",           Nothing,        "iz"   },
        {"#:Z",         "ES",           Nothing,        "iz"   },
        {"#:X",         "ES",           Nothing,        "iz"   },
        {"#:J",         "ES",           Nothing,        "iz"   },
        {"#:CH",        "ES",           Nothing,        "iz"   },
        {"#:SH",        "ES",           Nothing,        "iz"   },
        {"#:",          "E",            "S ",           Silent  },
        {"#:",          "ELY",          Nothing,        "lE"   },
        {"#:",          "EMENT",        Anything,       "ment" },
        {Anything,      "EFUL",         Anything,       "fUl"  },
        {Anything,      "EE",           Anything,       "E"    },
        {Anything,      "EARN",         Anything,       "2n"   },
        {Nothing,       "EAR",          "^",            "2"    },
        {Anything,      "EAD",          Anything,       "ed"   },
        {"#:",          "EA",           Nothing,        "E5"  },
        {Anything,      "EA",           "SU",           "e"    },
        {Anything,      "EA",           Anything,       "E"    },
        {Anything,      "EIGH",         Anything,       "A"    },
        {Anything,      "EI",           Anything,       "E"    },
        {Nothing,       "EYE",          Anything,       "I"    },
        {Anything,      "EY",           Anything,       "E"    },
        {Anything,      "EU",           Anything,       "yO"   },
        {Anything,      "E",            Anything,       "e"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule F_rules[] =
        {
        {Anything,      "FUL",          Anything,       "fUl"  },
        {Anything,      "F",            Anything,       "f"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule G_rules[] =
        {
        {Anything,      "GIV",          Anything,       "giv"  },
        {Nothing,       "G",            "I^",           "g"     },
        {Anything,      "GE",           "T",            "ge"   },
        {"SU",          "GGES",         Anything,       "gjes" },
        {Anything,      "GG",           Anything,       "g"     },
        {" B#",         "G",            Anything,       "g"     },
        {Anything,      "G",            "+",            "j"     },
        {Anything,      "GREAT",        Anything,       "grAt" },
        {"#",           "GH",           Anything,       Silent  },
        {Anything,      "G",            Anything,       "g"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule H_rules[] =
        {
        {Nothing,       "HAV",          Anything,       "hav"  },
        {Nothing,       "HERE",         Anything,       "hEr"  },
        {Nothing,       "HOUR",         Anything,       "32"  },
        {Anything,      "HOW",          Anything,       "h3"   },
        {Anything,      "H",            "#",            "h"     },
        {Anything,      "H",            Anything,       Silent  },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule I_rules[] =
        {
        {Nothing,       "IN",           Anything,       "in"   },
        {Nothing,       "I",            Nothing,        "I"    },
        {Anything,      "IN",           "D",            "In"   },
        {Anything,      "IER",          Anything,       "E2"  },
        {"#:R",         "IED",          Anything,       "Ed"   },
        {Anything,      "IED",          Nothing,        "Id"   },
        {Anything,      "IEN",          Anything,       "Een" },
        {Anything,      "IE",           "T",            "Ie"  },
        {" :",          "I",            "%",            "I"    },
        {Anything,      "I",            "%",            "E"    },
        {Anything,      "IE",           Anything,       "E"    },
        {Anything,      "I",            "^+:#",         "i"    },
        {Anything,      "IR",           "#",            "Ir"   },
        {Anything,      "IZ",           "%",            "Iz"   },
        {Anything,      "IS",           "%",            "Iz"   },
        {Anything,      "I",            "D%",           "I"    },
        {"+^",          "I",            "^+",           "i"    },
        {Anything,      "I",            "T%",           "I"    },
        {"#:^",         "I",            "^+",           "i"    },
        {Anything,      "I",            "^+",           "I"    },
        {Anything,      "IR",           Anything,       "2"    },
        {Anything,      "IGH",          Anything,       "I"    },
        {Anything,      "ILD",          Anything,       "Ild"  },
        {Anything,      "IGN",          Nothing,        "In"   },
        {Anything,      "IGN",          "^",            "In"   },
        {Anything,      "IGN",          "%",            "In"   },
        {Anything,      "IQUE",         Anything,       "Ek"   },
        {Anything,      "I",            Anything,       "i"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule J_rules[] =
        {
        {Anything,      "J",            Anything,       "j"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule K_rules[] =
        {
        {Nothing,       "K",            "N",            Silent  },
        {Anything,      "K",            Anything,       "k"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule L_rules[] =
        {
        {Anything,      "LO",           "C#",           "lo"   },
        {"L",           "L",            Anything,       Silent  },
        {"#:^",         "L",            "%",            "5l"   },
        {Anything,      "LEAD",         Anything,       "lEd"  },
        {Anything,      "L",            Anything,       "l"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule M_rules[] =
        {
        {Anything,      "MOV",          Anything,       "mOv"  },
        {Anything,      "M",            Anything,       "m"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule N_rules[] =
        {
        {"E",           "NG",           "+",            "nj"    },
        {Anything,      "NG",           "R",            "Ng"   },
        {Anything,      "NG",           "#",            "Ng"   },
        {Anything,      "NGL",          "%",            "Ng5l"},
        {Anything,      "NG",           Anything,       "N"    },
        {Anything,      "NK",           Anything,       "Nk"   },
        {Nothing,       "NOW",          Nothing,        "n3"   },
        {Anything,      "N",            Anything,       "n"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule O_rules[] =
        {
        {Anything,      "OF",           Nothing,        "5v"   },
        {Anything,      "OROUGH",       Anything,       "2o"  },
        {"#:",          "OR",           Nothing,        "2"    },
        {"#:",          "ORS",          Nothing,        "2z"   },
        {Anything,      "OR",           Anything,       "1r"   },
        {Nothing,       "ONE",          Anything,       "wun"  },
        {Anything,      "OW",           Anything,       "o"    },
        {Nothing,       "OVER",         Anything,       "ov2" },
        {Anything,      "OV",           Anything,       "uv"   },
        {Anything,      "O",            "^%",           "o"    },
        {Anything,      "O",            "^EN",          "o"    },
        {Anything,      "O",            "^I#",          "o"    },
        {Anything,      "OL",           "D",            "ol"   },
        {Anything,      "OUGHT",        Anything,       "1t"   },
        {Anything,      "OUGH",         Anything,       "uf"   },
        {Nothing,       "OU",           Anything,       "3"    },
        {"H",           "OU",           "S#",           "3"    },
        {Anything,      "OUS",          Anything,       "5s"   },
        {Anything,      "OUR",          Anything,       "1r"   },
        {Anything,      "OULD",         Anything,       "Ud"   },
        {"^",           "OU",           "^L",           "u"    },
        {Anything,      "OUP",          Anything,       "Op"   },
        {Anything,      "OU",           Anything,       "3"    },
        {Anything,      "OY",           Anything,       "Y"    },
        {Anything,      "OING",         Anything,       "oiN"},
        {Anything,      "OI",           Anything,       "Y"    },
        {Anything,      "OOR",          Anything,       "1r"   },
        {Anything,      "OOK",          Anything,       "Uk"   },
        {Anything,      "OOD",          Anything,       "Ud"   },
        {Anything,      "OO",           Anything,       "O"    },
        {Anything,      "O",            "E",            "o"    },
        {Anything,      "O",            Nothing,        "o"    },
        {Anything,      "OA",           Anything,       "o"    },
        {Nothing,       "ONLY",         Anything,       "onlE"},
        {Nothing,       "ONCE",         Anything,       "wuns" },
        {Anything,      "ON'T",         Anything,       "ont"  },
        {"C",           "O",            "N",            "4"    },
        {Anything,      "O",            "NG",           "1"    },
        {" :^",         "O",            "N",            "u"    },
        {"I",           "ON",           Anything,       "5n"   },
        {"#:",          "ON",           Nothing,        "5n"   },
        {"#^",          "ON",           Anything,       "5n"   },
        {Anything,      "O",            "ST ",          "o"    },
        {Anything,      "OF",           "^",            "1f"   },
        {Anything,      "OTHER",        Anything,       "uD2"},
        {Anything,      "OSS",          Nothing,        "1s"   },
        {"#:^",         "OM",           Anything,       "um"   },
        {Anything,      "O",            Anything,       "4"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule P_rules[] =
        {
        {Anything,      "PH",           Anything,       "f"     },
        {Anything,      "PEOP",         Anything,       "pEp"  },
        {Anything,      "POW",          Anything,       "p3"   },
        {Anything,      "PUT",          Nothing,        "pUt"  },
        {Anything,      "P",            Anything,       "p"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule Q_rules[] =
        {
        {Anything,      "QUAR",         Anything,       "kw1r" },
        {Anything,      "QU",           Anything,       "kw"    },
        {Anything,      "Q",            Anything,       "k"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule R_rules[] =
        {
        {Nothing,       "RE",           "^#",           "rE"   },
        {Anything,      "R",            Anything,       "r"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule S_rules[] =
        {
        {Anything,      "SH",           Anything,       "S"    },
        {"#",           "SION",         Anything,       "Z5n" },
        {Anything,      "SOME",         Anything,       "sum"  },
        {"#",           "SUR",          "#",            "Z2"  },
        {Anything,      "SUR",          "#",            "S2"  },
        {"#",           "SU",           "#",            "ZO"  },
        {"#",           "SSU",          "#",            "SO"  },
        {"#",           "SED",          Nothing,        "zd"    },
        {"#",           "S",            "#",            "z"     },
        {Anything,      "SAID",         Anything,       "sed"  },
        {"^",           "SION",         Anything,       "S5n" },
        {Anything,      "S",            "S",            Silent  },
        {".",           "S",            Nothing,        "z"     },
        {"#:.E",        "S",            Nothing,        "z"     },
        {"#:^##",       "S",            Nothing,        "z"     },
        {"#:^#",        "S",            Nothing,        "s"     },
        {"U",           "S",            Nothing,        "s"     },
        {" :#",         "S",            Nothing,        "z"     },
        {Nothing,       "SCH",          Anything,       "sk"    },
        {Anything,      "S",            "C+",           Silent  },
        {"#",           "SM",           Anything,       "zm"    },
        {"#",           "SN",           "'",            "z5n"  },
        {Anything,      "S",            Anything,       "s"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule T_rules[] =
        {
        {Nothing,       "THE",          Nothing,        "D5"  },
        {Anything,      "TO",           Nothing,        "tO"   },
        {Anything,      "THAT",         Nothing,        "Dat" },
        {Nothing,       "THIS",         Nothing,        "Dis" },
        {Nothing,       "THEY",         Anything,       "DA"  },
        {Nothing,       "THERE",        Anything,       "Der" },
        {Anything,      "THER",         Anything,       "D2"  },
        {Anything,      "THEIR",        Anything,       "Der" },
        {Nothing,       "THAN",         Nothing,        "Dan" },
        {Nothing,       "THEM",         Nothing,        "Dem" },
        {Anything,      "THESE",        Nothing,        "DEz" },
        {Nothing,       "THEN",         Anything,       "Den" },
        {Anything,      "THROUGH",      Anything,       "TrO" },
        {Anything,      "THOSE",        Anything,       "Doz" },
        {Anything,      "THOUGH",       Nothing,        "Do"  },
        {Nothing,       "THUS",         Anything,       "Dus" },
        {Anything,      "TH",           Anything,       "T"    },
        {"#:",          "TED",          Nothing,        "tid"  },
        {"S",           "TI",           "#N",           "C"    },
        {Anything,      "TI",           "O",            "S"    },
        {Anything,      "TI",           "A",            "S"    },
        {Anything,      "TIEN",         Anything,       "S5n" },
        {Anything,      "TUR",          "#",            "C2"  },
        {Anything,      "TU",           "A",            "CO"  },
        {Nothing,       "TWO",          Anything,       "tO"   },
        {Anything,      "T",            Anything,       "t"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule U_rules[] =
        {
        {Nothing,       "UN",           "I",            "yOn"  },
        {Nothing,       "UN",           Anything,       "un"   },
        {Nothing,       "UPON",         Anything,       "5p1n"},
        {"T",           "UR",           "#",            "Ur"   },
        {"S",           "UR",           "#",            "Ur"   },
        {"R",           "UR",           "#",            "Ur"   },
        {"D",           "UR",           "#",            "Ur"   },
        {"L",           "UR",           "#",            "Ur"   },
        {"Z",           "UR",           "#",            "Ur"   },
        {"N",           "UR",           "#",            "Ur"   },
        {"J",           "UR",           "#",            "Ur"   },
        {"TH",          "UR",           "#",            "Ur"   },
        {"CH",          "UR",           "#",            "Ur"   },
        {"SH",          "UR",           "#",            "Ur"   },
        {Anything,      "UR",           "#",            "yUr"  },
        {Anything,      "UR",           Anything,       "2"    },
        {Anything,      "U",            "^ ",           "u"    },
        {Anything,      "U",            "^^",           "u"    },
        {Anything,      "UY",           Anything,       "I"    },
        {" G",          "U",            "#",            Silent  },
        {"G",           "U",            "%",            Silent  },
        {"G",           "U",            "#",            "w"     },
        {"#N",          "U",            Anything,       "yO"   },
        {"T",           "U",            Anything,       "O"    },
        {"S",           "U",            Anything,       "O"    },
        {"R",           "U",            Anything,       "O"    },
        {"D",           "U",            Anything,       "O"    },
        {"L",           "U",            Anything,       "O"    },
        {"Z",           "U",            Anything,       "O"    },
        {"N",           "U",            Anything,       "O"    },
        {"J",           "U",            Anything,       "O"    },
        {"TH",          "U",            Anything,       "O"    },
        {"CH",          "U",            Anything,       "O"    },
        {"SH",          "U",            Anything,       "O"    },
        {Anything,      "U",            Anything,       "yO"   },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule V_rules[] =
        {
        {Anything,      "VIEW",         Anything,       "vyO"  },
        {Anything,      "V",            Anything,       "v"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule W_rules[] =
        {
        {Nothing,       "WERE",         Anything,       "w2"   },
        {Anything,      "WA",           "S",            "w4"   },
        {Anything,      "WA",           "T",            "w4"   },
        {Anything,      "WHERE",        Anything,       "Wer" },
        {Anything,      "WHAT",         Anything,       "W4t" },
        {Anything,      "WHOL",         Anything,       "hol"  },
        {Anything,      "WHO",          Anything,       "hO"   },
        {Anything,      "WH",           Anything,       "W"    },
        {Anything,      "WAR",          Anything,       "w1r"  },
        {Anything,      "WOR",          "^",            "w2"   },
        {Anything,      "WR",           Anything,       "r"     },
        {Anything,      "W",            Anything,       "w"     },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule X_rules[] =
        {
        {Anything,      "X",            Anything,       "ks"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule Y_rules[] =
        {
        {Anything,      "YOUNG",        Anything,       "yuN" },
        {Nothing,       "YOU",          Anything,       "yO"   },
        {Nothing,       "YES",          Anything,       "yes"  },
        {Nothing,       "Y",            Anything,       "y"     },
        {"#:^",         "Y",            Nothing,        "E"    },
        {"#:^",         "Y",            "I",            "E"    },
        {" :",          "Y",            Nothing,        "I"    },
        {" :",          "Y",            "#",            "I"    },
        {" :",          "Y",            "^+:#",         "i"    },
        {" :",          "Y",            "^#",           "I"    },
        {Anything,      "Y",            Anything,       "i"    },
        {Anything,      0,              Anything,       Silent  },
        };

/*
**      LEFT_PART       MATCH_PART      RIGHT_PART      OUT_PART
*/
static Rule Z_rules[] =
        {
        {Anything,      "Z",            Anything,       "z"     },
        {Anything,      0,              Anything,       Silent  },
        };

static Rule *Rules[] =
        {
        punct_rules,
        A_rules, B_rules, C_rules, D_rules, E_rules, F_rules, G_rules, 
        H_rules, I_rules, J_rules, K_rules, L_rules, M_rules, N_rules, 
        O_rules, P_rules, Q_rules, R_rules, S_rules, T_rules, U_rules, 
        V_rules, W_rules, X_rules, Y_rules, Z_rules
        };

/*
**	English to Phoneme translation.
**
**	Rules are made up of four parts:
**	
**		The left context.
**		The text to match.
**		The right context.
**		The phonemes to substitute for the matched text.
**
**	Procedure:
**
**		Seperate each block of letters (apostrophes included) 
**		and add a space on each side.  For each unmatched 
**		letter in the word, look through the rules where the 
**		text to match starts with the letter in the word.  If 
**		the text to match is found and the right and left 
**		context patterns also match, output the phonemes for 
**		that rule and skip to the next unmatched letter.
**
**
**	Special Context Symbols:
**
**		#	One or more vowels
**		:	Zero or more consonants
**		^	One consonant.
**		.	One of B, D, V, G, J, L, M, N, R, W or Z (voiced 
**			consonants)
**		%	One of ER, E, ES, ED, ING, ELY (a suffix)
**			(Right context only)
**		+	One of E, I or Y (a "front" vowel)
*/

int find_rule(const char *word, int index, Rule *rules, string&);
int leftmatch(const char *pattern, const char *context);
int rightmatch(const char *pattern, const char *context);

inline int isvowel(char chr)
{
  return (chr == 'A' || chr == 'E' || chr == 'I' || 
	  chr == 'O' || chr == 'U');
}

inline int isconsonant(char chr)
{
  return (isupper(chr) && !isvowel(chr));
}

string to_phoneme(const char *word)
{
  int index;	/* Current position in word */
  int type;	/* First letter of match part */

  string       out;

  index = 0;
  do
    {
      if (isupper(word[index]))
	type = word[index] - 'A' + 1;
      else
	type = 0;

      index = find_rule(word, index, Rules[type], out);
    }
  while (word[index] != '\0');

  return out;
}

int find_rule(const char *word, int index, Rule *rules, string &out)
{
  Rule *rule;
  char *left, *match, *right;
  int remainder;

  for (;;)	/* Search for the rule */
    {
      rule = rules++;
      match = (*rule)[1];

      if (match == 0)	/* bad symbol! */
	{
	  // fprintf(stderr,
	  //  "Error: Can't find rule for: '%c' in \"%s\"\n", word[index], word);
	  return index+1;	/* Skip it! */
	}

      for (remainder = index; *match != '\0'; match++, remainder++)
	{
	  if (*match != word[remainder])
	    break;
	}

      if (*match != '\0')	/* found missmatch */
	continue;
      left = (*rule)[0];
      right = (*rule)[2];

      if (!leftmatch(left, &word[index-1]))
	continue;
      if (!rightmatch(right, &word[remainder]))
	continue;
      out += (*rule)[3];
      return remainder;
    }
}


int leftmatch(const char *pattern,  /* first char of pattern to match in text */
	      const char *context)  /* last char of text to be matched */
{
  const char *pat;
  const char *text;
  int count;

  if (*pattern == '\0')	/* null string matches any context */
    {
      return TRUE;
    }

  /* point to last character in pattern string */
  count = strlen(pattern);
  pat = pattern + (count - 1);

  text = context;

  for (; count > 0; pat--, count--)
    {
      /* First check for simple text or space */
      if (isalpha(*pat) || *pat == '\'' || *pat == ' ')
	if (*pat != *text)
	  return FALSE;
	else
	  {
	    text--;
	    continue;
	  }

      switch (*pat)
	{
	case '#':	/* One or more vowels */
	  if (!isvowel(*text))
	    return FALSE;

	  text--;

	  while (isvowel(*text))
	    text--;
	  break;

	case ':':	/* Zero or more consonants */
	  while (isconsonant(*text))
	    text--;
	  break;

	case '^':	/* One consonant */
	  if (!isconsonant(*text))
	    return FALSE;
	  text--;
	  break;

	case '.':	/* B, D, V, G, J, L, M, N, R, W, Z */
	  if (*text != 'B' && *text != 'D' && *text != 'V'
	      && *text != 'G' && *text != 'J' && *text != 'L'
	      && *text != 'M' && *text != 'N' && *text != 'R'
	      && *text != 'W' && *text != 'Z')
	    return FALSE;
	  text--;
	  break;

	case '+':	/* E, I or Y (front vowel) */
	  if (*text != 'E' && *text != 'I' && *text != 'Y')
	    return FALSE;
	  text--;
	  break;

	case '%':
	default:
	  // fprintf(stderr, "Bad char in left rule: '%c'\n", *pat);
	  return FALSE;
	}
    }

  return TRUE;
}


int rightmatch(const char *pattern, /* first char of pattern to match in text */
	       const char *context) /* last char of text to be matched */
{
  const char *pat;
  const char *text;

  if (*pattern == '\0')	/* null string matches any context */
    return TRUE;

  pat = pattern;
  text = context;

  for (pat = pattern; *pat != '\0'; pat++)
    {
      /* First check for simple text or space */
      if (isalpha(*pat) || *pat == '\'' || *pat == ' ')
	if (*pat != *text)
	  return FALSE;
	else
	  {
	    text++;
	    continue;
	  }

      switch (*pat)
	{
	case '#':	/* One or more vowels */
	  if (!isvowel(*text))
	    return FALSE;

	  text++;

	  while (isvowel(*text))
	    text++;
	  break;

	case ':':	/* Zero or more consonants */
	  while (isconsonant(*text))
	    text++;
	  break;

	case '^':	/* One consonant */
	  if (!isconsonant(*text))
	    return FALSE;
	  text++;
	  break;

	case '.':	/* B, D, V, G, J, L, M, N, R, W, Z */
	  if (*text != 'B' && *text != 'D' && *text != 'V'
	      && *text != 'G' && *text != 'J' && *text != 'L'
	      && *text != 'M' && *text != 'N' && *text != 'R'
	      && *text != 'W' && *text != 'Z')
	    return FALSE;
	  text++;
	  break;

	case '+':	/* E, I or Y (front vowel) */
	  if (*text != 'E' && *text != 'I' && *text != 'Y')
	    return FALSE;
	  text++;
	  break;

	case '%':	/* ER, E, ES, ED, ING, ELY (a suffix) */
	  if (*text == 'E')
	    {
	      text++;
	      if (*text == 'L')
		{
		  text++;
		  if (*text == 'Y')
		    {
		      text++;
		      break;
		    }
		  else
		    {
		      text--; /* Don't gobble L */
		      break;
		    }
		}
	      else
		if (*text == 'R' || *text == 'S' 
		    || *text == 'D')
		  text++;
	      break;
	    }
	  else
	    if (*text == 'I')
	      {
		text++;
		if (*text == 'N')
		  {
		    text++;
		    if (*text == 'G')
		      {
			text++;
			break;
		      }
		  }
		return FALSE;
	      }
	    else
	      return FALSE;

	default:
	  // fprintf(stderr, "Bad char in right rule:'%c'\n", *pat);
	  return FALSE;
	}
    }

  return TRUE;
}
