// Copyright 1998 by Kevin Atkinson under the terms of the LGPL

// suggest.cc Suggestion code for Aspell

// The magic behind my spell checker comes from merging Lawrence
// Philips excellent metaphone algorithm and Ispell's near miss
// strategy which is inserting a space or hyphen, interchanging two
// adjacent letters, changing one letter, deleting a letter, or adding
// a letter.
// 
// The process goes something like this.
// 
// 1.     Convert the misspelled word to its soundslike equivalent (its
//        metaphone for English words).
// 
// 2.     Find words that have the same soundslike pattern.
//
// 3.     Find words that have similar soundslike patterns. A similar
//        soundlike pattern is a pattern that is obtained by
//        interchanging two adjacent letters, changing one letter,
//        deleting a letter, or adding a letter.
//
// 4.     Score the result list and return the words with the lowest
//        score. The score is roughly the weighed average of the edit
//        distance of the word to the misspelled word, the soundslike
//        equivalent of the two words, and the phoneme of the two words.
//        The edit distance is the weighed total of the number of
//        deletions, insertions, exchanges, or adjacent swaps needed to
//        make one string equivalent to the other.
//
// Please note that the soundlike equivalent is a rough approximation
// of how the words sounds. It is not the phoneme of the word by any
// means.  For more information on the metaphone algorithm please see
// the file metaphone.cc which included a detailed description of it.

#include "aspell.hh"
#include "approx.hh"
#include "approx-t.hh"
#include "language.hh"
#include "data.hh"
#include <list>
#include <hash_set>
#include "hash_string.hh"

#include "phoneme.hh"

template <class Iterator>
static Iterator preview_next (Iterator i) {
  return ++i;
}

inline double approx_match_dist(const string &w1, const string &w2,
				const dist_weights &w) {
  return approx_match_dist(w1.c_str(), w2.c_str(), w);
}

inline double approx_match_dist(const string &w1, const_string w2,
				const dist_weights &w) {
  return approx_match_dist(w1.c_str(), w2.c_str(), w);
}

static dist_weights sc_weights;

//
// OrignalWord stores infomation about the orignal misspelled word
//   for convince and speed.
//
struct OrignalWord {
  string   word;
  string   word_lower;
  string   soundslike;
  string   phoneme;
  bool     already_lower;
  int      case_pattern;
  OrignalWord() {}
  OrignalWord (const string &w, const string &sl, const string &p)
    : word(w), soundslike(sl), phoneme(p) {}
  OrignalWord (const string &w, const string &sl, const string &p,
	       const string &l, int cp)
    : word(w), word_lower(l), soundslike(sl), phoneme(p), case_pattern(cp) 
    {already_lower = word==word_lower;}
};

const bool dont_count = false;
const bool do_count = true;

//
// struct ScoreWordSound - used for storing the possible words while
//   there are being processed.
//

struct ScoreWordSound {
  SC_Word       word;
  string        soundslike;
  double        score;
  double        meta_score;
  bool          count;
  const SC_ReplList * repl_list;
  ScoreWordSound() {}
  ScoreWordSound(const SC_Word &w, const string &m) 
    :  word(w), soundslike(m), repl_list(0) {}
  ScoreWordSound(const SC_Word &w, const string &m, double mc, bool c) 
    : word(w), soundslike(m), meta_score(mc), count(c), repl_list(0) {}
  ScoreWordSound(const SC_Word &w, const string &m, 
		 double mc, const SC_ReplList *rl, bool c) 
    : word(w), soundslike(m), meta_score(mc), count(c), repl_list(rl) {}
};

double compare (const ScoreWordSound &lhs, const ScoreWordSound &rhs) {
  double temp = lhs.score - rhs.score;
  if (temp) return temp;
  return compare(lhs.word, rhs.word);
}

inline bool operator < (const ScoreWordSound &lhs, const ScoreWordSound &rhs) {
  return compare(lhs, rhs) < 0;
}

inline bool operator <= (const ScoreWordSound &lhs, const ScoreWordSound &rhs) {
  return compare(lhs, rhs) <= 0;
}

inline bool operator == (const ScoreWordSound &lhs, const ScoreWordSound &rhs) {
  return compare(lhs, rhs) == 0;
}

typedef list<ScoreWordSound> NearMisses;
 
typedef list<string>         Buffer;

class AspellScore {
protected:
  const SC_Language *lang;
  OrignalWord       orignal_word;

public:
  AspellScore(const SC_Language *l, const string &w)
    : lang(l), orignal_word(w, lang->to_soundslike(w.c_str()), 
			    lang->to_phoneme(w.c_str()),
			    lang->to_lower(w.c_str()), lang->case_pattern(w.c_str())) 
  {
    sc_weights.set(.9, // del1
		   .9, // del2
		   .8, // swaps
		   1); // subs
  }
  void score_meta(ScoreWordSound &to_score);
  void score(ScoreWordSound &to_score);
  void score_fast(ScoreWordSound &to_score);
  string fix_case(SC_Word word) {return lang->fix_case(orignal_word.case_pattern,word);}
};

void AspellScore::score_meta(ScoreWordSound &to_score) {
  to_score.meta_score = approx_match_dist(orignal_word.soundslike,
					  to_score.soundslike,
					  sc_weights);
}

void AspellScore::score(ScoreWordSound &to_score) {
  to_score.score = approx_match_dist(orignal_word.word, to_score.word, sc_weights);
  string lower = lang->to_lower(to_score.word);
  if (!orignal_word.already_lower || to_score.word != lower.c_str()) {
    to_score.score *= .33;
    to_score.score += approx_match_dist(orignal_word.word_lower, lower,sc_weights) * .67;
  }
  if (lang->have_phoneme()) {
    to_score.score *= .50;
    to_score.score += 
      approx_match_dist(orignal_word.phoneme,
			lang->to_phoneme(to_score.word),sc_weights)
      * .50;
  }
  if (to_score.meta_score == 0 && orignal_word.word.size() >= 4) { 
    to_score.score *= .45;
  } else {
    to_score.score *= .75;
    to_score.score += to_score.meta_score * .25;
  }
}

class SC_ToLower {
  const SC_Language *lang;
public:
  SC_ToLower(const SC_Language *l) : lang(l) {}
  char operator() (char c) const {return lang->to_lower(c);}
};

void AspellScore::score_fast(ScoreWordSound &to_score) {
  to_score.score = approx_match_dist_fast
    (orignal_word.word_lower.c_str(), static_cast<const char *>(to_score.word), 
					  DoNothing(), SC_ToLower(lang));
  if (to_score.meta_score == 0) { 
    to_score.score *= .45;
  } else {
    to_score.score *= .75;
    to_score.score += to_score.meta_score * .25;
  }
}  

class AspellSuggest : public AspellScore {

  // These thresholds are used relatively.
  static const double threshold_percent = .85; 

  const aspell        *sc;
  NearMisses           near_misses;
  Buffer               buffer;
  aspell::Suggestions *near_misses_return;

  void try_sound(const string&, double ms);
  void try_others();
  void score_list();
  void sort_and_transfer();
public:
  AspellSuggest(const aspell *s, const SC_Language *l,
		const string &w)
    : AspellScore(l,w), sc(s) {}
  void get_suggestions(aspell::Suggestions &sug);
  void get_suggestions_ultra(aspell::Suggestions &sug);
};

//
// try_sound - tries the soundslike string if there is a match add 
//    the possable words to near_misses
//

void AspellSuggest::try_sound (const string &m, double ms)  
{
  // sound is the object in the list which is a lot smaller than m

  {  // scope 1
    AS_WordLists::SWordList words 
                             = sc->word_lists().words_w_soundslike(m.c_str());
    for (AS_WordLists::SWordList::const_iterator i = words.begin();
	 i != words.end();
	 ++i) {
      near_misses.push_front(ScoreWordSound(*i, m, ms, do_count));
    }
  } {// scope 2
    AS_ReplLists::SWordList words 
      = sc->repl_lists().words_w_soundslike(m.c_str());
    if (words.size()) {
      for (AS_ReplLists::SWordList::const_iterator i = words.begin();
	   i != words.end();
	   ++i) {
	near_misses.push_front(ScoreWordSound((*i).misspelled_word().c_str(),m,ms,&(*i),do_count));
      }
    }
  }
}

//
// try_others - tries to come up with possible suggestions
//

void AspellSuggest::try_others () {
  const char *replace_list = lang->soundslike_chars();

  const string &word       = orignal_word.word;
  const string &soundslike = orignal_word.soundslike;

  string::size_type i, j;

  string new_soundslike;
  string &first = new_soundslike;
  string second;

  char a,b;
  const char* c;

  // Insert a space or hyphone

  for (i = 2; i < word.size()-1; ++i) {
    first.resize(0); second.resize(0);
    for (j = 0; j != i; ++j)
      first += word[j];
    for (j = i; j != word.size(); ++j)
      second += word[j];
    if (sc->check(first) && sc->check(second)) {
      buffer.push_front((first + " " + second).c_str());
      buffer.push_front((first + "-" + second).c_str());
      buffer.push_front((lang->to_soundslike(first) +
			 " " +
			 lang->to_soundslike(second)).c_str());
      Buffer::iterator ii = buffer.begin();
      near_misses.push_front(ScoreWordSound((*++ii).c_str(), 
					    buffer.front().c_str(), 1, dont_count));
      near_misses.push_front(ScoreWordSound((*++ii).c_str(), 
					    buffer.front().c_str(), 1, dont_count));
    }
  }

  try_sound(soundslike, 0);

  // Change one letter

  for (i = 0; i != soundslike.size(); ++i) {
    for (c=replace_list; *c; ++c) {
      if (*c == soundslike[i]) continue;
      new_soundslike = soundslike;
      new_soundslike[i] = *c;
      try_sound(new_soundslike,sc_weights.subs);
    }
  }

  // Delete one letter

  for (i = 0; i != soundslike.size(); ++i) {
    try_sound(string(soundslike).erase(i,1),sc_weights.del2);
  }

  // Interchange two adjacent letters.

  for (i = 0; i+1 != soundslike.size(); ++i) {
    new_soundslike = soundslike;
    a = new_soundslike[i];
    b = new_soundslike[i+1];
    new_soundslike[i] = b;
    new_soundslike[i+1] = a;
    try_sound(new_soundslike,sc_weights.swaps);
  }

  // Add one letter

  for (i = 0; i <= soundslike.size(); ++i) {
    for (c=replace_list; *c; ++c) {
      new_soundslike = soundslike;
      new_soundslike.insert(i, 1, *c);
      try_sound(new_soundslike,sc_weights.del1);
    }
  }
}

void AspellSuggest::score_list() {
  double min = 100;
  for (NearMisses::iterator i = near_misses.begin(); 
       i != near_misses.end();
       ++i) {
    score_fast(*i);
    if ((*i).score < min) min = (*i).score;
  }
  double threshold = orignal_word.word.size()*(1-threshold_percent) 
                     + min*threshold_percent;
  if (threshold < 1) threshold = 1;
  ++threshold;
  min = 100;
  for (NearMisses::iterator i = near_misses.begin(); 
       i != near_misses.end();
       ++i) {
    if ((*i).score > threshold) {
      (*i).score *= 1.25;
    } else {
      score(*i);
    }
    if ((*i).score < min) min = (*i).score;
  }
}

//
// sort_and_transfer
//

void AspellSuggest::sort_and_transfer() {
  //
  // now put the results into a vector and return that vector
  //
  if (near_misses.size()) {

    near_misses.sort();

    NearMisses::iterator ii = near_misses.begin();
    
    for (;!(*ii).count && preview_next(ii) != near_misses.end(); ++ii); 
                             // skip over ones that should't be counted
                             // in the threshold score.

    for(NearMisses::iterator jj = near_misses.begin(); 
	jj != near_misses.end() && !(*jj).count; 
	++jj)
      (*jj).score = (*ii).score * 1.05;

    if (!near_misses.front().count)
      // FIXME: Find a better way
      near_misses.sort();


    // length - (length-score)*threshold
    double threshold = orignal_word.word.size()*(1-threshold_percent) 
                                  + (*ii).score*threshold_percent;
    if (threshold < 1) threshold = 1;

    int c = 1;
    hash_set<string, hash_string<string> > duplicates_check;
    pair<hash_set<string, hash_string<string> >::iterator, bool> dup_pair;
    for (NearMisses::const_iterator i = near_misses.begin();
	 i != near_misses.end() && ( (*i).score <= threshold || c <= 3 );
	 ++i, ++c) {
      //cout << (*i).word << "  " << (*i).score 
      //     << "  " << (*i).soundslike 
      //     << "  " << lang->to_phoneme((*i).word) << endl;
      if (i->repl_list) {
	for (SC_ReplList::iterator j = i->repl_list->begin();
	     j != i->repl_list->end(); ++j) {
	  dup_pair = duplicates_check.insert(fix_case((*j).c_str()));
	  if (dup_pair.second)
	    near_misses_return->push_back(*dup_pair.first);
	}
      } else {
	dup_pair = duplicates_check.insert(fix_case(i->word));
	if (dup_pair.second)
	  near_misses_return->push_back(*dup_pair.first);
      }
    }
  }
}

void AspellSuggest::get_suggestions(aspell::Suggestions &sug) {
  near_misses_return = &sug;
  if (!orignal_word.soundslike.size()) return;
  try_others();
  score_list();
  sort_and_transfer();
}

void AspellSuggest::get_suggestions_ultra(aspell::Suggestions &sug) {
}

//
// Score Method
//

double aspell::score(const char *base, const char *other) const {
  AspellScore s(lang,base);
  string sl = lang->to_soundslike(other);
  ScoreWordSound sws(other, sl.c_str());
  s.score_meta(sws);
  s.score(sws);
  unsigned int len = 0;
  for (;base[len]; ++len);
  return (len-sws.score)/len;
}

//
// Finally the actual suggest methods!
//

aspell::Suggestions aspell::suggest(const string &word) const {
  Suggestions near_misses_return;
  AspellSuggest sug(this,lang,word);
  sug.get_suggestions(near_misses_return);
  return near_misses_return;
};

//
// This method basically tried every possible word and scores it
//

const aspell::Suggestions aspell::suggest_ultra(const string &word) const {
  Suggestions near_misses_return;
  AspellSuggest sug(this,lang,word);
  sug.get_suggestions_ultra(near_misses_return);
  return near_misses_return;
}
