/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Alex Pfaffe

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "system/sysprint.h"
#include "collp.h"
#include "collider.h"

int CD_num_cols_alloced = 0;
collision_pair *CD_contact = 0;

#define CD_MAX_COLLISION    1000

static int hits = 0;
// Array of hits.
static Collider* hitv[CD_MAX_COLLISION][2];
static int currHit;

///
Matrix3 Collider::mR;
Vector3 Collider::mT;
///
int Collider::trianglesTested = 0;
int Collider::boxesTested = 0;
int Collider::testLevel = 0;
bool Collider::firstHit = true;
int  Collider::numHits = 0;
Vector3 Collider::hitNormal;
float Collider::minBBoxDiam = 0.0;
///
Collider* Collider::_player = 0;
bool Collider::falling = false;
bool Collider::climbing = false;
bool Collider::init = false;
csWorld* Collider::_world = 0;

Moment *Moment::stack = 0;

Collider::Collider (PolygonSet *ps)
{
  _type = POLYGONSET;
  _ps = ps;
  _rm = NULL;
  bool init = false;

  int i, v;
  for (i = 0 ; i < _ps->num_polygon ; i++)
      {
      Polygon3D* p = (Polygon3D*)_ps->polygons[i];
      Portal* portal = p->get_portal();
      // Handle solid walls and mirrors.
      if (portal == NULL || portal->is_space_warped())
        {
        if (!init)
          {
          init = true;
          CHK (_rm = new CD_model);
          _rm->b = 0;
          _rm->num_boxes_alloced = 0;

          _rm->tris = 0;
          _rm->num_tris = 0;
          _rm->num_tris_alloced = 0;

          if (!_rm)
            return;  // Error
          } 
      
        // Collision detection only works with triangles.
        int *vt = p->get_vertices_idx();
        for (v = 2; v < p->get_num_vertices(); v++)
          {
	  _rm->addTriangle(_ps->wor_verts[vt[v-1]],_ps->wor_verts[vt[v]],_ps->wor_verts[vt[0]],v-2);
          }
        } // Is not portal
      } // For num_polygon

  if (init)
    {
    _rm->build_hierarchy();
    }
  // Set to be on initially.
  _cd_state = true; 
}

Collider::Collider( csSprite3D *sp )
{
  _type = SPRITE3D;
  _sp = sp;
  CHK (_rm = new CD_model);
  if (!_rm)
     return;  // Error
  _rm->b = 0;
  _rm->num_boxes_alloced = 0;

  _rm->tris = 0;
  _rm->num_tris = 0;
  _rm->num_tris_alloced = 0;

  // pick a frame for now.  This should of course be done for the
  // correct frame (some global CD outline or else we need to do it
  // for every frame which is too expensive.
  csFrame *cf = _sp->cur_action->get_frame(_sp->cur_frame);
  for (int i = 0; i < _sp->tpl->num_triangles; i++)
    {
    int v[3];
    v[0] = _sp->tpl->triangles[i].a;
    v[1] = _sp->tpl->triangles[i].b;
    v[2] = _sp->tpl->triangles[i].c;
    _rm->addTriangle(cf->get_vertice(v[0]),cf->get_vertice(v[1]),cf->get_vertice(v[2]),i);
    }

  _rm->build_hierarchy();
  _cd_state = true;
}

Collider::~Collider(void)
{
  if (_rm)
  {
    // the boxes pointed to should be deleted.
    CHK (delete [] _rm->b);

    // the triangles pointed to should be deleted.
    CHK (delete [] _rm->tris);

    CHK (delete _rm);
    _rm = NULL;
  }
}

// Return object's name.
char *Collider::get_name(void)
{
  if (_type == POLYGONSET) return _ps->get_name();
  if (_type == SPRITE3D)   return _sp->get_name();
  return "UNKNOWN";
}

void Collider::activate(bool on)
{
  _cd_state = on ? 1 : 0;
}

int Collider::collide_pair (Collider *p1, Collider *p2, Transform *t1, Transform *t2)
{
  // Skip inactive combinations.
  if (!p1->_cd_state || !p2->_cd_state) return 0;
  
  //call the low level collision detection routine.
  Collider::firstHit = true;

  BBox *b1 = p1->_rm->b;
  BBox *b2 = p2->_rm->b;
  
  Matrix3 tR1, tR2, R, R1, R2;
  Vector3 tT1, tT2, T, T1, T2;
  if (t1)
    {
    T1 = t1->v_world2this();
    R1 = t1->m_world2this();
    }
  if (t2)
    {
    R2 = t2->m_world2this();
    T2 = t2->v_world2this();
    }

  // [R1,T1] and [R2,T2] are how the two triangle sets
  // (i.e. models) are positioned in world space.  [tR1,tT1] and
  // [tR2,tT2] are how the top level boxes are positioned in world
  // space
  
  tR1 = R1 * b1->pR;
  tT1 = (R1 * b1->pT) + T1;
  tR2 = R2 * b2->pR;
  tT2 = (R2 * b2->pT) + T2;
  
  // [R,T] is the placement of b2's top level box within
  // the coordinate system of b1's top level box.

  R = tR1.gettranspose() * tR2;
  T = tR1.gettranspose() * (tT2 - tT1);

  // To transform tri's from model1's CS to model2's CS use this:
  //    x2 = mR . x1 + mT

  Collider::mR = R2.gettranspose() * R1;
  Collider::mT = R2.gettranspose() * (T1 - T2);

  // reset the report fields
  Collider::numHits = 0;
  Collider::trianglesTested = 0;
  Collider::boxesTested = 0;
  // make the call
  if (collide_recursive(b1, b2, R, T))
    {
      // Error.
    }
  else if (Collider::numHits != 0)
    {
    hitv[hits][0] = p1;
    hitv[hits][1] = p2;
    hits++;
    return 1;
    }
  return 0;
}

#define fastabs(x) ((x<0) ? -x : x)
inline float minval(float a, float b, float c)
{ return (a<b?(a<c?a:(c<b?c:b)):(b<c?b:c)); }
inline float maxval(float a, float b, float c)
{ return (a>b?(a>c?a:(c>b?c:b)):(b>c?b:c)); }

int project6(Vector3 ax, Vector3 p1, Vector3 p2, Vector3 p3,
             Vector3 q1, Vector3 q2, Vector3 q3)
{
   float P1 = ax * p1;
   float P2 = ax * p2;
   float P3 = ax * p3;
   float Q1 = ax * q1;
   float Q2 = ax * q2;
   float Q3 = ax * q3;

   float mx1 = maxval(P1,P2,P3);
   float mn1 = minval(P1,P2,P3);
   float mx2 = maxval(Q1,Q2,Q3);
   float mn2 = minval(Q1,Q2,Q3);
   
   if (mn1 > mx2) return 0;
   if (mn2 > mx1) return 0;
   return 1;
}

// very robust triangle intersection test
// uses no divisions
// works on coplanar triangles

bool tri_contact (Vector3 P1, Vector3 P2, Vector3 P3,
		  Vector3 Q1, Vector3 Q2, Vector3 Q3) 
{

  /*
     One triangle is (p1,p2,p3).  Other is (q1,q2,q3).
     Edges are (e1,e2,e3) and (f1,f2,f3).
     Normals are n1 and m1
     Outwards are (g1,g2,g3) and (h1,h2,h3).

     We assume that the triangle vertices are in the same coordinate system.

     First thing we do is establish a new c.s. so that p1 is at (0,0,0).
     */
  
  Vector3 p1 = P1 - P2;
  Vector3 p2 = P2 - P1;
  Vector3 p3 = P3 - P1;
  
  Vector3 q1 = Q1 - P1;
  Vector3 q2 = Q2 - P1;
  Vector3 q3 = Q3 - P1;
  
  Vector3 e1 = p2 - p1;
  Vector3 e2 = p3 - p2;
  Vector3 e3 = p1 - p3;

  Vector3 f1 = q2 - q1;
  Vector3 f2 = q3 - q2;
  Vector3 f3 = q1 - q3;
  
  Vector3 n1 = e1 % e2;
  Vector3 m1 = f1 % f2;

  Vector3 g1 = e1 % n1;
  Vector3 g2 = e2 % n1;
  Vector3 g3 = e3 % n1;
  Vector3 h1 = f1 % m1;
  Vector3 h2 = f2 % m1;
  Vector3 h3 = f3 % m1;

  Vector3 ef11 = e1 % f1;
  Vector3 ef12 = e1 % f2;
  Vector3 ef13 = e1 % f3;

  Vector3 ef21 = e2 % f1;
  Vector3 ef22 = e2 % f2;
  Vector3 ef23 = e2 % f3;

  Vector3 ef31 = e3 % f1;
  Vector3 ef32 = e3 % f2;
  Vector3 ef33 = e3 % f3;

  // now begin the series of tests

  if (!project6(n1, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(m1, p1, p2, p3, q1, q2, q3)) return false;
  
  if (!project6(ef11, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef12, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef13, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef21, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef22, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef23, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef31, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef32, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(ef33, p1, p2, p3, q1, q2, q3)) return false;

  if (!project6(g1, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(g2, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(g3, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(h1, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(h2, p1, p2, p3, q1, q2, q3)) return false;
  if (!project6(h3, p1, p2, p3, q1, q2, q3)) return false;

  return true;
}

int add_collision(int id1, int id2)
{
  if (!CD_contact)
    {
      CHK (CD_contact = new collision_pair[10]);
      if (!CD_contact) 
	{
	  return true;
	}
      CD_num_cols_alloced = 10;
      Collider::numHits = 0;
    }
  
  if (Collider::numHits == CD_num_cols_alloced)
    {
      CHK (collision_pair *t = new collision_pair[CD_num_cols_alloced*2]);
      if (!t)
	{
	  return true;
	}
      CD_num_cols_alloced *= 2;
      
      for(int i=0; i<Collider::numHits; i++) t[i] = CD_contact[i];
      CHK (delete [] CD_contact);
      CD_contact = t;
    }
  
  CD_contact[Collider::numHits].id1 = id1;
  CD_contact[Collider::numHits].id2 = id2;
  Collider::numHits++;

  return false;
}

int obb_disjoint(Matrix3 B, Vector3 T, Vector3 a, Vector3 b)
{
  register float t, s;
  register int r;
  Matrix3 Bf;
  const float reps = 1e-6;
  
  // Bf = fabs(B)
  Bf.m11 = fastabs(B.m11);  Bf.m11 += reps;
  Bf.m12 = fastabs(B.m12);  Bf.m12 += reps;
  Bf.m13 = fastabs(B.m13);  Bf.m13 += reps;
  Bf.m21 = fastabs(B.m21);  Bf.m21 += reps;
  Bf.m22 = fastabs(B.m22);  Bf.m22 += reps;
  Bf.m23 = fastabs(B.m23);  Bf.m23 += reps;
  Bf.m31 = fastabs(B.m31);  Bf.m31 += reps;
  Bf.m32 = fastabs(B.m32);  Bf.m32 += reps;
  Bf.m33 = fastabs(B.m33);  Bf.m33 += reps;
  
  // if any of these tests are one-sided, then the polyhedra are disjoint
  r = 1;

  // A1 x A2 = A0
  t = fastabs(T.x);
  
  r &= (t <= 
	  (a.x + b.x * Bf.m11 + b.y * Bf.m12 + b.z * Bf.m13));
  if (!r) return 1;
  
  // B1 x B2 = B0
  s = T.x*B.m11 + T.y*B.m21 + T.z*B.m31;
  t = fastabs(s);

  r &= ( t <=
	  (b.x + a.y * Bf.m11 + a.y * Bf.m21 + a.z * Bf.m31));
  if (!r) return 2;
    
  // A2 x A0 = A1
  t = fastabs(T.y);
  
  r &= ( t <= 
	  (a.y + b.x * Bf.m21 + b.y * Bf.m22 + b.z * Bf.m23));
  if (!r) return 3;

  // A0 x A1 = A2
  t = fastabs(T.z);

  r &= ( t <= 
	  (a.z + b.x * Bf.m31 + b.y * Bf.m32 + b.z * Bf.m33));
  if (!r) return 4;

  // B2 x B0 = B1
  s = T.x*B.m12 + T.y*B.m22 + T.z*B.m32;
  t = fastabs(s);

  r &= ( t <=
	  (b.y + a.x * Bf.m12 + a.y * Bf.m22 + a.z * Bf.m32));
  if (!r) return 5;

  // B0 x B1 = B2
  s = T.x*B.m13 + T.y*B.m23 + T.z*B.m33;
  t = fastabs(s);

  r &= ( t <=
	  (b.z + a.x * Bf.m13 + a.y * Bf.m23 + a.z * Bf.m33));
  if (!r) return 6;

  // A0 x B0
  s = T.z * B.m21 - T.y * B.m31;
  t = fastabs(s);
  
  r &= ( t <= 
	(a.y * Bf.m31 + a.z * Bf.m21 +
	 b.y * Bf.m13 + b.z * Bf.m12));
  if (!r) return 7;
  
  // A0 x B1
  s = T.z * B.m22 - T.y * B.m32;
  t = fastabs(s);

  r &= ( t <=
	(a.y * Bf.m32 + a.z * Bf.m22 +
	 b.x * Bf.m13 + b.z * Bf.m11));
  if (!r) return 8;

  // A0 x B2
  s = T.z * B.m23 - T.y * B.m33;
  t = fastabs(s);

  r &= ( t <=
	  (a.y * Bf.m33 + a.z * Bf.m23 +
	   b.x * Bf.m12 + b.y * Bf.m11));
  if (!r) return 9;

  // A1 x B0
  s = T.x * B.m31 - T.z * B.m11;
  t = fastabs(s);

  r &= ( t <=
	  (a.x * Bf.m31 + a.z * Bf.m11 +
	   b.y * Bf.m23 + b.z * Bf.m22));
  if (!r) return 10;

  // A1 x B1
  s = T.x * B.m32 - T.z * B.m12;
  t = fastabs(s);

  r &= ( t <=
	  (a.x * Bf.m32 + a.z * Bf.m12 +
	   b.x * Bf.m23 + b.z * Bf.m21));
  if (!r) return 11;

  // A1 x B2
  s = T.x * B.m33 - T.z * B.m13;
  t = fastabs(s);

  r &= (t <=
	  (a.x * Bf.m33 + a.z * Bf.m13 +
	   b.x * Bf.m22 + b.y * Bf.m21));
  if (!r) return 12;

  // A2 x B0
  s = T.y * B.m11 - T.x * B.m21;
  t = fastabs(s);

  r &= (t <=
	  (a.x * Bf.m21 + a.y * Bf.m11 +
	   b.y * Bf.m33 + b.z * Bf.m32));
  if (!r) return 13;

  // A2 x B1
  s = T.y * B.m12 - T.x * B.m22;
  t = fastabs(s);

  r &= ( t <=
	  (a.x * Bf.m22 + a.y * Bf.m12 +
	   b.x * Bf.m33 + b.z * Bf.m31));
  if (!r) return 14;

  // A2 x B2
  s = T.y * B.m13 - T.x * B.m23;
  t = fastabs(s);

  r &= ( t <=
	  (a.x * Bf.m23 + a.y * Bf.m13 +
	   b.x * Bf.m32 + b.y * Bf.m31));
  if (!r) return 15;

  return 0;  // should equal 0
}


int Collider::collide_recursive(BBox *b1, BBox *b2, Matrix3 R, Vector3 T)
{

  int rc;      // return codes
  
  if (Collider::firstHit && (Collider::numHits > 0)) return false;

  // test top level

  Collider::boxesTested++;

  int f1;
  
  f1 = obb_disjoint(R, T, b1->d, b2->d);

  if (f1 != 0) 
	{
	  return false;  // stop processing this test, go to top of loop
	}

      // contact between boxes
      if (b1->leaf() && b2->leaf()) 
	{
	  // it is a leaf pair - compare the polygons therein
          // tri_contact uses the model-to-model transforms stored in
	  // Collider::mR, Collider::mT.

	  // this will pass along any OUT_OF_MEMORY return codes which
	  // may be generated.
	  return BBox::tri_contact(b1, b2);
	}

      Matrix3 cR;
      Vector3 cT;
      
      // Currently, the transform from model 2 to model 1 space is
      // given by [B T], where y = [B T].x = B.x + T.

      if (b2->leaf() || (!b1->leaf() && (b1->size() > b2->size())))
	{
	  // here we descend to children of b1.  The transform from
	  // a child of b1 to b1 is stored in [b1->N->pR,b1->N->pT],
	  // but we will denote it [B1 T1] for short.

	  // Here, we compute [B1 T1]'[B T] = [B1'B B1'(T-T1)]
	  // for each child, and store the transform into the collision
	  // test queue.

          cR = b1->N->pR.gettranspose() * R;
          cT = b1->N->pR.gettranspose() * (T - b1->N->pT);
 
	  if ((rc = collide_recursive(b1->N, b2, cR, cT)) != false)
	    return rc;
	  
          cR = b1->P->pR.gettranspose() * R;
          cT = b1->P->pR.gettranspose() * (T - b1->P->pT);

	  if ((rc = collide_recursive(b1->P, b2, cR, cT)) != false)
	    return rc;
	  
	  return false;
	}
      else 
	{
	  // here we descend to the children of b2.  See comments for
	  // other 'if' clause for explanation.

          cR = R * b2->N->pR;
          cT = ( R * b2->N->pT) + T;
	  
	  if ((rc = collide_recursive(b1, b2->N, cR, cT)) != false)
	    return rc;
	  
          cR = R * b2->P->pR;
          cT = ( R * b2->P->pT) + T;

	  if ((rc = collide_recursive(b1, b2->P, cR, cT)) != false)
	    return rc;
	  
	  return false; 
	}
  
  return false;
}

void Collider::collide_reset(void)
{
  hits = 0;
  currHit = 0;
}

int Collider::report(Collider **id1, Collider **id2)
{
  if (currHit >= hits) return 0;

  *id1 = hitv[currHit][0];
  *id2 = hitv[currHit][1];
  currHit++;
  return 1;
}

Collider* Collider::find_collision(void)
{
  int i = 0;
  while (i < hits)
    if      (hitv[i][0] == this) return hitv[i][1];
    else if (hitv[i][1] == this) return hitv[i][0];
    else i++;
  return 0;
}

Collider* Collider::player(char *name)
{
  // Define the player bounding box.
  // The camera's lens or person's eye is assumed to be
  // at 0,0,0.  The height (DY), width (DX) and depth (DZ).
  // Is the size of the camera/person and the origin
  // coordinates (OX,OY,OZ) locate the bbox w.r.t. the eye.
  // This player is 1.8 metres tall (assuming 1cs unit = 1m) (6feet)
  #define DX    0.5
  #define DY    1.8
  #define DZ    0.3
  #define OX   -0.25
  #define OY   -1.1
  #define OZ    0.0
  
  Polygon3D *p;
  CHK (PolygonSet* player = new PolygonSet(name,CS_THING));

  player->add_vertex(OX,    OY,    OZ); 
  player->add_vertex(OX,    OY,    OZ+DZ); 
  player->add_vertex(OX,    OY+DY, OZ+DZ); 
  player->add_vertex(OX,    OY+DY, OZ); 
  player->add_vertex(OX+DX, OY,    OZ); 
  player->add_vertex(OX+DX, OY,    OZ+DZ); 
  player->add_vertex(OX+DX, OY+DY, OZ+DZ); 
  player->add_vertex(OX+DX, OY+DY, OZ); 

  // Left
  p = player->new_polygon ("",0);

  p->add_vertex(0); p->add_vertex(1); 
  p->add_vertex(2); p->add_vertex(3); 

  // Right
  p = player->new_polygon ("",0);
  p->add_vertex(4); p->add_vertex(5); 
  p->add_vertex(6); p->add_vertex(7); 

  // Front
  p = player->new_polygon ("",0);
  p->add_vertex(0); p->add_vertex(1); 
  p->add_vertex(5); p->add_vertex(4); 

  // Back
  p = player->new_polygon ("",0);
  p->add_vertex(3); p->add_vertex(2); 
  p->add_vertex(6); p->add_vertex(7); 

  // Top
  p = player->new_polygon ("",0);
  p->add_vertex(1); p->add_vertex(5); 
  p->add_vertex(6); p->add_vertex(2); 

  // Bottom
  p = player->new_polygon ("",0);
  p->add_vertex(0); p->add_vertex(4); 
  p->add_vertex(7); p->add_vertex(3); 

  CHK (Collider* c = new Collider(player));

  return c;
}

int Collider::init_world (csWorld* world)
{
  _world = world;
  Thing* tp;
  CsPrintf (MSG_INITIALIZATION, "Computing OBBs ...\n");

  int sn = world->sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* sp = (csSector*)world->sectors[sn];
    // Initialize the sector itself.
    CHK (sp->collider.cp = new Collider(sp));
    // Initialize the things in this sector.
    Thing* tp = sp->get_first_thing();
    while (tp)
    {
      CHK (tp->collider.cp = new Collider(tp));
      tp = (Thing*)(tp->get_next ());
    }
  }
  // Initialize all multisector objects that support
  // collision detection.
  MultiSectorObject* msop = world->get_first_mso();
  csObject*  op;
  csSprite3D* spp;

  while(msop)
    {
    op = msop->get_object();
    switch (op->get_type())
      {
      case CS_SPRITE3D:
        spp = (csSprite3D*) op;
        CHK (spp->collider.cp = new Collider(spp));
        break;
      case CS_THING:
        tp = (Thing *) op;
        CHK (tp->collider.cp = new Collider(tp));
        break;
      default:
        fprintf(stderr,"No CD built for unknown object %s\n",op->get_name());
        break;
      }
    msop = (MultiSectorObject*)(msop->get_next());
    }
  // Create a player object that follows the camera around.
  _player = Collider::player("Player");

  CsPrintf (MSG_INITIALIZATION, "DONE\n");
  return 0; // Ok.
}

void Collider::end_world(void)
{
  CHK (delete _player);
}

// TODO.  This is probably not the best way
static csCamera* _camera;
 
void Collider::set_camera(csCamera* camera) { _camera = camera; }
// ------------------------------------------------------------

int Collider::_collision_detect (csSector* sp, Transform *cdt)
{
  int hit = 0;
  // Check collision with this sector.
  hit += collide_pair(this,sp->collider.cp,cdt);

  // Check collision of with the things in this sector.
  Thing* tp = sp->get_first_thing();
  while (tp)
  {
    // TODO, if and when Things can move, their transform must be passed in.
    hit += collide_pair(this,tp->collider.cp,cdt);
    tp = (Thing*)(tp->get_next ());
    // TODO, should test which one is the closest.
  }

  // Check collision of with the MultiSectorObjects in this sector.
  SectorMSO* mso = sp->get_first_mso();
  csSprite3D* sp3d = 0;
  csObject* op = 0;
  Transform cds;

  while(mso)
    {
    op = mso->get_mso()->get_object();
    switch (op->get_type())
      {
      case CS_SPRITE3D:
        sp3d = (csSprite3D*) op;
        cds.set_world2this(sp3d->m_world2this());
        cds.set_origin(sp3d->v_world2this());
        if (sp3d->collider.cp) hit+= collide_pair(this,sp3d->collider.cp,cdt,&cds);
        break;
      case CS_THING:
        tp = (Thing *) op;
        if (tp->collider.cp) hit+= collide_pair(this,tp->collider.cp,cdt);
        break;
      default:
        break;
      }
    mso = mso->get_next_sector();
    }

  return hit;
}

int Collider::collision_detect( csSector* s, Transform *cdt, Collider *cd )
{
  // Cache the ground collider for next time.  If 0, we are falling.
  static Collider *ground = 0;
  static bool init    = false;
  // Previous sector.
  static csSector   *sold = 0;
  // Transformation matrix from world space to local space.
  static Matrix3 mold;
  // Location of the origin for the local transformation.
  static Vector3 vold;

  int hit = 0;
  collide_reset();

  if (!cd) cd = _player;

  if (!init)
    {
    init = true;
    mold = _camera->m_world2this();
    vold = _camera->v_world2this();
    sold = s;
    }

  
  Transform cdtnew(_camera->m_world2this(),_camera->v_world2this());
  // See if we are on solid ground.
  // First test against the current ground if there is one.
  // If there isn't one, test the sector.
  cdtnew.set_origin(_camera->v_world2this()+(0.2*VEC_DOWN));

  if (s!=sold||(ground&&collide_pair(cd,ground, &cdtnew))||cd->_collision_detect (s, &cdtnew))
  {
    falling = false;
    climbing = false;
    // TODO: Check which polygon we hit and what its normal is w.r.t. gravity,
    // If the angle more than 45* we should slip/bounce.
    //
    // We are on something solid, remember what the ground was.
    ground = cd->find_collision();
    // printf("Ground %s\n",ground->get_name());
    // Perform forward collision of player with this sector.
    collide_reset();
    hit = cd->_collision_detect (s, cdt);

    // Hack for now assumes constant frame rate.
    if (hit) // Return to previous position.
    {
      // Get the object we hit.
      Collider *cdhit = cd->find_collision();
      
      // Try to see if we can climb up.
      Transform cdtnew(_camera->m_world2this(),_camera->v_world2this()+(0.2*VEC_UP));
      
      // See if we still hit.  If so we are stuck, if not we may try to climb but
      // we need to retest the sector since we may still hit something else.
      if (!collide_pair(cd,cdhit, &cdtnew) && !cd->_collision_detect(s, &cdtnew))
      {
          collide_reset();
          climbing = true;
          _camera->set_position(cdtnew.v_world2this());
      }
      else // Try a steeper climb.
      {
        cdtnew.set_origin(_camera->v_world2this()+(0.4*VEC_UP));

        if (!collide_pair(cd,cdhit, &cdtnew) && !cd->_collision_detect(s, &cdtnew))
	  {
          collide_reset();
          climbing = true;
          _camera->set_position(cdtnew.v_world2this());
	  }
        else // Stay where we were, we definitly hit something.
        {
          climbing = false;
          _camera->set_position(vold);
          _camera->set_world2cam(mold);
          _camera->set_sector(sold);
        }
      }
    }
  }
  else // We are falling or descending.
  {
     // Assume constant rate.
     _camera->set_position(vold+(0.2*VEC_DOWN));
     _camera->set_sector(sold);
     ground = 0;
     falling = true;
     climbing = false;
  }
  // Store these locations for next frame.
  sold = s;
  mold = _camera->m_world2this();
  vold = _camera->v_world2this();
  return hit;
}

bool BBox::tri_contact(BBox *b1, BBox *b2)
{
  // assume just one triangle in each box.

  // the vertices of the CDTriangle in b2 is in model1 C.S.  The vertices of
  // the other triangle is in model2 CS.  Use Collider::mR, Collider::mT, and
  // to transform into model2 CS.

  int rc;  // return code

  Vector3 i1 = (( Collider::mR * b1->trp->p1) + Collider::mT);
  Vector3 i2 = (( Collider::mR * b1->trp->p2) + Collider::mT);
  Vector3 i3 = (( Collider::mR * b1->trp->p3) + Collider::mT);

  Collider::trianglesTested++;

  bool f = ::tri_contact(i1, i2, i3, b2->trp->p1, b2->trp->p2, b2->trp->p3);

  if (f) 
    {
      // add_collision may be unable to allocate enough memory,
      // so be prepared to pass along an OUT_OF_MEMORY return code.
      if ((rc = add_collision(b1->trp->id, b2->trp->id)) != false)
	return rc;
    }
  
  return false;
}
