/*
    Crystal Space COM Emulation Support
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Dan Ogles
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "com.h"
#include "util/memory.h"


#ifndef CS_STATIC_LINKED

// used to dynamically load a library.
extern CS_HLIBRARY csLoadLibrary (char* szLibName);
extern PROC csGetProcAddress (CS_HLIBRARY hLib, char* szProcName);
extern void csFreeAllLibraries ();

#endif // CS_STATIC_LINKED


#if defined(NO_COM_SUPPORT)

/**
 *  RegistryEntry</p>
 *
 *  This is how csCoCreateInstance() determines what DLL to load, or if we
 *  are working statically linked, which singleton class object that we
 *  are linked to. The CLSID is the globally unique identifier for the
 *  class object. The ProgID is a human-readable version of the CLSID which
 *  can also map to the DLL name or the singleton object (if we are statically
 *  linked).
 */
struct RegistryEntry
{
  /// the ProgID of the CoClass.
  char szProgID[128];    

#if defined(CS_STATIC_LINKED)
  /// a pointer to the class ID.
  const CLSID *clsid;
  /// a pointer to the singleton CoClass.
  IUnknown* pClass;
#else
  /// the CLSID of the CoClass.
  CLSID clsid;
  /// the InProc server name.
  char szDllName[256];     
#endif
};


#if defined(CS_STATIC_LINKED)
//------------------------------- Static linking emulation for non-COM platforms -
#pragma message ("Using COM static linking emulation.")

IID IID_IUnknown =
{ 0x00000000, 0x0000, 0x0000, { 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 } };

IID IID_IClassFactory =
{ 0x00000001, 0x0000, 0x0000, { 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 } };

#include "render/graph3d.h"
static csGraphics3DSoftwareFactory g3dFactorySoft;

// Seems that there is no easy alternative to these ... :-(
#if defined (DISP_X11)
#  include "system/xlib/driver2d/xlibg2d.h"
   extern const CLSID CLSID_Graphics2DXLib;
   static csGraphics2DXLibFactory g2dFactoryXlib;
#elif defined (DISP_DIVE)
#  include "system/os2/csdive/csdive.h"
   static csGraphics2DFactoryOS2DIVE g2dFactoryOS2DIVE;
#elif defined (DISP_RAW)
#  include "system/djgpp/raw/raw.h"
   static csGraphics2DFactoryDOSRAW g2dFactoryDOSRAW;
#endif
  
/**
 *  This is the array that emulates the COM registry when
 *  Crystal Space is statically linked to all components.
 *  All known CLSIDs are hard-coded into the executable through
 *  this array. The array must end with a NULL entry.
 */
const RegistryEntry StaticRegistry[] =
{
  { "crystalspace.graphics3d.software", &CLSID_Graphics3DSoftware, &g3dFactorySoft },
#ifdef DISP_X11
  { "crystalspace.graphics2d.default", &CLSID_Graphics2DXLib, &g2dFactoryXlib },
#endif
#ifdef DISP_DIVE
  { "crystalspace.graphics2d.default", &CLSID_csGraphics2DOS2DIVE, &g2dFactoryOS2DIVE },
#endif
#ifdef DISP_RAW
  { "crystalspace.graphics2d.default", &CLSID_csGraphics2DDOSRAW, &g2dFactoryDOSRAW },
#endif
  { "", NULL, NULL }
};

// csRegisterServer () 
HRESULT csRegisterServer (const DllRegisterData* dllRegData)
{
  (void)dllRegData;
  // this should already be registered (we are statically linked)
  return S_OK;
}

STDAPI csUnregisterServer (const DllRegisterData* dllRegData)
{
  (void)dllRegData;
  // we just ignore this (we are statically linked)
  return S_OK;
}

/*
 *  csCoInitialize ()
 *
 *  This method must be called before any COM calls are made.
 */
STDAPI csCoInitialize (void* lpReserved)
{
  (void)lpReserved;
  return S_OK;
}

/**
 *  csCoUninitialize ()
 *
 *  This must be called before exit.
 */
STDAPI_(void) csCoUninitialize ()
{
}

/**
 *  csCoCreateInstance ()
 *
 *  Creates an instance of a COM object given the CLSID and the
 *  IID. This is the statically-linked version.
 */
STDAPI csCoCreateInstance (REFCLSID rclsid, IUnknown* pUnkOuter, DWORD dwClsContext, REFIID riid, void** ppv)
{
  (void)dwClsContext;
  IUnknown* pClass = NULL;
  IClassFactory *piCF = NULL;
  HRESULT hRes;

  for (int i = 0; StaticRegistry[i].pClass != NULL; i++)
    if (rclsid == *StaticRegistry[i].clsid)
    {
      pClass = StaticRegistry[i].pClass;
      pClass->AddRef ();
      break;
    }

  if (!pClass)
    return REGDB_E_CLASSNOTREG;

  hRes = pClass->QueryInterface (IID_IClassFactory, (void**)&piCF);

  if (FAILED (hRes))
  {
    FINAL_RELEASE (piCF);
    FINAL_RELEASE (pClass);
    return hRes;
  }

  hRes = piCF->CreateInstance (pUnkOuter, riid, ppv);

  if (FAILED(hRes))
  {
    FINAL_RELEASE (piCF);
    FINAL_RELEASE (pClass);

    if (*ppv)
    {
      ((IUnknown*)*ppv)->Release ();
      *ppv = NULL;
    }
    return hRes;
  }

  pClass->Release ();
  return S_OK;
} 

STDAPI csCoGetClassObject (REFCLSID rclsid, DWORD dwClsContext, COSERVERINFO * pServerInfo, REFIID riid, void** ppv)
{
  (void)dwClsContext;
  (void)pServerInfo;
  IUnknown* pClass = NULL;
  HRESULT hRes;

  for (int i = 0; StaticRegistry[i].pClass != NULL; i++)
    if (rclsid == *StaticRegistry[i].clsid)
    {
      pClass = StaticRegistry[i].pClass;
      pClass->AddRef ();
      break;
    }

  if (!pClass)
    return REGDB_E_CLASSNOTREG;

  hRes = pClass->QueryInterface (riid, (void**)ppv);

  if (FAILED(hRes))
  {
    pClass->Release ();
    if (*ppv)
    {
      ((IUnknown*)*ppv)->Release ();
      *ppv = NULL;
    }
    return hRes;
  }

  pClass->Release ();
  return S_OK;
}


/**
 *  csCLSIDFromProgID
 *
 *  This converts a CLSID to a ProgID. A CLSID is a 64-bit unique identifier
 *  that identifies a coclass. A coclass is used to create instances of COM
 *  objects. A coclass exists as a singleton ('static') instance. A ProgID
 *  is a human-readable version that corresponds to a CLSID. It often comes
 *  in the form "libraryname.classname.version".
 *  For example: "CrystalSpace.Graphics3DDirect3D.1"
 */
STDAPI csCLSIDFromProgID ( char** lpszProgID, LPCLSID pclsid)
{
  bool bFoundCLSID = false;
  
  for (int i = 0; StaticRegistry[i].pClass != NULL; i++)
    if (!strcmp(*lpszProgID, StaticRegistry[i].szProgID))
    {
      memcpy(pclsid,  StaticRegistry[i].clsid, sizeof(CLSID));
      bFoundCLSID = true;
      break;
    }

  if (!bFoundCLSID)
    return E_UNEXPECTED;        // couldn't find the matching CLSID.

  return S_OK;
}   

STDAPI csProgIDFromCLSID (REFCLSID clsid, char** lpszProgID)
{
  bool bFoundProgID = false;

  for (int i = 0; StaticRegistry[i].pClass != NULL; i++)
    if (clsid == *StaticRegistry[i].clsid)
    {
      *lpszProgID = new char[strlen(StaticRegistry[i].szProgID)];
      strcpy (*lpszProgID, StaticRegistry[i].szProgID);
      bFoundProgID = true;
      break;
    }

  if (!bFoundProgID)
    return E_UNEXPECTED;        // couldn't find the matching CLSID.

  return S_OK;
} 
 
#else
//------------------------------ Dynamic linking emulation for non-COM platforms -

#pragma message ("Using COM dynamic linking emulation")

IID IID_IUnknown =
{ 0x00000000, 0x0000, 0x0000, { 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 } };

IID IID_IClassFactory =
{ 0x00000001, 0x0000, 0x0000, { 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 } };

/// This is the file where our private class registry is stored.
#define CRYST_REGFILE "cryst.reg"

typedef HRESULT (*GETCLASSOBJPROC)(const CLSID&, const IID&, void**);

/// This is used to keep track of what objects are already instantiated.
struct RunningObjectEntry
{
    /// the CLSID of this class
    CLSID clsid;
    /// a handle to the dynamic library.
    CS_HLIBRARY hLib;
    /// a pointer to the DllGetClassObject function in the library.
    GETCLASSOBJPROC pfnGCO;

    /// pointer to the next object.
    RunningObjectEntry* pNext;
};

/// the list of loaded coclasses.
static RunningObjectEntry* gb_pRunningObjectTable;

// the registry.
static RegistryEntry* gb_pRegistryEntries = NULL;
unsigned int gb_nRegistryEntries;

/**
 *  csCoInitialize ()
 *
 *  This method must be called before any COM calls are made.
 */
STDAPI csCoInitialize (void* lpReserved)
{
    FILE* pFile;

    gb_pRunningObjectTable = NULL;
    gb_pRegistryEntries = NULL;
    gb_nRegistryEntries = 0;

    // load the registry in from the file.
    // the registry by default is stored in CRYST_REGFILE.
    pFile = fopen(CRYST_REGFILE, "r");
    if (!pFile)
    {
        return E_UNEXPECTED;		// we failed.
    }

    fread (&gb_nRegistryEntries, sizeof(int), 1, pFile);

    if (gb_nRegistryEntries)
    {
        size_t cItems;

        CHK (gb_pRegistryEntries = new RegistryEntry[gb_nRegistryEntries]);

        // read the registry in.
        cItems = fread (gb_pRegistryEntries, sizeof(RegistryEntry), gb_nRegistryEntries, pFile);

        if (cItems != gb_nRegistryEntries)
        {
            delete gb_pRegistryEntries;
            gb_pRegistryEntries = NULL;
            gb_nRegistryEntries = 0;

            fclose(pFile);

            return E_UNEXPECTED;        // corrupted registry.
        }
    }

    fclose(pFile);

    return S_OK;
}


/**
 *  csCoUninitialize ()
 *
 *  This must be called before exit.
 */
STDAPI_(void) csCoUninitialize ()
{    
    RunningObjectEntry* pCur = gb_pRunningObjectTable;
    RunningObjectEntry* pNext;

    while (pCur)
    {
        pNext = pCur->pNext;
        delete pCur;

        pCur = pNext;
    }

    delete gb_pRegistryEntries;
    gb_nRegistryEntries = 0;
    gb_pRunningObjectTable = 0;

    csFreeAllLibraries ();
}


// csRegisterServer () -- Adds entries to CRYST_REGFILE.
STDAPI csRegisterServer (const DllRegisterData* pDllRegData)
{
	FILE* pFile;
	RegistryEntry regEntry;

	if (!pDllRegData)
		return E_UNEXPECTED;
	
	// update the number of registry entries
	pFile = fopen(CRYST_REGFILE, "w");
	if (!pFile)	
		return E_UNEXPECTED;

        gb_nRegistryEntries++;
	fwrite (&gb_nRegistryEntries, sizeof(int), 1, pFile);

        for (unsigned int i=0; i<gb_nRegistryEntries-1; i++)
        {
            fwrite (&gb_pRegistryEntries[i], sizeof(RegistryEntry), 1, pFile);
        }
		
	memcpy(&regEntry.clsid, &pDllRegData->clsid, sizeof(CLSID));
	strcpy(regEntry.szDllName, pDllRegData->szInProcServer);
	strcpy(regEntry.szProgID, pDllRegData->szProgID);

	fwrite( &regEntry, sizeof(RegistryEntry), 1, pFile );

	fclose(pFile);

        return csCoInitialize(0);
}

STDAPI csUnregisterServer (const DllRegisterData* pDllRegData)
{
	FILE* pFile;
	RegistryEntry* pNewRegistry;

	unsigned int i, c=0;
	bool bFoundEntry = false;

	CHK (pNewRegistry = new RegistryEntry[gb_nRegistryEntries-1]);

	for (i=0; i<gb_nRegistryEntries; i++)
	{
		if (gb_pRegistryEntries[i].clsid != pDllRegData->clsid)
		{
			memcpy(&pNewRegistry[c].clsid, &pDllRegData->clsid, sizeof(CLSID));
			strcpy(pNewRegistry[c].szDllName, pDllRegData->szInProcServer);
			strcpy(pNewRegistry[c].szProgID, pDllRegData->szProgID);
		
			c++;
		}
		else
		{
			bFoundEntry = true;
		}
	}

	if (!bFoundEntry)
	{
		delete [] pNewRegistry;	
		return S_FALSE;			// seems to already be unregistered.
	}

	// now rewrite the registry file.
	pFile = fopen(CRYST_REGFILE, "w");
	if (!pFile)
		return E_UNEXPECTED;
	
	fwrite (gb_pRegistryEntries, sizeof(RegistryEntry), gb_nRegistryEntries-1, pFile);
	fclose(pFile);

	// reset the registry currently loaded.
	gb_nRegistryEntries--;
	
	delete [] gb_pRegistryEntries;
	gb_pRegistryEntries = pNewRegistry;

        csCoInitialize(0);

	return S_OK;
}

/**
 *  csCoCreateInstance ()
 *
 *  Creates an instance of a COM object given the CLSID and the
 *  IID. This is the dynamic-linking version for non-COM platforms.
 */
STDAPI csCoCreateInstance (REFCLSID rclsid, IUnknown* pUnkOuter, unsigned long dwClsContext, REFIID riid, void** ppv)
{
    assert (gb_pRegistryEntries);

    IClassFactory *piCF = NULL;
    HRESULT hRes;
    CS_HLIBRARY hLib;
    RunningObjectEntry* pCurClass = gb_pRunningObjectTable;
    GETCLASSOBJPROC pfnGCO;

    unsigned int i;
    char szDllName[256];

    // first determine if the coclass is already loaded.
    while (pCurClass)
    {
        if (pCurClass->clsid == rclsid)
        {
            hRes = (*pCurClass->pfnGCO)(rclsid, IID_IClassFactory, (void**)&piCF);
            if (FAILED(hRes))
            {
                if (*ppv)
                {
                    ((IUnknown*)*ppv)->Release ();
                    *ppv = NULL;
                }
                return hRes;
            }

            hRes = piCF->CreateInstance (pUnkOuter, riid, ppv); 

            if (FAILED(hRes))
            {
                if (*ppv)
                {
                    ((IUnknown*)*ppv)->Release ();
                    *ppv = NULL;
                }

                return hRes;
            }

            return S_OK;
        }

        pCurClass = pCurClass->pNext;
    }

    szDllName[0] = 0;
    for (i=0; i< gb_nRegistryEntries; i++)
    {
        if (rclsid == gb_pRegistryEntries[i].clsid)
        {
            strcpy (szDllName, gb_pRegistryEntries[i].szDllName);
            break;
        }
    }

    if (szDllName[0] == 0)
    {
        *ppv = NULL;
        return REGDB_E_CLASSNOTREG;
    }

    hLib = csLoadLibrary (szDllName);
    if (!hLib)
    {
        *ppv = NULL;
        return CO_E_DLLNOTFOUND;
    }

    pfnGCO = (GETCLASSOBJPROC)csGetProcAddress(hLib, "DllGetClassObject");
    if (!pfnGCO)
    {
        *ppv = NULL;
        return CO_E_ERRORINDLL;     // couldn't find 'DllGetClassObject' global function.
    }

    // insert the class in the class table.
    if (!gb_pRunningObjectTable)
    {
        // create the head.
        CHK (gb_pRunningObjectTable = new RunningObjectEntry);
        gb_pRunningObjectTable->pNext = NULL;

        pCurClass = gb_pRunningObjectTable;
    }
    else
    {
        CHK (RunningObjectEntry* pNew = new RunningObjectEntry);
        pNew->pNext = gb_pRunningObjectTable;

        pCurClass = gb_pRunningObjectTable = pNew;
    }

    // fill in data in class table.
    memcpy (&pCurClass->clsid, &rclsid, sizeof(CLSID));
    pCurClass->hLib = hLib;
    pCurClass->pfnGCO = pfnGCO;

    // make the call to DllGetClassObject.
    hRes = (*pfnGCO)(rclsid, IID_IClassFactory, (void**)&piCF);

    if (FAILED(hRes))
    {
        if (piCF)
        {
            *ppv = NULL;
            piCF->Release ();
        }
        return hRes;
    }

    hRes = piCF->CreateInstance( pUnkOuter, riid, ppv ); 

    if (FAILED(hRes))
    {
        if (*ppv)
        {
            ((IUnknown*)*ppv)->Release ();
            *ppv = NULL;
        }

        return hRes;
    }

    return S_OK;
}

STDAPI csCoGetClassObject (REFCLSID rclsid, DWORD dwClsContext, COSERVERINFO * pServerInfo, REFIID riid, void** ppv)
{
    assert (gb_pRegistryEntries);

    HRESULT hRes;
    CS_HLIBRARY hLib;
    RunningObjectEntry* pCurClass = gb_pRunningObjectTable;
    GETCLASSOBJPROC pfnGCO;

    unsigned int i;
    char szDllName[256];

    // first determine if the coclass is already loaded.
    while (pCurClass)
    {
        if (pCurClass->clsid == rclsid)
        {
            hRes = (*pCurClass->pfnGCO)(rclsid, riid, (void**)ppv);
            if (FAILED(hRes))
            {
                if (*ppv)
                {
                    ((IUnknown*)*ppv)->Release ();
                    *ppv = NULL;
                }
                return hRes;
            }

            return S_OK;
        }

        pCurClass = pCurClass->pNext;
    }

    // it's not in the running object table, so load the appropriate library.

    szDllName[0] = 0;
    for (i=0; i< gb_nRegistryEntries; i++)
    {
        if (rclsid == gb_pRegistryEntries[i].clsid)
        {
            strcpy (szDllName, gb_pRegistryEntries[i].szDllName);
            break;
        }
    }

    if (szDllName[0] == 0)
    {
        *ppv = NULL;
        return REGDB_E_CLASSNOTREG;
    }

    hLib = csLoadLibrary (szDllName);
    if (!hLib)
    {
        *ppv = NULL;
        return CO_E_DLLNOTFOUND;
    }

    pfnGCO = (GETCLASSOBJPROC)csGetProcAddress(hLib, "DllGetClassObject");
    if (!pfnGCO)
    {
        *ppv = NULL;
        return CO_E_ERRORINDLL;     // couldn't find 'DllGetClassObject' global function.
    }

    // insert the class in the class table.
    if (!gb_pRunningObjectTable)
    {
        // create the head.
        CHK (gb_pRunningObjectTable = new RunningObjectEntry);
        gb_pRunningObjectTable->pNext = NULL;

        pCurClass = gb_pRunningObjectTable;
    }
    else
    {
        CHK (RunningObjectEntry* pNew = new RunningObjectEntry);
        pNew->pNext = gb_pRunningObjectTable;

        pCurClass = gb_pRunningObjectTable = pNew;
    }

    // fill in data in class table.
    memcpy (&pCurClass->clsid, &rclsid, sizeof(CLSID));
    pCurClass->hLib = hLib;
    pCurClass->pfnGCO = pfnGCO;

    // make the call to DllGetClassObject.
    hRes = (*pfnGCO)(rclsid, riid, (void**)ppv);

    if (FAILED(hRes))
    {
        if (*ppv)
        {
            ((IUnknown*)*ppv)->Release ();
            *ppv = NULL;
        }
        return hRes;
    }

    return S_OK;
}


/**
 *  csCLSIDFromProgID
 *
 *  This converts a CLSID to a ProgID. A CLSID is a 64-bit unique identifier
 *  that identifies a coclass. A coclass is used to create instances of COM
 *  objects. A coclass exists as a singleton ('static') instance. A ProgID
 *  is a human-readable version that corresponds to a CLSID. It often comes
 *  in the form "libraryname.classname.version".
 *  For example: "CrystalSpace.Graphics3DDirect3D.1"
 */
STDAPI csCLSIDFromProgID (char** lpszProgID, LPCLSID pclsid)
{
    unsigned int i = 0;
    bool bFoundCLSID = false;

    for (i=0; i<gb_nRegistryEntries; i++)
    {
        if (!strcmp(*lpszProgID, gb_pRegistryEntries[i].szProgID))
        {
            memcpy(pclsid,  &gb_pRegistryEntries[i].clsid, sizeof(CLSID));
            bFoundCLSID = true;
        }
    }

    if (!bFoundCLSID)
    {
        return E_UNEXPECTED;        // couldn't find the matching CLSID.
    }

    return S_OK;
}

STDAPI csProgIDFromCLSID (REFCLSID clsid, char** lpszProgID)
{
    unsigned int i = 0;
    bool bFoundProgID = false;

    for (i=0; i<gb_nRegistryEntries; i++)
    {
        if (clsid == gb_pRegistryEntries[i].clsid)
        {
            *lpszProgID = new char[strlen(gb_pRegistryEntries[i].szProgID)];
            strcpy (*lpszProgID, gb_pRegistryEntries[i].szProgID);

            bFoundProgID = true;
        }

        i++;
    }

    if (!bFoundProgID)
    {
        return E_UNEXPECTED;        // couldn't find the matching CLSID.
    }

    return S_OK;
}

#endif // CS_STATIC_LINKED

#else  // not NO_COM_SUPPORT

// Direct to COM for platforms that support it.


/**
 *  csCoInitialize ()
 *
 *  This method must be called before any COM calls are made.
 */
STDAPI csCoInitialize (void* lpReserved)
{
    return CoInitialize (lpReserved);
}


/**
 *  csCoUninitialize ()
 *
 *  This must be called before exit.
 */
STDAPI_(void) csCoUninitialize ()
{
    CoUninitialize ();
}

#define MAX_KEY_LEN 128

long WriteRegValue (const char* szKey, const char* szValueName, const char* szValue)
{
	HKEY hKey;
	
	long err = RegCreateKey(HKEY_CLASSES_ROOT, szKey, &hKey);
	if (err == ERROR_SUCCESS)
	{
		err = RegSetValue(hKey, NULL, REG_SZ, szValue, (strlen(szValue) + 1));
		RegCloseKey (hKey);
	}

	return err;
}

extern HINSTANCE DllHandle;

// csRegisterServer () -- Adds entries to the Windows registry.
STDAPI csRegisterServer (const DllRegisterData* pDllRegData)
{	
	LPOLESTR wszGUID;
	char szGUID[50];
	char szKey[MAX_KEY_LEN];
	const char szKeys[][256] =
	{
		"CLSID\\%s",
		"CLSID\\%s\\InprocServer32",
		"CLSID\\%s\\ProgID",
		"%s\\CLSID"
	};
	int i;	
	long err;
	
	//ASSERT (pDllRegData);

	// get the registry-style GUID representation, and convert to ANSI.
	VERIFY_SUCCESS (StringFromCLSID( pDllRegData->clsid, &wszGUID));
	WideCharToMultiByte (CP_ACP, 0, wszGUID, wcslen(wszGUID), szGUID, 50, NULL, NULL);
	LocalFree (wszGUID);

	// turn the brackets into braces
	for (i=0; i<(int)strlen(szGUID); i++)
	{
		if (szGUID[i] == '[')
		{
			szGUID[i] = '{';
		}
		else if (szGUID[i] == ']')
		{
			szGUID[i] = '}';
		}
	}
	szGUID[38] = '\0';

	// now write the keys
	
	// [HKCR\CLSID\{guid}]
	// @=friendlyname
	sprintf (szKey, szKeys[0], szGUID);

	err = WriteRegValue (szKey, 0, pDllRegData->szFriendlyName);
	if (err != ERROR_SUCCESS)
	{
		csUnregisterServer (pDllRegData);
		return E_UNEXPECTED;
	}

	// [HKCR\CLSID\{guid}\InprocServer32]
	// @=filename
	sprintf (szKey, szKeys[1], szGUID);

	char szFilename[MAX_PATH];    
	GetModuleFileName(DllHandle, szFilename, MAX_PATH);

	err = WriteRegValue (szKey, 0, szFilename);
	if (err != ERROR_SUCCESS)
	{
		csUnregisterServer (pDllRegData);
		return E_UNEXPECTED;
	}

	// [HKCR\CLSID\{guid}\ProgID]
	// @=progID
	sprintf (szKey, szKeys[2], szGUID);

	err = WriteRegValue (szKey, 0, pDllRegData->szProgID);
	if (err != ERROR_SUCCESS)
	{
		csUnregisterServer (pDllRegData);
		return E_UNEXPECTED;
	}

	// [HKCR\progID]
	// @=friendlyname
	err = WriteRegValue (pDllRegData->szProgID, 0, pDllRegData->szFriendlyName);
	if (err != ERROR_SUCCESS)
	{
		csUnregisterServer (pDllRegData);
		return E_UNEXPECTED;
	}
	
	// [HKCR\progID\CLSID]
	sprintf (szKey, szKeys[3], pDllRegData->szProgID);

	err = WriteRegValue (szKey, 0, szGUID);
	if (err != ERROR_SUCCESS)
	{
		csUnregisterServer (pDllRegData);
		return E_UNEXPECTED;
	}

	return S_OK;
}

STDAPI csUnregisterServer (const DllRegisterData* pDllRegData)
{
	OLECHAR wszGUID[MAX_KEY_LEN];
	char szGUID[40];
	char szKey[MAX_KEY_LEN];
	const char *szKeys[] =
	{
		{ "CLSID\\%s" },
		{ "CLSID\\%s\\InprocServer32" },
		{ "CLSID\\%s\\ProgID" },
		{ "%s\\CLSID" }
	};

	//ASSERT (pDllRegData);

	// get the registry-style GUID representation, and convert to ANSI.
	StringFromGUID2 (pDllRegData->clsid, wszGUID, 40 );
	wcstombs (szGUID, wszGUID, 40);

	// turn the brackets into braces
	for (int i=0; i<(int)strlen(szGUID); i++)
	{
		if (szGUID[i] == '[')
		{
			szGUID[i] = '{';
		}
		else if (szGUID[i] == ']')
		{
			szGUID[i] = '}';
		}
	}
	szGUID[38] = '\0';

	// now delete the entries.

	for (i=0; i<3; i++)
	{
		sprintf (szKey, szKeys[i], szGUID);
		RegDeleteKey(HKEY_CLASSES_ROOT, szKey);
	}

	RegDeleteKey(HKEY_CLASSES_ROOT, pDllRegData->szProgID);

	sprintf (szKey, szKeys[3], pDllRegData->szProgID);
	RegDeleteKey(HKEY_CLASSES_ROOT, szKey);

	return S_OK;
}


/**
 *  csCoCreateInstance ()
 *
 *  Creates an instance of a COM object given the CLSID and the
 *  IID. This is the dynamic-linking version for non-COM platforms.
 */
STDAPI csCoCreateInstance (REFCLSID rclsid, IUnknown* pUnkOuter, DWORD dwClsContext, REFIID riid, void** ppv)
{
    return CoCreateInstance (rclsid, pUnkOuter, dwClsContext, riid, ppv);
} 


STDAPI csCoGetClassObject (REFCLSID rclsid, DWORD dwClsContext, COSERVERINFO * pServerInfo, REFIID riid, void** ppv)
{
    return CoGetClassObject (rclsid, dwClsContext, pServerInfo, riid, ppv);
}

/**
 *  csCLSIDFromProgID
 *
 *  This converts a CLSID to a ProgID. A CLSID is a 64-bit unique identifier
 *  that identifies a coclass. A coclass is used to create instances of COM
 *  objects. A coclass exists as a singleton ('static') instance. A ProgID
 *  is a human-readable version that corresponds to a CLSID. It often comes
 *  in the form "libraryname.classname.version".
 *  For example: "CrystalSpace.Graphics3DDirect3D.1"
 */
STDAPI csCLSIDFromProgID (char** lpszProgID, LPCLSID pclsid)
{
	HRESULT hRes = S_OK;
	int nSize = MultiByteToWideChar (CP_ACP, NULL, *lpszProgID, -1, NULL, 0);
    LPOLESTR lpwcProgID = new OLECHAR[nSize];

	hRes = MultiByteToWideChar (CP_ACP, NULL, *lpszProgID, -1, lpwcProgID, nSize);
	if (FAILED(hRes))
		goto OnError;

    hRes = CLSIDFromProgID (lpwcProgID, pclsid);

OnError:

    delete [] lpwcProgID;
    return hRes;
}

STDAPI csProgIDFromCLSID (REFCLSID clsid, char** lpszProgID)
{
    HRESULT hRes;
	int nSize = strlen(*lpszProgID);
    LPOLESTR lpwcProgID = new OLECHAR[nSize];

    hRes = ProgIDFromCLSID (clsid, &lpwcProgID);
	if (FAILED(hRes))
		goto OnError;

    hRes = WideCharToMultiByte (CP_ACP, NULL, lpwcProgID, nSize, *lpszProgID, nSize, NULL, 0);
	if (FAILED(hRes))
		goto OnError;

OnError:
    delete lpwcProgID;
    return hRes;
}

void csErrorMessageFromResult (HRESULT hRes)
{
	LPVOID lpMsgBuf;
	
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL,
		hRes, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR) &lpMsgBuf, 0, NULL);
	
	MessageBox (NULL, (const char*)lpMsgBuf, "GetLastError", MB_OK|MB_ICONINFORMATION);
	
	LocalFree (lpMsgBuf);
}

#endif      // !NO_COM_SUPPORT

