/*
  Crystal Space Windowing System: Windowing System Application class
  Copyright (C) 1998 by Jorrit Tyberghein
  Written by Andrew Zabolotny <bit@eltech.ru>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "system/system.h"
#include "system/sysdriv.h"
#include "util/inifile.h"
#include "util/csstrvec.h"
#include "util/input/cseventq.h"
#include "cslistbx.h"
#include "csmouse.h"
#include "csmenu.h"
#include "cswindow.h"
#include "csdialog.h"
#include "csapp.h"
#include "cswsutil.h"

// Archive path of CSWS.CFG
#define CSWS_CFG "sys/csws.cfg"

// A special variation of SysSystemDriver for csApp
class appSystemDriver : public SysSystemDriver
{
  int curline, maxlines, maxwidth;
  char **textline;
  int *linecolor;
  int textcolor;

public:
  static csApp *application;
  static csWorld *init_world;
  static Textures *init_tex;

  appSystemDriver (csApp *parentapp) : SysSystemDriver ()
  {
    application = parentapp;
    textline = NULL;
  }

  virtual void Initialize (int argc, char *argv[]);

  virtual void DemoWrite (char* buf);

  void CloseConsole ()
  {
    Sleep (1000);
    Gfx2D->ClearAll (0);
    if (init_world) CHKB (delete init_world);
    if (init_tex) CHKB (delete init_tex);
  }

  virtual void NextFrame (long elapsed_time, long current_time)
  {
    SysSystemDriver::NextFrame (elapsed_time, current_time);
    application->NextFrame ();
  }

  virtual void Alert (char* msg);

  virtual void Warn (char* msg);
};

csApp *appSystemDriver::application = NULL;
csWorld *appSystemDriver::init_world = NULL;
Textures *appSystemDriver::init_tex = NULL;

void appSystemDriver::Initialize (int argc, char *argv[])
{
  csSystemDriver::Initialize (argc, argv);

  int Width, Height;
  GfxInfo->GetWidth (Width);
  GfxInfo->GetHeight (Height);

  // Initialize console data
  curline = 0;
  textcolor = cs_Color_White;
  maxwidth = Width / 4;
  maxlines = Height / (application->TextHeight () + 2);
  CHK (textline = new char *[maxlines]);
  CHK (linecolor = new int [maxlines]);
  for (int i = 0; i < maxlines; i++)
  {
    CHK (textline [i] = new char [maxwidth + 1]);
    memset (textline [i], 0, maxwidth + 1);
    linecolor [i] = cs_Color_Black;
  } /* endfor */

  // Now open temporary console
  CHK (init_world = new csWorld ());
  CHK (init_tex = new Textures (4));
  init_tex->initialize_limited ();

  // Open the visual, initialize keyboard and mouse
  System->Open (application->GetText ());

  // create a uniform palette: r(3)g(3)b(2)
  int r,g,b;
  for (r = 0; r < 8; r++)
    for (g = 0; g < 8; g++)
      for (b = 0; b < 4; b++)
        init_tex->alloc_rgb (r * 32, g * 32, b * 64, 0);

  application->SetWorld (init_world, init_tex);
};

/*
 * The routine to update the screen in demo mode.
 */
void appSystemDriver::DemoWrite (char* buf)
{
  char *crpos;
  while ((crpos = strchr (buf, '\n')))
  {
    *crpos = 0;
    int len = maxwidth - strlen (textline [curline]);
    if (len > 0)
    {
      strncat (textline [curline], buf, len);
      linecolor [curline] = textcolor;
    }
    buf = crpos + 1;
    curline++;
    if (curline >= maxlines)
    {
      for (int i = 1; i < maxlines; i++)
      {
        strcpy (textline [i - 1], textline [i]);
        linecolor [i - 1] = linecolor [i];
      } /* endfor */
      curline--;
    } /* endif */
    strcpy (textline [curline], "");
  } /* endwhile */
  strcpy (textline [curline], buf);
  linecolor [curline] = textcolor;

  if (Gfx2D)
  {
    if (SUCCEEDED (Gfx2D->BeginDraw ()))
    {
      Gfx2D->Clear (application->Pal [cs_Color_Gray_D]);
      Gfx2D->SetFontID (application->GetFont ());
      int fh = application->TextHeight () + 2;
      for (int i = 0; i < maxlines; i++)
        if (*textline [i])
        {
          Gfx2D->Write (1, 1 + i * fh,
            application->Pal [cs_Color_Black], -1, textline [i]);
          Gfx2D->Write (0, 0 + i * fh,
            application->Pal [linecolor [i]], -1, textline [i]);
        } /* endfor */
      Gfx2D->FinishDraw ();
      Gfx2D->Print (NULL);
    } /* endif */
  } /* endif */
}

void appSystemDriver::Alert (char* msg)
{
  if (DemoReady)
  {
    int oldcolor = textcolor;
    textcolor = cs_Color_Red_L;
    DemoWrite (msg);
    textcolor = oldcolor;
  }
  else
    csSystemDriver::Alert (msg);
}

void appSystemDriver::Warn (char* msg)
{
  if (DemoReady)
  {
    int oldcolor = textcolor;
    textcolor = cs_Color_Yellow;
    DemoWrite (msg);
    textcolor = oldcolor;
  }
  else
    csSystemDriver::Alert (msg);
}

/*
 * Do a large debug dump just before the program
 * exits. This function can be installed as a last signal
 * handler if the program crashes (for systems that support
 * this).
 */
void debug_dump ()
{
}

/*
 * This function is installed with atexit()
 */
void cleanup ()
{
  if (((appSystemDriver *)System)->application)
    CHKB (delete ((appSystemDriver *)System)->application);
}

//--//--//--//--//--//--//--//--//--//--//--//--//--//--//--//--//- csApp  -//--

csApp::csApp (char *AppTitle, int argc, char *argv[],
  csAppBackgroundStyle iBackgroundStyle) : csComponent (NULL)
{
  app = this;           // so that all inserted windows will inherit it
  world = NULL;         // No world so far
  text = NULL;
  SetText (AppTitle);   // application title string
  MouseOwner = NULL;    // no mouse owner
  KeyboardOwner = NULL; // no keyboard owner
  FocusOwner = NULL;    // no focus owner
  RedrawFlag = true;
  WindowListChanged = false;
  LoopLevel = 0;
  BackgroundStyle = iBackgroundStyle;
  mousetexturename = mousetextureparm =
  titletexturename = titletextureparm =
  dialogtexturename = dialogtextureparm = NULL;
  titlebardefs = dialogdefs =
  customtexturename = customtextureparm = NULL;
  insert = true;

  oldMouseCursorID = csmcNone;
  MouseCursorID = csmcArrow;

  SetPalette (CSPAL_APP);
  state |= CSS_VISIBLE | CSS_SELECTABLE | CSS_FOCUSED;

  CHK (Mouse = new csMouse (NULL));
  Mouse->app = this;
  CHK (GfxPpl = new csGraphicsPipeline ());

  CHK (System = new appSystemDriver (this));
  System->Initialize (argc, argv);
  EventQueue = System->EventQueue;

  atexit (cleanup);

  // Create the world
  CHK (csWorld *world = new csWorld ());
  char world_file[50];
  strcpy (world_file, config->GetStr ("World", "WORLDFILE", "csws.zip"));
  if (!world->load_file (NULL, world_file))
    exit (0);

  world->initialize ();
  SetWorld (world, NULL);

  int Width, Height;
  GfxInfo->GetWidth (Width);
  GfxInfo->GetHeight (Height);
  bound.Set (0, 0, Width, Height);
  dirty.Set (bound);
  WindowListWidth = Width / 3;
  WindowListHeight = Width / 6;

  // Tell CsPrintf() console is ready
  System->DemoReady = true;
}

csApp::~csApp ()
{
  // Delete all children prior to shutting down the system
  DeleteAll ();

  if (world)
    CHKB (delete world);
  if (mousetexturename)
    CHKB (free (mousetexturename));
  if (mousetextureparm)
    CHKB (free (mousetextureparm));
  if (titletexturename)
    CHKB (free (titletexturename));
  if (titletextureparm)
    CHKB (free (titletextureparm));
  if (titlebardefs)
    CHKB (delete titlebardefs);
  if (dialogtexturename)
    CHKB (free (dialogtexturename));
  if (dialogtextureparm)
    CHKB (free (dialogtextureparm));
  if (dialogdefs)
    CHKB (delete dialogdefs);
  if (customtexturename)
    CHKB (delete customtexturename);
  if (customtextureparm)
    CHKB (delete customtextureparm);

  if (Mouse)
    CHKB (delete Mouse);
  if (GfxPpl)
    CHKB (delete GfxPpl);

  if (System)
  {
    if (((appSystemDriver *)System)->application)
      ((appSystemDriver *)System)->application = NULL;
    System->Close ();
    CHK (delete System);
  }
}

void csApp::SetWorld (csWorld *AppWorld, Textures *AppTextures)
{
  world = AppWorld;
  textures = AppTextures;
  if (textures)
  {
    textures->alloc_palette ();
    setup_palette ();
  } /* endif */
}

bool csApp::InitialSetup ()
{
  load_config ();
  load_textures ();
  ((appSystemDriver *)System)->CloseConsole ();
  return true;
}

void csApp::Loop ()
{
  System->Loop();
}

void csApp::NextFrame ()
{
  Process ();

  // Now update screen
  if (GfxPpl->BeginDraw ())
  {
    Update ();
    GfxPpl->FinishDraw ();
  }
}

void csApp::Update ()
{
  if (MouseCursorID != oldMouseCursorID)
  {
    Mouse->SetCursor (MouseCursorID);
    oldMouseCursorID = MouseCursorID;
  } /* endif */

  // Save background/draw mouse cursor
  Mouse->Draw ();

  // Flush application graphics operations
  GfxPpl->Flush (GetPage ());

  // Restore previous background under mouse
  Mouse->Undraw ();
}

void csApp::load_config ()
{
  enum
  {
    tkMouseCursorImageFile = 1,
    tkMouseCursor,
    tkTitleButtonImageFile,
    tkTitleButton,
    tkDialogButtonImageFile,
    tkDialogButton,
    tkCustomTexture
  };
  static tokenDesc commands[] =
  {
    {tkMouseCursorImageFile,  "MOUSECURSORIMAGEFILE"},
    {tkMouseCursor,           "MOUSECURSOR"},
    {tkTitleButtonImageFile,  "TITLEBUTTONIMAGEFILE"},
    {tkTitleButton,           "TITLEBUTTON"},
    {tkDialogButtonImageFile, "DIALOGBUTTONIMAGEFILE"},
    {tkDialogButton,          "DIALOGBUTTON"},
    {tkCustomTexture,         "CUSTOMTEXTURE"},
    {0,                       NULL}
  };

  size_t cswscfglen;
  char *cswscfg = GetArchive ()->read (CSWS_CFG, &cswscfglen);
  if (!cswscfg)
    return;

  if (!customtexturename)
    CHKB (customtexturename = new csStrVector (16, 16));
  if (!customtextureparm)
    CHKB (customtextureparm = new csStrVector (16, 16));
  if (!titlebardefs)
    CHKB (titlebardefs = new csStrVector (16, 16));
  if (!dialogdefs)
    CHKB (dialogdefs = new csStrVector (16, 16));

  int cmd;
  char *oldcfg = cswscfg, *name, *params;
  while ((cmd = csGetObject (&cswscfg, commands, &name, &params)) > 0)
  {
    switch (cmd)
    {
      case tkMouseCursorImageFile:
      {
        mousetexturename = strdup (name);
        mousetextureparm = strdup (params);
        break;
      }
      case tkMouseCursor:
      {
        Mouse->NewPointer (name, params);
        break;
      }
      case tkTitleButtonImageFile:
      {
        titletexturename = strdup (name);
        titletextureparm = strdup (params);
        break;
      }
      case tkTitleButton:
      {
        CHK (char *tmp = new char [strlen (name) + 1 + strlen (params) + 1]);
        sprintf (tmp, "%s %s", name, params);
        titlebardefs->Push (tmp);
        break;
      }
      case tkDialogButtonImageFile:
      {
        dialogtexturename = strdup (name);
        dialogtextureparm = strdup (params);
        break;
      }
      case tkDialogButton:
      {
        CHK (char *tmp = new char [strlen (name) + 1 + strlen (params) + 1]);
        sprintf (tmp, "%s %s", name, params);
        dialogdefs->Push (tmp);
        break;
      }
      case tkCustomTexture:
      {
        customtexturename->Push (strdup (name));
        customtextureparm->Push (strdup (params));
        break;
      }
      default:
        CsPrintf (MSG_DEBUG_0, "Unknown token in csws.cfg! (%s)\n", cswscfg);
        fatal_exit (0, false);			// Unknown token
    }
  }
  CHK (delete[] oldcfg);
}

void csApp::load_textures ()
{
  csStrVector texlist (16, 32);
  GetTextures (&texlist);

  int i, txlen = strlen ("MAX_TEXTURES (####)\n") + 1;  // Up to 9999 textures
  for (i = 0; i < texlist.Length (); i++)
    txlen += strlen ((char *)texlist [i]);

  CHK (char *texcfg = new char [txlen]);
  sprintf (texcfg, "MAX_TEXTURES (%d)\n", texlist.Length ());
  for (i = 0; i < texlist.Length (); i++)
    strcat (texcfg, (char *)texlist[i]);

  world->load_textures (texcfg);
  textures = world->get_textures ();
  textures->initialize (world);
  textures->alloc_palette ();
  setup_palette ();
  CHK (delete [] texcfg);
  Mouse->Setup ();
}

void csApp::setup_palette ()
{
  Pal[cs_Color_Black]   = textures->find_rgb_real (  0,   0,   0);
  Pal[cs_Color_White]   = textures->find_rgb_real (255, 255, 255);
  Pal[cs_Color_Gray_D]  = textures->find_rgb_real (128, 128, 128);
  Pal[cs_Color_Gray_M]  = textures->find_rgb_real (160, 160, 160);
  Pal[cs_Color_Gray_L]  = textures->find_rgb_real (204, 204, 204);
  Pal[cs_Color_Blue_D]  = textures->find_rgb_real (  0,  20,  80);
  Pal[cs_Color_Blue_M]  = textures->find_rgb_real (  0,  44, 176);
  Pal[cs_Color_Blue_L]  = textures->find_rgb_real (  0,  64, 255);
  Pal[cs_Color_Green_D] = textures->find_rgb_real ( 20,  80,  20);
  Pal[cs_Color_Green_M] = textures->find_rgb_real ( 44, 176,  44);
  Pal[cs_Color_Green_L] = textures->find_rgb_real ( 64, 255,  64);
  Pal[cs_Color_Red_D]   = textures->find_rgb_real ( 80,   0,   0);
  Pal[cs_Color_Red_M]   = textures->find_rgb_real (176,   0,   0);
  Pal[cs_Color_Red_L]   = textures->find_rgb_real (255,   0,   0);
  Pal[cs_Color_Cyan_D]  = textures->find_rgb_real (  0,  60,  80);
  Pal[cs_Color_Cyan_M]  = textures->find_rgb_real (  0, 132, 176);
  Pal[cs_Color_Cyan_L]  = textures->find_rgb_real (  0, 192, 255);
  Pal[cs_Color_Brown_D] = textures->find_rgb_real ( 80,  60,  20);
  Pal[cs_Color_Brown_M] = textures->find_rgb_real (176, 132,  44);
  Pal[cs_Color_Brown_L] = textures->find_rgb_real (255, 192,  64);
}

static void add_texture (csStrVector *Textures, char *name, char *parm)
{
  int txlen = (name ? strlen (name) : 0) + (parm ? strlen (parm) : 0);
  CHK (char *tx = new char [strlen ("TEXTURE '' ()\n") + txlen + 1]);
  strcpy (tx, "TEXTURE '");
  if (name) strcat (tx, name);
  strcat (tx, "' (");
  if (parm) strcat (tx, parm);
  strcat (tx, ")\n");
  Textures->Push (tx);
}

void csApp::GetTextures (csStrVector *Textures)
{
  add_texture (Textures, mousetexturename, mousetextureparm);
  add_texture (Textures, titletexturename, titletextureparm);
  add_texture (Textures, dialogtexturename, dialogtextureparm);
  for (int i = 0; i < customtexturename->Length (); i++)
    add_texture (Textures, (char *)((*customtexturename)[i]),
      (char *)((*customtextureparm)[i]));
}

void csApp::Process ()
{
  bool did_some_work = false;
  csEvent *ev;

  // Broadcast a "pre-process" event
  {
    csEvent ev (csevBroadcast, cscmdPreProcess);
    HandleEvent (ev);
  }

  while ((ev = (EventQueue->Get ())))
  {
    did_some_work = true;
    if (!PreHandleEvent (*ev))
      if (!HandleEvent (*ev))
        PostHandleEvent (*ev);
    CHK (delete ev);
  }

  // Broadcast a "post-process" event
  {
    csEvent ev (csevBroadcast, cscmdPostProcess);
    HandleEvent (ev);
  }

  // Redraw all changed windows
  while (RedrawFlag)
  {
    did_some_work = true;
    RedrawFlag = false;
    csEvent ev (csevBroadcast, cscmdRedraw);
    HandleEvent (ev);
  }

  // If event queue is empty, sleep for one timeslice
  if (!did_some_work)
    Idle ();
}

void csApp::Idle ()
{
  System->Sleep (1);
}

bool csApp::PreHandleEvent (csEvent &Event)
{
  if (MouseOwner && IS_MOUSE_EVENT (Event))
  {
    // Bring mouse coordinates to child coordinate system
    int dX = 0, dY = 0;
    MouseOwner->LocalToGlobal (dX, dY);
    Event.Mouse.x -= dX;
    Event.Mouse.y -= dY;
    bool ret = MouseOwner->PreHandleEvent (Event);
    Event.Mouse.x += dX;
    Event.Mouse.y += dY;
    return ret;
  }
  else if (KeyboardOwner && IS_KEYBOARD_EVENT (Event))
    return KeyboardOwner->PreHandleEvent (Event);
  else if (FocusOwner)
    return FocusOwner->PreHandleEvent (Event);
  else
    return csComponent::PreHandleEvent (Event);
}

bool csApp::PostHandleEvent (csEvent &Event)
{
  if (MouseOwner && IS_MOUSE_EVENT (Event))
  {
    // Bring mouse coordinates to child coordinate system
    int dX = 0, dY = 0;
    MouseOwner->LocalToGlobal (dX, dY);
    Event.Mouse.x -= dX;
    Event.Mouse.y -= dY;
    bool ret = MouseOwner->PostHandleEvent (Event);
    Event.Mouse.x += dX;
    Event.Mouse.y += dY;
    return ret;
  }
  else if (KeyboardOwner && IS_KEYBOARD_EVENT (Event))
    return KeyboardOwner->PostHandleEvent (Event);
  else if (FocusOwner)
    return FocusOwner->PostHandleEvent (Event);
  else
    return csComponent::PostHandleEvent (Event);
}

bool csApp::HandleEvent (csEvent &Event)
{
  // Mouse should always receive events
  // to reflect current mouse position
  Mouse->HandleEvent (Event);

  if ((Event.Type == csevKeyDown)
   && (Event.Key.Code == CSKEY_INS))
    insert = !insert;

  // If mouse moves, reset mouse cursor to arrow
  if (Event.Type == csevMouseMove)
    SetMouse (csmcArrow);

  // If a component captured the mouse,
  // forward all mouse (and keyboard) events to it
  if (MouseOwner && (IS_MOUSE_EVENT (Event) || IS_KEYBOARD_EVENT (Event)))
  {
    if (IS_MOUSE_EVENT (Event))
      MouseOwner->GlobalToLocal (Event.Mouse.x, Event.Mouse.y);
    return MouseOwner->HandleEvent (Event);
  } /* endif */

  // If a component captured the keyboard, forward all keyboard events to it
  if (KeyboardOwner && IS_KEYBOARD_EVENT (Event))
  {
    return KeyboardOwner->HandleEvent (Event);
  } /* endif */

  // If a component captured events, forward all focused events to it
  if (FocusOwner && (IS_MOUSE_EVENT (Event) || IS_KEYBOARD_EVENT (Event)))
  {
    if (IS_MOUSE_EVENT (Event))
      FocusOwner->GlobalToLocal (Event.Mouse.x, Event.Mouse.y);
    return FocusOwner->HandleEvent (Event);
  } /* endif */

  // Send event to children as appropiate
  if (csComponent::HandleEvent (Event))
    return true;

  // Handle 'window list' event
  if ((Event.Type == csevMouseDown)
   && (System->Mouse->Button[1] && System->Mouse->Button[2]))
  {
    WindowList ();
    return true;
  } /* endif */

  return false;
}

void csApp::WindowList ()
{
  CHK (csWindowList *wl = new csWindowList (this));
  int x, y;
  GetMouse ()->GetPosition (x, y);
  int w = WindowListWidth;
  int h = WindowListHeight;
  csRect rect (x - w/2, y - h/2, x + w/2, y + h/2);
  if (rect.xmin < 0)
    rect.Move (-rect.xmin, 0);
  if (rect.ymin < 0)
    rect.Move (0, -rect.ymin);
  if (rect.xmax > bound.xmax)
    rect.Move (-(rect.xmax - bound.xmax), 0);
  if (rect.ymax > bound.ymax)
    rect.Move (0, -(rect.ymax - bound.ymax));
  ((csComponent *)wl)->SetRect (rect);
  wl->Select ();
}

void csApp::Draw ()
{
  switch (BackgroundStyle)
  {
    case csabsNothing:
      break;
    case csabsSolid:
      Box (0, 0, bound.Width (), bound.Height (), CSPAL_APP_WORKSPACE);
      break;
  } /* endswitch */
  csComponent::Draw ();
}

bool csApp::GetKeyState (int iKey)
{
  switch (iKey)
  {
    case CSKEY_ESC:       return System->Keyboard->Key.esc;
    case CSKEY_ENTER:     return System->Keyboard->Key.enter;
    case CSKEY_TAB:       return System->Keyboard->Key.tab;
    case CSKEY_BACKSPACE: return System->Keyboard->Key.backspace;
    case CSKEY_UP:        return System->Keyboard->Key.up;
    case CSKEY_DOWN:      return System->Keyboard->Key.down;
    case CSKEY_LEFT:      return System->Keyboard->Key.left;
    case CSKEY_RIGHT:     return System->Keyboard->Key.right;
    case CSKEY_PGUP:      return System->Keyboard->Key.pgup;
    case CSKEY_PGDN:      return System->Keyboard->Key.pgdn;
    case CSKEY_HOME:      return System->Keyboard->Key.home;
    case CSKEY_END:       return System->Keyboard->Key.end;
    case CSKEY_INS:       return System->Keyboard->Key.ins;
    case CSKEY_DEL:       return System->Keyboard->Key.del;
    case CSKEY_CTRL:      return System->Keyboard->Key.ctrl;
    case CSKEY_ALT:       return System->Keyboard->Key.alt;
    case CSKEY_SHIFT:     return System->Keyboard->Key.shift;
  }
  return false;
}

void csApp::Insert (csComponent *comp)
{
  csComponent::Insert (comp);
  WindowListChanged = true;
}

void csApp::Delete (csComponent *comp)
{
  csComponent::Delete (comp);
  WindowListChanged = true;
}

int csApp::Execute (csComponent *comp)
{
  csComponent *oldFocusOwner = CaptureFocus (comp);
  bool oldExitLoop = System->ExitLoop;
  System->ExitLoop = false;

  comp->Select ();
  comp->SetState (CSS_MODAL, true);
  LoopLevel++;
  System->Loop ();
  System->ExitLoop = oldExitLoop;
  LoopLevel--;
  comp->SetState (CSS_MODAL, false);

  CaptureFocus (oldFocusOwner);
  return DismissCode;
}

void csApp::Dismiss (int iCode)
{
  if (LoopLevel)
  {
    DismissCode = iCode;
    System->ExitLoop = true;
  } /* endif */
}
