/*
  Crystal Space Windowing System: Graphics Pipeline class
  Copyright (C) 1998 by Jorrit Tyberghein
  Written by Andrew Zabolotny <bit@eltech.ru>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <limits.h>
#include <assert.h>
#include "csgfxppl.h"
#include "render/graph2d.h"
#include "render/graph3d.h"

csGraphicsPipeline::csGraphicsPipeline ()
{
  MaxPage = 0;
  memset (SyncArea, 0, sizeof (SyncArea));
  ClearColor = -1;
  pipelen = 0;
}

csGraphicsPipeline::~csGraphicsPipeline ()
{
  Desync ();
}

void csGraphicsPipeline::Box (int xmin, int ymin, int xmax, int ymax, int color)
{
  csPipeEntry *pe = AllocOp (pipeopBOX);
  if (!pe) return;
  pe->Box.xmin = xmin;
  pe->Box.ymin = ymin;
  pe->Box.xmax = xmax;
  pe->Box.ymax = ymax;
  pe->Box.color = color;
}

void csGraphicsPipeline::Line (int x1, int y1, int x2, int y2, int color)
{
  csPipeEntry *pe = AllocOp (pipeopLINE);
  if (!pe) return;
  pe->Line.x1 = x1;
  pe->Line.y1 = y1;
  pe->Line.x2 = x2;
  pe->Line.y2 = y2;
  pe->Line.color = color;
}

void csGraphicsPipeline::Pixel (int x, int y, int color)
{
  csPipeEntry *pe = AllocOp (pipeopPIXEL);
  if (!pe) return;
  pe->Pixel.x = x;
  pe->Pixel.y = y;
  pe->Pixel.color = color;
}

void csGraphicsPipeline::Text (int x, int y, int fg, int bg, int font, char *s)
{
  csPipeEntry *pe = AllocOp (pipeopTEXT);
  if (!pe) return;
  pe->Text.x = x;
  pe->Text.y = y;
  pe->Text.fg = fg;
  pe->Text.bg = bg;
  pe->Text.font = font;
  CHK (pe->Text.string = new char[strlen (s) + 1]);
  strcpy (pe->Text.string, s);
}

void csGraphicsPipeline::Sprite2D (csSprite2D *s2d, int x, int y, int w, int h)
{
  csPipeEntry *pe = AllocOp (pipeopSPR2D);
  if (!pe) return;
  pe->Spr2D.s2d = s2d;
  pe->Spr2D.x = x;
  pe->Spr2D.y = y;
  pe->Spr2D.w = w;
  pe->Spr2D.h = h;
}

void csGraphicsPipeline::SaveArea (ImageArea **Area, int x, int y, int w, int h)
{
  csPipeEntry *pe = AllocOp (pipeopSAVAREA);
  if (!pe) return;
  pe->SavArea.Area = Area;
  pe->SavArea.x = x;
  pe->SavArea.y = y;
  pe->SavArea.w = w;
  pe->SavArea.h = h;
}

void csGraphicsPipeline::RestoreArea (ImageArea *Area, bool Free)
{
  csPipeEntry *pe = AllocOp (pipeopRESAREA);
  if (!pe) return;
  pe->ResArea.Area = Area;
  pe->ResArea.Free = Free;
}

void csGraphicsPipeline::Clear (int color)
{
  ClearColor = color;
  ClearPage = -1;
}

void csGraphicsPipeline::SetClipRect (int xmin, int ymin, int xmax, int ymax)
{
  csPipeEntry *pe = AllocOp (pipeopSETCLIP);
  if (!pe) return;
  pe->ClipRect.xmin = xmin;
  pe->ClipRect.xmax = xmax;
  pe->ClipRect.ymin = ymin;
  pe->ClipRect.ymax = ymax;
}

void csGraphicsPipeline::RestoreClipRect()
{
  AllocOp (pipeopRESCLIP);
}

// Theoretically Update() could contain a kind of optimizer which would
// minimize redraw operations, but as we for now don't care about speed
// it doesn't. Its complex anyway.
void csGraphicsPipeline::Flush (int iCurPage)
{
  CurPage = iCurPage;
  if (CurPage > MaxPage)
    MaxPage = CurPage;

  int old_font;
  int old_xmin, old_ymin, old_xmax, old_ymax;
  int clip_xmin, clip_ymin, clip_xmax, clip_ymax;
  GfxInfo->GetFontID (old_font);
  Gfx2D->GetClipRect (old_xmin, old_ymin, old_xmax, old_ymax);
  clip_xmin = old_xmin; clip_xmax = old_xmax;
  clip_ymin = old_ymin; clip_ymax = old_ymax;

#define INCLUDE_MIN_POINT(x, y)                         \
  {                                                     \
    int _x = x; if (_x < clip_xmin) _x = clip_xmin;     \
    int _y = y; if (_y < clip_ymin) _y = clip_ymin;     \
    if (xmin > _x) xmin = _x;                           \
    if (ymin > _y) ymin = _y;                           \
  }
#define INCLUDE_MAX_POINT(x, y)                         \
  {                                                     \
    int _x = x; if (_x > clip_xmax) _x = clip_xmax;     \
    int _y = y; if (_y > clip_ymax) _y = clip_ymax;     \
    if (xmax < _x) xmax = _x;                           \
    if (ymax < _y) ymax = _y;                           \
  }
#define INCLUDE_POINT(x, y)                             \
  INCLUDE_MIN_POINT (x, y);                             \
  INCLUDE_MAX_POINT (x + 1, y + 1);

  // First, synchronize image with other vide pages, if any
  RefreshRect.Set (INT_MAX, INT_MAX, INT_MIN, INT_MIN);
  Sync (CurPage, RefreshRect.xmin, RefreshRect.ymin, RefreshRect.xmax,
    RefreshRect.ymax);
  // Now paint everything user requested
  int xmin = INT_MAX, xmax = INT_MIN, ymin = INT_MAX, ymax = INT_MIN;
  for (int i = 0; i < pipelen; i++)
  {
    register csPipeEntry *op = &pipeline[i];
    switch (op->Op)
    {
      case pipeopNOP:
        break;
      case pipeopBOX:
      {
        int y;
        // Const expressions should be eliminated by GCC's CSE,
        // so don't care about them
        for (y = op->Box.ymin; y < op->Box.ymax; y++)
          Gfx2D->DrawHorizLine (op->Box.xmin, op->Box.xmax - 1, y, op->Box.color);
        INCLUDE_MIN_POINT (op->Box.xmin, op->Box.ymin);
        INCLUDE_MAX_POINT (op->Box.xmax, op->Box.ymax);
        break;
      }
      case pipeopLINE:
        if (op->Line.y1 == op->Line.y2)
          Gfx2D->DrawHorizLine (op->Line.x1, op->Line.x2, op->Line.y1,
            op->Line.color);
        else
          Gfx2D->DrawLine (op->Line.x1, op->Line.y1,
           op->Line.x2, op->Line.y2, op->Line.color);
        INCLUDE_POINT (op->Line.x1, op->Line.y1);
        INCLUDE_POINT (op->Line.x2, op->Line.y2);
        break;
      case pipeopPIXEL:
        Gfx2D->DrawPixel (op->Pixel.x, op->Pixel.y, op->Pixel.color);
        INCLUDE_POINT (op->Pixel.x, op->Pixel.y);
        break;
      case pipeopTEXT:
      {
        Gfx2D->SetFontID (op->Text.font);
        Gfx2D->Write (op->Text.x, op->Text.y,
         op->Text.fg, op->Text.bg, op->Text.string);
        int tmpx = op->Text.x + TextWidth (op->Text.string, op->Text.font);
        int tmpy = op->Text.y + TextHeight (op->Text.font);
        INCLUDE_MIN_POINT (op->Text.x, op->Text.y);
        INCLUDE_MAX_POINT (tmpx, tmpy);
        CHK (delete[] op->Text.string);
        break;
      }
      case pipeopSPR2D:
        op->Spr2D.s2d->Draw (op->Spr2D.x, op->Spr2D.y, op->Spr2D.w, op->Spr2D.h);
        INCLUDE_MIN_POINT (op->Spr2D.x, op->Spr2D.y);
        INCLUDE_MAX_POINT (op->Spr2D.x + op->Spr2D.w, op->Spr2D.y + op->Spr2D.h);
        break;
      case pipeopSAVAREA:
        Gfx2D->SaveArea (op->SavArea.Area, op->SavArea.x,
          op->SavArea.y, op->SavArea.w, op->SavArea.h);
        break;
      case pipeopRESAREA:
        INCLUDE_MIN_POINT (op->ResArea.Area->x, op->ResArea.Area->y);
        INCLUDE_MAX_POINT (op->ResArea.Area->x + op->ResArea.Area->w,
          op->ResArea.Area->y + op->ResArea.Area->h);
        Gfx2D->RestoreArea (op->ResArea.Area, op->ResArea.Free);
        break;
      case pipeopSETCLIP:
        Gfx2D->SetClipRect (clip_xmin = op->ClipRect.xmin,
          clip_ymin = op->ClipRect.ymin, clip_xmax = op->ClipRect.xmax,
          clip_ymax = op->ClipRect.ymax);
        break;
      case pipeopRESCLIP:
        Gfx2D->SetClipRect (clip_xmin = old_xmin, clip_ymin = old_ymin,
          clip_xmax = old_xmax, clip_ymax = old_ymax);
        break;
      default:
        assert ("unknown graphics pipeline operation in Pipeline::Flush()");
    } /* endswitch */
  } /* endfor */
  pipelen = 0;
  Gfx2D->SetClipRect (old_xmin, old_ymin, old_xmax, old_ymax);
  Gfx2D->SetFontID (old_font);

  if (xmin < 0) xmin = 0; if (xmax > FRAME_WIDTH)  xmax = FRAME_WIDTH;
  if (ymin < 0) ymin = 0; if (ymax > FRAME_HEIGHT) ymax = FRAME_HEIGHT;
  SyncRect [CurPage].xmin = xmin; SyncRect [CurPage].xmax = xmax;
  SyncRect [CurPage].ymin = ymin; SyncRect [CurPage].ymax = ymax;
  if (xmin < RefreshRect.xmin) RefreshRect.xmin = xmin;
  if (xmax > RefreshRect.xmax) RefreshRect.xmax = xmax;
  if (ymin < RefreshRect.ymin) RefreshRect.ymin = ymin;
  if (ymax > RefreshRect.ymax) RefreshRect.ymax = ymax;

  if (SyncArea [CurPage])
    Gfx2D->FreeArea (SyncArea [CurPage]);
  bool dblbuff;
  Gfx2D->GetDoubleBufferState (dblbuff);
  if (dblbuff
   && (xmin < xmax) && (ymin < ymax))
    Gfx2D->SaveArea (&SyncArea [CurPage], xmin, ymin, xmax - xmin, ymax - ymin);
  else
    SyncArea [CurPage] = NULL;
}

void csGraphicsPipeline::Desync ()
{
  int i;

  // Discard all backtracking rectangles
  for (i = 0; i < MAX_SYNC_PAGES; i++)
    if (SyncArea [i])
    {
      Gfx2D->FreeArea (SyncArea [i]);
      SyncArea [i] = NULL;
    } /* endif */

  // Discard all 'restore area' operations
  for (i = 0; i < pipelen; i++)
    if (pipeline[i].Op == pipeopRESAREA)
    {
      Gfx2D->FreeArea (pipeline[i].ResArea.Area);
      pipeline[i].Op = pipeopNOP;
    }
}

// This procedure propagates all changes made to other pages onto the
// current page.
void csGraphicsPipeline::Sync (int CurPage, int &xmin, int &ymin,
  int &xmax, int &ymax)
{
#undef INCLUDE_MIN_POINT
#define INCLUDE_MIN_POINT(x, y)                         \
  {                                                     \
    if (xmin > x) xmin = x;                             \
    if (ymin > y) ymin = y;                             \
  }
#undef INCLUDE_MAX_POINT
#define INCLUDE_MAX_POINT(x, y)                         \
  {                                                     \
    if (xmax < x) xmax = x;                             \
    if (ymax < y) ymax = y;                             \
  }

  if (ClearColor >= 0)
  {
    if (ClearPage == -1)
    {
      ClearPage = CurPage;
      Desync ();
    } else if (ClearPage == CurPage)
    {
      ClearColor = -1;
      goto sync;
    } /* endif */
    Gfx2D->Clear (ClearColor);
    INCLUDE_MIN_POINT (0, 0);
    INCLUDE_MAX_POINT (FRAME_WIDTH, FRAME_HEIGHT);
  } /* endif */

sync:
  int Page = CurPage - 1;
  Page = CurPage - 1;
  while (1)
  {
    if (Page < 0)
      Page = MaxPage;
    if (Page == CurPage)
      break;

    if (SyncArea [Page])
    {
      INCLUDE_MIN_POINT (SyncArea [Page]->x, SyncArea [Page]->y);
      INCLUDE_MAX_POINT (SyncArea [Page]->x + SyncArea [Page]->w,
        SyncArea [Page]->y + SyncArea [Page]->h);
      Gfx2D->RestoreArea (SyncArea [Page], false);
    }

    Page--;
  } /* endwhile */
}

int csGraphicsPipeline::TextWidth (char *text, int Font)
{
  if (FontList [Font].IndividualWidth)
  {
    int w = 0;
    for (int i = strlen (text); i > 0; i--, text++)
      w += FontList [Font].IndividualWidth[(unsigned char)*text];
    return w;
  } else
    return (strlen (text) * FontList [Font].Width);
}

int csGraphicsPipeline::TextHeight (int Font)
{
  return FontList [Font].Height;
}

bool csGraphicsPipeline::BeginDraw ()
{
  return SUCCEEDED (Gfx3D->BeginDraw (CSDRAW_2DGRAPHICS));
}

void csGraphicsPipeline::FinishDraw ()
{
  if (!RefreshRect.IsEmpty ())
    Gfx3D->Print (&RefreshRect);
  Gfx3D->FinishDraw ();
}
