/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef CSOBJECT_H
#define CSOBJECT_H

#include "geometry/math3d.h"
#include "engine/basic/csid.h"
#include "util/csbase.h"

class csScript;
class TriggerList;

// Crystal Space object type:
#define CS_SECTOR 0
#define CS_THING 1
#define CS_THINGTPL 2
#define CS_COLLECTION 3
#define CS_SCRIPT 4
#define CS_RUNSCRIPT 5
#define CS_POLYGON 6
#define CS_POLYPLANE 7
#define CS_STATLIGHT 8
#define CS_UNILIGHT 9
#define CS_DYNLIGHT 10
#define CS_FLTLIGHT 11
#define CS_TEXTURE 12
#define CS_SPRITE3D 13
#define CS_SPRITETPL 14
#define CS_LIBRARY 15
#define CS_WAVE 16

/**
 * The toplevel class for all objects in the Crystal
 * Space engine that can be manipulated by scripts. Currently the
 * supported csObjects are sectors, things and collections.<p>
 *
 * csObjects have names and a type (one of CS_...).
 * Activate triggers are supported for csObjects.<p>
 */
class csObject : public csBase
{
private:
  /// Unique ID for all CS objects.
  CS_ID id;

  /// Name of this object.
  char name[30];

  /// Type of this object (one of CS_...).
  int type;

public:
  /// Userdata for this object.
  void* userdata;

public:
  /// Create a new csObject with the given name and type.
  csObject (char* name, int type);

  ///
  virtual ~csObject () {}

  /**
   * Get the type of this csObject. This is one of CS_....
   */
  int get_type () { return type; }

  /**
   * Get the name of this csObject.
   */
  char* get_name () { return name; }
  
  /**
   * Get the unique ID of this csObject. This number is guaranteed
   * to be unique across all objects.
   */
  CS_ID get_id () { return id; }
};

/**
 * The toplevel class for all objects in Crystal Space that
 * need support for triggers (activation in the engine using
 * the space button currently).
 */
class csObjectTrigger : public csObject
{
private:
  /// List with all activation triggers for this object.
  TriggerList* activate_triggers;

public:
  /// Create a new csObjectTrigger with the given name and type.
  csObjectTrigger (char* name, int type);

  ///
  virtual ~csObjectTrigger ();

  /**
   * Add a new activate trigger to the list of triggers for the given
   * script and object.
   */
  void new_activate_trigger (csScript* script, csObject* ob = NULL);

  /**
   * Activate all the active triggers. This will cause the respective
   * scripts to be run.
   */
  void do_activate_triggers ();
};

#endif /*CSOBJECT_H*/
