/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/system.h"
#include "render/igraph3d.h"
#include "engine/basic/polyset.h"
#include "engine/polygon/polygon.h"
#include "engine/polygon/polytext.h"
#include "engine/texture.h"
#include "engine/light/dynlight.h"
#include "engine/world.h"
#include "engine/camera.h"
#include "engine/sector.h"
#include "geometry/bsp.h"
#include "engine/curve.h"
#include "engine/rview.h"
#include "engine/wirefrm.h"
#include "engine/stats.h"


//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( PolygonSet )

BEGIN_INTERFACE_TABLE( PolygonSet )
	IMPLEMENTS_COMPOSITE_INTERFACE( PolygonSet )
END_INTERFACE_TABLE()

PolygonSet::PolygonSet (char* name, int type) : PolygonParentInt (name, type)
{
  max_vertices = num_vertices = 0;
  max_polygon = num_polygon = 0;

  num_curves = max_curves = 0;
  curves = NULL;

  curves_center.x = curves_center.y = curves_center.z = 0;
  curves_scale=40;  
  curve_vertices = NULL;
  curve_texels = NULL;
  num_curve_vertices = max_curve_vertices = 0;

  wor_verts = NULL;
  obj_verts = NULL;
  cam_verts = NULL;
  polygons = NULL;
  bsp = NULL;
  draw_busy = 0;
  fog.enabled = false;
}

PolygonSet::~PolygonSet ()
{
  CHK (delete [] wor_verts);
  CHK (delete [] obj_verts);
  CHK (delete [] cam_verts);
  if (polygons)
  {
    for (int i = 0 ; i < num_polygon ; i++)
      CHKB (delete polygons [i]);
    CHK (delete [] polygons);
  }

  if (curves)
  {
    for (int i = 0 ; i < num_curves ; i++)
      CHKB (delete curves [i]);
    CHK (delete [] curves);
  }
  CHK (delete curve_vertices);
  CHK (delete curve_texels);

  CHK (delete bsp);
}


void PolygonSet::add_curve_vertex (Vector3& v, Vector2& t)
{
  if (!curve_vertices)
  {
    max_curve_vertices = 10;
    CHK (curve_vertices = new Vector3 [max_curve_vertices]);
    CHK (curve_texels   = new Vector2 [max_curve_vertices]);
  }
  while (num_curve_vertices >= max_curve_vertices)
  {
    max_curve_vertices += 10;
    CHK (Vector3* new_vertices = new Vector3 [max_curve_vertices]);
    CHK (Vector2* new_texels   = new Vector2 [max_curve_vertices]);
    memcpy (new_vertices, curve_vertices, sizeof (Vector3)*num_curve_vertices);
    memcpy (new_texels,   curve_texels,   sizeof (Vector2)*num_curve_vertices);
    CHK (delete [] curve_vertices);
    CHK (delete [] curve_texels);
    curve_vertices = new_vertices;
    curve_texels   = new_texels;
  }

  curve_vertices[num_curve_vertices] = v;
  curve_texels[num_curve_vertices] = t;
  num_curve_vertices++;
}

int PolygonSet::add_vertex (float x, float y, float z)
{
  if (!wor_verts)
  {
    max_vertices = 10;
    CHK (wor_verts = new Vector3 [max_vertices]);
    CHK (obj_verts = new Vector3 [max_vertices]);
    CHK (cam_verts = new Vector3 [max_vertices]);
  }
  while (num_vertices >= max_vertices)
  {
    max_vertices += 10;
    CHK (Vector3* new_wor_verts = new Vector3 [max_vertices]);
    CHK (Vector3* new_obj_verts = new Vector3 [max_vertices]);
    CHK (Vector3* new_cam_verts = new Vector3 [max_vertices]);
    memcpy (new_wor_verts, wor_verts, sizeof (Vector3)*num_vertices);
    memcpy (new_obj_verts, obj_verts, sizeof (Vector3)*num_vertices);
    memcpy (new_cam_verts, cam_verts, sizeof (Vector3)*num_vertices);

    CHK (delete [] wor_verts);
    CHK (delete [] obj_verts);
    CHK (delete [] cam_verts);

    wor_verts = new_wor_verts;
    obj_verts = new_obj_verts;
    cam_verts = new_cam_verts;
  }

  // By default all vertices are set with the same object space and world space.
  wor_verts[num_vertices].set (x, y, z);
  obj_verts[num_vertices].set (x, y, z);
  num_vertices++;
  return num_vertices-1;
}

int PolygonSet::add_vertex_smart (float x, float y, float z)
{
  if (!wor_verts) { add_vertex (x, y, z); return 0; }
  int i;
  for (i = 0 ; i < num_vertices ; i++)
    if (ABS (x-obj_verts[i].x) < SMALL_EPSILON &&
	ABS (y-obj_verts[i].y) < SMALL_EPSILON &&
	ABS (z-obj_verts[i].z) < SMALL_EPSILON)
    {
      return i;
    }
  add_vertex (x, y, z);
  return num_vertices-1;
}

Polygon3D* PolygonSet::get_polygon (char* name)
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
  {
    Polygon3D* p = (Polygon3D*)polygons[i];
    if (!strcmp (p->get_name (), name)) return p;
  }
  return NULL;
}

Polygon3D* PolygonSet::new_polygon (char* name, TextureMM* texture)
{
  CHK (Polygon3D* p = new Polygon3D (name, texture));
  p->set_sector (sector);
  add_polygon (p);
  return p;
}

void PolygonSet::add_polygon (PolygonInt* poly)
{
  if (!polygons)
  {
    max_polygon = 6;
    CHK (polygons = new PolygonInt* [max_polygon]);
  }
  while (num_polygon >= max_polygon)
  {
    max_polygon += 6;
    CHK (PolygonInt** new_polygons = new PolygonInt* [max_polygon]);
    memcpy (new_polygons, polygons, sizeof (PolygonInt*)*num_polygon);
    CHK (delete [] polygons);
    polygons = new_polygons;
  }

  polygons[num_polygon++] = poly;
  poly->set_parent (this);
}


csCurve* PolygonSet::get_curve (char* name)
{
  int i;
  for (i = 0 ; i < num_curves ; i++)
  {
    csCurve* p = (csCurve*)polygons[i];
    if (!strcmp (p->get_name (), name)) return p;
  }
  return NULL;
}

/*TODO weg?

csCurve* PolygonSet::new_bezier (char* name, TextureMM* texture)
{
  CHK (csBezier* p = new csBezier (name, 0x1234));
  p->setTexture(texture);
  //TODO??  p->set_sector (sector);
  add_curve (p);
  return p;
}
*/

void PolygonSet::add_curve (csCurve* curve)
{
  if (!curves)
  {
    max_curves = 6;
    CHK (curves = new csCurve* [max_curves]);
  }
  while (num_curves >= max_curves)
  {
    max_curves += 6;
    CHK (csCurve** new_curves = new csCurve* [max_curves]);
    memcpy (new_curves, curves, sizeof (csCurve*)*num_curves);
    CHK (delete [] curves);
    curves = new_curves;
  }
  curves[num_curves++] = curve;
  curve->set_parent (this);
}


void PolygonSet::use_bsp ()
{
  CHK (delete bsp);
  CHK (bsp = new BspTree (this));
}

struct IntersInfo
{
  const Vector3* start;
  const Vector3* end;
  Vector3* isect;
  float* pr;
};

void* test_bsp_intersection (PolygonParentInt*, PolygonInt** polygon, int num,
	void* data)
{
  IntersInfo* d = (IntersInfo*)data;
  int i;
  for (i = 0 ; i < num ; i++)
  {
    Polygon3D* p = (Polygon3D*)polygon[i];
    if (p->intersect_segment (*d->start, *d->end, *d->isect, d->pr))
      return (void*)p;
  }
  return NULL;
}

Polygon3D* PolygonSet::intersect_segment (const Vector3& start, 
        const Vector3& end, Vector3& isect, float* pr)
{
  int i;
  if (bsp)
  {
    IntersInfo d;
    d.start = &start;
    d.end = &end;
    d.isect = &isect;
    d.pr = pr;
    return (Polygon3D*) bsp->front2back (start, &test_bsp_intersection, (void*)&d);
  }
  else
    for (i = 0 ; i < num_polygon ; i++)
    {
      Polygon3D* p = (Polygon3D*)polygons[i];
      if (p->intersect_segment (start, end, isect, pr)) return p;
    }
  return NULL;
}

void PolygonSet::translate_vector (Vector3& trans)
{
  int i;
  for (i = 0 ; i < num_vertices ; i++)
    cam_verts[i] = wor_verts[i]-trans;
}

bool PolygonSet::transform_world2cam (csCamera& c)
{
  int i;

  for (i = 0 ; i < num_vertices ; i++)
    cam_verts[i] = c.world2this(wor_verts[i]);

  return true;
}

Vector3* PolygonSet::new_transformation ()
{
  if (draw_busy == 1)
    // This is the first time we need a transformation this session.
    // Just return NULL and use the default cam_verts array.
    return NULL;
  else
  {
    // Allocate a new cam_verts array.
    Vector3* old = cam_verts;
    CHK (cam_verts = new Vector3 [max_vertices]);
    return old;
  }
}

void PolygonSet::restore_transformation (Vector3* old)
{
  if (old)
  {
    if (cam_verts) CHKB (delete [] cam_verts);
    cam_verts = old;
  }
}


void PolygonSet::draw_polygon_array (PolygonInt** polygon, int num, RenderView* d,
	bool use_z_buf)
{
  Polygon3D* p;
  Portal* po;
  Vector3* verts;
  int num_verts;
  int i;
  
  WireFrame* wf = NULL;
  if (System->world->map_mode != MAP_OFF) wf = System->world->wf->get_wireframe ();
  
  for (i = 0 ; i < num ; i++)
  {
    Vector2 orig_triangle[3];
    p = (Polygon3D*)polygon[i];
    if ( !p->dont_draw &&
         p->clip_to_plane (d->clip_plane, d->v_origin(), verts, num_verts) && 
         p->do_perspective (*d, verts, num_verts, &Polygon2D::clipped, orig_triangle)  &&
         Polygon2D::clipped.clip_against (d->view) )
    {
      if (wf)
      {
        int j;
        wfPolygon* po = wf->add_polygon ();
        po->set_vis_color (wf->get_yellow ());
        po->set_num_vertices (p->get_num_vertices ());
        for (j = 0 ; j < p->get_num_vertices () ; j++) po->set_vertex (j, p->vwor (j));
        po->prepare ();
      }

      Stats::polygons_drawn++;

      po = p->get_portal ();
      if (csSector::do_portals && po)
      {
	bool filtered = false;
	// is_dest_fog is true if the portal points to a fogged sector.
	// is_this_fog is true if this sector is fogged itself.
	// fog_id is the id of the destination sector.
	bool is_dest_fog = false, is_this_fog = sector->has_fog ();
	CS_ID fog_id;
	if (po->portal_type () == PORTAL_CS)
	{
	  PortalCS* pcs = (PortalCS*)po;
	  is_dest_fog = pcs->get_sector () ? pcs->get_sector ()->has_fog () : false;
	  if (is_dest_fog)
	  {
	    fog_id = pcs->get_sector ()->get_id ();
	    Gfx3D->OpenFogObject (fog_id, &pcs->get_sector ()->get_fog ());
	  }
	}
              
        // If there is filtering (alpha mapping or something like that) we need to keep the
        // 2D polygon and texture plane so that it can be drawn after the sector has been drawn.
        // We can't just use 'clipped' because calling Portal->Draw will (possibly)
        // reuse the clipped global variable.
        // The texture plane needs to be kept because this polygon may be rendered again
        // (through mirrors) possibly overwriting the plane.
        Polygon2D* keep_clipped = NULL;
        PolyPlane* keep_plane = NULL;
              
        long do_transp;
        Gfx3D->GetRenderState (G3DRENDERSTATE_TRANSPARENCYENABLE, do_transp);
	if (do_transp)
          filtered = p->get_polytex (0)->get_texture ()->get_parent ()->get_transparent () != -1 || p->get_alpha ();
              
        if (filtered || is_dest_fog || is_this_fog)
        {
          CHK (keep_clipped = new Polygon2D (Polygon2D::clipped));
          CHK (keep_plane = new PolyPlane (*(p->get_plane ())));
        }
              
        // Draw through the portal. If this fails we draw the original polygon
        // instead. Drawing through a portal can fail because we have reached
        // the maximum number that a sector is drawn (for mirrors).
	// Note that use_z_buf is given to this routine as well but this
	// parameter has nothing to do with the Z buffer but with the fact that
	// we have a portal in the middle of a sector. use_z_buf happens to be
	// true for Things and false for sectors so this works now.
        if (po->draw (&Polygon2D::clipped, &p->get_plane ()->get_camera_plane (), use_z_buf, *d))
        {
	  if (is_dest_fog) keep_clipped->add_fog_polygon (p, keep_plane, d->is_mirrored (), fog_id, CS_FOG_FRONT);
          if (filtered)
	    keep_clipped->draw_filled (p, keep_plane, d->is_mirrored (),
	    	use_z_buf, orig_triangle);
	  if (is_this_fog) keep_clipped->add_fog_polygon (p, keep_plane, d->is_mirrored (), sector->get_id (), CS_FOG_BACK);
        }
        else Polygon2D::clipped.draw_filled (p, p->get_plane (), d->is_mirrored (),
		use_z_buf, orig_triangle);
              
        // Cleanup.
        if (keep_clipped)
        {
          CHK (delete keep_clipped);
          CHK (delete keep_plane);
        }
	if (is_dest_fog)
	{
	  Gfx3D->CloseFogObject (fog_id);
	}
      }
      else Polygon2D::clipped.draw_filled (p, p->get_plane (), d->is_mirrored (),
      	use_z_buf, orig_triangle);
          
      long do_edges;
      Gfx3D->GetRenderState (G3DRENDERSTATE_EDGESENABLE, do_edges);
      if (do_edges) Polygon2D::clipped.draw (System->world->get_textures ()->white ());
    }
  }
}


//---------------------------------------------------------------------------
