/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef CAMERA_H
#define CAMERA_H

#include "com/com.h"
#include "geometry/math3d.h"
#include "engine/icamera.h"

#define VEC_FORWARD   Vector3(0,0,1)
#define VEC_BACKWARD  Vector3(0,0,-1)
#define VEC_RIGHT     Vector3(1,0,0)
#define VEC_LEFT      Vector3(-1,0,0)
#define VEC_UP        (-Vector3(0,-1,0))
#define VEC_DOWN      (-Vector3(0,1,0))

#define VEC_ROT_RIGHT  Vector3(0,1,0)
#define VEC_ROT_LEFT   Vector3(0,-1,0)
#define VEC_TILT_RIGHT (-Vector3(0,0,1))
#define VEC_TILT_LEFT  (-Vector3(0,0,-1))
#define VEC_TILT_UP    (-Vector3(1,0,0))
#define VEC_TILT_DOWN  (-Vector3(-1,0,0))

class csSector;
class Polygon3D;
class Vertex;
class csWorld;
class Dumper;

/**
 * A camera positioned in the 3D world.
 */
class csCamera : public OrthoTransform
{
  friend class Dumper;

private:
  /// The sector the camera is in.
  csSector* sector;
  /// If true we are in a mirrored world.
  bool mirror;

public:
  ///
  static int aspect;
  ///
  static float inv_aspect;

public:
  ///
  csCamera ();
  ///
  virtual ~csCamera ();

  /// Save the camera position to a file.
  void save_file (char* filename);
  /// Load the camera position from a file.
  bool load_file (csWorld* world, char* filename);

  /**
   * Check if there is a polygon in front of us within a distance
   * of v.
   */
  Polygon3D* get_hit (Vector3& v);

  /// Set the sector that the camera resides in.
  void set_sector (csSector *s) { sector = s; }
  /// Get the current sector of the camera.
  csSector* get_sector () { return sector; }

  /// Returns true if we are in a mirrored world.
  bool is_mirrored () { return mirror; }
  /// Set the mirrored state of this camera.
  void set_mirror (bool m) { mirror = m; }

  /// Sets the absolute position of the camera inside the sector
  inline void set_position (const Vector3& v) { set_origin(v); }
  /// Set the world to camera transformation matrix
  inline void set_world2cam(const Matrix3& m) { set_world2this(m); } 
  /// Set the camera to world transformation matrix
  inline void set_cam2world(const Matrix3& m) { set_this2world(m); } 

  /// Return the world to camera transformation matrix
  inline Matrix3 m_world2cam() const { return m_world2this(); }
  /// Return the camera to world transformation matrix
  inline Matrix3 m_cam2world() const { return m_this2world(); }
  /// Return the world to camera translation
  inline Vector3 v_world2cam() const { return v_world2this(); }

  /// Transform a worldspace point to camera space
  inline Vector3 world2cam(const Vector3& v) const { return world2this(v); }
  /// Transform a camera space point to worldspace
  inline Vector3 cam2world(const Vector3& v) const { return this2world(v); }
  /// Transform a camera space point to worldspace, relative to camera position
  inline Vector3 cam2world_rel(const Vector3& v) const 
  { return this2world_rel(v); }

  /**
   * Moves the camera a relative amount in world coordinates.
   */
  void move_w (const Vector3& v);

  /**
   * Moves the camera a relative amount in camera coordinates.
   */
  void move (const Vector3& v) { move_w(m_t2w * v); }

  /**
   * Moves the camera a relative amount in world coordinates,
   * ignoring portals and walls
   */
  void moveip_w (const Vector3& v) { v_w2t += v; }

  /**
   * Moves the camera a relative amount in camera coordinates,
   * ignoring portals and walls
   */
  void moveip (const Vector3& v) { v_w2t += m_t2w * v; }

  /**
   * Rotate the camera by the angle (radians) around the given vector,
   * in world coordinates.
   * Note: rotates the camera, not the coordinate system
   */
  void rotate_w (const Vector3& v, float angle);

  /**
   * Rotate the camera by the angle (radians) around the given vector,
   * in camera coordinates.
   * Note: rotates the camera, not the coordinate system
   */
  void rotate (const Vector3& v, float angle);

  /**
   * Use the given transformation matrix, in worldspace,
   * to reorient the camera
   * Note: rotates the camera, not the coordinate system
   */
  void rotate_w (const Matrix3& m) { set_this2world(m * m_t2w); }

  /**
   * Use the given transformation matrix, in camera space,
   * to reorient the camera.
   * Note: rotates the camera, not the coordinate system
   */
  void rotate (const Matrix3& m) { set_this2world(m_t2w * m); }

  /**
   * Have the camera look at the given (x,y,z) point, using up as
   * the up-vector
   */
  void look_at (const Vector3& v, const Vector3& up);

  /**
   * Eliminate roundoff error by snapping the camera orientation to a
   * grid of density n
   */
  void correct (int n);

private:
  ///
  void correct (int n, float* vals[]);

  DECLARE_INTERFACE_TABLE( csCamera )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( Camera )
};

#define GetICameraFromCamera(a)  &a->m_xCamera
#define GetCameraFromICamera(a)  ((csCamera*)((size_t)a - offsetof(csCamera, m_xCamera)))

#endif /*CAMERA_H*/
