/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef CURVE_H
#define CURVE_H

#include "com/com.h"
#include "util/def.h"
#include "geometry/math3d.h"
#include "geometry/math2d.h"
#include "engine/basic/csobject.h"

class TextureMM;

/**
 * Vertex resulting from a tesselated curve.
 */
class csCurveVertex
{
public:
  /// Object space coordinates.
  Vector3 object_coord;
  /// World space coordinates.
  Vector3 world_coord;
  /// Camera space coordinates.
  Vector3 camera_coord;
  /// Texture coordinates.
  Vector2 txt_coord;
};

/**
 * Triangle resulting from a tesselated curve.
 */
class csCurveTriangle
{
public:
  /// Indices in triangle list.
  int i1, i2, i3;
};

/**
 * Tesselated curve. This is basicly a list of triangles.
 */
class csCurveTesselated
{
private:
  csCurveVertex* vertices;
  int num_vertices;
  csCurveTriangle* triangles;
  int num_triangles;

public:
  /**
   * Allocate a new tesselated curve with the given
   * number of vertices and triangles.
   */
  csCurveTesselated (int num_v, int num_t);
  ///
  ~csCurveTesselated ();

  ///
  int getNumVertices () { return num_vertices; }
  ///
  int getNumTriangles () { return num_triangles; }
  ///
  csCurveVertex& getVertex (int i) { return vertices[i]; }
  ///
  csCurveTriangle& getTriangle (int i) { return triangles[i]; }
};

/**
 * This is an abstract class for all curves in Crystal Space.
 */
class PolygonSet;
class ThingTemplate;
class csCurve : public csObject
{
private:
  TextureMM* texture;

public:
  PolygonSet* parent;

public:
  csCurve(char* name, int type);
  ///
  virtual ~csCurve () { }

  void set_parent (PolygonSet* p) { parent = p; }
  /**
   * Tesselate this curve with the given resolution.
   * This function will allocated and return a csCurveTesselated object.
   */
  virtual csCurveTesselated* tesselate (int res) = 0;
  /**
   * set control index for a control point (referring to the controls
   * in the parent PolygonSet)
   */
  virtual void set_control_point(int index, int control_id) = 0;
  ///
  TextureMM* getTexture () { return texture; }

  ///
  void setTexture (TextureMM* t) { texture = t; }
};

inline csCurve::csCurve(char* name, int type)
  : csObject (name, type),texture(NULL),parent(NULL) {}; 

/**
 * A curve template.
 */
class csCurveTemplate : public csObject
{
protected:
  ThingTemplate *parent;

  TextureMM* texture;

public:
  csCurveTemplate(char* name, int type);

  virtual void set_parent(ThingTemplate *p) { parent=p; }

  virtual csCurve* makeCurve(char* name, int type) = 0;

  /// Tesselate this curve.
  virtual void set_vertex(int index, int ver_ind) = 0;
  ///
  virtual int get_vertex(int index)  = 0;
  virtual int num_vertices() = 0;
  ///
  TextureMM* get_texture () { return texture; }

  ///
  void set_texture (TextureMM* t) { texture = t; }
};


/**
 * A specific curve implementation for Bezier curve template.
 */
class csBezierTemplate : public csCurveTemplate
{
private:
  int ver_id[9];

public:
  csBezierTemplate(char* name, int type);

  virtual csCurve* makeCurve(char* name, int type);

  /// Tesselate this curve.
  virtual void set_vertex(int index, int ver_ind);
  ///
  virtual int get_vertex(int index);
  virtual int num_vertices();
};




/**
 * A specific curve implementation for Bezier curves.
 */
class csBezier : public csCurve
{
private:
  /// The control points of this curve.
  Vector3 points[3][3];
  /// The texture coordinates of every control point.
  Vector2 texture_coords[3][3];

  csCurveTesselated* previous_tesselation;
  int previous_resolution;

public:
  csBezier (char* name, int type);
  ~csBezier ();

  /// Tesselate this curve.
  virtual csCurveTesselated* tesselate (int res);

  /// Load a curve from disk.
  void load (char* buf);

  virtual void set_control_point(int index, int control_id);

  /// Get a curve point.
  inline Vector3& getControlPoint (int i) { return points[i/3][i-(i/3)*3]; }

  /// Get the texture coordinate of a curve point.
  inline Vector2& getTextureCoord (int i) { return texture_coords[i/3][i-(i/3)*3]; }
};


#endif /*CURVE_H*/

