
#include "com/com.h"
#include "util/types.h"
#include "util/gfx/csimage.h"

#ifndef __ITEXTURE_H__
#define __ITEXTURE_H__

/// The internal texture mapping modes.
#define TXT_GLOBAL 0    // Textures are mapped with a single palette
#define TXT_PRIVATE 1   // Every texture has it's own 256-color palette
#define TXT_24BIT 2     // Every texture is represented in 24-bits (RGB)

// External data types.
struct HighColorCache_Data;

// Forward declarations
interface IMipMapContainer;
interface ITextureContainer;
interface ITextureMap;
interface IWorld;

/// Interfaces ////

extern const IID IID_IMipMapContainer;

/** 
 *  IMipMapContainer</p>
 *  The container for a texture and it's mip-maps.
 */
interface IMipMapContainer : public IUnknown
{
  DECLARE_IUNKNOWN()

  /// Whether this texture is cached in video memory.
  STDMETHOD(GetInVideoMemory) ( bool& retval );
  /// Notify the texture that it is entering/leaving video memory.
  STDMETHOD(SetInVideoMemory) ( bool newval );

  /** Returns the HighColorCache_Data*, which is information about how  
   *  the texture is stored in video memory.
   */
  STDMETHOD(GetHighColorCache) (HighColorCache_Data** retval);
  
  /// Sets the HighColorCache_Data for this texture.
  STDMETHOD(SetHighColorCache) (HighColorCache_Data* newval);

  /// Blend mipmap level 0.
  STDMETHOD(BlendMipmapZero)(ITextureContainer* tex);

  /// Create all mipmapped bitmaps from the first level.
  STDMETHOD(CreateMipmaps)(ITextureContainer* tex);

  /// Free all information that is only needed for loading the image.
  STDMETHOD(FreeLoadInformation)();

  /** Return S_OK if the texture has been loaded correctly,
   *  S_FALSE otherwise.
   */
  STDMETHOD(LoadedCorrectly)();

  /// Get the mipmapped texture at some level (0..3).
  STDMETHOD(GetTexture)(int lev, ITextureMap** retval);

  /// Get the private colormap if any.
  STDMETHOD(GetPrivateColorMap)(unsigned char** retval);

  /// Get the conversion table from private to global colormap if any.
  STDMETHOD(GetPrivateToGlobalConversion)(unsigned char** retval);

  /// Set the transparent color.
  STDMETHOD(SetTransparent)(int red, int green, int blue);

  /// Get the transparent index (-1 if no transparency, 0 if transparency).
  STDMETHOD(GetTransparent)(int& retval);

  /// Get the mean color.
  STDMETHOD(GetMeanColor)(int& retval);

  /// Get the average red value.
  STDMETHOD(GetMeanRed)(float& retval);
  /// Get the average green value.
  STDMETHOD(GetMeanGreen)(float& retval);
  /// Get the average blue value.
  STDMETHOD(GetMeanBlue)(float& retval);

  /// Returns the number of colors in this texture.
  STDMETHOD(GetNumberOfColors)(int& retval);

  /// Compute the 'usage' table.
  STDMETHOD(ComputeColorUsage)();

  /// Get the usage table.
  STDMETHOD(GetColorUsage)( int idx, RGBPalEntry& retval );
  
  /**
   * Remap the palette of this texture according to the internal
   * texture mode.
   */
  STDMETHOD(RemapPalette)(ITextureContainer* new_palette);

  /**
   * Remap the palette of this texture using the given global
   * palette (inside Textures).
   */
  STDMETHOD(RemapPaletteGlobal)(ITextureContainer* new_palette);

  /**
   * Remap the palette of this texture using a private
   * palette and make a mapping of this palette to RGB
   * and to the global palette (inside Textures).
   */
  STDMETHOD(RemapPalettePrivate)(ITextureContainer* new_palette);

  /// Get the name of this texture.
  STDMETHOD(GetName)(char** retval);
};

typedef UShort RGB16map[256];
typedef unsigned char RGB8map[256];

// Colors are encoded in a 16-bit short using the following
// distribution (only for 8-bit mode):
#define BITS_RED 6
#define BITS_GREEN 6
#define BITS_BLUE 4
#define MASK_RED ((1<<BITS_RED)-1)
#define MASK_GREEN ((1<<BITS_GREEN)-1)
#define MASK_BLUE ((1<<BITS_BLUE)-1)
#define NUM_RED (1<<BITS_RED)
#define NUM_GREEN (1<<BITS_GREEN)
#define NUM_BLUE (1<<BITS_BLUE)

#define TABLE_WHITE 0
#define TABLE_RED 1
#define TABLE_GREEN 2
#define TABLE_BLUE 3
#define TABLE_WHITE_HI 4
#define TABLE_RED_HI 5
#define TABLE_GREEN_HI 6
#define TABLE_BLUE_HI 7

#define MIPMAP_UGLY 0
#define MIPMAP_DEFAULT 1
#define MIPMAP_NICE 2
#define MIPMAP_VERYNICE 3

#define MIX_TRUE_RGB 0
#define MIX_FAST_WRB 1
#define MIX_FAST_WRG 2
#define MIX_FAST_WBG 3
#define MIX_NOCOLOR 4

/**
 * Lookup table entry corresponding to one palette entry.
 * 'red', 'green', and 'blue' are tables giving the red, green,
 * and blue components for all light levels of that palette index.
 */
struct PalIdxLookup
{
  RGB16map red;
  RGB16map green;
  RGB16map blue;
};

/**
 * Lookup table class for true_rgb mode
 * (used for display output of 8 and 16-bit and texture width
 * of 8-bit and with true_rgb mode).
 * The tables in this class convert an 8-bit palette index
 * to a red, green, or blue 16-bit value (using the specific
 * mask for each color). The result of these three tables
 * can then be or'ed together for a 16-bit truecolor value
 * which can then in turn be translated (using another lookup
 * table) to 8-bit if needed.
 */
struct TextureTablesTrueRgb
{
  /// Lookup table.
  PalIdxLookup lut[256];
};

/**
 * Lookup table class for true_rgb mode and TXT_PRIVATE textures.
 * (used for display output of 8 and 16-bit and texture width
 * of 8-bit/private colormap and with true_rgb mode).
 * The tables in this class convert an 8-bit color component
 * to a red, green, or blue 16-bit value (using the specific
 * mask for each color). The result of these three tables
 * can then be or'ed together for a 16-bit truecolor value
 * which can then in turn be translated (using another lookup
 * table) to 8-bit if needed.
 */
struct TextureTablesTrueRgbPriv
{
  /// Lookup table.
  PalIdxLookup lut[256];
};

/**
 * Lookup tables which translate a palette entry to another
 * palette entry with the right amount of colored light added.
 * (used for display output of 8-bit and texture width of 8-bit
 * and with any of the FAST_xxx modes).
 */
struct TextureTablesFastXxx
{
  /// The first light table.
  RGB8map light1[256];
  /// The second light table.
  RGB8map light2[256];
  /// The third light table.
  RGB8map light3[256];
};

/**
 * Two tables with white light. They add white light to a palette
 * index and return a 16-bit truecolor value.
 * (used for display output of 16-bit and texture width of 8-bit).
 */
struct TextureTablesWhite16
{
  /// White light table with index 0 as default (no light added).
  RGB16map white1_light[256];
  /// White light table with index 200 as default (no light added).
  RGB16map white2_light[256];
};

/**
 * Two tables with white light. They add white light to a palette
 * index and return a new 8-bit palette index.
 * (used for display output of 8-bit and texture width of 8-bit).
 */
struct TextureTablesWhite8
{
  /// White light table with index 0 as default (no light added).
  RGB8map white1_light[256];
  /// White light table with index 200 as default (no light added).
  RGB8map white2_light[256];
};

/**
 * Lookup table to convert a 16-bit truecolor value to
 * an 8-bit palette index and the other way around.
 * (used for display output of 8-bit/16-bit and texture width of 8-bit).
 */
struct TextureTablesPalette
{
  /// Lookup table for true_rgb mode.
  unsigned char true_to_pal[65536];
  /// From palette to truecolor value.
  UShort pal_to_true[256];
};

/**
 * Lookup table for alpha mapping. Converts two palette entries
 * to a new palette entry (alpha blended).
 * (used for display output of 8-bit).
 */
struct TextureTablesAlpha
{
  /// Alpha table for 50%
  RGB8map alpha_map50[256];
  /// Alpha table for 25% and 75%
  RGB8map alpha_map25[256];
};


extern const IID IID_ITextureContainer;
/**
 * ITextureContainer</p>
 * The interface for the collection of all textures and also maintains
 * the shared palette.
 */
interface ITextureContainer : public IUnknown
{

   DECLARE_IUNKNOWN()

  /**
   * Initialize the texture system. This function has to be called
   * AFTER the general system has been set up (it needs to know
   * the depth of the display and the bitformat of the RGB pixels
   * in 15/16 bit truecolor mode). After you have called this
   * function you can call alloc_palette().<br>
   * Returns true if successful.
   */
  STDMETHOD(Initialize)(IWorld* world);

  /**
   * Initialize the texture system in a limited way. This is
   * useful if you want to use the texture system for unlighted
   * drawing only.
   */
  STDMETHOD(InitializeLimited)();

  /**
   * Find an rgb value using the palette directly (not use
   * the faster lookup tables).
   */
  STDMETHOD(FindRGBPrecise)(int r, int g, int b, int& retval);

  /**
   * Allocate a new RGB color.
   */
  STDMETHOD(AllocateNewRGB)(int r, int g, int b, int dist);

  /// Find a texture given a name.
  STDMETHOD(GetMipMapContainer)(char* name, IMipMapContainer** retval);

  /// Find a texture given an idx.
  STDMETHOD(GetMipMapContainerByIndex)(int idx, IMipMapContainer** retval);

  /**
   * Find an rgb value using the faster lookup tables.
   */
  STDMETHOD(FindRGB)(int r, int g, int b, int& retval);

  /**
   * Return the index for some color. This works in 8-bit
   * (returns an index in the 256-color table) and in 15/16-bit
   * (returns a 15/16-bit encoded RGB value).
   */
  STDMETHOD(FindColor)(int r, int g, int b, int& retval);

  /**
   * This version of find_rgb finds some r,g,b value AFTER gamma
   * correction is applied. This is useful for console messages
   * that always need the same color regardless of the gamma
   * correction.
   */
  STDMETHOD(FindRealRGB)(int r, int g, int b, int& retval);

  /**
   * Compute the 'best' palette for all loaded textures.
   * This function will exactly behave the same in 16-bit mode
   * since we still need the common 256-color palette (although
   * it will not be installed on the display).
   */
  STDMETHOD(ComputePalette)();

  /**
   * Compute the light tables using the previously compute palette.
   */
  STDMETHOD(ComputeLightTables)(IWorld* world);

  /**
   * Really allocate the palette on the system.
   */
  STDMETHOD(AllocatePalette)();

  /**
   * Get the palette.
   */
  STDMETHOD(GetPalette)(RGBcolor** retval);

  ///
  STDMETHOD(GetPaletteAllocated)(int idx, int& retval);
  ///
  STDMETHOD(GetRedPalette)(int idx, int& retval);
  ///
  STDMETHOD(GetGreenPalette)(int idx, int& retval);
  ///
  STDMETHOD(GetBluePalette)(int idx, int& retval);

  ///
  STDMETHOD(GetRedLevel)(int& retval);
  ///
  STDMETHOD(GetGreenLevel)(int& retval);
  ///
  STDMETHOD(GetBlueLevel)(int& retval);

  /**
   * Load all the texture descriptions from the world
   * file (no actual images).
   */
  STDMETHOD(Load)(char* buf, IWorld* world);

  ///
  STDMETHOD(GetRed)(int& retval);
  ///
  STDMETHOD(GetBlue)(int& retval);
  ///
  STDMETHOD(GetYellow)(int& retval);
  ///
  STDMETHOD(GetGreen)(int& retval);
  ///
  STDMETHOD(GetWhite)(int& retval);
  ///
  STDMETHOD(GetBlack)(int& retval);

  /**
   * GAC: Use a slower method of mixing lights that produces a better result
   * Accepts 16.16 rgb values and a palette entry
   * Returns palette entry of color with lights added
   * Assumes lights are RGB format (Not WRB).
   */
  STDMETHOD(MixLights)(int r, int g, int b, int p, unsigned char& retval);
  
  /**
   * Add some light to some light component. This is used for TXT_PRIVATE
   * mode. This returns a 16-bit truecolor component for the specific color.
   */
  STDMETHOD(AddRedLightPrivate)(int r, int l, int& retval);

  ///
  STDMETHOD(AddGreenLightPrivate)(int g, int l, int& retval);

  ///
  STDMETHOD(AddBlueLightPrivate)(int b, int l, int& retval);

  /**
   * Version of mix_lights for 16-bit mode. This is slightly faster since
   * the last lookup is not needed.
   */
  STDMETHOD(MixLights16Bit)(int r, int g, int b, int p, unsigned short& retval);
  /// Lookup table.
  STDMETHOD(GetPaletteLUT)(TextureTablesPalette** retval);
  /// Lookup table.
  STDMETHOD(GetAlphaLUT)(TextureTablesAlpha** retval);
  /// Lookup table.
  STDMETHOD(GetTrueRGBLUT)(TextureTablesTrueRgb** retval);
  /// Lookup table.
  STDMETHOD(GetPrivateRGBLUT)(TextureTablesTrueRgbPriv** retval);
  /// Lookup table.
  STDMETHOD(GetCurrentRGBLUT)(PalIdxLookup** retval);
  /// Lookup table.
  STDMETHOD(GetFastXXXLUT)(TextureTablesFastXxx** retval);
  /// Lookup table.
  STDMETHOD(Get16BitWhiteLUT)(TextureTablesWhite16** retval);
  /// Lookup table.
  STDMETHOD(Get8BitWhiteLUT)(TextureTablesWhite8** retval);
  /// Gets the texture-palette scheme - global or private.
  STDMETHOD(GetMode)(int& retval);
  /// Gets the mixing type.
  STDMETHOD(GetMixingType)(int& retval);
  /// Gets the alpha mask.
  STDMETHOD(GetAlphaMask)(unsigned short& retval);
  /// Are we only rendering lightmaps in the cache? (@@@Move flag to cache)
  STDMETHOD(GetLightMapsOnly)(bool& retval);
  /// Render grid on top of textures in cache? (@@@Move flag to cache)
  STDMETHOD(GetLightMapsGrid)(bool& retval);
};

/**
 * Define a small (3 pixels) margin at the bottom and top of
 * the texture. This is the easiest way I could find to 'fix' the
 * overflow problems with the texture mapper.
 */
#define H_MARGIN 3

extern const IID IID_ITextureMap;
/// ITextureMap -- Interface for a singular texture map.
interface ITextureMap : public IUnknown
{
  /// Gets the bits of the texture.
  STDMETHOD(GetBitmap) (unsigned char** retval) = 0;
  /// Gets the width of the texture.
  STDMETHOD(GetWidth) (int& retval) = 0;
  /// Gets the height of the texture
  STDMETHOD(GetHeight) (int& retval) = 0;
  /// Gets the width bit shift of the texture.
  STDMETHOD(GetWidthShift) (int& retval) = 0;
  /// Gets the height bit shift of the texture.
  STDMETHOD(GetHeightShift) (int& retval) = 0;
  /// Gets the width mask of the texture.
  STDMETHOD(GetWidthMask) (int& retval) = 0;
  /// Gets the height mask of the texture.
  STDMETHOD(GetHeightMask) (int& retval) = 0;
  /// Returns the mip-map container of this texture.
  STDMETHOD(GetParent) (IMipMapContainer** retval) = 0;
};


class Texture8;
class Texture32;

/// composite implementation for 8-bit singular texturemap.
class XTextureMap8 : public ITextureMap
{
public:
  XTextureMap8( Texture8* pThis )
  {
      m_pThis = pThis;
  }

  DECLARE_IUNKNOWN()
  /// Gets the bits of the texture.
  STDMETHOD(GetBitmap) (unsigned char** retval);
  /// Gets the width of the texture.
  STDMETHOD(GetWidth) (int& retval);
  /// Gets the height of the texture
  STDMETHOD(GetHeight) (int& retval);
  /// Gets the width bit shift of the texture.
  STDMETHOD(GetWidthShift) (int& retval);
  /// Gets the height bit shift of the texture.
  STDMETHOD(GetHeightShift) (int& retval);
  /// Gets the width mask of the texture.
  STDMETHOD(GetWidthMask) (int& retval);
  /// Gets the height mask of the texture.
  STDMETHOD(GetHeightMask) (int& retval);
  /// Returns the mip-map container of this texture.
  STDMETHOD(GetParent) (IMipMapContainer** retval);

private:

    Texture8* m_pThis;
};

/// XTextureMap32 -- composite implementation for 16-bit singular texturemap.
class XTextureMap32 : public ITextureMap
{
public:
  XTextureMap32( Texture32* pThis )
  {
      m_pThis = pThis;
  }

  DECLARE_IUNKNOWN()
  /// Gets the bits of the texture.
  STDMETHOD(GetBitmap) (unsigned char** retval);
  /// Gets the width of the texture.
  STDMETHOD(GetWidth) (int& retval);
  /// Gets the height of the texture
  STDMETHOD(GetHeight) (int& retval);
  /// Gets the width bit shift of the texture.
  STDMETHOD(GetWidthShift) (int& retval);
  /// Gets the height bit shift of the texture.
  STDMETHOD(GetHeightShift) (int& retval);
  /// Gets the width mask of the texture.
  STDMETHOD(GetWidthMask) (int& retval);
  /// Gets the height mask of the texture.
  STDMETHOD(GetHeightMask) (int& retval);
  /// Returns the mip-map container of this texture.
  STDMETHOD(GetParent) (IMipMapContainer** retval);

private:

    Texture32* m_pThis;
};


#endif
