/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "engine/light/lghtmap.h"
#include "engine/texture.h"
#include "engine/polygon/polygon.h"
#include "engine/sector.h"
#include "util/archive.h"
#include "engine/light/light.h"
#include "engine/world.h"
#include "system/csendian.h"

//---------------------------------------------------------------------------

ShadowMap::ShadowMap ()
{
  map = NULL;
}

ShadowMap::~ShadowMap ()
{
  if (map) CHKB (delete [] map);
}

void ShadowMap::alloc (csLight*, int w, int h, int lms)
{
  int size = w*h;
  int i;
  if (map) { CHK (delete [] map); map = NULL; }

  int lw = w/lms+2;
  int lh = h/lms+2;
  long lm_size = lw*lh;
  CHK (map = new unsigned char [lm_size]);
  for (i = 0 ; i < lm_size ; i++) map[i] = 0;
}

void ShadowMap::mipmap_lightmap (int w, int h, int lms, ShadowMap* source, int w2, int h2, int lms2)
{
  int size = w*h;

  if (size > 1000000) return;
  int lw = w/lms+2;
  int lh = h/lms+2;
  int lw2 = w2/lms2+2;
  int lh2 = h2/lms2+2;
  int u, v, uv, uv2;

  uv2 = 0;
  for (v = 0 ; v < lh ; v++)
    for (u = 0 ; u < lw ; u++)
    {
      uv = v*lw + u;

      if (u+u >= lw2 || v+v >= lh2)
      {
	if (u+u >= lw2) uv2 = (v+v)*lw2 + lw2-1;
	else if (v+v >= lh2) uv2 = (lh2-1)*lw2 + u+u;
      }
      else uv2 = (v+v)*lw2 + u+u;

      map[uv] = source->map[uv2];
    }

  light = source->light;
}

void ShadowMap::copy_lightmap (ShadowMap* source, int size)
{
  if (map) CHKB (delete [] map);
  CHK (map = new unsigned char [size]);
  memcpy (map, source->map, size);
  light = source->light;
}

//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( LightMap )

BEGIN_INTERFACE_TABLE( LightMap )
	IMPLEMENTS_COMPOSITE_INTERFACE( LightMap )
END_INTERFACE_TABLE()

int LightMap::setting = MIPMAP_SHADOW_ACCURATE;

LightMap::LightMap ()
{
  first_smap = NULL;
  hicolorcache = NULL;
  in_memory = false;
}

LightMap::~LightMap ()
{
  while (first_smap)
  {
    ShadowMap* smap = first_smap->next;
    CHK (delete first_smap);
    first_smap = smap;
  }
}

void LightMap::del_smap (ShadowMap* smap)
{
  first_smap = smap->next;
  CHK (delete smap);
}

ShadowMap* LightMap::new_smap (csLight* light, int w, int h, int lms)
{
  if (size > 1000000) return NULL;

  CHK (ShadowMap* smap = new ShadowMap ());
  smap->light = light;
  smap->next = first_smap;
  first_smap = smap;

  smap->alloc (light, w, h, lms);

  return smap;
}

ShadowMap* LightMap::find_smap (csLight* light)
{
  ShadowMap* smap = first_smap;
  while (smap)
  {
    if (smap->light == light) return smap;
    smap = smap->next;
  }
  return NULL;
}

// finds nearest number that is a power of two
int find_nearest_po2 (int n)
{
  int w=1;

  while (n > w)  w <<= 1;

  return w;
}


// returns true if n is a power of two
bool is_po2(int n)
{
  int i;
  int po2 = 1;

  if (!n) return false; 
    
  for (i=0; i<32; i++, po2<<=1)
    if (n==po2) return true;

  return false;
}

void LightMap::set_size (int w, int h, int lms)
{
  size = w*h;
  lwidth = w/lms+2;
  lheight = h/lms+2;
  lm_size = lwidth*lheight;
  rwidth = lwidth;
  rheight = lheight;
}

void LightMap::alloc (int w, int h, int lms, Polygon3D* poly)
{
  set_size (w, h, lms);
  int i;
  static_lm.clear ();
  real_lm.clear ();

  if (size > 1000000) return;

  if (Textures::mixing == MIX_NOCOLOR)
  {
    static_lm.allocR (lm_size);
    real_lm.allocR (lm_size);
  }
  else
  {
    static_lm.allocR (lm_size);
    static_lm.allocG (lm_size);
    static_lm.allocB (lm_size);
    real_lm.allocR (lm_size);
    real_lm.allocG (lm_size);
    real_lm.allocB (lm_size);
  }

  if (poly)
  {
    csSector* s = poly->get_sector ();
    if (s)
      for (i = 0 ; i < lm_size ; i++)
      {
	int r, g, b;
	s->get_ambient_color (r, g, b);
        static_lm.mapR[i] = r;
        if (static_lm.mapG) static_lm.mapG[i] = g;
        if (static_lm.mapB) static_lm.mapB[i] = b;
      }
  }
}

void LightMap::mipmap_lightmap (int w, int h, int lms, LightMap* source, int w2, int h2, int lms2)
{
  alloc (w, h, lms, NULL);

  if (size > 1000000 || source->size > 1000000) return;
  int lw2 = source->get_width ();
  int lh2 = source->get_height ();
  int u, v, uv, uv2;

  uv2 = 0;
  for (v = 0 ; v < lheight ; v++)
    for (u = 0 ; u < lwidth ; u++)
    {
      uv = v*lwidth + u;

      if (u+u >= lw2 || v+v >= lh2)
      {
	if (u+u >= lw2) uv2 = (v+v)*lw2 + lw2-1;
	else if (v+v >= lh2) uv2 = (lh2-1)*lw2 + u+u;
      }
      else uv2 = (v+v)*lw2 + u+u;

      static_lm.mapR[uv] = source->static_lm.mapR[uv2];
      if (static_lm.mapG) static_lm.mapG[uv] = source->static_lm.mapG[uv2];
      if (static_lm.mapB) static_lm.mapB[uv] = source->static_lm.mapB[uv2];
    }

  ShadowMap* smap, * smap2;
  while (first_smap)
  {
    smap = first_smap->next;
    CHK (delete first_smap);
    first_smap = smap;
  }

  smap = source->first_smap;
  while (smap)
  {
    smap2 = new_smap (smap->light, w, h, lms);
    smap2->mipmap_lightmap (w, h, lms, smap, w2, h2, lms2);
    smap = smap->next;
  }
}

void LightMap::copy_lightmap (LightMap* source)
{
  lm_size = source->lm_size;
  size = source->size;
  static_lm.copy (source->static_lm, lm_size);
  real_lm.copy (source->real_lm, lm_size);
  lwidth = source->lwidth;
  lheight = source->lheight;
  rwidth = source->rwidth;
  rheight = source->rheight;

  ShadowMap* smap, * smap2;
  while (first_smap)
  {
    smap = first_smap->next;
    CHK (delete first_smap);
    first_smap = smap;
  }

  smap = source->first_smap;
  while (smap)
  {
    CHK (smap2 = new ShadowMap ());
    smap2->next = first_smap;
    first_smap = smap2;
    smap2->copy_lightmap (smap, lm_size);
    smap = smap->next;
  }
}

struct PolySave
{
  char header[4];
  float x1, y1, z1;	// Coordinate of vertex 1
  float x2, y2, z2;	// Coordinate of vertex 2
  long lm_size;		// Size of lightmap
  long lm_cnt;		// Number of non-dynamic lightmaps (normally 3)
};
#define POLYSAVE_LEN (4*9)

struct LightSave
{
  float x, y, z, dist;
};
#define LIGHTSAVE_LEN (4*4)

struct LightHeader
{
  char header[4];
  long dyn_cnt;		// Number of dynamic maps
};
#define LIGHTHDR_LEN (4*2)

void cache_name (char* buf, Polygon3D* poly, char* suffix)
{
  PolygonSet* poly_set = (PolygonSet*)(poly->get_parent ());
  sprintf (buf, "LM_%s_%s_%d%s",
    	poly_set->get_name (),
    	poly->get_name (),
	poly->get_id (),
	suffix);
}

bool LightMap::read_from_cache (int w, int h, int lms, Polygon3D* poly, csWorld* world)
{
  char buf[200];
  PolySave ps;
  LightHeader lh;
  LightSave ls;
  csLight* light;
  int i;
  Archive* ar = world->get_world_file ();

  set_size (w, h, lms);

  cache_name (buf, poly, "");

  char* data = ar->read (buf);
  if (!data) return false;

  char* d = data;
  memcpy (ps.header, d, 4); d += 4;
  memcpy (&ps.x1, d, 4); d += 4;
  memcpy (&ps.y1, d, 4); d += 4;
  memcpy (&ps.z1, d, 4); d += 4;
  memcpy (&ps.x2, d, 4); d += 4;
  memcpy (&ps.y2, d, 4); d += 4;
  memcpy (&ps.z2, d, 4); d += 4;
  memcpy (&ps.lm_size, d, 4); d += 4;
  memcpy (&ps.lm_cnt, d, 4); d += 4;
  ps.x1 = convert_endian (ps.x1);
  ps.y1 = convert_endian (ps.y1);
  ps.z1 = convert_endian (ps.z1);
  ps.x2 = convert_endian (ps.x2);
  ps.y2 = convert_endian (ps.y2);
  ps.z2 = convert_endian (ps.z2);
  ps.lm_size = convert_endian (ps.lm_size);
  ps.lm_cnt = convert_endian (ps.lm_cnt);

  //-------------------------------
  // Check if cached item is still valid.
  //-------------------------------
  if (	ABS (ps.x1 - poly->vobj (0).x) > EPSILON ||
  	ABS (ps.y1 - poly->vobj (0).y) > EPSILON ||
  	ABS (ps.z1 - poly->vobj (0).z) > EPSILON ||
  	ABS (ps.x2 - poly->vobj (1).x) > EPSILON ||
  	ABS (ps.y2 - poly->vobj (1).y) > EPSILON ||
  	ABS (ps.z2 - poly->vobj (1).z) > EPSILON ||
	ps.lm_size != lm_size)
  {
    // Invalid.
    CHK (delete [] data);
    return false;
  }

  //-------------------------------
  // The cached item is valid.
  //-------------------------------
  static_lm.clear ();

  static_lm.allocR (lm_size);
  memcpy (static_lm.mapR, d, lm_size);
  d += lm_size;

  if (Textures::mixing != MIX_NOCOLOR)
  {
    static_lm.allocG (lm_size);
    memcpy (static_lm.mapG, d, lm_size);
    d += lm_size;
    static_lm.allocB (lm_size);
    memcpy (static_lm.mapB, d, lm_size);
  }

  CHK (delete [] data);

  //-------------------------------
  // Now load the dynamic data.
  //-------------------------------
  cache_name (buf, poly, "_d");
  data = ar->read (buf);
  if (!data) return true;	// No dynamic data. @@@ Recalculate dynamic data?

  d = data;
  memcpy (lh.header, d, 4); d += 4;
  memcpy (&lh.dyn_cnt, d, 4); d += 4;
  lh.dyn_cnt = convert_endian (lh.dyn_cnt);

  for (i = 0 ; i < lh.dyn_cnt ; i++)
  {
    memcpy (&ls.x, d, 4); d += 4;
    memcpy (&ls.y, d, 4); d += 4;
    memcpy (&ls.z, d, 4); d += 4;
    memcpy (&ls.dist, d, 4); d += 4;
    ls.x = convert_endian (ls.x);
    ls.y = convert_endian (ls.y);
    ls.z = convert_endian (ls.z);
    ls.dist = convert_endian (ls.dist);

    light = world->find_light (ls.x, ls.y, ls.z, ls.dist);
    if (light->get_type () == CS_STATLIGHT)
    {
      csStatLight* slight = (csStatLight*)light;
      slight->register_polygon (poly);
    }
    ShadowMap* smap = new_smap (light, w, h, lms);
    memcpy (smap->map, d, lm_size);
    d += lm_size;
  }

  CHK (delete [] data);

  return true;
}

void LightMap::cache (Polygon3D* poly, csWorld* world)
{
  char buf[200];
  PolySave ps;
  void* entry;
  long l;
  float f;
  Archive* ar = world->get_world_file ();

  strcpy (ps.header, "MAPL");
  ps.x1 = poly->vobj (0).x;
  ps.y1 = poly->vobj (0).y;
  ps.z1 = poly->vobj (0).z;
  ps.x2 = poly->vobj (1).x;
  ps.y2 = poly->vobj (1).y;
  ps.z2 = poly->vobj (1).z;
  ps.lm_size = lm_size;
  ps.lm_cnt = 0;
  if (static_lm.mapR) ps.lm_cnt++;
  if (static_lm.mapG) ps.lm_cnt++;
  if (static_lm.mapB) ps.lm_cnt++;

  //-------------------------------
  // Write the normal lightmap data.
  //-------------------------------
  cache_name (buf, poly, "");
  entry = ar->new_file (buf, POLYSAVE_LEN+lm_size*3);
  ar->write (entry, ps.header, 4);
  f = convert_endian (ps.x1); ar->append (entry, (char*)&f, 4);
  f = convert_endian (ps.y1); ar->append (entry, (char*)&f, 4);
  f = convert_endian (ps.z1); ar->append (entry, (char*)&f, 4);
  f = convert_endian (ps.x2); ar->append (entry, (char*)&f, 4);
  f = convert_endian (ps.y2); ar->append (entry, (char*)&f, 4);
  f = convert_endian (ps.z2); ar->append (entry, (char*)&f, 4);
  l = convert_endian (ps.lm_size); ar->append (entry, (char*)&l, 4);
  l = convert_endian (ps.lm_cnt); ar->append (entry, (char*)&l, 4);

  if (static_lm.mapR) ar->append (entry, (char*)static_lm.mapR, lm_size);
  if (static_lm.mapG) ar->append (entry, (char*)static_lm.mapG, lm_size);
  if (static_lm.mapB) ar->append (entry, (char*)static_lm.mapB, lm_size);

  //-------------------------------
  // Write the dynamic data.
  //-------------------------------
  LightHeader lh;

  ShadowMap* smap = first_smap;
  if (smap)
  {
    strcpy (lh.header, "DYNL");
    lh.dyn_cnt = 0;
    while (smap) { lh.dyn_cnt++; smap = smap->next; }
    smap = first_smap;

    cache_name (buf, poly, "_d");
    entry = ar->new_file (buf, LIGHTHDR_LEN+lh.dyn_cnt*(LIGHTSAVE_LEN+lm_size));
    ar->write (entry, lh.header, 4);
    l = convert_endian (lh.dyn_cnt); ar->append (entry, (char*)&l, 4);
  }
  while (smap)
  {
    csLight* light = smap->light;
    if (smap->map)
    {
      LightSave ls;
      ls.x = light->get_center ().x;
      ls.y = light->get_center ().y;
      ls.z = light->get_center ().z;
      ls.dist = light->get_dist ();
      f = convert_endian (ls.x); ar->append (entry, (char*)&f, 4);
      f = convert_endian (ls.y); ar->append (entry, (char*)&f, 4);
      f = convert_endian (ls.z); ar->append (entry, (char*)&f, 4);
      f = convert_endian (ls.dist); ar->append (entry, (char*)&f, 4);
      ar->append (entry, (char*)(smap->map), lm_size);
    }
    smap = smap->next;
  }
}

void LightMap::convert_to_mixing_mode ()
{
  if (Textures::mixing == MIX_TRUE_RGB) return;
  else convert_to_mixing_mode (static_lm.mapR, static_lm.mapG, static_lm.mapB, lm_size);
}

void LightMap::convert_to_mixing_mode (unsigned char* mr, unsigned char* mg,
				       unsigned char* mb, int sz)
{
  if (!mr || !mg || !mb) return; // If we have a dynamic light there is no conversion needed

  while (sz > 0)
  {
    csLight::mixing_dependent_strengths (mr, mg, mb);
    mr++;
    mg++;
    mb++;
    sz--;
  }
}

#define realresize

// Only works for expanding a map.
void resize_map (unsigned char* old_map, int oldw, int oldh,
		 unsigned char* new_map, int neww, int newh)
{
#if !defined(realresize)
  int row;
  for (row = 0 ; row < oldh ; row++)
    memcpy (new_map+neww*row, old_map+oldw*row, oldw);
#else
  float rrow=(float)(oldh-0.5)/(float)newh;
  float rcol=(float)(oldw-0.5)/(float)neww;

  for (int row = 0 ; row < newh ; row++)
    for(int col = 0; col < neww; col++)
      new_map[neww*row+col] = old_map[oldw*((int)(row*rrow))+(int)(col*rcol)];
#endif
}

void LightMap::convert_for_3d_driver (bool requirePO2, int maxAspect)
{
  if (!requirePO2) return; // Nothing to do.
  int oldw = lwidth, oldh = lheight;
  lwidth = find_nearest_po2 (lwidth);
  lheight = find_nearest_po2 (lheight);
  while (lwidth/lheight > maxAspect) lheight += lheight;
  while (lheight/lwidth > maxAspect) lwidth += lwidth;
  if (oldw == lwidth && oldh == lheight) return;	// Already ok, nothing to do.

  // Move the old data to o_stat and o_real.
  RGBLightMap o_stat, o_real;
  o_stat.mapR = static_lm.mapR; static_lm.mapR = NULL;
  o_stat.mapG = static_lm.mapG; static_lm.mapG = NULL;
  o_stat.mapB = static_lm.mapB; static_lm.mapB = NULL;
  o_real.mapR = real_lm.mapR; real_lm.mapR = NULL;
  o_real.mapG = real_lm.mapG; real_lm.mapG = NULL;
  o_real.mapB = real_lm.mapB; real_lm.mapB = NULL;

  lm_size = lwidth*lheight;
#if defined(realresize)
  rwidth=lwidth;
  rheight=lheight;
#endif
 
  // Allocate new data and transform old to new.
  if (Textures::mixing == MIX_NOCOLOR)
  {
    static_lm.allocR (lm_size);
    resize_map (o_stat.mapR, oldw, oldh, static_lm.mapR, lwidth, lheight);
    real_lm.allocR (lm_size);
    resize_map (o_real.mapR, oldw, oldh, real_lm.mapR, lwidth, lheight);
  }
  else
  {
    static_lm.allocR (lm_size);
    resize_map (o_stat.mapR, oldw, oldh, static_lm.mapR, lwidth, lheight);
    static_lm.allocG (lm_size);
    resize_map (o_stat.mapG, oldw, oldh, static_lm.mapG, lwidth, lheight);
    static_lm.allocB (lm_size);
    resize_map (o_stat.mapB, oldw, oldh, static_lm.mapB, lwidth, lheight);

    real_lm.allocR (lm_size);
    resize_map (o_real.mapR, oldw, oldh, real_lm.mapR, lwidth, lheight);
    real_lm.allocG (lm_size);
    resize_map (o_real.mapG, oldw, oldh, real_lm.mapG, lwidth, lheight);
    real_lm.allocB (lm_size);
    resize_map (o_real.mapB, oldw, oldh, real_lm.mapB, lwidth, lheight);
  }

  // Convert all shadowmaps.
  ShadowMap* smap = first_smap;
  while (smap)
  {
    unsigned char* old_map = smap->map;
    CHK (smap->map = new unsigned char [lm_size]);
    resize_map (old_map, oldw, oldh, smap->map, lwidth, lheight);
    CHK (delete [] old_map);
    smap = smap->next;
  }
}

//---------------------------------------------------------------------------
