/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef LIGHTMAP_H
#define LIGHTMAP_H

#include "com/com.h"
#include "engine/light/ilghtmap.h"

class PolyTexture;
class Polygon3D;
class csLight;
class csWorld;
class Dumper;

/**
 * There are three possible levels of mipmapping accuracy.
 * This accuracy has nothing to do with the quality of the
 * mipmapped texture itself but it controls the quality
 * of the shadows.<p>
 *
 * With MIPMAP_SHADOW_ACCURATE the shadows
 * are 'perfect' at all distances. Even polygons that are
 * far away (mipmap level 3) will have correct shadows.
 * The disadvantage of this is that this is slower for the
 * texture cache. It will not influence stationary frame rate,
 * but far away textures will take somewhat longer to load
 * in the cache. Once in the cache, there is no difference.<p>
 *
 * MIPMAP_SHADOW_INACCURATE gives the fastest way for the
 * texture cache but shadows will be noticeably worse when
 * polygons gets farther away. This is rather ugly. This
 * mode is included only so that you can see the difference :-)<p>
 *
 * MIPMAP_SHADOW_REASONABLE is somewhat in between. Polygons
 * far away will have reasonably accurate shadows (not as
 * accurate as with MIPMAP_SHADOW_ACCURATE) and they will
 * draw somewhat faster (not as fast as with MIPMAP_SHADOW_INACCURATE).<p>
 *
 * Note that none of these options affect textures that are
 * really close by.<p>
 *
 * Note that changing this option (by pressing 'alt-t') will only
 * be visible when a texture is removed from the texture cache and
 * reinserted again. The texture cache can be cleared by pressing 'r'.
 */
#define MIPMAP_SHADOW_ACCURATE 0
#define MIPMAP_SHADOW_INACCURATE 1
#define MIPMAP_SHADOW_REASONABLE 2

class PolyTexture;
struct HighColorCache_Data;

/**
 * This is a shadow-map for a pseudo-dynamic light.
 */
class ShadowMap
{
  ///
  friend class LightMap;
  ///
  friend class PolyTexture;

private:
  ShadowMap* next;
  unsigned char* map;

  /**
   * The pseudo-dynamic light.
   */
  csLight* light;

  ///
  ShadowMap ();
  ///
  virtual ~ShadowMap ();

  ///
  unsigned char* get_shadow_map () { return map; }

  ///
  void alloc (csLight* light, int w, int h, int lms);
  ///
  void mipmap_lightmap (int w, int h, int lms, ShadowMap* source, int w2, int h2, int lms2);
  ///
  void copy_lightmap (ShadowMap* source, int size);
};

/**
 * A small class encapsulating an RGB lightmap.
 */
class RGBLightMap
{
public:
  unsigned char* mapR;
  unsigned char* mapG;
  unsigned char* mapB;

  ///
  RGBLightMap () { mapR = mapG = mapB = NULL; }
  ///
  ~RGBLightMap () { clear (); }

  ///
  void clear ()
  {
    if (mapR) { CHK (delete [] mapR); mapR = NULL; }
    if (mapG) { CHK (delete [] mapG); mapG = NULL; }
    if (mapB) { CHK (delete [] mapB); mapB = NULL; }
  }

  ///
  void allocR (int size)
  {
    CHK (mapR = new unsigned char [size]);
  }

  ///
  void allocG (int size)
  {
    CHK (mapG = new unsigned char [size]);
  }

  ///
  void allocB (int size)
  {
    CHK (mapB = new unsigned char [size]);
  }

  ///
  void copy (RGBLightMap& other, int size)
  {
    clear ();
    if (other.mapR) { allocR (size); memcpy (mapR, other.mapR, size); }
    if (other.mapG) { allocG (size); memcpy (mapG, other.mapG, size); }
    if (other.mapB) { allocB (size); memcpy (mapB, other.mapB, size); }
  }
};

/**
 * The static and all dynamic lightmaps for one or more mipmap-levels of a polygon.
 */
class LightMap
{
  ///
  friend class PolyTexture;
  ///
  friend class Dumper;

private:
  /**
   * A color lightmap containing all static lighting information
   * for the static lights (no pseudo-dynamic lights are here).
   */
  RGBLightMap static_lm;

  /**
   * The final lightmap that is going to be used for rendering.
   * In many cases this is just a copy of static_lm. But it may
   * contain extra lighting information obtained from all the
   * pseudo-dynamic shadowmaps and also the true dynamic lights.
   */
  RGBLightMap real_lm;

  /**
   * Linked list of shadow-maps (for the pseudo-dynamic lights).
   * These shadowmaps are applied to static_lm to get real_lm.
   */
  ShadowMap* first_smap;

  /// Size of the lighted texture.
  long size;
  /// Size of the lightmap.
  long lm_size;

  /// Lightmap dims (possibly po2 depending on used 3D driver).
  int lwidth, lheight;
  /// Original lightmap dims (non-po2 possibly).
  int rwidth, rheight;
  
  /// The hicolor cache ptr.
  HighColorCache_Data *hicolorcache;
  
  /**
   * Convert three lightmap tables to the right mixing mode.
   * This function assumes that mixing mode is one of FAST_Wxx
   * and will not function correctly if called with NOCOLOR, TRUE_RGB
   * or FAST_RGB.<br>
   * This function correctly recognizes a dynamic lightmap which only
   * contains shadow information and does not do the conversion in that
   * case.
   */
  void convert_to_mixing_mode (unsigned char* mr, unsigned char* mg,
			       unsigned char* mb, int sz);

  /**
   * Calculate the sizes of this lightmap.
   */
  void set_size (int w, int h, int lms);

public:
  /// One of MIPMAP_SHADOW_...
  static int setting;

public:
  ///
  LightMap ();
  ///
  virtual ~LightMap ();

  ///
  RGBLightMap& get_static_map () { return static_lm; }
  ///
  RGBLightMap& get_real_map () { return real_lm; }

  ///
  int get_width() { return lwidth; }
  ///
  int get_height() { return lheight; }
  ///
  int get_realwidth() { return rwidth; }
  ///
  int get_realheight() { return rheight; }
  ///
  long get_size () { return lm_size; }

  // DAN: High color cache specific stuff
  bool in_memory;
  ///
  HighColorCache_Data *get_hicolorcache() { return hicolorcache; }
  ///
  void set_hicolorcache(HighColorCache_Data *d) { hicolorcache = d; }

  /**
   * Allocate the lightmap. 'w' and 'h' are the size of the
   * bounding box in lightmap space.
   */
  void alloc (int w, int h, int lms, Polygon3D* poly);

  /**
   * Allocate this lightmap by mipmapping the given source lightmap.
   */
  void mipmap_lightmap (int w, int h, int lms, LightMap* source, int w2, int h2, int lms2);

  /// Copy a lightmap.
  void copy_lightmap (LightMap* source);

  /**
   * Create a ShadowMap for this LightMap.
   */
  ShadowMap* new_smap (csLight* light, int w, int h, int lms);

  /**
   * Allocate the static RGBLightMap.
   */
  void alloc_static_lm (int w, int h, int lms);

  /**
   * Find a ShadowMap for a specific pseudo-dynamic light.
   */
  ShadowMap* find_smap (csLight* light);

  /**
   * Delete a ShadowMap. NOTE!!! This function only works
   * if the ShadowMap was the LAST CREATED for this LightMap!!!
   * It is ment for dynamic lights that do not reach the polygon
   * but this can only be seen after trying.
   */
  void del_smap (ShadowMap* plm);

  /**
   * Read from the cache. Return true if succesful.
   */
  bool read_from_cache (int w, int h, int lms, Polygon3D* poly, csWorld* world);

  /**
   * Cache the lightmaps in the precalculation area.
   */
  void cache (Polygon3D* poly, csWorld* world);

  /**
   * Convert the lightmaps to the correct mixing mode.
   * This function does nothing unless the mixing mode is one
   * of fast_xxx or nocolor.
   */
  void convert_to_mixing_mode ();

  /**
   * Convert the lightmaps to a 3D driver dependent size.
   */
  void convert_for_3d_driver (bool requirePO2, int maxAspect = 32767);

  DECLARE_INTERFACE_TABLE( LightMap )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( LightMap )
};

#define GetILightMapFromLightMap(a)  &a->m_xLightMap;
#define GetLightMapFromILightMap(a)  ((LightMap*)((size_t)a - offsetof(LightMap, m_xLightMap)))

#endif /*LIGHTMAP_H*/

