/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/sysprint.h"
#include "engine/light/light.h"
#include "engine/sector.h"
#include "engine/polygon/polygon.h"
#include "engine/texture.h"

//---------------------------------------------------------------------------

csLight::csLight (float x, float y, float z, float dist,
	      float red, float green, float blue,
	      char* name, int cso_type) : csObject (name, cso_type)
{
  center.x = x;
  center.y = y;
  center.z = z;
  csLight::dist = dist;
  csLight::sqdist = dist*dist;
  color.red = red;
  color.green = green;
  color.blue = blue;

  halo_enabled = false;
  halo_intensity = 0.0f;
  halo_max_intensity = 1.0f;
  halo_ref_count = 0;
  in_halo_queue = false;
}

csLight::~csLight ()
{
}

void csLight::mixing_dependent_strengths (unsigned char* rp, unsigned char* gp,
					unsigned char* bp)
{
  if (Textures::mixing == MIX_TRUE_RGB)
    return;
  else if (Textures::mixing == MIX_NOCOLOR)
  {
    unsigned char w = (*rp+*gp+*bp)/3;
    *rp = w;
    *gp = w;
    *bp = w;
  }
  else
  {
    int r = *rp;
    int g = *gp;
    int b = *bp;
    int w = r;
    if (g < w) w = g;
    if (b < w) w = b;
    r -= w;
    g -= w;
    b -= w;

    *rp = w;

    if (Textures::color_table2 == TABLE_RED) *gp = r;
    else if (Textures::color_table2 == TABLE_GREEN) *gp = g;
    else if (Textures::color_table2 == TABLE_BLUE) *gp = b;

    if (Textures::color_table3 == TABLE_RED) *bp = r;
    else if (Textures::color_table3 == TABLE_GREEN) *bp = g;
    else if (Textures::color_table3 == TABLE_BLUE) *bp = b;
  }
}

void csLight::mixing_dependent_strengths (float* rp, float* gp, float* bp)
{
  if (Textures::mixing == MIX_TRUE_RGB)
    return;
  else if (Textures::mixing == MIX_NOCOLOR)
  {
    float w = (*rp+*gp+*bp)/3;
    *rp = w;
    *gp = w;
    *bp = w;
  }
  else
  {
    float r = *rp;
    float g = *gp;
    float b = *bp;
    float w = r;
    if (g < w) w = g;
    if (b < w) w = b;
    r -= w;
    g -= w;
    b -= w;

    *rp = w;

    if (Textures::color_table2 == TABLE_RED) *gp = r;
    else if (Textures::color_table2 == TABLE_GREEN) *gp = g;
    else if (Textures::color_table2 == TABLE_BLUE) *gp = b;

    if (Textures::color_table3 == TABLE_RED) *bp = r;
    else if (Textures::color_table3 == TABLE_GREEN) *bp = g;
    else if (Textures::color_table3 == TABLE_BLUE) *bp = b;
  }
}

//---------------------------------------------------------------------------

csStatLight::csStatLight (float x, float y, float z, float dist,
	      float red, float green, float blue, bool dynamic)
	      : csLight (x, y, z, dist, red, green, blue, "", CS_STATLIGHT)
{
  csStatLight::dynamic = dynamic;
  polygons = NULL;
}

csStatLight::~csStatLight ()
{
  if (polygons) CHKB (delete [] polygons);
}

void csStatLight::shine_lightmaps ()
{
  //CsPrintf (MSG_INITIALIZATION, "  Shine light (%f,%f,%f).\n", center.x, center.y, center.z);
  LightView lview;
  lview.l = this;
  lview.center = center;
  lview.mirror = false;
  lview.r = get_red ();
  lview.g = get_green ();
  lview.b = get_blue ();
  lview.dynamic = false;
  sector->shine_lightmaps (lview);
}

void csStatLight::dump_frustrum (Transform& t)
{
  sector->dump_frustrum (this, NULL, 0, t);
}

int csStatLight::hit_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* dist)
{
  return sector->hit_beam (start, end, poly, dist);
}

void csStatLight::register_polygon (Polygon3D* poly)
{
  if (dynamic && !polygons)
  {
    num_polygon = 0;
    CHK (polygons = new Polygon3D* [MAX_NUM_POLYGON]);
  }

  int i;
  for (i = 0 ; i < num_polygon ; i++)
    if (polygons[i] == poly) return;
  polygons[num_polygon++] = poly;
  if (num_polygon >= MAX_NUM_POLYGON)
  {
    CsPrintf (MSG_WARNING, "Overflow number of polygons for dynamic light!\n");
  }
}

void csStatLight::set_color (csColor& col)
{
  csLight::set_color (col);
  int i;
  for (i = 0 ; i < num_polygon ; i++)
    polygons[i]->make_dirty_dynamic_lights ();
}

//---------------------------------------------------------------------------

LightPatch::LightPatch ()
{
  next_poly = prev_poly = NULL;
  next_light = prev_light = NULL;
  num_vertices = 0;
  vertices = NULL;
  polygon = NULL;
  light = NULL;
}

LightPatch::~LightPatch ()
{
  if (vertices) { CHK (delete [] vertices); }
  if (polygon) polygon->unlink_lightpatch (this);
  if (light) light->unlink_lightpatch (this);
}

//---------------------------------------------------------------------------

csDynLight::csDynLight (float x, float y, float z, float dist,
	      float red, float green, float blue)
	      : csLight (x, y, z, dist, red, green, blue, "-", CS_DYNLIGHT)
{
  lightpatches = NULL;
}

csDynLight::~csDynLight ()
{
  while (lightpatches) CHKB (delete lightpatches);
}

void csDynLight::setup ()
{
  while (lightpatches) CHKB (delete lightpatches);
  LightView lview;
  lview.l = this;
  lview.center = center;
  lview.frustrum = NULL;
  lview.num_frustrum = 0;
  lview.mirror = false;
  lview.r = get_red ();
  lview.g = get_green ();
  lview.b = get_blue ();
  lview.dynamic = true;
  sector->shine_lightmaps (lview);
}

void csDynLight::move (csSector* sector, float x, float y, float z)
{
  if (sector == csDynLight::sector && center.x == x && center.y == y && center.z == z)
  {
    // No move. Just return.
    return;
  }
  set_sector (sector);
  center.x = x;
  center.y = y;
  center.z = z;
}

void csDynLight::unlink_lightpatch (LightPatch* lp)
{
  if (lp->next_light) lp->next_light->prev_light = lp->prev_light;
  if (lp->prev_light) lp->prev_light->next_light = lp->next_light;
  else lightpatches = lp->next_light;
  lp->prev_light = lp->next_light = NULL;
  lp->light = NULL;
}

void csDynLight::add_lightpatch (LightPatch* lp)
{
  lp->next_light = lightpatches;
  lp->prev_light = NULL;
  if (lightpatches) lightpatches->prev_light = lp;
  lightpatches = lp;
  lp->light = this;
}


//---------------------------------------------------------------------------
