/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef LIGHT_H
#define LIGHT_H

#include "geometry/math3d.h"
#include "engine/basic/csobject.h"
#include "engine/cscolor.h"

class csSector;
class Polygon3D;
class csDynLight;
class Dumper;

/// Light level that is used when there is no light on the texture.
#define DEFAULT_LIGHT_LEVEL 20
/// Light level that corresponds to a normally lit texture.
#define NORMAL_LIGHT_LEVEL 200

/// Max number of polygons that can be lit by one light. (bad practice !!!@@@)
#define MAX_NUM_POLYGON 600

/**
 * Superclass of all positional lights.
 * A light subclassing from this has a color, a position
 * and a radius.<P>
 *
 * First some terminology about all the several types of lights
 * that Crystal Space supports:
 * <ul>
 * <li>Static light. This is a normal static light that cannot move
 *     and cannot change intensity/color. All lighting information from
 *     all static lights is collected in one static lightmap.
 * <li>Pseudo-dynamic light. This is a static light that still cannot
 *     move but the intensity/color can change. The shadow information
 *     from every pseudo-dynamic light is kept in a seperate shadow-map.
 *     Shadowing is very accurate with pseudo-dynamic lights since they
 *     use the same algorithm as static lights.
 * <li>Dynamic light. This is a light that can move and change
 *     intensity/color. These lights are the most flexible. All lighting
 *     information from all dynamic lights is collected in one dynamic
 *     lightmap (seperate from the pseudo-dynamic shadow-maps).
 *     Shadows for dynamic lights will be less accurate because things
 *     will not cast accurate shadows (due to computation speed limitations).
 * <li>Floating light. This is also a dynamic light but it is more a hack.
 *     It is kept here because in some cases it may be more efficient
 *     then the real dynamic lights. We will see how this turns out. If
 *     dynamic lights are fast enough we may remove floating lights.
 * </ul>
 * Note that static and pseudo-dynamic lights are represented by the
 * same csStatLight class.
 */
class csLight : public csObject
{
protected:
  /// Home sector of the light.
  csSector* sector;
  /// Position of the light.
  Vector3 center;
  /// Radius of light.
  float dist;
  /// Squared radius of light.
  float sqdist;
  /// Color.
  csColor color;

  /// If a halo is enabled for this light.
  bool halo_enabled;
  /// the current intensity of the attached halo.
  float halo_intensity;
  /// the maximum intensity of the attached halo.
  float halo_max_intensity;
  /// the reference count (for halo fading purposes).
  int halo_ref_count;
  /// whether this light is in the halo queue or not.
  bool in_halo_queue;
  

public:
  ///
  csLight (float x, float y, float z, float dist,
  	 float red, float green, float blue,
	 char* name, int cso_type);
  ///
  virtual ~csLight ();

  ///
  virtual void set_sector (csSector* sector) { csLight::sector = sector; }
  ///
  csSector* get_sector () { return sector; }

  ///
  Vector3& get_center () { return center; }
  ///
  float get_dist () { return dist; }
  ///
  float get_sq_dist () { return sqdist; }
  ///
  csColor& get_color () { return color; }
  ///
  float get_red () { return color.red; }
  ///
  float get_green () { return color.green; }
  ///
  float get_blue () { return color.blue; }

  ///
  void enable_halo () { halo_enabled = true; }
  ///
  void disable_halo () { halo_enabled = false; }
  ///
  bool is_halo_enabled () { return halo_enabled; }
  ///
  float get_halo_max_intensity () { return halo_max_intensity; }
  ///
  float get_halo_intensity () { return halo_intensity; }
  /// 
  void set_halo_intensity (float newI) { halo_intensity = newI; }
  ///
  int get_reference_count () { return halo_ref_count; }
  ///
  void add_reference () { halo_ref_count++; }
  ///
  void remove_reference () { if (halo_ref_count > 0) halo_ref_count--; }
  ///
  bool get_halo_in_queue () { return in_halo_queue; }
  ///
  void set_halo_in_queue(bool bNew) { in_halo_queue = bNew; }
  

  ///
  virtual void set_color (csColor& col)
  {
    color = col;
  }

  /**
   * Change the given r, g, b value to the current mixing mode
   * (FAST_W.. or NOCOLOR). This function correctly works in TRUE_RGB
   * as well (it checks first).
   */
  static void mixing_dependent_strengths (unsigned char* rp, unsigned char* gp,
					  unsigned char* bp);

  /// Version with floating point colors.
  static void mixing_dependent_strengths (float* rp, float* gp, float* bp);
};

/**
 * Class for a static light. These lights cast shadows (against
 * sector boundaries and with things), they support three different
 * colors (R,G,B). They cannot move and they can only vary in
 * intensity with some memory trade-offs.
 */
class csStatLight : public csLight
{
private:
  /**
   * The following three variables are used if the light intensity
   * can vary. 'dynamic' is set to true in that case.
   * 'num_polygon' and 'polygons' indicate all polygons that are
   * possibly lit by this light.
   */
  bool dynamic;

  /// Number of polygons affected by this dynamic light.
  int num_polygon;

  /// List of polygons that are affected by this dynamic light.
  Polygon3D** polygons;

public:
  ///
  csStatLight (float x, float y, float z, float dist,
  	 float red, float green, float blue,
	 bool dynamic);
  ///
  virtual ~csStatLight ();

  ///
  bool is_dynamic () { return dynamic; }
  ///
  virtual void set_color (csColor& col);

  /// Register a polygon for a dynamic light.
  void register_polygon (Polygon3D* poly);

  ///
  int hit_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* dist);

  /**
   * Shine this light on all polygons visible from the light.
   * This routine will update the lightmaps of all polygons.
   * It correctly takes dynamic lights into account and will then
   * update the corresponding dynamic lightmap.
   */
  void shine_lightmaps ();

  /**
   * This is a debugging function that will show the outlines
   * on all polygons that are hit by this light.
   * It will draw perspective correct outlines so it is meant to
   * be called with a camera transformation.
   */
  void dump_frustrum (Transform& t);
};

/// Dynamic lights can light maximum 200 polygons (@@@ Remove this arbitrary limit!)
#define MAX_DYN_POLYGON 200

/**
 * A light patch. This is a 3D polygon which fits on a world level 3D
 * polygon and defines where the light hits the polygon.
 * There is a list of light patches in every polygon (all dynamic lights
 * hitting a polygon will give rise to a seperate light patch) and there
 * is a list of light patches in every dynamic light (representing all
 * polygons that are hit by that particular light).
 */
class LightPatch
{
  friend class Polygon3D;
  friend class PolyTexture;
  friend class csDynLight;
  friend class Dumper;

private:
  ///
  LightPatch* next_poly;
  ///
  LightPatch* prev_poly;
  ///
  LightPatch* next_light;
  ///
  LightPatch* prev_light;

  ///
  int num_vertices;
  ///
  Vector3* vertices;

  /// Polygon that this light patch is for.
  Polygon3D* polygon;
  /// Light that this light patch originates from.
  csDynLight* light;

public:
  ///
  LightPatch ();
  ///
  ~LightPatch ();

  ///
  Polygon3D* get_polygon () { return polygon; }
  ///
  csDynLight* get_light () { return light; }

  ///
  LightPatch* get_next_poly () { return next_poly; }
  ///
  LightPatch* get_next_light () { return next_light; }
};

/**
 * Class for a dynamic light. These lights don't cast shadows (in contrast
 * with static lights) but they can move and vary in intensity.
 * They also don't work currently :-(
 */
class csDynLight : public csLight
{
private:
  ///
  csDynLight* next;
  ///
  csDynLight* prev;

  /// List of light patches for this dynamic light.
  LightPatch* lightpatches;

public:
  ///
  csDynLight (float x, float y, float z, float dist,
  	 float red, float green, float blue);
  ///
  virtual ~csDynLight ();

  /**
   * Initial placement of the light. This routine generates a view
   * frustrum as seen from the light. The clipped polygons that
   * result from this are light patches and are put in the
   * lightpatches list. This routine needs to be called whenever
   * the light moves.
   */
  void setup ();

  /**
   * Move the light. This will NOT automatically recalculate the
   * view frustrum. You still need to call setup() after this.
   */
  void move (csSector* sector, Vector3& v) { move (sector, v.x, v.y, v.z); }

  /// Move the light.
  void move (csSector* sector, float x, float y, float z);

  /// Resize the light.
  void resize (float radius) { dist = radius; sqdist = dist*dist; }

  /**
   * Unlink a light patch from the light patch list.
   * Warning! This function does not test if the light patch
   * is really on the list!
   */
  void unlink_lightpatch (LightPatch* lp);

  /**
   * Add a light patch to the light patch list.
   */
  void add_lightpatch (LightPatch* lp);

  ///
  void set_next (csDynLight* n) { next = n; }
  ///
  void set_prev (csDynLight* p) { prev = p; }
  ///
  csDynLight* get_next () { return next; }
  ///
  csDynLight* get_prev () { return prev; }
};

#endif /*LIGHT_H*/
