/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef MSECTOBJ_H
#define MSECTOBJ_H

#include "util/def.h"
#include "geometry/math3d.h"

class csObject;
class csSector;
class MultiSectorObject;
class csWorld;

/**
 * Link between csSectors and MultiSectorObjects. This
 * link is seperate because we have a N-N relationship.
 */
class SectorMSO
{
  friend class csSector;
  friend class MultiSectorObject;

private:
  /// Next in Sector list.
  SectorMSO* next_sector;
  ///
  SectorMSO* prev_sector;
  /// Next in MultiSectorObject list.
  SectorMSO* next_mso;
  ///
  SectorMSO* prev_mso;

  /// csSector that this link is for.
  csSector* sector;
  /// MultiSectorObject that this link is for.
  MultiSectorObject* mso;

  ///
  SectorMSO ();
  ///
  ~SectorMSO ();

public:
  /// Return the next MSO in this sector.
  SectorMSO* get_next_sector() { return next_sector; }

  /// Return the object for this SectorMSO.
  MultiSectorObject* get_mso() { return mso; }
};

/**
 * A MultiSectorObject represents a stub for an object
 * that can be in more than one sector at the same time.
 * Currently Things and 3D sprites can do this so they
 * should use a MultiSectorObject.<p>
 *
 * A MultiSectorObject uses SectorMSO objects to be able
 * to define a N-N relation between Sectors and MultiSectorObjects.<p>
 *
 * All MultiSectorObjects are also linked in the world so that
 * we can find a global list of them quickly.
 */
class MultiSectorObject
{
private:
  /// Linked list in the world.
  MultiSectorObject* next;
  ///
  MultiSectorObject* prev;
  ///
  csWorld* world;

  /// List of Sectors that this MultiSectorObject represents.
  SectorMSO* first_sector;

  /**
   * Object that this MultiSectorObject represents (Thing
   * or 3D sprite).
   */
  csObject* object;

public:
  ///
  MultiSectorObject (csObject* obj);
  ///
  MultiSectorObject ();
  ///
  virtual ~MultiSectorObject ();

  ///
  csObject* get_object () { return object; }
  ///
  void set_object (csObject* obj) { object = obj; }

  ///
  csWorld* get_world () { return world; }
  ///
  void set_world (csWorld* w) { world = w; }

  ///
  void set_next (MultiSectorObject* mso) { next = mso; }
  ///
  MultiSectorObject* get_next () { return next; }
  ///
  void set_prev (MultiSectorObject* mso) { prev = mso; }
  ///
  MultiSectorObject* get_prev () { return prev; }

  /**
   * Unlink a SectorMSO from this MultiSectorObject.
   * Warning! This function does not test if the object
   * is really on the list!
   */
  void unlink_mso (SectorMSO* lp);

  /**
   * Add a SectorMSO to the list.
   */
  void add_mso (SectorMSO *lp);

  /**
   * Conveniance function to initialize this MSO with one sector.
   * @@@ RETHINK!
   */
  void init_one_sector (csSector* s);

  /**
   * Conveniance function to get the first sector of this MSO.
   * @@@ RETHINK!
   */
  csSector* get_first_sector () { return first_sector->sector; }
};

#endif /*MSECTOBJ_H*/
