/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "system/system.h"
#include "system/sysprint.h"
#include "engine/objects/cssprite.h"
#include "engine/camera.h"
#include "engine/world.h"
#include "engine/texture.h"
#include "util/parser.h"
#include "util/util.h"
#include "render/igraph3d.h"
#include "geometry/polyclip.h"
#include "general/loader/csloader.h"

csFrame::csFrame (int num_vertices)
{
  CHK (texels = new Vector2 [num_vertices]);
  CHK (vertices = new Vector3 [num_vertices]);
  name = NULL;
  max_vertex = num_vertices;
}

csFrame::~csFrame ()
{
  if (texels) CHKB (delete [] texels);
  if (vertices) CHKB (delete [] vertices);
}

void csFrame::add_vertex(int num)
{
  if(num<1) return;

  CHK(Vector2 * new_texels = new Vector2 [max_vertex + num]);
  CHK(Vector3 * new_vertices = new Vector3 [max_vertex + num]);
    
  memcpy(new_texels, texels, max_vertex*sizeof(Vector2));
  memcpy(new_vertices, vertices, max_vertex*sizeof(Vector3));
  CHK(delete [] texels);
  CHK(delete [] vertices);

  max_vertex += num;

  texels = new_texels;
  vertices = new_vertices;
}

csSpriteAction::csSpriteAction()
{
  name = NULL;
  num_frameset = max_frameset = 0;
  frameset = NULL;
  delay = NULL;
}

csSpriteAction::~csSpriteAction()
{
  if(frameset) CHKB (delete frameset);
  if(delay) CHKB (delete [] delay);
}

void csSpriteAction::add_frame(csFrame * f, int d)
{
  int plus = 5;
  if(num_frameset >= max_frameset)
  {
    CHK(csFrame ** new_frameset = new csFrame * [max_frameset + plus]);
    if(frameset)
    {
      memcpy(new_frameset, frameset, max_frameset*sizeof(csFrame *));
      CHK(delete [] frameset);
    }
    frameset = new_frameset;

    CHK(int * new_delay = new int [max_frameset + plus]);
    if(delay)
    {
      memcpy(new_delay, delay, max_frameset*sizeof(int));
      CHK(delete [] delay);
    }
    delay = new_delay;

    max_frameset += plus;
  }
  frameset[num_frameset] = f;
  delay[num_frameset] = d;
  num_frameset++;
}

csSpriteTemplate::csSpriteTemplate (char* name) : csObject (name, CS_SPRITETPL)
{
  texture = NULL;
  triangles = NULL;
  max_triangles = num_triangles = 0;
  frames = NULL;
  num_frames = 0;
  num_vertices = 0;
  actions = NULL;
  num_actions = 0;
}

csSpriteTemplate::~csSpriteTemplate ()
{
  if (triangles) CHKB (delete [] triangles);
  if (frames)
  {
    int i;
    for (i = 0 ; i < num_frames ; i++)
    {
      CHK (delete frames[i]);
    }
    CHK (delete [] frames);
  }
}

void csSpriteTemplate::add_triangle (int a, int b, int c)
{
  if (num_triangles >= max_triangles)
  {
    CHK (Triangle* new_triangles = new Triangle [max_triangles+8]);
    if (triangles)
    {
      memcpy (new_triangles, triangles, sizeof (Triangle)*max_triangles);
      CHK (delete [] triangles);
    }
    triangles = new_triangles;
    max_triangles += 8;
  }
  triangles[num_triangles].a = a;
  triangles[num_triangles].b = b;
  triangles[num_triangles].c = c;
  num_triangles++;
}

csFrame* csSpriteTemplate::add_frame ()
{
  CHK (csFrame** new_frames = new csFrame* [num_frames+1]);
  if (frames)
  {
    memcpy (new_frames, frames, sizeof (csFrame*)*num_frames);
    CHK (delete [] frames);
  }
  frames = new_frames;
  CHK (csFrame* fr = new csFrame (num_vertices));
  frames[num_frames++] = fr;
  return fr;
}

csFrame* csSpriteTemplate::find_frame(char *n)
{
  for(int i=0; i<num_frames; i++)
    if(strcmp(frames[i]->get_name(), n) == 0) return frames[i];

  return NULL;
}

csSpriteAction* csSpriteTemplate::add_action ()
{
  CHK (csSpriteAction** new_actions = new csSpriteAction *[num_actions+1]);
  if (actions)
  {
    memcpy (new_actions, actions, sizeof (csSpriteAction*)*num_actions);
    CHK (delete [] actions);
  }
  actions = new_actions;
  CHK(csSpriteAction* a = new csSpriteAction ());
  actions[num_actions++] = a;
  return a;
}

void csSpriteTemplate::set_texture (Textures* textures, char *texname)
{
  if (!textures)
  {
    CsPrintf (MSG_FATAL_ERROR, "There are no textures defined in this world file!\n");
    fatal_exit (0, true);
    return;
  }
  texture = textures->get_texture_mm (texname);
  if (texture == NULL)
  {
    CsPrintf (MSG_FATAL_ERROR, "Couldn't find texture named '%s'!\n", texname);
    fatal_exit (0, true);
    return;
  }
}

csSpriteAction* csSpriteTemplate::find_action(char *n)
{
  for(int i=0; i<num_actions; i++)
    if(strcmp(actions[i]->get_name(), n) == 0) return actions[i];

  return NULL;
}

enum { kTokenSTplTexNr = 1, kTokenSTplFrame,
       kTokenSTplTriangle, kTokenSTplAction };

void csSpriteTemplate::load (char* buf, Textures* textures)
{
  static tokenDesc commands[] = {
    {kTokenSTplTexNr, "TEXNR"},
    {kTokenSTplFrame, "FRAME"},
    {kTokenSTplAction, "ACTION"},
    {kTokenSTplTriangle, "TRIANGLE"},
    {0,0}};
  static tokenDesc tok_frame[] = {{2, "V"}, {0,0}};
  static tokenDesc tok_frameset[] = {{2, "F"}, {0,0}};
  char* name;
    
  long cmd;
  char* params;
  char* params2;
  char str[255];

  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    switch (cmd)
    {
      case kTokenSTplTexNr:
        {
          ScanStr (params, "%s", str);
          set_texture (textures, str);
        }
        break;

      case kTokenSTplAction:
        {
          csSpriteAction* act = add_action ();
          act->set_name(name);
          int d;
          char fn[64];
          while ((cmd = csGetObject (&params, tok_frameset, &name, &params2)) > 0)
          {
            switch (cmd)
            {
            case 2:
              ScanStr (params2, "%s,%d", &fn, &d);
              csFrame * ff = find_frame(fn);
              if(!ff)
              {
                CsPrintf (MSG_FATAL_ERROR, "Error! Trying to add a unknown frame '%s' in %s action !\n", fn, act->get_name());
                fatal_exit (0, false);
              }
              act->add_frame(ff, d);
              break;
            }
          }
        }
        break;

      case kTokenSTplFrame:
        {
          csFrame* fr = add_frame ();
          fr->set_name(name);
          int i = 0;
          float x, y, z, u, v;
          while ((cmd = csGetObject (&params, tok_frame, &name, &params2)) > 0)
          {
            switch (cmd)
            {
            case 2:
              ScanStr (params2, "%f,%f,%f:%f,%f", &x, &y, &z, &u, &v);
              // check if it's the first frame
              if(num_frames==1)
              {
                // add vertice/texel in current frame
                if(num_vertices >= fr->get_max_vertex())
                {
                  int more = 1;
                  num_vertices+=more;
                  fr->add_vertex(more);
                }
              }
              else if (i >= num_vertices)
              {
                CsPrintf (MSG_FATAL_ERROR, "Error! Trying to add too many vertices in frame '%s'!\n", fr->get_name());
                fatal_exit (0, false);
              }
              fr->set_vertice (i, x, y, z);
              fr->set_texel (i, u, v);
              i++;
              break;
            }
          }
          if (i < num_vertices)
          {
            CsPrintf (MSG_FATAL_ERROR, "Error! Too few vertices in frame '%s'! (%d %d)\n", fr->get_name(), i, num_vertices);
            fatal_exit (0, false);
          }
        }
        break;

      case kTokenSTplTriangle:
        {
          int a, b, c;
          ScanStr (params, "%d,%d,%d", &a, &b, &c);
          add_triangle (a, b, c);
        }
        break;
    }
  }
}

//=============================================================================

csSprite3D::csSprite3D (char* name) : csObject (name, CS_SPRITE3D)
{
  mso.set_object (this);
  v_obj2world.x = 0;
  v_obj2world.y = 0;
  v_obj2world.z = 0;
  cur_frame = 0;
  tr_frame = NULL;
  persp = NULL;
  visible = NULL;
  tpl = NULL;
  force_otherskin = false;
  cur_action = NULL;
  vertex_colors = NULL;
}

csSprite3D::~csSprite3D ()
{
  CHK (delete tr_frame);
  CHK (delete [] persp);
  CHK (delete [] visible);
  CHK (delete [] vertex_colors);
}


void csSprite3D::set_move (float x, float y, float z)
{
  v_obj2world.x = -x;
  v_obj2world.y = -y;
  v_obj2world.z = -z;
}

void csSprite3D::set_transform (Matrix3& matrix)
{
  m_obj2world = matrix;
  m_world2obj = m_obj2world.getinverse();
}

void csSprite3D::move (float dx, float dy, float dz)
{
  v_obj2world.x += dx;
  v_obj2world.y += dy;
  v_obj2world.z += dz;
}

void csSprite3D::transform (Matrix3& matrix)
{
  Matrix3 m = matrix;
  m *= m_obj2world;
  m_obj2world = m;
  m_world2obj = m_obj2world.getinverse();
}

void csSprite3D::set_template (csSpriteTemplate* tmpl)
{
  tpl = tmpl;
}

void csSprite3D::set_texture(char * name, Textures* textures)
{
  force_otherskin = true;
  texture = textures->get_texture_mm (name);
  if (texture == NULL)
  {
    CsPrintf (MSG_FATAL_ERROR, "Couldn't find texture named '%s'!\n", name);
    exit (0);
  }
}

void csSprite3D::set_vertex_color (int i, csColor& col)
{
  if (!vertex_colors)
  {
    CHK (vertex_colors = new csColor [tpl->get_num_vertices ()]);
    int j;
    for (j = 0 ; j < tpl->get_num_vertices (); j++)
      vertex_colors[j].set (0, 0, 0);
  }
  vertex_colors[i] = col;
}

void csSprite3D::reset_vertex_colors ()
{
  CHK (delete [] vertex_colors);
  vertex_colors = NULL;
}

void csSprite3D::draw (RenderView& rview)
{
  if (!tpl->texture)
  {
    CsPrintf (MSG_FATAL_ERROR, "Error! Trying to draw a sprite with no texture!\n");
    fatal_exit (0, false);
  }

  int i;

  // First create the transformation from object to camera space directly:
  //   W = Mow * O - Vow;
  //   C = Mwc * (W - Vwc)
  // ->
  //   C = Mwc * (Mow * O - Vow - Vwc)
  //   C = Mwc * Mow * O - Mwc * (Vow + Vwc)
  Matrix3 m_o2c = rview.m_world2this();
  m_o2c *= m_obj2world;
  Vector3 v_o2c = rview.world2this(-v_obj2world);

  csFrame * cframe = cur_action->get_frame(cur_frame);

  // Transform all vertices from object to camera space and do perspective
  // correction as well.
  for (i = 0 ; i < tpl->num_vertices ; i++)
  {
    Vector3 v;
    v = m_o2c * ( cframe->get_vertice (i) ) + v_o2c;
    tr_frame->set_vertice (i, v.x, v.y, v.z);
    tr_frame->set_texel(i, cframe->get_texel (i).x, cframe->get_texel (i).y);
    if (v.z >= SMALL_Z)
    {
      float iz = csCamera::aspect/v.z;
      persp[i].x = v.x * iz + (FRAME_WIDTH/2);
      persp[i].y = v.y * iz + (FRAME_HEIGHT/2);
      visible[i] = true;
    }
    else
      visible[i] = false;
  }

  // Clipped polygon (assume it cannot have more than 64 vertices)
  G3DPolygon poly;
  memset( &poly, 0, sizeof(G3DPolygon));

  CHK (poly.pi_texcoords = new G3DPolygon::poly_texture_def [64]);
  // The triangle in question
  Vector2 triangle [3];
  Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERTESTENABLE, true);
  Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERFILLENABLE, true);
  // Draw all triangles.
  for (i = 0 ; i < tpl->num_triangles ; i++)
  {
    int a = tpl->triangles[i].a;
    int b = tpl->triangles[i].b;
    int c = tpl->triangles[i].c;
    if (visible[a] && visible[b] && visible[c])
    {
      //-----
      // Do backface culling.
      //-----
      if (Vector2::Area2 (persp [a].x, persp [a].y,
                          persp [b].x, persp [b].y,
                          persp [c].x, persp [c].y) >= 0)
        continue;

      triangle [0] = persp[a];
      triangle [1] = persp[b];
      triangle [2] = persp[c];
      // Clip triangle
      int rescount;
      Vector2 *cpoly = rview.view->Clip (triangle, 3, rescount);
      if (!cpoly)
        continue;

      // Compute U & V in vertices of the polygon
      // First find the topmost triangle vertex
      int top;
      if (triangle [0].y < triangle [1].y)
        if (triangle [0].y < triangle [2].y) top = 0;
        else top = 2;
      else
        if (triangle [1].y < triangle [2].y) top = 1;
        else top = 2;

      poly.num = rescount;
      
      if(force_otherskin)
        poly.pi_texture = GetIMipMapContainerFromTextureMM(texture);
      else
        poly.pi_texture = GetIMipMapContainerFromTextureMM(tpl->texture);

      poly.pi_triangle = (Vector2 *)&triangle;
      int trivert [3] = { a, b, c };
      int j;
      for (j = 0; j < 3; j++)
      {
        poly.pi_tritexcoords [j].z = 1 / tr_frame->get_vertice (trivert [j]).z;
        poly.pi_tritexcoords [j].u = tr_frame->get_texel (trivert [j]).x;
        poly.pi_tritexcoords [j].v = tr_frame->get_texel (trivert [j]).y;
	if (vertex_colors)
	{
	  poly.pi_tritexcoords[j].r = vertex_colors[trivert[j]].red;
	  poly.pi_tritexcoords[j].g = vertex_colors[trivert[j]].green;
	  poly.pi_tritexcoords[j].b = vertex_colors[trivert[j]].blue;
	}
      } /* endfor */
      for (j = 0; j < rescount; j++)
      {
        float x = poly.vertices [j].sx = cpoly [j].x;
        float y = poly.vertices [j].sy = cpoly [j].y;

        // Find the triangle left/right edges between which
        // the given x,y, point is located.
        int vtl, vtr, vbl, vbr;
        vtl = vtr = top;
        vbl = (vtl + 1) % 3;
        vbr = (vtl + 3 - 1) % 3;
        if (y > triangle [vbl].y)
        {
          vtl = vbl;
          vbl = (vbl + 1) % 3;
        } else if (y > triangle [vbr].y)
        {
          vtr = vbr;
          vbr = (vbr + 3 - 1) % 3;
        } /* endif */

        // Now interpolate Z,U,V by Y
        float tL = triangle [vbl].y - triangle [vtl].y;
        if (tL)
          tL = (y - triangle [vtl].y) / tL;
        float tR = triangle [vbr].y - triangle [vtr].y;
        if (tR)
          tR = (y - triangle [vtr].y) / tR;
        float xL = triangle [vtl].x + tL * (triangle [vbl].x - triangle [vtl].x);
        float xR = triangle [vtr].x + tR * (triangle [vbr].x - triangle [vtr].x);
        float tX = xR - xL;
        if (tX)
          tX = (x - xL) / tX;

        #define INTERPOLATE(val,tl,bl,tr,br)	\
        {					\
          float vl,vr;				\
          if (tl != bl)				\
            vl = tl + (bl - tl) * tL;		\
          else					\
            vl = tl;				\
          if (tr != br)				\
            vr = tr + (br - tr) * tR;		\
          else					\
            vr = tr;				\
          val = vl + (vr - vl) * tX;		\
        }

        // Calculate Z
        INTERPOLATE (poly.pi_texcoords [j].z,
          poly.pi_tritexcoords [vtl].z, poly.pi_tritexcoords [vbl].z,
          poly.pi_tritexcoords [vtr].z, poly.pi_tritexcoords [vbr].z);
        // Calculate U
        INTERPOLATE (poly.pi_texcoords [j].u,
          poly.pi_tritexcoords [vtl].u, poly.pi_tritexcoords [vbl].u,
          poly.pi_tritexcoords [vtr].u, poly.pi_tritexcoords [vbr].u);
        // Calculate V
        INTERPOLATE (poly.pi_texcoords [j].v,
          poly.pi_tritexcoords [vtl].v, poly.pi_tritexcoords [vbl].v,
          poly.pi_tritexcoords [vtr].v, poly.pi_tritexcoords [vbr].v);
	if (vertex_colors)
	{
          // Calculate R
          INTERPOLATE (poly.pi_texcoords [j].r,
            poly.pi_tritexcoords [vtl].r, poly.pi_tritexcoords [vbl].r,
            poly.pi_tritexcoords [vtr].r, poly.pi_tritexcoords [vbr].r);
          // Calculate G
          INTERPOLATE (poly.pi_texcoords [j].g,
            poly.pi_tritexcoords [vtl].g, poly.pi_tritexcoords [vbl].g,
            poly.pi_tritexcoords [vtr].g, poly.pi_tritexcoords [vbr].g);
          // Calculate B
          INTERPOLATE (poly.pi_texcoords [j].b,
            poly.pi_tritexcoords [vtl].b, poly.pi_tritexcoords [vbl].b,
            poly.pi_tritexcoords [vtr].b, poly.pi_tritexcoords [vbr].b);
	}
      } /* endfor */
      CHK (delete[] cpoly);
      // Draw resulting polygon
      Gfx3D->DrawPolygonQuick (poly, vertex_colors != NULL);
    }
  }
  CHK (delete [] poly.pi_texcoords);
}

enum { kTokenSpriteTemplate = 1, kTokenSpriteMove, kTokenSpriteTexnr};

void csSprite3D::load (csWorld* w, char* buf, Textures* textures)
{
  static tokenDesc commands[] = {
  	{kTokenSpriteTemplate, "TEMPLATE"},
  {kTokenSpriteTexnr, "TEXNR"},
	{kTokenSpriteMove, "MOVE"},
	{0,0}};
  static tokenDesc tok_matvec[] = {{1, "MATRIX"}, {2, "V"}, {0,0}};
  char* name;

  long cmd;
  char* params;
  char str[255], str2[255];

  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    switch (cmd)
    {
      case kTokenSpriteMove:
        {
          char* params2;
          m_obj2world = Matrix3(); /* identity matrix */
          v_obj2world.x = v_obj2world.y = v_obj2world.z = 0;
          while ((cmd = csGetObject (&params, tok_matvec, &name, &params2)) > 0)
          {
            switch (cmd)
            {
             case 1: m_obj2world =  CSLoader::load_matrix (params2);  break;
             case 2: v_obj2world = -CSLoader::load_vector (params2);  break;
            }
          }
        }
        break;

      case kTokenSpriteTemplate:
        memset(str, 0, 255);
        memset(str2, 0, 255);
        ScanStr (params, "%s,%s", str, str2);
        tpl = w->get_sprite_template (str, true);
        if (tpl == NULL)
        {
          CsPrintf (MSG_FATAL_ERROR, "Couldn't find template named '%s'!\n", str);
          fatal_exit (0, false);
        }
        if(!(cur_action = tpl->find_action(str2)))
        {
          CsPrintf (MSG_FATAL_ERROR, "Couldn't find action named '%s' in %s template !\n", str2, str);
          fatal_exit (0, false);
        }
        break;

      case kTokenSpriteTexnr:
        memset(str, 0, 255);
        ScanStr (params, "%s", str);
        set_texture(str, textures);
//        unset_texture();
        break;
    }
  }

  init_sprite ();
}

void csSprite3D::init_sprite ()
{
  if (!tpl)
  {
    CsPrintf (MSG_FATAL_ERROR, "There is no defined template for this sprite!\n");
    fatal_exit (0, false);
  }

  if (!cur_action) { set_frame (0); cur_action = tpl->get_first_action (); }

  CHK (tr_frame = new csFrame (tpl->num_vertices));
  CHK (persp = new Vector2 [tpl->num_vertices]);
  CHK (visible = new bool [tpl->num_vertices]);

  last_time = SystemGetTime();

  m_world2obj = m_obj2world.getinverse();
}

bool csSprite3D::next_frame (long current_time, bool onestep, bool stoptoend)
{
  bool ret = false;

  if(onestep)
  {
    if(current_time>last_time+cur_action->get_frame_delay(cur_frame))
    {
      last_time = current_time;
      cur_frame++;
      if (cur_frame >= cur_action->get_num_frameset())
      {
        if(stoptoend) cur_frame --;
        else cur_frame = 0;
        ret = true;
      }
    }
  }
  else
  {
    while(1)
    {
      if(current_time>last_time+cur_action->get_frame_delay(cur_frame))
      {
        last_time += cur_action->get_frame_delay(cur_frame);
        cur_frame++;
        if (cur_frame >= cur_action->get_num_frameset())
        {
          if(stoptoend)
          {
            cur_frame--;
            ret = true;
            break;
          }
          cur_frame = 0;
          ret = true;
        }        
      }
      else break;
    }
  }

  return ret;
}

