/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef CSSPRITE_H
#define CSSPRITE_H

#include "geometry/math3d.h"
#include "geometry/math2d.h"
#include "engine/basic/csobject.h"
#include "engine/msectobj.h"
#include "engine/rview.h"
#include "engine/cscolor.h"
#include "colldet/collp.h"

class Textures;
class Dumper;

/**
 * A frame for a 3D sprite animation.
 */
class csFrame
{
private:
  Vector3* vertices;
  Vector2* texels;
  char *name;
  int max_vertex;

public:
  ///
  csFrame (int max_vertices);
  ///
  virtual ~csFrame ();

  ///
  void set_vertice (int i, float x, float y, float z)
  {
    vertices[i].x = x;
    vertices[i].y = y;
    vertices[i].z = z;
  }
  ///
  void set_texel (int i, float u, float v)
  {
    texels[i].x = u;
    texels[i].y = v;
  }
  ///
  void set_name(char * n)
  {
    if (name) CHKB (delete [] name);
    if (n)
    {
      CHK (name = new char [strlen (n)+1]);
      strcpy (name, n);
    }
    else
      name = n;
  }

  ///
  Vector3& get_vertice (int i) { return vertices[i]; }
  ///
  Vector2& get_texel (int i) { return texels[i]; }
  ///
  char * get_name() {return name;}

  ///
  void add_vertex(int num_vertices);
  ///
  int get_max_vertex() {return max_vertex;}
};

/**
 * A Action frameset for a 3D sprite animation.
 */
class csSpriteAction
{
public:
  ///
  csSpriteAction();
  ///
  virtual ~csSpriteAction();

  ///
  void add_frame(csFrame * frame, int delay);
  ///
  void set_name(char *n)
  {
    if(name) CHKB (delete [] name);
    CHK (name = new char [strlen (n)+1]);
    strcpy (name, n);
  }
  ///
  char * get_name() {return name;}
  ///
  csFrame * get_frame(int f)
  {
    if(f<num_frameset) return frameset[f];
    return NULL;
  }
  ///
  int get_num_frameset() {return num_frameset;}
  ///
  int get_frame_delay(int i) {return delay[i]; }

private:
  csFrame  ** frameset;
  int *delay;
  int num_frameset;
  int max_frameset;

  char *name;
};

/// A triangle for 3D sprites.
struct Triangle
{
  int a, b, c;
};

class csSprite3D;

/**
 * A 3D sprite based on a triangle mesh with a single texture.
 * Animation is done with frames.
 * This class represents a template from which a csSprite3D
 * class can be made.
 */
class csSpriteTemplate : public csObject
{
  friend Dumper;

private:
  friend class csSprite3D;
  friend class Collider;

  /// The texture.
  TextureMM* texture;

  /// The vertices.
  int num_vertices;

  /// The triangles.
  Triangle* triangles;
  int num_triangles;
  int max_triangles;

  /// The frames.
  csFrame** frames;
  int num_frames;

  /// The actions.
  csSpriteAction **actions;
  int num_actions;

public:
  /// Create the sprite template
  csSpriteTemplate (char* name);
  /// Destroy the template
  virtual ~csSpriteTemplate ();

  /// Add a triangle to the sprite.
  void add_triangle (int a, int b, int c);
  /// Query the array of sprite polygons
  Triangle* get_triangles () { return triangles; }
  /// Query the number of sprite polygons
  int get_num_triangles () { return num_triangles; }

  /// Query the number of vertices
  int get_num_vertices () { return num_vertices; }

  /// Create and add a new frame to the sprite.
  csFrame * add_frame ();
  /// find a named frame into the sprite.
  csFrame * find_frame(char * name);
  /// Query the number of frames
  int get_num_frames () { return num_frames; }
  /// Query the array of frames
  csFrame** get_frames () { return frames; }

  /// Create and add a new action frameset to the sprite.
  csSpriteAction * add_action ();
  /// find a named action into the sprite.
  csSpriteAction * find_action (char * name);
  /// Get the first action.
  csSpriteAction * get_first_action () { return actions[0]; }
  /// Get number of actions in sprite
  int get_num_actions () { return num_actions; }
  /// Get action number No
  csSpriteAction * get_action (int No) { return actions [No]; }

  /// Load the sprite template from the world file.
  void load (char* buf, Textures* textures);

  /// Query the texture used for this sprite
  TextureMM *get_texture () const { return texture; }
  /// Set the texture used for this sprite
  void set_texture (Textures* textures, char *texname);
};

/**
 * A 3D sprite based on a triangle mesh with a single texture.
 * Animation is done with frames.
 */
class csSprite3D : public csObject
{
  friend Dumper;
  friend Collider;

private:
  /// A Sprite is a MultiSectorObject.
  MultiSectorObject mso;

  /// Object to world transformation.
  Vector3 v_obj2world;
  /// Object to world transformation.
  Matrix3 m_obj2world;
  /// World to object transformation.
  Matrix3 m_world2obj;

  /**
   * Array of colors for the vertices. If not set then this
   * sprite does not have colored vertices.
   */
  csColor* vertex_colors;

  /// The template.
  csSpriteTemplate* tpl;

  /// The texture skin.
  TextureMM* texture;

  /// The current frame number.
  int cur_frame;
  /// The current action.
  csSpriteAction * cur_action;

  /// The transformed frame (from object->camera space).
  csFrame* tr_frame;

  /// The perspective corrected vertices (screen space) for tr_frame.
  Vector2* persp;

  /// Array which indicates which vertices are visible and which are not.
  bool* visible;

  /// The last frame time action
  int last_time;

  //
  bool force_otherskin;

public:
  ///
  csSprite3D (char* name);
  ///
  virtual ~csSprite3D ();

  ///
  void set_template (csSpriteTemplate* tmpl);

  ///
  csSpriteTemplate* get_template () { return tpl; }

  ///
  MultiSectorObject& get_mso () { return mso; }

  /// force a new texture skin other than default
  void set_texture(char * name, Textures* textures);

  /**
   * Set a color for a vertex.
   * As soon as you use this function this sprite will be rendered
   * using gouroud shading. Calling this function for the first time
   * will initialize all colors to black.
   */
  void set_vertex_color (int i, csColor& col);

  /**
   * Reset the color list. If you call this function then the
   * sprite will no longer use gouroud shading.
   */
  void reset_vertex_colors ();

  ///
  void unset_texture()
  { force_otherskin = false; }

  /**
   * Set the transformation vector to move sprite to some position.
   */
  void set_move (Vector3& v) { set_move (v.x, v.y, v.z); }

  /**
   * Set the transformation vector to move sprite to some position.
   */
  void set_move (float x, float y, float z);

  /**
   * Set the transformation matrix to rotate the sprite in some
   * orientation.
   */
  void set_transform (Matrix3& matrix);

  /**
   * Relative move.
   */
  void move (float dx, float dy, float dz);

  /**
   * Relative move.
   */
  void move (Vector3& v) { move (v.x, v.y, v.z); }

  /**
   * Relative transform.
   */
  void transform (Matrix3& matrix);

  /**
   * Draw this sprite given a camera transformation.
   */
  void draw (RenderView& rview);

  /**
   * Go to the next frame depending on the current time in milliseconds.
   */
  bool next_frame (long current_time, bool onestep = false, bool stoptoend = false);

  /**
   * Go to a specified frame.
   */
  void set_frame (int f)
  {
    if (cur_action && f < cur_action->get_num_frameset()) cur_frame = f;
  }

  /**
   * Get the current frame number.
   */
  int get_cur_frame () { return cur_frame; }

  /**
   * Get the current frame number.
   */
  csSpriteAction * get_cur_action () { return cur_action; }

  /**
   * Get the number of frames.
   */
  int get_num_frames () { return cur_action->get_num_frameset(); }

  /**
   * Select an action.
   */
  void set_action (char * name)
  {
    csSpriteAction *act;
    if((act=tpl->find_action(name)))
    {
      set_frame(0);
      cur_action = act;
    }
  }

  /**
   * Load the sprite from the world file.
   */
  void load (csWorld* w, char* buf, Textures* textures);

  /**
   * Initialize a sprite. This function is called automatically
   * from within 'load'. However you should call it directly
   * if you created the sprite on the fly (without 'load').
   */
  void init_sprite ();

  /// Get world to local transformation matrix
  inline Matrix3 m_world2this() const { return m_world2obj; }
  /// Get world to local translation
  inline Vector3 v_world2this() const { return -v_obj2world; }

  /**
   * Collision Detection data.
   */
  ColliderP collider;
};

#endif /*CSSPRITE_H*/
