/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/system.h"
#include "engine/objects/thing.h"
#include "engine/objects/thingtpl.h"
#include "engine/polygon/polygon.h"
#include "engine/polygon/polytext.h"
#include "engine/light/light.h"
#include "engine/texture.h"
#include "engine/world.h"
#include "engine/stats.h"
#include "render/igraph3d.h"
#include "engine/sector.h"
#include "engine/wirefrm.h"
#include "engine/curve.h"


//---------------------------------------------------------------------------

Thing::Thing (char* name) : PolygonSet (name, CS_THING), obj() 
{
  moveable = false;
  merged = NULL;
  convex = false;
  center_idx = -1;
}

Thing::~Thing ()
{
  if (merged)
  {
    int i;
    for (i = 0 ; i < num_polygon ; i++)
      polygons[i] = NULL;
  }
}

void Thing::setConvex (bool c)
{
  convex = c;
  if (convex)
  {
    if (center_idx == -1) center_idx = add_vertex (0, 0, 0);
    int i;
    float minx = 1000000000., miny = 1000000000., minz = 1000000000.;
    float maxx = -1000000000., maxy = -1000000000., maxz = -1000000000.;
    for (i = 0 ; i < num_vertices ; i++)
      if (i != center_idx)
      {
        if (obj_verts[i].x < minx) minx = obj_verts[i].x;
	if (obj_verts[i].x > maxx) maxx = obj_verts[i].x;
        if (obj_verts[i].y < miny) miny = obj_verts[i].y;
	if (obj_verts[i].y > maxy) maxy = obj_verts[i].y;
        if (obj_verts[i].z < minz) minz = obj_verts[i].z;
	if (obj_verts[i].z > maxz) maxz = obj_verts[i].z;
      }
    obj_verts[center_idx].set ((minx+maxx)/2, (miny+maxy)/2, (minz+maxz)/2);
    wor_verts[center_idx].set ((minx+maxx)/2, (miny+maxy)/2, (minz+maxz)/2);
  }
}

void Thing::transform ()
{
  int i;
  for (i = 0 ; i < num_vertices ; i++)
    wor_verts[i] = obj.this2world(obj_verts[i]);
  for (i = 0 ; i < num_polygon ; i++)
  {
    Polygon3D* p = (Polygon3D*)polygons[i];
    p->object_to_world (obj);
  }
}

void Thing::set_move (csSector* home, float x, float y, float z)
{
  obj.set_origin(Vector3(x,y,z));
  sector = home;
}

void Thing::set_transform (Matrix3& matrix)
{
  obj.set_this2world(matrix);
}

void Thing::move (float dx, float dy, float dz)
{
  obj.translate(Vector3(dx,dy,dz));
}

void Thing::transform (Matrix3& matrix)
{
  obj.set_this2world(matrix * obj.m_this2world());
}

void Thing::mipmap_settings (int setting)
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
  {
    Polygon3D* p = (Polygon3D*)polygons[i];
    p->mipmap_settings (setting);
  }
}

Polygon3D* Thing::intersect_sphere (Vector3& center, float radius, float* pr)
{
  float d, min_d = radius;
  int i;
  Polygon3D* p, * min_p = NULL;
  PolyPlane* pl;
  Vector3 hit;
  float A, B, C, D;

  for (i = 0 ; i < num_polygon ; i++)
  {
    p = (Polygon3D*)polygons[i];
    pl = p->get_plane ();
    d = pl->distance (center);
    if (d < min_d && pl->visible_from_point (center))
    {
      pl->get_world_normal (&A, &B, &C, &D);
      hit.x = d*(-A-center.x)+center.x;
      hit.y = d*(-B-center.y)+center.y;
      hit.z = d*(-C-center.z)+center.z;
      if (p->intersect_ray (center, hit))
      {
	min_d = d;
	min_p = p;
      }
    }
  }
  if (pr) *pr = min_d;
  return min_p;
}


#	define INTERPOLATE(val,tl,bl,tr,br)	\
        {					\
          float vl,vr;				\
          if (tl != bl)				\
            vl = tl + (bl - tl) * tL;		\
          else					\
            vl = tl;				\
          if (tr != br)				\
            vr = tr + (br - tr) * tR;		\
          else					\
            vr = tr;				\
          val = vl + (vr - vl) * tX;		\
        }


void Thing::draw_curves (RenderView& rview, bool use_z_buf)
{
  int i;
  int res=1;
  
  // Calculate tesselation resolution
  Vector3 wv = curves_center;
  Vector3 world_coord = obj.this2world (wv);
  Vector3 camera_coord = rview.world2this (world_coord);
  
  if (camera_coord.z >= SMALL_Z)
    {
      res=(int)(curves_scale/camera_coord.z);
      if (res<2)
	res=2;
      else if (res>9)
	res=9;
    }
  else
    res=9;

  //CsPrintf(MSG_WARNING,"res=%d\n",res);
  
  // Loop over all curves
  csCurve* c;
  for (i = 0 ; i < num_curves ; i++)
    {
      int j;
      c = curves[i];
      csCurveTesselated* tess = c->tesselate (res);
      
      // First I transform all tesselated vertices from object to
      // world space and then from world to camera space. @@@ NOTE!
      // These transformations should be combined for better efficiency!
      for (j = 0 ;  j < tess->getNumVertices () ; j++)
	{
	  csCurveVertex& cv = tess->getVertex (j);
	  cv.world_coord = obj.this2world (cv.object_coord);
	  cv.camera_coord = rview.world2this (cv.world_coord);
	}
      
      // Clipped polygon (assume it cannot have more than 64 vertices)
      
      // Draw all the triangles within this curve
      G3DPolygon poly;
      memset( &poly, 0, sizeof(G3DPolygon));
      Vector2 persp[3];
      CHK (poly.pi_texcoords = new G3DPolygon::poly_texture_def [64]);
      Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERTESTENABLE, use_z_buf);
      Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERFILLENABLE, true);
      for (j = 0 ; j < tess->getNumTriangles () ; j++)
	{
	  csCurveTriangle& ct = tess->getTriangle (j);
	  
	  // Transform all vertices from camera space to perspective 
	  // correct coords
	  
	  Vector3* varr[3];
	  varr[0] = &tess->getVertex (ct.i1).camera_coord;
	  varr[1] = &tess->getVertex (ct.i2).camera_coord;
	  varr[2] = &tess->getVertex (ct.i3).camera_coord;
	  
	  Vector2* texes[3];
	  texes[0] = &tess->getVertex (ct.i1).txt_coord;
	  texes[1] = &tess->getVertex (ct.i2).txt_coord;
	  texes[2] = &tess->getVertex (ct.i3).txt_coord;
	  
	  int visible = true;
	  int k;
	  for (k = 0 ; k < 3 ; k++)
	    {
	      if (varr[k]->z >= SMALL_Z)
		{
		  float iz = csCamera::aspect/varr[k]->z;
		  persp[k].x = varr[k]->x * iz + (FRAME_WIDTH/2);
		  persp[k].y = varr[k]->y * iz + (FRAME_HEIGHT/2);
		}
	      else
		visible = false;
	    }
	  
	  // Draw all triangles.
	  if (visible)
	    {
	      //-----
	      // Do backface culling.
	      //-----
	      if (Vector2::Area2 (persp [0].x, persp [0].y,
				  persp [1].x, persp [1].y,
				  persp [2].x, persp [2].y) >= 0)
		continue;

	      
	      // Clip triangle
	      int rescount;
	      Vector2 *cpoly = rview.view->Clip (persp, 3, rescount);
	      if (!cpoly)
		{
		  continue;
		}
	      
	      // Compute U & V in vertices of the polygon
	      // Find first the topmost triangle vertex
	      int top;
	      if (persp [0].y < persp [1].y)
		if (persp [0].y < persp [2].y)
		  top = 0;
		else
		  top = 2;
	      else
		if (persp [1].y < persp [2].y)
		  top = 1;
		else
		  top = 2;
	      
	      poly.num = rescount;
	      poly.pi_texture = GetIMipMapContainerFromTextureMM(c->getTexture());
	      
	      poly.pi_triangle = (Vector2 *)&persp;
	      int j;
	      for (j = 0; j < 3; j++)
		{
		  if (varr[j]->z > 0.0001)
		    poly.pi_tritexcoords [j].z = 1 / varr[j]->z;
		  poly.pi_tritexcoords [j].u = texes[j]->x;
		  poly.pi_tritexcoords [j].v = texes[j]->y;
		} // endfor 
	      
	      for (j = 0; j < rescount; j++)
		{
		  float x = poly.vertices [j].sx = cpoly [j].x;
		  float y = poly.vertices [j].sy = cpoly [j].y;
		  
		  // Find the triangle left/right edges between which
		  // is located given x,y point
		  int vtl, vtr, vbl, vbr;
		  vtl = vtr = top;
		  vbl = (vtl + 1) % 3;
		  vbr = (vtl + 3 - 1) % 3;
		  if (y > persp [vbl].y)
		    {
		      vtl = vbl;
		      vbl = (vbl + 1) % 3;
		    } 
		  else if (y > persp [vbr].y)
		    {
		      vtr = vbr;
		      vbr = (vbr + 3 - 1) % 3;
		    } // endif 
		  
		  // Now interpolate Z,U,V by Y
		  float tL = persp [vbl].y - persp [vtl].y;
		  if (tL< -0.001 || tL>0.001)
		    tL = (y - persp [vtl].y) / tL;
		  float tR = persp [vbr].y - persp [vtr].y;
		  if (tR< -0.001 || tR>0.001)
		    tR = (y - persp [vtr].y) / tR;
		  float xL = persp [vtl].x + tL * (persp [vbl].x - persp [vtl].x);
		  float xR = persp [vtr].x + tR * (persp [vbr].x - persp [vtr].x);
		  float tX = xR - xL;
		  if (tX< -0.001 || tX>0.001)
		    tX = (x - xL) / tX;
		  
		  // Calculate Z
		  INTERPOLATE (poly.pi_texcoords [j].z,
			       poly.pi_tritexcoords [vtl].z, poly.pi_tritexcoords [vbl].z,
			       poly.pi_tritexcoords [vtr].z, poly.pi_tritexcoords [vbr].z);
		  // Calculate U
		  INTERPOLATE (poly.pi_texcoords [j].u,
			       poly.pi_tritexcoords [vtl].u, poly.pi_tritexcoords [vbl].u,
			       poly.pi_tritexcoords [vtr].u, poly.pi_tritexcoords [vbr].u);
		  // Calculate V
		  INTERPOLATE (poly.pi_texcoords [j].v,
			       poly.pi_tritexcoords [vtl].v, poly.pi_tritexcoords [vbl].v,
			       poly.pi_tritexcoords [vtr].v, poly.pi_tritexcoords [vbr].v);
		} // endfor 
	      CHK (delete[] cpoly);
	      // Draw resulting polygon
	      
	      Gfx3D->DrawPolygonQuick (poly, false);
	    }
	}
      CHK (delete [] poly.pi_texcoords);
    }
}

void Thing::draw (RenderView& rview, bool use_z_buf)
{
  draw_busy++;
  Vector3* old = new_transformation ();

  WireFrame* wf = NULL;
  if (System->world->map_mode != MAP_OFF) wf = System->world->wf->get_wireframe ();

  if (transform_world2cam (rview))
  {
    Stats::polygons_considered += num_polygon;

    draw_curves (rview, use_z_buf);

    int res=1;

    // Calculate tesselation resolution
    // TODO should depend on a point in space, and a scale param in place of 40
    if (num_vertices>0)
    {
      Vector3 wv = wor_verts[0];
      Vector3 world_coord = obj.this2world (wv);
      Vector3 camera_coord = rview.world2this (world_coord);
	
      if (camera_coord.z > 0.0001)
      {
        res=(int)(40.0/camera_coord.z);
        if (res<1) res=1;
        else if (res>9) res=9;
      }
    }

    draw_polygon_array (polygons, num_polygon, &rview, use_z_buf);
  }

  restore_transformation (old);
  draw_busy--;
}

void Thing::draw_foggy (RenderView& d)
{
  draw_busy++;
  Vector3* old = new_transformation ();
  Polygon3D* p;
  Vector3* verts;
  int num_verts;
  int i;

  WireFrame* wf = NULL;
  if (System->world->map_mode != MAP_OFF) wf = System->world->wf->get_wireframe ();

  if (transform_world2cam (d))
  {
    Vector2 orig_triangle[3];
    Gfx3D->OpenFogObject (get_id (), &get_fog ());
    Stats::polygons_considered += num_polygon;

    d.set_mirror (!d.is_mirrored ());
    for (i = 0 ; i < num_polygon ; i++)
    {
      p = (Polygon3D*)polygons[i];
      if (p->dont_draw) continue;
      bool front = p->get_plane ()->visible_from_point (d.v_origin ());

      if ( !front &&
           p->clip_to_plane (d.clip_plane, d.v_origin (), verts, num_verts, false) &&
           p->do_perspective (d, verts, num_verts, &Polygon2D::clipped, orig_triangle) &&
	   Polygon2D::clipped.clip_against (d.view) )
      {
        if (wf)
        {
          int j;
          wfPolygon* po = wf->add_polygon ();
          po->set_vis_color (wf->get_yellow ());
          po->set_num_vertices (p->get_num_vertices ());
          for (j = 0 ; j < p->get_num_vertices () ; j++)
            po->set_vertex (j, p->vwor (j));
          po->prepare ();
        }
        Stats::polygons_drawn++;

	Polygon2D::clipped.add_fog_polygon (p, p->get_plane (), d.is_mirrored (),
		get_id (), CS_FOG_BACK);

        long do_edges;
        Gfx3D->GetRenderState (G3DRENDERSTATE_EDGESENABLE, do_edges);
	if (do_edges)
          Polygon2D::clipped.draw (System->world->get_textures ()->white ());
      }
    }
    d.set_mirror (!d.is_mirrored ());
    for (i = 0 ; i < num_polygon ; i++)
    {
      p = (Polygon3D*)polygons[i];
      if (p->dont_draw) continue;
      bool front = p->get_plane ()->visible_from_point (d.v_origin ());

      if ( front &&
           p->clip_to_plane (d.clip_plane, d.v_origin (), verts, num_verts, true) &&
           p->do_perspective (d, verts, num_verts, &Polygon2D::clipped, orig_triangle) &&
	   Polygon2D::clipped.clip_against (d.view) )
      {
        if (wf)
        {
          int j;
          wfPolygon* po = wf->add_polygon ();
          po->set_vis_color (wf->get_yellow ());
          po->set_num_vertices (p->get_num_vertices ());
          for (j = 0 ; j < p->get_num_vertices () ; j++)
            po->set_vertex (j, p->vwor (j));
          po->prepare ();
        }
        Stats::polygons_drawn++;

	Polygon2D::clipped.add_fog_polygon (p, p->get_plane (), d.is_mirrored (),
		get_id (), CS_FOG_FRONT);

        long do_edges;
        Gfx3D->GetRenderState (G3DRENDERSTATE_EDGESENABLE, do_edges);
	if (do_edges)
          Polygon2D::clipped.draw (System->world->get_textures ()->white ());
      }
    }
    Gfx3D->CloseFogObject (get_id ());
  }

  restore_transformation (old);
  draw_busy--;
}


void Thing::shine_lightmaps (LightView& lview)
{
  Polygon3D* p;
  int i;

  draw_busy++;
  Vector3* old = new_transformation ();

  translate_vector (lview.center);
  for (i = 0 ; i < num_polygon ; i++)
  {
    p = (Polygon3D*)polygons[i];
    Vector3* new_frustrum = NULL;
    int new_num_frustrum = 0;
    if (p->clip_frustrum (lview.center, lview.frustrum, lview.num_frustrum, lview.mirror,
    	&new_frustrum, &new_num_frustrum))
    {
      // First we check if the light can reach the polygon. If not then there is no
      // need to include it in the view frustrum. This is especially useful for portals
      // since then whole sectors can be culled away efficiently.
      float dist_to_plane = p->get_poly_plane ()->distance (lview.center);

      bool faraway = true;
      if (dist_to_plane < lview.l->get_dist ())
      {
        // The light is close enough to the plane of the polygon. Now we calculate
	// an accurate minimum squared distance from the light to the polygon. Note
	// that we use the new_frustrum which is relative to the center of the light.
	// So this algorithm works with the light position at (0,0,0) (@@@ we should
	// use this to optimize this algorithm further).
	Vector3 o (0, 0, 0);
	float min_sqdist = SqDist::point_poly (o, new_frustrum, new_num_frustrum,
		*(p->get_poly_plane ()), dist_to_plane*dist_to_plane);
	if (min_sqdist < lview.l->get_sq_dist ()) faraway = false;	// Light reaches polygon.
      }

      if (!faraway)
      {
        LightView new_lview = lview;
        new_lview.frustrum = new_frustrum;
        new_lview.num_frustrum = new_num_frustrum;
        p->shine_lightmaps (new_lview);
        if (!new_lview.dynamic && csSector::do_radiosity)
        {
          // If there is no portal we simulate radiosity by creating
	  // a dummy portal for this polygon which reflects light.
	  PortalCS mirror;
	  mirror.set_sector (p->get_sector ());
	  mirror.set_alpha (10);
	  mirror.set_filter (
		  p->get_texture ()->get_mean_r ()/4,
		  p->get_texture ()->get_mean_g ()/4,
		  p->get_texture ()->get_mean_b ()/4);

	  mirror.set_warp (p->get_poly_plane ()->get_reflect ());
	  mirror.shine_lightmaps (new_lview);
        }
      }
      else
      {
        if (new_frustrum) CHKB (delete [] new_frustrum);
      }
    }
  }

  restore_transformation (old);
  draw_busy--;
}

void Thing::dump_frustrum (csStatLight* l, Vector3* frustrum, int num_frustrum,
	Transform& t)
{
  draw_busy++;
  Vector3* old = new_transformation ();

  translate_vector (l->get_center ());

  int i;
  Polygon3D* p;
  for (i = 0 ; i < num_polygon ; i++)
  {
    p = (Polygon3D*)polygons[i];
    Vector3* new_frustrum = NULL;
    int new_num_frustrum = 0;
    if (p->clip_frustrum (l->get_center (), frustrum, num_frustrum, false/*@@@unsupported*/,
    	&new_frustrum, &new_num_frustrum))
    {
      int white = System->world->get_textures ()->white ();
      int red = System->world->get_textures ()->red ();
      int j;
      Vector3 light_cam;
      Vector3 v0, v1, v2;
      light_cam = t.world2this(l->get_center ());

      for (j = 0 ; j < new_num_frustrum ; j++)
      {
        v0 = new_frustrum[j] + l->get_center ();
        v1 = t.world2this(v0);
        v0 = new_frustrum[(j+1)%new_num_frustrum] + l->get_center ();
        v2 = t.world2this(v0);
        Gfx3D->DrawLine (light_cam, v1, red);
        Gfx3D->DrawLine (light_cam, v2, red);
        Gfx3D->DrawLine (v1, v2, white);
      }
    }
    if (new_frustrum) CHKB (delete [] new_frustrum);
  }

  restore_transformation (old);
  draw_busy--;
}

void Thing::init_lightmaps (bool do_cache)
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
    ((Polygon3D*)polygons[i])->init_lightmaps (do_cache);
}

void Thing::cache_lightmaps ()
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
    ((Polygon3D*)polygons[i])->cache_lightmaps ();
}

void Thing::merge (Thing* other)
{
  int i, j;
  other->merged = this;
  CHK (int *merge_vertices = new int [other->get_num_vertices ()+1]);
  for (i = 0 ; i < other->get_num_vertices () ; i++)
    merge_vertices[i] = add_vertex_smart (other->vwor (i));

  for (i = 0 ; i < other->get_num_polygon () ; i++)
  {
    Polygon3D* p = (Polygon3D*)(other->get_polygon (i));
    int* idx = p->get_vertices_idx ();
    for (j = 0 ; j < p->get_num_vertices () ; j++)
      idx[j] = merge_vertices[idx[j]];
    p->set_parent (this);
    add_polygon (p);
  }

  CHK (delete [] merge_vertices);
}


void Thing::merge_template (ThingTemplate* tpl,
	TextureMM* default_texture, float default_texlen,
	CLights* default_lightx)
{
  (void)default_texlen;
  int i, j;
  int* merge_vertices;

  //TODO should merge? take averages or something?
  curves_center = tpl->curves_center;
  curves_scale = tpl->curves_scale;

  CHK (merge_vertices = new int [tpl->get_num_vertices ()+1]);
  for (i = 0 ; i < tpl->get_num_vertices () ; i++)
    merge_vertices[i] = add_vertex_smart (tpl->vtex (i));
  
  for (i = 0 ; i < tpl->get_num_polygon () ; i++)
  {
    PolygonTemplate* pt = tpl->get_polygon (i);
    Polygon3D* p;
    p = new_polygon (pt->get_name (), pt->get_texture ());
    if (!pt->get_texture ()) p->set_texture (default_texture);
    p->theDynLight = default_lightx;
    int* idx = pt->get_vertices_idx ();
    for (j = 0 ; j < pt->get_num_vertices () ; j++)
      p->add_vertex (merge_vertices[idx[j]]);
    p->set_no_lighting (pt->get_no_lighting ());
    p->set_no_mipmap (pt->get_no_mipmap ());
    p->set_texture_space (pt->get_texture_matrix (), pt->get_texture_vector ());
  }

  for (i=0;i< tpl->get_num_curve_vertices();i++)
    {
      add_curve_vertex (tpl->curve_vertex(i),tpl->curve_texel(i));
    }

  for (i = 0 ; i < tpl->get_num_curves () ; i++)
  {
    csCurveTemplate* pt = tpl->get_curve (i);
    csCurve* p = pt->makeCurve(pt->get_name (), 0x1234);
    p->set_parent(this);

    if (!pt->get_texture ()) p->setTexture (default_texture);
    for (j = 0 ; j < pt->num_vertices() ; j++)
    {
      p->set_control_point(j,pt->get_vertex(j));
    }
    add_curve(p);
  }

  CHK (delete [] merge_vertices);
}

//---------------------------------------------------------------------------
