/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef THINGTPL_H
#define THINGTPL_H

#include "geometry/math3d.h"
#include "geometry/math2d.h"  // texel coords
#include "engine/basic/csobject.h"
#include "engine/basic/fog.h"

class PolygonTemplate;
class TextureMM;

class csCurveTemplate;

/**
 * This class represents a template for a Thing.
 * Using this template one can construct a Thing later.
 */
class ThingTemplate : public csObject
{
private:
  /// Vertices of this template.
  Vector3* vertices;
  /// Number of vertices.
  int num_vertices;
  /// Maximum number of vertices.
  int max_vertices;

  /**
   * List of polygontemplates.
   */
  PolygonTemplate** polygon;
  /// Number of polygons.
  int num_polygon;
  /// Maximum number of polygons.
  int max_polygon;

  /**
   * List of curve templates.
   */
  csCurveTemplate** curves;
  /// Number of curves.
  int num_curves;
  /// Maximum number of polygons.
  int max_curves;


  /// Vertices to be used by curves
  Vector3* curve_vertices;
  /// Texture coords of curve vertices
  Vector2* curve_texels;

  /// Number of vertices.
  int num_curve_vertices;
  /// Maximum number of vertices.
  int max_curve_vertices;

  /// Fog structure.
  csFog fog;

 public:
  /**
   * Global thing parameters used for determining tesselation resolution
   */
  /// Center of thing to determine distance from
  Vector3 curves_center;
  /// scale param (the larger this param it, the more 
  /// the curves are tesselated)
  float   curves_scale;  

public:
  /// Add a vertex to this template.
  void add_vertex (float x, float y, float z);
  /// Add a vertex to this template.
  void add_vertex (Vector3& v) { add_vertex (v.x, v.y, v.z); }
  /// Add a polygon template to this thing template.
  void add_polygon (PolygonTemplate* p);

  void add_curve (csCurveTemplate* p);
  int get_num_curves () { return num_curves; }
  csCurveTemplate* get_curve (int i) { return curves[i]; }

  ///
  int get_num_curve_vertices () { return num_curve_vertices; }
  ///
  Vector3& curve_vertex (int i) { return curve_vertices[i]; }
  ///
  Vector2& curve_texel (int i) { return curve_texels[i]; }
  ///
  void add_curve_vertex (Vector3& v, Vector2& t);
  ///

public:
  ///
  ThingTemplate (char* name);
  ///
  virtual ~ThingTemplate ();

  ///
  int get_num_vertices () { return num_vertices; }
  ///
  Vector3& vtex (int i) { return vertices[i]; }
  ///
  int get_num_polygon () { return num_polygon; }
  ///
  PolygonTemplate* get_polygon (int i) { return polygon[i]; }

  /// Return true if this has fog.
  bool has_fog () { return fog.enabled; }

  /// Return fog structure.
  csFog& get_fog () { return fog; }
};

/**
 * A PolygonTemplate is used by a ThingTemplate to describe a
 * template for a polygon.
 */
class PolygonTemplate
{
private:
  /// A table of indices into the vertices of the parent Polygon
  int* vertices_idx;
  /// Number of vertices for this polygon.
  int num_vertices;
  /// Maximum number of vertices.
  int max_vertices;
  ///
  char name[30];

  /// Should mipmapping be used for this polygon?
  int no_mipmap;
  /// Should this polygon be lighted?
  int no_lighting;
  /// The texture used for this polygon.
  TextureMM* texture;
  /// The transformation from object space to texture space.
  Matrix3 m_obj2tex;
  /// The transformation from object space to texture space.
  Vector3 v_obj2tex;

  ///
  ThingTemplate* parent;

public:
  ///
  PolygonTemplate (ThingTemplate* parent, char* name, TextureMM* texture = NULL);
  ///
  virtual ~PolygonTemplate ();

  ///
  int get_max_vertices () { return max_vertices; }
  ///
  int get_num_vertices () { return num_vertices; }
  ///
  int* get_vertices_idx () { return vertices_idx; }
  ///
  void add_vertex (int v);
  ///
  char* get_name () { return name; }

  /// Compute the plane normal of this polygon.
  void PlaneNormal (float* A, float* B, float* C);

  ///
  void set_texture (TextureMM* texture) { PolygonTemplate::texture = texture; }
  /// Set the transformation from object to texture space.
  void set_texture_space (Matrix3& tx_matrix, Vector3& tx_vector);
  /// Transform the texture tranformation using the given matrix/vector.
  void transform (Matrix3& m, Vector3& v);
  ///
  TextureMM* get_texture () { return texture; }
  ///
  Matrix3& get_texture_matrix () { return m_obj2tex; }
  ///
  Vector3& get_texture_vector () { return v_obj2tex; }

  ///
  void set_no_lighting (int no) { no_lighting = no; }
  ///
  void set_no_mipmap (int no) { no_mipmap = no; }
  ///
  int get_no_lighting () { return no_lighting; }
  ///
  int get_no_mipmap () { return no_mipmap; }
};

#endif /*THINGTPL_H*/
