/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/system.h"
#include "system/sysprint.h"
#include "engine/polygon/polygon.h"
#include "engine/polygon/polytext.h"
#include "engine/texture.h"
#include "engine/polygon/polyplan.h"
#include "engine/sector.h"
#include "engine/world.h"
#include "engine/light/light.h"
#include "engine/light/dynlight.h"
#include "engine/light/lghtmap.h"
#include "engine/camera.h"
#include "render/igraph3d.h"
#include "engine/polygon/portal.h"
#include "general/dumper.h"

// Option variable: force lightmap recalculation?
bool Polygon3D::do_force_recalc = false;
// Option variables: mipmap distances
float Polygon3D::zdist_mipmap1 = ZDIST_MIPMAP1;
float Polygon3D::zdist_mipmap2 = ZDIST_MIPMAP2;
float Polygon3D::zdist_mipmap3 = ZDIST_MIPMAP3;
// Option variable: shadow mipmap size
int Polygon3D::def_mipmap_size = 16;

//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( Polygon3D )

BEGIN_INTERFACE_TABLE( Polygon3D )
	IMPLEMENTS_COMPOSITE_INTERFACE( Polygon3D )
END_INTERFACE_TABLE()


Polygon3D::Polygon3D (char* name, TextureMM* texture) : PolygonInt (name, CS_POLYGON), theDynLight (0) // SJI
{
  vertices_idx = NULL;
  max_vertices = num_vertices = 0;

  CHK (tex = new PolyTexture ());
  tex->set_polygon (this);
  CHK (tex1 = new PolyTexture ());
  tex1->set_polygon (this);
  CHK (tex2 = new PolyTexture ());
  tex2->set_polygon (this);
  CHK (tex3 = new PolyTexture ());
  tex3->set_polygon (this);

  txtMM = texture;
  if (texture) set_texture (texture);
  lightmap = lightmap1 = lightmap2 = lightmap3 = NULL;

  delete_tex_info = true;

  plane = NULL;
  delete_plane = false;

  portal = NULL;
  delete_portal = false;

  orig_poly = NULL;
  dont_draw = false;
  serial = 0;

  no_mipmap = false;
  no_lighting = false;
  cosinus_factor = -1;
  lightmap_up_to_date = false;

  lightpatches = NULL;
  uv_coords = NULL;
  colors = NULL;
}

Polygon3D::Polygon3D (Polygon3D& poly) : PolygonInt (poly.get_name (), CS_POLYGON)
{
  theDynLight = poly.theDynLight;
  vertices_idx = NULL;
  max_vertices = num_vertices = 0;

  poly_set = poly.poly_set;
  sector = poly.sector;

  portal = poly.portal;
  delete_portal = false;	// This polygon is no owner

  plane = poly.plane;
  delete_plane = false;		// This polygon is no owner

  txtMM = poly.txtMM;

  tex = poly.tex;
  tex1 = poly.tex1;
  tex2 = poly.tex2;
  tex3 = poly.tex3;

  lightmap = lightmap1 = lightmap2 = lightmap3 = NULL;

  delete_tex_info = false;
  poly.dont_draw = true;
  dont_draw = false;
  orig_poly = poly.orig_poly ? poly.orig_poly : &poly;
  serial = orig_poly->serial;
  orig_poly->serial++;

  no_mipmap = poly.no_mipmap;
  no_lighting = poly.no_lighting;
  cosinus_factor = poly.cosinus_factor;
  lightmap_up_to_date = false;

  lightpatches = NULL;
  uv_coords = NULL;
  colors = NULL;
}

Polygon3D::~Polygon3D ()
{
  CHK (delete [] uv_coords);
  CHK (delete [] colors);
  CHK (delete [] vertices_idx);
  if (delete_tex_info)
  {
    CHK (delete tex);
    CHK (delete tex1);
    CHK (delete tex2);
    CHK (delete tex3);
    CHK (delete lightmap);
    CHK (delete lightmap1);
    CHK (delete lightmap2);
    CHK (delete lightmap3);
  }
  if (plane && delete_plane) CHKB (delete plane);
  if (portal && delete_portal) CHKB (delete portal);
  while (lightpatches) CHKB (delete lightpatches);
}

void Polygon3D::reset ()
{
  num_vertices = max_vertices = 0;
  CHK (delete [] vertices_idx);
  vertices_idx = NULL;
  reset_uv ();
}

void Polygon3D::set_uv (int i, float u, float v)
{
  if (!uv_coords) CHKB (uv_coords = new Vector2 [num_vertices]);
  uv_coords[i].x = u;
  uv_coords[i].y = v;
}

void Polygon3D::reset_uv ()
{
  CHK (delete [] uv_coords);
  uv_coords = NULL;
  CHK (delete [] colors);
  colors = NULL;
}

void Polygon3D::set_color (int i, float r, float g, float b)
{
  if (!colors)
  {
    CHK (colors = new csColor [num_vertices]);
    int j;
    for (j = 0 ; j < num_vertices ; j++)
      colors[j].set (0, 0, 0);
  }
  colors[i].set (r, g, b);
}

void Polygon3D::set_cs_portal (csSector* sector)
{
  if (portal && portal->portal_type () == PORTAL_CS)
  {
    if (((PortalCS*)portal)->get_sector () == sector) return;
  }
  if (portal && delete_portal) { CHK (delete portal); portal = NULL; }
  if (!sector) return;
  PortalCS* csp;
  CHK (csp = new PortalCS);
  portal = (Portal*)csp;
  delete_portal = true;
  portal->disable_space_warping ();
  csp->set_sector (sector);
  portal->set_texture (txtMM);
}

void Polygon3D::set_portal (Portal* prt)
{
  if (portal && delete_portal) { CHK (delete portal); portal = NULL; }
  portal = prt;
  delete_portal = true;
}

void Polygon3D::split_with_plane (Vector3* front, int& front_n,
				  Vector3* back, int& back_n, Plane& plane)
{
  Vector3 ptA, ptB;
  float sideA, sideB;
  ptA = vwor (num_vertices - 1);
  sideA = plane.classify (ptA);

  front_n = back_n = 0;

  for (int i = -1 ; ++i < num_vertices ; )
  {
    ptB = vwor (i);
    sideB = plane.classify (ptB);
    if (sideB > 0)
    {
      if (sideA < 0)
      {
	// Compute the intersection point of the line
	// from point A to point B with the partition
	// plane. This is a simple ray-plane intersection.
	Vector3 v = ptB; v -= ptA;
	float sect = - plane.classify (ptA) / ( plane.normal () * v ) ;
	v *= sect; v += ptA;
	*back++ = *front++ = v;
	back_n++; front_n++;
      }
      *back++ = ptB;
      back_n++;
    }
    else if (sideB < 0)
    {
      if (sideA > 0)
      {
	// Compute the intersection point of the line
	// from point A to point B with the partition
	// plane. This is a simple ray-plane intersection.
	Vector3 v = ptB; v -= ptA;
	float sect = - plane.classify (ptA) / ( plane.normal () * v );
	v *= sect; v += ptA;
	*back++ = *front++ = v;
	back_n++; front_n++;
      }
      *front++ = ptB;
      front_n++;
    }
    else
    {
      *back++ = *front++ = ptB;
      back_n++; front_n++;
    }
    ptA = ptB;
    sideA = sideB;
  }
}

void Polygon3D::set_texture (TextureMM* texture)
{
  txtMM = texture;
  tex->set_texture (texture->get_texture (0));
  tex1->set_texture (texture->get_texture (1));
  tex2->set_texture (texture->get_texture (2));
  tex3->set_texture (texture->get_texture (3));
}

PolyTexture* Polygon3D::get_polytex (int mipmap)
{
  switch (mipmap)
  {
    case 0: return tex;
    case 1: return tex1;
    case 2: return tex2;
    case 3: return tex3;
  }
  return tex;
}

PolyTexture* Polygon3D::get_polytex (float z_dist)
{
  if (z_dist < zdist_mipmap1) return tex;
  if (z_dist < zdist_mipmap2) return tex1;
  if (z_dist < zdist_mipmap3) return tex2;
  return tex3;
}

bool Polygon3D::same_plane (PolygonInt* p)
{
  if (((Polygon3D*)p)->plane == plane) return true;
  return ((Polygon3D*)p)->plane->close (plane);
}

int Polygon3D::classify (PolygonInt* spoly)
{
  if (same_plane (spoly)) return POL_SAME_PLANE;

  int i;
  int front = 0, back = 0;
  Polygon3D* poly = (Polygon3D*)spoly;
  PolyPlane* pl = poly->get_plane ();

  for (i = 0 ; i < num_vertices ; i++)
  {
    float dot = pl->classify (vwor (i));
    if (ABS (dot) < SMALL_EPSILON) dot = 0;
    if (dot > 0) back++;
    else if (dot < 0) front++;
  }
  if (back == 0) return POL_FRONT;
  if (front == 0) return POL_BACK;
  return POL_SPLIT_NEEDED;
}

void Polygon3D::compute_normal ()
{
  float A, B, C, D;
  PlaneNormal (&A, &B, &C);
  D = -A*vobj (0).x - B*vobj (0).y - C*vobj (0).z;

  // By default the world space normal is equal to the object space normal.
  plane->get_object_plane ().set (A, B, C, D);
  plane->get_world_plane ().set (A, B, C, D);
}

void Polygon3D::object_to_world (const ReversibleTransform& t)
{
  plane->object_to_world (t, vwor (0));
  if (portal) portal->object_to_world (t);
}

#define TEXW(t) ((t)->w_orig)
#define TEXH(t) ((t)->h)

void Polygon3D::precalculate ()
{
  if (orig_poly) return;
  if (uv_coords) return;

  tex->create_bounding_texture_box (); tex->lm = NULL;
  tex1->create_bounding_texture_box (); tex1->lm = NULL;
  tex2->create_bounding_texture_box (); tex2->lm = NULL;
  tex3->create_bounding_texture_box (); tex3->lm = NULL;

  lightmap = lightmap1 = lightmap2 = lightmap3 = NULL;
  if (!no_lighting && TEXW(tex)*TEXH(tex) < 1000000)
  {
    CHK (lightmap = new LightMap ());
    lightmap->alloc (TEXW(tex), TEXH(tex), def_mipmap_size, this);
    tex->set_mipmap_size (def_mipmap_size); tex->lm = lightmap;
  }
}

#define USE_SUBTEX_OPT 1

void Polygon3D::mipmap_settings (int setting)
{
  if (!lightmap) return;

  if (lightmap1) { CHK (delete lightmap1); lightmap1 = NULL; }
  if (lightmap2) { CHK (delete lightmap2); lightmap2 = NULL; }
  if (lightmap3) { CHK (delete lightmap3); lightmap3 = NULL; }

  if (orig_poly) return;

  switch (setting)
  {
    case MIPMAP_SHADOW_ACCURATE:
      if (Textures::mipmap_nice == MIPMAP_VERYNICE)
        tex1->set_mipmap_size (def_mipmap_size);
      else
        tex1->set_mipmap_size (def_mipmap_size>>1);
#     if !USE_SUBTEX_OPT
      tex1->lm = lightmap;
#     else
      CHK (lightmap1 = new LightMap ());
      tex1->lm = lightmap1;
      lightmap1->copy_lightmap (lightmap);
#     endif

      tex2->set_mipmap_size (tex1->mipmap_size>>1);

      // @@@ Normally the two mipmap levels can share the same lightmap.
      // But with the new sub-texture optimization in combination with the dynamic
      // lighting extension this gives some problems. For the moment we just copy
      // the lightmaps even though they are actually the same.
#     if !USE_SUBTEX_OPT
      tex2->lm = lightmap;
#     else
      CHK (lightmap2 = new LightMap ());
      tex2->lm = lightmap2;
      lightmap2->copy_lightmap (lightmap);
#     endif

      if (tex2->mipmap_size < 1)
      {
        tex2->set_mipmap_size (1);
#       if !USE_SUBTEX_OPT
	CHK (delete lightmap2);
#       endif
	CHK (lightmap2 = new LightMap ());
	lightmap2->mipmap_lightmap (TEXW(tex2), TEXH(tex2), 1, lightmap, TEXW(tex1), TEXH(tex1), tex1->mipmap_size);
	tex2->lm = lightmap2;

        tex3->set_mipmap_size (1);
	CHK (lightmap3 = new LightMap ());
	lightmap3->mipmap_lightmap (TEXW(tex3), TEXH(tex3), 1, lightmap2, TEXW(tex2), TEXH(tex2), tex2->mipmap_size);
	tex3->lm = lightmap3;
      }
      else
      {
        tex3->set_mipmap_size (tex2->mipmap_size>>1);
#       if !USE_SUBTEX_OPT
        tex3->lm = lightmap;
#       else
	CHK (lightmap2 = new LightMap ());
	tex3->lm = lightmap2;
	lightmap2->copy_lightmap (lightmap);
#       endif

        if (tex3->mipmap_size < 1)
        {
          tex3->set_mipmap_size (1);
#	  if !USE_SUBTEX_OPT
	  CHK (delete lightmap3);
#	  endif
	  CHK (lightmap3 = new LightMap ());
	  lightmap3->mipmap_lightmap (TEXW(tex3), TEXH(tex3), 1, lightmap, TEXW(tex2), TEXH(tex2), tex2->mipmap_size);
	  tex3->lm = lightmap3;
        }
      }
      break;
    case MIPMAP_SHADOW_INACCURATE:
      if (Textures::mipmap_nice == MIPMAP_VERYNICE)
      {
#	if !USE_SUBTEX_OPT
        tex1->lm = lightmap;
#	else
	CHK (lightmap1 = new LightMap ());
	tex1->lm = lightmap1;
	lightmap1->copy_lightmap (lightmap);
#	endif
      }
      else
      {
        CHK (lightmap1 = new LightMap ());
        lightmap1->mipmap_lightmap (TEXW(tex1), TEXH(tex1), def_mipmap_size, lightmap, TEXW(tex), TEXH(tex), def_mipmap_size);
	tex1->lm = lightmap1;
      }

      CHK (lightmap2 = new LightMap ());
      lightmap2->mipmap_lightmap (TEXW(tex2), TEXH(tex2), def_mipmap_size, tex1->lm, TEXW(tex1), TEXH(tex1), def_mipmap_size);
      tex2->set_mipmap_size (def_mipmap_size); tex2->lm = lightmap2;

      CHK (lightmap3 = new LightMap ());
      lightmap3->mipmap_lightmap (TEXW(tex3), TEXH(tex3), def_mipmap_size, lightmap2, TEXW(tex2), TEXH(tex2), def_mipmap_size);
      tex3->set_mipmap_size (def_mipmap_size); tex3->lm = lightmap3;

      break;
    case MIPMAP_SHADOW_REASONABLE:
      if (Textures::mipmap_nice == MIPMAP_VERYNICE)
        tex1->set_mipmap_size (def_mipmap_size);
      else
        tex1->set_mipmap_size (def_mipmap_size>>1);
#     if !USE_SUBTEX_OPT
      tex1->lm = lightmap;
#     else
      CHK (lightmap1 = new LightMap ());
      tex1->lm = lightmap1;
      lightmap1->copy_lightmap (lightmap);
#     endif

      CHK (lightmap2 = new LightMap ());
      lightmap2->mipmap_lightmap (TEXW(tex1), TEXH(tex1), def_mipmap_size, lightmap, TEXW(tex), TEXH(tex), def_mipmap_size);
      tex2->set_mipmap_size (def_mipmap_size>>1); tex2->lm = lightmap2;

#     if !USE_SUBTEX_OPT
      tex3->set_mipmap_size (def_mipmap_size>>2); tex3->lm = lightmap2;
#     else
      CHK (lightmap3 = new LightMap ());
      tex3->lm = lightmap3;
      lightmap3->copy_lightmap (lightmap2);
#     endif
      break;
  }

  int aspect;
  bool po2 = (Gfx3D->NeedsPO2Maps() == S_OK);

  Gfx3D->GetMaximumAspectRatio(aspect);

  if (lightmap) lightmap->convert_for_3d_driver (po2, aspect);
  if (lightmap1) lightmap1->convert_for_3d_driver (po2, aspect);
  if (lightmap2) lightmap2->convert_for_3d_driver (po2, aspect);
  if (lightmap3) lightmap3->convert_for_3d_driver (po2, aspect);
}

void Polygon3D::set_texture_space (PolyPlane* pl)
{
  if (plane && delete_plane) CHKB (delete plane);
  plane = pl;
  delete_plane = false;

  compute_normal ();

  precalculate ();
}

void Polygon3D::set_texture_space (Polygon3D* copy_from)
{
  if (plane && delete_plane) CHKB (delete plane);
  plane = copy_from->plane;
  delete_plane = false;

  compute_normal ();

  precalculate ();
}

void Polygon3D::set_texture_space (Matrix3& tx_matrix, Vector3& tx_vector)
{
  if (plane && delete_plane) CHKB (delete plane);
  CHK (plane = new PolyPlane ("-"));
  delete_plane = true;

  compute_normal ();
  plane->set_texture_space (tx_matrix, tx_vector);

  precalculate ();
}

void Polygon3D::set_texture_space (Vector3& v_orig, Vector3& v1, float len1)
{
  float xo = v_orig.x;
  float yo = v_orig.y;
  float zo = v_orig.z;
  float x1 = v1.x;
  float y1 = v1.y;
  float z1 = v1.z;
  set_texture_space (xo, yo, zo, x1, y1, z1, len1);
}

void Polygon3D::set_texture_space (
	float xo, float yo, float zo,
	float x1, float y1, float z1,
	float len1)
{
  if (plane && delete_plane) CHKB (delete plane);
  CHK (plane = new PolyPlane ("-"));
  delete_plane = true;

  compute_normal ();
  plane->set_texture_space (xo, yo, zo, x1, y1, z1, len1);

  precalculate ();
}

void Polygon3D::set_texture_space (
	float xo, float yo, float zo,
	float x1, float y1, float z1,
	float len1,
	float x2, float y2, float z2,
	float len2)
{
  if (plane && delete_plane) CHKB (delete plane);
  CHK (plane = new PolyPlane ("-"));
  delete_plane = true;

  compute_normal ();
  plane->set_texture_space (xo, yo, zo, x1, y1, z1, len1, x2, y2, z2, len2);

  precalculate ();
}


void Polygon3D::make_dirty_dynamic_lights ()
{
  if (tex) tex->make_dirty_dynamic_lights ();
  if (tex1) tex1->make_dirty_dynamic_lights ();
  if (tex2) tex2->make_dirty_dynamic_lights ();
  if (tex3) tex3->make_dirty_dynamic_lights ();
}

void Polygon3D::unlink_lightpatch (LightPatch* lp)
{
  if (lp->next_poly) lp->next_poly->prev_poly = lp->prev_poly;
  if (lp->prev_poly) lp->prev_poly->next_poly = lp->next_poly;
  else lightpatches = lp->next_poly;
  lp->prev_poly = lp->next_poly = NULL;
  lp->polygon = NULL;
  make_dirty_dynamic_lights ();
}

void Polygon3D::add_lightpatch (LightPatch* lp)
{
  lp->next_poly = lightpatches;
  lp->prev_poly = NULL;
  if (lightpatches) lightpatches->prev_poly = lp;
  lightpatches = lp;
  lp->polygon = this;
  make_dirty_dynamic_lights ();
}

int Polygon3D::add_vertex (int v)
{
  if (!vertices_idx)
  {
    max_vertices = 4;
    CHK (vertices_idx = new int [max_vertices]);
  }
  while (num_vertices >= max_vertices)
  {
    max_vertices += 2;
    CHK (int* new_vertices_idx = new int [max_vertices]);
    memcpy (new_vertices_idx, vertices_idx, sizeof (int)*num_vertices);
    CHK (delete [] vertices_idx);
    vertices_idx = new_vertices_idx;
  }

  vertices_idx[num_vertices++] = v;
  return num_vertices-1;
}

int Polygon3D::add_vertex (Vector3& v)
{
  int i = poly_set->add_vertex_smart (v);
  add_vertex (i);
  return i;
}

int Polygon3D::add_vertex (float x, float y, float z)
{
  int i = poly_set->add_vertex_smart (x, y, z);
  add_vertex (i);
  return i;
}


enum InFlag { IN_P, IN_Q, IN_UNKNOWN };

void Polygon3D::PlaneNormal (float* yz, float* zx, float* xy)
{
  float ayz = 0;
  float azx = 0;
  float axy = 0;
  int i, i1;
  float x1, y1, z1, x, y, z;

  i1 = num_vertices-1;
  for (i = 0 ; i < num_vertices ; i++)
  {
    x = vwor (i).x;
    y = vwor (i).y;
    z = vwor (i).z;
    x1 = vwor (i1).x;
    y1 = vwor (i1).y;
    z1 = vwor (i1).z;
    ayz += (z1+z) * (y-y1);
    azx += (x1+x) * (z-z1);
    axy += (y1+y) * (x-x1);
    i1 = i;
  }

  float d = sqrt (ayz*ayz + azx*azx + axy*axy);

  if (d < SMALL_EPSILON) d = SMALL_EPSILON;

  *yz = ayz / d;
  *zx = azx / d;
  *xy = axy / d;
}


bool Polygon3D::clip_poly (Vector3* frustrum, int m, bool mirror, Vector3** dest, int* num_dest)
{
  int i, i1;
  if (!frustrum)
  {
    // Infinite frustrum.
    CHK (Vector3* dd = new Vector3 [num_vertices]);
    *dest = dd;
        if(!mirror)
      for (i = 0 ; i < num_vertices ; i++)
        dd[i] = vcam (i);
        else
      for (i = 0 ; i < num_vertices ; i++)
        dd[i] = vcam (num_vertices-i-1);
    *num_dest = num_vertices;
    return true;
  }

  CHK (Vector3* dd = new Vector3 [num_vertices+m]);     // For safety
  *dest = dd;
  if(!mirror)
    for (i = 0 ; i <num_vertices ; i++)
      dd[i] = vcam (i);
  else
    for (i = 0 ; i <num_vertices ; i++)
      dd[i] = vcam (num_vertices-i-1);
  *num_dest = num_vertices;

  i1 = -1;
  Vector3 *p = frustrum + m;

  for (i = -m-1 ; ++i ; )
  {
    clip_poly_plane (dd, num_dest, mirror, p[i1], p[i]);
    if ((*num_dest) < 3)
    {
      CHK (delete [] dd);
      *dest = NULL;
      return false;
    }
    i1 = i;
  }

  return true;
}

// Clip a polygon against a plane.
void Polygon3D::clip_poly_plane (Vector3* verts, int* num, bool mirror,
  Vector3& v1, Vector3& v2)
{
  int cw_offset = -1;
  int ccw_offset;
  bool first_vertex_side;
  Vector3 isect_cw,isect_ccw;
  Vector3 Plane_Normal;
  int i;

  //  do the check only once at the beginning instead of twice during the routine.
  if (mirror)
    Plane_Normal = v2%v1;
  else
    Plane_Normal = v1%v2;

  //  On which side is the first vertex?
  first_vertex_side = (Plane_Normal*(verts[(*num) -1] -v1) > 0);

  for (i = 0; i < (*num) - 1; i++)
  {
    if ( (Plane_Normal*(verts[i] - v1) > 0) != first_vertex_side)
      {cw_offset = i;break;}
  }

  if (cw_offset == -1)
  {
    //  Return , if there is no intersection.
    if (first_vertex_side)
      {*num = 0;}      // The whole polygon is behind the plane because the first is.
    return;
  }

  for (ccw_offset = (*num) -2; ccw_offset >= 0; ccw_offset--)
  {
    if ((Plane_Normal*(verts[ccw_offset] - v1) > 0) != first_vertex_side)
      {break;}
  }

  // calculate the intersection points.
  i = cw_offset - 1;
  if (i < 0) {i = (*num) -1;}
  Intersect3::plane (verts[cw_offset], verts[i], Plane_Normal, v1,isect_cw);
  Intersect3::plane (verts[ccw_offset], verts[ccw_offset + 1],Plane_Normal, v1, isect_ccw);

  // remove the obsolete point and insert the intersection points.
  if (first_vertex_side)
  {
    for (i = 0; i < ccw_offset - cw_offset + 1; i++)
      verts[i] = verts[i + cw_offset];
    verts[i] = isect_ccw;
    verts[i+1] = isect_cw;
    *num = 3 + ccw_offset - cw_offset;
  }
  else
  {
    if (cw_offset + 1 < ccw_offset)
      for (i = 0; i < (*num) - ccw_offset - 1; i++)
        verts[cw_offset + 2 + i] = verts[ccw_offset + 1 + i];
    else if (cw_offset + 1 > ccw_offset)
      for (i = (*num) - 2 - ccw_offset;i >= 0; i--)
        verts[cw_offset + 2 + i] = verts[ccw_offset + 1 + i];

    verts[cw_offset] = isect_cw;
    verts[cw_offset+1] = isect_ccw;
    *num = 2 + cw_offset + (*num) - ccw_offset - 1;
  }
}

bool Polygon3D::clip_frustrum (Vector3& center, Vector3* frustrum, int num_frustrum, bool mirror,
	Vector3** new_frustrum, int* new_num_frustrum)
{
  if (!plane->visible_from_point (center)) return false;
  return clip_poly (frustrum, num_frustrum, mirror, new_frustrum, new_num_frustrum);
}

bool Polygon3D::clip_to_plane (Plane* portal_plane, const Vector3& v_w2c,
	Vector3*& pverts, int& num_verts, bool cw)
{
  int i, i1, cnt_vis;
  float r;
  bool zs, z1s;

  // Assume maximum 100 vertices! (@@@ HARDCODED LIMIT)
  static Vector3 verts[100];
  bool vis[100];

  // Count the number of visible vertices for this polygon (note
  // that the transformation from world to camera space for all the
  // vertices has been done earlier).
  // If there are no visible vertices this polygon need not be drawn.
  cnt_vis = 0;
  for (i = 0 ; i < num_vertices ; i++)
    if (vcam (i).z >= SMALL_Z) cnt_vis++;
  if (cnt_vis == 0) return false;

  // Perform backface culling.
  // Note! The plane normal needs to be correctly calculated for this
  // to work!
  if (plane->visible_from_point (v_w2c) != cw) return false;

  // Copy the vertices to verts.
  for (i = 0 ; i < num_vertices ; i++) verts[i] = vcam (i);
  pverts = verts;

  // If there is no portal polygon then everything is ok.
  if (!portal_plane) { num_verts = num_vertices; return true; }

  // Otherwise we will have to clip this polygon in 3D against the
  // portal polygon. This is to make sure that objects behind the
  // portal polygon are not accidently rendered.

  // First count how many vertices are before the portal polygon
  // (so are visible as seen from the portal).
  cnt_vis = 0;
  for (i = 0 ; i < num_vertices ; i++)
  {
    vis[i] = portal_plane->visible_from_point (vcam (i));
    if (vis[i]) cnt_vis++;
  }

  if (cnt_vis == 0) return false; // Polygon is not visible.

  // If all vertices are visible then everything is ok.
  if (cnt_vis == num_vertices) { num_verts = num_vertices; return true; }

  // We really need to clip.
  num_verts = 0;

  float A = portal_plane->A ();
  float B = portal_plane->B ();
  float C = portal_plane->C ();
  float D = portal_plane->D ();

  i1 = num_vertices-1;
  for (i = 0 ; i < num_vertices ; i++)
  {
    zs = !vis[i];
    z1s = !vis[i1];

    if (z1s && !zs)
    {
      Intersect3::plane (vcam (i1), vcam (i), A, B, C, D, verts[num_verts], r);
      num_verts++;
      verts[num_verts++] = vcam (i);
    }
    else if (!z1s && zs)
    {
      Intersect3::plane (vcam (i1), vcam (i), A, B, C, D, verts[num_verts], r);
      num_verts++;
    }
    else if (!z1s && !zs)
    {
       verts[num_verts++] = vcam (i);
    }
    i1 = i;
  }

  return true;
}

bool Polygon3D::do_perspective (const Transform& trans,
	Vector3* source, int num_verts, Polygon2D* dest, Vector2* orig_triangle)
{
  Vector3 *ind, *end = source+num_verts;

  if (num_verts==0) return false;
  dest->make_empty();

  // Classify all points as NORMAL (z>=SMALL_Z), NEAR (0<=z<SMALL_Z), or
  // BEHIND (z<0).  Use several processing algorithms: trivially accept if all
  // points are NORMAL, mixed process if some points are NORMAL and some
  // are not, special process if there are no NORMAL points, but some are
  // NEAR.  Assume that the polygon has already been culled if all points
  // are BEHIND.

  // Handle the trivial acceptance case:
  ind = source;
  while (ind < end)
  {
    if (ind->z >= SMALL_Z) dest->add_perspective (*ind);
    else break;
    ind++;
  }

  // Check if special or mixed processing is required
  if (ind != end)
  {
    // If we are processing a triangle (uv_coords != NULL) then
    // we stop here because the triangle is only visible if all
    // vertices are visible (this is not exactly true but it is
    // easier this way! @@@ CHANGE IN FUTURE).
    if (uv_coords) return false;

    Vector3 *exit = NULL, *exitn = NULL, *reenter = NULL, *reentern = NULL;
    Vector2 *evert = NULL, *rvert = NULL;

    if (ind == source)
    {
      while (ind < end)
      {
        if (ind->z >= SMALL_Z) { reentern = ind;  reenter = ind-1;  break; }
        ind++;
      }
    }
    else
    {
      exit = ind;
      exitn = ind-1;
      evert = dest->get_last();
    }

    // Check if mixed processing is required
    if (exit || reenter)
    {
     bool needfinish = false;

     if (exit)
     {
      // we know where the polygon is no longer NORMAL, now we need to
      // to find out on which edge it becomes NORMAL again.
      while (ind < end)
      {
       if (ind->z >= SMALL_Z) { reentern = ind;  reenter = ind-1;  break; }
       ind++;
      }
      if (ind == end) { reentern = source;  reenter = ind-1; }
       else needfinish = true;
     } /* if (exit) */
     else
     {
      // we know where the polygon becomes NORMAL, now we need to
      // to find out on which edge it ceases to be NORMAL.
      while (ind < end)
      {
       if (ind->z >= SMALL_Z) dest->add_perspective(*ind);
       else { exit = ind;  exitn = ind-1;  break; }
       ind++;
      }
      if (ind == end) { exit = source;  exitn = ind-1; }
      evert = dest->get_last();
     }

     // Add the NEAR points appropriately.
#    define MAX_VALUE 1000000.

     // First, for the exit point.
     float ex, ey, epointx, epointy;
     int esec;
     ex = exitn->z * exit->x - exitn->x * exit->z;
     ey = exitn->z * exit->y - exitn->y * exit->z;
     if (ABS(ex) < SMALL_EPSILON && ABS(ey) < SMALL_EPSILON)
     {
      // Uncommon special case:  polygon passes through origin.
      plane->world_to_camera (trans, source[0]);
      ex = plane->get_camera_plane().A();
      ey = plane->get_camera_plane().B();
      if (ABS(ex) < SMALL_EPSILON && ABS(ey) < SMALL_EPSILON)
      {
       // Downright rare case:  polygon near parallel with viewscreen.
       ex = exit->x - exitn->x;
       ey = exit->y - exitn->y;
      }
     }
     if (ABS(ex) > ABS(ey))
     {
       if (ex>0) { epointx = MAX_VALUE; esec = 3; }
       else { epointx = -MAX_VALUE; esec = 1; }
       epointy = (epointx - evert->x)*ey/ex + evert->y;
     }
     else
     {
       if (ey>0) { epointy = MAX_VALUE; esec = 0; }
       else { epointy = -MAX_VALUE; esec = 2; }
       epointx = (epointy - evert->y)*ex/ey + evert->x;
     }

     // Next, for the reentry point.
     float rx, ry, rpointx, rpointy;
     int rsec;

     // Obligatory ugly hack :)
     // Can be fixed if someone writes a function something like this:
     // Vector2 get_perspective(const Vector3&)
     Polygon2D p(1);
     p.add_perspective(*reentern);
     rvert = p.get_first();

     rx = reentern->z * reenter->x - reentern->x * reenter->z;
     ry = reentern->z * reenter->y - reentern->y * reenter->z;
     if (ABS(rx) < SMALL_EPSILON && ABS(ry) < SMALL_EPSILON)
     {
      // Uncommon special case:  polygon passes through origin.
      plane->world_to_camera (trans, source[0]);
      rx = plane->get_camera_plane().A();
      ry = plane->get_camera_plane().B();
      if (ABS(rx) < SMALL_EPSILON && ABS(ry) < SMALL_EPSILON)
      {
       // Downright rare case:  polygon near parallel with viewscreen.
       rx = reenter->x - reentern->x;
       ry = reenter->y - reentern->y;
      }
     }
     if (ABS(rx) > ABS(ry))
     {
       if (rx>0) { rpointx = MAX_VALUE; rsec = 3; }
       else { rpointx = -MAX_VALUE; rsec = 1; }
       rpointy = (rpointx - rvert->x)*ry/rx + rvert->y;
     }
     else
     {
       if (ry>0) { rpointy = MAX_VALUE; rsec = 0; }
       else { rpointy = -MAX_VALUE; rsec = 2; }
       rpointx = (rpointy - rvert->y)*rx/ry + rvert->x;
     }

    dest->add_vertex(epointx,epointy);
    dest->add_vertex(rpointx,rpointy);

     // Add the rest of the vertices, which are all NORMAL points.
     if (needfinish) while (ind < end)
      dest->add_perspective(*ind++);

    } /* if (exit || reenter) */

    // Do special processing (all points are NEAR or BEHIND)
    else
    {
     // @@@
     return false;
    }

  } /* if (ind != end) */

  if (uv_coords)
  {
    for (int i = 0 ; i < 3 ; i++)
      orig_triangle[i] = *dest->get_vertex (i);
  }

  plane->world_to_camera (trans, source[0]);
  return true;
}

bool Polygon3D::intersect_ray (const Vector3& start, const Vector3& end)
{
  // First we do backface culling on the polygon with respect to
  // the starting point of the beam.
  Plane& pl = plane->get_world_plane ();
  float dot1 = pl.D () + pl.A ()*start.x + pl.B ()*start.y + pl.C ()*start.z;
  if (dot1 > 0) return false;

  // If this vector is perpendicular to the plane of the polygon we
  // need to catch this case here.
  float dot2 = pl.D () + pl.A ()*end.x + pl.B ()*end.y + pl.C ()*end.z;
  if (ABS (dot1-dot2) < SMALL_EPSILON) return false;

  // Now we generate a plane between the starting point of the ray and
  // every edge of the polygon. With the plane normal of that plane we
  // can then check if the end of the ray is on the same side for all
  // these planes.
  Vector3 normal, relend;
  relend = end;
  relend -= start;

  int i, i1;
  i1 = num_vertices-1;
  for (i = 0 ; i < num_vertices ; i++)
  {
    normal = Plane::calc_plane_normal (start, vwor (i1), vwor (i));
    if ( (relend * normal) > 0) return false;
    i1 = i;
  }

  return true;
}

bool Polygon3D::intersect_segment (const Vector3& start, const Vector3& end,
                                   Vector3& isect, float* pr)
{
  if (!intersect_ray (start, end)) return false;
  return plane->intersect_segment (start, end, isect, pr);
}

void Polygon3D::init_lightmaps (bool do_cache)
{
  if (orig_poly) return;
  if (!tex->lm) return;
  if (!do_cache) { lightmap_up_to_date = false; return; }
  if (do_force_recalc) { tex->init_lightmaps (); lightmap_up_to_date = false; }
  else if (!tex->lm->read_from_cache (TEXW(tex), TEXH(tex), def_mipmap_size, this, System->world))
  {
    tex->init_lightmaps ();
    lightmap_up_to_date = true;
  }
  else lightmap_up_to_date = true;
}

void Polygon3D::shine_lightmaps (LightView& lview)
{
  if (orig_poly) return;

  if (uv_coords)
  {
    // We are working for a vertex lighted polygon.
    csLight* light = (csLight*)lview.l;
    float rd = lview.r / light->get_dist ();
    float gd = lview.g / light->get_dist ();
    float bd = lview.b / light->get_dist ();

    if (lview.dynamic)
    {
      // Currently not yet supported. @@@
      return;
    }
    else
    {
      csColor col;
      int i;
      float cosfact = cosinus_factor;
      if (cosfact == -1) cosfact = PolyTexture::cfg_cosinus_factor;

      for (i = 0 ; i < num_vertices ; i++)
      {
        if (colors) col = colors[i];
	else col.set (0, 0, 0);
        float d = Vector3::sqdist (light->get_center (), vwor (i));
	if (d >= light->get_sq_dist ()) continue;
	d = sqrt (d);
	float cosinus = (vwor (i)-light->get_center ())*get_poly_plane ()->normal ();
	cosinus /= d;
	cosinus += cosfact;
	if (cosinus < 0) cosinus = 0;
	else if (cosinus > 1) cosinus = 1;
	col.red = col.red + cosinus * rd*(light->get_dist () - d);
	if (col.red > 1) col.red = 1;
	col.green = col.green + cosinus * gd*(light->get_dist () - d);
	if (col.green > 1) col.green = 1;
	col.blue = col.blue + cosinus * bd*(light->get_dist () - d);
	if (col.blue > 1) col.blue = 1;
        set_color (i, col);
      }
    }
    return;
  }

  if (lview.dynamic)
  {
    // We are working for a dynamic light. In this case we create
    // a light patch for this polygon.
    CHK (LightPatch* lp = new LightPatch ());
    add_lightpatch (lp);
    csDynLight* dl = (csDynLight*)lview.l;
    dl->add_lightpatch (lp);

    lp->num_vertices = lview.num_frustrum;
    CHK (lp->vertices = new Vector3 [lview.num_frustrum]);

    int i, mi;
    for (i = 0 ; i < lview.num_frustrum ; i++)
    {
      if (lview.mirror) mi = lview.num_frustrum-i-1;
      else mi = i;
      //lp->vertices[i] = lview.frustrum[mi] + lview.center;
      lp->vertices[i] = lview.frustrum[mi];
    }

  }
  else
  {
    if (lightmap_up_to_date) return;
    tex->shine_lightmaps (lview);
    // If there is already a lightmap1 this means that we are
    // redoing lighting at run-time. In that case we also
    // need to update the other mipmap levels.
    // @@@ We need specific functionality for this.
    // Doing this with mipmap_settings is NOT efficient.
    if (lightmap1) mipmap_settings (LightMap::setting);
  }
}

void Polygon3D::cache_lightmaps ()
{
  if (orig_poly) return;
  if (!tex->lm) return;
  if (!lightmap_up_to_date)
  {
    lightmap_up_to_date = true;
    tex->lm->cache (this, System->world);
  }
  tex->lm->convert_to_mixing_mode ();
}

//---------------------------------------------------------------------------

Polygon2D Polygon2D::clipped (MAX_VERTICES, true);

Polygon2D::Polygon2D (int max, bool use_uv)
{
  max_vertices = max;
  CHK (vertices = new Vector2 [max]);
  if (use_uv) { CHK (uv_coords = new Vector2 [max]); }
  else uv_coords = NULL;
  make_empty ();
}

Polygon2D::Polygon2D (Polygon2D& copy)
{
  max_vertices = copy.max_vertices;
  CHK (vertices = new Vector2 [max_vertices]);
  num_vertices = copy.num_vertices;
  memcpy (vertices, copy.vertices, sizeof (Vector2)*num_vertices);
  bbox = copy.bbox;
  if (copy.uv_coords)
  {
    CHK (uv_coords = new Vector2 [max_vertices]);
    memcpy (uv_coords, copy.uv_coords, sizeof (Vector2)*num_vertices);
  }
  else uv_coords = NULL;
}

Polygon2D::~Polygon2D ()
{
  CHK (delete [] vertices);
  CHK (delete [] uv_coords);
}

void Polygon2D::make_empty ()
{
  num_vertices = 0;
  bbox.start_bounding_box ();
}

void Polygon2D::add_vertex (float x, float y)
{
  vertices[num_vertices].x = x;
  vertices[num_vertices].y = y;
  num_vertices++;
  bbox.add_bounding_vertex (x, y);
  if (num_vertices > max_vertices)
  {
    CsPrintf (MSG_FATAL_ERROR, "OVERFLOW add_vertex!\n");
    fatal_exit (0, false);
  }
}

void Polygon2D::add_perspective (float x, float y, float z)
{
  float iz = csCamera::aspect/z;
  float px, py;

  if (num_vertices >= max_vertices)
  {
    CsPrintf (MSG_FATAL_ERROR, "OVERFLOW add_perspective!\n");
    fatal_exit (0, false);
  }

  px = x * iz + (FRAME_WIDTH/2);
  py = y * iz + (FRAME_HEIGHT/2);
  vertices[num_vertices].x = px;
  vertices[num_vertices].y = py;

  num_vertices++;
  bbox.add_bounding_vertex (px, py);
}

#if USE_OCCLUSION
bool Polygon2D::clip_to_occlusion (Occlusion* o)
{
  if (o->hull_is_empty ()) return true;

  int i, in;
  for (i = 0 ; i < num_vertices ; i++)
  {
    in = o->point_in_hull (vertices[i]);
    if (in == POLY_OUT) return true;
  }
  return false;
}
#endif

void Polygon2D::draw (int col)
{
  int i;
  int x1, y1, x2, y2;

  if (!num_vertices) return;

  x1 = QRound (vertices[num_vertices-1].x);
  y1 = QRound (vertices[num_vertices-1].y);
  for (i = 0 ; i < num_vertices ; i++)
  {
    x2 = QRound (vertices[i].x);
    y2 = QRound (vertices[i].y);
    Gfx2D->DrawLine (x1, FRAME_HEIGHT - 1 - y1, x2, FRAME_HEIGHT - 1 - y2, col);

    x1 = x2;
    y1 = y2;
  }
}

G3DPolygon g3dpoly;

//@@@ for debugging
Polygon3D* hilight;
bool do_coord_check;
Vector2 coord_check_vector;

void Polygon2D::draw_filled (Polygon3D* poly, PolyPlane* plane, bool mirror, bool use_z_buf,
	Vector2* orig_triangle)
{
  int i;
  bool debug = false;

  // If do_coord_check is true we are in a debugging mode. Here we don't
  // draw the polygon but check if the given coordinate is inside the
  // polygon. If so we dump all information about the polygon.
  if (do_coord_check)
  {
    CHK (Polygon2D* pp = new Polygon2D (max_vertices));
    if (mirror)
      for (i = 0 ; i < num_vertices ; i++)
        pp->add_vertex (vertices[num_vertices-i-1]);
    else
      for (i = 0 ; i < num_vertices ; i++)
        pp->add_vertex (vertices[i]);
    if (coord_check_vector.in_poly_2d (pp->get_vertices (),
    	pp->get_num_vertices (), &pp->get_bounding_box ()) != POLY_OUT)
    {
      PolygonSet* ps = (PolygonSet*)(poly->get_parent ());
      CsPrintf (MSG_DEBUG_0, "Hit polygon '%s/%s'\n", ps->get_name (),
      	poly->get_name ());
      Dumper::dump (this, "Polygon2D");
      Dumper::dump (poly);
      debug = true;
      hilight = poly;
    }

    CHK (delete pp);
    if (!debug) return;
    // If debug we go to the rest of the processing and
    // call a debugging function in Graph3D.
  }

  if (use_z_buf)
  {
    Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERTESTENABLE, true);
    Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERFILLENABLE, true);
  }
  else
  {
    Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERTESTENABLE, false);
    Gfx3D->SetRenderState (G3DRENDERSTATE_ZBUFFERFILLENABLE, true);
  }

  memset(&g3dpoly, 0, sizeof(G3DPolygon));
  g3dpoly.num = num_vertices;

  if (poly->get_uv_coords ())
  {
    // We are going to use DrawPolygonQuick.

    // First we have to find the u,v coordinates for every
    // point in the clipped polygon. We know we started
    // from orig_triangle and that texture mapping is not perspective correct.

    // Compute U & V in vertices of the polygon
    // First find the topmost triangle vertex
    int top;
    if (orig_triangle [0].y < orig_triangle [1].y)
      if (orig_triangle [0].y < orig_triangle [2].y) top = 0;
      else top = 2;
    else
      if (orig_triangle [1].y < orig_triangle [2].y) top = 1;
      else top = 2;

    g3dpoly.pi_texture = GetIMipMapContainerFromTextureMM (poly->get_texture ());
    g3dpoly.pi_triangle = orig_triangle;
    g3dpoly.pi_tritexcoords[0].z = 1. / poly->vcam (0).z;
    g3dpoly.pi_tritexcoords[0].u = poly->get_uv_coords ()[0].x;
    g3dpoly.pi_tritexcoords[0].v = poly->get_uv_coords ()[0].y;
    g3dpoly.pi_tritexcoords[1].z = 1. / poly->vcam (1).z;
    g3dpoly.pi_tritexcoords[1].u = poly->get_uv_coords ()[1].x;
    g3dpoly.pi_tritexcoords[1].v = poly->get_uv_coords ()[1].y;
    g3dpoly.pi_tritexcoords[2].z = 1. / poly->vcam (2).z;
    g3dpoly.pi_tritexcoords[2].u = poly->get_uv_coords ()[2].x;
    g3dpoly.pi_tritexcoords[2].v = poly->get_uv_coords ()[2].y;
    csColor* po_colors = poly->get_colors ();
    if (po_colors)
    {
      g3dpoly.pi_tritexcoords[0].r = po_colors[0].red;
      g3dpoly.pi_tritexcoords[0].g = po_colors[0].green;
      g3dpoly.pi_tritexcoords[0].b = po_colors[0].blue;
      g3dpoly.pi_tritexcoords[1].r = po_colors[1].red;
      g3dpoly.pi_tritexcoords[1].g = po_colors[1].green;
      g3dpoly.pi_tritexcoords[1].b = po_colors[1].blue;
      g3dpoly.pi_tritexcoords[2].r = po_colors[2].red;
      g3dpoly.pi_tritexcoords[2].g = po_colors[2].green;
      g3dpoly.pi_tritexcoords[2].b = po_colors[2].blue;
    }
    int j;
    CHK (g3dpoly.pi_texcoords = new G3DPolygon::poly_texture_def [64]);
    for (j = 0 ; j < num_vertices ; j++)
    {
      float x = g3dpoly.vertices[j].sx = vertices[j].x;
      float y = g3dpoly.vertices[j].sy = vertices[j].y;

      // Find the triangle left/right edges between which
      // the given x,y, point is located.
      int vtl, vtr, vbl, vbr;
      vtl = vtr = top;
      vbl = (vtl + 1) % 3;
      vbr = (vtl + 3 - 1) % 3;
      if (y > orig_triangle [vbl].y)
      {
        vtl = vbl;
        vbl = (vbl + 1) % 3;
      }
      else if (y > orig_triangle [vbr].y)
      {
        vtr = vbr;
        vbr = (vbr + 3 - 1) % 3;
      }

      // Now interpolate Z,U,V by Y
      float tL = orig_triangle [vbl].y - orig_triangle [vtl].y;
      if (tL) tL = (y - orig_triangle [vtl].y) / tL;
      float tR = orig_triangle [vbr].y - orig_triangle [vtr].y;
      if (tR) tR = (y - orig_triangle [vtr].y) / tR;
      float xL = orig_triangle [vtl].x + tL * (orig_triangle [vbl].x - orig_triangle [vtl].x);
      float xR = orig_triangle [vtr].x + tR * (orig_triangle [vbr].x - orig_triangle [vtr].x);
      float tX = xR - xL;
      if (tX) tX = (x - xL) / tX;

#     define INTERPOLATE(val,tl,bl,tr,br)	\
        {					\
          float vl,vr;				\
          if (tl != bl)				\
            vl = tl + (bl - tl) * tL;		\
          else					\
            vl = tl;				\
          if (tr != br)				\
            vr = tr + (br - tr) * tR;		\
          else					\
            vr = tr;				\
          val = vl + (vr - vl) * tX;		\
        }

      // Calculate Z
      INTERPOLATE (g3dpoly.pi_texcoords [j].z,
          g3dpoly.pi_tritexcoords [vtl].z, g3dpoly.pi_tritexcoords [vbl].z,
          g3dpoly.pi_tritexcoords [vtr].z, g3dpoly.pi_tritexcoords [vbr].z);
      // Calculate U
      INTERPOLATE (g3dpoly.pi_texcoords [j].u,
          g3dpoly.pi_tritexcoords [vtl].u, g3dpoly.pi_tritexcoords [vbl].u,
          g3dpoly.pi_tritexcoords [vtr].u, g3dpoly.pi_tritexcoords [vbr].u);
      // Calculate V
      INTERPOLATE (g3dpoly.pi_texcoords [j].v,
          g3dpoly.pi_tritexcoords [vtl].v, g3dpoly.pi_tritexcoords [vbl].v,
          g3dpoly.pi_tritexcoords [vtr].v, g3dpoly.pi_tritexcoords [vbr].v);
      if (po_colors)
      {
        // Calculate R
        INTERPOLATE (g3dpoly.pi_texcoords [j].r,
            g3dpoly.pi_tritexcoords [vtl].r, g3dpoly.pi_tritexcoords [vbl].r,
            g3dpoly.pi_tritexcoords [vtr].r, g3dpoly.pi_tritexcoords [vbr].r);
        // Calculate G
        INTERPOLATE (g3dpoly.pi_texcoords [j].g,
            g3dpoly.pi_tritexcoords [vtl].g, g3dpoly.pi_tritexcoords [vbl].g,
            g3dpoly.pi_tritexcoords [vtr].g, g3dpoly.pi_tritexcoords [vbr].g);
        // Calculate B
        INTERPOLATE (g3dpoly.pi_texcoords [j].b,
            g3dpoly.pi_tritexcoords [vtl].b, g3dpoly.pi_tritexcoords [vbl].b,
            g3dpoly.pi_tritexcoords [vtr].b, g3dpoly.pi_tritexcoords [vbr].b);
      }
    }
    Gfx3D->DrawPolygonQuick (g3dpoly, po_colors != NULL);
    CHK (delete [] g3dpoly.pi_texcoords);
  }
  else
  {
    // We are going to use DrawPolygon.
    if (mirror)
      for (i = 0 ; i < num_vertices ; i++)
      {
        g3dpoly.vertices[num_vertices-i-1].sx = vertices[i].x;
        g3dpoly.vertices[num_vertices-i-1].sy = vertices[i].y;
      }
    else
      for (i = 0 ; i < num_vertices ; i++)
      {
        g3dpoly.vertices[i].sx = vertices[i].x;
        g3dpoly.vertices[i].sy = vertices[i].y;
      }
    g3dpoly.polygon = GetIPolygon3DFromPolygon3D(poly);
    g3dpoly.id = poly->get_id ();

    g3dpoly.plane.m_cam2tex = &plane->m_cam2tex;
    g3dpoly.plane.v_cam2tex = &plane->v_cam2tex;

    float Ac, Bc, Cc, Dc;
    plane->get_camera_normal (&Ac, &Bc, &Cc, &Dc);
    g3dpoly.normal.A = Ac;
    g3dpoly.normal.B = Bc;
    g3dpoly.normal.C = Cc;
    g3dpoly.normal.D = Dc;

    if (debug)
      Gfx3D->DrawPolygonDebug (g3dpoly);
    else
      Gfx3D->DrawPolygon (g3dpoly);
  }

  if (hilight == poly) draw (System->world->get_textures ()->white ());
}

void Polygon2D::add_fog_polygon (Polygon3D* poly, PolyPlane* plane, bool mirror, CS_ID id, int fogtype)
{
  int i;

  memset(&g3dpoly, 0, sizeof(G3DPolygon));
  g3dpoly.num = num_vertices;
  if (mirror)
    for (i = 0 ; i < num_vertices ; i++)
    {
      g3dpoly.vertices[num_vertices-i-1].sx = vertices[i].x;
      g3dpoly.vertices[num_vertices-i-1].sy = vertices[i].y;
    }
  else
    for (i = 0 ; i < num_vertices ; i++)
    {
      g3dpoly.vertices[i].sx = vertices[i].x;
      g3dpoly.vertices[i].sy = vertices[i].y;
    }
  g3dpoly.polygon = GetIPolygon3DFromPolygon3D(poly);
  g3dpoly.id = poly->get_id ();

  g3dpoly.plane.m_cam2tex = &plane->m_cam2tex;
  g3dpoly.plane.v_cam2tex = &plane->v_cam2tex;

  float Ac, Bc, Cc, Dc;
  plane->get_camera_normal (&Ac, &Bc, &Cc, &Dc);
  g3dpoly.normal.A = Ac;
  g3dpoly.normal.B = Bc;
  g3dpoly.normal.C = Cc;
  g3dpoly.normal.D = Dc;

  Gfx3D->SetRenderState(G3DRENDERSTATE_ZBUFFERTESTENABLE, true);
  Gfx3D->SetRenderState(G3DRENDERSTATE_ZBUFFERFILLENABLE, true);
  Gfx3D->AddFogPolygon (id, g3dpoly, fogtype);
}



//---------------------------------------------------------------------------
