/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef POLYGON_H
#define POLYGON_H

#include "com/com.h"
#include "util/def.h"
#include "geometry/math3d.h"
#include "geometry/math2d.h"
#include "geometry/polyint.h"
#include "geometry/polyclip.h"
#include "engine/polygon/polyplan.h"
#include "engine/basic/polyset.h"
#include "engine/polygon/portal.h"
#include "engine/cscolor.h"

class csSector;
class StatLight;
class CLights;
class TextureMM;
class PolyPlane;
class Polygon2D;
class LightMap;
class LightPatch;
class Dumper;
class PolyTexture;
class PolygonSet;
#if USE_OCCLUSION
class Occlusion;
#endif

// Values returned by classify.
#define POL_SAME_PLANE 0
#define POL_FRONT 1
#define POL_BACK 2
#define POL_SPLIT_NEEDED 3

/**
 * This is our main 3D polygon class. Polygons are used to construct the
 * outer hull of sectors and the faces of 3D things.
 * Polygons can be transformed in 3D (usually they are transformed so
 * that the camera position is at (0,0,0) and the Z-axis is forward).
 * Polygons cannot be transformed in 2D. That's what Polygon2D is for.
 * It is possible to convert a Polygon3D to a Polygon2D though, at
 * which point processing continues with the Polygon2D object.<p>
 *
 * Polygons have a texture and lie on a plane. The plane does not
 * define the orientation of the polygon but is derived from it. The plane
 * does define how the texture is scaled and translated accross the surface
 * of the polygon.
 * Several planes can be shared for different polygons. As a result of this
 * their textures will be correctly aligned.<p>
 *
 * If a polygon is part of a sector it can be a portal to another sector.
 * A portal-polygon is a see-through polygon that defines a view to another
 * sector. Normally the texture for a portal-polygon is not drawn unless
 * the texture is filtered in which case it is drawn on top of the other
 * sector.<p>
 */
class Polygon3D : public PolygonInt
{
  friend class Dumper;

public:
  /// SJI - the dynamic light for this poly
  CLights* theDynLight;

private:
  /// A table of indices into the vertices of the parent PolygonSet (container).
  int* vertices_idx;
  /// Number of vertices.
  int num_vertices;
  /// Maximum number of vertices.
  int max_vertices;

  /**
   * The following two fields are somewhat related. 'poly_set' is the real
   * parent (container) of this polygon. It is either a 3D thing or a sector.
   * 'sector' is always a sector. If this polygon belongs to a sector ('poly_set'
   * is a sector) then 'sector' will have the same value as 'poly_set'. If
   * this polygon belongs to a thing then 'sector' will be set to the sector
   * containing the thing.<p>
   *
   * @@@ Note! We have to reconsider this. If a thing moves to another sector
   * we would have to update this variable for all polygons of the thing.
   */
  PolygonSet* poly_set;
  /// The csSector that this polygon is in.
  csSector* sector;

  /**
   * If not-null, this polygon is a portal.
   */
  Portal* portal;

  /**
   * If 'delete_portal' is true this portal was allocated by
   * this Polygon and it should also be deleted by it.
   */
  bool delete_portal;

  /**
   * The PolygonPlane for this polygon.
   */
  PolyPlane* plane;

  /**
   * If 'delete_plane' is true this plane was allocated by
   * this Polygon and it should also be deleted by it.
   */
  bool delete_plane;

  /// The texture.
  TextureMM* txtMM;

  /**
   * The textures and lightmaps for the four mipmap levels.
   * The transformed texture for this polygon.
   */
  PolyTexture* tex;
  /// A mipmapped texture: one step further.
  PolyTexture* tex1;
  /// A mipmapped texture: two steps further.
  PolyTexture* tex2;
  /// A mipmapped texture: three steps further.
  PolyTexture* tex3;
  /// The light-map for this polygon.
  LightMap* lightmap;
  /// The light-map for this polygon (mipmap level 1)
  LightMap* lightmap1;
  /// The light-map for this polygon (mipmap level 2)
  LightMap* lightmap2;
  /// The light-map for this polygon (mipmap level 3)
  LightMap* lightmap3;

  /**
   * If 'delete_tex_info' is true this polygon is responsible for
   * cleaning the PolyTextures.
   */
  bool delete_tex_info;

  /**
   * This bool indicates if the lightmap is up-to-date (read from the
   * cache). If set to false the polygon still needs to be recalculated.
   */
  bool lightmap_up_to_date;

  /**
   * The original polygon. This is useful when a BSP tree
   * has split a polygon. In that case, orig_poly will indicate the
   * original polygon that this polygon was split from.
   * If not split then this will be NULL.
   */
  Polygon3D* orig_poly;

  /**
   * This is a serial number that is assigned to all copies of a split
   * polygon. The original polygon gets number 0. The other polygons
   * have sequential numbers. This number is used by the lighting cache
   * in CSA files to distinguish the different lightmaps.
   */
  int serial;

  /**
   * This flag indicates that we want no mipmapping for this polygon.
   * This is useful for sky-textures (for example).
   */
  int no_mipmap;

  /**
   * This flag indicates that we want no lighting for this polygon.
   */
  int no_lighting;

  /**
   * This field describes how the light hitting this polygon is affected
   * by the angle by which the beam hits the polygon. If this value is
   * equal to -1 (default) then the global PolyTexture::cfg_cosinus_factor
   * will be used.
   */
  float cosinus_factor;

  /**
   * List of light patches for this polygon.
   */
  LightPatch* lightpatches;

  /**
   * If the following arrays is non-NULL then this polygon is drawn
   * using a different technique (all previous mipmap and lightmap stuff
   * is ignored). The polygon will be drawn with perspective incorrect
   * texture mapping and (in the future) Gouroud shading. The following
   * array specifies the u,v coordinates for every vertex of the polygon.
   */
  Vector2* uv_coords;

  /**
   * If uv_coords is given then this can be an optional array of vertex
   * colors. If this array is given then gouroud shading is used. Otherwise
   * flatshading.
   */
  csColor* colors;

  /**
   * Precompute the plane normal. Normally this is done automatically by
   * set_texture_space but if needed you can call this function again when
   * something has changed.
   */
  void compute_normal ();

  /**
   * Return twice the signed area of the polygon in world space coordinates using
   * the yz, zx, and xy components. In effect this calculates the (P,Q,R) or the
   * plane normal of the polygon.
   */
  void PlaneNormal (float* yz, float* zx, float* xy);

public:
  /// Option variable: force lightmap recalculation?
  static bool do_force_recalc;
  /// Option variables: mipmap distances
  static float zdist_mipmap1, zdist_mipmap2, zdist_mipmap3;
  /// Option variable: shadow mipmap size
  static int def_mipmap_size;

  /**
   * If this flag is true then this polygon will never be drawn.
   * This is useful for polygons which have been split. The original
   * unsplit polygon is still kept because it holds the shared
   * information about the lighted texture and lightmaps (all split
   * children refer to the original polygon for that).
   */
  bool dont_draw;

  /**
   * Construct a new polygon with the given name and texture.
   */
  Polygon3D (char* name, TextureMM* texture);

  /**
   * Construct a new polygon and copy from the given polygon.
   * Note! Several fields will reflect that a copy was made!
   * New polytextures will be allocated.
   */
  Polygon3D (Polygon3D& poly);

  ///
  virtual ~Polygon3D ();

  /**
   * Clone this polygon.
   */
  PolygonInt* clone () { CHK (PolygonInt* p = (PolygonInt*)new Polygon3D (*this)); return p; }

  /**
   * Clear the polygon (remove all vertices).
   * Also remove the uv coordinates if any.
   */
  void reset ();

  /**
   * Add a vertex from the container (polygonset) to the polygon.
   */
  int add_vertex (int v);

  /**
   * Add a vertex in a smart way: check if the vertex
   * already exists in the containing PolygonSet. If so,
   * return that index. Otherwise add the new vertex.
   */
  int add_vertex (Vector3& v);

  ///
  int add_vertex (float x, float y, float z);

  /**
   * Finish adding vertices.
   */
  void finish () { precalculate (); }

  /**
   * Set an (u,v) texture coordinate for the specified vertex
   * of this polygon. This function may only be called after all
   * vertices have been added. As soon as this function is used
   * this polygon will be rendered using a different technique
   * (perspective incorrect texture mapping and Gouroud shading).
   * This is useful for triangulated objects for which the triangles
   * are very small and also for drawing very far away polygons
   * for which perspective correctness is not needed (sky polygons for
   * example).
   */
  void set_uv (int i, float u, float v);

  /**
   * Reset the (u,v) texture coordinates which may have been set by set_uv().
   * After calling this function normal texture mapping (perspective correct) will
   * be used. Also clear the color table if present.
   */
  void reset_uv ();

  /// Get the pointer to the uv coordinates.
  Vector2* get_uv_coords () { return uv_coords; }

  /// Get the pointer to the color table.
  csColor* get_colors () { return colors; }

  /**
   * Set a color. Only use after 'set_uv' has been called for this index.
   * When this function is called for the first time the color array will
   * be initialized to all black.
   */
  void set_color (int i, float r, float g, float b);
  /// Set a color.
  void set_color (int i, csColor& c) { set_color (i, c.red, c.green, c.blue); }

  /**
   * If the polygon is a portal this will set the sector
   * that this portal points to. This function will possibly
   * create a PortalCS.
   */
  void set_cs_portal (csSector* sector);

  /**
   * Set a pre-created portal on this polygon.
   */
  void set_portal (Portal* prt);

  /**
   * Get the portal structure (if this is a portal).
   */
  Portal* get_portal () { return portal; }

  /**
   * Set the polygonset (container) that this polygon belongs to.
   */
  void set_parent (PolygonParentInt* poly_set) { Polygon3D::poly_set = (PolygonSet*)poly_set; }

  /**
   * Get the polygonset (container) that this polygons belongs to.
   */
  PolygonParentInt* get_parent () { return (PolygonParentInt*)poly_set; }

  /**
   * Get the sector that this polygon belongs to.
   */
  csSector* get_sector () { return sector; }

  /**
   * Set the sector that this polygon belongs to.
   */
  void set_sector (csSector* sector) { Polygon3D::sector = sector; }

  /**
   * Return the original polygon if this polygon was created
   * by copying from another polygon.
   */
  Polygon3D* get_orig_poly () { return orig_poly; }

  /**
   * Return a serial number that is updated when a polygon
   * is copied.
   */
  int get_serial () { return serial; }

  /**
   * Return the plane of this polygon.
   */
  PolyPlane* get_plane () { return plane; }

  /**
   * Return the plane of this polygon.
   */
  Plane* get_poly_plane () { return &plane->get_world_plane (); }

  /**
   * Return the maximum number of vertices in this polygon.
   */
  int get_max_vertices () { return max_vertices; }

  /**
   * Return the current number of vertices in this polygon.
   */
  int get_num_vertices () { return num_vertices; }

  /**
   * Get the table with all indeces of the vertices.
   */
  int* get_vertices_idx () { return vertices_idx; }

  /**
   * Set the warping transformation for the portal (if any).
   */
  void set_warp (const Transform& t) { if (portal) portal->set_warp (t); }

  /**
   * Set the warping transformation for the portal (if any).
   */
  void set_warp (const Matrix3& m_w, const Vector3& v_w_before, const Vector3& v_w_after)
  {
    if (portal) portal->set_warp (m_w, v_w_before, v_w_after);
  }

  /**
   * 'idx' is a local index into the vertices table of the polygon.
   * This index is translated to the index in the parent container and
   * a reference to the vertex is returned.
   */
  Vector3& vwor (int idx) { return poly_set->vwor (vertices_idx[idx]); }

  /**
   * 'idx' is a local index into the vertices table of the polygon.
   * This index is translated to the index in the parent container and
   * a reference to the vertex is returned.
   */
  Vector3& vobj (int idx) { return poly_set->vobj (vertices_idx[idx]); }

  /**
   * 'idx' is a local index into the vertices table of the polygon.
   * This index is translated to the index in the parent container and
   * a reference to the vertex is returned.
   */
  Vector3& vcam (int idx) { return poly_set->vcam (vertices_idx[idx]); }

  /**
   * Set the texture for this polygon.
   */
  void set_texture (TextureMM* texture);

  /**
   * Get the texture.
   */
  TextureMM* get_texture () { return txtMM; }

  /**
   * Get the polytexture (lighted texture) with a given mipmap level (0, 1, 2 or 3).
   */
  PolyTexture* get_polytex (int mipmap);

  /**
   * Get the polytexture (lighted texture) with a given distance.
   */
  PolyTexture* get_polytex (float z_dist);

  /**
   * Enable/disable lighting for this polygon.
   */
  void set_no_lighting (int no) { no_lighting = no; }

  /**
   * Enable/disable mipmapping for this polygon.
   */
  void set_no_mipmap (int no) { no_mipmap = no; }

  /**
   * Get the cosinus factor.
   */
  float get_cosinus_factor () { return cosinus_factor; }

  /**
   * Set the cosinus factor.
   */
  void set_cosinus_factor (float f) { cosinus_factor = f; }

  /**
   * Set the alpha transparency value for this polygon (only if
   * it is a portal).
   * Currently only 25, 50 and 75 are supported.
   */
  void set_alpha (int da) { if (portal) portal->set_alpha (da); }

  /**
   * Set the fog level (only useful for portals).
   */
  void set_fog (int f) { if (portal) portal->set_fog (f); }

  /**
   * Is lighting enabled?
   */
  int get_no_lighting () { return no_lighting; }

  /**
   * Is mipmapping enabled?
   */
  int get_no_mipmap () { return no_mipmap; }

  /**
   * Get the alpha transparency value for this polygon.
   */
  int get_alpha () { return portal ? portal->get_alpha () : 0; }

  /**
   * Get the fog level for this polygon.
   */
  int get_fog () { return portal ? portal->get_fog () : 0; }

  /**
   * One of the set_texture_space functions should be called after
   * adding all vertices to the polygon (not before) and before
   * doing any processing on the polygon (not after)!
   * It makes sure that the plane normal is correctly computed and
   * the texture and plane are correctly initialized.
   * 'precalculate ()' is called after that.<p>
   *
   * Internally the transformation from 3D to texture space is
   * represented by a matrix and a vector. You can supply this
   * matrix directly or let it be calculated from other parameters.
   * If you supply another Polygon or a PolyPlane to this function
   * it will automatically share the plane.
   */

  /**
   * Copy the plane from the other polygon. The plane is automatically shared.
   */
  void set_texture_space (Polygon3D* copy_from);

  /**
   * Use the given plane.
   */
  void set_texture_space (PolyPlane* plane);

  /**
   * Calculate the matrix using two vertices (which are preferably on the
   * plane of the polygon and are possibly (but not necessarily) two vertices
   * of the polygon). The first vertex is seen as the origin and the second
   * as the u-axis of the texture space coordinate system. The v-axis is
   * calculated on the plane of the polygon and orthogonal to the given
   * u-axis. The length of the u-axis and the v-axis is given as the 'len1'
   * parameter.<p>
   *
   * For example, if 'len1' is equal to 2 this means that texture will
   * be tiled exactly two times between vertex 'v_orig' and 'v1'.
   * I hope this explanation is clear since I can't seem to make it
   * any clearer :-)
   */
  void set_texture_space (Vector3& v_orig, Vector3& v1, float len1);
  ///
  void set_texture_space (float xo, float yo, float zo,
			  float x1, float y1, float z1, float len1);

  /**
   * Use 'v1' and 'len1' for the u-axis and 'v2' and 'len2' for the
   * v-axis. Otherwise this function is the same as the previous one.
   */
  void set_texture_space (float xo, float yo, float zo,
			  float x1, float y1, float z1, float len1,
			  float x2, float y2, float z2, float len2);

  /**
   * The most general function. With these you provide the matrix
   * directly.
   */
  void set_texture_space (Matrix3& tx_matrix, Vector3& tx_vector);

  /**
   * After the plane normal and the texture matrices have been set
   * up this routine makes some needed pre-calculations for this polygon.
   * It will create a texture space bounding box that
   * is going to be used for lighting and the texture cache.
   * Then it will allocate the light map tables for this polygons.
   * You also need to call this function if you make a copy of a
   * polygon (using the copy constructor) or if you change the vertices
   * in a polygon.
   */
  void precalculate ();

  /**
   * Apply the current mipmap settings to this polygon.
   * This function also converts the lightmaps to the correct
   * format required by the 3D driver.
   */
  void mipmap_settings (int setting);

  /**
   * Get the lightmap belonging with this polygon.
   */
  LightMap* get_lightmap () { return orig_poly ? orig_poly->lightmap : lightmap; }

  /**
   * A dynamic light has changed (this can be either an
   * intensity/color change of a pseudo-dynamic light or else
   * a real dynamic light change).
   */
  void make_dirty_dynamic_lights ();

  /**
   * Unlink a light patch from the light patch list.
   * Warning! This function does not test if the light patch
   * is really on the list!
   */
  void unlink_lightpatch (LightPatch* lp);

  /**
   * Add a light patch to the light patch list.
   */
  void add_lightpatch (LightPatch *lp);

  /**
   * Get the list of light patches for this polygon.
   */
  LightPatch* get_lightpatches () { return lightpatches; }

  /**
   * Clip a polygon against a frustrum in camera space.
   * The frustrum is defined with (0,0,0) as one of the
   * points of the plane.
   * If the frustrum == NULL then it is considered infinite.
   * This function returns false (and does not allocate the
   * clipped polygon) if the polygon is completely clipped
   * away. Otherwise it will allocated a new array
   * of Vector3 in dest.
   */
  bool clip_poly (Vector3* frustrum, int num_frustrum, bool mirror, Vector3** dest, int* num_dest);

  /**
   * Clip a polygon against a plane (in camera space).
   * The plane is defined as going through v1, v2, and (0,0,0).
   * The 'verts' array is modified and 'num' is also modified if needed.
   */
  void clip_poly_plane (Vector3* verts, int* num, bool mirror, Vector3& v1, Vector3& v2);

  /**
   * See if a polygon is visible from the given center (in world space
   * using backface culling) and then clip against the frustrum and return
   * a new frustrum (in camera space with the frustrum center at (0,0,0)).
   * This function returns false if the polygon is completely clipped away
   * or if it is not visible. No new_frustrum will be allocated in that case.
   * If 'mirror' is true the given frustrum is mirrored (vertices in anti-clockwise
   * order). This function correctly handles that case and will return a new
   * frustrum that is also mirrored.
   */
  bool clip_frustrum (Vector3& center, Vector3* frustrum, int num_frustrum, bool mirror,
  	Vector3** new_frustrum, int* new_num_frustrum);

  /**
   * Initialize the lightmaps for this polygon.
   * Should be called before calling shine_lightmaps and before
   * calling cache_lightmaps ().
   * This function will try to read the lightmap from the
   * cache in the level archive.
   * If do_cache == false this function will not try to read the lightmap
   * from the cache.
   */
  void init_lightmaps (bool do_cache = true);

  /**
   * Update the lightmap of this polygon according to the given light and
   * the frustrum. The light is given in world space coordinates. The
   * view frustrum is given in camera space (with (0,0,0) the origin
   * of the frustrum). The camera space used is just world space translated
   * so that the center of the light is at (0,0,0).
   * If the lightmaps were cached in the level archive this function will
   * do nothing.
   */
  void shine_lightmaps (LightView& lview);

  /**
   * Call after calling init_lightmaps and shine_lightmaps to cache
   * the calculated lightmap to the level archive. This function does
   * nothing if the cached lightmap was already up-to-date.
   */
  void cache_lightmaps ();

  /**
   * Transform the plane of this polygon from object space to world space.
   * This is mainly used for things since sectors currently have
   * identical object and world space coordinates.
   */
  void object_to_world (const ReversibleTransform& t);

  /**
   * Clip this camera space polygon to the given plane. 'plane' can be NULL
   * in which case no clipping happens.<p>
   *
   * If this function returns false then the polygon is not visible (backface
   * culling, no visible vertices, ...) and 'verts' will be NULL. Otherwise
   * this function will return true and 'verts' will point to the new
   * clipped polygon (this is a pointer to a static table of vertices.
   * WARNING! Because of this you cannot do new clip_to_plane calls until
   * you have processed the 'verts' array!).<p>
   *
   * If 'cw' is true the polygon has to be oriented clockwise in order to be
   * visible. Otherwise it is the other way around.
   */
  bool clip_to_plane (Plane* portal_plane, const Vector3& v_w2c, Vector3*& pverts,
    int& num_verts, bool cw = true);

  /**
   * This is the link between Polygon3D and Polygon2D (see below for
   * more info about Polygon2D). It should be used after the parent
   * container has been transformed from world to camera space.
   * It will fill the given Polygon2D with a perspective corrected
   * polygon that is also clipped to the view plane (Z=SMALL_Z).
   * If all vertices are behind the view plane the polygon will not
   * be visible and it will return false.
   * 'do_perspective' will also do back-face culling and returns false
   * if the polygon is not visible because of this.
   * If the polygon is deemed to be visible it will first transform
   * the plane from world to camera space and return true.
   */
  bool do_perspective (const Transform& trans,
  	Vector3* source, int num_verts, Polygon2D* dest, Vector2* orig_triangle);

  /**
   * Classify a polygon with regards to this one (in world space). If this poly
   * is on same plane as this one it returns POL_SAME_PLANE. If this poly is
   * completely in front of the given poly it returnes POL_FRONT. If this poly
   * is completely back of the given poly it returnes POL_BACK. Otherwise it
   * returns POL_SPLIT_NEEDED.
   */
  int classify (PolygonInt* spoly);

  /**
   * Split this polygon with the given plane (A,B,C,D) and put the
   * 'front' polygon in the 'front' array and the 'back' polygon in the
   * 'back' array.
   */
  void split_with_plane (Vector3* front, int& front_n, Vector3* back, int& back_n, Plane& plane);

  /**
   * Return true if this polygon and the given polygon are on the same
   * plane. If their planes are shared this is automatically the case.
   * Otherwise this function will check their respective plane equations
   * to test for equality.
   */
  bool same_plane (PolygonInt* p);

  /**
   * Intersect world-space segment with this polygon. Return
   * true if it intersects and the intersection point in world coordinates.
   */
  bool intersect_segment (const Vector3& start, const Vector3& end, 
                          Vector3& isect, float* pr = NULL);

  /**
   * Intersect world-space ray with this polygon. This function
   * is similar to intersect_segment except that it doesn't keep the lenght
   * of the ray in account. It just tests if the ray intersects with the
   * interior of the polygon.
   */
  bool intersect_ray (const Vector3& start, const Vector3& end);

  /**
   * Debugging dump of this polygon.
   */
  void dump ();

  // COM Stuff //

  DECLARE_INTERFACE_TABLE( Polygon3D )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( Polygon3D )
};

#define GetIPolygon3DFromPolygon3D( poly3d )  &poly3d->m_xPolygon3D;
#define GetPolygon3DFromIPolygon3D( iP3d )  ((Polygon3D*)((size_t)iP3d - offsetof(Polygon3D, m_xPolygon3D)))


/**
 * The maximum number of vertices that any Polygon2D can have. This should
 * be considerably larger than what you expect to be the real maximum
 * number of vertices in a 3D polygon since clipping may add new vertices.
 * Note that 100 is probably overkill but this is no problem since there
 * are not going to be many of these Polygon2D objects anyway.
 */
#define MAX_VERTICES 100

/**
 * The following class represents a 2D polygon (the 2D coordinates are
 * perspective corrected coordinates).<p>
 *
 * This class is used as the main driver for the engine pipeline.
 * The source Polygon is first converted to 2D using Polygon3D::do_perspective.
 */
class Polygon2D
{
  friend class Dumper;

private:
  /// The 2D vertices.
  Vector2* vertices;
  ///
  int num_vertices;
  ///
  int max_vertices;

  /**
   * The u,v coordinates. This array is used only when the original
   * 3D polygon also has an u,v coordinate array.
   */
  Vector2* uv_coords;

  /// A 2D bounding box that is maintained automatically.
  Box bbox;

public:
  /// Scratchpad polygon
  static Polygon2D clipped;

  ///
  Polygon2D (int max, bool use_uv = false);
  ///
  Polygon2D (Polygon2D& copy);
  ///
  virtual ~Polygon2D ();

  /**
   * Initialize the Polygon2D to empty.
   */
  void make_empty ();

  /**
   * Get the number of vertices.
   */
  int get_num_vertices () { return num_vertices; }

  /**
   * Get the maximum number of vertices.
   */
  int get_max_vertices () { return max_vertices; }

  /**
   * Get the array with all vertices.
   */
  Vector2* get_vertices () { return vertices; }

  /**
   * Get the specified vertex.
   */
  Vector2* get_vertex(int i) 
  {
    if (i<0 || i>=num_vertices) return NULL;
    return &vertices[i];
  }

  /**
   * Get the first vertex.
   */
  Vector2* get_first()
  { if (num_vertices<=0) return NULL;  else return vertices; }

  /**
   * Get the last vertex.
   */
  Vector2* get_last()
  { if (num_vertices<=0) return NULL;  else return &vertices[num_vertices-1]; }

  /**
   * Add a vertex (2D) to the polygon
   */
  void add_vertex (Vector2& v) { add_vertex (v.x, v.y); }

  /**
   * Add a vertex (2D) to the polygon
   */
  void add_vertex (float x, float y);

  /**
   * Set all polygon vertices at once
   */
  void set_vertices (Vector2 *v, int num)
  { memcpy (vertices, v, (num_vertices = num) * sizeof (Vector2)); }

  /**
   * Compute the perspective transformation of a 3D vertex and add it to the polygon.
   */
  void add_perspective (Vector3& v) { add_perspective (v.x, v.y, v.z); }

  /**
   * Compute the perspective transformation of a 3D vertex and add it to the polygon.
   */
  void add_perspective (float x, float y, float z);

  /**
   * Set the uv coordinate for a specified vertex.
   */
  void set_uv (int i, float u, float v) { uv_coords[i].x = u; uv_coords[i].y = v; }

  ///
  Box& get_bounding_box () { return bbox; }
  ///
  int overlap_bounding_box (const Box& box) { return bbox.overlap (box); }
  ///
  int overlap_bounding_box (Polygon2D* poly) { return bbox.overlap (poly->bbox); }

  /**
   * Clipping routines. They return false if the resulting polygon is not
   * visible for some reason.
   * Note that these routines must not be called if the polygon is not visible.
   * These routines will not check that.
   * Note that these routines will put the resulting clipped 2D polygon
   * in place of the original 2D polygon.
   */
  bool clip_against (csClipper* view)
  { return view->Clip (vertices, num_vertices, max_vertices, &bbox); }

# if USE_OCCLUSION
  /**
   * Clip this 2D polygon to the occlusion object. This is a very
   * approximate clipper that preserves convexity without having
   * to split the polygon in seperate parts.
   * This routine returns false if the polygon is completely clipped
   * away.
   */
  bool clip_to_occlusion (Occlusion* o);
# endif

  /**
   * Draw the polygon.
   */
  void draw (int col);
  
  /**
   * Draw a texture mapped polygon.
   * 'plane' should be a correctly transformed plane (transformed to camera
   * space).
   * 'poly' is only used for debugging. The plane and vertices are not used.
   */
  void draw_filled (Polygon3D* poly, PolyPlane* plane, bool mirror, bool use_z_buf = false,
  	Vector2* orig_triangle = NULL);

  /**
   * Add a back or front polygon of a fog object.
   * NOTE! Don't forget to open the fog object first with Gfx3D->OpenFogObject ();
   */
  void add_fog_polygon (Polygon3D* poly, PolyPlane* plane, bool mirror, CS_ID id, int fog_type);

  /**
   * Debugging dump of this polygon.
   */
  void dump (char* name);
};

#endif /*POLYGON_H*/
