/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef POLYPLANE_H
#define POLYPLANE_H

#include "geometry/math3d.h"
#include "engine/basic/csobject.h"

class Dumper;

/**
 * This class represents a texture plane. This is a plane
 * that defines the orientation and offset of a texture. It can
 * be used by several polygons to let the textures fit perfectly.
 */
class PolyPlane : public csObject
{
  ///
  friend class Polygon3D;
  ///
  friend class Polygon2D;
  ///
  friend class PolyTexture;
  ///
  friend class Dumper;

private:
  /// The object space plane equation (this is fixed).
  Plane plane_obj;
  /// The world space plane equation.
  Plane plane_wor;
  /// The camera space plane equation.
  Plane plane_cam;

  /// Transformation from object to texture space.
  Matrix3 m_obj2tex;
  /// Translation from object to texture space.
  Vector3 v_obj2tex;

  /// Transformation from world to texture space.
  Matrix3 m_world2tex;
  /// Translation from world to texture space.
  Vector3 v_world2tex;

  /**
   * Transformed texture space transformation. This transforms a
   * coordinate from camera space to texture space. This transformation
   * is obtained by using m_world2tex and v_world2tex with the camera transformation.
   */
  Matrix3 m_cam2tex;
  ///
  Vector3 v_cam2tex;

public:
  ///
  PolyPlane (char* name);
  ///
  virtual ~PolyPlane ();

  /**
   * Transform this plane from object space to world space using
   * the given matrices. One vertex on the plane is also given so
   * that we can more easily recompute the 'D' component of the plane.
   * The given vertex should be in world space.
   */
  void object_to_world (const ReversibleTransform& obj, Vector3& vertex1);

  /**
   * Transform the same way from world space to camera space.
   * The given vertex should be in camera space.
   */
  void world_to_camera (const ReversibleTransform& t, Vector3& vertex1);

  ///
  void set_texture_space (Vector3& v_orig,
  			  Vector3& v1, float len1,
  			  Vector3& v2, float len2);
  ///
  void set_texture_space (float xo, float yo, float zo,
			  float x1, float y1, float z1, float len1,
			  float x2, float y2, float z2, float len2);
  ///
  void set_texture_space (float xo, float yo, float zo,
  			  float x1, float y1, float z1,
  			  float len);
  ///
  void set_texture_space (Vector3& v_orig, Vector3& v1, float len);
  ///
  void set_texture_space (Vector3& v_orig, Vector3& v_u, Vector3& v_v);
  ///
  void set_texture_space (float xo, float yo, float zo,
  			  float xu, float yu, float zu,
  			  float xv, float yv, float zv);
  ///
  void set_texture_space (float xo, float yo, float zo,
  			  float xu, float yu, float zu,
  			  float xv, float yv, float zv,
  			  float xw, float yw, float zw);
  ///
  void set_texture_space (Matrix3& tx_matrix, Vector3& tx_vector);

  /// Get the transformation from object to texture space.
  void get_texture_space (Matrix3& tx_matrix, Vector3& tx_vector);

  /**
   * Check if the plane in world space is visible from the given point.
   * To do this it will only do back-face culling.
   */
  int visible_from_point (const Vector3& p)
  {
    return plane_wor.visible_from_point (p);
  }

  /**
   * Return the minimum squared distance from the plane to
   * a point in 3D space (using world coordinates).
   */
  float sq_distance (Vector3& v) { float d = distance (v); return d*d; }

  /**
   * Return the minimum distance from the plane to
   * a point in 3D space (using world coordinates).
   */
  float distance (Vector3& v)
  {
    // The normal is normalized so...
    return plane_wor.distance (v);
  }

  /**
   * Check if this plane is close to another one in
   * world space coordinates.
   */
  int close (PolyPlane* plane)
  {
    return plane_wor.close (plane->plane_wor);
  }

  /**
   * Classify a vector with regards to this plane in world space.
   */
  float classify (Vector3& pt)
  {
    return plane_wor.classify (pt);
  }

  /**
   * Return the closest point on the plane to a point
   * in 3D space (using world coordinates).
   */
  void closest_point (Vector3& v, Vector3& isect);

  /**
   * Run a segment through this plane (in world space) and
   * see where it intersects. Put value in 'pr' between 0 and 1 for
   * position on segment (0 is at start, 1 is at end) and also
   * put intersection point in 'isect'. Return false if there is
   * no intersection.
   */
  bool intersect_segment (const Vector3& start, const Vector3& end, 
                          Vector3& isect, float* pr);

  /**
   * Get the object version of the plane.
   */
  Plane& get_object_plane () { return plane_obj; }

  /**
   * Get the world version of the plane.
   */
  Plane& get_world_plane () { return plane_wor; }

  /**
   * Get the camera version of the plane.
   */
  Plane& get_camera_plane () { return plane_cam; }

  /**
   * Get the normal in object space.
   */
  void get_object_normal (float* p_A, float* p_B, float* p_C, float* p_D);

  /**
   * Get the normal in world space.
   */
  void get_world_normal (float* p_A, float* p_B, float* p_C, float* p_D);

  /**
   * Get the normal in camera space.
   */
  void get_camera_normal (float* p_A, float* p_B, float* p_C, float* p_D);
};

#endif /*POLYPLANE_H*/
