/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "util/def.h"
#include "engine/polygon/portal.h"
#include "engine/polygon/polygon.h"
#include "engine/sector.h"
#include "engine/stats.h"
#include "engine/texture.h"


Portal::Portal ()
{
  cfg_alpha = 0;
  cfg_fog = 0;
  do_warp_space = false;
  do_mirror = false;
  txtMM = NULL;
  filter_r = 1;
  filter_g = 1;
  filter_b = 1;
  static_dest = false;
}

void Portal::object_to_world (const ReversibleTransform& t)
{
  if (static_dest)
    warp_wor = warp_obj * t;
  else
    warp_wor = t.getinverse () * warp_obj * t;
}

void Portal::set_warp (const Transform& t)
{
  do_warp_space = true;
  warp_obj = t;
  Matrix3 m = warp_obj.m_world2this();
  do_mirror = ( ( ( m.col1() % m.col2() ) * m.col3() ) < 0 );

  warp_wor = warp_obj;
}

void Portal::set_warp (const Matrix3& m_w, const Vector3& v_w_before, const Vector3& v_w_after)
{
  do_warp_space = true;

  warp_obj = Transform(m_w.getinverse(), v_w_after - m_w * v_w_before);

  // If the three colunms of the transformation matrix are taken
  // as vectors, V1, V2, and V3, then V1 x V2 = ( + or - ) V3.
  // The result is positive for non-mirroring transforms, and
  // negative for mirroring transforms.  Thus, (V1 x V2) * V3 
  // will equal +1 or -1, depending on whether the transform is
  // mirroring.
  Matrix3 m = warp_obj.m_world2this();
  do_mirror = ( ( ( m.col1() % m.col2() ) * m.col3() ) < 0 );

  warp_wor = warp_obj;
}

void Portal::warp_space (ReversibleTransform& t, bool& mirror)
{
  // warp_wor is a world -> warp space transformation.
  // t is a world -> camera space transformation.
  // Set t to equal a warp -> camera space transformation by
  // reversing warp and then applying the old t.
  t /= warp_wor;
  if (do_mirror) mirror = !mirror;
}

bool PortalCS::draw (Polygon2D* new_clipper, Plane* portal_plane, bool loose_end, RenderView& rview)
{
  if (sector->draw_busy >= 5)
    return false;

  Stats::portals_drawn++;
  if (!new_clipper->get_num_vertices ())
    return false;

  csPolygonClipper new_view (new_clipper->get_vertices (),
    new_clipper->get_num_vertices (), rview.is_mirrored (), true);

  RenderView new_rview = rview;
  new_rview.view = &new_view;

  if (do_warp_space)
  {
    if (new_rview.clip_plane || loose_end || sector->is_bsp ())
    {
      Plane pl = *portal_plane;
      pl.invert();
      new_rview.clip_plane = &pl;
    }
    bool mirror = new_rview.is_mirrored ();
    warp_space (new_rview, mirror);
    new_rview.set_mirror (mirror);
    sector->draw (new_rview);
  }
  else
  {
    if (new_rview.clip_plane) 
    {
      Plane pl = *portal_plane; 
      pl.invert();
      new_rview.clip_plane = &pl;
    }
    sector->draw (new_rview);
  }

  return true;
}

Polygon3D* PortalCS::hit_beam (Vector3& start, Vector3& end)
{
  return sector->hit_beam (start, end);
}

Polygon3D* PortalCS::intersect_sphere (Vector3& center, float radius, float* pr)
{
  return sector->intersect_sphere (center, radius, pr);
}

csSector* PortalCS::follow_segment (ReversibleTransform& t,
				  Vector3& new_position, bool& mirror)
{
  if (do_warp_space)
  {
    warp_space (t, mirror); 
    new_position = warp_wor.world2this(new_position);
  }
  return sector->follow_segment (t, new_position, mirror);
}

bool PortalCS::blocking_things (Vector3& start, Vector3& end, Polygon3D** poly)
{
  if (sector->beam_busy >= csSector::cfg_reflections) return false;
  if (do_warp_space)
  {
    Vector3 start2 = warp_wor.world2this(start);
    Vector3 end2 = warp_wor.world2this(end);
    return sector->blocking_things (start2, end2, poly);
  }
  else
    return sector->blocking_things (start, end, poly);
}

Polygon3D* PortalCS::follow_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* sqdist)
{
  if (sector->beam_busy >= csSector::cfg_reflections) return NULL;
  if (do_warp_space)
  {
    Vector3 start2 = warp_wor.world2this(start);
    Vector3 end2 = warp_wor.world2this(end);
    return sector->follow_beam (start2, end2, poly, sqdist);
  }
  else
    return sector->follow_beam (start, end, poly, sqdist);
}

void PortalCS::shine_lightmaps (LightView& lview)
{
  if (do_warp_space)
  {
    LightView new_lview;
    new_lview.copy (lview);

    new_lview.center = warp_wor.world2this (lview.center);

    int i;
    for (i = 0 ; i < lview.num_frustrum ; i++)
      new_lview.frustrum[i] = warp_wor.world2this_rel (lview.frustrum[i]);

    if (do_mirror) new_lview.mirror = !lview.mirror;
    if (cfg_alpha)
    {
      //float a = cfg_alpha;
      //new_lview.r = (a * lview.r + (100-a) * filter_r) / 100.;
      //new_lview.g = (a * lview.g + (100-a) * filter_g) / 100.;
      //new_lview.b = (a * lview.b + (100-a) * filter_b) / 100.;
      float fr, fg, fb;
      if (txtMM)
      {
        fr = txtMM->get_mean_r ();
        fg = txtMM->get_mean_g ();
        fb = txtMM->get_mean_b ();
      }
      else
      {
        fr = filter_r;
        fg = filter_g;
        fb = filter_b;
      }
      new_lview.r = lview.r * fr;
      new_lview.g = lview.g * fg;
      new_lview.b = lview.b * fb;
    }

    // Don't go further if the light intensity is almost zero.
    if (new_lview.r >= SMALL_EPSILON || new_lview.g >= SMALL_EPSILON || new_lview.b >= SMALL_EPSILON)
      sector->shine_lightmaps (new_lview);
  }
  else sector->shine_lightmaps (lview);
}

void PortalCS::dump_frustrum (csStatLight* light, Vector3* frustrum, int num_frustrum,
	Transform& t)
{
  if (do_warp_space)
  {
    // @@@ Currently not supported yet.
    return;
  }
  else sector->dump_frustrum (light, frustrum, num_frustrum, t);
}

