/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef RVIEW_H
#define RVIEW_H

#include "geometry/math3d.h"
#include "engine/camera.h"

class csClipper;
class Matrix3;
class Vector3;
class csLight;

/**
 * This structure represents all information needed for drawing
 * a scene. It is closely related to the csCamera class and modified
 * while rendering according to portals/warping portals and such.
 */
class RenderView : public csCamera
{
public:
  
  /// The 2D polygon describing how everything drawn inside should be clipped.
  csClipper* view;

  /**
   * This variable holds the plane of the portal through which the camera
   * is looking. If NULL this is not relevant (the initial portal).
   * If not NULL this polygon will be used to clip every object so that
   * only the objects in front of the plane remain.<p>
   *
   * This is needed for portals that arive somewhere in the middle of a
   * sector. This can happen with BSP sectors or with Things containing
   * portals. In that case everything behind the portal plane should
   * be clipped away.<p>
   *
   * Since this feature is not needed for portals on sector walls of
   * non-BSP sectors this variable is set to NULL in those cases (for
   * efficiency).
   */
  Plane* clip_plane;

  ///
  RenderView() : csCamera(), view(NULL), clip_plane(NULL) {}
  ///
  RenderView(const csCamera& c) : csCamera(c), view(NULL), clip_plane(NULL) {}
  ///
  RenderView(const csCamera& c, csClipper* v, Plane* p) : 
   csCamera(c), view(v), clip_plane(p) {}

  ///
  void set_view(csClipper* v) { view = v; }
  ///
  void set_clip_plane(Plane* p) { clip_plane = p; }
};

/**
 * This structure represents all information needed for static lighting.
 * It is the basic information block that is passed between the various
 * static lighting routines.
 */
class LightView
{
public:
  /// The light that we're processing.
  csLight* l;

  /**
   * The center of the light. This may differ from the actual center
   * of 'l' because space warping may have relocated it. Routines should
   * always use this center.
   */
  Vector3 center;

  /**
   * The current color of the light. Initially this is the same as the
   * light in csStatLight but portals may change this.
   */
  float r, g, b;

  /// If space is mirrored.
  bool mirror;

  /**
   * If this structure is used for dynamic light frustrum calculation
   * then this flag is true.
   */
  bool dynamic;

  /**
   * The current view frustrum as seen from the light (relative
   * to the center of the light). If frustrum == NULL then everything is
   * visible (the full frustrum).
   */
  Vector3* frustrum;
  ///
  int num_frustrum;

public:
  ///
  LightView () { frustrum = NULL; num_frustrum = 0; }
 
  ///
  ~LightView ()
  {
    if (frustrum) CHKB (delete [] frustrum);
  }

  ///
  void set_num_frustrum (int n)
  {
    num_frustrum = n;
    if (frustrum) CHKB (delete [] frustrum);
    CHK (frustrum = new Vector3 [n]);
  }

  ///
  int get_num_frustrum () { return num_frustrum; }

  ///
  Vector3* get_frustrum () { return frustrum; }

  ///
  LightView& copy (LightView& other)
  {
    if (&other == this) return *this;
    *this = other;
    if (other.frustrum)
    {
      CHK (frustrum = new Vector3 [num_frustrum]);
      memcpy (frustrum, other.frustrum, num_frustrum*sizeof (Vector3));
    }
    return *this;
  }

  ///
  LightView& copy_nofrust (LightView& other)
  {
    if (&other == this) return *this;
    *this = other;
    frustrum = NULL,
    num_frustrum = 0;
    return *this;
  }
};


#endif /*RVIEW_H*/

