/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/system.h"
#include "system/sysprint.h"
#include "engine/sector.h"
#include "engine/objects/thing.h"
#include "engine/objects/cssprite.h"
#include "engine/polygon/polygon.h"
#include "engine/polygon/polytext.h"
#include "engine/light/dynlight.h"
#include "engine/light/light.h"
#include "engine/camera.h"
#include "engine/texture.h"
#include "engine/world.h"
#include "engine/stats.h"
#include "render/igraph3d.h"
#include "engine/wirefrm.h"
#include "geometry/bsp.h"
#include "render/ihalo.h"

// Option variable: render portals?
bool csSector::do_portals = true;
// Option variable: render things?
bool csSector::do_things = true;
// Configuration variable: number of allowed reflections for static lighting.
int csSector::cfg_reflections = 1;
// Option variable: do pseudo radiosity?
bool csSector::do_radiosity = false;

//---------------------------------------------------------------------------

csSector::csSector (char* name) : PolygonSet (name, CS_SECTOR)
{
  first_thing = NULL;
  first_mso = NULL;
  sector = this;
  beam_busy = 0;
  level_r = level_g = level_b = 0;
  static_bsp = NULL;
  static_thing = NULL;
}

csSector::~csSector ()
{
  while (first_thing)
  {
    Thing* n = (Thing*)(first_thing->get_next ());
    CHK (delete first_thing);
    first_thing = n;
  }
  CHK (delete static_bsp);
  CHK (delete static_thing);
  while (first_mso) { CHK (delete first_mso); }
  lights.DeleteAll ();
}

void csSector::add_thing (Thing* thing)
{
  thing->set_next ((PolygonSet*)first_thing);
  first_thing = thing;
}

void csSector::add_sprite (csWorld* world, csSprite3D* sprite)
{
  sprite->get_mso ().init_one_sector (this);
  world->add (&sprite->get_mso ());
}

void csSector::add_light (csStatLight* light)
{
  lights.Push (light);
  light->set_sector (this);
}

void csSector::use_static_bsp ()
{
  CHK (delete bsp); bsp = NULL;
  CHK (delete static_bsp); static_bsp = NULL;

  CHK (delete static_thing);
  CHK (static_thing = new Thing ("__static__"));

  Thing* sp = first_thing;
  while (sp)
  {
    if (!sp->isMoveable () && !sp->get_fog ().enabled) static_thing->merge (sp);
    sp = (Thing*)(sp->get_next ());
  }

  CHK (static_bsp = new BspTree (static_thing));
}

Polygon3D* csSector::hit_beam (Vector3& start, Vector3& end)
{
  Vector3 isect;
  Polygon3D* p;

  // First check the things of this sector and return the one with
  // the closest distance.
  Thing* sp = first_thing;
  float sq_dist, min_sq_dist = 100000000.;
  Polygon3D* min_poly = NULL;
  while (sp)
  {
    p = sp->intersect_segment (start, end, isect);
    if (p)
    {
      sq_dist = (isect.x-start.x)*(isect.x-start.x) +
	(isect.y-start.y)*(isect.y-start.y) +
	(isect.z-start.z)*(isect.z-start.z);
      if (sq_dist < min_sq_dist) { min_sq_dist = sq_dist; min_poly = p; }
    }
    sp = (Thing*)(sp->get_next ());
  }

  if (min_poly) return min_poly;

  p = intersect_segment (start, end, isect);
  if (p)
  {
    Portal* po = p->get_portal ();
    if (po) return po->hit_beam (start, end);
    else return p;
  }
  else return NULL;
}

void csSector::mipmap_settings (int setting)
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
  {
    Polygon3D* p = (Polygon3D*)polygons[i];
    p->mipmap_settings (setting);
  }

  Thing* sp = first_thing;
  while (sp)
  {
    sp->mipmap_settings (setting);
    sp = (Thing*)(sp->get_next ());
  }
}

Polygon3D* csSector::intersect_segment (const Vector3& start, const Vector3& end,
				      Vector3& isect, float* pr)
{
  Thing* sp = first_thing;
  while (sp)
  {
    Polygon3D* p = sp->intersect_segment (start, end, isect, pr);
    if (p) return p;
    sp = (Thing*)(sp->get_next ());
  }

  return PolygonSet::intersect_segment (start, end, isect, pr);
}

csSector* csSector::follow_segment (ReversibleTransform& t, Vector3& new_position,
				bool& mirror)
{
  Vector3 isect;
  Polygon3D* p = sector->intersect_segment (t.v_origin(), new_position, isect);
  Portal* po;

  if (p)
  {
    po = p->get_portal ();
    if (po)
      return po->follow_segment (t, new_position, mirror);
    else
      new_position = isect;
  }

  return this;
}


Polygon3D* csSector::intersect_sphere (Vector3& center, float radius, float* pr)
{
  float d, min_d = radius;
  int i;
  Polygon3D* p, * min_p = NULL;
  PolyPlane* pl;
  Vector3 hit;
  float A, B, C, D;

  for (i = 0 ; i < num_polygon ; i++)
  {
    p = (Polygon3D*)polygons[i];
    pl = p->get_plane ();
    d = pl->distance (center);
    if (d < min_d && pl->visible_from_point (center))
    {
      pl->get_world_normal (&A, &B, &C, &D);
      hit.x = d*(-A-center.x)+center.x;
      hit.y = d*(-B-center.y)+center.y;
      hit.z = d*(-C-center.z)+center.z;
      if (p->intersect_ray (center, hit))
      {
	Portal* po = p->get_portal ();
	if (po)
	{
	  p = po->intersect_sphere (center, min_d, &d);
	  if (p)
	  {
	    min_d = d;
	    min_p = p;
	  }
	}
	else
	{
	  min_d = d;
	  min_p = p;
	}
      }
    }
  }

  Thing* sp = first_thing;
  while (sp)
  {
    p = sp->intersect_sphere (center, radius, &d);
    if (p && d < min_d)
    {
      min_d = d;
      min_p = p;
    }
    sp = (Thing*)(sp->get_next ());
  }

  if (pr) *pr = min_d;
  return min_p;
}

void* csSector::draw_polygons (PolygonParentInt* pi, PolygonInt** polygon, int num, void* data)
{
  RenderView* d = (RenderView*)data;
  csSector* sector = (csSector*)pi;
  sector->draw_polygon_array (polygon, num, d, false);
  return NULL;
}

int compare_z_thing (const void* p1, const void* p2)
{
  Thing* sp1 = *(Thing**)p1;
  Thing* sp2 = *(Thing**)p2;
  float z1 = sp1->vcam (sp1->getCenter ()).z;
  float z2 = sp2->vcam (sp2->getCenter ()).z;
  if (z1 < z2) return -1;
  else if (z1 > z2) return 1;
  return 0;
}

void csSector::draw (RenderView& rview)
{
  draw_busy++;
  Vector3* old = new_transformation ();

  transform_world2cam (rview);
  Stats::polygons_considered += num_polygon;

  if (bsp)
    bsp->back2front (rview.v_origin(), &draw_polygons, (void*)&rview);
  else
  {
    draw_polygons ((PolygonParentInt*)this, polygons, num_polygon, (void*)&rview);
    if (static_thing && do_things)
    {
      Vector3* old = static_thing->new_transformation ();
      static_thing->transform_world2cam (rview);
      static_bsp->back2front (rview.v_origin(), &draw_polygons, (void*)&rview);
      static_thing->restore_transformation (old);
    }
  }

  if (do_things)
  {
    // All Things which are not merged with the static bsp still need to
    // be drawn. If there is a static bsp then we cannot do something special
    // for convex things (hulls). In that case we first draw the non-merged
    // things using normal Z-buffer and only Z sort the foggy things and draw
    // them back to front. If there is no static bsp then we can also Z sort
    // the hulls and draw then back to front using Z-fill.
    //
    // A consequence of this algorithm is that if you use a static bsp then
    // you can only use portals in Things that do not move (i.e. the Thing
    // needs to be merged with the bsp). But the Thing need not be convex in
    // that case. If you don't use a static bsp then you can use portals in
    // moving Things but the Things containing portals need to be convex.
    // a static bsp then you can only use portals 
    //
    // We should see if there are better alternatives to Z-sort which are
    // more accurate in more cases (@@@).
    Thing* sort_list[100];	// @@@HARDCODED == BAD == EASY!
    int sort_idx = 0;
    int i;
    Thing* sp = first_thing;
    if (static_thing)
    {
      // Here we have a static bsp. So we draw all Things using the
      // Z-buffer and put all foggy Things in the sort_list.
      while (sp)
      {
        if (!sp->isMerged ())
          if (sp->get_fog ().enabled) sort_list[sort_idx++] = sp;
          else sp->draw (rview);
        sp = (Thing*)(sp->get_next ());
      }
    }
    else
    {
      // Here we don't have a static bsp. In this case we put all
      // hulls (convex Things) and foggy Things in the sort_list.
      while (sp)
      {
        if (sp->isConvex () || sp->get_fog ().enabled)
	  sort_list[sort_idx++] = sp;
        sp = (Thing*)(sp->get_next ());
      }
    }

    if (sort_idx)
    {
      // Now sort the objects in sort_list.
      qsort (sort_list, sort_idx, sizeof (Thing*), compare_z_thing);

      // Draw them back to front.
      for (i = 0 ; i < sort_idx ; i++)
      {
        sp = sort_list[i];
        if (sp->get_fog ().enabled) sp->draw_foggy (rview);
        else if (!sp->isMerged () && sp->isConvex ())
      	  sp->draw (rview, false);
      }
    }

    // If there is no static bsp then we still need to draw the remaining
    // non-convex Things.
    if (!static_thing)
    {
      sp = first_thing;
      while (sp)
      {
        if (!sp->isConvex () && !sp->get_fog ().enabled) sp->draw (rview);
        sp = (Thing*)(sp->get_next ());
      }
    }

    // Draw sprites.
    SectorMSO* mso = first_mso;
    while (mso)
    {
      csSprite3D* sp3d = (csSprite3D*)(mso->mso->get_object ());
      sp3d->draw (rview);
      mso = mso->next_sector;
    }
  }

  // queue all halos in this sector to be drawn.
  IHaloRasterizer* piHR = System->world->getHaloRastizer ();
  if (piHR)
  {
    int numlights = lights.Length();
    
    for (int i = 0; i<numlights; i++)
    {
      csStatLight* light = (csStatLight*)(lights.Get(i));
      if (!light->is_halo_enabled ()) continue;
      
      // this light is already in the queue.
      if (light->get_halo_in_queue())
	continue;
      
      csHaloInformation* cshaloinfo = new csHaloInformation; 
      
      cshaloinfo->v = rview.world2cam(light->get_center());
      
      if (cshaloinfo->v.z > SMALL_Z)
      {
	float iz = csCamera::aspect/cshaloinfo->v.z;
	cshaloinfo->v.x = cshaloinfo->v.x * iz + (FRAME_WIDTH/2);
	cshaloinfo->v.y = FRAME_HEIGHT - 1 - (cshaloinfo->v.y * iz + (FRAME_HEIGHT/2));

	cshaloinfo->pLight = light;
	
	cshaloinfo->r = light->get_red();
	cshaloinfo->g = light->get_green();
	cshaloinfo->b = light->get_blue();
	cshaloinfo->intensity = 0.0f;
	
	if (piHR->TestHalo(&cshaloinfo->v) == S_OK)
	{
	  piHR->CreateHalo(cshaloinfo->r, cshaloinfo->g, cshaloinfo->b, &cshaloinfo->haloinfo);
	  System->world->add_halo(cshaloinfo);
	}
      }
      else
      {
	delete cshaloinfo;
      }
    }
  }

  WireFrame* wf = NULL;
  if (System->world->map_mode != MAP_OFF)
  {
    wf = System->world->wf->get_wireframe ();
    int i;
    for (i = 0 ; i < lights.Length () ; i++)
    {
      wfVertex* vt = wf->add_vertex (((csStatLight*)lights[i])->get_center ());
      vt->set_color (wf->get_red ());
    }
  }

  long do_edges;
  Gfx3D->GetRenderState(G3DRENDERSTATE_EDGESENABLE, do_edges);

  if (do_edges)
  {
    extern bool do_coord_check;
//  extern Vector2 coord_check_vector;
    int i;
    Vector3 v;
    float iz;
    int px, py, r;
    int red = System->world->get_textures ()->red ();
    int white = System->world->get_textures ()->white ();
    for (i = 0 ; i < lights.Length () ; i++)
    {
      v = rview.world2this(((csStatLight*)lights[i])->get_center ());
      if (v.z > SMALL_Z)
      {
        iz = csCamera::aspect/v.z;
        px = QInt (v.x * iz + (FRAME_WIDTH/2));
        py = FRAME_HEIGHT - 1 - QInt (v.y * iz + (FRAME_HEIGHT/2));
	r = QInt (.3 * iz);
	if (do_coord_check)
	{
     // DAN: Commented out this code for now, until we decide
     // where to put light selection. 10.05.98
/*	  if (ABS (coord_check_vector.x - px) < 5 && ABS (coord_check_vector.y - (FRAME_HEIGHT-1-py)) < 5)
	  {
	    Gfx3D->selected_light = lights[i];
	    CsPrintf (MSG_CONSOLE, "Selected light %s/(%f,%f,%f).\n", get_name (), lights[i]->get_center ().x, lights[i]->get_center ().y, lights[i]->get_center ().z);
	    CsPrintf (MSG_DEBUG_0, "Selected light %s/(%f,%f,%f).\n", get_name (), lights[i]->get_center ().x, lights[i]->get_center ().y, lights[i]->get_center ().z);
	  }*/
	}
        Gfx2D->DrawLine (px-r, py-r, px+r, py+r, white);
        Gfx2D->DrawLine (px+r, py-r, px-r, py+r, white);
        Gfx2D->DrawLine (px, py-2, px, py+2, red);
        Gfx2D->DrawLine (px+2, py, px-2, py, red);
      }
    }
  }

  restore_transformation (old);
  draw_busy--;
}


struct SectorShineInfo
{
  csStatLight* light;
  Vector3 center;	// The center of the light (possibly warped)
  bool mirror;		// If everything is mirrored.
  Vector3* frustrum;	// A (possibly warped) view frustrum for the light.
  int num_frustrum;
  // These two are only used for dump_frustrum
  Transform *trans;
};

void* csSector::shine_polygons (PolygonParentInt*, PolygonInt** polygon, int num, void* data)
{
  Polygon3D* p;
  Portal* po;
  LightView* d = (LightView*)data;
  int i;
  for (i = 0 ; i < num ; i++)
  {
    p = (Polygon3D*)polygon[i];
    Vector3* new_frustrum = NULL;
    int new_num_frustrum = 0;
    if (p->clip_frustrum (d->center, d->frustrum, d->num_frustrum, d->mirror, &new_frustrum, &new_num_frustrum))
    {
      // First we check if the light can reach the polygon. If not then there is no
      // need to include it in the view frustrum. This is especially useful for portals
      // since then whole sectors can be culled away efficiently.

      // First compute the distance from the center of the light to the plane of
      // the polygon. If this distance is greater then the radius of the light then
      // we have a trivial case (no hit possible).
      float dist_to_plane = p->get_poly_plane ()->distance (d->center);

      bool faraway = true;
      if (dist_to_plane < d->l->get_dist ())
      {
        // The light is close enough to the plane of the polygon. Now we calculate
	// an accurate minimum squared distance from the light to the polygon. Note
	// that we use the new_frustrum which is relative to the center of the light.
	// So this algorithm works with the light position at (0,0,0) (@@@ we should
	// use this to optimize this algorithm further).
	Vector3 o (0, 0, 0);
	float min_sqdist = SqDist::point_poly (o, new_frustrum, new_num_frustrum,
		*(p->get_poly_plane ()), dist_to_plane*dist_to_plane);
	if (min_sqdist < d->l->get_sq_dist ()) faraway = false;	// Light reaches polygon.
      }

      if (!faraway)
      {
        LightView new_lview;
	new_lview.copy_nofrust (*d);
        new_lview.frustrum = new_frustrum;
        new_lview.num_frustrum = new_num_frustrum;
        p->shine_lightmaps (new_lview);
        po = p->get_portal ();
        if (po) po->shine_lightmaps (new_lview);
        else if (!new_lview.dynamic && do_radiosity)
        {
          // If there is no portal we simulate radiosity by creating
	  // a dummy portal for this polygon which reflects light.
	  PortalCS mirror;
	  mirror.set_sector (p->get_sector ());
	  mirror.set_alpha (10);
	  mirror.set_filter (
		  p->get_texture ()->get_mean_r ()/4,
		  p->get_texture ()->get_mean_g ()/4,
		  p->get_texture ()->get_mean_b ()/4);

	  mirror.set_warp (p->get_poly_plane ()->get_reflect ());
	  mirror.shine_lightmaps (new_lview);
        }
      }
      else
      {
        if (new_frustrum) CHKB (delete [] new_frustrum);
      }
    }
  }
  return NULL;
}

void csSector::shine_lightmaps (LightView& lview)
{
  if (draw_busy >= cfg_reflections) return;
  int i;

  draw_busy++;
  Vector3* old = new_transformation ();

  translate_vector (lview.center);

  LightView new_lview;
  new_lview.copy (lview);

  // First we slightly expand the frustrum. This is to make sure that
  // boundary cases are also catched correctly.
  if (new_lview.frustrum)
  {
    // First calculate the center point of the frustrum. We will
    // expand around that point.
    Vector3 minf, maxf, cent;
    minf.x = minf.y = minf.z = 1000000;
    maxf.x = maxf.y = maxf.z = -1000000;
    for (i = 0 ; i < lview.num_frustrum ; i++)
    {
      if (lview.frustrum[i].x < minf.x) minf.x = lview.frustrum[i].x;
      if (lview.frustrum[i].y < minf.y) minf.y = lview.frustrum[i].y;
      if (lview.frustrum[i].z < minf.z) minf.z = lview.frustrum[i].z;
      if (lview.frustrum[i].x > maxf.x) maxf.x = lview.frustrum[i].x;
      if (lview.frustrum[i].y > maxf.y) maxf.y = lview.frustrum[i].y;
      if (lview.frustrum[i].z > maxf.z) maxf.z = lview.frustrum[i].z;
    }
    cent = (minf+maxf) / 2.0;

    // Scale the frustrum.
    for (i = 0 ; i < lview.num_frustrum ; i++)
      new_lview.frustrum[i] = (lview.frustrum[i]-cent) * 1.15 + cent;
  }

  if (bsp)
    bsp->back2front (new_lview.center, &shine_polygons, (void*)&new_lview);
  else
    shine_polygons ((PolygonParentInt*)this, polygons, num_polygon, (void*)&new_lview);

  if (static_thing) static_thing->shine_lightmaps (lview);
  Thing* sp = first_thing;
  while (sp)
  {
    if (!sp->isMerged ()) sp->shine_lightmaps (lview);
    sp = (Thing*)(sp->get_next ());
  }

  restore_transformation (old);
  draw_busy--;
}

void* csSector::dump_frustrum_polygons (PolygonParentInt*, PolygonInt** polygon, int num, void* data)
{
  Polygon3D* p;
  Portal* po;
  SectorShineInfo* d = (SectorShineInfo*)data;
  int i;
  for (i = 0 ; i < num ; i++)
  {
    p = (Polygon3D*)polygon[i];
    Vector3* new_frustrum = NULL;
    int new_num_frustrum = 0;
    if (p->clip_frustrum (d->light->get_center (), d->frustrum, d->num_frustrum, false/*@@@unsupported*/,
    	&new_frustrum, &new_num_frustrum))
    {
      int white = System->world->get_textures ()->white ();
      int red = System->world->get_textures ()->red ();
      int j;
      Vector3 light_cam;
      Vector3 v0, v1, v2;
      light_cam = d->trans->world2this( d->light->get_center () );

      for (j = 0 ; j < new_num_frustrum ; j++)
      {
        v0 = new_frustrum[j] + d->light->get_center ();
        v1 = d->trans->world2this(v0);
        v0 = new_frustrum[(j+1)%new_num_frustrum] + d->light->get_center ();
        v2 = d->trans->world2this(v0);
        Gfx3D->DrawLine (light_cam, v1, red);
        Gfx3D->DrawLine (light_cam, v2, red);
        Gfx3D->DrawLine (v1, v2, white);
      }

      po = p->get_portal ();
      if (po)
      {
        po->dump_frustrum (d->light, new_frustrum, new_num_frustrum, *(d->trans));
      }
      if (new_frustrum) CHKB (delete [] new_frustrum);
    }
  }
  return NULL;
}

void csSector::dump_frustrum (csStatLight* l, Vector3* frustrum, int num_frustrum,
	Transform& t)
{
  if (draw_busy) return;

  draw_busy++;
  Vector3* old = new_transformation ();

  translate_vector (l->get_center ());

  SectorShineInfo sh;
  sh.light = l;
  sh.frustrum = frustrum;
  sh.num_frustrum = num_frustrum;
  sh.trans = &t;

  if (bsp)
    bsp->back2front (l->get_center (), &dump_frustrum_polygons, (void*)&sh);
  else
    dump_frustrum_polygons ((PolygonParentInt*)this, polygons, num_polygon, (void*)&sh);

  Thing* sp = first_thing;
  while (sp)
  {
    sp->dump_frustrum (l, frustrum, num_frustrum, t);
    sp = (Thing*)(sp->get_next ());
  }

  restore_transformation (old);
  draw_busy--;
}

void csSector::init_lightmaps (bool do_cache)
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
    ((Polygon3D*)polygons[i])->init_lightmaps (do_cache);

  Thing* sp = first_thing;
  while (sp)
  {
    sp->init_lightmaps (do_cache);
    sp = (Thing*)(sp->get_next ());
  }
}

void csSector::cache_lightmaps ()
{
  int i;
  for (i = 0 ; i < num_polygon ; i++)
    ((Polygon3D*)polygons[i])->cache_lightmaps ();

  Thing* sp = first_thing;
  while (sp)
  {
    sp->cache_lightmaps ();
    sp = (Thing*)(sp->get_next ());
  }
}

Thing* csSector::get_thing (char* name)
{
  Thing* s = first_thing;
  while (s)
  {
    if (!strcmp (name, s->get_name ())) return s;
    s = (Thing*)(s->get_next ());
  }
  return NULL;
}

//---------------------------------------------------------------------------
// Everything needed for the precomputation of the lighting.
//---------------------------------------------------------------------------

struct BeamInfo
{
  Vector3 start;
  Vector3 end;
  Polygon3D* poly;
  float sqdist;
};

void* csSector::beam_polygons (PolygonParentInt*, PolygonInt** polygon, int num, void* data)
{
  BeamInfo* d = (BeamInfo*)data;
  int i;
  void* rc;
  Portal* po;
  Polygon3D* p;

  for (i = 0 ; i < num ; i++)
  {
    p = (Polygon3D*)polygon[i];
    if (p->intersect_ray (d->start, d->end))
    {
      po = p->get_portal ();
      rc = NULL;
      if (po && p != d->poly)
        return po->follow_beam (d->start, d->end, d->poly, &d->sqdist);
      else
      {
        Vector3 isect;
	p->get_plane ()->intersect_segment (d->start, d->end, isect, NULL);
        d->sqdist = Vector3::sqdist (isect, d->start);
        return (void*)p;
      }
    }
  }
  return NULL;
}

void* csSector::check_for_things (PolygonParentInt*, PolygonInt** polygon, int num, void* data)
{
  BeamInfo* d = (BeamInfo*)data;
  int i;
  Portal* po;
  Polygon3D* p;
  bool block = false;
  Polygon3D* block_poly;
  bool noblock = false;

  for (i = 0 ; i < num ; i++)
  {
    p = (Polygon3D*)polygon[i];
    void* rc = NULL;
    bool isc = p->intersect_ray (d->start, d->end);
    if (isc)
    {
      po = p->get_portal ();
      if (po)
      {
         rc = (void*)po->blocking_things (d->start, d->end, &(d->poly));
         if (rc) return rc;
         // Set noblock to true to indicate that there is no block (no
         // matter what 'block' says).
         noblock = true;
      }
      else
      {
        // There is an intersection but it is not with a portal.
        // Remember this for later. If we find no other portals to go through
        // we will return true that there is something blocking our path.
        block = true;
	block_poly = p;
      }
    }
  }
return NULL;//@@@

  if (noblock) return NULL;
  else if (block)
  {
    d->poly = block_poly;
    return (void*)true;
  }
  return NULL;
}


Polygon3D* csSector::follow_beam (Vector3& start, Vector3& end, Polygon3D* poly,
				float* sqdist)
{
  BeamInfo beam;
  beam.start = start;
  beam.end = end;
  beam.poly = poly;

  beam_busy++;
  Polygon3D* p;
  visited++;
  if (bsp) p = (Polygon3D*)bsp->front2back (start, &beam_polygons, (void*)&beam);
  else p = (Polygon3D*)beam_polygons ((PolygonParentInt*)this, polygons, num_polygon, (void*)&beam);
  if (!p) visited--;
  beam_busy--;

  // Check all the things.
  Polygon3D* p2;
  float sqdist2;
  Vector3 isect2;
  Thing* sp = first_thing;
  while (sp)
  {
    p2 = sp->intersect_segment (start, end, isect2);
    if (p2)
    {
      sqdist2 = Vector3::sqdist (isect2, start);
      if (!p || sqdist2 < beam.sqdist)
      {
        p = p2;
	beam.sqdist = sqdist2;
      }
    }
    sp = (Thing*)(sp->get_next ());
  }

  if (sqdist) *sqdist = beam.sqdist;
  return p;
}


// hit_beam is called by PolyTexture::shine with 'start' equal to
// the original world space start of the light and 'end' the corresponding
// world space coordinate of the end of the light beam. This 'end' point
// does not necessarily correspond with the point on the polygon.

bool csSector::hit_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* sqdist)
{
  // First we set visited to 0 in the sector of the polygon we are
  // looking for. 'follow_beam' will increase visited by 1 for all sectors
  // that the beam passed to get to the specific polygon.
  poly->get_sector ()->visited = 0;

  Polygon3D* p = follow_beam (start, end, poly, sqdist);

  if (!p) return false;

  if (p->same_plane (poly))
  {
    // If we have the same plane then we have a hit.
    return true;
  }

  // exp_sqdist is what we would expect the distance to our polygon
  // to be if we have a hit.
  float exp_sqdist = Vector3::sqdist (start, end);
  
  if (!poly->get_sector ()->bsp && poly->get_sector () == p->get_sector () &&
  	(PolygonSet*)(p->get_parent ()) == (PolygonSet*)(p->get_sector ()))
  {
    // If the polygon that is hit is a member of the same sector as
    // our sector, and if the sector does NOT use a bsp (it is convex),
    // and if the polygon that is hit does not belong to a Thing, then
    // we also consider it a hit. This takes care of misses where the
    // miss is just on an adjacent polygon.
    *sqdist = exp_sqdist;
    return true;
  }

  if (*sqdist > exp_sqdist && poly->get_sector ()->visited)
  {
    // We have missed this polygon and got a hit on some polygon that is
    // further away than this one. Since the original purpose of
    // this routine is to aim a beam of light at the polygon, we can't
    // really miss. So we assume it is a hit (easy, isn't it :-)
    // To make sure this test really works in all cases we also check
    // that the portal of our polygon is at least visited by the
    // beam.
    *sqdist = exp_sqdist;
    return true;
  }

  // The only case that is not fixed is similar to the last case but
  // for sectors that do use a BSP. Here we can't just use the same test.

  return false;
}

bool csSector::blocking_things (Vector3& start, Vector3& end, Polygon3D** poly, bool same_sector)
{
  // Check all the things in the current sector.
  Vector3 isect2;
  Polygon3D* p;
  Thing* sp = first_thing;
  while (sp)
  {
    p = sp->intersect_segment (start, end, isect2);
    if (p) { if (poly) *poly = p; return true; }
    sp = (Thing*)(sp->get_next ());
  }

  // If no things are blocking the path then check in the other sectors.
  BeamInfo beam;
  beam.start = start;
  beam.end = end;
  void* rc = NULL;

  beam_busy++;

  // Verify here if the maximum cfg_reflections have been reached to avoid
  // to loose time to find the portal from which the light is coming
  // for nothing.
  if (beam_busy < cfg_reflections)
  {
    if (bsp) rc = bsp->front2back (start, &check_for_things, (void*)&beam);
    else if (!same_sector)
      rc = check_for_things ((PolygonParentInt*)this, polygons, num_polygon, (void*)&beam);
  }

  beam_busy--;

  if (rc) { if (poly) *poly = beam.poly; return true; }

  return false;
}


void csSector::shine_lights ()
{
  int i;
  for (i = 0 ; i < lights.Length () ; i++)
    ((csStatLight*)lights[i])->shine_lightmaps ();
}

csStatLight* csSector::find_light (float x, float y, float z, float dist)
{
  int i;
  for (i = 0 ; i < lights.Length () ; i++)
  {
    csStatLight* l = (csStatLight*)lights[i];
    if (ABS (x-l->get_center ().x) < SMALL_EPSILON &&
    	ABS (y-l->get_center ().y) < SMALL_EPSILON &&
    	ABS (z-l->get_center ().z) < SMALL_EPSILON &&
    	ABS (dist-l->get_dist ()) < SMALL_EPSILON)
      return l;
  }
  return NULL;
}

void csSector::unlink_mso (SectorMSO* lp)
{
  if (lp->next_sector) lp->next_sector->prev_sector = lp->prev_sector;
  if (lp->prev_sector) lp->prev_sector->next_sector = lp->next_sector;
  else first_mso = lp->next_sector;
  lp->prev_sector = lp->next_sector = NULL;
  lp->sector = NULL;
}

void csSector::add_mso (SectorMSO* lp)
{
  lp->next_sector = first_mso;
  lp->prev_sector = NULL;
  if (first_mso) first_mso->prev_sector = lp;
  first_mso = lp;
  lp->sector = this;
}

//---------------------------------------------------------------------------
