/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SECTOR_H
#define SECTOR_H

#include "util/def.h"
#include "engine/basic/csobjvec.h"
#include "engine/basic/polyset.h"
#include "geometry/math3d.h"
#include "engine/rview.h"

struct LightInfo;
class Thing;
class csStatLight;
class Polygon3D;
class Polygon2D;
class csCamera;
class csDynLight;
class Portal;
class csSprite3D;
class SectorMSO;
class Dumper;
class BspTree;
#if USE_OCCLUSION
class Occlusion;
#endif

/**
 * A sector is a convex hull of polygons. It is
 * one of the base classes for the portal engine.
 */
class csSector : public PolygonSet
{
  friend class Dumper;

private:
  /// First thing in this sector.
  Thing* first_thing;

  /// First MultiSectorObject in this sector.
  SectorMSO* first_mso;

#if USE_OCCLUSION
  /// First occlusion object in this sector.
  Occlusion* first_occlusion;
#endif
  /// All static lights in this sector.
  csObjVector lights;

  /// Ambient light level for red in this sector.
  int level_r;
  /// Ambient light level for green in this sector.
  int level_g;
  /// Ambient light level for blue in this sector.
  int level_b;

  ///
  int visited;

  /**
   * If this variable is not NULL then it is a BSP tree in this
   * sector which includes all static (non-moving) Things.
   */
  BspTree* static_bsp;

  /**
   * If static_bsp is not NULL then this is a pointer to the Thing
   * which holds all polygons of the non-moving Things.
   */
  Thing* static_thing;

  /**
   * This function is called by the BSP tree traversal routine
   * to draw a number of polygons.
   */
  static void* draw_polygons (PolygonParentInt*, PolygonInt** polygon,
  	int num, void* data);

  /**
   * This function is called by the BSP tree traversal routine
   * to update the lightmaps for a number of polygons.
   */
  static void* shine_polygons (PolygonParentInt*, PolygonInt** polygon,
  	int num, void* data);

  /**
   * This function is called by the BSP tree traversal routine
   * to dump the frustrum (debugging).
   */
  static void* dump_frustrum_polygons (PolygonParentInt*, PolygonInt** polygon,
  	int num, void* data);

public:
  /// Option variable: render portals?
  static bool do_portals;
  /// Option variable: render things?
  static bool do_things;
  /// Configuration variable: number of allowed reflections for static lighting.
  static int cfg_reflections;
  /// Option variable: do pseudo-radiosity?
  static bool do_radiosity;

  /// How many times are we shining a specific light through this sector.
  int beam_busy;

  ///
  csSector (char* name);
  ///
  virtual ~csSector ();

  /// Add a thing to this sector.
  void add_thing (Thing* thing);
  /// Add a sprite to this sector.
  void add_sprite (csWorld* world, csSprite3D* sprite);
#if USE_OCCLUSION
  /// Add a occlusion to this sector.
  void add_occlusion (Occlusion* occlusion);
#endif
  /// Add a light to this sector.
  void add_light (csStatLight* light);
  ///
  csStatLight* get_light (int nr) { return (csStatLight*)lights[nr]; }
  ///
  int get_num_lights () { return lights.Length (); }
  /// Find a light with the given position and radius.
  csStatLight* find_light (float x, float y, float z, float dist);

  /// Find a thing with the given name.
  Thing* get_thing (char* name);
  ///
  Thing* get_first_thing() { return first_thing; }

  /**
   * Call this function to generate a BSP tree for all Things
   * in this sector.
   */
  void use_static_bsp ();

  ///
  void get_ambient_color (int& r, int& g, int& b) { r = level_r; g = level_g; b = level_b; }
  ///
  void set_ambient_color (int r, int g, int b) { level_r = r; level_g = g; level_b = b; }

  ///
  Polygon3D* hit_beam (Vector3& start, Vector3& end);

  ///
  void mipmap_settings (int setting);

  /**
   * Draw the sector in the given view and with the given transformation.
   */
  void draw (RenderView& rview);

  /**
   * Init the lightmaps for all polygons in this sector.
   * If do_cache == false this function will not try to read from
   * the cache.
   */
  void init_lightmaps (bool do_cache = true);

  /**
   * Update the lightmaps for all polygons in this sector and possibly
   * traverse through portals to other sectors.
   * If lview.frustrum == NULL it is equivalent to full view (everything is
   * visible).
   */
  void shine_lightmaps (LightView& lview);

  /**
   * This is a debugging function that will show the outlines
   * on all polygons that are hit by a light in this sector.
   * It will draw perspective correct outlines so it is meant to
   * be called with a camera transformation.
   */
  void dump_frustrum (csStatLight* l, Vector3* frustrum, int num_frustrum,
  	Transform& t);

  /**
   * Cache the lightmaps for all polygons in this sector.
   */
  void cache_lightmaps ();

  /**
   * Intersects world-space sphere with polygons of this set. Return
   * polygon it hits with (or NULL) and the intersection point
   * in world coordinates. It will also return the polygon with the
   * closest hit (the most nearby polygon).
   * If 'pr' != NULL it will also return the distance where the
   * intersection happened.
   * Note. This function correctly accounts for portal polygons
   * and could thus return a polygon not belonging to this sector.
   */
  Polygon3D* intersect_sphere (Vector3& center, float radius, float* pr = NULL);

  /**
   * Follow a segment starting at this sector. If the segment intersects
   * with a polygon it will stop there unless the polygon is a portal in which
   * case it will recursively go to that sector (possibly applying warping
   * transformations) and continue there.<p>
   *
   * This routine will modify all the given parameters to reflect space warping.
   * These should be used as the new camera transformation when you decide to
   * really go to the new position.<p>
   *
   * This function returns the resulting sector and new_position will be set
   * to the last position that you can go to before hitting a wall.
   */
  csSector* follow_segment (ReversibleTransform& t, Vector3& new_position, 
                          bool& mirror);

  /**
   * Intersect world-space segment with polygons of this sector. Return
   * polygon it intersects with (or NULL) and the intersection point
   * in world coordinates.<p>
   *
   * If 'pr' != NULL it will also return a value between 0 and 1
   * indicating where on the 'start'-'end' vector the intersection
   * happened.<p>
   *
   * This function is an extension of PolygonSet::intersect_segment in
   * that it will also test for hits against things.
   */
  virtual Polygon3D* intersect_segment (const Vector3& start,
                                        const Vector3& end, Vector3& isect,
					float* pr = NULL);

  //------------------------------------------------
  // Everything for setting up the lighting system.
  //------------------------------------------------

  /**
   * The whole setup starts with csWorld::shine_lights calling
   * csSector::shine_lights for every sector in the world.
   * This function will call csStatLight::shine_lightmaps for every
   * light in the sector.
   * csStatLight::shine_light will generate a view frustrum from the
   * center of the light and use that to light all polygons that
   * are hit by the frustrum.
   */
  void shine_lights ();

  /**
   * Follow a beam of light starting in this sector at 'start' and
   * ending on the given 'poly' at 'end'. Return true if the polygon
   * is actually reached and also return the squared distance in 'sqdist'
   * where the beam of light hits with the polygon.
   */
  bool hit_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* sqdist);

  /**
   * Check for Things along the way. This function assumes that there
   * is valid path from 'start' to 'end' through portals (and it will
   * follow that path). It will return false if there are no Things
   * blocking the path, otherwise it will return true and put the first polygon
   * that is found to be blocking the beam in 'poly' (only Polygons of Things
   * are considered).<br>
   * This function also returns true (but with 'poly' set to NULL) if it
   * could not go through a portal before it reached 'end' (there was
   * an error in the path somewhere).
   * If same_sector is true we know that we are looking for a hit in the
   * same sector only.
   */
  bool blocking_things (Vector3& start, Vector3& end, Polygon3D** poly, bool same_sector = false);

  /**
   * Follow a beam of light with a given start and end point and return
   * the first polygon that is hit. Also return the squared distance of the
   * path that was followed to get at the polygon (this correctly takes care
   * of mirroring portals and other warping portals).
   */
  Polygon3D* follow_beam (Vector3& start, Vector3& end, Polygon3D* poly, float* sqdist);

  /**
   * Unlink a SectorMSO from this sector.
   * Warning! This function does not test if the object
   * is really on the list!
   */
  void unlink_mso (SectorMSO* lp);

  /**
   * Add a SectorMSO to the list.
   */
  void add_mso (SectorMSO *lp);

  /**
   * Return the first Multi Sector Object for this sector.
   */
  SectorMSO* get_first_mso () { return first_mso; }

private:
  /**
   * BSP function used by follow_beam which does the actual work.
   */
  static void* beam_polygons (PolygonParentInt*, PolygonInt** polygon,
  	int num, void* data);

  /**
   * BSP function used by blocking_things which does the actual work.
   */
  static void* check_for_things (PolygonParentInt*, PolygonInt** polygon,
  	int num, void* data);
};

#endif /*SECTOR_H*/
