/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <time.h>

#include "system/system.h"
#include "system/sysprint.h"
#include "util/util.h"
#include "engine/texture.h"
#include "engine/polygon/polytext.h"
#include "engine/light/light.h"
#include "util/token.h"
#include "util/parser.h"
#include "util/inifile.h"
#include "engine/polygon/polygon.h"
#include "util/archive.h"
#include "engine/world.h"
#include "util/gfx/csimage.h"
#include "render/igraph2d.h"
#include "render/igraph3d.h"
#include "general/loader/csloader.h"
#include "engine/inv_cmap.h"

#define RESERVED_COLOR(c) ((c == 0) || (c == 255))

// Texture gamma
float Textures::Gamma = 1.0;
// Ignore image file loading errors?
bool TextureMM::fIgnoreLoadingErrors = false;

//---------------------------------------------------------------------------

TxtCmapPrivate::TxtCmapPrivate ()
{
  int i;
  for (i = 0 ; i < 256 ; i++)
  {
    alloc[i] = false;
    rgb_values[(i<<2)+0] = 0;
    rgb_values[(i<<2)+1] = 0;
    rgb_values[(i<<2)+2] = 0;
    rgb_values[(i<<2)+3] = 0;
    priv_to_global[i] = 0;
  }
}

int TxtCmapPrivate::find_rgb (int r, int g, int b)
{
  if (r>255) r=255; else if (r<0) r=0;
  if (g>255) g=255; else if (g<0) g=0;
  if (b>255) b=255; else if (b<0) b=0;

  int i, min, mindist;
  mindist = 1000L*256*256;
  min = -1;
  register int red, green, blue, dist;
  for (i = 1 ; i < 256 ; i++) // Color 0 is not used, it is reserved for transparency
    if (alloc[i])
    {
      red = r - rgb_values[(i<<2)+0];
      green = g - rgb_values[(i<<2)+1];
      blue = b - rgb_values[(i<<2)+2];
      dist = (299*red*red) + (587*green*green) + (114*blue*blue);
      if (dist == 0) return i;
      if (dist < mindist) { mindist = dist; min = i; }
    }
  return min;
}

int TxtCmapPrivate::alloc_rgb (int r, int g, int b, int dist)
{
  int d, j;
  if (r < 0) r = 0; else if (r > 255) r = 255;
  if (g < 0) g = 0; else if (g > 255) g = 255;
  if (b < 0) b = 0; else if (b > 255) b = 255;

  d = 0;
  int i = find_rgb (r, g, b);
  if (i != -1)
  {
    d = 299*(r-rgb_values[(i<<2)+0])*(r-rgb_values[(i<<2)+0])+
      587*(g-rgb_values[(i<<2)+1])*(g-rgb_values[(i<<2)+1])+
      114*(b-rgb_values[(i<<2)+2])*(b-rgb_values[(i<<2)+2]);
  }
  if (i == -1 || d > dist)
  {
    for (j = 1 ; j < 256 ; j++) // Color 0 is not used, reserved for transparency
      if (!alloc[j])
      {
        alloc[j] = true;
        rgb_values[(j<<2)+0] = r;
        rgb_values[(j<<2)+1] = g;
        rgb_values[(j<<2)+2] = b;
        return j;
      }
    return i; // We couldn't allocate a new color, return best fit
  }
  else return i;
}

//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( TextureMM )

BEGIN_INTERFACE_TABLE( TextureMM )
        IMPLEMENTS_COMPOSITE_INTERFACE( MipMapContainer )
END_INTERFACE_TABLE()


TextureMM::TextureMM (char* name, ImageFile* image) : csObject (name, CS_TEXTURE)
{
  t1 = t2 = t3 = t4 = NULL;
  priv_cmap = NULL;

  usage = NULL;
  transparent = -1;
  ifile = NULL;
  hicolorcache=NULL;
  in_memory = false;

  ifile = image;
}

TextureMM::~TextureMM ()
{
  if (ifile) CHKB (delete ifile);

  if (t1) CHKB (delete t1);
  if (t2) CHKB (delete t2);
  if (t3) CHKB (delete t3);
  if (t4) CHKB (delete t4);
  if (priv_cmap) CHKB (delete priv_cmap);

  if (usage) CHKB (delete usage);
}

int get_red_glob (void* pal, int idx)
{
  return ((RGBcolor*)pal)[idx].red;
}

int get_green_glob (void* pal, int idx)
{
  return ((RGBcolor*)pal)[idx].green;
}

int get_blue_glob (void* pal, int idx)
{
  return ((RGBcolor*)pal)[idx].blue;
}

int get_red_priv (void* pal, int idx)
{
  return ((unsigned char*)pal)[(idx<<2)+0];
}

int get_green_priv (void* pal, int idx)
{
  return ((unsigned char*)pal)[(idx<<2)+1];
}

int get_blue_priv (void* pal, int idx)
{
  return ((unsigned char*)pal)[(idx<<2)+2];
}

void box_filter3 (Filter3x3& f, unsigned char* bm, int x, int y, int w, int h, void* pal,
        int (*fr)(void* pal, int idx),
        int (*fg)(void* pal, int idx),
        int (*fb)(void* pal, int idx),
        int* pr, int* pg, int* pb)
{
  int m11, m12, m13;
  int m21, m22, m23;
  int m31, m32, m33;
  int r, g, b;
  m11 = ((x-1+w)%w) + ((y-1+h)%h) * w;  m11 = bm[m11];
  m12 = x           + ((y-1+h)%h) * w;  m12 = bm[m12];
  m13 = ((x+1)%w)   + ((y-1+h)%h) * w;  m13 = bm[m13];
  m21 = ((x-1+w)%w) + y * w;            m21 = bm[m21];
  m22 = x           + y * w;            m22 = bm[m22];
  m23 = ((x+1)%w)   + y * w;            m23 = bm[m23];
  m31 = ((x-1+w)%w) + ((y+1)%h) * w;    m31 = bm[m31];
  m32 = x           + ((y+1)%h) * w;    m32 = bm[m32];
  m33 = ((x+1)%w)   + ((y+1)%h) * w;    m33 = bm[m33];
  r = f.f11*fr (pal, m11) + f.f12*fr (pal, m12) + f.f13*fr (pal, m13) +
      f.f21*fr (pal, m21) + f.f22*fr (pal, m22) + f.f23*fr (pal, m23) +
      f.f31*fr (pal, m31) + f.f32*fr (pal, m32) + f.f33*fr (pal, m33);
  g = f.f11*fg (pal, m11) + f.f12*fg (pal, m12) + f.f13*fg (pal, m13) +
      f.f21*fg (pal, m21) + f.f22*fg (pal, m22) + f.f23*fg (pal, m23) +
      f.f31*fg (pal, m31) + f.f32*fg (pal, m32) + f.f33*fg (pal, m33);
  b = f.f11*fb (pal, m11) + f.f12*fb (pal, m12) + f.f13*fb (pal, m13) +
      f.f21*fb (pal, m21) + f.f22*fb (pal, m22) + f.f23*fb (pal, m23) +
      f.f31*fb (pal, m31) + f.f32*fb (pal, m32) + f.f33*fb (pal, m33);
  *pr = r/f.tot;
  *pg = g/f.tot;
  *pb = b/f.tot;
}

// Version which works for RGB bitmaps.
void box_filter3 (Filter3x3& f, ULong* bm, int x, int y, int w, int h, int* pr, int* pg, int* pb)
{
  ULong m11, m12, m13;
  ULong m21, m22, m23;
  ULong m31, m32, m33;
  int r, g, b;
  m11 = ((x-1+w)%w) + ((y-1+h)%h) * w;  m11 = bm[m11];
  m12 = x           + ((y-1+h)%h) * w;  m12 = bm[m12];
  m13 = ((x+1)%w)   + ((y-1+h)%h) * w;  m13 = bm[m13];
  m21 = ((x-1+w)%w) + y * w;            m21 = bm[m21];
  m22 = x           + y * w;            m22 = bm[m22];
  m23 = ((x+1)%w)   + y * w;            m23 = bm[m23];
  m31 = ((x-1+w)%w) + ((y+1)%h) * w;    m31 = bm[m31];
  m32 = x           + ((y+1)%h) * w;    m32 = bm[m32];
  m33 = ((x+1)%w)   + ((y+1)%h) * w;    m33 = bm[m33];
  r = f.f11*R24 (m11) + f.f12*R24 (m12) + f.f13*R24 (m13) +
      f.f21*R24 (m21) + f.f22*R24 (m22) + f.f23*R24 (m23) +
      f.f31*R24 (m31) + f.f32*R24 (m32) + f.f33*R24 (m33);
  g = f.f11*G24 (m11) + f.f12*G24 (m12) + f.f13*G24 (m13) +
      f.f21*G24 (m21) + f.f22*G24 (m22) + f.f23*G24 (m23) +
      f.f31*G24 (m31) + f.f32*G24 (m32) + f.f33*G24 (m33);
  b = f.f11*B24 (m11) + f.f12*B24 (m12) + f.f13*B24 (m13) +
      f.f21*B24 (m21) + f.f22*B24 (m22) + f.f23*B24 (m23) +
      f.f31*B24 (m31) + f.f32*B24 (m32) + f.f33*B24 (m33);
  *pr = r/f.tot;
  *pg = g/f.tot;
  *pb = b/f.tot;
}

void box_filter5 (Filter5x5& f, unsigned char* bm, int x, int y, int w, int h, void* p,
        int (*fr)(void* pal, int idx),
        int (*fg)(void* pal, int idx),
        int (*fb)(void* pal, int idx),
        int* pr, int* pg, int* pb)
{
  int m00, m01, m02, m03, m04;
  int m10, m11, m12, m13, m14;
  int m20, m21, m22, m23, m24;
  int m30, m31, m32, m33, m34;
  int m40, m41, m42, m43, m44;
  int r, g, b;
  m00 = ((x-2+w)%w) + ((y-2+h)%h) * w;  m00 = bm[m00];
  m01 = ((x-1+w)%w) + ((y-2+h)%h) * w;  m01 = bm[m01];
  m02 = x           + ((y-2+h)%h) * w;  m02 = bm[m02];
  m03 = ((x+1)%w)   + ((y-2+h)%h) * w;  m03 = bm[m03];
  m04 = ((x+2)%w)   + ((y-2+h)%h) * w;  m04 = bm[m04];
  m10 = ((x-2+w)%w) + ((y-1+h)%h) * w;  m10 = bm[m10];
  m11 = ((x-1+w)%w) + ((y-1+h)%h) * w;  m11 = bm[m11];
  m12 = x           + ((y-1+h)%h) * w;  m12 = bm[m12];
  m13 = ((x+1)%w)   + ((y-1+h)%h) * w;  m13 = bm[m13];
  m14 = ((x+2)%w)   + ((y-1+h)%h) * w;  m14 = bm[m14];
  m20 = ((x-2+w)%w) + y * w;            m20 = bm[m20];
  m21 = ((x-1+w)%w) + y * w;            m21 = bm[m21];
  m22 = x           + y * w;            m22 = bm[m22];
  m23 = ((x+1)%w)   + y * w;            m23 = bm[m23];
  m24 = ((x+2)%w)   + y * w;            m24 = bm[m24];
  m30 = ((x-2+w)%w) + ((y+1)%h) * w;    m30 = bm[m30];
  m31 = ((x-1+w)%w) + ((y+1)%h) * w;    m31 = bm[m31];
  m32 = x           + ((y+1)%h) * w;    m32 = bm[m32];
  m33 = ((x+1)%w)   + ((y+1)%h) * w;    m33 = bm[m33];
  m34 = ((x+2)%w)   + ((y+1)%h) * w;    m34 = bm[m34];
  m40 = ((x-2+w)%w) + ((y+2)%h) * w;    m40 = bm[m40];
  m41 = ((x-1+w)%w) + ((y+2)%h) * w;    m41 = bm[m41];
  m42 = x           + ((y+2)%h) * w;    m42 = bm[m42];
  m43 = ((x+1)%w)   + ((y+2)%h) * w;    m43 = bm[m43];
  m44 = ((x+2)%w)   + ((y+2)%h) * w;    m44 = bm[m44];
  r = f.f00*fr (p, m00) + f.f01*fr (p, m01) + f.f02*fr (p, m02) + f.f03*fr (p, m03) + f.f04*fr (p, m04) +
      f.f10*fr (p, m10) + f.f11*fr (p, m11) + f.f12*fr (p, m12) + f.f13*fr (p, m13) + f.f14*fr (p, m14) +
      f.f20*fr (p, m20) + f.f21*fr (p, m21) + f.f22*fr (p, m22) + f.f23*fr (p, m23) + f.f24*fr (p, m24) +
      f.f30*fr (p, m30) + f.f31*fr (p, m31) + f.f32*fr (p, m32) + f.f33*fr (p, m33) + f.f34*fr (p, m34) +
      f.f40*fr (p, m40) + f.f41*fr (p, m41) + f.f42*fr (p, m42) + f.f43*fr (p, m43) + f.f44*fr (p, m44);
  g = f.f00*fg (p, m00) + f.f01*fg (p, m01) + f.f02*fg (p, m02) + f.f03*fg (p, m03) + f.f04*fg (p, m04) +
      f.f10*fg (p, m10) + f.f11*fg (p, m11) + f.f12*fg (p, m12) + f.f13*fg (p, m13) + f.f14*fg (p, m14) +
      f.f20*fg (p, m20) + f.f21*fg (p, m21) + f.f22*fg (p, m22) + f.f23*fg (p, m23) + f.f24*fg (p, m24) +
      f.f30*fg (p, m30) + f.f31*fg (p, m31) + f.f32*fg (p, m32) + f.f33*fg (p, m33) + f.f34*fg (p, m34) +
      f.f40*fg (p, m40) + f.f41*fg (p, m41) + f.f42*fg (p, m42) + f.f43*fg (p, m43) + f.f44*fg (p, m44);
  b = f.f00*fb (p, m00) + f.f01*fb (p, m01) + f.f02*fb (p, m02) + f.f03*fb (p, m03) + f.f04*fb (p, m04) +
      f.f10*fb (p, m10) + f.f11*fb (p, m11) + f.f12*fb (p, m12) + f.f13*fb (p, m13) + f.f14*fb (p, m14) +
      f.f20*fb (p, m20) + f.f21*fb (p, m21) + f.f22*fb (p, m22) + f.f23*fb (p, m23) + f.f24*fb (p, m24) +
      f.f30*fb (p, m30) + f.f31*fb (p, m31) + f.f32*fb (p, m32) + f.f33*fb (p, m33) + f.f34*fb (p, m34) +
      f.f40*fb (p, m40) + f.f41*fb (p, m41) + f.f42*fb (p, m42) + f.f43*fb (p, m43) + f.f44*fb (p, m44);
  *pr = r/f.tot;
  *pg = g/f.tot;
  *pb = b/f.tot;
}

// Version which works for RGB bitmaps.
void box_filter5 (Filter5x5& f, ULong* bm, int x, int y, int w, int h, int* pr, int* pg, int* pb)
{
  ULong m00, m01, m02, m03, m04;
  ULong m10, m11, m12, m13, m14;
  ULong m20, m21, m22, m23, m24;
  ULong m30, m31, m32, m33, m34;
  ULong m40, m41, m42, m43, m44;
  int r, g, b;
  m00 = ((x-2+w)%w) + ((y-2+h)%h) * w;  m00 = bm[m00];
  m01 = ((x-1+w)%w) + ((y-2+h)%h) * w;  m01 = bm[m01];
  m02 = x           + ((y-2+h)%h) * w;  m02 = bm[m02];
  m03 = ((x+1)%w)   + ((y-2+h)%h) * w;  m03 = bm[m03];
  m04 = ((x+2)%w)   + ((y-2+h)%h) * w;  m04 = bm[m04];
  m10 = ((x-2+w)%w) + ((y-1+h)%h) * w;  m10 = bm[m10];
  m11 = ((x-1+w)%w) + ((y-1+h)%h) * w;  m11 = bm[m11];
  m12 = x           + ((y-1+h)%h) * w;  m12 = bm[m12];
  m13 = ((x+1)%w)   + ((y-1+h)%h) * w;  m13 = bm[m13];
  m14 = ((x+2)%w)   + ((y-1+h)%h) * w;  m14 = bm[m14];
  m20 = ((x-2+w)%w) + y * w;            m20 = bm[m20];
  m21 = ((x-1+w)%w) + y * w;            m21 = bm[m21];
  m22 = x           + y * w;            m22 = bm[m22];
  m23 = ((x+1)%w)   + y * w;            m23 = bm[m23];
  m24 = ((x+2)%w)   + y * w;            m24 = bm[m24];
  m30 = ((x-2+w)%w) + ((y+1)%h) * w;    m30 = bm[m30];
  m31 = ((x-1+w)%w) + ((y+1)%h) * w;    m31 = bm[m31];
  m32 = x           + ((y+1)%h) * w;    m32 = bm[m32];
  m33 = ((x+1)%w)   + ((y+1)%h) * w;    m33 = bm[m33];
  m34 = ((x+2)%w)   + ((y+1)%h) * w;    m34 = bm[m34];
  m40 = ((x-2+w)%w) + ((y+2)%h) * w;    m40 = bm[m40];
  m41 = ((x-1+w)%w) + ((y+2)%h) * w;    m41 = bm[m41];
  m42 = x           + ((y+2)%h) * w;    m42 = bm[m42];
  m43 = ((x+1)%w)   + ((y+2)%h) * w;    m43 = bm[m43];
  m44 = ((x+2)%w)   + ((y+2)%h) * w;    m44 = bm[m44];
  r = f.f00*R24 (m00) + f.f01*R24 (m01) + f.f02*R24 (m02) + f.f03*R24 (m03) + f.f04*R24 (m04) +
      f.f10*R24 (m10) + f.f11*R24 (m11) + f.f12*R24 (m12) + f.f13*R24 (m13) + f.f14*R24 (m14) +
      f.f20*R24 (m20) + f.f21*R24 (m21) + f.f22*R24 (m22) + f.f23*R24 (m23) + f.f24*R24 (m24) +
      f.f30*R24 (m30) + f.f31*R24 (m31) + f.f32*R24 (m32) + f.f33*R24 (m33) + f.f34*R24 (m34) +
      f.f40*R24 (m40) + f.f41*R24 (m41) + f.f42*R24 (m42) + f.f43*R24 (m43) + f.f44*R24 (m44);
  g = f.f00*G24 (m00) + f.f01*G24 (m01) + f.f02*G24 (m02) + f.f03*G24 (m03) + f.f04*G24 (m04) +
      f.f10*G24 (m10) + f.f11*G24 (m11) + f.f12*G24 (m12) + f.f13*G24 (m13) + f.f14*G24 (m14) +
      f.f20*G24 (m20) + f.f21*G24 (m21) + f.f22*G24 (m22) + f.f23*G24 (m23) + f.f24*G24 (m24) +
      f.f30*G24 (m30) + f.f31*G24 (m31) + f.f32*G24 (m32) + f.f33*G24 (m33) + f.f34*G24 (m34) +
      f.f40*G24 (m40) + f.f41*G24 (m41) + f.f42*G24 (m42) + f.f43*G24 (m43) + f.f44*G24 (m44);
  b = f.f00*B24 (m00) + f.f01*B24 (m01) + f.f02*B24 (m02) + f.f03*B24 (m03) + f.f04*B24 (m04) +
      f.f10*B24 (m10) + f.f11*B24 (m11) + f.f12*B24 (m12) + f.f13*B24 (m13) + f.f14*B24 (m14) +
      f.f20*B24 (m20) + f.f21*B24 (m21) + f.f22*B24 (m22) + f.f23*B24 (m23) + f.f24*B24 (m24) +
      f.f30*B24 (m30) + f.f31*B24 (m31) + f.f32*B24 (m32) + f.f33*B24 (m33) + f.f34*B24 (m34) +
      f.f40*B24 (m40) + f.f41*B24 (m41) + f.f42*B24 (m42) + f.f43*B24 (m43) + f.f44*B24 (m44);
  *pr = r/f.tot;
  *pg = g/f.tot;
  *pb = b/f.tot;
}


// This function blends a mipmap from one bitmap into another
// bitmap (two mipmaps have the same size).
void TextureMM::create_blended_mipmap (Textures* tex, unsigned char* bm, int w, int h,
        unsigned char* bm2)
{
  if (!ifile)
    return;
  if (Textures::txtMode == TXT_GLOBAL)
    create_blended_mipmap_global (tex, bm, w, h, bm2);
  else if (Textures::txtMode == TXT_24BIT)
    create_blended_mipmap_24bit (tex, bm, w, h, bm2);
  else
    create_blended_mipmap_private (tex, bm, w, h, bm2);
}

void TextureMM::create_blended_mipmap_global (Textures* tex,
        unsigned char* bm, int w, int h,
        unsigned char* bm2)
{
  int x, y, r, g, b;
  RGBcolor* pal = tex->get_palette ();

  // If a texture is transparent we don't blend but just copy because
  // color 0 (transparent) needs to be preserved (maybe use a smarter
  // algorithm for this @@@).
  bool tran = get_transparent () != -1;
  if (tran)
  {
    for (y = 0 ; y < h ; y++)
      for (x = 0 ; x < w ; x++)
        *bm2++ = *bm++;
    return;
  }

  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      box_filter3 (Textures::blend_filter, bm, x, y, w, h, (void*)pal,
        get_red_glob, get_green_glob, get_blue_glob,
        &r, &g, &b);
      *bm2++ = tex->find_rgb (r, g, b);
    }
}

void TextureMM::create_blended_mipmap_private (Textures* /*tex*/, unsigned char* bm, int w, int h,
        unsigned char* bm2)
{
  int x, y, r, g, b;
  unsigned char* pal = priv_cmap->rgb_values;

  // If a texture is transparent we don't blend but just copy because
  // color 0 (transparent) needs to be preserved (maybe use a smarter
  // algorithm for this @@@).
  bool tran = get_transparent () != -1;
  if (tran)
  {
    for (y = 0 ; y < h ; y++)
      for (x = 0 ; x < w ; x++)
        *bm2++ = *bm++;
    return;
  }

  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      box_filter3 (Textures::blend_filter, bm, x, y, w, h, (void*)pal,
        get_red_priv, get_green_priv, get_blue_priv,
        &r, &g, &b);
      *bm2++ = priv_cmap->find_rgb (r, g, b);
    }
}

void TextureMM::create_blended_mipmap_24bit (Textures* /*tex*/, unsigned char* bm, int w, int h,
        unsigned char* bm2)
{
  int x, y, r, g, b;
  ULong* bml = (ULong*)bm;
  ULong* bml2 = (ULong*)bm2;

  // If a texture is transparent we don't blend but just copy because
  // color 0 (transparent) needs to be preserved (maybe use a smarter
  // algorithm for this @@@).
  bool tran = get_transparent () != -1;
  if (tran)
  {
    for (y = 0 ; y < h ; y++)
      for (x = 0 ; x < w ; x++)
        *bml2++ = *bml++;
    return;
  }

  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      box_filter3 (Textures::blend_filter, bml, x, y, w, h, &r, &g, &b);
      *bml2++ = (r<<16) | (g<<8) | b;
    }
}


// This function create a mipmap from one bitmap into another
// bitmap.
void TextureMM::create_mipmap_bitmap (Textures* tex,
        unsigned char* bm, int w, int h,        // Specs source bitmap
        int steps,                              // Nr of levels to mipmap
        unsigned char* bm2, int w2, int h2)     // Specs dest bitmap
{
  if (!ifile)
    return;
  if (Textures::txtMode == TXT_GLOBAL)
    create_mipmap_bitmap_global (tex, bm, w, h, steps, bm2, w2, h2);
  else if (Textures::txtMode == TXT_24BIT)
    create_mipmap_bitmap_24bit (tex, bm, w, h, steps, bm2, w2, h2);
  else
    create_mipmap_bitmap_private (tex, bm, w, h, steps, bm2, w2, h2);
}

void TextureMM::create_mipmap_bitmap_global (Textures* tex,
        unsigned char* bm, int w, int h,
        int steps,
        unsigned char* bm2, int w2, int h2)
{
  int x, y, r, g, b;
  RGBcolor* p = tex->get_palette ();

  // If a texture is transparent we only allow MIPMAP_UGLY because
  // the other mipmap settings possibly change the colors (averaging them).
  bool tran = get_transparent () != -1;
  if (tran || Textures::mipmap_nice == MIPMAP_UGLY)
  {
    if (steps == 1)
      for (y = 0 ; y < h2 ; y++)
      {
        for (x = 0 ; x < w2 ; x++)
        {
          *bm2++ = *bm;
          bm += 2;
        }
        bm += w;
      }
    else if (steps == 2)
      for (y = 0 ; y < h2 ; y++)
      {
        for (x = 0 ; x < w2 ; x++)
        {
          *bm2++ = *bm;
          bm += 4;
        }
        bm += w*3;
      }
  }
  else if (Textures::mipmap_nice == MIPMAP_DEFAULT)
  {
    if (steps == 1)
      for (y = 0 ; y < h2 ; y++)
      {
        for (x = 0 ; x < w2 ; x++)
        {
          // @@@ Consider a more accurate algorithm that shifts the source bitmap one
          // half pixel. In the current implementation there is a small shift in the
          // texture data.
          r = p[*bm].red + p[*(bm+1)].red + p[*(bm+w)].red + p[*(bm+w+1)].red;
          g = p[*bm].green + p[*(bm+1)].green + p[*(bm+w)].green + p[*(bm+w+1)].green;
          b = p[*bm].blue + p[*(bm+1)].blue + p[*(bm+w)].blue + p[*(bm+w+1)].blue;
          *bm2++ = tex->find_rgb (r/4, g/4, b/4);
          bm += 2;
        }
        bm += w;
      }
    else if (steps == 2)
      for (y = 0 ; y < h2 ; y++)
      {
        for (x = 0 ; x < w2 ; x++)
        {
          r = p[*bm].red         + p[*(bm+1)].red       + p[*(bm+2)].red       + p[*(bm+3)].red +
              p[*(bm+w)].red     + p[*(bm+w+1)].red     + p[*(bm+w+2)].red     + p[*(bm+w+3)].red +
              p[*(bm+w+w)].red   + p[*(bm+w+w+1)].red   + p[*(bm+w+w+2)].red   + p[*(bm+w+w+3)].red +
              p[*(bm+w*3)].red   + p[*(bm+w*3+1)].red   + p[*(bm+w*3+2)].red   + p[*(bm+w*3+3)].red;
          g = p[*bm].green         + p[*(bm+1)].green       + p[*(bm+2)].green       + p[*(bm+3)].green +
              p[*(bm+w)].green     + p[*(bm+w+1)].green     + p[*(bm+w+2)].green     + p[*(bm+w+3)].green +
              p[*(bm+w+w)].green   + p[*(bm+w+w+1)].green   + p[*(bm+w+w+2)].green   + p[*(bm+w+w+3)].green +
              p[*(bm+w*3)].green   + p[*(bm+w*3+1)].green   + p[*(bm+w*3+2)].green   + p[*(bm+w*3+3)].green;
          b = p[*bm].blue         + p[*(bm+1)].blue       + p[*(bm+2)].blue       + p[*(bm+3)].blue +
              p[*(bm+w)].blue     + p[*(bm+w+1)].blue     + p[*(bm+w+2)].blue     + p[*(bm+w+3)].blue +
              p[*(bm+w+w)].blue   + p[*(bm+w+w+1)].blue   + p[*(bm+w+w+2)].blue   + p[*(bm+w+w+3)].blue +
              p[*(bm+w*3)].blue   + p[*(bm+w*3+1)].blue   + p[*(bm+w*3+2)].blue   + p[*(bm+w*3+3)].blue;
          *bm2++ = tex->find_rgb (r/16, g/16, b/16);
          bm += 4;
        }
        bm += w*3;
      }
  }
  else
  {
    if (steps == 1)
      for (y = 0 ; y < h ; y += 2)
        for (x = 0 ; x < w ; x += 2)
        {
          box_filter3 (Textures::mipmap_filter_1, bm, x, y, w, h, (void*)p,
            get_red_glob, get_green_glob, get_blue_glob,
            &r, &g, &b);
          *bm2++ = tex->find_rgb (r, g, b);
        }
    else if (steps == 2)
      for (y = 0 ; y < h ; y += 4)
        for (x = 0 ; x < w ; x += 4)
        {
          box_filter5 (Textures::mipmap_filter_2, bm, x+1, y+1, w, h, (void*)p,
            get_red_glob, get_green_glob, get_blue_glob,
            &r, &g, &b);
          *bm2++ = tex->find_rgb (r, g, b);
        }
  }
}

void TextureMM::create_mipmap_bitmap_private (Textures* /*tex*/,
        unsigned char* bm, int w, int h,
        int steps,
        unsigned char* bm2, int w2, int h2)
{
  int x, y, r, g, b;
  unsigned char* pal = priv_cmap->rgb_values;

  // If a texture is transparent we only allow MIPMAP_UGLY because
  // the other mipmap settings possibly change the colors (averaging them).
  bool tran = get_transparent () != -1;
  if (tran || Textures::mipmap_nice == MIPMAP_UGLY)
    for (y = 0 ; y < h2 ; y++)
    {
      for (x = 0 ; x < w2 ; x++)
      {
        *bm2++ = *bm;
        bm += 2;
      }
      bm += w;
    }
  else if (Textures::mipmap_nice == MIPMAP_DEFAULT)
    for (y = 0 ; y < h2 ; y++)
    {
      for (x = 0 ; x < w2 ; x++)
      {
        int pi = (*bm)<<2;
        int pi_1 = (*(bm+1))<<2;
        int pi_w1 = (*(bm+w+1))<<2;
        int pi_w = (*(bm+w))<<2;
        // @@@ Consider a more accurate algorithm that shifts the source bitmap one
        // half pixel. In the current implementation there is a small shift in the
        // texture data.
        r = pal[pi+0] + pal[pi_1+0] + pal[pi_w+0] + pal[pi_w1+0];
        g = pal[pi+1] + pal[pi_1+1] + pal[pi_w+1] + pal[pi_w1+1];
        b = pal[pi+2] + pal[pi_1+2] + pal[pi_w+2] + pal[pi_w1+2];
        *bm2++ = priv_cmap->find_rgb (r/4, g/4, b/4);
        bm += 2;
      }
      bm += w;
    }
  else
  {
    if (steps == 1)
      for (y = 0 ; y < h ; y += 2)
        for (x = 0 ; x < w ; x += 2)
        {
          box_filter3 (Textures::mipmap_filter_1, bm, x, y, w, h, (void*)pal,
            get_red_priv, get_green_priv, get_blue_priv,
            &r, &g, &b);
          *bm2++ = priv_cmap->find_rgb (r, g, b);
        }
    else if (steps == 2)
      for (y = 0 ; y < h ; y += 4)
        for (x = 0 ; x < w ; x += 4)
        {
          box_filter5 (Textures::mipmap_filter_2, bm, x+1, y+1, w, h, (void*)pal,
            get_red_priv, get_green_priv, get_blue_priv,
            &r, &g, &b);
          *bm2++ = priv_cmap->find_rgb (r, g, b);
        }
  }
}

void TextureMM::create_mipmap_bitmap_24bit (Textures* /*tex*/,
        unsigned char* bm, int w, int h,
        int steps,
        unsigned char* bm2, int w2, int h2)
{
  int x, y, r, g, b;
  ULong* bml = (ULong*)bm;
  ULong* bml2 = (ULong*)bm2;

  // If a texture is transparent we only allow MIPMAP_UGLY because
  // the other mipmap settings possibly change the colors (averaging them).
  bool tran = get_transparent () != -1;
  if (tran || Textures::mipmap_nice == MIPMAP_UGLY)
    for (y = 0 ; y < h2 ; y++)
    {
      for (x = 0 ; x < w2 ; x++)
      {
        *bml2++ = *bml;
        bml += 2;
      }
      bml += w;
    }
  else if (Textures::mipmap_nice == MIPMAP_DEFAULT)
    for (y = 0 ; y < h2 ; y++)
    {
      for (x = 0 ; x < w2 ; x++)
      {
        ULong pi = *bml;
        ULong pi_1 = *(bml+1);
        ULong pi_w1 = *(bml+w+1);
        ULong pi_w = *(bml+w);
        // @@@ Consider a more accurate algorithm that shifts the source bitmap one
        // half pixel. In the current implementation there is a small shift in the
        // texture data.
        r = R24 (pi) + R24 (pi_1) + R24 (pi_w) + R24 (pi_w1);
        g = G24 (pi) + G24 (pi_1) + G24 (pi_w) + G24 (pi_w1);
        b = B24 (pi) + B24 (pi_1) + B24 (pi_w) + B24 (pi_w1);
        *bml2++ = ((r/4)<<16)|((g/4)<<8)|(b/4);
        bml += 2;
      }
      bml += w;
    }
  else
  {
    if (steps == 1)
      for (y = 0 ; y < h ; y += 2)
        for (x = 0 ; x < w ; x += 2)
        {
          box_filter3 (Textures::mipmap_filter_1, bml, x, y, w, h, &r, &g, &b);
          *bml2++ = (r<<16)|(g<<8)|b;
        }
    else if (steps == 2)
      for (y = 0 ; y < h ; y += 4)
        for (x = 0 ; x < w ; x += 4)
        {
          box_filter5 (Textures::mipmap_filter_2, bml, x+1, y+1, w, h, &r, &g, &b);
          *bml2++ = (r<<16)|(g<<8)|b;
        }
  }
}

void TextureMM::alloc_mipmaps ()
{
  if (!ifile)
    return;

  if (t1) { CHK (delete t1); t1 = NULL; }
  if (t2) { CHK (delete t2); t2 = NULL; }
  if (t3) { CHK (delete t3); t3 = NULL; }
  if (t4) { CHK (delete t4); t4 = NULL; }

  int w, h;
  w = ifile->get_width ();
  h = ifile->get_height ();

  TextureFactory8 tf8;
  TextureFactory32 tf32;
  TextureFactory* tfact;
  if (Textures::txtMode == TXT_24BIT) tfact = &tf32;
  else tfact = &tf8;

  t1 = tfact->new_texture (this, w, h);

  // Mipmap level 1.
  if (Textures::mipmap_nice != MIPMAP_VERYNICE) { w /= 2; h /= 2; }
  t2 = tfact->new_texture (this, w, h);

  // Mipmap level 2.
  w /= 2; h /= 2;
  t3 = tfact->new_texture (this, w, h);

  // Mipmap level 3.
  w /= 2; h /= 2;
  t4 = tfact->new_texture (this, w, h);
}

void TextureMM::blend_mipmap0 (Textures* tex)
{
  if (!ifile)
    return;

  TextureFactory8 tf8;
  TextureFactory32 tf32;
  TextureFactory* tfact;
  if (Textures::txtMode == TXT_24BIT) tfact = &tf32;
  else tfact = &tf8;

  Texture* tt;
  int w, h;
  w = ifile->get_width ();
  h = ifile->get_height ();

  tt = tfact->new_texture (this, w, h);
  create_blended_mipmap (tex, t1->get_bitmap8 (), w, h, tt->get_bitmap8 ());
  t1->copy (tt);

  CHK (delete tt);
}

void TextureMM::create_mipmaps (Textures* tex)
{
  if (!ifile)
    return;

  int w, h;
  w = ifile->get_width ();
  h = ifile->get_height ();

  if (Textures::mipmap_nice == MIPMAP_VERYNICE)
  {
    // Mipmap level 1 (the same size of texture as level 0 but blended a little)
    create_blended_mipmap (tex, t1->get_bitmap8 (), w, h, t2->get_bitmap8 ());

    // Mipmap level 2 (mipmap starting from the blended version).
    create_mipmap_bitmap (tex, t2->get_bitmap8 (), w, h, 1, t3->get_bitmap8 (), w/2, h/2);

    // Mipmap level 3 (mipmap starting from the blended version).
    create_mipmap_bitmap (tex, t2->get_bitmap8 (), w, h, 2, t4->get_bitmap8 (), w/4, h/4);

    // Size of last mipmap (for computation of mean value).
    w /= 4; h /= 4;
  }
  else
  {
    // Mipmap level 1 (mipmap starting from original version).
    create_mipmap_bitmap (tex, t1->get_bitmap8 (), w, h, 1, t2->get_bitmap8 (), w/2, h/2);

    // Mipmap level 2 (mipmap starting from original version).
    create_mipmap_bitmap (tex, t1->get_bitmap8 (), w, h, 2, t3->get_bitmap8 (), w/4, h/4);

    // Mipmap level 3 (mipmap starting from level 1).
    create_mipmap_bitmap (tex, t2->get_bitmap8 (), w/2, h/2, 2, t4->get_bitmap8 (), w/8, h/8);

    // Size of last mipmap (for computation of mean value).
    w /= 8; h /= 8;
  }

  // Compute one mean value for this texture. This is then used when texture
  // mapping is disabled.
  if (Textures::txtMode == TXT_GLOBAL)
  {
    int x, y, r, g, b;
    unsigned char* d = t4->get_bitmap8 ();
    RGBcolor* pal = tex->get_palette ();
    r = g = b = 0;
    for (x = 0 ; x < w ; x++)
      for (y = 0 ; y < h ; y++)
      {
        r += pal[*d].red;
        g += pal[*d].green;
        b += pal[*d].blue;
        d++;
      }
    r /= (w*h);
    g /= (w*h);
    b /= (w*h);

    mean_r = (float)r / 255.;
    mean_g = (float)g / 255.;
    mean_b = (float)b / 255.;
    mean_color = tex->find_color (r, g, b);
  }
  else if (Textures::txtMode == TXT_24BIT)
  {
    int x, y, r, g, b;
    ULong* d = t4->get_bitmap32 ();
    r = g = b = 0;
    for (x = 0 ; x < w ; x++)
      for (y = 0 ; y < h ; y++)
      {
        r += R24 (*d);
        g += G24 (*d);
        b += B24 (*d);
        d++;
      }
    r /= (w*h);
    g /= (w*h);
    b /= (w*h);

    mean_r = (float)r / 255.;
    mean_g = (float)g / 255.;
    mean_b = (float)b / 255.;
    mean_color = tex->find_color (r, g, b);
  }
  else
  {
    int x, y, r, g, b;
    unsigned char* d = t4->get_bitmap8 ();
    unsigned char* pal = priv_cmap->rgb_values;
    r = g = b = 0;
    for (x = 0 ; x < w ; x++)
      for (y = 0 ; y < h ; y++)
      {
        r += pal[((*d)<<2)+0];
        g += pal[((*d)<<2)+1];
        b += pal[((*d)<<2)+2];
        d++;
      }
    r /= (w*h);
    g /= (w*h);
    b /= (w*h);
    //mean_color = priv_cmap->find_color (r, g, b);
    mean_r = (float)r / 255.;
    mean_g = (float)g / 255.;
    mean_b = (float)b / 255.;
    mean_color = tex->find_color (r, g, b);
  }
}

// After loading and creating the subtextures we don't need
// the load information anymore.
void TextureMM::free_load_info ()
{
  if (ifile) { CHK (delete ifile); ifile = NULL; }
  if (usage) { CHK (delete usage); usage = NULL; }
}

// This function must be called BEFORE color remapping.
void TextureMM::set_transparent (int red, int green, int blue)
{
  transp_red = red;
  transp_green = green;
  transp_blue = blue;
  transparent = 0;
}

Texture* TextureMM::get_texture (int lev)
{
  switch (lev)
  {
    case 0: return t1;
    case 1: return t2;
    case 2: return t3;
    case 3: return t4;
  }
  return NULL;
}

void TextureMM::compute_color_usage ()
{
  if (!ifile) return;
  if (usage) return;
  CHK (usage = new ImageColorInfo(*ifile) );
}

void TextureMM::remap_palette (Textures* new_palette)
{
  if (!ifile) return;
  if (Textures::txtMode == TXT_GLOBAL)
    remap_palette_global (new_palette);
  else if (Textures::txtMode == TXT_24BIT)
    remap_palette_24bit (new_palette);
  else
    remap_palette_private (new_palette);
}

void TextureMM::remap_palette_global (Textures* new_palette)
{
  compute_color_usage ();
  if (!usage)
    return;

  int* trans;
  int num_col = usage->get_num_colors ();
  int i;

  CHK (trans = new int [num_col]);
  for (i = 0 ; i < num_col ; i++)
  {
    if (get_transparent () != -1 && get_usage(i).red == transp_red &&
        get_usage(i).green == transp_green && get_usage(i).blue == transp_blue)
      trans[i] = 0;
    else
      trans[i] = new_palette->find_rgb (get_usage(i).red, get_usage(i).green,
                                                          get_usage(i).blue);
  }

  int w = ifile->get_width ();
  int h = ifile->get_height ();

  unsigned char* dest = t1->get_bitmap8 ();

  const RGBPixel* src = ifile->get_image ();

  int x, y;
  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      for (i = 0 ; i < num_col ; i++)
        if (src->red == get_usage(i).red && src->green == get_usage(i).green && src->blue == get_usage(i).blue) break;
      *dest++ = trans[i];
      src++;
    }

  CHK (delete [] trans);
}

void TextureMM::remap_palette_private (Textures* new_palette)
{
  compute_color_usage ();
  if (!usage)
    return;

  int* trans;
  int num_col = usage->get_num_colors ();
  int i;

  if (priv_cmap) { CHK (delete priv_cmap); priv_cmap = NULL; }
  CHK (priv_cmap = new TxtCmapPrivate);

  // First allocate colors in the private palette.
  CHK (trans = new int [num_col]);

  // Minimum allowed distance, the most used colors should use
  // a very low distance.
  int dist, ddist;
  if (num_col < 256)
  {
    // If the number of colors in our texture is less than 256 we
    // set the allocator to exact mode. This will simply copy the
    // palette.
    dist = 0;
    ddist = 0;
  }
  else if (num_col < 400)
  {
    // If we have less than 400 colors then we prefer a lower distance
    // in general.
    dist = 500; // @@@ Experiment with these values
    ddist = 50;
  }
  else
  {
    dist = 1000; // @@@ Experiment with these values
    ddist = 100;
  }
  for (i = 0 ; i < num_col ; i++)
  {
    if (get_transparent () != -1 && get_usage(i).red == transp_red &&
        get_usage(i).green == transp_green && get_usage(i).blue == transp_blue)
      trans[i] = 0;
    else
      trans[i] = priv_cmap->alloc_rgb (
                     get_usage(i).red,
                     get_usage(i).green,
                     get_usage(i).blue,
                     dist);
    dist += ddist;
  }

  int w = ifile->get_width ();
  int h = ifile->get_height ();

  unsigned char* dest = t1->get_bitmap8 ();
  const RGBPixel* src = ifile->get_image ();

  // Map the texture to the private palette.
  int x, y;
  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      for (i = 0 ; i < num_col ; i++)
        if (src->red == get_usage(i).red && src->green == get_usage(i).green && src->blue == get_usage(i).blue) break;
      *dest++ = trans[i];
      src++;
    }

  CHK (delete [] trans);

  // Make a table to convert the private colormap to the global colormap.
  for (i = 0 ; i < num_col ; i++)
    priv_cmap->priv_to_global[i] = new_palette->find_rgb (
                priv_cmap->rgb_values[(i<<2)+0],
                priv_cmap->rgb_values[(i<<2)+1],
                priv_cmap->rgb_values[(i<<2)+2]);
}

void TextureMM::remap_palette_24bit (Textures*)
{
  compute_color_usage ();
  if (!usage)
    return;

  int w = ifile->get_width ();
  int h = ifile->get_height ();

  ULong* dest = t1->get_bitmap32 ();
  const RGBPixel* src = ifile->get_image ();

  // Map the texture to the RGB palette.
  int x, y;
  for (y = 0 ; y < h ; y++)
    for (x = 0 ; x < w ; x++)
    {
      *dest++ = (src->red<<16) | (src->green<<8) | src->blue;
      src++;
    }
}

//---------------------------------------------------------------------------

Texture::Texture (TextureMM* p, int w, int h)
{
  parent = p;
  width = w;
  height = h;
  m_pxTextureMap = NULL;
  init ();
}


// COM IUnknown implementation
STDMETHODIMP_(ULong) Texture::AddRef()
{
        return ++m_cRef;
}

STDMETHODIMP_(ULong) Texture::Release()
{
        // no destroy
        return --m_cRef;
}

STDMETHODIMP Texture::QueryInterface( REFIID riid, void** ppv )
{
        if (riid == IID_IUnknown)
                *ppv = static_cast<IUnknown*>(GetITextureMapFromTexture(this));
        else if (riid == IID_ITextureMap)
                *ppv = static_cast<ITextureMap*>(GetITextureMapFromTexture(this));
        else
        {
                *ppv = NULL;
                return E_NOINTERFACE;
        }

        ((IUnknown*)*ppv)->AddRef();
        return S_OK;
}

Texture8::Texture8 (TextureMM* p, int w, int h) : Texture (p, w, h)
{
  CHK (bitmap = new unsigned char [w*h]);

  m_pxTextureMap = new XTextureMap8(this);
}

Texture32::Texture32 (TextureMM* p, int w, int h) : Texture (p, w, h)
{
  CHK (bitmap = new ULong [w*h]);

  m_pxTextureMap = new XTextureMap32(this);
}

void Texture::init ()
{
  switch (width)
  {
    case 4: shf_w = 2; and_w = 0x3; break;
    case 8: shf_w = 3; and_w = 0x7; break;
    case 16: shf_w = 4; and_w = 0xf; break;
    case 32: shf_w = 5; and_w = 0x1f; break;
    case 64: shf_w = 6; and_w = 0x3f; break;
    case 128: shf_w = 7; and_w = 0x7f; break;
    case 256: shf_w = 8; and_w = 0xff; break;
    case 512: shf_w = 9; and_w = 0x1ff; break;
    case 1024: shf_w = 10; and_w = 0x3ff; break;
  }
  switch (height)
  {
    case 4: shf_h = 2; and_h = 0x3; break;
    case 8: shf_h = 3; and_h = 0x7; break;
    case 16: shf_h = 4; and_h = 0xf; break;
    case 32: shf_h = 5; and_h = 0x1f; break;
    case 64: shf_h = 6; and_h = 0x3f; break;
    case 128: shf_h = 7; and_h = 0x7f; break;
    case 256: shf_h = 8; and_h = 0xff; break;
    case 512: shf_h = 9; and_h = 0x1ff; break;
    case 1024: shf_h = 10; and_h = 0x3ff; break;
  }
}

Texture::~Texture ()
{
    delete m_pxTextureMap;
}

Texture8::~Texture8 ()
{
  if (bitmap) CHKB (delete [] bitmap);
}

Texture32::~Texture32 ()
{
  if (bitmap) CHKB (delete [] bitmap);
}

//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( Textures )

BEGIN_INTERFACE_TABLE( Textures )
        IMPLEMENTS_COMPOSITE_INTERFACE( TextureContainer )
END_INTERFACE_TABLE()

UShort Textures::alpha_mask;
int Textures::mipmap_nice = 0;
bool Textures::do_blend_mipmap0 = false;
bool Textures::do_lightmapgrid = false;
bool Textures::do_lightmaponly = false;
int Textures::color_table1;
int Textures::color_table2;
int Textures::color_table3;
int Textures::mixing = MIX_TRUE_RGB;
int Textures::force_mix = -1;
int Textures::txtMode = TXT_GLOBAL;
int Textures::force_txtMode = -1;
int Textures::ambient_red = DEFAULT_LIGHT_LEVEL;
int Textures::ambient_green = DEFAULT_LIGHT_LEVEL;
int Textures::ambient_blue = DEFAULT_LIGHT_LEVEL;
int Textures::ambient_white = DEFAULT_LIGHT_LEVEL;
bool Textures::use_rgb = true;
bool Textures::accurate_rgb_mapper = false;
int Textures::force_accurate_rgb_map = -1;

Filter3x3 Textures::mipmap_filter_1 =
{
  1, 2, 1,
  2, 4, 2,
  1, 2, 1,
  16
};
Filter5x5 Textures::mipmap_filter_2 =
{
  1, 1, 1, 1, 1,
  1, 2, 4, 2, 1,
  1, 4, 8, 4, 1,
  1, 2, 4, 2, 1,
  1, 1, 1, 1, 1,
  48
};
Filter3x3 Textures::blend_filter =
{
  1, 2, 1,
  2, 4, 2,
  1, 2, 1,
  16
};


Textures::Textures (int max)
{
  max_textures = max;
  num_textures = 0;
  CHK (textures = new TextureMM* [max]);

  lt_truergb = NULL;
  lt_truergb_private = NULL;
  lt_fastxxx = NULL;
  lt_white16 = NULL;
  lt_white8 = NULL;
  lt_pal = NULL;
  lt_alpha = NULL;

  GfxInfo->GetPixelFormat (&pfmt);

  truecolor = pfmt.PalEntries == 0;
  if (!truecolor)
  {
    // If we don't have truecolor we simulate 6:6:4 bits
    // for R:G:B in the masks anyway because we still need the
    // 16-bit format for our light mixing (in true_rgb mode).
    pfmt.RedMask = MASK_RED << (BITS_GREEN+BITS_BLUE);
    pfmt.RedShift = BITS_GREEN+BITS_BLUE;
    pfmt.GreenMask = MASK_GREEN << BITS_BLUE;
    pfmt.GreenShift = BITS_BLUE;
    pfmt.BlueMask = MASK_BLUE;
    pfmt.BlueShift = 0;
    long m;
    pfmt.RedBits = 0;  m = pfmt.RedMask >> pfmt.RedShift;    while (m) { pfmt.RedBits++; m >>= 1; }
    pfmt.GreenBits = 0;m = pfmt.GreenMask >> pfmt.GreenShift;while (m) { pfmt.GreenBits++; m >>= 1; }
    pfmt.BlueBits = 0; m = pfmt.BlueMask >> pfmt.BlueShift;  while (m) { pfmt.BlueBits++; m >>= 1; }
  }
  num_red = (pfmt.RedMask >> pfmt.RedShift) + 1;
  num_green = (pfmt.GreenMask >> pfmt.GreenShift) + 1;
  num_blue = (pfmt.BlueMask >> pfmt.BlueShift) + 1;

  read_config ();

  alpha_mask = 0;
  alpha_mask |= 1 << (pfmt.RedShift);
  alpha_mask |= 1 << (pfmt.GreenShift);
  alpha_mask |= 1 << (pfmt.BlueShift);
  alpha_mask = ~alpha_mask;

  // By allocating black at 0 and white at 255 we are compatible with
  // the Windows palette (those two colors cannot be modified).
  // This is important even for X Windows because X Windows can be run
  // in Windows NT (or 95) under X emulation and the same limitation
  // is also present then.
  //
  // Note that we can't use these colors for textures since it is
  // possible that we apply gamma correction on the colormap and this
  // is not possible for those two entries. However, they are very
  // useful as fixed black and white entries no matter what the
  // palette is (for messages and such).
  //
  // Another useful trick is to map color 0 to 0 in every lightlevel.
  // Using this feature transparent colors in textures remain transparent
  // even with lighting added.
  memset (alloc, 0, sizeof (alloc));
  memset (pal, 0, sizeof (pal));
  alloc[0] = true;
  pal[0].red = 0;
  pal[0].green = 0;
  pal[0].blue = 0;
  alloc[255] = true;
  pal[255].red = 255;
  pal[255].green = 255;
  pal[255].blue = 255;
}

bool Textures::force_txtmode (char* p)
{
  if (!strcmp (p, "global")) force_txtMode = TXT_GLOBAL;
  else if (!strcmp (p, "private")) force_txtMode = TXT_PRIVATE;
  else if (!strcmp (p, "24bit")) force_txtMode = TXT_24BIT;
  //{
    //CsPrintf (MSG_FATAL_ERROR, "TXTMODE '24bit' currently not supported! Please use 'global' or 'private'.\n");
    //return false;
  //}
  else
  {
    CsPrintf (MSG_FATAL_ERROR, "Bad value '%s' for TXTMODE (use 'global', 'private', or '24bit')!\n", p);
    return false;
  }
  return true;
}

bool Textures::force_mixing (char* mix)
{
  if (!strcmp (mix, "true_rgb")) force_mix = MIX_TRUE_RGB;
  else if (!strcmp (mix, "fast_wrb")) force_mix = MIX_FAST_WRB;
  else if (!strcmp (mix, "fast_wrg")) force_mix = MIX_FAST_WRG;
  else if (!strcmp (mix, "fast_wbg")) force_mix = MIX_FAST_WBG;
  else if (!strcmp (mix, "nocolor")) force_mix = MIX_NOCOLOR;
  else
  {
    CsPrintf (MSG_FATAL_ERROR, "Bad value '%s' for 'mixing' (use 'true_rgb', 'fast_wrb',\n'fast_wrg', 'fast_wbg', or 'nocolor')!\n", mix);
    return false;
  }
  return true;
}

bool Textures::force_accurate_rgb_mapper (bool rgb)
{
  force_accurate_rgb_map = (bool)rgb;
  return true;
}

void Textures::read_config ()
{
  char* p;
  Archive* ar = NULL;
  csIniFile* cfg = NULL;

  if (System->world) ar = System->world->get_world_file ();
  if (ar) CHKB (cfg = new csIniFile (ar, "cryst.cfg"));

  accurate_rgb_mapper = !config->GetYesNo ("World", "FAST_RGB_ALLOC", true);
  if (cfg) accurate_rgb_mapper = !cfg->GetYesNo ("World", "FAST_RGB_ALLOC", !accurate_rgb_mapper);
  if (force_accurate_rgb_map != -1) accurate_rgb_mapper = (bool)force_accurate_rgb_map;

  do_blend_mipmap0 = config->GetYesNo ("TextureMapper", "BLEND_MIPMAP", false);

  p = config->GetStr ("TextureMapper", "MIPMAP_FILTER_1", "-");
  if (*p != '-')
  {
    ScanStr (p, "%d,%d,%d,%d,%d,%d,%d,%d,%d",
      &mipmap_filter_1.f11, &mipmap_filter_1.f12, &mipmap_filter_1.f13,
      &mipmap_filter_1.f21, &mipmap_filter_1.f22, &mipmap_filter_1.f23,
      &mipmap_filter_1.f31, &mipmap_filter_1.f32, &mipmap_filter_1.f33);
    mipmap_filter_1.tot =
      mipmap_filter_1.f11+mipmap_filter_1.f12+mipmap_filter_1.f13+
      mipmap_filter_1.f21+mipmap_filter_1.f22+mipmap_filter_1.f23+
      mipmap_filter_1.f31+mipmap_filter_1.f32+mipmap_filter_1.f33;
  }
  p = config->GetStr ("TextureMapper", "MIPMAP_FILTER_2", "-");
  if (*p != '-')
  {
    ScanStr (p, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
      &mipmap_filter_2.f00, &mipmap_filter_2.f01, &mipmap_filter_2.f02, &mipmap_filter_2.f03, &mipmap_filter_2.f04,
      &mipmap_filter_2.f10, &mipmap_filter_2.f11, &mipmap_filter_2.f12, &mipmap_filter_2.f13, &mipmap_filter_2.f14,
      &mipmap_filter_2.f20, &mipmap_filter_2.f21, &mipmap_filter_2.f22, &mipmap_filter_2.f23, &mipmap_filter_2.f24,
      &mipmap_filter_2.f30, &mipmap_filter_2.f31, &mipmap_filter_2.f32, &mipmap_filter_2.f33, &mipmap_filter_2.f34,
      &mipmap_filter_2.f40, &mipmap_filter_2.f41, &mipmap_filter_2.f42, &mipmap_filter_2.f43, &mipmap_filter_2.f44);
    mipmap_filter_2.tot =
      mipmap_filter_2.f00+mipmap_filter_2.f01+mipmap_filter_2.f02+mipmap_filter_2.f03+mipmap_filter_2.f04+
      mipmap_filter_2.f10+mipmap_filter_2.f11+mipmap_filter_2.f12+mipmap_filter_2.f13+mipmap_filter_2.f14+
      mipmap_filter_2.f20+mipmap_filter_2.f21+mipmap_filter_2.f22+mipmap_filter_2.f23+mipmap_filter_2.f24+
      mipmap_filter_2.f30+mipmap_filter_2.f31+mipmap_filter_2.f32+mipmap_filter_2.f33+mipmap_filter_2.f34+
      mipmap_filter_2.f40+mipmap_filter_2.f41+mipmap_filter_2.f42+mipmap_filter_2.f43+mipmap_filter_2.f44;
  }
  p = config->GetStr ("TextureMapper", "BLEND_FILTER", "-");
  if (*p != '-')
  {
    ScanStr (p, "%d,%d,%d,%d,%d,%d,%d,%d,%d",
      &blend_filter.f11, &blend_filter.f12, &blend_filter.f13,
      &blend_filter.f21, &blend_filter.f22, &blend_filter.f23,
      &blend_filter.f31, &blend_filter.f32, &blend_filter.f33);
    blend_filter.tot =
      blend_filter.f11+blend_filter.f12+blend_filter.f13+
      blend_filter.f21+blend_filter.f22+blend_filter.f23+
      blend_filter.f31+blend_filter.f32+blend_filter.f33;
  }

  prefered_dist = config->GetInt ("World", "RGB_DIST", PREFERED_DIST);
  prefered_col_dist = config->GetInt ("World", "RGB_COL_DIST", PREFERED_COL_DIST);
  p = config->GetStr ("TextureMapper", "MIPMAP_NICE", "nice");
  if (!strcmp (p, "nice"))
  {
    mipmap_nice = MIPMAP_NICE;
    CsPrintf (MSG_INITIALIZATION, "Mipmap calculation 'nice'.\n");
  }
  else if (!strcmp (p, "ugly"))
  {
    mipmap_nice = MIPMAP_UGLY;
    CsPrintf (MSG_INITIALIZATION, "Mipmap calculation 'ugly'.\n");
  }
  else if (!strcmp (p, "default"))
  {
    mipmap_nice = MIPMAP_DEFAULT;
    CsPrintf (MSG_INITIALIZATION, "Mipmap calculation 'default'.\n");
  }
  else if (!strcmp (p, "verynice"))
  {
    G3D_CAPS caps;
    Gfx3D->GetCaps (&caps);

    if (caps.SupportsArbitraryMipMapping)
    {
      mipmap_nice = MIPMAP_VERYNICE;
      CsPrintf (MSG_INITIALIZATION, "Mipmap calculation 'verynice'\n  (Note: this is expensive for the texture cache)\n");
    }
    else
    {
      mipmap_nice = MIPMAP_NICE;
      CsPrintf (MSG_INITIALIZATION, "Mipmap calculation 'nice' ('verynice' not available for hardware accelerators).\n");
    }
  }
  else
  {
    CsPrintf (MSG_FATAL_ERROR, "Bad value '%s' for MIPMAP_NICE!\n(Use 'verynice', 'nice', 'ugly', or 'default')\n", p);
    fatal_exit (0, false);
  }

  if (force_mix != -1) mixing = force_mix;
  else
  {
    char buf[100];
    strcpy (buf, config->GetStr ("World", "MIXLIGHTS", "true_rgb"));
    if (cfg) p = cfg->GetStr ("World", "MIXLIGHTS", buf);
    else p = buf;

    if (!strcmp (p, "true_rgb")) mixing = MIX_TRUE_RGB;
    else if (!strcmp (p, "fast_wrb")) mixing = MIX_FAST_WRB;
    else if (!strcmp (p, "fast_wrg")) mixing = MIX_FAST_WRG;
    else if (!strcmp (p, "fast_wbg")) mixing = MIX_FAST_WBG;
    else if (!strcmp (p, "nocolor")) mixing = MIX_NOCOLOR;
    else
    {
      CsPrintf (MSG_FATAL_ERROR, "Bad value '%s' for MIXLIGHTS (use 'true_rgb', 'fast_wrb',\n'fast_wrg', 'fast_wbg', or 'nocolor')!\n", p);
      fatal_exit (0, false);
    }
  }

  G3D_COLORMAPFORMAT cmf;
  Gfx3D->GetColormapFormat(cmf);
  if (cmf!=G3DCOLORFORMAT_ANY && force_txtMode==-1)
  {
    if (cmf==G3DCOLORFORMAT_GLOBAL) force_txtMode = TXT_GLOBAL;
    else if (cmf==G3DCOLORFORMAT_PRIVATE) force_txtMode = TXT_PRIVATE;
  }

  if (force_txtMode != -1) txtMode = force_txtMode;
  else
  {
    char buf[100];
    strcpy (buf, config->GetStr ("World", "TXTMODE", "global"));
    if (cfg) p = cfg->GetStr ("World", "TXTMODE", buf);
    else p = buf;

    if (!strcmp (p, "global")) txtMode = TXT_GLOBAL;
    else if (!strcmp (p, "private")) txtMode = TXT_PRIVATE;
    else if (!strcmp (p, "24bit")) txtMode = TXT_24BIT;
    //{
      //CsPrintf (MSG_FATAL_ERROR, "TXTMODE '24bit' currently not supported! Please use 'global' or 'private'.\n");
      //fatal_exit (0, false);
    //}
    else
    {
      CsPrintf (MSG_FATAL_ERROR, "Bad value '%s' for TXTMODE (use 'global', 'private', or '24bit')!\n", p);
      fatal_exit (0, false);
    }
  }

  if (!truecolor && txtMode == TXT_24BIT)
  {
    CsPrintf (MSG_FATAL_ERROR, "TXTMODE '24bit' not yet supported in 8-bit display mode!\n");
    fatal_exit (0, false);
  }

  if (truecolor)
    CsPrintf (MSG_INITIALIZATION, "Truecolor mode (15/16 bit).\n");
  if (truecolor && mixing != MIX_TRUE_RGB)
  {
    CsPrintf (MSG_INITIALIZATION, "  Mixing mode forced to true_rgb.\n");
    mixing = MIX_TRUE_RGB;
  }
  if (truecolor && accurate_rgb_mapper)
  {
    CsPrintf (MSG_INITIALIZATION, "  Accurate rgb mapper not needed.\n");
    accurate_rgb_mapper = false;
  }
  if (txtMode != TXT_GLOBAL && mixing != MIX_TRUE_RGB)
  {
    CsPrintf (MSG_INITIALIZATION, "Private or 24-bit textures: mixing mode forced to true_rgb.\n");
    mixing = MIX_TRUE_RGB;
  }

  ambient_red = config->GetInt ("World", "AMBIENT_RED", DEFAULT_LIGHT_LEVEL);
  ambient_green = config->GetInt ("World", "AMBIENT_GREEN", DEFAULT_LIGHT_LEVEL);
  ambient_blue = config->GetInt ("World", "AMBIENT_BLUE", DEFAULT_LIGHT_LEVEL);
  ambient_white = config->GetInt ("World", "AMBIENT_WHITE", DEFAULT_LIGHT_LEVEL);

  ambient_red += ambient_white;
  ambient_green += ambient_white;
  ambient_blue += ambient_white;
  level_r = ambient_red;
  level_g = ambient_green;
  level_b = ambient_blue;

  if (cfg) CHKB (delete cfg);
}

Textures::~Textures ()
{
  for (int i = 0; i < num_textures; i++)
    CHKB (delete textures[i]);
  if (textures) CHKB (delete [] textures);
  if (lt_truergb) CHKB (delete lt_truergb);
  if (lt_truergb_private) CHKB (delete lt_truergb_private);
  if (lt_fastxxx) CHKB (delete lt_fastxxx);
  if (lt_white16) CHKB (delete lt_white16);
  if (lt_white8) CHKB (delete lt_white8);
  if (lt_pal) CHKB (delete lt_pal);
  if (lt_alpha) CHKB (delete lt_alpha);
}

TextureMM* Textures::new_texture (char* name, ImageFile* image)
{
  CHK (TextureMM* tm = new TextureMM (name, image));
  if (tm->loaded_correctly ())
    add_texture (tm);
  return tm;
}

void Textures::add_texture (TextureMM* tm)
{
  if (num_textures >= max_textures)
  {
    int max_textures = num_textures + 8;
    CHK (TextureMM** newtextures = new TextureMM* [max_textures]);
    if (textures)
    {
      memcpy (newtextures, textures, num_textures * sizeof (TextureMM*));
      CHK (delete [] textures);
    }
    textures = newtextures;
  }
  textures [num_textures++] = tm;
}

Texture* Textures::get_texture (int idx, int lev)
{
  return textures[idx]->get_texture (lev);
}

int Textures::get_texture_idx (char* name)
{
  int i;
  for (i = 0 ; i < num_textures ; i++)
    if (!strcmp (textures[i]->get_name (), name)) return i;
  return -1;
}

TextureMM* Textures::get_texture_mm (char* name)
{
  int idx = get_texture_idx (name);
  if (idx == -1) return NULL;
  else return textures[idx];
}

void Textures::delete_texture (int idx, bool del)
{
  if ((idx < 0)
   || (idx >= num_textures))
    return;

  if (del)
    delete textures [idx];
  memmove (&textures [idx], &textures [idx + 1],
    (--num_textures - idx) * sizeof (TextureMM *));

  if (max_textures > num_textures + 16)
  {
    max_textures = num_textures + 8;
    CHK (TextureMM** newtextures = new TextureMM* [max_textures]);
    memcpy (newtextures, textures, num_textures * sizeof (TextureMM*));
    CHK (delete [] textures);
    textures = newtextures;
  } /* endif */
}

#define GAMMA(c) ((int)(256.*pow (((float)(c))/256., 1./Gamma)))

int Textures::find_rgb_real (int r, int g, int b)
{
  int num_pal_entries;
  GfxInfo->GetNumPalEntries( num_pal_entries );

  if (num_pal_entries != 0)
  {
    if (r == 0 && g == 0 && b == 0) return 0;
    if (r == 255 && g == 255 && b == 255) return 255;

    int i, min, dist, mindist;
    mindist = 1000*256*256; min = -1;
    int pr, pg, pb;

    for(i = 1 ; i < 255 ; i++) // Exclude colors 0 and 255
      if (alloc[i])
      {
        pr = GAMMA (pal[i].red); if (pr < 0) pr = 0; else if (pr > 255) pr = 255;
        pg = GAMMA (pal[i].green); if (pg < 0) pg = 0; else if (pg > 255) pg = 255;
        pb = GAMMA (pal[i].blue); if (pb < 0) pb = 0; else if (pb > 255) pb = 255;
        dist = 299*(r-pr)*(r-pr)+587*(g-pg)*(g-pg)+114*(b-pb)*(b-pb);
        if (dist == 0) return i;
        if (dist < mindist) { mindist = dist; min = i; }
      }
    return min;
  } else
    return encode_rgb (r, g, b);
}

UShort Textures::encode_rgb (int r, int g, int b)
{
  return
    ((r >> (8-pfmt.RedBits))   << pfmt.RedShift) |
    ((g >> (8-pfmt.GreenBits)) << pfmt.GreenShift) |
    ((b >> (8-pfmt.BlueBits))  << pfmt.BlueShift);
}

int Textures::find_rgb (int r, int g, int b)
{
  //GAC Now use lookup table
  if (r>255) r=255; else if (r<0) r=0;
  if (g>255) g=255; else if (g<0) g=0;
  if (b>255) b=255; else if (b<0) b=0;

  return lt_palette_table[encode_rgb (r, g, b)];
}

int Textures::find_color (int r, int g, int b)
{
  //GAC Now use lookup table
  if (r>255) r=255; else if (r<0) r=0;
  if (g>255) g=255; else if (g<0) g=0;
  if (b>255) b=255; else if (b<0) b=0;

  if (truecolor) return encode_rgb (r, g, b);
  else return lt_palette_table[encode_rgb (r, g, b)];
}

UShort Textures::encode_rgb_safe (int r, int g, int b)
{
  //GAC Now use lookup table
  if (r>255) r=255; else if (r<0) r=0;
  if (g>255) g=255; else if (g<0) g=0;
  if (b>255) b=255; else if (b<0) b=0;

  return encode_rgb (r, g, b);
}

int Textures::find_rgb_slow (int r, int g, int b)
{
  if (r>255) r=255; else if (r<0) r=0;
  if (g>255) g=255; else if (g<0) g=0;
  if (b>255) b=255; else if (b<0) b=0;

  int i, min, mindist;
  mindist = 1000L*256*256;
  min = -1;
  register int red, green, blue, dist;
  for(i = 1 ; i < 255 ; i++) // Exclude colors 0 and 255
    if (alloc[i])
    {
      red = r - pal[i].red;
      green = g - pal[i].green;
      blue = b - pal[i].blue;
      dist = (299*red*red) + (587*green*green) + (114*blue*blue);
      if (dist == 0) return i;
      if (dist < mindist) { mindist = dist; min = i; }
    }
  return min;
}

int Textures::alloc_rgb (int r, int g, int b, int dist)
{
  int d, j;
  if (r < 0) r = 0; else if (r > 255) r = 255;
  if (g < 0) g = 0; else if (g > 255) g = 255;
  if (b < 0) b = 0; else if (b > 255) b = 255;

  d = 0;
  int i = find_rgb_slow (r, g, b);
  if (i != -1)
  {
    d = 299*(r-pal[i].red)*(r-pal[i].red)+
      587*(g-pal[i].green)*(g-pal[i].green)+
      114*(b-pal[i].blue)*(b-pal[i].blue);
  }
  if (i == -1 || d > dist)
  {
    for (j = 1 ; j < 255 ; j++) // Exclude color 0 and 255
      if (!alloc[j])
      {
        alloc[j] = true;
        pal[j].red = r;
        pal[j].green = g;
        pal[j].blue = b;
        return j;
      }
    return i; // We couldn't allocate a new color, return best fit
  }
  else return i;
}

void Textures::create_lt_palette ()
{
  // GAC Fill palette_table for fast lookup using slow version of lookup
  // JTY This works by encoding the R,G,B values into one 16-bit word
  // (by default this is 6:6:4 bits for R:G:B in 8-bit mode). This lowers
  // the precision of our RGB mapper a bit which means that the display
  // will have slightly less quality (you have to look very closely to
  // be able to see this).
  //
  // In truecolor mode we do something similar. Only the number of bits
  // for every color can vary.

  if (lt_pal) CHKB (delete lt_pal);
  CHK (lt_pal = new TextureTablesPalette);
  lt_palette_table = lt_pal->true_to_pal;

  int i;
  int r, g, b;

#ifdef SLOW_create_lt_palette
  int tr,tg,tb;
  for (tr=0; tr<num_red; tr++)
  {
    for (tg=0; tg<num_green; tg++)
    {
      for (tb=0; tb<num_blue; tb++)
      {
        r = tr<<(8-pfmt.RedBits);
        g = tg<<(8-pfmt.GreenBits);
        b = tb<<(8-pfmt.BlueBits);
        lt_palette_table[encode_rgb (r,g,b)] = find_rgb_slow(r, g, b);
      }
    }
  }
#else
  // Greg Ewing, 12 Oct 1998
  unsigned char *colormap[3];
  unsigned long *dist_buf;
  for (i = 0; i < 3; i++)
    CHKB (colormap[i] = new unsigned char[256]);
  for (i = 0; i < 256; i++) {
    colormap[0][i] = pal[i].red;
    colormap[1][i] = pal[i].green;
    colormap[2][i] = pal[i].blue;
  }
  CHK (dist_buf = new unsigned long[num_red * num_green * num_blue]);
  inv_cmap(256, colormap, pfmt.RedBits, pfmt.GreenBits, pfmt.BlueBits,
           dist_buf, lt_palette_table);
  for (i = 0; i < 3; i++)
    CHKB (delete [] colormap[i]);
  CHK (delete [] dist_buf);
  // Color number 0 is reserved for transparency
  lt_palette_table [encode_rgb (pal [0].red, pal [0].green, pal [0].blue)] =
    find_rgb_slow (pal [0].red, pal [0].green, pal [0].blue);
#endif

  for (i = 0 ; i < 256 ; i++)
  {
    r = pal[i].red;
    g = pal[i].green;
    b = pal[i].blue;
    lt_pal->pal_to_true[i] = encode_rgb (r, g, b);
  }
}

void Textures::compute_palette ()
{
  int i, t;

  CsPrintf (MSG_INITIALIZATION, "Computing palette...\n");

  for (i = 1 ; i < 255 ; i++)
    alloc[i] = false;

  // First compute the usage of all colors of all textures.
  // These usage lists will be sorted so that the most frequently
  // used color of each textures is put first.
  for (i = 0 ; i < num_textures ; i++)
    textures[i]->compute_color_usage ();

  // The following code is for allocating an extra color for each
  // color table. This is so that we have a palette which is better
  // suited for colored lighting.
  // This is not needed for truecolor mode.
  int pr1, pg1, pb1, pr2, pg2, pb2, pr3, pg3, pb3;
  if (color_table1 == TABLE_RED)        { pr1 = 50; pg1 = 0; pb1 = 0; }
  else if (color_table1 == TABLE_GREEN) { pr1 = 0; pg1 = 50; pb1 = 0; }
  else if (color_table1 == TABLE_BLUE)  { pr1 = 0; pg1 = 0; pb1 = 50; }
  else { pr1 = 0; pg1 = 0; pb1 = 0; }
  if (color_table2 == TABLE_RED)        { pr2 = 50; pg2 = 0; pb2 = 0; }
  else if (color_table2 == TABLE_GREEN) { pr2 = 0; pg2 = 50; pb2 = 0; }
  else if (color_table2 == TABLE_BLUE)  { pr2 = 0; pg2 = 0; pb2 = 50; }
  else { pr2 = 0; pg2 = 0; pb2 = 0; }
  if (color_table3 == TABLE_RED)        { pr3 = 50; pg3 = 0; pb3 = 0; }
  else if (color_table3 == TABLE_GREEN) { pr3 = 0; pg3 = 50; pb3 = 0; }
  else if (color_table3 == TABLE_BLUE)  { pr3 = 0; pg3 = 0; pb3 = 50; }
  else { pr3 = 0; pg3 = 0; pb3 = 0; }

  // Then allocate colors for all textures at the same time.
  for (i = 0 ; i < 256 ; i++)
  {
    for (t = 0 ; t < num_textures ; t++)
      if (i < textures[t]->get_num_colors ())
      {
        alloc_rgb (
                     textures[t]->get_usage (i).red,
                     textures[t]->get_usage (i).green,
                     textures[t]->get_usage (i).blue,
                     prefered_dist);
        if (!truecolor)
        {
          if (pr1 || pg1 || pb1)
            alloc_rgb (
                       textures[t]->get_usage (i).red+pr1,
                       textures[t]->get_usage (i).green+pg1,
                       textures[t]->get_usage (i).blue+pb1,
                       prefered_col_dist);
          if (pr2 || pg2 || pb2)
            alloc_rgb (
                       textures[t]->get_usage (i).red+pr2,
                       textures[t]->get_usage (i).green+pg2,
                       textures[t]->get_usage (i).blue+pb2,
                       prefered_col_dist);
          if (pr3 || pg3 || pb3)
            alloc_rgb (
                       textures[t]->get_usage (i).red+pr3,
                       textures[t]->get_usage (i).green+pg3,
                       textures[t]->get_usage (i).blue+pb3,
                       prefered_col_dist);
        }
      }
  }

  create_lt_palette ();

  // Remap all textures according to the new colormap.
  for (i = 0 ; i < num_textures ; i++)
    textures[i]->remap_palette (this);

  black_color = 0;
  if (truecolor)
  {
    white_color = encode_rgb (255, 255, 255);
    red_color = encode_rgb (255, 0, 0);
    blue_color = encode_rgb (0, 0, 255);
    yellow_color = encode_rgb (255, 255, 0);
    green_color = encode_rgb (0, 255, 0);
  }
  else
  {
    white_color = 255;
    red_color = find_rgb (255, 0, 0);
    blue_color = find_rgb (0, 0, 255);
    yellow_color = find_rgb (255, 255, 0);
    green_color = find_rgb (0, 255, 0);
  }

  CsPrintf (MSG_INITIALIZATION, "DONE\n");
}

int Textures::find_rgb_map (int r, int g, int b, int map_type, int l)
{
  int nr = r, ng = g, nb = b;

  switch (map_type)
  {
    case TABLE_WHITE_HI:
      nr = (200+l)*r / 200;
      ng = (200+l)*g / 200;
      nb = (200+l)*b / 200;
      break;
    case TABLE_WHITE:
      nr = l*r / 200;
      ng = l*g / 200;
      nb = l*b / 200;
      break;
    case TABLE_RED_HI:
      nr = (200+l)*r / 200;
      break;
    case TABLE_GREEN_HI:
      ng = (200+l)*g / 200;
      break;
    case TABLE_BLUE_HI:
      nb = (200+l)*b / 200;
      break;
    case TABLE_RED:
      if (use_rgb) nr = l*r / 200;
      else nr = r+l*200/256;
      break;
    case TABLE_GREEN:
      if (use_rgb) ng = l*g / 200;
      else ng = g+l*200/256;
      break;
    case TABLE_BLUE:
      if (use_rgb) nb = l*b / 200;
      else nb = b+l*200/256;
      break;
  }
  return truecolor ? encode_rgb_safe (nr, ng, nb) :
    (accurate_rgb_mapper ? find_rgb_slow (nr, ng, nb) : find_rgb (nr, ng, nb));
}

void Textures::create_lt_white16 (Archive*)
{
  if (lt_white16) CHKB (delete lt_white16);
  CHK (lt_white16 = new TextureTablesWhite16 ());

  int i, l;
  int r, g, b;

  CsPrintf (MSG_INITIALIZATION, "  Calculate 16-bit white tables...\n");

  for (i = 1 ; i < 256 ; i++)
    if (alloc[i])
    {
      r = pal[i].red;
      g = pal[i].green;
      b = pal[i].blue;
      for (l = 0 ; l < 256 ; l++)
      {
        lt_white16->white1_light[l][i] = find_rgb_map (r, g, b, TABLE_WHITE_HI, l);
        lt_white16->white2_light[l][i] = find_rgb_map (r, g, b, TABLE_WHITE, l);
      }
    }

    // Color 0 is mapped to 0 in all cases to make sure that transparent
    // colors remain transparent.
    for (l = 0 ; l < 256 ; l++)
    {
      lt_white16->white1_light[l][0] = 0;
      lt_white16->white2_light[l][0] = 0;
    }
}

void Textures::create_lt_white8 (Archive* ar)
{
  if (lt_white8) CHKB (delete lt_white8);
  CHK (lt_white8 = new TextureTablesWhite8 ());

  int i, l;
  bool recalc_needed = false;
  size_t size;
  int r, g, b;

  // If accurate_rgb_mapper == false we don't bother to use the
  // cached light tables.

  if (accurate_rgb_mapper)
  {
    if (ar->file_exists ("table_white8", &size))
    {
      bool err = false;
      if (size != 256*2*sizeof (RGB8map)) err = true;
      if (!err)
      {
        char* data = ar->read ("table_white8");
        char* d = data;

        for (i = 0 ; i < 256 ; i++)
        {
          memcpy (lt_white8->white1_light[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
          memcpy (lt_white8->white2_light[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
        }
        CHK (delete [] data);

        for (i = 0 ; i < 50 ; i++)
        {
          int l, j, m;
          l = (rand () >> 2) & 0xff;
          j = (rand () >> 2) & 0xff;
          if (!RESERVED_COLOR(j) && alloc[j])
          {
            m = find_rgb_map (pal[j].red, pal[j].green, pal[j].blue, TABLE_WHITE_HI, l);
            if (lt_white8->white1_light[l][j] != m) { err = true; break; }
            m = find_rgb_map (pal[j].red, pal[j].green, pal[j].blue, TABLE_WHITE, l);
            if (lt_white8->white2_light[l][j] != m) { err = true; break; }
          }
        }
      }
      if (err)
      {
        CsPrintf (MSG_INITIALIZATION, "  'table_white8' does not conform to current world file.\n  \
The file will automatically be converted.\n  \
Please be patient...\n");
        recalc_needed = true;
      }
      else
      {
        CsPrintf (MSG_INITIALIZATION, "  Succesfully loaded 'table_white8'.\n");
        recalc_needed = false;
      }
    }
    else
    {
      CsPrintf (MSG_INITIALIZATION, "  Can't find precalculated 'table_white8'.\n  \
Will make it now. Please be patient...\n");
      recalc_needed = true;
    }
  }
  else
  {
    CsPrintf (MSG_INITIALIZATION, "  Calculate 8-bit white tables...\n");
    recalc_needed = true; // If no accurate_rgb_mapper we recalc in any case
  }

  if (recalc_needed)
  {
    for (i = 1 ; i < 256 ; i++)
      if (alloc[i])
      {
        r = pal[i].red;
        g = pal[i].green;
        b = pal[i].blue;
        for (l = 0 ; l < 256 ; l++)
        {
          lt_white8->white1_light[l][i] = find_rgb_map (r, g, b, TABLE_WHITE_HI, l);
          lt_white8->white2_light[l][i] = find_rgb_map (r, g, b, TABLE_WHITE, l);
        }
      }

    // Color 0 is mapped to 0 in all cases to make sure that transparent
    // colors remain transparent.
    for (l = 0 ; l < 256 ; l++)
    {
      lt_white8->white1_light[l][0] = 0;
      lt_white8->white2_light[l][0] = 0;
    }

    // If accurate_rgb_mapper == false then there is no need to save the tables
    // because they can be recalculated quickly enough.
    if (accurate_rgb_mapper)
    {
      void* entry = ar->new_file ("table_white8", 256*2*sizeof (RGB8map));
      for (i = 0 ; i < 256 ; i++)
      {
        ar->append (entry, (char*)lt_white8->white1_light[i], sizeof (RGB8map));
        ar->append (entry, (char*)lt_white8->white2_light[i], sizeof (RGB8map));
      }
      if (!ar->write_archive ())
        CsPrintf (MSG_WARNING, "  WARNING: error updating tables in archive %s!\n",
                     ar->GetFilename ());
    }
  }
}

void Textures::create_lt_fastxxx (Archive* ar)
{
  int i, l;
  size_t size;
  bool recalc_needed = false;
  int r, g, b;

  if (lt_fastxxx) CHKB (delete lt_fastxxx);
  CHK (lt_fastxxx = new TextureTablesFastXxx);

  // If accurate_rgb_mapper == false we don't bother to use the
  // cached light tables.

  if (accurate_rgb_mapper)
  {
    if (ar->file_exists ("table_fast_xxx", &size))
    {
      bool err = false;
      if (size != 256*3*sizeof (RGB8map)) err = true;
      if (!err)
      {
        char* data = ar->read ("table_fast_xxx");
        char* d = data;

        for (i = 0 ; i < 256 ; i++)
        {
          memcpy (lt_fastxxx->light1[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
          memcpy (lt_fastxxx->light2[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
          memcpy (lt_fastxxx->light3[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
        }
        CHK (delete [] data);

        for (i = 0 ; i < 50 ; i++)
        {
          int l, j, m;
          l = (rand () >> 2) & 0xff;
          j = (rand () >> 2) & 0xff;
          if (!RESERVED_COLOR(j) && alloc[j])
          {
            m = find_rgb_map (pal[j].red, pal[j].green, pal[j].blue, color_table1, l);
            if (lt_fastxxx->light1[l][j] != m) { err = true; break; }
            m = find_rgb_map (pal[j].red, pal[j].green, pal[j].blue, color_table2, l);
            if (lt_fastxxx->light2[l][j] != m) { err = true; break; }
            m = find_rgb_map (pal[j].red, pal[j].green, pal[j].blue, color_table3, l);
            if (lt_fastxxx->light3[l][j] != m) { err = true; break; }
          }
        }
      }
      if (err)
      {
        CsPrintf (MSG_INITIALIZATION, "  'table_fast_xxx' does not conform to current world file.\n"
                  "  The file will automatically be converted.\n"
                  "  Please be patient...\n");
        recalc_needed = true;
      }
      else
      {
        CsPrintf (MSG_INITIALIZATION, "  Succesfully loaded 'table_fast_xxx'.\n");
        recalc_needed = false;
      }
    }
    else
    {
      CsPrintf (MSG_INITIALIZATION, "  Can't find precalculated 'table_fast_xxx'.\n  \
Will make it now. Please be patient...\n");
      recalc_needed = true;
    }
  }
  else
  {
    CsPrintf (MSG_INITIALIZATION, "  Calculate tables for fast_xxx mode...\n");
    recalc_needed = true; // If no accurate_rgb_mapper we recalc in any case
  }

  if (recalc_needed)
  {
    for (i = 1 ; i < 256 ; i++)
      if (alloc[i])
      {
        r = pal[i].red;
        g = pal[i].green;
        b = pal[i].blue;
        for (l = 0 ; l < 256 ; l++)
        {
          lt_fastxxx->light1[l][i] = find_rgb_map (r, g, b, color_table1, l);
          lt_fastxxx->light2[l][i] = find_rgb_map (r, g, b, color_table2, l);
          lt_fastxxx->light3[l][i] = find_rgb_map (r, g, b, color_table3, l);
        }
      }

    // Color 0 is mapped to 0 in all cases to make sure that transparent
    // colors remain transparent.
    for (l = 0 ; l < 256 ; l++)
    {
      lt_fastxxx->light1[l][0] = 0;
      lt_fastxxx->light2[l][0] = 0;
      lt_fastxxx->light3[l][0] = 0;
    }

    // If accurate_rgb_mapper == false then there is no need to save the tables
    // because they can be recalculated quickly enough.
    if (accurate_rgb_mapper)
    {
      void* entry = ar->new_file ("table_fast_xxx", 256*3*sizeof (RGB8map));
      for (i = 0 ; i < 256 ; i++)
      {
        ar->append (entry, (char*)lt_fastxxx->light1[i], sizeof (RGB8map));
        ar->append (entry, (char*)lt_fastxxx->light2[i], sizeof (RGB8map));
        ar->append (entry, (char*)lt_fastxxx->light3[i], sizeof (RGB8map));
      }
      if (!ar->write_archive ())
        CsPrintf (MSG_WARNING, "  WARNING: error updating tables in archive %s!\n",
                     ar->GetFilename ());
    }
  }
}

void Textures::create_lt_truergb (Archive*)
{
  if (lt_truergb) CHKB (delete lt_truergb);
  CHK (lt_truergb = new TextureTablesTrueRgb);
  lt_light = lt_truergb->lut;
  int i, l;
  int r, g, b;
  int tr, tg, tb;  // Temp working vars

  CsPrintf (MSG_INITIALIZATION, "  Calculate tables for true_rgb mode...\n");

  for (i = 1 ; i < 256 ; i++)
    if (alloc[i])
    {
      r = pal[i].red;
      g = pal[i].green;
      b = pal[i].blue;
      for (l = 0 ; l < 256 ; l++)
      {
        tr = r*l/200;
        tr = tr<0 ? 0 : (tr>255 ? 255 : tr);
        lt_light[i].red[l] = encode_rgb (tr,0,0);

        tg = g*l/200;
        tg = tg<0 ? 0 : (tg>255 ? 255 : tg);
        lt_light[i].green[l] = encode_rgb (0,tg,0);

        tb = b*l/200;
        tb = tb<0 ? 0 : (tb>255 ? 255 : tb);
        lt_light[i].blue[l] = encode_rgb (0,0,tb);

        // Because truecolor 0 (completely black) is used for transparency
        // we can't allow 0 in the normal mappings. So if all lt_light[i].xxx
        // are 0 for one value we will set them to '1'.
        if (lt_light[i].red[l] == 0 && lt_light[i].green[l] == 0 && lt_light[i].blue[l] == 0)
        {
          lt_light[i].red[l] = 1<<pfmt.RedShift;
          lt_light[i].green[l] = 1<<pfmt.GreenShift;
          lt_light[i].blue[l] = 1<<pfmt.BlueShift;
        }
      }
    }

  // Color 0 is mapped to 0 in all cases to make sure that transparent
  // colors remain transparent.
  for (l = 0 ; l < 256 ; l++)
  {
    lt_light[0].red[l] = 0;
    lt_light[0].green[l] = 0;
    lt_light[0].blue[l] = 0;
  }
}

void Textures::create_lt_truergb_private (Archive*)
{
  if (lt_truergb_private) CHKB (delete lt_truergb_private);
  CHK (lt_truergb_private = new TextureTablesTrueRgbPriv);
  lt_light = lt_truergb_private->lut;
  int i, l;
  int tr, tg, tb;  // Temp working vars

  CsPrintf (MSG_INITIALIZATION, "  Calculate tables for true_rgb/private mode...\n");

  for (i = 0 ; i < 256 ; i++)
    if (alloc[i])
    {
      for (l = 0 ; l < 256 ; l++)
      {
        tr = i*l/200;
        tr = tr<0 ? 0 : (tr>255 ? 255 : tr);
        lt_light[i].red[l] = encode_rgb (tr,0,0);

        tg = i*l/200;
        tg = tg<0 ? 0 : (tg>255 ? 255 : tg);
        lt_light[i].green[l] = encode_rgb (0,tg,0);

        tb = i*l/200;
        tb = tb<0 ? 0 : (tb>255 ? 255 : tb);
        lt_light[i].blue[l] = encode_rgb (0,0,tb);

        // Because truecolor 0 (completely black) is used for transparency
        // we can't allow 0 in the normal mappings. So if all lt_light[i].xxx
        // are 0 for one value we will set them to '1'.
        if (lt_light[i].red[l] == 0 && lt_light[i].green[l] == 0 && lt_light[i].blue[l] == 0)
        {
          lt_light[i].red[l] = 1<<pfmt.RedShift;
          lt_light[i].green[l] = 1<<pfmt.GreenShift;
          lt_light[i].blue[l] = 1<<pfmt.BlueShift;
        }
      }
    }

  // Color 0 is mapped to 0 in all cases to make sure that transparent
  // colors remain transparent. @@@ RECHECK FOR PRIVATE COLORMAPS
  for (l = 0 ; l < 256 ; l++)
  {
    lt_light[0].red[l] = 0;
    lt_light[0].green[l] = 0;
    lt_light[0].blue[l] = 0;
  }
}

void Textures::create_lt_alpha (Archive* ar)
{
  if (lt_alpha) CHKB (delete lt_alpha);
  CHK (lt_alpha = new TextureTablesAlpha ());

  int i, j;
  int r, g, b;
  bool recalc_needed = false;
  size_t size;

  // If accurate_rgb_mapper == false we don't bother to use the
  // cached light tables.

  if (accurate_rgb_mapper)
  {
    if (ar->file_exists ("table_alpha", &size))
    {
      bool err = false;
      if (size != 256*2*sizeof (RGB8map)) err = true;
      if (!err)
      {
        char* data = ar->read ("table_alpha");
        char* d = data;

        for (i = 0 ; i < 256 ; i++)
        {
          memcpy (lt_alpha->alpha_map50[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
          memcpy (lt_alpha->alpha_map25[i], d, sizeof (RGB8map)); d += sizeof (RGB8map);
        }
        CHK (delete [] data);

        for (i = 0 ; i < 25 ; i++)
        {
          int m;
          i = (rand () >> 2) & 0xff;
          j = (rand () >> 2) & 0xff;

          if (RESERVED_COLOR(j) || !alloc [j])
            continue;

          r = (pal[i].red + pal[j].red) / 2;
          g = (pal[i].green + pal[j].green) / 2;
          b = (pal[i].blue + pal[j].blue) / 2;
          if (accurate_rgb_mapper) m = find_rgb_slow (r, g, b);
          else m = find_rgb (r, g, b);
          if (lt_alpha->alpha_map50[i][j] != m) { err = true; break; }

          r = (pal[i].red*250 + pal[j].red*750) / 1000;
          g = (pal[i].green*250 + pal[j].green*750) / 1000;
          b = (pal[i].blue*250 + pal[j].blue*750) / 1000;
          if (accurate_rgb_mapper) m = find_rgb_slow (r, g, b);
          else m = find_rgb (r, g, b);
          if (lt_alpha->alpha_map25[i][j] != m) { err = true; break; }
        }
      }
      if (err)
      {
        CsPrintf (MSG_INITIALIZATION, "  'table_alpha' does not conform to current world file.\n  \
The file will automatically be converted.\n  \
Please be patient...\n");
        recalc_needed = true;
      }
      else
      {
        CsPrintf (MSG_INITIALIZATION, "  Succesfully loaded 'table_alpha'.\n");
        recalc_needed = false;
      }
    }
    else
    {
      CsPrintf (MSG_INITIALIZATION, "  Can't find precalculated 'table_alpha'.\n  \
Will make it now. Please be patient...\n");
      recalc_needed = true;
    }
  }
  else
  {
    CsPrintf (MSG_INITIALIZATION, "  Calculate alpha tables...\n");
    recalc_needed = true; // If no accurate_rgb_mapper we recalc in any case
  }

  if (recalc_needed)
  {
    for (i = 0 ; i < 256 ; i++)
      for (j = 0 ; j < 256 ; j++)
      {
        r = (pal[i].red + pal[j].red) / 2;
        g = (pal[i].green + pal[j].green) / 2;
        b = (pal[i].blue + pal[j].blue) / 2;
        if (accurate_rgb_mapper)
          lt_alpha->alpha_map50[i][j] = find_rgb_slow (r, g, b);
        else
          lt_alpha->alpha_map50[i][j] = find_rgb (r, g, b);

        r = (pal[i].red*250 + pal[j].red*750) / 1000;
        g = (pal[i].green*250 + pal[j].green*750) / 1000;
        b = (pal[i].blue*250 + pal[j].blue*750) / 1000;
        if (accurate_rgb_mapper)
          lt_alpha->alpha_map25[i][j] = find_rgb_slow (r, g, b);
        else
          lt_alpha->alpha_map25[i][j] = find_rgb (r, g, b);
      }

    // If accurate_rgb_mapper == false then there is no need to save the tables
    // because they can be recalculated quickly enough.
    if (accurate_rgb_mapper)
    {
      void* entry = ar->new_file ("table_alpha", 256*2*sizeof (RGB8map));
      for (i = 0 ; i < 256 ; i++)
      {
        ar->append (entry, (char*)lt_alpha->alpha_map50[i], sizeof (RGB8map));
        ar->append (entry, (char*)lt_alpha->alpha_map25[i], sizeof (RGB8map));
      }
      if (!ar->write_archive ())
        CsPrintf (MSG_WARNING, "  WARNING: error updating tables in archive %s!\n",
                     ar->GetFilename ());
    }
  }
}

void Textures::compute_light_tables (csWorld* world)
{
  // Light level 200 is normal

  CsPrintf (MSG_INITIALIZATION, "Computing all needed tables...\n");

  Archive* ar = world->get_world_file ();

  if (mixing != MIX_TRUE_RGB && mixing != MIX_NOCOLOR) create_lt_fastxxx (ar);
  if (txtMode == TXT_PRIVATE) create_lt_truergb_private (ar);
  else if (mixing == MIX_TRUE_RGB) create_lt_truergb (ar);
  if (truecolor) create_lt_white16 (ar);
  else create_lt_white8 (ar);
  if (!truecolor) create_lt_alpha (ar);

  CsPrintf (MSG_INITIALIZATION, "DONE\n");
}

void Textures::alloc_palette ()
{
  int i;
  int r, g, b;
  for (i = 0 ; i < 256 ; i++)
    if (alloc[i])
    {
      r = GAMMA (pal[i].red); if (r < 0) r = 0; else if (r > 255) r = 255;
      g = GAMMA (pal[i].green); if (g < 0) g = 0; else if (g > 255) g = 255;
      b = GAMMA (pal[i].blue); if (b < 0) b = 0; else if (b > 255) b = 255;
      Gfx2D->SetRGB (i, r, g, b);
    }

  if (truecolor)
  {
    red_color = encode_rgb (255, 0, 0);
    blue_color = encode_rgb (0, 0, 255);
    yellow_color = encode_rgb (255, 255, 0);
    green_color = encode_rgb (0, 255, 0);
    white_color = encode_rgb (255, 255, 255);
    black_color = 0;
  }
  else
  {
    red_color = find_rgb_slow (255, 0, 0);
    blue_color = find_rgb_slow (0, 0, 255);
    yellow_color = find_rgb_slow (255, 255, 0);
    green_color = find_rgb_slow (0, 255, 0);
    white_color = 255;
    black_color = 0;
  }
}

enum { kTokenTexsMaxT = 1, kTokenTexsTex };

bool Textures::load (char* buf, csWorld* world, Archive* ar)
{
  static tokenDesc commands[] = {
        {kTokenTexsMaxT, "MAX_TEXTURES"},
        {kTokenTexsTex, "TEXTURE"},
        {0,0}};
  char* name;
  long cmd;
  char* params;
  TextureMM* tex;

  int old_num = num_textures;

  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    switch (cmd)
    {
      case kTokenTexsMaxT:
        // ignored for backward compatibility
        break;
      case kTokenTexsTex:
        {
          ImageFile *image = CSLoader::load_image (name, world, ar);
          tex = new_texture (name, image);
          if (!tex->loaded_correctly ())
            return false;
          CSLoader::txt_process(*tex, params);
        }
        break;
    }
  }

  // Allocate all mipmaped textures (but don't initialize them yet).
  int i;
  for (i = old_num ; i < num_textures ; i++)
    textures[i]->alloc_mipmaps ();

  return true;
}

bool Textures::initialize_limited ()
{
  create_lt_palette ();
  return true;
}

bool Textures::initialize (csWorld* world)
{
  int i;

  if (force_accurate_rgb_map != -1) accurate_rgb_mapper = (bool)force_accurate_rgb_map;
  if (force_mix != -1) mixing = force_mix;
  if (force_txtMode != -1) txtMode = force_txtMode;

  if (truecolor)
  {
    mixing = MIX_TRUE_RGB;
    accurate_rgb_mapper = false;
  }
  if (txtMode != TXT_GLOBAL)
  {
    mixing = MIX_TRUE_RGB;
  }

  use_rgb = mixing == MIX_TRUE_RGB;

  switch (txtMode)
  {
    case TXT_GLOBAL:
      CsPrintf (MSG_INITIALIZATION, "One global palette for all textures.\n");
      break;
    case TXT_PRIVATE:
      CsPrintf (MSG_INITIALIZATION, "One private palette for every texture.\n");
      break;
    case TXT_24BIT:
      CsPrintf (MSG_INITIALIZATION, "Code all textures in 24-bit.\n");
      break;
  }

  switch (mixing)
  {
    case MIX_TRUE_RGB:
      CsPrintf (MSG_INITIALIZATION, "Use accurate but slow RGB light mixing.\n");
      color_table1 = TABLE_RED;
      color_table2 = TABLE_GREEN;
      color_table3 = TABLE_BLUE;
      break;
    case MIX_FAST_WRB:
      CsPrintf (MSG_INITIALIZATION, "WRB (white, red, and blue) light mixing.\n");
      color_table1 = TABLE_WHITE;
      color_table2 = TABLE_RED;
      color_table3 = TABLE_BLUE;
      break;
    case MIX_FAST_WRG:
      CsPrintf (MSG_INITIALIZATION, "WRG (white, red, and green) light mixing.\n");
      color_table1 = TABLE_WHITE;
      color_table2 = TABLE_RED;
      color_table3 = TABLE_GREEN;
      break;
    case MIX_FAST_WBG:
      CsPrintf (MSG_INITIALIZATION, "WBG (white, blue, and green) light mixing.\n");
      color_table1 = TABLE_WHITE;
      color_table2 = TABLE_BLUE;
      color_table3 = TABLE_GREEN;
      break;
    case MIX_NOCOLOR:
      CsPrintf (MSG_INITIALIZATION, "Use colorless lights.\n");
      color_table1 = TABLE_WHITE;
      color_table2 = TABLE_WHITE;
      color_table3 = TABLE_WHITE;
      break;
  }

  if (accurate_rgb_mapper)
  {
    if (mixing == MIX_TRUE_RGB)
    {
      accurate_rgb_mapper = false;
      CsPrintf (MSG_INITIALIZATION, "true_rgb mode: fast RGB mapper forced.\n");
    }
    else
      CsPrintf (MSG_INITIALIZATION, "Accurate RGB mapper chosen.\n");
  }
  else
    CsPrintf (MSG_INITIALIZATION, "Fast RGB mapper chosen (slightly lower display quality).\n");

  compute_palette ();
  compute_light_tables (world);

  if (do_blend_mipmap0)
  {
    CsPrintf (MSG_INITIALIZATION, "Blend textures...\n");
    for (i = 0 ; i < num_textures ; i++)
      textures[i]->blend_mipmap0 (this);
  }

  CsPrintf (MSG_INITIALIZATION, "Computing mipmapped textures...\n");
  for (i = 0 ; i < num_textures ; i++)
  {
    textures[i]->create_mipmaps (this);
    textures[i]->free_load_info ();
  }
  CsPrintf (MSG_INITIALIZATION, "DONE!\n");

  world->set_textures(this);
  Gfx3D->SetWorld(GetIWorldFromWorld(world));

  return true;
}

//---------------------------------------------------------------------------
