/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef TEXTURE_H
#define TEXTURE_H

#include "util/types.h"
#include "engine/basic/csobject.h"
#include "render/igraph2d.h"
#include "engine/itexture.h"

/// The default MipMap distances.
#define ZDIST_MIPMAP1 15.
#define ZDIST_MIPMAP2 30.
#define ZDIST_MIPMAP3 60.

/// The prefered distances to use for the color matching.
#define PREFERED_DIST 16333
#define PREFERED_COL_DIST 133333

#define R24(rgb) (((rgb)>>16)&0xff)
#define G24(rgb) (((rgb)>>8)&0xff)
#define B24(rgb) ((rgb)&0xff)

class Texture;
class Textures;
class PolyTexture;
class csWorld;
class Archive;
struct HighColorCache_Data;

/**
 * A matrix for a 3x3 box filter.
 */
struct Filter3x3
{
  int f11, f12, f13;
  int f21, f22, f23;
  int f31, f32, f33;
  int tot;
};

/**
 * A matrix for a 5x5 box filter.
 */
struct Filter5x5
{
  int f00, f01, f02, f03, f04;
  int f10, f11, f12, f13, f14;
  int f20, f21, f22, f23, f24;
  int f30, f31, f32, f33, f34;
  int f40, f41, f42, f43, f44;
  int tot;
};

/**
 * A class containing the private colormap and translation
 * to the global colormap for a single texture. This class
 * is only used in TXT_PRIVATE mode (private colormap for
 * every texture).
 */
class TxtCmapPrivate
{
  friend class TextureMM;
  friend interface IMipMapContainer;

private:
  /**
   * Array with RGB values for every color of the palette.
   * An RGB value has four bytes: R, G, B, unused.
   * So there are 256*4 bytes in this array.
   */
  unsigned char rgb_values[256*4];

  /**
   * Array which maps the private palette index to the global
   * palette index.
   */
  unsigned char priv_to_global[256];

  /// Colors which are allocated.
  unsigned char alloc[256];

private:
  /// Constructor
  TxtCmapPrivate ();

  /**
   * Find an RGB value from the private map and return the
   * index in the private colormap.
   */
  int find_rgb (int r, int g, int b);

  /**
   * Allocate a new RGB color in the private colormap.
   */
  int alloc_rgb (int r, int g, int b, int dist);
};


/**
 * TextureMM represents a texture and all its mipmapped
 * variants.
 */
class TextureMM : public csObject
{
private:
  /// The corresponding ImageFile.
  ImageFile* ifile;
  /// A sorted table with all used colors in the image.
  ImageColorInfo* usage;
  ///
  int transp_red;
  ///
  int transp_green;
  ///
  int transp_blue;
  /// Transparent (-1 if not, 0 if transparency used).
  int transparent;
  /// Mean color used when texture mapping is disabled.
  int mean_color;
  /// Mean color in r,g,b components.
  float mean_r, mean_g, mean_b;

  /// Texture for mipmap level 0
  Texture* t1;
  /// Texture for mipmap level 1
  Texture* t2;
  /// Texture for mipmap level 2
  Texture* t3;
  /// Texture for mipmap level 3
  Texture* t4;

  /// Private colormap if needed.
  TxtCmapPrivate* priv_cmap;

  ///
  HighColorCache_Data *hicolorcache;
 
  /**
   * Create a mipmapped bitmap from a previous level.
   * 'bm', 'w', and 'h' are the specifications of the source bitmap.
   * 'bm2', 'w2', and 'h2' are the specifications for the destination bitmap.
   * 'steps' is the number of steps to mipmap (only 1 or 2 supported)
   */
  void create_mipmap_bitmap (Textures* tex,
  	unsigned char* bm, int w, int h,
	int steps,
  	unsigned char* bm2, int w2, int h2);
  /// Version for global colormap.
  void create_mipmap_bitmap_global (Textures* tex,
  	unsigned char* bm, int w, int h,
	int steps,
  	unsigned char* bm2, int w2, int h2);
  /// Version for private colormap.
  void create_mipmap_bitmap_private (Textures* tex,
  	unsigned char* bm, int w, int h,
	int steps,
  	unsigned char* bm2, int w2, int h2);
  /// Version for 24bit.
  void create_mipmap_bitmap_24bit (Textures* tex,
  	unsigned char* bm, int w, int h,
	int steps,
  	unsigned char* bm2, int w2, int h2);

  /// Create a blended mipmap with the same size as the previous level.
  void create_blended_mipmap (Textures* tex,
  	unsigned char* bm, int w, int h,	// Specs source bitmap
	unsigned char* bm2);			// Dest bitmap
  /// Version for global colormap.
  void create_blended_mipmap_global (Textures* tex,
  	unsigned char* bm, int w, int h,
	unsigned char* bm2);
  /// Version for private colormap.
  void create_blended_mipmap_private (Textures* tex,
  	unsigned char* bm, int w, int h,
	unsigned char* bm2);
  /// Version for 24bit.
  void create_blended_mipmap_24bit (Textures* tex,
  	unsigned char* bm, int w, int h,
	unsigned char* bm2);

public:
  /// Set this to TRUE to ignore image loading errors
  static bool fIgnoreLoadingErrors;

  ///
  TextureMM (char* name, ImageFile* image);
  ///
  virtual ~TextureMM ();

  // DAN: High color cache specific stuff
  ///
  bool in_memory;
  ///
  HighColorCache_Data *get_hicolorcache() { return hicolorcache; }
  ///
  void set_hicolorcache(HighColorCache_Data *d) { hicolorcache = d; }

  /// Allocate all mipmap levels for this texture.
  void alloc_mipmaps ();

  /// Blend mipmap level 0.
  void blend_mipmap0 (Textures* tex);

  /// Create all mipmapped bitmaps from the first level.
  void create_mipmaps (Textures* tex);

  /// Free all information that is only needed for loading the image.
  void free_load_info ();

  /// Return true if the texture has been loaded correctly.
  bool loaded_correctly () { return fIgnoreLoadingErrors || (ifile != NULL); }

  /// Get the mipmapped texture at some level (0..3).
  Texture* get_texture (int lev);

  /// Get the private colormap if any.
  unsigned char* get_colormap_private () { return priv_cmap->rgb_values; }

  /// Get the conversion table from private to global colormap if any.
  unsigned char* get_private_to_global () { return priv_cmap->priv_to_global; }

  /// Set the transparent color.
  void set_transparent (int red, int green, int blue);

  /// Get the transparent index (-1 if no transparency, 0 if transparency).
  int get_transparent () { return transparent; }

  /// Get the transparent color
  void get_transparent (int &red, int &green, int &blue)
  { red = transp_red; green = transp_green; blue = transp_blue; }

  /// Get the mean color.
  int get_mean_color () { return mean_color; }

  ///
  float get_mean_r () { return mean_r; }
  ///
  float get_mean_g () { return mean_g; }
  ///
  float get_mean_b () { return mean_b; }

  ///
  int get_num_colors () { return usage->get_num_colors (); }

  /// Compute the 'usage' table.
  void compute_color_usage ();

  ///
  const RGBPalEntry& get_usage (int idx) 
  { return usage->get_color_table()[idx]; }

  /**
   * Remap the palette of this texture according to the internal
   * texture mode.
   */
  void remap_palette (Textures* new_palette);

  /**
   * Remap the palette of this texture using the given global
   * palette (inside Textures).
   */
  void remap_palette_global (Textures* new_palette);

  /**
   * Remap the palette of this texture using a private
   * palette and make a mapping of this palette to RGB
   * and to the global palette (inside Textures).
   */
  void remap_palette_private (Textures* new_palette);

  /**
   * This function does not really remap but it converts
   * the format to an ULong format suitable for 24-bit
   * internal texture format.
   */
  void remap_palette_24bit (Textures* new_palette);

  DECLARE_INTERFACE_TABLE( TextureMM )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( MipMapContainer )
};

#define GetIMipMapContainerFromTextureMM(a)  &a->m_xMipMapContainer
#define GetTextureMMFromIMipMapContainer(a)  ((TextureMM*)((size_t)a - offsetof(TextureMM, m_xMipMapContainer)))


/**
 * A simple Texture.
 */
class Texture
{
    friend class XTextureMap8;
    friend class XTextureMap32;

protected:
  ///
  TextureMM* parent;
  ///
  int width;
  ///
  int height;
  ///
  int shf_w;
  ///
  int shf_h;
  ///
  int and_w;
  ///
  int and_h;

private:
  ///
  void init ();

protected:
  /**
   * Create a texture with a width and height.
   * This constructor is protected because Texture objects
   * should not be declared directly.
   */
  Texture (TextureMM* p, int w, int h);

public:
  ///
  virtual ~Texture ();

  ///
  int get_width () { return width; }
  ///
  int get_height () { return height; }
  ///
  int get_w_shift () { return shf_w; }
  ///
  int get_h_shift () { return shf_h; }
  ///
  int get_w_mask () { return and_w; }
  ///
  int get_h_mask () { return and_h; }
  ///
  TextureMM* get_parent () { return parent; }
  ///
  virtual unsigned char* get_bitmap8 () = 0;
  ///
  virtual ULong* get_bitmap32 () = 0;
  ///
  virtual void copy (Texture* src) = 0;

// COM stuff //
  DECLARE_IUNKNOWN()

  ITextureMap*  m_pxTextureMap;
};


#define GetITextureMapFromTexture(a)  a->m_pxTextureMap
#define GetTextureFromITextureMap(a)  ((Texture*)a->m_pThis)

/**
 * 8-bit version of this texture.
 */
class Texture8 : public Texture
{
    friend class XTextureMap8;

private:
  /// Raw data.
  unsigned char* bitmap;

public:
  /// Create a texture with a width and height.
  Texture8 (TextureMM* p, int w, int h);
  ///
  virtual ~Texture8 ();
  ///
  virtual unsigned char* get_bitmap8 () { return bitmap; }
  ///
  virtual ULong* get_bitmap32 () { return (ULong*)bitmap; }
  ///
  virtual void copy (Texture* src)
  {
    memcpy (bitmap, src->get_bitmap8 (), width*height);
  }
};

/**
 * 32-bit version of this texture.
 */
class Texture32 : public Texture
{
    friend class XTextureMap32;

private:
  /// Raw data.
  ULong* bitmap;

public:
  /// Create a texture with a width and height.
  Texture32 (TextureMM* p, int w, int h);
  ///
  virtual ~Texture32 ();
  ///
  virtual unsigned char* get_bitmap8 () { return (unsigned char*)bitmap; }
  ///
  virtual ULong* get_bitmap32 () { return bitmap; }
  ///
  virtual void copy (Texture* src)
  {
    memcpy (bitmap, src->get_bitmap32 (), width*height*sizeof (ULong));
  }

  friend class IXTextureMap32;
};

/**
 * Create textures.
 */
class TextureFactory
{
public:
  /// Create a texture.
  virtual Texture* new_texture (TextureMM* parent, int w, int h) = 0;
};

/**
 * 8-bit texture creator.
 */
class TextureFactory8 : public TextureFactory
{
public:
  /// Create a texture.
  virtual Texture* new_texture (TextureMM* parent, int w, int h)
  {
    CHK (Texture8* txt = new Texture8 (parent, w, h));
    return txt;
  }
};

/**
 * 32-bit texture creator.
 */
class TextureFactory32 : public TextureFactory
{
public:
  /// Create a texture.
  virtual Texture* new_texture (TextureMM* parent, int w, int h)
  {
    CHK (Texture32* txt = new Texture32 (parent, w, h));
    return txt;
  }
};

/**
 * This class contains all textures and also maintains
 * the shared palette.
 */
class Textures
{
private:
  /// List of textures.
  TextureMM** textures;
  ///
  int num_textures;
  ///
  int max_textures;

  /// Pixel format.
  csPixelFormat pfmt;
  int num_red, num_green, num_blue;

  /// True if truecolor mode is enabled.
  bool truecolor;

  /**
   * The shared palette. This palette is also used in truecolor (16-bit) mode.
   * It then contains the single palette that is shared by all textures.
   */
  RGBcolor pal[256];
  /// Which colors are allocated and which are not?
  int alloc[256];

  int red_color;
  int yellow_color;
  int green_color;
  int blue_color;
  int white_color;
  int black_color;
  /// Ambient color for red.
  int level_r;
  /// Ambient color for green.
  int level_g;
  /// Ambient color for blue.
  int level_b;

  /// Configuration values for color matching.
  int prefered_dist;
  /// Configuration values for color matching.
  int prefered_col_dist;

  /// Read configuration values from config file.
  void read_config ();

  /**
   * Encode RGB values to a 16-bit word (for 16-bit mode).
   */
  UShort encode_rgb (int r, int g, int b);

  /**
   * Encode RGB values to 16-bit word (safe mode).
   */
  UShort encode_rgb_safe (int r, int g, int b);

  /// For optimization: points to table in lt_pal.
  unsigned char* lt_palette_table;
  /// For optimization: points to tables in lt_truergb or in lt_truergb_priv.
  PalIdxLookup* lt_light;

  /**
   * Create the palette lookup table. The colormap must
   * be created before this can be used.
   */
  void create_lt_palette ();

  /**
   * Create the lookup tables for any of the fast_xxx modes.
   * This function will possibly read the lookup tables from disk
   * (with the name: table_fast_xxx) and also write it on disk
   * if needed.
   */
  void create_lt_fastxxx (Archive* ar);

  /**
   * Create the lookup tables for the white tables in 16-bit display
   * mode.
   */
  void create_lt_white16 (Archive* ar);

  /**
   * Create the lookup tables for the white tables in 8-bit display mode.
   * The tables are cached if needed (with the name 'table_white8').
   */
  void create_lt_white8 (Archive* ar);

  /**
   * Create the truergb 16-bit tables.
   */
  void create_lt_truergb (Archive* ar);

  /**
   * Create the truergb 16-bit tables for TXT_PRIVATE mode.
   */
  void create_lt_truergb_private (Archive* ar);

  /**
   * Create the alpha tables.
   * The tables are cached if needed (with the name 'table_alpha').
   */
  void create_lt_alpha (Archive* ar);

public:
  /// Lookup table.
  TextureTablesTrueRgb* lt_truergb;
  /// Lookup table.
  TextureTablesTrueRgbPriv* lt_truergb_private;
  /// Lookup table.
  TextureTablesFastXxx* lt_fastxxx;
  /// Lookup table.
  TextureTablesWhite16* lt_white16;
  /// Lookup table.
  TextureTablesWhite8* lt_white8;
  /// Lookup table.
  TextureTablesPalette* lt_pal;
  /// Lookup table.
  TextureTablesAlpha* lt_alpha;
  /// Texture gamma
  static float Gamma;

  /// Configuration value: how is mipmapping done (one of MIPMAP_...)
  static int mipmap_nice;

  /// Configuration value: blend mipmap level 0.
  static bool do_blend_mipmap0;

  /// For debugging: show lightmapgrid.
  static bool do_lightmapgrid;

  /// For debugging: don't show textures but only map lightmaps.
  static bool do_lightmaponly;

  /// Filter to use when mipmapping 1 level.
  static Filter3x3 mipmap_filter_1;

  /// Filter to use when mipmapping 2 levels.
  static Filter5x5 mipmap_filter_2;

  /// Filter to use when performing blending.
  static Filter3x3 blend_filter;

  /// Alpha mask used for 16-bit mode.
  static UShort alpha_mask;

  /// How is table 1 used (one of TABLE_...)
  static int color_table1;
  /// How is table 2 used (one of TABLE_...)
  static int color_table2;
  /// How is table 3 used (one of TABLE_...)
  static int color_table3;

  /// How do we mix colors (one of MIX_...)
  static int mixing;
  /// Force value (set by commandline) (-1 = no force)
  static int force_mix;

  /// How are texture represented internally.
  static int txtMode;
  /// Force value (set by commandline) (-1 = no force)
  static int force_txtMode;

  /// Ambient red value.
  static int ambient_red;
  /// Ambient green value.
  static int ambient_green;
  /// Ambient blue value.
  static int ambient_blue;
  /// Ambient white value.
  static int ambient_white;
  /// If true we have R,G,B lights.
  static bool use_rgb;
  /**
   * If true we use the accurate RGB mapper. This field
   * can only be true if we are using any of the FAST_xxx
   * or NOCOLOR mixing modes. In all other modes it is not
   * relevant to use the accurate rgb mapper.
   */
  static bool accurate_rgb_mapper;

  /// Force value (set by commandline) (-1 = no force)
  static int force_accurate_rgb_map;

private:
  ///
  int get_texture_idx (char* name);
  ///
  Texture* get_texture (int idx, int lev);

  /**
   * Find rgb for a specific map type and apply an intensity.
   * 'map_type' is one of TABLE_....
   */
  int find_rgb_map (int r, int g, int b, int map_type, int l);

public:
  ///
  Textures (int max);
  ///
  virtual ~Textures ();

  /**
   * Initialize the texture system. This function has to be called
   * AFTER the general system has been set up (it needs to know
   * the depth of the display and the bitformat of the RGB pixels
   * in 15/16 bit truecolor mode). After you have called this
   * function you can call alloc_palette().<br>
   * Returns true if successful.
   */
  bool initialize (csWorld* world);

  /**
   * Initialize the texture system in a limited way. This is
   * useful if you want to use the texture system for unlighted
   * drawing only.
   */
  bool initialize_limited ();

  /// Create a new texture.
  TextureMM* new_texture (char* name, ImageFile* image);

  /// Return number of textures
  int get_num_textures () { return num_textures; }

  /// Return texture by index
  TextureMM* get_texture_mm (int idx) { return textures[idx]; }

  /// Find a texture given a name.
  TextureMM* get_texture_mm (char* name);

  /// Add a texture
  void add_texture (TextureMM* tm);

  /// Delete a texture
  void delete_texture (int idx, bool del = true);

  /**
   * Find an rgb value using the palette directly (not use
   * the faster lookup tables).
   */
  int find_rgb_slow (int r, int g, int b);

  /**
   * Allocate a new RGB color.
   */
  int alloc_rgb (int r, int g, int b, int dist);

  ///
  static bool force_mixing (char* mix);

  ///
  static bool force_accurate_rgb_mapper (bool rgb);

  ///
  static bool force_txtmode (char* txtmode);

  /**
   * Find an rgb value using the faster lookup tables.
   */
  int find_rgb (int r, int g, int b);

  /**
   * Return the index for some color. This works in 8-bit
   * (returns an index in the 256-color table) and in 15/16-bit
   * (returns a 15/16-bit encoded RGB value).
   */
  int find_color (int r, int g, int b);

  /**
   * This version of find_rgb finds some r,g,b value AFTER gamma
   * correction is applied. This is useful for console messages
   * that always need the same color regardless of the gamma
   * correction.
   */
  int find_rgb_real (int r, int g, int b);

  /**
   * Compute the 'best' palette for all loaded textures.
   * This function will exactly behave the same in 16-bit mode
   * since we still need the common 256-color palette (although
   * it will not be installed on the display).
   */
  void compute_palette ();

  /**
   * Compute the light tables using the previously compute palette.
   */
  void compute_light_tables (csWorld* world);

  /**
   * Really allocate the palette on the system.
   */
  void alloc_palette ();

  /**
   * Get the palette.
   */
  RGBcolor* get_palette () { return pal; }

  ///
  int get_pal_alloced (int idx) { return alloc[idx]; }
  ///
  int get_pal_red (int idx) { return pal[idx].red; }
  ///
  int get_pal_green (int idx) { return pal[idx].green; }
  ///
  int get_pal_blue (int idx) { return pal[idx].blue; }

  ///
  int get_level_red () { return level_r; }
  ///
  int get_level_green () { return level_g; }
  ///
  int get_level_blue () { return level_b; }

  /**
   * Load all the texture descriptions from the world
   * file (no actual images).
   * If the optional 'ar' is given, that archive is checked
   * first before the others are checked.
   */
  bool load (char* buf, csWorld* world, Archive* ar = NULL);

  ///
  int red () { return red_color; }
  ///
  int blue () { return blue_color; }
  ///
  int yellow () { return yellow_color; }
  ///
  int green () { return green_color; }
  ///
  int white () { return white_color; }
  ///
  int black () { return black_color; }

  /**
   * GAC: Use a slower method of mixing lights that produces a better result
   * Accepts 16.16 rgb values and a palette entry
   * Returns palette entry of color with lights added
   * Assumes lights are RGB format (Not WRB).
   */
  int mix_lights (int r, int g, int b, int p)
  {
    PalIdxLookup* pil = lt_light+p;
    return lt_palette_table[pil->red[r>>16] |
			   pil->green[g>>16] |
			   pil->blue[b>>16]];
  }

  /**
   * Add some light to some light component. This is used for TXT_PRIVATE
   * mode. This returns a 16-bit truecolor component for the specific color.
   */
  int add_light_red_private (int r, int l)
  {
    return lt_light[r].red[l>>16];
  }
  ///
  int add_light_green_private (int g, int l)
  {
    return lt_light[g].green[l>>16];
  }
  ///
  int add_light_blue_private (int b, int l)
  {
    return lt_light[b].blue[l>>16];
  }

  /**
   * Version of mix_lights for 16-bit mode. This is slightly faster since
   * the last lookup is not needed.
   */
  UShort mix_lights_16 (int r, int g, int b, int p)
  {
    PalIdxLookup* pil = lt_light+p;
    return pil->red[r>>16] |
	   pil->green[g>>16] |
	   pil->blue[b>>16];
  }

  DECLARE_INTERFACE_TABLE( Textures )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( TextureContainer )
};

#define GetITextureContainerFromTextures(a)  &a->m_xTextureContainer;
#define GetTexturesFromITextureContainer(a)  ((Textures*)((size_t)a - offsetof(Textures, m_xTextureContainer)))

#endif /*TEXTURE_H*/
