/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "util/def.h"
#include "system/system.h"
#include "util/types.h"
#include "geometry/math3d.h"
#include "geometry/csrect.h"
#include "render/graph2d.h"
#include "system/sysprint.h"
#include "engine/texture.h"
#include "engine/wirefrm.h"
#include "engine/camera.h"
#include "render/igraph3d.h"

wfColor::wfColor (Textures* txt, int r, int g, int b)
{
  wfColor::r = r;
  wfColor::g = g;
  wfColor::b = b;
  int i;
  for (i = 0 ; i < 16 ; i++)
    col_idx[i] = txt->find_color (r*(20-i)/20, g*(20-i)/20, b*(20-i)/20);
}

int wfColor::get_color (float z)
{
  int zi = QInt (z);
  zi /= 2;
  if (zi > 15) zi = 15;
  return col_idx[zi];
}


wfObject::wfObject ()
{
  next = prev = NULL;
}

wfObject::~wfObject ()
{
}

bool wfObject::perspective (csCamera* c, Vector3& v, Vector2& persp, float radius, float& pradius)
{
  Vector3 cam;
  cam = c->world2this(v);
  if (cam.z > SMALL_Z)
  {
    float iz = csCamera::aspect/cam.z;
    persp.x = cam.x * iz + (FRAME_WIDTH/2);
    persp.y = FRAME_HEIGHT - 1 - (cam.y * iz + (FRAME_HEIGHT/2));
    pradius = radius * iz;
    return true;
  }
  else return false;
}

void wfVertex::draw (IGraphics3D* g, csCamera* c)
{
  Vector2 persp;
  float rad;
  int r, px, py;

  if (perspective (c, loc, persp, PLANE_VERTEX_RADIUS, rad))
  {
    Vector3 cam;
    cam = c->world2this (loc);
    int col = color->get_color (cam.z);

    r = QInt (rad);
    px = QInt (persp.x);
    py = QInt (persp.y);
    Gfx2D->DrawLine (px-r, py-r, px+r, py+r, col);
    Gfx2D->DrawLine (px+r, py-r, px-r, py+r, col);
    //if (cross)
    //{
      //Gfx2D->DrawLine (px-r, py, px+r, py, col);
      //Gfx2D->DrawLine (px, py-r, px, py+r, col);
    //}
  }
}

void wfLine::draw (IGraphics3D* g, csCamera* c)
{
  Vector3 cam1, cam2;
  cam1 = c->world2this (v1);
  cam2 = c->world2this (v2);
  g->DrawLine (cam1, cam2, color->get_color ((cam1.z+cam2.z)/2));
}

wfPolygon::wfPolygon ()
{
  vertices = NULL;
}

wfPolygon::~wfPolygon ()
{
  if (vertices) CHKB (delete [] vertices);
}

void wfPolygon::set_num_vertices (int n)
{
  num_vertices = n;
  if (vertices) CHKB (delete [] vertices);
  CHK (vertices = new Vector3 [num_vertices]);
}

void wfPolygon::set_vertex (int i, Vector3& v)
{
  if (i >= num_vertices)
  {
    CsPrintf (MSG_INTERNAL_ERROR, "Bad vertex number %d in wfPolygon::set_vertex!\n", i);
    fatal_exit (0, false);
  }
  vertices[i] = v;
}

void wfPolygon::prepare ()
{
  A = 0;
  B = 0;
  C = 0;
  int i, i1;
  float x1, y1, z1, x, y, z;

  Vector3 vmin (1000000,1000000,1000000);
  Vector3 vmax (-1000000,-10000000,-1000000);
  i1 = num_vertices-1;
  for (i = 0 ; i < num_vertices ; i++)
  {
    x = vertices[i].x;
    y = vertices[i].y;
    z = vertices[i].z;
    if (x < vmin.x) vmin.x = x;
    if (y < vmin.y) vmin.y = y;
    if (z < vmin.z) vmin.z = z;
    if (x > vmax.x) vmax.x = x;
    if (y > vmax.y) vmax.y = y;
    if (z > vmax.z) vmax.z = z;
    x1 = vertices[i1].x;
    y1 = vertices[i1].y;
    z1 = vertices[i1].z;
    A += (z1+z) * (y-y1);
    B += (x1+x) * (z-z1);
    C += (y1+y) * (x-x1);
    i1 = i;
  }

  float d = sqrt (A*A + B*B + C*C);

  if (d < SMALL_EPSILON) d = SMALL_EPSILON;

  A /= d;
  B /= d;
  C /= d;

  x = vertices[0].x;
  y = vertices[0].y;
  z = vertices[0].z;
  D = -(A*x+B*y+C*z);

  center.x = (vmin.x+vmax.x)/2;
  center.y = (vmin.y+vmax.y)/2;
  center.z = (vmin.z+vmax.z)/2;
}

bool wfPolygon::visible (csCamera* camera)
{
  return Plane::classify (A, B, C, D, camera->v_origin()) < 0;
}

void wfPolygon::draw (IGraphics3D* g, csCamera* c)
{
  int i;
  Vector3 cam1, cam2, cen;
  bool vis = visible (c);
  cen = c->world2this (center);
  int col = color->get_color (cen.z);
  int vcol = vcolor->get_color (cen.z);
  for (i = 0 ; i < num_vertices ; i++)
  {
    cam1 = c->world2this (vertices[i]);
    cam2 = c->world2this (vertices[(i+1)%num_vertices]);
    g->DrawLine (cam1, cam2, col);
    if (vis) g->DrawLine (cam1, cen, vcol);
  }
}

//=================================================================================================

WireFrame::WireFrame (Textures* txt)
{
  objects = NULL;
  numObjects = 0;
  WireFrame::txt = txt;
  colors = NULL;
  white = register_color (255, 255, 255);
  red = register_color (255, 0, 0);
  green = register_color (0, 255, 0);
  blue = register_color (0, 0, 255);
  yellow = register_color (255, 255, 0);
}

WireFrame::~WireFrame ()
{
  clear ();
  while (colors)
  {
    wfColor* n = colors->next;
    CHK (delete colors);
    colors = n;
  }
}

void WireFrame::clear ()
{
  while (objects)
  {
    wfObject* n = objects->get_next ();
    CHK (delete objects);
    objects = n;
  }
  numObjects = 0;
}

wfColor* WireFrame::find_color (int r, int g, int b)
{
  wfColor* c = colors;
  while (c)
  {
    if (c->r == r && c->g == g && c->b == b) return c;
    c = c->next;
  }
  return NULL;
}

wfColor* WireFrame::register_color (int r, int g, int b)
{
  wfColor* c = find_color (r, g, b);
  if (c) return c;
  CHK (c = new wfColor (txt, r, g, b));
  c->next = colors;
  colors = c;
  return c;
}

wfVertex* WireFrame::add_vertex (Vector3& v)
{
  CHK (wfVertex* vt = new wfVertex ());
  vt->set_location (v);
  vt->set_color (white);
  vt->set_next (objects);
  if (objects) objects->set_prev (vt);
  objects = vt;

  numObjects++;

  return vt;
}

wfLine* WireFrame::add_line (Vector3& v1, Vector3& v2)
{
  CHK (wfLine* li = new wfLine ());
  li->set_line (v1, v2);
  li->set_color (white);
  li->set_next (objects);
  if (objects) objects->set_prev (li);
  objects = li;

  numObjects++;

  return li;
}

wfPolygon* WireFrame::add_polygon ()
{
  CHK (wfPolygon* po = new wfPolygon ());
  po->set_color (white);
  po->set_next (objects);
  if (objects) objects->set_prev (po);
  objects = po;

  numObjects++;

  return po;
}

void WireFrame::draw (IGraphics3D* g, csCamera* c)
{
  wfObject* o = objects;
  while (o)
  {
    o->draw (g, c);
    o = o->get_next ();
  }
}

void WireFrame::apply( void (*func)( wfObject*, void*), void *param )
{
  wfObject* o = objects;
  while (o)
  {
    func(o, param);
    o = o->get_next ();
  }
}

WireFrameCam::WireFrameCam (Textures* txt)
{
  CHK (wf = new WireFrame (txt));
  CHK (c = new csCamera ());
}

WireFrameCam::~WireFrameCam ()
{
  CHK (delete wf);
  CHK (delete c);
}

void WireFrameCam::key_up (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->moveip (.01*VEC_UP);
    else if (shift) c->moveip (.4*VEC_UP);
    else c->moveip (.2*VEC_UP);
  }
  else if (ctrl) c->moveip (.01*VEC_FORWARD);
  else if (shift) c->moveip (1.2*VEC_FORWARD);
  else c->moveip (.6*VEC_FORWARD);
}

void WireFrameCam::key_down (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->moveip (.01*VEC_DOWN);
    else if (shift) c->moveip (.4*VEC_DOWN);
    else c->moveip (.2*VEC_DOWN);
  }
  else if (ctrl) c->moveip (.01*VEC_BACKWARD);
  else if (shift) c->moveip (1.2*VEC_BACKWARD);
  else c->moveip (.6*VEC_BACKWARD);
}

void WireFrameCam::key_left (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->moveip (.01*VEC_LEFT);
    else if (shift) c->moveip (.4*VEC_LEFT);
    else c->moveip (.2*VEC_LEFT);
  }
  else if (ctrl) c->rotate (VEC_ROT_LEFT, .005);
  else if (shift) c->rotate (VEC_ROT_LEFT, .2);
  else c->rotate (VEC_ROT_LEFT, .1);
}

void WireFrameCam::key_right (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->moveip (.01*VEC_RIGHT);
    else if (shift) c->moveip (.4*VEC_RIGHT);
    else c->moveip (.2*VEC_RIGHT);
  }
  else if (ctrl) c->rotate (VEC_ROT_RIGHT, .005);
  else if (shift) c->rotate (VEC_ROT_RIGHT, .2);
  else c->rotate (VEC_ROT_RIGHT, .1);
}

void WireFrameCam::key_pgup (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->rotate (VEC_TILT_LEFT, .005);
    else c->rotate (VEC_TILT_LEFT, .1);
  }
  else if (ctrl) c->rotate (VEC_TILT_UP, .005);
  else if (shift) c->rotate (VEC_TILT_UP, .2);
  else c->rotate (VEC_TILT_UP, .1);
}

void WireFrameCam::key_pgdn (bool shift, bool alt, bool ctrl)
{
  if (alt)
  {
    if (ctrl) c->rotate (VEC_TILT_RIGHT, .005);
    else c->rotate (VEC_TILT_RIGHT, .1);
  }
  else if (ctrl) c->rotate (VEC_TILT_DOWN, .005);
  else if (shift) c->rotate (VEC_TILT_DOWN, .2);
  else c->rotate (VEC_TILT_DOWN, .1);
}

