/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef WIREFRM_H
#define WIREFRM_H

#include "util/def.h"
#include "geometry/math3d.h"
#include "geometry/math2d.h"

class Textures;
interface IGraphics3D;
class csCamera;

#define PLANE_VERTEX_RADIUS .1

/**
 * A color for an object in a WireFrame.
 * This color object is special in that it provides support
 * for depth based dimming.
 */
class wfColor
{
public:
  wfColor* next;	// Next color.
  int r, g, b;		// Base color.
  int col_idx[16];	// 16-levels based on distance.

public:
  ///
  wfColor (Textures* txt, int r, int g, int b);
  /// Get a color based on depth.
  int get_color (float z);
};

/**
 * Object in a WireFrame.
 */
class wfObject
{
private:
  wfObject* next, * prev;

protected:
  wfColor* color;

public:
  ///
  wfObject ();
  ///
  virtual ~wfObject ();

  /// Set the prefered color.
  void set_color (wfColor* col) { color = col; }
  ///
  wfColor* get_color () { return color; }

  ///
  wfObject* get_next () { return next; }
  ///
  wfObject* get_prev () { return prev; }
  ///
  void set_next (wfObject* ob) { next = ob; }
  ///
  void set_prev (wfObject* ob) { prev = ob; }

  /// Draw this object given a camera.
  virtual void draw (IGraphics3D* g, csCamera* c) = 0;

  /// Do the perspective transform.
  bool perspective (csCamera* c, Vector3& v, Vector2& persp, float radius, float& pradius);
};

/**
 * Vertex in a WireFrame.
 */
class wfVertex : public wfObject
{
private:
  Vector3 loc;

public:
  ///
  Vector3& get_location () { return loc; }

  ///
  void set_location (Vector3& v) { loc = v; }
  ///
  void set_location (float x, float y, float z) { loc.x = x; loc.y = y; loc.z = z; }

  ///
  virtual void draw (IGraphics3D* g, csCamera* c);
};

/**
 * Line in a WireFrame.
 */
class wfLine : public wfObject
{
private:
  Vector3 v1, v2;

public:
  ///
  Vector3& get_vertex1 () { return v1; }
  ///
  Vector3& get_vertex2 () { return v2; }
  ///
  void set_line (Vector3& vv1, Vector3& vv2) { v1 = vv1; v2 = vv2; }

  ///
  virtual void draw (IGraphics3D* g, csCamera* c);
};

/**
 * A polygon in a WireFrame.
 */
class wfPolygon : public wfObject
{
private:
  int num_vertices;
  Vector3* vertices;
  Vector3 center;		// Calculated center of polygon.
  float A, B, C, D;		// Plane equation of polygon.
  wfColor* vcolor;		// Visibility color.

public:
  ///
  wfPolygon ();
  ///
  virtual ~wfPolygon ();

  ///
  void set_num_vertices (int n);
  ///
  int get_num_vertices () { return num_vertices; }
  ///
  void set_vertex (int i, Vector3& v);

  /// Set the prefered color for the visibility lines.
  void set_vis_color (wfColor* vcol) { vcolor = vcol; }

  /// This function should be called after all vertices have been set.
  void prepare ();

  /// Return true if object is visible from a given camera.
  bool visible (csCamera* camera);

  ///
  virtual void draw (IGraphics3D* g, csCamera* c);
};

/**
 * A WireFrame is a set of lines, vertices and polygons in 3D.
 * You can draw it given a camera.
 */
class WireFrame
{
private:
  wfObject* objects;
  Textures* txt;
  wfColor* colors;
  wfColor* white;
  wfColor* red;
  wfColor* green;
  wfColor* blue;
  wfColor* yellow;

  size_t numObjects;

public:
  ///
  WireFrame (Textures* txt);
  ///
  virtual ~WireFrame ();

  ///
  void clear ();

  ///
  size_t entries() { return numObjects; }

  /// Get the default white color.
  wfColor* get_white () { return white; }

  /// Get the default red color.
  wfColor* get_red () { return red; }

  /// Get the default blue color.
  wfColor* get_blue () { return blue; }

  /// Get the default green color.
  wfColor* get_green () { return green; }

  /// Get the default yellow color.
  wfColor* get_yellow () { return yellow; }

  /**
   * Find a registered color.
   */
  wfColor* find_color (int r, int g, int b);

  /**
   * Register a color so that it can be used by objects in
   * the wireframe. This is needed because colors are
   * represented by an array (darker colors for objects in the
   * distance).<p>
   * The returned value is the one that should be given to the
   * wfObject::set_color function.
   */
  wfColor* register_color (int r, int g, int b);

  ///
  wfVertex* add_vertex (Vector3& v);
  ///
  wfLine* add_line (Vector3& v1, Vector3& v2);
  ///
  wfPolygon* add_polygon ();

  ///
  void draw (IGraphics3D* g, csCamera* c);

  /// Apply a function to all objects contained in the WireFrame
  void apply( void (*func)( wfObject*, void*), void* );
};

/**
 * A WireFrame associated with a camera.
 */
class WireFrameCam
{
private:
  WireFrame* wf;
  csCamera* c;

public:
  ///
  WireFrameCam (Textures* txt);
  ///
  virtual ~WireFrameCam ();

  ///
  WireFrame* get_wireframe () { return wf; }
  ///
  csCamera* get_camera () { return c; }

  ///
  void key_up (bool shift, bool alt, bool ctrl);
  ///
  void key_down (bool shift, bool alt, bool ctrl);
  ///
  void key_left (bool shift, bool alt, bool ctrl);
  ///
  void key_right (bool shift, bool alt, bool ctrl);
  ///
  void key_pgdn (bool shift, bool alt, bool ctrl);
  ///
  void key_pgup (bool shift, bool alt, bool ctrl);
};

#endif /*WIREFRM_H*/

