/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/system.h"
#include "system/sysprint.h"
#include "engine/world.h"
#include "engine/camera.h"
#include "engine/light/light.h"
#include "engine/light/dynlight.h"
#include "engine/polygon/polyplan.h"
#include "engine/polygon/polygon.h"
#include "engine/objects/thingtpl.h"
#include "engine/objects/thing.h"
#include "engine/objects/cssprite.h"
#include "engine/basic/csobject.h"
#include "engine/basic/cscoll.h"
#include "engine/sector.h"
#include "engine/library.h"
#include "scripts/primscri.h"
#include "scripts/intscri.h"
#include "engine/texture.h"
#include "util/token.h"
#include "engine/light/lghtmap.h"
#include "util/util.h"
#include "util/parser.h"
#include "engine/stats.h"
#include "scripts/csscript.h"
#include "render/igraph3d.h"
#include "util/archive.h"
#include "util/inifile.h"
#include "engine/wirefrm.h"
#include "general/loader/csloader.h"
#include "render/ihalo.h"

//---------------------------------------------------------------------------

// COM IUnknown implementation
IMPLEMENT_UNKNOWN_NODELETE( csWorld )

BEGIN_INTERFACE_TABLE( csWorld )
	IMPLEMENTS_COMPOSITE_INTERFACE( World )
END_INTERFACE_TABLE()

csWorld::csWorld ()
{
  first_dyn_lights = NULL;
  first_mso = NULL;
  textures = NULL;
  world_file = NULL;
  Gfx3D->SetWorld (GetIWorldFromWorld(this));
  System->world = this;
  wf = NULL;
  map_mode = MAP_OFF;
  start_sector = NULL;
  start_vec.x = start_vec.y = start_vec.z = 0;
  piHR = NULL;
}

csWorld::~csWorld ()
{
  clear ();
  CHK (delete wf);
  if (piHR) { FINAL_RELEASE (piHR); piHR = NULL; }
}

void csWorld::clear ()
{
  sectors.DeleteAll ();
  libraries.DeleteAll ();
  planes.DeleteAll ();
  scripts.DeleteAll ();
  collections.DeleteAll ();
  sprite_templates.DeleteAll ();
  thing_templates.DeleteAll ();
  waves.DeleteAll ();

  while (first_dyn_lights)
  {
    csDynLight* dyn = first_dyn_lights->get_next ();
    CHK (delete first_dyn_lights);
    first_dyn_lights = dyn;
  }
  while (first_mso)
  {
    MultiSectorObject* mso = first_mso->get_next ();
    if (first_mso->get_object ()) { CHK (delete first_mso->get_object ()); }
    CHK (delete first_mso);
    first_mso = mso;
  }
  CHK (delete textures); textures = NULL;
  CHK (delete world_file); world_file = NULL;
  CHK (delete [] start_sector); start_sector = NULL;
}

csSpriteTemplate* csWorld::get_sprite_template (char* name, bool use_libs)
{
  // if use_libs == true we should support qualified names (like standard.bullet).
  // with the library first and then the name of the object.
  int sn = sprite_templates.Length ();
  while (sn > 0)
  {
    sn--;
    csSpriteTemplate* s = (csSpriteTemplate*)sprite_templates[sn];
    if (!strcmp (name, s->get_name ())) return s;
  }

  if (use_libs)
  {
    int nl = libraries.Length ();
    while (nl > 0)
    {
      nl--;
      Library* lib = (Library*)libraries[nl];
      csSpriteTemplate* s = (csSpriteTemplate*)lib->sprite_templates.FindName (name);
      if (s) return s;
    }
  }

  return NULL;
}

ThingTemplate* csWorld::get_thing_template (char* name, bool use_libs)
{
  int tn = thing_templates.Length ();
  while (tn > 0)
  {
    tn--;
    ThingTemplate* s = (ThingTemplate*)thing_templates[tn];
    if (!strcmp (name, s->get_name ())) return s;
  }

  if (use_libs)
  {
    int nl = libraries.Length ();
    while (nl > 0)
    {
      nl--;
      Library* lib = (Library*)libraries[nl];
      ThingTemplate* s = (ThingTemplate*)lib->thing_templates.FindName (name);
      if (s) return s;
    }
  }
  return NULL;
}

MultiSectorObject* csWorld::get_mso (char* name)
{
  MultiSectorObject* s = first_mso;
  while (s)
  {
    csObject* o = s->get_object ();
    if (o && !strcmp (name, o->get_name ())) return s;
    s = s->get_next ();
  }
  return NULL;
}

void csWorld::add (MultiSectorObject* mso)
{
  if (mso)
  {
    mso->set_next (first_mso);
    if (first_mso) first_mso->set_prev (mso);
    first_mso = mso;
    first_mso->set_prev (NULL);
    mso->set_world (this);
  }
}

void csWorld::remove (MultiSectorObject* mso)
{
  if (mso->get_next ()) mso->get_next ()->set_prev (mso->get_prev ());
  if (mso->get_prev ()) mso->get_prev ()->set_next (mso->get_next ());
  else first_mso = mso->get_next ();
  mso->set_world (NULL);
  mso->set_next (NULL);
  mso->set_prev (NULL);
}

csSector* csWorld::new_sector (char* name)
{
  CHK (csSector* s = new csSector (name));
  s->set_ambient_color (textures->get_level_red (), textures->get_level_green (), textures->get_level_blue ());
  sectors.Push (s);
  return s;
}

csScript* csWorld::new_script (LanguageLayer* layer, char* name, char* params)
{
  if (!layer)
    return NULL;

  csScript* s = NULL;
  char* par = strchr (params, ':');
  if (!par)
  {
    CsPrintf (MSG_FATAL_ERROR, "Missing language qualifier for script '%s'!\n", name);
    fatal_exit (0, false);
  }

  if (scripts.FindName (name))
  {
    CsPrintf (MSG_FATAL_ERROR, "A script with the name '%s' is already defined!\n", name);
    fatal_exit (0, false);
  }

  // All different script languages should be recognized here.
  if (!strncmp (params, "prim", (int)(par-params)))
  {
    CHK (PrimScript* sc = new PrimScript (layer, name));
    par++;
    sc->load (&par);
    s = (csScript*)sc;
  }
  else if (!strncmp (params, "int", (int)(par-params)))
  {
    CHK (IntScript* sc = new IntScript (layer, name));
    par++;
    sc->load (par);
    s = (csScript*)sc;
  }
  else
  {
    CsPrintf (MSG_FATAL_ERROR, "Unknown script qualifier for script '%s'!\n", name);
    fatal_exit (0, false);
    return NULL;
  }

  s->prepare ();
  scripts.Push (s);

  return s;
}

Thing* csWorld::get_thing (char* name)
{
  int i = sectors.Length ();
  while (i > 0)
  {
    i--;
    csSector* s = (csSector*)sectors[i];
    Thing* t = s->get_thing (name);
    if (t) return t;
  }
  return NULL;
}

void csWorld::trigger_activate (csCamera& c)
{
  Vector3 where;
  where = c.this2world (3*VEC_FORWARD);
  Polygon3D* p = c.get_hit (where);
  if (p)
  {
    CsPrintf (MSG_CONSOLE, "Activate polygon '%s' ", p->get_name ());
    PolygonSet* ob = (PolygonSet*)(p->get_parent ());
    CsPrintf (MSG_CONSOLE, "in set '%s'\n", ob->get_name ());
    ob->do_activate_triggers ();
  }
}

enum { kTokenWorldSector = 1, kTokenWorldPlane, kTokenWorldCol,
       kTokenWorldScript, kTokenWorldTexs, kTokenWorldLightX,
       kTokenWorldRoom, kTokenWorldThing, kTokenWorldSixface,
       kTokenWorldSpriteTmp, kTokenWorldLibrary, kTokenWorldStart,
       kTokenWorldWaves };

bool csWorld::load (LanguageLayer* layer, char* buf)
{
  static tokenDesc tokens[] = {{1, "WORLD"}, {0,0}};
  static tokenDesc commands[] = {
  	{kTokenWorldSector, "SECTOR"},
  	{kTokenWorldRoom, "ROOM"},
  	{kTokenWorldPlane, "PLANE"},
  	{kTokenWorldCol, "COLLECTION"},
  	{kTokenWorldScript, "SCRIPT"},
  	{kTokenWorldTexs, "TEXTURES"},
  	{kTokenWorldLightX, "LIGHTX"},
  	{kTokenWorldThing, "THING"},
  	{kTokenWorldSixface, "SIXFACE"},
  	{kTokenWorldSpriteTmp, "SPRITE"},
  	{kTokenWorldLibrary, "LIBRARY"},
  	{kTokenWorldStart, "START"},
  	{kTokenWorldWaves, "WAVES"},
	{0,0}};
  char* name, * data;

  if (csGetObject (&buf, tokens, &name, &data))
  {
    long cmd;
    char* params;

    while ((cmd = csGetObject (&data, commands, &name, &params)) > 0)
    {
      switch (cmd)
      {
  	case kTokenWorldSpriteTmp:
	  {
	    csSpriteTemplate* t = get_sprite_template (name);
	    if (!t)
	    {
	      CHK (t = new csSpriteTemplate (name));
	      sprite_templates.Push (t);
	    }
	    t->load (params, textures);
	  }
	  break;
        case kTokenWorldThing:
          if (!get_thing_template(name))
            thing_templates.Push ( CSLoader::load_thingtpl(name, params, textures) );
	  break;
        case kTokenWorldSixface:
          if (!get_thing_template(name))
            thing_templates.Push ( CSLoader::load_sixtpl(name, params, textures) );
	  break;
        case kTokenWorldSector:
          if (!sectors.FindName (name))
            sectors.Push( CSLoader::load_sector(name, this, params, textures) );
	  break;
        case kTokenWorldPlane:
          planes.Push ( CSLoader::load_polyplane (params, name) );
	  break;
        case kTokenWorldCol:
          collections.Push ( CSLoader::load_collection(name,this,params) );
	  break;
        case kTokenWorldScript:
          new_script (layer, name, params);
	  break;
        case kTokenWorldTexs:
	  {
            if (!load_textures (params))
              return false;
	    Library* lib = (Library*)libraries.FindName ("standard");
	    if (!lib)
	    {
	      CHK (lib = new Library ("standard"));
	      libraries.Push (lib);
	      CSLoader::load_library_def (lib, this, "standard.zip", textures);
	    }
	  }
	  break;
	case kTokenWorldWaves:
	  if (!load_waves (params)) return false;
	  break;
	case kTokenWorldRoom:
          // Not an object but it is translated to a special sector.
          if (!sectors.FindName (name))
            sectors.Push ( CSLoader::load_room(name, this, params, textures) );
	  break;
        case kTokenWorldLightX:
          CLights::LoadFromBuffer (name, params);
	  break;
	case kTokenWorldLibrary:
          {
  	    char str[255];
	    ScanStr (params, "%s", str);
	    Library* lib = (Library*)libraries.FindName (name);
	    if (!lib)
	    {
	      // Library was not already loaded.
	      CHK (lib = new Library (name));
	      libraries.Push (lib);
	      CSLoader::load_library_def (lib, this, str, textures);
	    }
	  }
	  break;
	case kTokenWorldStart:
	  {
	    char str[255];
	    ScanStr (params, "%s,%f,%f,%f", str, &start_vec.x, &start_vec.y, &start_vec.z);
	    CHK (delete start_sector);
	    CHK (start_sector = new char [strlen (str)+1]);
	    strcpy (start_sector, str);
	  }
	  break;
      }
    }
  }

  int sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    for (PolygonSet *ps = s;  ps;  
         (ps==s) ? (ps = s->get_first_thing()) : (ps = ps->get_next()) )
      for (int i=0;  i < ps->get_num_polygon();  i++)
      {
        Polygon3D* p = (Polygon3D*)( ps->get_polygon(i) );
        if (p && p->get_portal())
        {
          PortalCS* portal = (PortalCS*)( p->get_portal() );
          csSector *stmp = portal->get_sector();
          csSector *snew = (csSector*)sectors.FindName (stmp->get_name());
          portal->set_sector(snew);
          CHK( delete stmp );
        }
      }
  }
  
  return true;
}

bool csWorld::initialize ()
{
  Gfx3D->ClearCache ();
  shine_lights ();
  mipmap_settings (LightMap::setting);

  if (!SUCCEEDED (Gfx3D->QueryInterface(IID_IHaloRasterizer, (void**)&piHR))) piHR = NULL;

  return true;
}

Archive* csWorld::get_world_file ()
{
  if (!world_file) CHKB (world_file = new Archive ("precalc.zip"));
  return world_file;
}

bool csWorld::load_textures (char* buf, Archive* ar)
{
  if (textures) CHKB (delete textures);
  CHK (textures = new Textures (50));
  if (!textures->load (buf, this, ar)) return false;

  Gfx3D->ClearCache ();
  Gfx3D->SetWorld( GetIWorldFromWorld(this) );
  return true;
}

bool csWorld::append_textures (char* buf, Archive* ar)
{
  if (!textures->load (buf, this, ar)) return false;

  Gfx3D->ClearCache ();
  return true;
}

bool csWorld::load_waves (char* buf, Archive* ar)
{
  enum { kTokenWavesWav = 1 };

  static tokenDesc commands[] = {
        {kTokenWavesWav, "WAVE"},
        {0,0}};
  char* name;
  long cmd;
  char* params;

  while ((cmd = csGetObject (&buf, commands, &name, &params)) > 0)
  {
    switch (cmd)
    {
      case kTokenWavesWav:
        {
          Wave *wav = get_wave (name);
          if (!wav)
	  {
	    wav = CSLoader::load_wave (name, this, ar);
	    if (wav) waves.Push (wav);
	  }
        }
        break;
    }
  }

  return true;
}

Wave *csWorld::get_wave (char *name)
{
  return (Wave *)waves.FindName (name);
}

void csWorld::shine_lights ()
{
  if (Polygon3D::do_force_recalc)
  {
    CsPrintf (MSG_INITIALIZATION, "Recalculation of lightmaps forced.\n");
    CsPrintf (MSG_INITIALIZATION, "  Pseudo-radiosity system %s.\n", csSector::do_radiosity ? "enabled" : "disabled");
    CsPrintf (MSG_INITIALIZATION, "  Maximum number of visits per sector = %d.\n", csSector::cfg_reflections);
  }
  else
  {
    // If no recalculation is forced we set these variables to default to
    // make sure that we don't do too many unneeded calculations.
    csSector::do_radiosity = false;
    csSector::cfg_reflections = 1;
  }

  int sn;
  sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    s->init_lightmaps ();
  }
  sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    s->shine_lights ();
  }
  sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    s->cache_lightmaps ();
  }

  Archive* ar = get_world_file ();
  if (ar)
    if (!ar->write_archive ())
      CsPrintf (MSG_WARNING, "WARNING: error updating lighttable cache in archive %s!\n", ar->GetFilename ());
}

void csWorld::mipmap_settings (int setting)
{
  int sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    s->mipmap_settings (setting);
  }
}

void csWorld::start_world ()
{
  clear ();
  CSResetID ();
  Stats::polygons_loaded = 0;
  Stats::portals_loaded = 0;
  Stats::sectors_loaded = 0;
  Stats::things_loaded = 0;
  Stats::lights_loaded = 0;
}

bool csWorld::load_file (LanguageLayer* layer, char* file)
{
  start_world ();

  FILE* fp;
  fp = fopen (file, "rb");
  if (!fp)
  {
    CsPrintf (MSG_FATAL_ERROR, "Could not open world file '%s'!\n\
The world file should either be a .ZIP level file or else an ASCII\n\
file describing how the world looks.\n", file);
    return false;
  }

  char* buf;

  // First check if the world file is a ZIP archive.
  char hdr[5];
  fread (hdr, 4, 1, fp);
  hdr[4] = 0;
  if (!strcmp (hdr, "PK\003\004"))
  {
    fclose (fp);
    fp = NULL;
    CHK (world_file = new Archive (file));
    buf = world_file->read ("world");
    if (!buf)
    {
      CsPrintf (MSG_FATAL_ERROR, "Could not find 'world' file inside ZIP archive '%s'!\n\
Are you sure '%s' is a valid Crystal Space level archive?\n", file, file);
      CHK (delete world_file);
      return false;
    }
  }
  else
  {
    CHK (world_file = new Archive ("precalc.zip"));
    fseek (fp, 0, SEEK_END);
    long off = ftell (fp);
    fseek (fp, 0, SEEK_SET);
    CHK (buf = new char [off+1]);
    fread (buf, 1, off, fp);
    buf[off] = 0;
  }

  csIniFile* cfg = NULL;
  if (world_file) CHKB (cfg = new csIniFile (world_file, "cryst.cfg"));
  Polygon3D::def_mipmap_size = config->GetInt ("TextureMapper", "LIGHTMAP_SIZE", 16);
  if (cfg) Polygon3D::def_mipmap_size = cfg->GetInt ("TextureMapper", "LIGHTMAP_SIZE", Polygon3D::def_mipmap_size);
  if (cfg) CHKB (delete cfg);
  CsPrintf (MSG_INITIALIZATION, "Lightmap grid size = %dx%d.\n", Polygon3D::def_mipmap_size,
      Polygon3D::def_mipmap_size);

  if (!load (layer, buf)) return false;

  if (Stats::polygons_loaded)
  {
    CsPrintf (MSG_INITIALIZATION, "Loaded world file:\n");
    CsPrintf (MSG_INITIALIZATION, "  %d polygons (%d portals),\n", Stats::polygons_loaded,
      Stats::portals_loaded);
    CsPrintf (MSG_INITIALIZATION, "  %d sectors and %d things,\n", Stats::sectors_loaded,
      Stats::things_loaded);
    CsPrintf (MSG_INITIALIZATION, "  %d lights.\n", Stats::lights_loaded);
  } /* endif */

  CHK (delete [] buf);

  if (fp) fclose (fp);
  if (get_world_file ()) get_world_file ()->write_archive ();

  return true;
}

void csWorld::draw (csCamera* c, csClipper* view)
{
  Stats::polygons_considered = 0;
  Stats::polygons_drawn = 0;
  Stats::portals_drawn = 0;

  RenderView rview(*c,view,NULL);

  if (map_mode != MAP_OFF) wf->get_wireframe ()->clear ();
  csSector* s = c->get_sector ();
  if (s->has_fog ()) Gfx3D->OpenFogObject (s->get_id (), &s->get_fog ());
  s->draw (rview);
  if (s->has_fog ())
  {
    // Since there is fog in the current sector we simulate this by adding
    // the view plane polygon.
    G3DPolygon g3dpoly;
    g3dpoly.num = view->GetNumVertices ();
    int i;
    for (i = 0 ; i < g3dpoly.num ; i++)
    {
      g3dpoly.vertices[g3dpoly.num-i-1].sx = view->GetVertex (i).x;
      g3dpoly.vertices[g3dpoly.num-i-1].sy = view->GetVertex (i).y;
    }
    Gfx3D->AddFogPolygon (s->get_id (), g3dpoly, CS_FOG_VIEW);
    Gfx3D->CloseFogObject (s->get_id ());
  }

  // draw all halos on the screen
  IHaloRasterizer* piHR = NULL;
  bool supports_halos;

  supports_halos = Gfx3D->QueryInterface (IID_IHaloRasterizer, (void**)&piHR) == S_OK ? true : false;
  
  csHaloInformation* pinfo;
  
  for (int cntHalos = 0; cntHalos <halos.Length(); cntHalos++)
  {
    HRESULT hres = S_FALSE;

    pinfo = (csHaloInformation*)halos.Get(cntHalos);

    float hintensity = pinfo->pLight->get_halo_intensity();

    if (pinfo->pLight->get_reference_count() == 0)
    {      
      hintensity -= .15f;

      // this halo is completely invisible. kill it.
      if (hintensity <= 0)
      {
	halos.Delete(cntHalos);
	pinfo->pLight->set_halo_in_queue(false);
	
	piHR->DestroyHalo(pinfo->haloinfo);
	delete pinfo;

	cntHalos--;
	continue;
      }

      pinfo->pLight->set_halo_intensity(hintensity);
    }
    else
    {

      if (hintensity < pinfo->pLight->get_halo_max_intensity())
	hintensity += .15f;
      
      if (hintensity > pinfo->pLight->get_halo_max_intensity())
	hintensity = pinfo->pLight->get_halo_max_intensity();

      pinfo->pLight->set_halo_intensity(hintensity);
    }
  
    if (supports_halos)
    {
      // project the halo.
      pinfo->v = rview.world2cam(pinfo->pLight->get_center());
      
      if (pinfo->v.z > SMALL_Z)
      {
	float iz = csCamera::aspect/pinfo->v.z;
	pinfo->v.x = pinfo->v.x * iz + (FRAME_WIDTH/2);
	pinfo->v.y = FRAME_HEIGHT - 1 - (pinfo->v.y * iz + (FRAME_HEIGHT/2));
	
	pinfo->intensity = pinfo->pLight->get_halo_intensity();

	hres = piHR->DrawHalo(&pinfo->v, pinfo->intensity, pinfo->haloinfo);
      }
    }

    // was this halo actually drawn?
    if (hres == S_FALSE)
      pinfo->pLight->remove_reference();
  }   

  // DAN: this is commneted out until we figure out where light selection logic should go.
/*  if (Gfx3D->do_light_frust && Gfx3D->selected_light)
    Gfx3D->selected_light->dump_frustrum (*c); */
}

void csWorld::add_halo(csHaloInformation* pinfo)
{
  pinfo->pLight->add_reference();
  pinfo->pLight->set_halo_in_queue(true);
  halos.Push(pinfo);
}

bool csWorld::has_halo(csLight* pLight)
{
  for (int i=0; i < halos.Length(); i++)
  {
    if (((csHaloInformation*)halos[i])->pLight == pLight)
      return true;
  }

  return false;
}

csStatLight* csWorld::find_light (float x, float y, float z, float dist)
{
  csStatLight* l;
  int sn = sectors.Length ();
  while (sn > 0)
  {
    sn--;
    csSector* s = (csSector*)sectors[sn];
    l = s->find_light (x, y, z, dist);
    if (l) return l;
  }
  return NULL;
}

void csWorld::add_dyn_light (csDynLight* dyn)
{
  dyn->set_next (first_dyn_lights);
  dyn->set_prev (NULL);
  if (first_dyn_lights) first_dyn_lights->set_prev (dyn);
  first_dyn_lights = dyn;
}

void csWorld::remove_dyn_light (csDynLight* dyn)
{
  if (dyn->get_next ()) dyn->get_next ()->set_prev (dyn->get_prev ());
  if (dyn->get_prev ()) dyn->get_prev ()->set_next (dyn->get_next ());
  else if (dyn == first_dyn_lights) first_dyn_lights = dyn->get_next ();
  dyn->set_next (NULL);
  dyn->set_prev (NULL);
}

void csWorld::advance_sprite_frames (long current_time)
{
  MultiSectorObject* mso = first_mso;
  while (mso)
  {
    csObject* o = mso->get_object ();
    if (o->get_type () == CS_SPRITE3D)
      ((csSprite3D*)o)->next_frame (current_time);
    mso = mso->get_next ();
  }
}

//---------------------------------------------------------------------------

