/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef WORLD_H
#define WORLD_H

#include "geometry/math3d.h"
#include "engine/iworld.h"
#include "engine/basic/csobjvec.h"
#include "render/ihalo.h"

class csSector;
class Textures;
class PolyPlane;
class csScript;
class Polygon3D;
class Polygon2D;
class csCamera;
class PolygonSet;
class Thing;
class ThingTemplate;
class Collection;
class csObject;
class LanguageLayer;
class Portal;
class csStatLight;
class csDynLight;
class Archive;
class csSpriteTemplate;
class WireFrameCam;
class MultiSectorObject;
class Library;
class csClipper;
class Dumper;
class Wave;
class csLight;

/// Several map modes.
#define MAP_OFF 0
#define MAP_OVERLAY 1
#define MAP_ON 2

/// This is used to keep track of halo information.
class csHaloInformation
{
public:
  ///
  Vector3 v;
  ///
  csLight* pLight;

  ///
  float r, g, b;
  ///
  float intensity;
  ///
  HALOINFO haloinfo;

  ///
  csHaloInformation()
  {
    pLight = NULL;
    r = g = b = intensity = 0.0f;
    haloinfo = NULL;
  }

};

/**
 * The world!
 */
class csWorld
{
  friend class Dumper;

public:
  /// List of sectors in the world.
  csObjVector sectors;
  /// List of libraries.
  csObjVector libraries;
  /// List of planes.
  csObjVector planes;
  /// List of scripts.
  csObjVector scripts;
  /// List of collections.
  csObjVector collections;
  /// List of sprite templates.
  csObjVector sprite_templates;
  /// List of thing templates.
  csObjVector thing_templates;
  /// List of waves.
  csObjVector waves;

private:
  /// Texture and color information object.
  Textures* textures;
  /// Linked list of all MultiSectorObjects (Things and Sprites).
  MultiSectorObject* first_mso;
  /// Linked list of dynamic lights.
  csDynLight* first_dyn_lights;
  /**
   * If world file was loaded from a ZIP archive.
   * If not this points to the 'precalc.zip' archive.
   */
  Archive* world_file;
  /// List of halos (csHaloInformation).
  csVector halos;  
  /// The Halo rasterizer. If NULL halo's are not supported by the rasterizer.
  IHaloRasterizer* piHR;

public:
  /// A WireFrame object for the map.
  WireFrameCam* wf;
  /// Map mode.
  int map_mode;

  /// Starting sector.
  char* start_sector;
  /// Starting position.
  Vector3 start_vec;

public:
  ///
  csWorld ();
  ///
  virtual ~csWorld ();

  /**
   * Initialize the world. This function needs to be called after the
   * texture system has been correctly initilialized.
   */
  bool initialize ();

  /**
   * Prepare for creation of a world. This function should be called
   * before you start to add sectors. load_world automatically calls
   * this function.
   */
  void start_world ();

  /// Get the Halo Rasterizer COM interface if supported (NULL if not).
  IHaloRasterizer* getHaloRastizer () { return piHR; }

  /**
   * Get the default world file. If there is none this function
   * will open precalc.zip and use that.
   */
  Archive* get_world_file ();

  /// Clear everything in the world.
  void clear ();

  ///
  csSector* new_sector (char* name);

  ///
  csSpriteTemplate* get_sprite_template (char* name, bool use_libs = false);

  ///
  ThingTemplate* get_thing_template (char* name, bool use_libs = false);
  ///
  Thing* get_thing (char* name);

  ///
  Textures* get_textures () { return textures; }
  ///
  void set_textures (Textures* txt) { textures = txt; }

  /**
   * Load waves from a WAVES(...) argument.
   * If 'ar' is given optionally load from that archive as well.
   */
  bool load_waves (char* buf, Archive* ar = NULL);
  ///
  Wave* get_wave (char* name);

  ///
  MultiSectorObject* get_mso (char* name);
  ///
  MultiSectorObject* get_first_mso () { return first_mso; }
  ///
  void add (MultiSectorObject* mso);
  ///
  void remove (MultiSectorObject* mso);

  ///
  void add_dyn_light (csDynLight* dyn);
  ///
  void remove_dyn_light (csDynLight* dyn);
  ///
  csDynLight* get_first_dyn_light () { return first_dyn_lights; }

  ///
  void add_halo(csHaloInformation* pinfo);
  ///
  bool has_halo(csLight* pLight);

  /**
   * Create a new script. 'params' should start with a
   * qualifier indicating the language of the script.
   * The syntax is as follows:<p>
   *     SCRIPT 'name' (prim: dim 1,0,0)<p>
   *	 SCRIPT 'name' (C:script1.c)
   */
  csScript* new_script (LanguageLayer* layer, char* name, char* params);

  /// Activate all triggers valid at this camera position.
  void trigger_activate (csCamera& c);

  /// Draw the world given a view and a camera.
  void draw (csCamera* c, csClipper* view);

  /// Load the world from the world file.
  bool load (LanguageLayer* layer, char* buf);
  /// Load the world from the world file.
  bool load_file (LanguageLayer* layer, char* name);

  /**
   * Load textures from a TEXTURES(...) argument.
   * If 'ar' is given optionally load from that archive as well.
   */
  bool load_textures (char* buf, Archive* ar = NULL);

  /**
   * Append textures from a TEXTURES(...) argument (for libraries).
   * If 'ar' is given optionally load from that archive as well.
   */
  bool append_textures (char* buf, Archive* ar = NULL);

  ///
  void shine_lights ();
  ///
  void mipmap_settings (int setting);
  ///
  csStatLight* find_light (float x, float y, float z, float dist);

  /**
   * Advance the frames of all sprites.
   */
  void advance_sprite_frames (long current_time);

  DECLARE_INTERFACE_TABLE( csWorld )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( World )
};

#define GetIWorldFromWorld(a)  &a->m_xWorld
#define GetWorldFromIWorld(a)  ((csWorld*)((size_t)a - offsetof(csWorld, m_xWorld)))

#endif /*WORLD_H*/
