/*
  Maze editor: 3D model class
  Copyright (C) 1998 by Jorrit Tyberghein
  Written by Andrew Zabolotny <bit@eltech.ru>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <limits.h>
#include "me_draft.h"
#include "me_model.h"
#include "me_text.h"
#include "me_util.h"
#include "system/system.h"
#include "engine/objects/thingtpl.h"
#include "engine/objects/cssprite.h"

//--------------------------------------------------// 3D polygon class //----//

mz3DPolygon::mz3DPolygon (mz3DModel *iParent) : V (8, 8)
{
  parent = iParent;
  Color = -1;
  selected = false;
  Name = NULL;

  // Generate a name for current polygon
  char tmp [16];
  sprintf (tmp, "p%lX", csSystemDriver::Time ());
  SetName (tmp);
}

mz3DPolygon::mz3DPolygon (mz3DPolygon &iCopy) : V (8, 8)
{
  parent = iCopy.parent;
  Color = iCopy.Color;
  selected = iCopy.selected;
  Name = NULL;

  // Generate a name for current polygon
  char tmp [16];
  sprintf (tmp, "p%lX", csSystemDriver::Time ());
  SetName (tmp);

  for (int i = 0; i < iCopy.Vertices (); i++)
    V [i] = iCopy.V [i];
}

mz3DPolygon::~mz3DPolygon ()
{
  if (Name)
    delete [] Name;
}

void mz3DPolygon::SetName (char *iName)
{
  if (Name)
    delete [] Name;
  Name = strnew (iName);
}

bool mz3DPolygon::NotifyVertexDeleted (int iVertex)
{
  for (int i = 0; i < Vertices (); i++)
    if ((*this) [i] > iVertex)
      (*this) [i]--;
    else if ((*this) [i] == iVertex)
      return false;
  return true;
}

bool mz3DPolygon::operator == (const mz3DPolygon &p) const
{
  if (p.Vertices () != Vertices ())
    return false;
  int i, v1 = 0, v2 = 0, m1 = INT_MAX, m2 = INT_MAX;
  int s = Vertices ();
  mz3DPolygon &thisv = (mz3DPolygon &)(*this);
  mz3DPolygon &pv = (mz3DPolygon &)p;
  for (i = 0; i < s; i++)
  {
    if (thisv [i] < m1)
    {
      m1 = thisv [i];
      v1 = i;
    } /* endif */
    if (pv [i] < m2)
    {
      m2 = pv [i];
      v2 = i;
    } /* endif */
  } /* endfor */

  for (i = 0; i < s; i++)
    if (thisv [(v1 + i) % s] != pv [(v2 + i) % s])
      return false;
  return true;
}

csPolyError mz3DPolygon::Check ()
{
  if (Vertices () < 3)
    return cspeUnderflow;
  if (Vertices () > MAX_POLYGON_VERTICES)
    return cspeOverflow;

  // Check if some vertex is mentioned twice
  int i, j;
  for (i = 0; i < Vertices (); i++)
    for (j = i + 1; j < Vertices (); j++)
    {
      if ((*this) [i] == (*this) [j])
        return cspeDuplicateVertex;
    } /* endfor */

  // Check if we have a valid parent model
  if (!parent)
    return cspeNoParent;

  // Check if all vertices lies on the same plane
  mz3DVertex v1 = Vertex (1) - Vertex (0);
  mz3DVertex v2 = Vertex (2) - Vertex (0);
  float epsilon1 = EPSILON, epsilon2 = -EPSILON;
  for (i = 3; i < Vertices (); i++)
  {
    mz3DVertex v3 = Vertex (i) - Vertex (0);
    float val = v1.x * v2.y * v3.z
              + v1.y * v2.z * v3.x
              + v1.z * v2.x * v3.y
              - v1.x * v2.z * v3.y
              - v1.z * v2.y * v3.x
              - v1.y * v2.x * v3.z;
    if ((val > epsilon1)
     || (val < epsilon2))
      return cspeNonCoplanar;
    v1 = v2;
    v2 = v3;
  } /* endfor */

  // And finally, check if polygon is convex
  // For this, check for all polygon edges that all vertices lies
  // on one side of the plane that is perpendicular to polygon's plane
  // and crosses it through the line that contains that edge
  for (i = 0; i < Vertices (); i++)
  {
    // First compute the plane that is perpendicular to polygon plane
    // and crosses it through the line that contains current edge
    int v = i, vn = (i + 1) % Vertices (), vp = (i + Vertices () - 1) % Vertices ();
    Vector3 pp = Vertex (v) + ((Vertex (vp) - Vertex (v))
                             % (Vertex (vn) - Vertex (v)));
    Vector3 normal;
    float D;

    // Normalize pp so that we will avoid precision lose
    if (pp.x > 0) pp.x = sqrt (pp.x); else pp.x = -sqrt (-pp.x);
    if (pp.y > 0) pp.y = sqrt (pp.y); else pp.y = -sqrt (-pp.y);
    if (pp.z > 0) pp.z = sqrt (pp.z); else pp.z = -sqrt (-pp.z);

    Plane::calc_plane (Vertex (v), Vertex (vn), pp, normal, D);
    Plane plane (normal, D);

    // Compute maximum allowable bias
    epsilon1 = fabs (plane.classify (Vertex (v )));
    epsilon2 = fabs (plane.classify (Vertex (vn)));
    if (epsilon1 > epsilon2)
      epsilon2 = -epsilon1;
    else
      epsilon1 = -epsilon2;

    // Now look if all polygon vertices are on one side of plane
    // defined by three vectors e1, e2 and e3 (or normal n)
    bool oneside = false, otherside = false;
    for (j = 0; j < Vertices (); j++)
    {
      if ((j == v) || (j == vn))
        continue;
      float c = plane.classify (Vertex (j));
      if (c > epsilon1)
        oneside = true;
      if (c < epsilon2)
        otherside = true;
    } /* endfor */
    if (oneside && otherside)
      return cspeNonConvex;
  } /* endfor */

  // Everything looks okay
  return cspeOK;
}

mz3DVertex& mz3DPolygon::Vertex (int iVertex)
{
  return parent->Vertex ((*this) [iVertex]);
}

char *mz3DPolygon::ErrorText (csPolyError iCode)
{
  switch (iCode)
  {
    case cspeUnderflow:
      return TEXT_POLYERRUNDERFLOW;
    case cspeOverflow:
      return TEXT_POLYERROVERFLOW;
    case cspeDuplicateVertex:
      return TEXT_POLYERRDUPVERTEX;
    case cspeNoParent:
      return TEXT_POLYERRNOPARENT;
    case cspeNonCoplanar:
      return TEXT_POLYERRNONCOPLANAR;
    case cspeNonConvex:
      return TEXT_POLYERRNONCONVEX;
    default:
      return NULL;
  } /* endswitch */
}

void mz3DPolygon::Reverse ()
{
  for (int i = Vertices () / 2; i > 0; i--)
  {
    int idx = (*this) [i - 1];
    (*this) [i - 1] = (*this) [Vertices () - i];
    (*this) [Vertices () - i] = idx;
  } /* endif */
}

//-----------------------------------------------// Sprite Action class //----//

mzSpriteAction::mzSpriteAction (csSpriteAction *iAction) : Names (8, 8),
  Delays (8, 8)
{
  Name = strnew (iAction->get_name ());
  int Frames = iAction->get_num_frameset ();
  for (int i = 0; i < Frames; i++)
  {
    csFrame *f = iAction->get_frame (i);
    int d = iAction->get_frame_delay (i);
    int frames = Names.Length ();
    Names.Push (strnew (f->get_name ()));
    if (frames != Names.Length ())
      Delays.Push ((void *)d);
  } /* endfor */
}

mzSpriteAction::~mzSpriteAction ()
{
  if (Name)
    delete [] Name;
}

void mzSpriteAction::NotifyFrameDeleted (char *iFrameName)
{
  for (int i = Names.Length () - 1; i >= 0; i--)
    if (strcmp (iFrameName, (char *)Names [i]) == 0)
    {
      Names.Delete (i);
      Delays.Delete (i);
    } /* endif */
}

bool mzSpriteAction::operator == (const mzSpriteAction &a)
{
  if (Frames () != a.Frames ())
    return false;
  if (strcmp (Name, a.Name))
    return false;
  int i;
  for (i = Frames () - 1; i >= 0; i--)
  {
    if (Delay (i) != a.Delay (i))
      return false;
    if (FrameName (i) != a.FrameName (i))
      return false;
  } /* endfor */
  return true;
}

bool mzSpriteAction::SetFrameName (int iFrame, char *iName)
{
  if ((iFrame < 0) || (iFrame > Frames ()))
    return false;
  if (Names [iFrame])
    delete [] FrameName (iFrame);
  Names [iFrame] = strnew (iName);
  return true;
}

//----------------------------------------------------// 3D model utils //----//

mz3DModel::csVertexVector::~csVertexVector ()
{
  DeleteAll ();
}

void mz3DModel::csVertexVector::FreeItem (csSome Item)
{
  CHK (delete (mz3DVertex *)Item);
}

mz3DModel::csPolygonVector::~csPolygonVector ()
{
  DeleteAll ();
}

void mz3DModel::csPolygonVector::FreeItem (csSome Item)
{
  CHK (delete (mz3DPolygon *)Item);
}

mz3DModel::csLightVector::~csLightVector ()
{
  DeleteAll ();
}

void mz3DModel::csLightVector::FreeItem (csSome Item)
{
  (void)Item;
}

//----------------------------------------------------// 3D model class //----//

mz3DModel::mz3DModel (csModelType iType, char *iFileName)
  : V (8, 8), A (8, 8), FrameName (8, 8)
{
  Type = iType;
  Color = -1;
  Name = FileName = NULL;
  Texture = NULL;
  Frame = -1;
  SetFrame (0);

  // Generate a name for model
  char tmp [16];
  char *format = iType == csmtRoom ? "r%lX" :
                 iType == csmtThing ? "t%lX" :
                 iType == csmtSprite ? "s%lX" :
                 "m%lX";
  sprintf (tmp, format, csSystemDriver::Time ());
  SetName (tmp);

  if (iFileName)
    SetFileName (iFileName);
}

mz3DModel::~mz3DModel ()
{
  SetName (NULL);
  SetFileName (NULL);
}

void mz3DModel::SetName (char *iName)
{
  if (Name)
    delete [] Name;
  Name = strnew (iName);
}

void mz3DModel::SetFileName (char *iFileName)
{
  if (FileName)
    delete [] FileName;
  FileName = strnew (iFileName);
}

void mz3DModel::SetFrameName (int iFrame, char *iName)
{
  if ((iFrame < 0)
   || (iFrame > Frames ()))
    return;

  if (iFrame < FrameName.Length ())
    FrameName.FreeItem (FrameName [iFrame]);
  FrameName [iFrame] = strnew (iName);
}

bool mz3DModel::SetFrame (int iFrame, char *iFrameName)
{
  if ((Type != csmtSprite)
   && (iFrame != 0))
    return false;
  if ((iFrame < 0)
   || (iFrame > V.Length ()))
    return false;
  int oldFrame = Frame;
  Frame = iFrame;
  if (iFrameName)
    SetFrameName (Frame, iFrameName);
  if (Frame == V.Length ())
  {
    V.Push (new csVertexVector ());
    if (!iFrameName)
    {
      char tmp [16];
      sprintf (tmp, "f%d", Frame);
      SetFrameName (Frame, tmp);
    } /* endif */
    // duplicate previous frame, if any
    if (oldFrame >= 0)
    {
      int vc = Vertices (oldFrame);
      for (int vertex = 0; vertex < vc; vertex++)
        ((csVertexVector *)V [Frame])->Push (new mz3DVertex (Vertex (vertex, oldFrame)));
    } /* endif */
  }

  // duplicate selection
  if (oldFrame >= 0)
  {
    int vc = Vertices (oldFrame);
    for (int vertex = 0; vertex < vc; vertex++)
      Vertex (vertex).selected = Vertex (vertex, oldFrame).selected;
  } /* endif */
  return true;
}

void mz3DModel::DeleteFrame (int iFrame)
{
  if ((iFrame < 0) || (iFrame >= Frames ()))
    return;
  for (int i = A.Length () - 1; i >= 0; i--)
  {
    mzSpriteAction *sa = (mzSpriteAction *)A [i];
    sa->NotifyFrameDeleted ((char *)FrameName [iFrame]);
  } /* endfor */
  V.Delete (iFrame);
  FrameName.Delete (iFrame);
  if (Frame >= V.Length ())
    Frame = V.Length () - 1;
  if (Frame < 0)
    SetFrame (0);
}

bool mz3DModel::InsertVertex (mz3DVertex *iVertex, int *iNum)
{
  int i;
  for (i = Vertices () - 1; i >= 0; i--)
    if (Vertex (i) == *iVertex)
    {
      if (iNum)
        *iNum = i;
      CHK (delete iVertex);
      return false;
    } /* endif */
  if (iNum)
    *iNum = Vertices ();
  for (i = Frames () - 1; i >= 0; i--)
    ((csVertexVector *)V [i])->Push (iVertex);
  return true;
}

bool mz3DModel::DeleteVertex (int iVertex)
{
  if ((iVertex < 0)
   || (iVertex >= Vertices ()))
    return false;

  int i;
  for (i = Frames () - 1; i >= 0; i--)
    ((csVertexVector *)V [i])->Delete (iVertex);
  for (i = Polygons () - 1; i >= 0; i--)
    if (!Polygon (i).NotifyVertexDeleted (iVertex))
      DeletePolygon (i);
  return true;
}

bool mz3DModel::InsertPolygon (mz3DPolygon *iPolygon, int *iNum)
{
  for (int i = P.Length () - 1; i >= 0; i--)
    if (Polygon (i) == *iPolygon)
    {
      if (iNum)
        *iNum = i;
      delete iPolygon;
      return false;
    } /* endif */
  if (iNum)
    *iNum = P.Length ();
  iPolygon->parent = this;
  P.Push (iPolygon);
  return true;
}

bool mz3DModel::DeletePolygon (int iPolygon)
{
  if ((iPolygon < 0)
   || (iPolygon >= Polygons ()))
    return false;

  P.Delete (iPolygon);
  return true;
}

bool mz3DModel::InsertAction (mzSpriteAction *iAction, int *iNum)
{
  for (int i = Actions () - 1; i >= 0; i--)
    if (Action (i) == *iAction)
    {
      if (iNum)
        *iNum = i;
      delete iAction;
      return false;
    } /* endif */
  if (iNum)
    *iNum = A.Length ();
  A.Push (iAction);
  return true;
}

bool mz3DModel::DeleteAction (int iAction)
{
  if ((iAction < 0)
   || (iAction >= Actions ()))
    return false;

  A.Delete (iAction);
  return true;
}

void mz3DModel::Clear (bool iVertices, bool iFaces, bool iLights, bool iFrames,
  bool iActions)
{
  if (iVertices)
  {
    P.DeleteAll ();
    for (int i = Frames () - 1; i >= 0; i--)
      ((csVertexVector *)V [i])->DeleteAll ();
  }
  if (iFaces)
    P.DeleteAll ();
  if (iLights)
    L.DeleteAll ();
  if (iFrames)
  {
    V.DeleteAll ();
    FrameName.DeleteAll ();
    Frame = -1;
    SetFrame (0);
  }
  if (iActions)
    A.DeleteAll ();
}

bool mz3DModel::Save (int (*Put) (void *iStream, void *iData, int iSize), void *iStream)
{
#define PUTS(s)					\
  {						\
    int _len = strlen (s);			\
    if (Put (iStream, s, _len) != _len)		\
      return false;				\
  }

  char *header;
  switch (Type)
  {
    case csmtRoom:
      return false; // not supported for now
    case csmtThing:
      header = "  THING '%s'\n  (\n";
      break;
    case csmtSprite:
      header = "  SPRITE '%s'\n  (\n";
      break;
    default:
      return false;
  } /* endswitch */

  char buff [256];
  sprintf (buff, header, Name);
  PUTS (buff);

  int i, j;
  switch (Type)
  {
    case csmtRoom:
      break;
    case csmtThing:
      for (i = 0; i < Vertices (); i++)
      {
        strcpy (buff, "    VERTEX (");
        ftoa (&buff [strlen (buff)], Vertex (i).x);
        strcat (buff, ",");
        ftoa (&buff [strlen (buff)], Vertex (i).y);
        strcat (buff, ",");
        ftoa (&buff [strlen (buff)], Vertex (i).z);
        strcat (buff, ")\n");
        PUTS (buff);
      } /* endfor */
      for (i = 0; i < Polygons (); i++)
      {
        mz3DPolygon &p = Polygon (i);
        sprintf (buff, "    POLYGON '%s' (VERTICES (", p.Name);
        for (int j = 0; j < p.Vertices (); j++)
        {
          char tmp [10];
          char *format;
          if (j == 0)
            format = "%d";
          else
            format = ",%d";
          sprintf (tmp, format, p [j]);
          strcat (buff, tmp);
        } /* endfor */
        strcat (buff, ")");

        strcat (buff, ")\n");
        PUTS (buff);
      } /* endfor */
      break;
    case csmtSprite:
      if (Texture)
        sprintf (buff, "    TEXNR ('%s')\n", Texture->get_name ());
      PUTS (buff);
      for (i = 0; i < Frames (); i++)
      {
        sprintf (buff, "    FRAME '%s' (", GetFrameName (i));
        PUTS (buff);
        for (j = 0; j < Vertices (); j++)
        {
          if (j)
            strcpy (buff, " V(");
          else
            strcpy (buff, "V(");
          ftoa (&buff [strlen (buff)], Vertex (j, i).x);
          strcat (buff, ",");
          ftoa (&buff [strlen (buff)], Vertex (j, i).y);
          strcat (buff, ",");
          ftoa (&buff [strlen (buff)], Vertex (j, i).z);
          strcat (buff, ":");
          ftoa (&buff [strlen (buff)], Vertex (j, i).y);
          strcat (buff, ",");
          ftoa (&buff [strlen (buff)], Vertex (j, i).y);
          strcat (buff, ")");
          PUTS (buff);
        } /* endfor */
        PUTS (")\n");
      } /* endfor */
      for (i = 0; i < Actions (); i++)
      {
        mzSpriteAction &a = Action (i);
        sprintf (buff, "    ACTION '%s' (", a.Name);
        PUTS (buff);
        for (j = 0; j < a.Frames (); j++)
        {
          sprintf (buff, "F(%s,%d) ", a.FrameName (j), a.Delay (j));
          if (j == a.Frames () - 1)
            buff [strlen (buff) - 1] = 0;
          PUTS (buff);
        } /* endfor */
        PUTS (")\n");
      } /* endfor */
      for (i = 0; i < Polygons (); i++)
      {
        mz3DPolygon &p = Polygon (i);
        sprintf (buff, "    TRIANGLE (%d,%d,%d)\n", p [0], p [1], p [2]);
        PUTS (buff);
      } /* endfor */
      break;
    default:
      break;
  } /* endswitch */

  PUTS ("  )\n");
  return true;
}

void mz3DModel::Draw (mzDraftEditor *Canvas, bool iActive, bool iModified,
  bool iCopy)
{
  // Private integer 2D point structure
  struct Point { int x, y, xm, ym; bool selected; };
  int i, j;
  CHK (Point *vert = new Point [Vertices ()]);

  if (iModified)
  {
    for (i = 0; i < Vertices (); i++)
      vert [i].selected = Vertex (i).selected;
    for (i = 0; i < Polygons (); i++)
    {
      mz3DPolygon &p = Polygon (i);
      if (p.selected)
        for (j = p.Vertices () - 1; j >= 0; j--)
          vert [p [j]].selected = true;
    } /* endfor */
  } /* endif */

  // First project all vertices of the model
  for (i = 0; i < Vertices (); i++)
  {
    Canvas->world2win (Vertex (i), vert [i].x, vert [i].y);
    if (iModified && vert [i].selected)
      Canvas->world2winM (Vertex (i), vert [i].xm, vert [i].ym);
    else
    {
      vert [i].xm = vert [i].x;
      vert [i].ym = vert [i].y;
    } /* endif */
  }

  // First project and paint all vertices of the model
  Vector2 poly [MAX_POLYGON_VERTICES];
  Vector2 polym [MAX_POLYGON_VERTICES];
  bool drawsel = false;
  do
  {
    for (i = 0; i < Polygons (); i++)
    {
      mz3DPolygon &CurrentPoly = Polygon (i);
      if (drawsel != CurrentPoly.selected)
        continue;

      int vert_in_poly = CurrentPoly.Vertices ();
      for (j = 0; j < vert_in_poly; j++)
      {
        Point &p = vert [CurrentPoly [j]];
        poly [j].set (p.x, p.y);
        polym [j].set (p.xm, p.ym);
      } /* endfor */

      int color;
      if (iActive)
      {
        if (CurrentPoly.selected)
          color = Canvas->palette [CSPAL_DRAFT_SELPOLY];
        else
        {
          color = CurrentPoly.Color;
          if (color < 0)
            color = Color;
          if (color < 0)
            color = Canvas->palette [CSPAL_DRAFT_ACTPOLY];
        }
      }
      else
        color = Canvas->palette [CSPAL_DRAFT_POLY];
      Canvas->SetColor (CSPAL_DRAFT_SCRATCHPAD, color);

      for (j = 0; j < vert_in_poly; j++)
      {
        int prev = (j == 0 ? vert_in_poly - 1 : j - 1);
        if (!iModified || !iActive)
          Canvas->Line ((int)poly [prev].x, (int)poly [prev].y,
            (int)poly [j].x, (int)poly [j].y, CSPAL_DRAFT_SCRATCHPAD);
        else
        {
          if (iCopy)
            Canvas->Line ((int)poly [prev].x, (int)poly [prev].y,
              (int)poly [j].x, (int)poly [j].y, CSPAL_DRAFT_ACTPOLY);
          int linecolor = CSPAL_DRAFT_SCRATCHPAD;
          if (iActive
           && (vert [CurrentPoly [j]].selected
            || vert [CurrentPoly [prev]].selected))
            linecolor = CSPAL_DRAFT_MODIFIED;
          Canvas->Line ((int)polym [prev].x, (int)polym [prev].y,
            (int)polym [j].x, (int)polym [j].y, linecolor);
        } /* endif */
      } /* endfor */
    } /* endfor */
    drawsel ^= true;
  } while (drawsel); /* enddo */

  // And now paint all vertices of the model
  for (i = 0; i < Vertices (); i++)
  {
    int x = vert [i].x, y = vert [i].y;
    int xm = vert [i].xm, ym = vert [i].ym;
    int col = iActive ? (Vertex (i).selected ? CSPAL_DRAFT_SELVERTEX
                                             : CSPAL_DRAFT_VERTEX)
                      : CSPAL_DRAFT_INACTVERTEX;
    if (!iModified || !iActive)
      if (F_DRAFT_CROSSVERTEX)
      {
        Canvas->Line (x-1, y-1, x+1, y+1, col);
        Canvas->Line (x+1, y-1, x-1, y+1, col);
      }
      else
        Canvas->Line (x, y, x, y, col);
    else
    {
      if (iCopy)
        if (F_DRAFT_CROSSVERTEX)
        {
          Canvas->Line (x-1, y-1, x+1, y+1, CSPAL_DRAFT_VERTEX);
          Canvas->Line (x+1, y-1, x-1, y+1, CSPAL_DRAFT_VERTEX);
        }
        else
          Canvas->Line (x, y, x, y, CSPAL_DRAFT_VERTEX);
      if (F_DRAFT_CROSSVERTEX)
      {
        Canvas->Line (xm-1, ym-1, xm+1, ym+1, CSPAL_DRAFT_VERTEX);
        Canvas->Line (xm+1, ym-1, xm-1, ym+1, CSPAL_DRAFT_VERTEX);
      }
      else
        Canvas->Line (xm, ym, xm, ym, col);
    } /* endif */
  } /* endfor */
  CHK (delete [] vert);
}

void mz3DModel::Modify (mzDraftEditor *Canvas, bool iCopy)
{
  // Private integer 2D point structure
  int i, j;
  CHK (bool *vert = new bool [Vertices ()]);

  // Find which vertices should be modified
  for (i = 0; i < Vertices (); i++)
    vert [i] = Vertex (i).selected;
  for (i = 0; i < Polygons (); i++)
  {
    mz3DPolygon &p = Polygon (i);
    if (p.selected)
      for (j = p.Vertices () - 1; j >= 0; j--)
        vert [p [j]] = true;
  } /* endfor */

  if (iCopy)
  {
    CHK (int *vertidx = new int [Vertices ()]);
    int nvert = Vertices ();
    for (i = 0; i < nvert; i++)
      if (vert [i])
      {
        mz3DVertex *v = new mz3DVertex (Vertex (i));
        Vertex (i).selected = false;
        Canvas->Modify (*v);
        InsertVertex (v, &vertidx [i]);
        Vertex (vertidx [i]).selected = true;
      }
      else
        vertidx [i] = i;
    int npoly = Polygons ();
    for (i = 0; i < npoly; i++)
      if (Polygon (i).selected)
      {
        mz3DPolygon *p = new mz3DPolygon (Polygon (i));
        Polygon (i).selected = false;
        for (j = 0; j < p->Vertices (); j++)
          (*p) [j] = vertidx [(*p) [j]];
        int polyidx;
        InsertPolygon (p, &polyidx);
        Polygon (polyidx).selected = true;
      } /* endfor */
    CHK (delete [] vertidx);
    CHK (delete [] vert);
    return;
  } /* endif */

  // Modify all selected vertices of the model
  for (i = 0; i < Vertices (); i++)
    if (vert [i])
      Canvas->Modify (Vertex (i));
  CHK (delete [] vert);
}

bool mz3DModel::HasSelection ()
{
  int i;
  // Find which vertices should be modified
  for (i = 0; i < Vertices (); i++)
    if (Vertex (i).selected)
      return true;
  for (i = 0; i < Polygons (); i++)
    if (Polygon (i).selected)
      return true;
  return false;
}

bool mz3DModel::Load (csSector* iSector)
{
  Clear (true, true, true, true, true);
  char name [256];
  strncpy (name, iSector->get_name (), sizeof (name));
  int i, sl = strlen (name) - 1;
  for (i = sl; i >= 0; i--)
    if ((name [i] >= '0') && (name [i] <= '9'))
      sl = i;
    else if (name [i] == '#')
    {
      name [i] = 0;
      break;
    }
    else
      break;
  SetName (name);
  return Join (iSector, true);
}

bool mz3DModel::Join (csSector *iSector, bool iForce)
{
  if (Type != csmtRoom)
    return false;
  if (!iForce)
  {
    // Check if this sector is a part of this room
    char *SectName = iSector->get_name ();
    int nl = strlen (Name);
    int sl = strlen (SectName);
    if (nl > sl)
      return false;			// csSector name should be based on Name
    if (strncmp (Name, SectName, nl) != 0)
      return false;			// Not a subset of this room
    if (SectName [nl] != '#')
      return false;			// No '#' after basename
    while (++nl < sl)
      if ((SectName [nl] < '0') || (SectName [nl] > '9'))
        return false;
  } /* endif */

  int i, j;
  int *vert = new int [iSector->get_num_vertices ()];
  for (i = 0; i < iSector->get_num_vertices (); i++)
  {
    CHK (mz3DVertex *v = new mz3DVertex (iSector->vobj (i)));
    InsertVertex (v, &vert [i]);
  } /* endfor */
  for (i = 0; i < iSector->get_num_polygon (); i++)
  {
    Polygon3D *pt = (Polygon3D *)iSector->get_polygon (i);
    CHK (mz3DPolygon *p = new mz3DPolygon (this));
    p->SetName (pt->get_name ());
    for (j = 0; j < pt->get_num_vertices (); j++)
      (*p) [j] = vert [pt->get_vertices_idx () [j]];
    InsertPolygon (p);
  } /* endfor */
  return true;
}

bool mz3DModel::Load (ThingTemplate* iThing)
{
  Clear (true, true, true, true, true);
  SetName (iThing->get_name ());
  return Join (iThing);
}

bool mz3DModel::Join (ThingTemplate* iThing)
{
  int i, j;
  int *vert = new int [iThing->get_num_vertices ()];
  for (i = 0; i < iThing->get_num_vertices (); i++)
  {
    CHK (mz3DVertex *v = new mz3DVertex (iThing->vtex (i)));
    InsertVertex (v, &vert [i]);
  } /* endfor */
  for (i = 0; i < iThing->get_num_polygon (); i++)
  {
    PolygonTemplate *pt = iThing->get_polygon (i);
    CHK (mz3DPolygon *p = new mz3DPolygon (this));
    p->SetName (pt->get_name ());
    for (j = 0; j < pt->get_num_vertices (); j++)
      (*p) [j] = vert [pt->get_vertices_idx () [j]];
    InsertPolygon (p);
  } /* endfor */
  CHK (delete [] vert);
  return true;
}

bool mz3DModel::Load (csSpriteTemplate *iSprite)
{
  Clear (true, true, true, true, true);
  SetName (iSprite->get_name ());
  SetTexture (iSprite->get_texture ());
  return Join (iSprite);
}

bool mz3DModel::Join (csSpriteTemplate *iSprite)
{
  csFrame** frames = iSprite->get_frames ();
  int *vert = new int [iSprite->get_num_vertices ()];
  for (int frame = 0; frame < iSprite->get_num_frames (); frame++)
  {
    SetFrame (frame, frames [frame]->get_name ());
    for (int vertex = 0; vertex < iSprite->get_num_vertices (); vertex++)
    {
      if (frame == 0)
      {
        CHK (mz3DVertex *v = new mz3DVertex (frames [frame]->get_vertice (vertex)));
        InsertVertex (v, &vert [vertex]);
      }
      else
        Vertex (vert [vertex]) = frames [frame]->get_vertice (vertex);
      Vertex (vert [vertex]).SetTexel (frames [frame]->get_texel (vertex));
    } /* endfor */
  } /* endfor */

  Triangle* triangles = iSprite->get_triangles ();
  for (int triangle = 0; triangle < iSprite->get_num_triangles (); triangle++)
  {
    CHK (mz3DPolygon *p = new mz3DPolygon (this));
    p->SetName (NULL);
    (*p) [0] = vert [triangles [triangle].a];
    (*p) [1] = vert [triangles [triangle].b];
    (*p) [2] = vert [triangles [triangle].c];
    InsertPolygon (p, NULL);
  } /* endfor */
  CHK (delete [] vert);
  SetFrame (0);
  return true;
}

void mz3DModel::SelectAllVertices (int iCode)
{
  for (int i = Vertices () - 1; i >= 0; i--)
  {
    switch (iCode)
    {
      case cscmdMzVertexSelect:
      case cscmdMzVertexSelectAll:
        Vertex (i).selected = true;
        break;
      case cscmdMzVertexDeselect:
      case cscmdMzVertexDeselectAll:
        Vertex (i).selected = false;
        break;
      case cscmdMzVertexInvert:
      case cscmdMzVertexInvertAll:
        Vertex (i).selected = !Vertex (i).selected;
        break;
    } /* endswitch */
  } /* endfor */
}

void mz3DModel::SelectAllPolygons (int iCode)
{
  for (int i = Polygons () - 1; i >= 0; i--)
  {
    switch (iCode)
    {
      case cscmdMzPolySelect:
      case cscmdMzPolySelectAll:
        Polygon (i).selected = true;
        break;
      case cscmdMzPolyDeselect:
      case cscmdMzPolyDeselectAll:
        Polygon (i).selected = false;
        break;
      case cscmdMzPolyInvert:
      case cscmdMzPolyInvertAll:
        Polygon (i).selected = !Polygon (i).selected;
        break;
    } /* endswitch */
  } /* endfor */
}

void mz3DModel::DelSelectedVertices ()
{
  for (int i = Vertices () - 1; i >= 0; i--)
    if (Vertex (i).selected)
      DeleteVertex (i);
}

void mz3DModel::DelSelectedPolygons ()
{
  for (int i = Polygons () - 1; i >= 0; i--)
    if (Polygon (i).selected)
      DeletePolygon (i);
}

void mz3DModel::GetBoundingBox (Vector3 &iMin, Vector3 &iMax, bool iOnlySelected)
{
  // Private integer 2D point structure
  int i, j;
  CHK (bool *vert = new bool [Vertices ()]);

  if (iOnlySelected)
  {
    // Find which vertices should be modified
    for (i = 0; i < Vertices (); i++)
      vert [i] = Vertex (i).selected;
    for (i = 0; i < Polygons (); i++)
    {
      mz3DPolygon &p = Polygon (i);
      if (p.selected)
        for (j = p.Vertices () - 1; j >= 0; j--)
          vert [p [j]] = true;
    } /* endfor */
  } /* endif */

  iMin.set (+9999999999., +9999999999., +9999999999.);
  iMax.set (-9999999999., -9999999999., -9999999999.);

  for (i = Vertices () - 1; i >= 0; i--)
  {
    if (!iOnlySelected || vert [i])
    {
      mz3DVertex &v = Vertex (i);
      if (v.x < iMin.x) iMin.x = v.x;
      if (v.x > iMax.x) iMax.x = v.x;
      if (v.y < iMin.y) iMin.y = v.y;
      if (v.y > iMax.y) iMax.y = v.y;
      if (v.z < iMin.z) iMin.z = v.z;
      if (v.z > iMax.z) iMax.z = v.z;
    } /* endif */
  } /* endfor */
  CHK (delete [] vert);
}

void mz3DModel::GetBoundingBox (mzDraftEditor *Canvas, Vector3 &iMin,
  Vector3 &iMax, bool iOnlySelected)
{
  // Private integer 2D point structure
  int i, j;
  CHK (bool *vert = new bool [Vertices ()]);

  if (iOnlySelected)
  {
    // Find which vertices should be modified
    for (i = 0; i < Vertices (); i++)
      vert [i] = Vertex (i).selected;
    for (i = 0; i < Polygons (); i++)
    {
      mz3DPolygon &p = Polygon (i);
      if (p.selected)
        for (j = p.Vertices () - 1; j >= 0; j--)
          vert [p [j]] = true;
    } /* endfor */
  } /* endif */

  iMin.set (+9999999999., +9999999999., +9999999999.);
  iMax.set (-9999999999., -9999999999., -9999999999.);

  for (i = Vertices () - 1; i >= 0; i--)
  {
    if (!iOnlySelected || vert [i])
    {
      Vector3 v;
      Canvas->world2win (Vertex (i), v);
      if (v.x < iMin.x) iMin.x = v.x;
      if (v.x > iMax.x) iMax.x = v.x;
      if (v.y < iMin.y) iMin.y = v.y;
      if (v.y > iMax.y) iMax.y = v.y;
      if (v.z < iMin.z) iMin.z = v.z;
      if (v.z > iMax.z) iMax.z = v.z;
    } /* endif */
  } /* endfor */
}

void mz3DModel::GetTextures (csStrVector &iTexList)
{
  if (Type == csmtSprite)
    AddTexture (iTexList, Texture);
}

void mz3DModel::AddTexture (csStrVector &iTexList, TextureMM *iTexture)
{
  if (!iTexture)
    return;

  char buff [256];
  sprintf (buff, "TEXTURE '%s' (", iTexture->get_name ());
  if (iTexture->get_transparent () >= 0)
  {
    int tr, tg, tb;
    iTexture->get_transparent (tr, tg, tb);
    sprintf ((char *)&buff [strlen (buff)], "TRANSPARENT(%d,%d,%d)", tr, tg, tb);
  } /* endif */
  strcat (buff, ")");
  iTexList.Push (strnew (buff));
}
