/*
  Maze editor: 3D model class
  Copyright (C) 1998 by Jorrit Tyberghein
  Written by Andrew Zabolotny <bit@eltech.ru>

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __ME_MODEL_H__
#define __ME_MODEL_H__

#include "util/csvector.h"
#include "engine/objects/cssprite.h"
#include "geometry/math3d.h"

class mzDraftEditor;

/// Maximum allowed vertices in a polygon
#define MAX_POLYGON_VERTICES 16

/**
 * The mz3DVertex class represents a vertex of the 3D model
 */
class mz3DVertex : public Vector3
{
public:
  /// Vertex is selected?
  bool selected;
  /// Texture X & Y
  float u,v;

  /// Initialize needed fields
  mz3DVertex ()
  { selected = false; u = v = 0; }

  /// Initialize a 3D vertex from a 3D vector
  mz3DVertex (const Vector3 &copy)
  { x = copy.x; y = copy.y; z = copy.z; selected = false; u = v = 0; }

  inline void SetTexel (Vector2 &iTexel)
  { u = iTexel.x; v = iTexel.y; }

  /// Assign a 3D vector to 3D vertex
  inline void operator = (const Vector3 &v)
  { x = v.x; y = v.y; z = v.z; selected = false; u = this->v = 0; }

  /// Compare two vertices for equality
  inline bool operator == (const mz3DVertex &v) const
  { return (v.x == x) && (v.y == y) && (v.z == z); }
};

/// Polygon error codes
enum csPolyError
{
  /// Polygon is okay
  cspeOK,
  /// Polygon does not have enough vertices
  cspeUnderflow,
  /// Polygon has more than MAX_POLYGON_VERTICES vertices
  cspeOverflow,
  /// Polygon has some vertex included more than once
  cspeDuplicateVertex,
  /// Polygon has no parent model
  cspeNoParent,
  /// Polygon vertices are not coplanar
  cspeNonCoplanar,
  /// Polygon is not convex
  cspeNonConvex
};

class mz3DModel;

/**
 * This class represents a polygon of the 3D model.<p>
 * It contains a list of indexes into parent model object's Vertex
 * array; that is, mz3DPolygon's Vertex array contains just the indices of
 * corresponding mz3DVector objects in parent model's Vertex array.
 */
class mz3DPolygon
{
  /// The vertices that form this polygon
  csVector V;

public:
  /// Polygon is selected?
  bool selected;
  /// Parent model
  mz3DModel *parent;
  /// Polygon color (one of cs_Color_XXX) or -1 for default
  int Color;
  /// Polygon name (initially automatically generated)
  char *Name;

  /// Initialize the polygon object
  mz3DPolygon (mz3DModel *iParent);

  /// Copy constructor
  mz3DPolygon (mz3DPolygon &iCopy);

  /// Destructor
  virtual ~mz3DPolygon ();

  /// Get a reference to vertex iVertex
  inline int& operator [] (int iVertex)
  { return (int&)V [iVertex]; }

  /// Get a reference to vertex by index into this polygon
  mz3DVertex& Vertex (int iVertex);

  /// Get number of vertices in polygon
  inline int Vertices () const
  { return V.Length (); }

  /**
   * This method is called to notify polygon that a vertex has been deleted
   * from parent model. Polygon should re-scan its Vertex array upon reception
   * of this model and decrement indices of all vertices which's indices are
   * above index of the vertex that has been deleted. If polygon contains
   * deleted vertex, it should be deleted too and returns "false".
   */
  bool NotifyVertexDeleted (int iVertex);

  /// Compare two polygons for equality
  bool operator == (const mz3DPolygon &p) const;

  /// Check polygon for correctness and return error code
  csPolyError Check ();

  /// Return a descriptive string for given error code or NULL
  char *ErrorText (csPolyError iCode);

  /// Set name of this polygon
  void SetName (char *iName);

  /// Delete a vertex
  void DeleteVertex (int iVertex)
  { V.Delete (iVertex); }

  /// Reverse polygon's visible side
  void Reverse ();
};

/**
 * This class is used to store actions for 3D sprites. An action is a
 * specific sequence of frames with specific delays. Actions are kept
 * in a csObjVector object in mz3DModel class.
 */
class mzSpriteAction : public csBase
{
  /// Array of frame names
  csStrVector Names;
  /// Array of frame delays
  csVector Delays;
public:
  /// Name of this action
  char *Name;

  /// Create a sprite action object from a csSpriteAction object
  mzSpriteAction (csSpriteAction *iAction);
  /// Destroy a sprite action object
  virtual ~mzSpriteAction ();

  /// Notify action that a frame has been deleted
  void NotifyFrameDeleted (char *iFrameName);

  /// Check if two actions are same
  bool operator == (const mzSpriteAction &a);

  /// Return number of frames in action
  int Frames () const { return Names.Length (); }

  /// Return a reference to given frame delay
  int& Delay (int iFrame) const { return (int &)Delays.Get (iFrame); }

  /// Return given frame name
  char *FrameName (int iFrame) const { return (char *)Names.Get (iFrame); }

  /// Set given frame name
  bool SetFrameName (int iFrame, char *iName);
};

/// Model type: room, thing or sprite
enum csModelType
{
  csmtUnknown,
  csmtRoom,
  csmtThing,
  csmtSprite
};

/**
 * This class encapsulates a 3D model object. It contains all routines
 * needed for editing, manipulating, loading and storing an abstract
 * 3D model.
 */
class mz3DModel : public csBase
{
  /// Private class which defines a private vector used to store 3D vertices
  class csVertexVector : public csVector
  {
  public:
    /// Initialize vertex vector object
    csVertexVector () : csVector (64, 64)
    { ; }

    /// Delete all vertices before deleting object
    virtual ~csVertexVector ();

    /// Virtual function which frees a vector element
    virtual void FreeItem (csSome Item);
  };

  /// Private class which defines a private vector used to store 3D polygons
  class csPolygonVector : public csVector
  {
  public:
    /// Initialize polygon vector object
    csPolygonVector () : csVector (16, 16)
    { ; }

    /// Delete all polygons before deleting object
    virtual ~csPolygonVector ();

    /// Virtual function which frees a vector element
    virtual void FreeItem (csSome Item);
  };

  /// Private class which defines a private vector used to store lights
  class csLightVector : public csVector
  {
  public:
    /// Initialize polygon vector object
    csLightVector () : csVector (16, 16)
    { ; }

    /// Delete all polygons before deleting object
    virtual ~csLightVector ();

    /// Virtual function which frees a vector element
    virtual void FreeItem (csSome Item);
  };

  /// All vertices in this model
  csObjVector V;
  /// All polygons in this model
  csPolygonVector P;
  /// All lights in this model
  csLightVector L;
  /// All 3D sprite actions
  csObjVector A;
  /// Model file pathname
  char *FileName;
  // Active model frame
  int Frame;
  /// Frame names
  csStrVector FrameName;
  /// Texture (only for sprites)
  TextureMM *Texture;

public:
  /// Model type: room, thing or sprite
  csModelType Type;
  /// Model color (one of cs_Color_XXX) or -1 for default
  int Color;
  /// Model name
  char *Name;

  /// Create the 3D model object, optionaly read from text buffer
  mz3DModel (csModelType iType, char *iFileName);
  /// Destroy the 3D model object
  virtual ~mz3DModel ();

  /// Query total number of vertices
  inline int Vertices (int iFrame = 0)
  { return ((csVertexVector *)V [iFrame])->Length (); }
  /// Query total number of polygons
  inline int Polygons ()
  { return P.Length (); }
  /// Query total number of lights
  inline int Lights ()
  { return L.Length (); }
  /// Query total number of frames
  inline int Frames ()
  { return V.Length (); }
  inline int Actions ()
  { return A.Length (); }

  /// Return a reference to given vertex
  inline mz3DVertex& Vertex (int iVertex, int iFrame = -1)
  { return *(mz3DVertex *) (*(csVertexVector *)V [iFrame >= 0 ? iFrame : Frame])[iVertex]; }
  /// Return a reference to given polygon
  inline mz3DPolygon& Polygon (int iPolygon)
  { return *(mz3DPolygon *) P [iPolygon]; }
  /// Return a reference to given action
  inline mzSpriteAction& Action (int iAction)
  { return *(mzSpriteAction *) A [iAction]; }

  /// Insert a vertex and return its number
  bool InsertVertex (mz3DVertex *iVertex, int *iNum = NULL);
  /// Delete a vertex
  bool DeleteVertex (int iVertex);

  /// Insert a polygon and return its number
  bool InsertPolygon (mz3DPolygon *iPolygon, int *iNum = NULL);
  /// Delete a polygon
  bool DeletePolygon (int iPolygon);

  /// Insert a sprite action
  bool InsertAction (mzSpriteAction *iAction, int *iNum = NULL);
  /// Delete a sprite action
  bool DeleteAction (int iAction);

  /// Save this model into a stream
  bool Save (int (*Put) (void *iStream, void *iData, int iSize), void *iStream);
  /// Load the model from a text buffer
  bool Load (char *iBuffer);

  /// Set name of this model
  void SetName (char *iName);
  /// Get model file name
  char *GetFileName ()
  { return FileName; }
  /// Set file name and path of this model
  void SetFileName (char *iFileName);

  /// Get texture (sprites only)
  TextureMM *GetTexture () { return Texture; }
  /// Set texture (sprites only)
  void SetTexture (TextureMM *iTexture) { Texture = iTexture; }

  /// Set name of given frame
  void SetFrameName (int iFrame, char *iName);
  /// Get name of given frame
  inline char *GetFrameName (int iFrame)
  { return (char *) FrameName [iFrame]; }

  /// Set active frame number
  bool SetFrame (int iFrame, char *iFrameName = NULL);
  /// Get active frame number
  int GetFrame () { return Frame; }
  /// Delete a frame or clear frame if there are no more frames
  void DeleteFrame (int iFrame);

  /// Clear model
  void Clear (bool iVertices, bool iFaces, bool iLights, bool iFrames, bool iActions);

  /// Draw the model
  virtual void Draw (mzDraftEditor *Canvas, bool iActive, bool iModified, bool iCopy);

  /// Modify selected model vertices/polys using Canvas's modification parameters
  void Modify (mzDraftEditor *Canvas, bool iCopy);

  /// Load model from a sector object
  bool Load (csSector *iSector);
  /// Load model from a thing template object
  bool Load (ThingTemplate *iThing);
  /// Load model from a sprite object
  bool Load (csSpriteTemplate *iSprite);

  /// Join a sector to this model (if it is a room)
  bool Join (csSector *iSector, bool iForce = false);
  /// Join a thing to this model (if it is a thing)
  bool Join (ThingTemplate *iThing);
  /// Join a sprite to this model (if it is a sprite)
  bool Join (csSpriteTemplate *iSprite);

  /// Select, deselect or invert all vertices of model
  void SelectAllVertices (int iCode);
  /// Select, deselect or invert all polygons of model
  void SelectAllPolygons (int iCode);

  /// Delete all selected vertices
  void DelSelectedVertices ();
  /// Delete all selected polygons
  void DelSelectedPolygons ();

  /// Check if model has any selected elements
  bool HasSelection ();

  /// Find the minimal and maximal coordinates of model's vertices
  void GetBoundingBox (Vector3 &iMin, Vector3 &iMax, bool iOnlySelected);
  /// Find the minimal and maximal coordinates of model's vertices in window space
  void GetBoundingBox (mzDraftEditor *Canvas, Vector3 &iMin, Vector3 &iMax,
    bool iOnlySelected);

  /// Add all textures used in model to a list
  void GetTextures (csStrVector &iTexList);

private:
  void AddTexture (csStrVector &iTexList, TextureMM *iTexture);
};

#endif // __ME_MODEL_H__
