/*
    A very primitive editor for editing vertices and polygons.
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "general/cs.h"
#include "engine/texture.h"
#include "util/util.h"
#include "util/inifile.h"
#include "system/sysdriv.h"

// Configuration file
#define MODE_NORMAL 0
#define MODE_EDIT_PLANE 1
#define MODE_EDIT_VERTEX 2
#define MODE_EDIT_POLYGON 3

#define PLANE_VERTEX_RADIUS .1

#define SPEED_FAST   .4
#define SPEED_NORMAL .2
#define PSPEED_NORMAL 1.0
#define PSPEED_FAST   2.0
#define PSPEED_SLOW   0.5
//=================================================================================================

class PEObject
{
public:
  PEObject* next, * prev;
  PEObject* sel_prev, * sel_next;

  PEObject ()
  {
    next = prev = sel_prev = sel_next = NULL;
  };
  virtual ~PEObject () { }

  virtual void move (csCamera* camera, Vector3& pos) = 0;
  virtual void transform (csCamera* camera, Matrix3& tr) = 0;
  virtual bool can_be_deleted () { return true; }
};

//=================================================================================================

class PEPlane : public PEObject
{
public:
  Vector3 v0, v1, v2;
  Vector3 help1_v1, help1_v2;
  Vector3 help2_v1, help2_v2;
  Vector3 help3_v1, help3_v2;

  PEPlane () {}

  void update_plane ();
  virtual void move (csCamera* camera, Vector3& pos);
  virtual void transform (csCamera* camera, Matrix3& tr);

  static void intersect (PEPlane* p1, PEPlane* p2, Vector3& V, Vector3& N3);
};

void PEPlane::move (csCamera* camera, Vector3& pos)
{
  v0 += camera->cam2world_rel(pos);
  v1 += camera->cam2world_rel(pos);
  v2 += camera->cam2world_rel(pos);
  update_plane ();
}

void PEPlane::transform (csCamera*, Matrix3& tr)
{
  v1 = tr*(v1-v0) + v0;
  v2 = tr*(v2-v0) + v0;
  update_plane ();
}

void PEPlane::update_plane ()
{
  help1_v1 = .2 * (v1 - v0) + .8 * (v2 - v0) + v0;
  help1_v2 = .8 * (v1 - v0) + .2 * (v2 - v0) + v0;
  help2_v1 = .15 * (v1 - v0) + .75 * (v2 - v0) + v0;
  help2_v2 = .75 * (v1 - v0) + .15 * (v2 - v0) + v0;
  help3_v1 = .10 * (v1 - v0) + .70 * (v2 - v0) + v0;
  help3_v2 = .70 * (v1 - v0) + .10 * (v2 - v0) + v0;
}

void PEPlane::intersect (PEPlane* p1, PEPlane* p2, Vector3& V, Vector3& N3)
{
  Vector3 N1, N2;
  float D1, D2, D3;
  Plane::calc_plane (p1->v0, p1->v1, p1->v2, N1, D1);
  Plane::calc_plane (p2->v0, p2->v1, p2->v2, N2, D2);
  D3 = 0;
  N3 = N1 % N2;		// The direction of the line is perpendicular to the normals of both planes.
  Matrix3 N (N1.x, N1.y, N1.z, N2.x, N2.y, N2.z, N3.x, N3.y, N3.z);
  N.invert ();
  Vector3 K;
  K.x = -D1; K.y = -D2; K.z = -D3;
  V = N * K;		// This gives the location of the intersection of the two planes and
  			// a third plane perpendicular to the two planes and going through (0,0,0)
}

//=================================================================================================

class PEVertex : public PEObject
{
public:
  Vector3 v;
  int used;		// Usage count for polygons.
  int num;		// Number used while saving for indicating the index of this vertex.
  int dirty_polygon;	// Number of polygons that need updating after this vertex moved.

  PEVertex () { used = dirty_polygon = 0; }
  virtual void move (csCamera* camera, Vector3& pos);
  virtual void transform (csCamera* camera, Matrix3& tr);
  virtual bool can_be_deleted () { return used == 0; }
  void move (const Vector3& vn);
};

void PEVertex::move (csCamera* camera, Vector3& pos)
{
  v += camera->cam2world_rel(pos);
  dirty_polygon = used;
}

void PEVertex::move (const Vector3& vn)
{
  v = vn;
  dirty_polygon = used;
}

void PEVertex::transform (csCamera*, Matrix3&)
{
}

//=================================================================================================

class PEPolygon : public PEObject
{
public:
  int num_vertices;
  PEVertex** vertices;
  Vector3 center;		// Calculated center of polygon.
  float A, B, C, D;		// Plane equation of polygon.

  PEPolygon ()
  {
    num_vertices = 0;
    vertices = NULL;
  }
  virtual ~PEPolygon ();
  void update ();
  void update_if_needed ();
  void clear ();
  virtual void move (csCamera* camera, Vector3& pos);
  virtual void transform (csCamera* camera, Matrix3& tr);

  void set_num_vertices (int num);
  void set_vertex (int i, PEVertex* v);

  bool visible (csCamera* camera);
  PEPlane* create_plane ();
  void align_plane (PEPlane* p);
};

PEPlane* PEPolygon::create_plane ()
{
  CHK (PEPlane* p = new PEPlane ());
  align_plane (p);
  return p;
}

void PEPolygon::align_plane (PEPlane* p)
{
  Vector3 normal (A, B, C);
  p->v0 = vertices[0]->v;
  p->v1 = Vector3::unit (vertices[1]->v - vertices[0]->v) + vertices[0]->v;
  p->v2 = (normal % (p->v1-p->v0)) + p->v0;
}

void PEPolygon::set_num_vertices (int num)
{
  clear ();
  CHK (vertices = new PEVertex* [num]);
  num_vertices = num;
}

void PEPolygon::set_vertex (int i, PEVertex* v)
{
  vertices[i] = v;
  v->used++;
}

void PEPolygon::clear ()
{
  int i;
  for (i = 0 ; i < num_vertices ; i++)
    vertices[i]->used--;
  if (vertices) { CHK (delete [] vertices); vertices = NULL; }
}

PEPolygon::~PEPolygon ()
{
  clear ();
}

void PEPolygon::move (csCamera* camera, Vector3& pos)
{
  int i;
  for (i = 0 ; i < num_vertices ; i++)
    vertices[i]->move (camera, pos);
}

void PEPolygon::transform (csCamera*, Matrix3&)
{
}

void PEPolygon::update ()
{
  A = 0;
  B = 0;
  C = 0;
  int i, i1;
  float x1, y1, z1, x, y, z;

  Vector3 vmin (1000000,1000000,1000000);
  Vector3 vmax (-1000000,-10000000,-1000000);
  i1 = num_vertices-1;
  for (i = 0 ; i < num_vertices ; i++)
  {
    x = vertices[i]->v.x;
    y = vertices[i]->v.y;
    z = vertices[i]->v.z;
    if (x < vmin.x) vmin.x = x;
    if (y < vmin.y) vmin.y = y;
    if (z < vmin.z) vmin.z = z;
    if (x > vmax.x) vmax.x = x;
    if (y > vmax.y) vmax.y = y;
    if (z > vmax.z) vmax.z = z;
    x1 = vertices[i1]->v.x;
    y1 = vertices[i1]->v.y;
    z1 = vertices[i1]->v.z;
    A += (z1+z) * (y-y1);
    B += (x1+x) * (z-z1);
    C += (y1+y) * (x-x1);
    i1 = i;
  }

  float d = sqrt (A*A + B*B + C*C);

  if (d < SMALL_EPSILON) d = SMALL_EPSILON;

  A /= d;
  B /= d;
  C /= d;

  x = vertices[0]->v.x;
  y = vertices[0]->v.y;
  z = vertices[0]->v.z;
  D = -(A*x+B*y+C*z);

  center.x = (vmin.x+vmax.x)/2;
  center.y = (vmin.y+vmax.y)/2;
  center.z = (vmin.z+vmax.z)/2;
}

void PEPolygon::update_if_needed ()
{
  int i;
  bool dirty = false;
  for (i = 0 ; i < num_vertices ; i++)
    if (vertices[i]->dirty_polygon)
    {
      vertices[i]->dirty_polygon--;
      dirty = true;
      break;
    }
  if (dirty) update ();
}

bool PEPolygon::visible (csCamera* camera)
{
  update_if_needed ();
  return Plane::classify (A, B, C, D, camera->v_origin()) < 0;
}

//=================================================================================================

class PEList
{
private:
  PEObject* objects;
  PEObject* sel_objects;

public:
  PEList () { objects = sel_objects = NULL; }
  ~PEList () { delete_objects (); }

  void delete_objects ();
  bool delete_object (PEObject* object, bool force = false);
  void add_new_object (PEObject* object);
  PEObject* first () { return objects; }

  void delete_selected ();
  void clear_selected ();
  PEObject* first_selected () { return sel_objects; }
  int count_selected ();
  bool is_selected (PEObject* v);
  void deselect (PEObject* v);
  void select (PEObject* v);
};

void PEList::delete_objects ()
{
  while (objects) delete_object (objects, true);
  sel_objects = NULL;
}

bool PEList::delete_object (PEObject* p, bool force)
{
  if (!force && !p->can_be_deleted ())
    return false;

  deselect (p);

  if (p->next) p->next->prev = p->prev;
  if (p->prev) p->prev->next = p->next;
  else objects = p->next;
  CHK (delete p);
  return true;
}

void PEList::add_new_object (PEObject* object)
{
  object->next = objects;
  object->prev = NULL;
  if (objects) objects->prev = object;
  objects = object;
}

void PEList::delete_selected ()
{
  PEObject* s = sel_objects;
  int cnt = 0;
  while (s)
  {
    if (!s->can_be_deleted ())
    {
      CsPrintf (MSG_STDOUT, "Some objects can't be deleted!\n");
      return;
    }
    cnt++;
    s = s->sel_next;
  }
  while (sel_objects) delete_object (sel_objects);
  CsPrintf (MSG_STDOUT, "Deleted %d objects.\n", cnt);
}

void PEList::clear_selected ()
{
  while (sel_objects) deselect (sel_objects);
}

int PEList::count_selected ()
{
  int cnt = 0;
  PEObject* p = sel_objects;
  while (p) { cnt++; p = p->sel_next; }
  return cnt;
}

bool PEList::is_selected (PEObject* v)
{
  return v->sel_prev || v->sel_next || v == sel_objects;
}

void PEList::deselect (PEObject* v)
{
  if (!is_selected (v)) return;

  if (v->sel_next) v->sel_next->sel_prev = v->sel_prev;
  if (v->sel_prev) v->sel_prev->sel_next = v->sel_next;
  else sel_objects = v->sel_next;
  v->sel_prev = v->sel_next = NULL;
}

void PEList::select (PEObject* v)
{
  if (is_selected (v)) return;

  // Add to selection.
  v->sel_next = sel_objects;
  v->sel_prev = NULL;
  if (sel_objects) sel_objects->sel_prev = v;
  sel_objects = v;
}


//=================================================================================================

class PrimEdit : public SysSystemDriver
{
public:
  csCamera* camera;
  Textures* txt;

  PEList* planes;
  PEList* vertices;
  PEList* polygons;
  PEVertex* drag_v;
  bool drag_border;
  int border1_x, border1_y;
  int border2_x, border2_y;

  bool grid;
  bool discrete;
  int mode;
  int vmode;
  PEList* list (int m = -1);

  PrimEdit ();
  ~PrimEdit ();
  virtual bool Open (char* title);

  virtual void NextFrame (long elapsed_time, long current_time);

  void DrawFrame ();
  void draw_vertex (Vector3& v, int col, bool cross = false);
  void draw_line (Vector3& v1, Vector3& v2, int col);

  void eatkeypress (int key, bool shift, bool alt, bool ctrl);
  void eatmousedown (int button, int x, int y, bool shift, bool alt, bool ctrl);
  void eatmouseup (int button, int x, int y);
  void eatmousemove (int x, int y);
  void eatmousedown1 (int x, int y, bool shift, bool alt, bool ctrl);
  void eatmousedown2 (int x, int y, bool shift, bool alt, bool ctrl);
  void eatmousedown3 (int x, int y, bool shift, bool alt, bool ctrl);
  void handle_key_forward (float speed, bool shift, bool alt, bool ctrl);
  void handle_key_backwards (float speed, bool shift, bool alt, bool ctrl);
  void handle_key_left (float speed, bool shift, bool alt, bool ctrl);
  void handle_key_right (float speed, bool shift, bool alt, bool ctrl);
  void handle_key_pgup (float, bool shift, bool alt, bool ctrl);
  void handle_key_pgdn (float, bool shift, bool alt, bool ctrl);

  PEPlane* search_plane (int x, int y);
  PEVertex* search_vertex (int x, int y);
  PEPolygon* search_polygon (int x, int y);
  void select_vertices_box (int x1, int y1, int x2, int y2);
  bool perspective (Vector3& v, Vector2& persp, float radius, float& pradius);
  bool hit_vertex (Vector3& v, float radius, int x, int y);
  bool hit_vertex (Vector3& v, float radius, int x1, int y1, int x2, int y2);
  bool hit_line (Vector3& v1, Vector3& v2, int x, int y);
  void new_polygon ();
  void select_polygons ();
  void select_vertices ();
  void clone_vertices ();

  void align_plane ();
  bool constrain (PEVertex* v, int x, int y);
  void constrain (PEVertex* v, PEPlane* p, int x, int y);
  void constrain (PEVertex* v, PEPlane* p1, PEPlane* p2, int x, int y);

  void clear_world ();
  void save_world ();
  void load_world ();
};

PrimEdit::PrimEdit () : SysSystemDriver ()
{
  CHK (planes = new PEList ());
  CHK (vertices = new PEList ());
  CHK (polygons = new PEList ());
  mode = MODE_NORMAL;
  vmode = MODE_NORMAL;
  drag_v = NULL;
  grid = true;
  drag_border = false;
  discrete = false;

  CHK (camera = new csCamera ());
  if (csCamera::aspect == 0)
    csCamera::aspect = FRAME_HEIGHT;
  csCamera::inv_aspect = 1./csCamera::aspect;

  txt = NULL;
}

void PrimEdit::clear_world ()
{
  planes->delete_objects ();
  vertices->delete_objects ();
  polygons->delete_objects ();
}

void PrimEdit::load_world ()
{
  clear_world ();
  FILE* fp = fopen ("world.pri", "r");
  if (!fp)
  {
    CsPrintf (MSG_STDOUT, "Error opening file 'world.pri' for reading!\n");
    return;
  }

  int num, i;
  char buf[1024];
  PEVertex** verts;
  CHK (verts = new PEVertex* [1000]);	// @@@ BAD hardcoded limit!

  fgets (buf, 1023, fp);		// Skip WORLD.
  fgets (buf, 1023, fp);		// Skip TEXTURES....
  fgets (buf, 1023, fp);		// Skip empty line.

  while (fgets (buf, 1023, fp))
  {
    if (!*buf || *buf == 13 || *buf == 10) break;
    fgets (buf, 1023, fp);	// Skip PLANE command.
    float xo, yo, zo, x1, y1, z1, x2, y2, z2;
    sscanf (buf, "    ORIG (%f,%f,%f) FIRST (%f,%f,%f) SECOND (%f,%f,%f)\n", &xo, &yo, &zo, &x1, &y1, &z1, &x2, &y2, &z2);
    fgets (buf, 1023, fp);	// Skip FIRST_LEN command.
    fgets (buf, 1023, fp);	// Skip ).
    CHK (PEPlane* v = new PEPlane ());
    planes->add_new_object (v);
    v->v0.x = xo; v->v0.y = yo; v->v0.z = zo;
    v->v1.x = x1; v->v1.y = y1; v->v1.z = z1;
    v->v2.x = x2; v->v2.y = y2; v->v2.z = z2;
    v->update_plane ();
  }

  fgets (buf, 1023, fp);		// Skip SECTOR.
  fgets (buf, 1023, fp);		// Skip TEXNR.

  num = 0;
  while (fgets (buf, 1023, fp))
  {
    if (!*buf || *buf == 13 || *buf == 10) break;
    float x, y, z;
    sscanf (buf, "    VERTEX (%f,%f,%f)\n", &x, &y, &z);
    CHK (PEVertex* v = new PEVertex ());
    vertices->add_new_object (v);
    verts[num] = v; num++;
    v->v.x = x; v->v.y = y; v->v.z = z;
  }

  while (fgets (buf, 1023, fp))
  {
    if (!*buf || *buf == 13 || *buf == 10) break;
    fgets (buf, 1023, fp);	// Skip POLYGON command.
    int list[100];
    ScanStr (buf, "      VERTICES (%D)\n", list, &num);
    fgets (buf, 1023, fp);	// Skip ).
    CHK (PEPolygon* v = new PEPolygon ());
    polygons->add_new_object (v);
    v->set_num_vertices (num);
    for (i = 0 ; i < num ; i++)
      v->set_vertex (i, verts[list[i]]);
    v->update ();
  }


  fclose (fp);
  CHK (delete [] verts);
  CsPrintf (MSG_STDOUT, "World file loaded from 'world.pri'.\n");
}

void PrimEdit::save_world ()
{
  FILE* fp = fopen ("world.pri", "w");
  if (!fp)
  {
    CsPrintf (MSG_STDOUT, "Error opening file 'world.pri' for writing!\n");
    return;
  }
  int num, i;

  fprintf (fp, "WORLD (\n  TEXTURES (MAX_TEXTURES (4) TEXTURE 'stone4.gif' ())\n\n");

  PEPlane* p = (PEPlane*)planes->first ();
  num = 1;
  while (p)
  {
    fprintf (fp, "  PLANE 'pl%d' (\n", num); num++;
    fprintf (fp, "    ORIG (%f,%f,%f) FIRST (%f,%f,%f) SECOND (%f,%f,%f)\n",
    	p->v0.x, p->v0.y, p->v0.z, p->v1.x, p->v1.y, p->v1.z, p->v2.x, p->v2.y, p->v2.z);
    fprintf (fp, "    FIRST_LEN (1) SECOND_LEN (1)\n");
    fprintf (fp, "  )\n");
    p = (PEPlane*)p->next;
  }

  fprintf (fp, "\n  SECTOR 'room' (\n");
  fprintf (fp, "    TEXNR ('stone4.gif') TEXLEN (1)\n");

  num = 0;
  PEVertex* v = (PEVertex*)vertices->first ();
  while (v)
  {
    fprintf (fp, "    VERTEX (%f,%f,%f)\n", v->v.x, v->v.y, v->v.z);
    v->num = num;
    num++;
    v = (PEVertex*)v->next;
  }

  fprintf (fp, "\n");

  PEPolygon* po = (PEPolygon*)polygons->first ();
  num = 1;
  while (po)
  {
    fprintf (fp, "    POLYGON 'po%d' (\n", num); num++;
    fprintf (fp, "      VERTICES (");
    fprintf (fp, "%d", po->vertices[0]->num);
    for (i = 1 ; i < po->num_vertices ; i++)
      fprintf (fp, ",%d", po->vertices[i]->num);
    fprintf (fp, ")\n    )\n");
    po = (PEPolygon*)po->next;
  }

  fprintf (fp, "\n    LIGHT (0,0,0:10,1,1,1,0)\n");
  fprintf (fp, "\n  )\n)\n");

  fclose (fp);
  CsPrintf (MSG_STDOUT, "World file saved to 'world.pri'.\n");
}

PEList* PrimEdit::list (int m)
{
  if (m == -1) m = mode;
  switch (mode)
  {
    case MODE_EDIT_PLANE: return planes;
    case MODE_EDIT_VERTEX: return vertices;
    case MODE_EDIT_POLYGON: return polygons;
  }
  return NULL;
}

PrimEdit::~PrimEdit ()
{
  if (camera) CHKB (delete camera);
  CHK (delete planes);
  CHK (delete vertices);
  CHK (delete polygons);
  if (txt) CHKB (delete txt);
}

bool PrimEdit::Open (char* title)
{
  bool rc = SysSystemDriver::Open (title);
  if (rc)
  {
    CHK (txt = new Textures (4));
    txt->alloc_rgb (0, 0, 0, 0);
    txt->alloc_rgb (255, 255, 255, 0);
    txt->alloc_rgb (255, 0, 0, 0);
    txt->alloc_rgb (0, 255, 0, 0);
    txt->alloc_rgb (0, 0, 255, 0);
    txt->alloc_rgb (255, 0, 255, 0);
    txt->alloc_rgb (255, 255, 0, 0);
    txt->alloc_rgb (0, 255, 255, 0);
    txt->alloc_rgb (128, 128, 128, 0);
    txt->alloc_rgb (128, 0, 0, 0);
    txt->alloc_palette ();
    txt->initialize_limited ();
    if (csCamera::aspect == 0)
      csCamera::aspect = FRAME_HEIGHT;
    csCamera::inv_aspect = 1./csCamera::aspect;
  }
  return rc;
}
 
void PrimEdit::handle_key_forward (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    Vector3 pos (0, 0, 0);
    if (alt) pos.y = shift ? SPEED_FAST : SPEED_NORMAL;
    else pos.z = shift ? SPEED_FAST : SPEED_NORMAL;
    while (p) { p->move (camera, pos); p = p->sel_next; }
  }
  else if (alt)
  {
    if (ctrl) camera->move (.01*VEC_UP);
    else if (shift) camera->move (.4*VEC_UP);
    else camera->move (.2*VEC_UP);
  }
  else if (ctrl) camera->move (.01*VEC_FORWARD);
  else if (shift) camera->move (1.2*VEC_FORWARD);
  else camera->move (.6*VEC_FORWARD);
}

void PrimEdit::handle_key_backwards (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    Vector3 pos (0, 0, 0);
    if (alt) pos.y = -(shift ? SPEED_FAST : SPEED_NORMAL);
    else pos.z = -(shift ? SPEED_FAST : SPEED_NORMAL);
    while (p) { p->move (camera, pos); p = p->sel_next; }
  }
  else if (alt)
  {
    if (ctrl) camera->move (.01*VEC_DOWN);
    else if (shift) camera->move (.4*VEC_DOWN);
    else camera->move (.2*VEC_DOWN);
  }
  else if (ctrl) camera->move (.01*VEC_BACKWARD);
  else if (shift) camera->move (1.2*VEC_BACKWARD);
  else camera->move (.6*VEC_BACKWARD);
}

void PrimEdit::handle_key_left (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    if (alt)
    {
      Vector3 pos (-(shift ? SPEED_FAST : SPEED_NORMAL), 0, 0);
      while (p) { p->move (camera, pos); p = p->sel_next; }
    }
    else
    {
      Matrix3 rot (0, 0, 0, 0, 1, 0, 0, 0, 0);
      rot.m11 =  cos (shift ? .2 : .1);
      rot.m13 = -sin (shift ? .2 : .1);
      rot.m31 =  sin (shift ? .2 : .1);
      rot.m33 =  cos (shift ? .2 : .1);
      while (p) { p->transform (camera, rot); p = p->sel_next; }
    }
  }
  else if (alt)
  {
    if (ctrl) camera->move (.01*VEC_LEFT);
    else if (shift) camera->move (.4*VEC_LEFT);
    else camera->move (.2*VEC_LEFT);
  }
  else if (discrete) camera->rotate (VEC_ROT_LEFT, M_PI/4);
  else if (ctrl) camera->rotate (VEC_ROT_LEFT, .005);
  else if (shift) camera->rotate (VEC_ROT_LEFT, .2);
  else camera->rotate (VEC_ROT_LEFT, .1);
}

void PrimEdit::handle_key_right (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    if (alt)
    {
      Vector3 pos ((shift ? SPEED_FAST : SPEED_NORMAL), 0, 0);
      while (p) { p->move (camera, pos); p = p->sel_next; }
    }
    else
    {
      Matrix3 rot (0, 0, 0, 0, 1, 0, 0, 0, 0);
      rot.m11 =  cos (-(shift ? .2 : .1));
      rot.m13 = -sin (-(shift ? .2 : .1));
      rot.m31 =  sin (-(shift ? .2 : .1));
      rot.m33 =  cos (-(shift ? .2 : .1));
      while (p) { p->transform (camera, rot); p = p->sel_next; }
    }
  }
  else if (alt)
  {
    if (ctrl) camera->move (.01*VEC_RIGHT);
    else if (shift) camera->move (.4*VEC_RIGHT);
    else camera->move (.2*VEC_RIGHT);
  }
  else if (discrete) camera->rotate (VEC_ROT_RIGHT, M_PI/4);
  else if (ctrl) camera->rotate (VEC_ROT_RIGHT, .005);
  else if (shift) camera->rotate (VEC_ROT_RIGHT, .2);
  else camera->rotate (VEC_ROT_RIGHT, .1);
}

void PrimEdit::handle_key_pgup (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    Matrix3 rot (0, 0, 0, 0, 0, 0, 0, 0, 0);
    if (alt)
    {
      rot.m33 = 1;
      rot.m11 =  cos (-(shift ? .2 : .1));
      rot.m12 = -sin (-(shift ? .2 : .1));
      rot.m21 =  sin (-(shift ? .2 : .1));
      rot.m22 =  cos (-(shift ? .2 : .1));
    }
    else
    {
      rot.m11 = 1;
      rot.m22 =  cos (-(shift ? .2 : .1));
      rot.m23 = -sin (-(shift ? .2 : .1));
      rot.m32 =  sin (-(shift ? .2 : .1));
      rot.m33 =  cos (-(shift ? .2 : .1));
    }
    while (p) { p->transform (camera, rot); p = p->sel_next; }
  }
  else if (alt)
  {
    if (ctrl) camera->rotate (VEC_TILT_LEFT, .005);
    else camera->rotate (VEC_TILT_LEFT, .1);
  }
  else if (discrete) camera->rotate (VEC_TILT_UP, M_PI/4);
  else if (ctrl) camera->rotate (VEC_TILT_UP, .005);
  else if (shift) camera->rotate (VEC_TILT_UP, .2);
  else camera->rotate (VEC_TILT_UP, .1);
}

void PrimEdit::handle_key_pgdn (float, bool shift, bool alt, bool ctrl)
{
  if (mode != MODE_NORMAL)
  {
    PEObject* p = list ()->first_selected ();
    Matrix3 rot (0, 0, 0, 0, 0, 0, 0, 0, 0);
    if (alt)
    {
      rot.m33 = 1;
      rot.m11 =  cos ((shift ? .2 : .1));
      rot.m12 = -sin ((shift ? .2 : .1));
      rot.m21 =  sin ((shift ? .2 : .1));
      rot.m22 =  cos ((shift ? .2 : .1));
    }
    else
    {
      rot.m11 = 1;
      rot.m22 =  cos ((shift ? .2 : .1));
      rot.m23 = -sin ((shift ? .2 : .1));
      rot.m32 =  sin ((shift ? .2 : .1));
      rot.m33 =  cos ((shift ? .2 : .1));
    }
    while (p) { p->transform (camera, rot); p = p->sel_next; }
  }
  else if (alt)
  {
    if (ctrl) camera->rotate (VEC_TILT_RIGHT, .005);
    else camera->rotate (VEC_TILT_RIGHT, .1);
  }
  else if (discrete) camera->rotate (VEC_TILT_DOWN, M_PI/4);
  else if (ctrl) camera->rotate (VEC_TILT_DOWN, .005);
  else if (shift) camera->rotate (VEC_TILT_DOWN, .2);
  else camera->rotate (VEC_TILT_DOWN, .1);
}

bool PrimEdit::constrain (PEVertex* v, int x, int y)
{
  PEPlane* p1, * p2;
  PEPolygon* po1, * po2;

  p1 = (PEPlane*)planes->first_selected ();
  if (p1) p2 = (PEPlane*)p1->sel_next;
  else p2 = NULL;

  po1 = (PEPolygon*)polygons->first_selected ();
  if (po1) po2 = (PEPolygon*)po1->sel_next;
  else po2 = NULL;

  if (p1 && !p2) { po2 = po1; }

  if ((p1 || po1) && (p2 || po2))
  {
    if (!p1) p1 = po1->create_plane ();
    else po1 = NULL;
    if (!p2) p2 = po2->create_plane ();
    else po2 = NULL;
    constrain (v, p1, p2, x, y);
    if (po1) CHKB (delete p1);
    if (po2) CHKB (delete p2);
  }
  else if (p1 || po1)
  {
    if (!p1) p1 = po1->create_plane ();
    else po1 = NULL;
    constrain (v, p1, x, y);
    if (po1) CHKB (delete p1);
  }
  else return false;
  return true;
}

void PrimEdit::constrain (PEVertex* v, PEPlane* p, int x, int y)
{
  // Constrain to one plane only.
  Vector3 cam, cam0, cam1, cam2;
  cam.x = x-FRAME_WIDTH/2;				// Inverse perspective transform
  cam.y = -(y-FRAME_HEIGHT/2);
  cam.z = csCamera::aspect;

  cam0 = camera->world2cam(p->v0);
  cam1 = camera->world2cam(p->v1);
  cam2 = camera->world2cam(p->v2);

  Vector3 normal;
  float A, B, C, D;
  Plane::calc_plane (cam0, cam1, cam2, normal, D);
  A = normal.x; B = normal.y; C = normal.z;

  // From the plane equation and the equation of the line:
  // Ax+By+Cz+D = 0
  // x = t*cam.x;
  // y = t*cam.y;
  // z = t*cam.z;
  float t = - D / (A*cam.x + B*cam.y + C*cam.z);

  v->move (camera->cam2world(t*cam));
}

void PrimEdit::constrain (PEVertex* v, PEPlane* p1, PEPlane* p2, int x, int y)
{
  // Constrain to the intersection of two planes.
  Vector3 cam;
  cam.x = x-FRAME_WIDTH/2;				// Inverse perspective transform
  cam.y = -(y-FRAME_HEIGHT/2);
  cam.z = csCamera::aspect;

  Vector3 N1, N2, N3;
  float D1, D2, D3;
  Plane::calc_plane (p1->v0, p1->v1, p1->v2, N1, D1);
  Plane::calc_plane (p2->v0, p2->v1, p2->v2, N2, D2);
  D3 = 0;
  N3 = N1 % N2;		// The direction of the line is perpendicular to the normals of both planes.
  Matrix3 N (N1.x, N1.y, N1.z, N2.x, N2.y, N2.z, N3.x, N3.y, N3.z);
  N.invert ();
  Vector3 V, K;
  K.x = -D1; K.y = -D2; K.z = -D3;
  V = N * K;		// This gives the location of the intersection of the two planes and
  			// a third plane perpendicular to the two planes and going through (0,0,0)

  // So the intersection line in world space is a vector going through V and
  // with direction N3.

  // Convert to camera space.
  Vector3 Vc, Nc;
  Vc = camera->world2cam(V);
  Nc = camera->world2cam(V+N3);
  Nc -= Vc;

  // We now have a vector in camera space starting at (0,0,0) and
  // going through 'cam'. This is the vector representing where we clicked (V1).
  // We also have a vector in camera space starting at Vc with
  // direction Nc. This is the intersection of the two planes (V2).
  // Now we have to find the point on the second vector that is closest
  // to the first vector. Is the intersection of vector V3 (the vector
  // along the closest path between V1 and V2) and V2.
  Vector3 D = cam % Nc;		// Direction of the vector between V1 and V2.

  // Calculate the plane perpendicular to V1 and V3 and through 0.
  Vector3 P1 = cam % D;
  D1 = 0;
  // Calculate the plane perpendicular to V2 and V3 and through Vc.
  Vector3 P2 = Nc % D;
  D2 = - (P2 * Vc);
  // Calculate the plane through Vc with normal D.
  Vector3 P3 = D;
  D3 = - (P3 * Vc);

  // Calculate the intersection of the three previous planes.

  // V1 = cam * t;		Equation of vector 1.
  // V2 = Nc * t + Vc;		Equation of vector 2.
  // V3 = D * t + Vd;		Equation of the closest distance vector between V1 and V2.

  Matrix3 P (P1.x, P1.y, P1.z, P2.x, P2.y, P2.z, P3.x, P3.y, P3.z);
  P.invert ();
  K.x = -D1; K.y = -D2; K.z = -D3;
  V = P * K;

  v->move (camera->cam2world(V));
}

void PrimEdit::eatmouseup (int button, int, int)
{
  if (button == 1)
  {
    if (drag_v) drag_v = NULL;
    if (drag_border)
    {
      drag_border = false;
      if (border1_x > border2_x) { int s = border1_x; border1_x = border2_x; border2_x = s; }
      if (border1_y > border2_y) { int s = border1_y; border1_y = border2_y; border2_y = s; }
      select_vertices_box (border1_x, border1_y, border2_x, border2_y);
    }
  }
}

void PrimEdit::eatmousemove (int x, int y)
{
  if (drag_v) constrain (drag_v, x, y);
  if (drag_border)
  {
    border2_x = x;
    border2_y = y;
  }
}

void PrimEdit::eatmousedown (int button, int x, int y, bool shift, bool alt, bool ctrl)
{
  switch (button)
  {
    case 1: eatmousedown1 (x, y, shift, alt, ctrl); break;
    case 2: eatmousedown2 (x, y, shift, alt, ctrl); break;
    case 3: eatmousedown3 (x, y, shift, alt, ctrl); break;
  }
}

void PrimEdit::eatmousedown3 (int, int, bool, bool, bool)
{
}

void PrimEdit::eatmousedown2 (int x, int y, bool, bool, bool)
{
  CHK (PEVertex* v = new PEVertex ());
  if (constrain (v, x, y))
  {
    vertices->add_new_object (v);
    CsPrintf (MSG_STDOUT, "Created constrained vertex at (%f,%f,%f).\n", v->v.x, v->v.y, v->v.z);
  }
  else
  {
    CHK (delete v);
    CsPrintf (MSG_STDOUT, "Could not create constrained vertex!\n");
  }
}

void PrimEdit::eatmousedown1 (int x, int y, bool shift, bool, bool)
{
  PEVertex* v = NULL;
  PEPlane* p = NULL;
  PEPolygon* po = NULL;

  if (!shift)
  {
    polygons->clear_selected ();
    planes->clear_selected ();
    vertices->clear_selected ();
  }

  if (vmode == MODE_NORMAL || vmode == MODE_EDIT_VERTEX) v = search_vertex (x, y);
  if (!v)
  {
    if (vmode == MODE_NORMAL || vmode == MODE_EDIT_PLANE) p = search_plane (x, y);
    if (!p)
      if (vmode == MODE_NORMAL || vmode == MODE_EDIT_POLYGON) po = search_polygon (x, y);
  }

  if (p)
  {
    if (planes->is_selected (p)) planes->deselect (p);
    else planes->select (p);
  }

  if (v)
  {
    if (vertices->is_selected (v)) vertices->deselect (v);
    else vertices->select (v);
  }

  if (po)
  {
    if (polygons->is_selected (po)) polygons->deselect (po);
    else polygons->select (po);
  }

  if (v) drag_v = v;
  else drag_v = NULL;

  if (!drag_v)
  {
    drag_border = true;
    border1_x = x;
    border1_y = y;
    border2_x = x;
    border2_y = y;
  }
}

void PrimEdit::eatkeypress (int key, bool shift, bool alt, bool ctrl)
{
  float speed = PSPEED_NORMAL;
  if (ctrl) speed = PSPEED_SLOW;
  if (shift) speed = PSPEED_FAST;

  switch (key)
  {
    case 'X':
      {
        CHK (PEPlane* p = new PEPlane ());
	planes->add_new_object (p);
	p->v0.x = p->v0.y = p->v0.z = 0;
	p->v1.x = p->v1.y = 0; p->v1.z = 1;
	p->v2.x = p->v2.z = 0; p->v2.y = 1;
        p->update_plane ();
        CsPrintf (MSG_STDOUT, "Created new X plane.\n");
      }
      break;
    case 'Y':
      {
        CHK (PEPlane* p = new PEPlane ());
	planes->add_new_object (p);
	p->v0.x = p->v0.y = p->v0.z = 0;
	p->v1.x = p->v1.y = 0; p->v1.z = 1;
	p->v2.y = p->v2.z = 0; p->v2.x = 1;
        p->update_plane ();
        CsPrintf (MSG_STDOUT, "Created new Y plane.\n");
      }
      break;
    case 'Z':
      {
        CHK (PEPlane* p = new PEPlane ());
	planes->add_new_object (p);
	p->v0.x = p->v0.y = p->v0.z = 0;
	p->v1.x = p->v1.z = 0; p->v1.y = 1;
	p->v2.y = p->v2.z = 0; p->v2.x = 1;
        p->update_plane ();
        CsPrintf (MSG_STDOUT, "Created new Z plane.\n");
      }
      break;
    case 'V':
      {
        CHK (PEVertex* v = new PEVertex ());
	vertices->add_new_object (v);
	v->v.x = 0; v->v.y = 0; v->v.z = 0;
        CsPrintf (MSG_STDOUT, "Created new vertex (%f,%f,%f).\n", v->v.x, v->v.y, v->v.z);
      }
      break;
    case 'g':
      grid = !grid;
      CsPrintf (MSG_STDOUT, "Set grid %s.\n", grid ? "on" : "off");
      break;
    case 'b':
      if (mode == MODE_NORMAL) camera->rotate (VEC_ROT_RIGHT, M_PI);
      break;
    case 'N':
      if (mode == MODE_NORMAL) 
       camera->set_world2cam(Matrix3() /*identity*/);
      break;
    case 'P':
      new_polygon ();
      break;
    case 'D':
      if (mode != MODE_NORMAL) list ()->delete_selected ();
      else CsPrintf (MSG_STDOUT, "Can't delete in this mode!\n");
      break;
    case 'a':
      align_plane ();
      break;
    case 'p':
      if (ctrl)
      {
        select_polygons ();
      }
      else if (alt)
      {
        vmode = MODE_EDIT_PLANE;
        CsPrintf (MSG_STDOUT, "Show only planes.\n");
      }
      else
      {
        mode = MODE_EDIT_PLANE;
        CsPrintf (MSG_STDOUT, "Plane editing mode.\n");
      }
      break;
    case 'o':
      if (alt)
      {
        vmode = MODE_EDIT_POLYGON;
        CsPrintf (MSG_STDOUT, "Show only polygons.\n");
      }
      else
      {
        mode = MODE_EDIT_POLYGON;
        CsPrintf (MSG_STDOUT, "Polygon editing mode.\n");
      }
      break;
    case 'n':
      if (ctrl)
      {
        discrete = !discrete;
        CsPrintf (MSG_STDOUT, "Movements in multiples of 45 degrees %s.\n", discrete ? "on" : "off");
      }
      else if (alt)
      {
	vmode = MODE_NORMAL;
        CsPrintf (MSG_STDOUT, "Show everything.\n");
      }
      else
      {
        mode = MODE_NORMAL;
        CsPrintf (MSG_STDOUT, "Normal mode.\n");
      }
      break;
    case 'C':
      clone_vertices ();
      break;
    case 'v':
      if (ctrl)
      {
        select_vertices ();
      }
      else if (alt)
      {
        vmode = MODE_EDIT_VERTEX;
        CsPrintf (MSG_STDOUT, "Show only vertices.\n");
      }
      else
      {
        mode = MODE_EDIT_VERTEX;
        CsPrintf (MSG_STDOUT, "Vertex editing mode.\n");
      }
      break;
    case 'q':
      Shutdown = true;
      break;
    case 's':
      if (ctrl) save_world ();
      break;
    case 'S':
      load_world ();
      break;
  }
}

void PrimEdit::clone_vertices ()
{
  // Clone all selected vertices. Unselect the old vertices
  // and select the new ones.
  int cnt = vertices->count_selected ();
  if (cnt < 1)
  {
    CsPrintf (MSG_STDOUT, "Need at least one selected vertex!\n");
    return;
  }
  PEVertex* v;

  // First clear all 'num' to 0.
  v = (PEVertex*)vertices->first ();
  while (v) { v->num = 0; v = (PEVertex*)v->next; }

  // Clone the selected vertices.
  v = (PEVertex*)vertices->first_selected ();
  while (v)
  {
    CHK (PEVertex* vn = new PEVertex ());
    vertices->add_new_object (vn);
    vn->move (v->v);
    vn->num = 1;
    v = (PEVertex*)v->sel_next;
  }

  // Unselect the old vertices and select the new ones.
  v = (PEVertex*)vertices->first ();
  while (v) { if (v->num) vertices->select (v); else vertices->deselect (v); v = (PEVertex*)v->next; }
  CsPrintf (MSG_STDOUT, "Cloned %d vertices.\n", cnt);
}

void PrimEdit::select_polygons ()
{
  // Select all polygons that connect to all the
  // selected vertices.
  int cnt = vertices->count_selected ();
  if (cnt < 1)
  {
    CsPrintf (MSG_STDOUT, "Need at least one selected vertex!\n");
    return;
  }
  PEPolygon* po = (PEPolygon*)polygons->first ();
  while (po)
  {
    int i, sel = 0;
    for (i = 0 ; i < po->num_vertices ; i++)
      if (vertices->is_selected (po->vertices[i])) sel++;
    if (sel == cnt) polygons->select (po);
    po = (PEPolygon*)po->next;
  }
}

void PrimEdit::select_vertices ()
{
  // Select all vertices for the selected polygons.
  int cnt = polygons->count_selected ();
  if (cnt < 1)
  {
    CsPrintf (MSG_STDOUT, "Need at least one selected polygon!\n");
    return;
  }
  PEPolygon* po = (PEPolygon*)polygons->first_selected ();
  while (po)
  {
    int i;
    for (i = 0 ; i < po->num_vertices ; i++)
      vertices->select (po->vertices[i]);
    po = (PEPolygon*)po->sel_next;
  }
}

void PrimEdit::new_polygon ()
{
  int cnt = vertices->count_selected ();
  if (cnt < 3)
  {
    CsPrintf (MSG_STDOUT, "Need minimum three vertices to create a polygon!\n");
    return;
  }
  PEPolygon* po;
  CHK (po = new PEPolygon ());
  polygons->add_new_object (po);

  po->set_num_vertices (cnt);
  int i;
  PEVertex* v = (PEVertex*)vertices->first_selected ();
  for (i = 0 ; i < cnt ; i++)
  {
    po->set_vertex (cnt-i-1, v);
    v = (PEVertex*)v->sel_next;
  }
  po->update ();

  CsPrintf (MSG_STDOUT, "Created new polygon.\n");
}

void PrimEdit::align_plane ()
{
  if (planes->count_selected () > 1)
  {
    CsPrintf (MSG_STDOUT, "Need zero or one selected plane!\n");
    return;
  }
  int cnt_v = vertices->count_selected ();
  int cnt_p = polygons->count_selected ();
  if ((cnt_v < 1 || cnt_v > 3) && cnt_p != 1)
  {
    CsPrintf (MSG_STDOUT, "Need one to three selected vertices or else one polygon.\n");
    return;
  }

  PEVertex* v1 = (PEVertex*)vertices->first_selected ();
  PEVertex* v2 = v1 ? (PEVertex*)v1->sel_next : (PEVertex*)NULL;
  PEVertex* v3 = v2 ? (PEVertex*)v2->sel_next : (PEVertex*)NULL;
  PEPlane* p = (PEPlane*)planes->first_selected ();
  PEPolygon* po = (PEPolygon*)polygons->first_selected ();

  if (v3)
  {
    // We have three selected vertices.
    Vector3 normal;
    normal = Plane::calc_plane_normal (v1->v, v2->v, v3->v);
    if (!p)
    {
      CHK (p = new PEPlane ());
      planes->add_new_object (p);
    }
    p->v0 = v1->v;
    p->v1 = Vector3::unit (v2->v - v1->v) + v1->v;
    Vector3 unormal = Vector3::unit (normal);
    p->v2 = (unormal % (p->v1-v1->v)) + v1->v;
  }
  else if (v2)
  {
    // We have two selected vertices. Align the plane
    // straight up with the two vertices as one base.
    //@@@
    CsPrintf (MSG_STDOUT, "Not implemented yet!\n");
  }
  else if (v1)
  {
    // We have one selected vertex.
    if (!p)
    {
      CsPrintf (MSG_STDOUT, "We need a plane if only one vertex is selected!\n");
      return;
    }
    p->v1 = p->v1 - p->v0 + v1->v;
    p->v2 = p->v2 - p->v0 + v1->v;
    p->v0 = v1->v;
  }
  else if (po)
  {
    // We have a selected polygon. Use that to create or align the plane.
    if (p) po->align_plane (p);
    else
    {
      p = po->create_plane ();
      planes->add_new_object (p);
    }
  }
  if (p) p->update_plane ();
}

void PrimEdit::NextFrame (long elapsed_time, long current_time)
{
  SysSystemDriver::NextFrame (elapsed_time, current_time);

  bool alt, shift, ctrl;
  float speed = PSPEED_NORMAL;

  alt = Keyboard->Key.alt;
  ctrl = Keyboard->Key.ctrl;
  shift = Keyboard->Key.shift;
  if (ctrl) speed = PSPEED_SLOW;
  if (shift) speed = PSPEED_FAST;

  if (Keyboard->Key.up)    handle_key_forward (speed, shift, alt, ctrl);
  if (Keyboard->Key.down)  handle_key_backwards (speed, shift, alt, ctrl);
  if (Keyboard->Key.left)  handle_key_left (speed, shift, alt, ctrl);
  if (Keyboard->Key.right) handle_key_right (speed, shift, alt, ctrl);
  if (Keyboard->Key.pgup)  handle_key_pgup (speed, shift, alt, ctrl);
  if (Keyboard->Key.pgdn)  handle_key_pgdn (speed, shift, alt, ctrl);

  csEvent *Event;
  while ((Event = EventQueue->Get ()))
  {
    switch (Event->Type)
    {
      case csevKeyDown:
        eatkeypress (Event->Key.Code, Event->Key.ShiftKeys & CSMASK_SHIFT,
          Event->Key.ShiftKeys & CSMASK_ALT, Event->Key.ShiftKeys & CSMASK_CTRL);
        break;
      case csevMouseDown:
        eatmousedown (Event->Mouse.Button, Event->Mouse.x, Event->Mouse.y,
        	Event->Mouse.ShiftKeys & CSMASK_SHIFT, Event->Mouse.ShiftKeys & CSMASK_ALT, Event->Mouse.ShiftKeys & CSMASK_CTRL);
        break;
      case csevMouseMove:
        eatmousemove (Event->Mouse.x, Event->Mouse.y);
        break;
      case csevMouseUp:
        eatmouseup (Event->Mouse.Button, Event->Mouse.x, Event->Mouse.y);
        break;
    }
    CHK (delete Event);
  }

  DrawFrame ();
}

void PrimEdit::DrawFrame ()
{
  if (!Gfx3D->BeginDraw (CSDRAW_2DGRAPHICS))
    return;

  Gfx2D->Clear (0);
  int v_color, l_color;
  bool cross;

  Vector3 v; v.x = v.y = v.z = 0;
  draw_vertex (v, txt->blue (), true);

  if (vmode == MODE_NORMAL || vmode == MODE_EDIT_PLANE)
  {
    PEPlane* p = (PEPlane*)planes->first ();
    int cnt = 0;
    while (p)
    {
      if (planes->is_selected (p)) { v_color = txt->red (); l_color = txt->red (); cnt++; }
      else { v_color = txt->yellow (); l_color = txt->yellow (); }

      draw_line (p->help1_v1, p->help1_v2, l_color);
      draw_line (p->help2_v1, p->help2_v2, l_color);
      draw_line (p->help3_v1, p->help3_v2, l_color);

      draw_line (p->v0, p->v1, l_color);
      draw_line (p->v0, p->v2, l_color);
      draw_vertex (p->v0, v_color);
      draw_vertex (p->v1, v_color);
      draw_vertex (p->v2, v_color);
      p = (PEPlane*)p->next;
    }

    if (cnt == 1 && grid)
    {
      // Draw a grid raster
      p = (PEPlane*)planes->first_selected ();
      int i, j;
      for (i = -10 ; i < 10 ; i++)
        for (j = -10 ; j < 10 ; j++)
        {
	  v = p->v0 + ((float)i)*(p->v1-p->v0) + ((float)j)*(p->v2-p->v0);
	  draw_vertex (v, txt->white ());
        }
    }
    else if (cnt == 2 && grid)
    {
      // Draw a grid line on the intersection
      p = (PEPlane*)planes->first_selected ();
      PEPlane* p2 = (PEPlane*)p->sel_next;
      Vector3 V, D;
      PEPlane::intersect (p, p2, V, D);
      int i;
      for (i = -10 ; i < 10 ; i++)
      {
        v = V + ((float)i)*D;
        draw_vertex (v, txt->white ());
      }
    }
  }

  if (vmode == MODE_NORMAL || vmode == MODE_EDIT_VERTEX)
  {
    PEVertex* ve = (PEVertex*)vertices->first ();
    while (ve)
    {
      if (vertices->is_selected (ve)) { v_color = txt->red (); cross = vertices->first_selected () == ve; }
      else { v_color = txt->green (); cross = false; }
      draw_vertex (ve->v, v_color, cross);
      ve = (PEVertex*)ve->next;
    }
  }

  if (vmode == MODE_NORMAL || vmode == MODE_EDIT_POLYGON)
  {
    PEPolygon* po = (PEPolygon*)polygons->first ();
    while (po)
    {
      if (polygons->is_selected (po)) { l_color = txt->red (); v_color = txt->find_color (128, 0, 0); }
      else { l_color = txt->white (); v_color = txt->find_color (128, 128, 128); }
      int i;
      bool vis = po->visible (camera);
      for (i = 0 ; i < po->num_vertices ; i++)
      {
        draw_line (po->vertices[i]->v, po->vertices[(i+1)%po->num_vertices]->v, l_color);
        if (vis) draw_line (po->vertices[i]->v, po->center, v_color);
      }
      po = (PEPolygon*)po->next;
    }
  }

  if (drag_border)
  {
    Gfx2D->DrawLine (border1_x, border1_y, border1_x, border2_y, txt->white ());
    Gfx2D->DrawLine (border2_x, border1_y, border2_x, border2_y, txt->white ());
    Gfx2D->DrawLine (border1_x, border1_y, border2_x, border1_y, txt->white ());
    Gfx2D->DrawLine (border1_x, border2_y, border2_x, border2_y, txt->white ());
  }

  char buffer[40];
  char* buf, * vbuf;
  switch (mode)
  {
    case MODE_NORMAL: buf = "NORMAL"; break;
    case MODE_EDIT_PLANE: buf = "PLANE"; break;
    case MODE_EDIT_VERTEX: buf = "VERTEX"; break;
    case MODE_EDIT_POLYGON: buf = "POLYGON"; break;
    default: buf = "(?)"; break;
  }
  switch (vmode)
  {
    case MODE_NORMAL: vbuf = "all"; break;
    case MODE_EDIT_PLANE: vbuf = "planes"; break;
    case MODE_EDIT_VERTEX: vbuf = "verts"; break;
    case MODE_EDIT_POLYGON: vbuf = "polys"; break;
    default: vbuf = "(?)"; break;
  }
  sprintf (buffer, "(%2.2f,%2.2f,%2.2f) %s(%s) [%s][%s]",
        camera->v_world2cam().x, camera->v_world2cam().y, camera->v_world2cam().z, buf, vbuf,
	grid ? "grid" : "nogrid", discrete ? "45o" : "free");
  Gfx2D->Write (10, FRAME_HEIGHT-10, txt->white (), -1, buffer);

  Gfx3D->FinishDraw ();
  // Blit the image to screen
  Gfx3D->Print (NULL);
}

bool PrimEdit::perspective (Vector3& v, Vector2& persp, float radius, float& pradius)
{
  Vector3 cam;
  cam = camera->world2cam(v);
  if (cam.z > SMALL_Z)
  {
    float iz = csCamera::aspect/cam.z;
    persp.x = cam.x * iz + (FRAME_WIDTH/2);
    persp.y = FRAME_HEIGHT - 1 - (cam.y * iz + (FRAME_HEIGHT/2));
    pradius = radius * iz;
    return true;
  }
  else return false;
}

bool PrimEdit::hit_vertex (Vector3& v, float radius, int x, int y)
{
  Vector2 persp;
  float rad;
  int r, px, py;
  if (perspective (v, persp, radius, rad))
  {
    r = QInt (rad);
    if (r < 2) r = 2;
    px = QInt (persp.x);
    py = QInt (persp.y);
    if (x >= px-r && x <= px+r && y >= py-r && y <= py+r)
      return true;
  }
  return false;
}

bool PrimEdit::hit_vertex (Vector3& v, float radius, int x1, int y1, int x2, int y2)
{
  Vector2 persp;
  float rad;
  int r, px, py;
  if (perspective (v, persp, radius, rad))
  {
    r = QInt (rad);
    if (r < 2) r = 2;
    px = QInt (persp.x);
    py = QInt (persp.y);
    if (x1 <= px+r && x2 >= px-r && y1 <= py+r && y2 >= py-r)
      return true;
  }
  return false;
}

bool PrimEdit::hit_line (Vector3& v1, Vector3& v2, int x, int y)
{
  Vector3 cam1, cam2;
  cam1 = camera->world2cam(v1);
  cam2 = camera->world2cam(v2);
  if (cam1.z < SMALL_Z && cam2.z < SMALL_Z) return false;

  float x1 = cam1.x, y1 = cam1.y, z1 = cam1.z;
  float x2 = cam2.x, y2 = cam2.y, z2 = cam2.z;
  float t;

  if (z1 < SMALL_Z)
  {
    // x = t*(x2-x1)+x1;
    // y = t*(y2-y1)+y1;
    // z = t*(z2-z1)+z1;
    t = (SMALL_Z-z1) / (z2-z1);
    x1 = t*(x2-x1)+x1;
    y1 = t*(y2-y1)+y1;
    z1 = SMALL_Z;
  }
  else if (z2 < SMALL_Z)
  {
    // x = t*(x2-x1)+x1;
    // y = t*(y2-y1)+y1;
    // z = t*(z2-z1)+z1;
    t = (SMALL_Z-z1) / (z2-z1);
    x2 = t*(x2-x1)+x1;
    y2 = t*(y2-y1)+y1;
    z2 = SMALL_Z;
  }

  float iz1 = csCamera::aspect/z1;
  float px1 = x1 * iz1 + (FRAME_WIDTH/2);
  float py1 = FRAME_HEIGHT - 1 - (y1 * iz1 + (FRAME_HEIGHT/2));
  float iz2 = csCamera::aspect/z2;
  float px2 = x2 * iz2 + (FRAME_WIDTH/2);
  float py2 = FRAME_HEIGHT - 1 - (y2 * iz2 + (FRAME_HEIGHT/2));
  float px = (float)x;
  float py = (float)y;

  if (ABS (px2 - px1) > ABS (py2 - py1))
  {
    t = (px - px1) / (px2 - px1);
    if (t < 0 || t > 1) return false;
    return (ABS (t * (py2-py1) + py1 - py) < 3);
  }
  else
  {
    t = (py - py1) / (py2 - py1);
    if (t < 0 || t > 1) return false;
    return (ABS (t * (px2-px1) + px1 - px) < 3);
  }
}

void PrimEdit::draw_vertex (Vector3& v, int col, bool cross)
{
  Vector2 persp;
  float rad;
  int r, px, py;

  if (perspective (v, persp, PLANE_VERTEX_RADIUS, rad))
  {
    r = QInt (rad);
    px = QInt (persp.x);
    py = QInt (persp.y);
    Gfx2D->DrawLine (px-r, py-r, px+r, py+r, col);
    Gfx2D->DrawLine (px+r, py-r, px-r, py+r, col);
    if (cross)
    {
      Gfx2D->DrawLine (px-r, py, px+r, py, col);
      Gfx2D->DrawLine (px, py-r, px, py+r, col);
    }
  }
}

void PrimEdit::draw_line (Vector3& v1, Vector3& v2, int col)
{
  Vector3 cam1, cam2;
  cam1 = camera->world2cam(v1);
  cam2 = camera->world2cam(v2);
  Gfx3D->DrawLine (cam1, cam2, col);
}

PEPlane* PrimEdit::search_plane (int x, int y)
{
  PEPlane* p = (PEPlane*)planes->first ();
  while (p)
  {
    if (hit_vertex (p->v0, PLANE_VERTEX_RADIUS, x, y)) return p;
    if (hit_vertex (p->v1, PLANE_VERTEX_RADIUS, x, y)) return p;
    if (hit_vertex (p->v2, PLANE_VERTEX_RADIUS, x, y)) return p;
    if (hit_line (p->v0, p->v1, x, y)) return p;
    if (hit_line (p->v0, p->v2, x, y)) return p;
    if (hit_line (p->help1_v1, p->help1_v2, x, y)) return p;
    if (hit_line (p->help2_v1, p->help2_v2, x, y)) return p;
    if (hit_line (p->help3_v1, p->help3_v2, x, y)) return p;
    p = (PEPlane*)p->next;
  }
  return NULL;
}

void PrimEdit::select_vertices_box (int x1, int y1, int x2, int y2)
{
  PEVertex* v = (PEVertex*)vertices->first ();
  while (v)
  {
    if (hit_vertex (v->v, PLANE_VERTEX_RADIUS, x1, y1, x2, y2))
      vertices->select (v);
    v = (PEVertex*)v->next;
  }
}

PEVertex* PrimEdit::search_vertex (int x, int y)
{
  PEVertex* v = (PEVertex*)vertices->first ();
  while (v)
  {
    if (hit_vertex (v->v, PLANE_VERTEX_RADIUS, x, y)) return v;
    v = (PEVertex*)v->next;
  }
  return NULL;
}


PEPolygon* PrimEdit::search_polygon (int x, int y)
{
  PEPolygon* v = (PEPolygon*)polygons->first ();
  while (v)
  {
    int i;
    bool vis = v->visible (camera);
    for (i = 0 ; i <v->num_vertices ; i++)
    {
      if (hit_line (v->vertices[i]->v, v->vertices[(i+1)%v->num_vertices]->v, x, y)) return v;
      if (vis) if (hit_line (v->vertices[i]->v, v->center, x, y)) return v;
    }
    v = (PEPolygon*)v->next;
  }
  return NULL;
}

//=================================================================================================

int main (int argc, char *argv[])
{
  config = new csIniFile ("primedit.cfg");

  // Application object
  PrimEdit app;

  // print some fancy info
  CsPrintf (MSG_INITIALIZATION, "Crystal Space Primitive Editor.\n");
  CsPrintf (MSG_INITIALIZATION, "Created by Jorrit Tyberghein.\n\n");

  // Initialize application
  app.Initialize (argc, argv);

  // Open the graphics context
  app.Open ("Crystal Space Primitive Editor");

  // Main loop
  app.Loop();
}

void debug_dump () { }
void cleanup () { }

