/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "system/sysprint.h"
#include "general/walktest/main.h"
#include "general/walktest/scon.h"
#include "general/walktest/bot.h"
#include "engine/camera.h"
#include "general/command.h"
#include "render/igraph3d.h"
#include "engine/world.h"
#include "general/csview.h"
#include "engine/wirefrm.h"
#include "engine/objects/cssprite.h"
#include "engine/light/light.h"
#include "engine/sector.h"
#include "sound/cssnddrv.h"
#include "util/util.h"

#ifdef COMP_VC
	// disable annoying warning of unsafe mix of bool and int.
#	pragma warning(disable: 4805)
#endif

csKeyMap* mapping = NULL;

Bot* bot = NULL;
extern Wave* wMissile_boom;
extern Wave* wMissile_whoosh;

#define DYN_TYPE_MISSILE 1
#define DYN_TYPE_RANDOM 2
#define DYN_TYPE_EXPLOSION 3

struct LightStruct
{
  int type;
};

struct MissileStruct
{
  int type;		// type == DYN_TYPE_MISSILE
  OrthoTransform dir;
  csSprite3D* sprite;
};

struct ExplosionStruct
{
  int type;		// type == DYN_TYPE_EXPLOSION
  float radius;
  int dir;
};

struct RandomLight
{
  int type;		// type == DYN_TYPE_RANDOM
  float dyn_move_dir;
  float dyn_move;
  float dyn_r1, dyn_g1, dyn_b1;
};

void RandomColor (float& r, float& g, float& b)
{
  switch ((rand ()>>3) % 3)
  {
    case 0:
      r = (float)(900+(rand () % 100))/1000.;
      g = (float)(rand () % 1000)/1000.;
      b = (float)(rand () % 1000)/1000.;
      break;
    case 1:
      r = (float)(rand () % 1000)/1000.;
      g = (float)(900+(rand () % 100))/1000.;
      b = (float)(rand () % 1000)/1000.;
      break;
    case 2:
      r = (float)(rand () % 1000)/1000.;
      g = (float)(rand () % 1000)/1000.;
      b = (float)(900+(rand () % 100))/1000.;
      break;
  }
}

void HandleDynLight (csDynLight* dyn)
{
  LightStruct* ls = (LightStruct*)(dyn->userdata);
  switch (ls->type)
  {
    case DYN_TYPE_MISSILE:
    {
      MissileStruct* ms = (MissileStruct*)(dyn->userdata);
      Vector3 v (0, 0, 2.5);
      Vector3 old = dyn->get_center ();
      v = old + ms->dir.m_this2world () * v;
      csSector* s = dyn->get_sector ();
      bool mirror = false;
      Vector3 old_v = v;
      s = s->follow_segment (ms->dir, v, mirror);
      if (ABS (v.x-old_v.x) > SMALL_EPSILON || ABS (v.y-old_v.y) > SMALL_EPSILON && ABS (v.z-old_v.z) > SMALL_EPSILON)
      {
        v = old;
        if (ms->sprite)
	{
	  cs->world->remove (&ms->sprite->get_mso ());
	  CHK (delete ms->sprite);
	}
        CHK (delete ms);
	CHK (ExplosionStruct* es = new ExplosionStruct);
	Sys->Sound->playEphemeral (wMissile_boom);
	es->type = DYN_TYPE_EXPLOSION;
	es->radius = 2;
	es->dir = 1;
	dyn->userdata = (void*)es;
        return;
      }
      dyn->move (s, v.x, v.y, v.z);
      dyn->setup ();
      if (ms->sprite)
      {
        ms->sprite->set_move (v);
	ms->sprite->get_mso ().init_one_sector (s);
      }
      break;
    }
    case DYN_TYPE_EXPLOSION:
    {
      ExplosionStruct* es = (ExplosionStruct*)(dyn->userdata);
      if (es->dir == 1)
      {
        es->radius += 3;
	if (es->radius > 6) es->dir = -1;
      }
      else
      {
        es->radius -= 2;
	if (es->radius < 1)
	{
	  CHK (delete es);
          cs->world->remove_dyn_light (dyn);
          CHK (delete dyn);
	  return;
	}
      }
      dyn->resize (es->radius);
      dyn->setup ();
      break;
    }
    case DYN_TYPE_RANDOM:
    {
      RandomLight* rl = (RandomLight*)(dyn->userdata);
      rl->dyn_move += rl->dyn_move_dir;
      if (rl->dyn_move < 0 || rl->dyn_move > 2) rl->dyn_move_dir = -rl->dyn_move_dir;
      if (ABS (rl->dyn_r1-dyn->get_red ()) < .01 && ABS (rl->dyn_g1-dyn->get_green ()) < .01 && ABS (rl->dyn_b1-dyn->get_blue ()) < .01)
        RandomColor (rl->dyn_r1, rl->dyn_g1, rl->dyn_b1);
      else dyn->set_color (csColor ((rl->dyn_r1+7.*dyn->get_red ())/8., (rl->dyn_g1+7.*dyn->get_green ())/8., (rl->dyn_b1+7.*dyn->get_blue ())/8.));
      dyn->move (dyn->get_sector (), dyn->get_center ().x, dyn->get_center ().y+rl->dyn_move_dir, dyn->get_center ().z);
      dyn->setup ();
      break;
    }
  }
}

void map_key (char* keyname, csKeyMap* map)
{
  map->shift = 0;
  map->alt = 0;
  map->ctrl = 0;
  char* dash = strchr (keyname, '-');
  while (dash)
  {
    *dash = 0;
    if (!strcmp (keyname, "shift")) map->shift = 1;
    else if (!strcmp (keyname, "alt")) map->alt = 1;
    else if (!strcmp (keyname, "ctrl")) map->ctrl = 1;
    else CsPrintf (MSG_CONSOLE, "Bad modifier '%s'!\n", keyname);

    *dash = '-';
    keyname = dash+1;
    dash = strchr (dash+1, '-');
  }

  if (!strcmp (keyname, "tab")) map->key = CSKEY_TAB;
  else if (!strcmp (keyname, "space")) map->key = ' ';
  else if (!strcmp (keyname, "esc")) map->key = CSKEY_ESC;
  else if (!strcmp (keyname, "enter")) map->key = CSKEY_ENTER;
  else if (!strcmp (keyname, "bs")) map->key = CSKEY_BACKSPACE;
  else if (!strcmp (keyname, "up")) map->key = CSKEY_UP;
  else if (!strcmp (keyname, "down")) map->key = CSKEY_DOWN;
  else if (!strcmp (keyname, "right")) map->key = CSKEY_RIGHT;
  else if (!strcmp (keyname, "left")) map->key = CSKEY_LEFT;
  else if (!strcmp (keyname, "pgup")) map->key = CSKEY_PGUP;
  else if (!strcmp (keyname, "pgdn")) map->key = CSKEY_PGDN;
  else if (!strcmp (keyname, "home")) map->key = CSKEY_HOME;
  else if (!strcmp (keyname, "end")) map->key = CSKEY_END;
  else if (!strcmp (keyname, "ins")) map->key = CSKEY_INS;
  else if (!strcmp (keyname, "del")) map->key = CSKEY_DEL;
  else if (!strcmp (keyname, "f1")) map->key = CSKEY_F1;
  else if (!strcmp (keyname, "f2")) map->key = CSKEY_F2;
  else if (!strcmp (keyname, "f3")) map->key = CSKEY_F3;
  else if (!strcmp (keyname, "f4")) map->key = CSKEY_F4;
  else if (!strcmp (keyname, "f5")) map->key = CSKEY_F5;
  else if (!strcmp (keyname, "f6")) map->key = CSKEY_F6;
  else if (!strcmp (keyname, "f7")) map->key = CSKEY_F7;
  else if (!strcmp (keyname, "f8")) map->key = CSKEY_F8;
  else if (!strcmp (keyname, "f9")) map->key = CSKEY_F9;
  else if (!strcmp (keyname, "f10")) map->key = CSKEY_F10;
  else if (!strcmp (keyname, "f11")) map->key = CSKEY_F11;
  else if (!strcmp (keyname, "f12")) map->key = CSKEY_F12;
  else if (*(keyname+1) != 0) CsPrintf (MSG_CONSOLE, "Bad key '%s'!\n", keyname);
  else if ((*keyname >= 'A' && *keyname <= 'Z') || strchr ("!@#$%^&*()_+", *keyname))
  {
    map->shift = 1;
    map->key = *keyname;
  }
  else if (*keyname >= 'a' && *keyname <= 'z')
  {
    if (map->shift) map->key = (*keyname)+'A'-'a';
    else map->key = *keyname;
  }
  else map->key = *keyname;
}

char* keyname (csKeyMap* map)
{
  static char buf[100];
  buf[0] = 0;
  if (map->shift) strcat (buf, "shift-");
  if (map->ctrl) strcat (buf, "ctrl-");
  if (map->alt) strcat (buf, "alt-");
  switch (map->key)
  {
    case CSKEY_TAB: strcat (buf, "tab"); break;
    case ' ': strcat (buf, "space"); break;
    case CSKEY_ESC: strcat (buf, "esc"); break;
    case CSKEY_ENTER: strcat (buf, "enter"); break;
    case CSKEY_BACKSPACE: strcat (buf, "bs"); break;
    case CSKEY_UP: strcat (buf, "up"); break;
    case CSKEY_DOWN: strcat (buf, "down"); break;
    case CSKEY_RIGHT: strcat (buf, "right"); break;
    case CSKEY_LEFT: strcat (buf, "left"); break;
    case CSKEY_PGUP: strcat (buf, "pgup"); break;
    case CSKEY_PGDN: strcat (buf, "pgdn"); break;
    case CSKEY_HOME: strcat (buf, "home"); break;
    case CSKEY_END: strcat (buf, "end"); break;
    case CSKEY_INS: strcat (buf, "ins"); break;
    case CSKEY_DEL: strcat (buf, "del"); break;
    case CSKEY_F1: strcat (buf, "f1"); break;
    case CSKEY_F2: strcat (buf, "f2"); break;
    case CSKEY_F3: strcat (buf, "f3"); break;
    case CSKEY_F4: strcat (buf, "f4"); break;
    case CSKEY_F5: strcat (buf, "f5"); break;
    case CSKEY_F6: strcat (buf, "f6"); break;
    case CSKEY_F7: strcat (buf, "f7"); break;
    case CSKEY_F8: strcat (buf, "f8"); break;
    case CSKEY_F9: strcat (buf, "f9"); break;
    case CSKEY_F10: strcat (buf, "f10"); break;
    case CSKEY_F11: strcat (buf, "f11"); break;
    case CSKEY_F12: strcat (buf, "f12"); break;
    default:
    {
      char* s = strchr (buf, 0);
      *s++ = map->key;
      *s = 0;
    }
  }
  return buf;
}

csKeyMap* find_mapping (char* keyname)
{
  csKeyMap map;
  map_key (keyname, &map);

  csKeyMap* m = mapping;
  while (m)
  {
    if (map.key == m->key && map.shift == m->shift && map.ctrl == m->ctrl && map.alt == m->alt)
      return m;
    m = m->next;
  }
  return NULL;
}

void bind_key (char* arg)
{
  if (!arg)
  {
    csKeyMap* map = mapping;
    while (map)
    {
      CsPrintf (MSG_CONSOLE, "Key '%s' bound to '%s'.\n", keyname (map), map->cmd);
      map = map->next;
    }
    return;
  }
  char* space = strchr (arg, ' ');
  if (space)
  {
    *space = 0;
    csKeyMap* map = find_mapping (arg);
    if (map)
    {
      CHK (delete [] map->cmd);
    }
    else
    {
      CHK (map = new csKeyMap ());
      map->next = mapping;
      map->prev = NULL;
      if (mapping) mapping->prev = map;
      mapping = map;
      map_key (arg, map);
    }
    CHK (map->cmd = new char [strlen (space+1)+1]);
    strcpy (map->cmd, space+1);
    *space = ' ';
  }
  else
  {
    csKeyMap* map = find_mapping (arg);
    if (map) CsPrintf (MSG_CONSOLE, "Key bound to '%s'!\n", map->cmd);
    else CsPrintf (MSG_CONSOLE, "Key not bound!\n");
  }
}

//--//--//--//--//--//--//--//--//--//--//-- Handle our additional commands --//
static bool ExtraHandler (char *cmd, char *arg)
{
  if (!strcasecmp (cmd, "help"))
  {
    Command::perform (cmd, arg);
    CsPrintf (MSG_CONSOLE, "-*- Additional commands -*-\n");
    CsPrintf (MSG_CONSOLE, " bind, fclear, addlight, dellight, dellights\n");
    CsPrintf (MSG_CONSOLE, " picklight, droplight, physics, colldet, stats, hi\n");
    CsPrintf (MSG_CONSOLE, " fps, perftest, capture, coordshow, zbuf\n");
    CsPrintf (MSG_CONSOLE, " map, fire, debug0, debug1, debug2, p_alpha\n");
    CsPrintf (MSG_CONSOLE, " addbot, snd_play, snd_volume, s_fog\n");
  }
  else if (!strcasecmp (cmd, "bind"))
    bind_key (arg);
  else if (!strcasecmp (cmd, "fclear"))
    Command::change_boolean (arg, &do_clear, "fclear");
  else if (!strcasecmp (cmd, "fps"))
    Command::change_boolean (arg, &do_fps, "fps");
  else if (!strcasecmp (cmd, "physics"))
    Command::change_boolean (arg, &do_physics, "physics");
  else if (!strcasecmp (cmd, "colldet"))
    Command::change_boolean (arg, &do_cd, "colldet");
  else if (!strcasecmp (cmd, "zbuf"))
    Command::change_boolean (arg, &do_show_z, "zbuf");
  else if (!strcasecmp (cmd, "stats"))
  {
    Command::change_boolean (arg, &do_stats, "stats");
    if (do_stats) do_show_coord = false;
  }
  else if (!strcasecmp (cmd, "coordshow"))
  {
    Command::change_boolean (arg, &do_show_coord, "coordshow");
    if (do_show_coord) do_stats = false;
  }
  else if (!strcasecmp (cmd, "hi"))
  {
    hilight = arg ? cs->camera->get_sector ()->get_polygon (arg) : (Polygon3D*)NULL;
    if (hilight) CsPrintf (MSG_CONSOLE, "Hilighting polygon: '%s'\n", arg);
    else CsPrintf (MSG_CONSOLE, "Disabled hilighting.\n");
  }
  else if (!strcasecmp (cmd, "p_alpha"))
  {
    if (hilight)
    {
      if (hilight->get_portal ())
      {
        int a = hilight->get_alpha ();
        Command::change_int (arg, &a, "portal alpha", 0, 100);
	hilight->set_alpha (a);
      }
      else CsPrintf (MSG_CONSOLE, "Only for portals!\n");
    }
    else CsPrintf (MSG_CONSOLE, "No polygon selected!\n");
  }
  else if (!strcasecmp (cmd, "s_fog"))
  {
    csFog& f = cs->camera->get_sector ()->get_fog ();
    if (!arg)
    {
      CsPrintf (MSG_CONSOLE, "Fog in current sector (%f,%f,%f) density=%f\n",
      	f.red, f.green, f.blue, f.density);
    }
    else
    {
      float r, g, b, dens;
      if (ScanStr (arg, "%f,%f,%f,%f", &r, &g, &b, &dens) != 4)
      {
        CsPrintf (MSG_CONSOLE, "Expected r,g,b,density. Got something else!\n");
        return false;
      }
      f.enabled = true;
      f.density = dens;
      f.red = r;
      f.green = g;
      f.blue = b;
    }
  }
  else if (!strcasecmp (cmd, "capture"))
    CaptureScreen ();
  else if (!strcasecmp (cmd, "perftest"))
    perf_test ();
  else if (!strcasecmp (cmd, "debug0"))
  {
    CsPrintf (MSG_CONSOLE, "No debug0 implementation in this version.\n");
  }
  else if (!strcasecmp (cmd, "debug1"))
  {
    CsPrintf (MSG_CONSOLE, "No debug1 implementation in this version.\n");
  }
  else if (!strcasecmp (cmd, "debug2"))
  {
    CsPrintf (MSG_CONSOLE, "No debug2 implementation in this version.\n");
  }
  else if (!strcasecmp (cmd, "fire"))
  {
    Vector3 pos;
    Vector3 dir(0,0,0);
    pos = cs->camera->cam2world (dir);
    float r, g, b;
    RandomColor (r, g, b);
    CHK (csDynLight* dyn = new csDynLight (pos.x, pos.y, pos.z, 4, r, g, b));
    cs->world->add_dyn_light (dyn);
    dyn->set_sector (cs->camera->get_sector ());
    dyn->setup ();
    CHK (MissileStruct* ms = new MissileStruct);
    Sys->Sound->playEphemeral (wMissile_whoosh);
    ms->type = DYN_TYPE_MISSILE;
    ms->dir = (OrthoTransform)*(cs->camera);
    ms->sprite = NULL;
    dyn->userdata = (void*)ms;

    char misname[10];
    sprintf (misname, "missile%d", ((rand () >> 3) & 1)+1);

    csSpriteTemplate* tmpl = cs->world->get_sprite_template (misname, true);
    if (!tmpl)
      CsPrintf (MSG_CONSOLE, "Could not find '%s' sprite template!\n", misname);
    else
    {
      CHK (csSprite3D* sp = new csSprite3D ("missile"));
      cs->camera->get_sector ()->add_sprite (cs->world, sp);
      ms->sprite = sp;
      sp->set_template (tmpl);
      sp->set_move (pos);
      Matrix3 m = ms->dir.m_this2world ();
      sp->set_transform (m);
      sp->set_action ("default");
      int i;
      for (i = 0 ; i < tmpl->get_num_vertices () ; i++)
      {
        float r, g, b;
        RandomColor (r, g, b);
        sp->set_vertex_color (i, csColor (r, g, b));
      }
      sp->init_sprite ();
    }
  }
  else if (!strcasecmp (cmd, "addbot"))
  {
    csSpriteTemplate* tmpl = cs->world->get_sprite_template ("missile1", true);
    CHK (delete bot);
    CHK (bot = new Bot (tmpl, "bot"));
    bot->get_mso ().init_one_sector (cs->camera->get_sector ());
    cs->camera->get_sector ()->add_sprite (cs->world, bot);
    bot->set_move (cs->camera->v_origin ());
    bot->set_action ("default");
    bot->init_sprite ();
  }
  else if (!strcasecmp (cmd, "addlight"))
  {
    Vector3 pos;
    Vector3 dir(0,0,0);
    pos = cs->camera->cam2world (dir);
    CHK (csDynLight* dyn = new csDynLight (pos.x, pos.y, pos.z, 6, 1, 1, 1));
    cs->world->add_dyn_light (dyn);
    dyn->set_sector (cs->camera->get_sector ());
    dyn->setup ();
    CHK (RandomLight* rl = new RandomLight);
    rl->type = DYN_TYPE_RANDOM;
    rl->dyn_move_dir = .2;
    rl->dyn_move = 0;
    rl->dyn_r1 = rl->dyn_g1 = rl->dyn_b1 = 1;
    dyn->userdata = (void*)rl;
    CsPrintf (MSG_CONSOLE, "Dynamic light added.\n");
  }
  else if (!strcasecmp (cmd, "dellight"))
  {
    csDynLight* dyn;
    if ((dyn = cs->world->get_first_dyn_light ()) != NULL)
    {
      cs->world->remove_dyn_light (dyn);
      CHK (delete dyn);
      CsPrintf (MSG_CONSOLE, "Dynamic light deleted.\n");
    }
  }
  else if (!strcasecmp (cmd, "dellights"))
  {
    csDynLight* dyn;
    while ((dyn = cs->world->get_first_dyn_light ()) != NULL)
    {
      cs->world->remove_dyn_light (dyn);
      CHK (delete dyn);
    }
    CsPrintf (MSG_CONSOLE, "All dynamic lights deleted.\n");
  }
  else if (!strcasecmp (cmd, "picklight"))
  {
#   if 0
    pickup_light = cs->world->get_first_flt_light ();
    if (pickup_light) CsPrintf (MSG_CONSOLE, "Floating light taken.\n");
    else CsPrintf (MSG_CONSOLE, "No floating light to take.\n");
#   endif
  }
  else if (!strcasecmp (cmd, "droplight"))
  {
#   if 0
    if (pickup_light) CsPrintf (MSG_CONSOLE, "Floating light dropped.\n");
    else CsPrintf (MSG_CONSOLE, "No floating light to drop.\n");
    pickup_light = NULL;
#   endif
  }
  else if (!strcasecmp (cmd, "map"))
  {
    char* choices[4] = { "off", "overlay", "on", NULL };
    Command::change_choice (arg, &System->world->map_mode, "map", choices, 3);
  }
  else if (!strcasecmp (cmd, "snd_play"))
  {
    Wave *w=cs->world->get_wave(arg);
    if (w)
    {
      w->convertTo(Sys->Sound->getFrequency(), Sys->Sound->is16Bits(), false);
      Sys->Sound->playEphemeral(w);
    }
    else CsPrintf (MSG_CONSOLE, "Wave '%s' not found!\n", arg);
  }
  else if (!strcasecmp (cmd, "snd_volume"))
  {
    float vol = Sys->Sound->getVolume ();
    Command::change_float (arg, &vol, "snd_volume", 0.0, 1.0);
    Sys->Sound->setVolume (vol);
  }
  else
    return false;
  return true;
}

//-----------------------------------------------------------------------------

char WalkTest::world_file[100];

WalkTest::WalkTest () : SysSystemDriver ()
{
  Command::ExtraHandler = ExtraHandler;
}

void WalkTest::ParseArg (int argc, char* argv[], int& i)
{
  if (argv[i][0] != '-')
  {
    strcpy (world_file, argv[i]);
  }
  if (strcasecmp ("-clear", argv[i]) == 0)
  {
    do_clear = true;
    CsPrintf (MSG_INITIALIZATION, "Screen will be cleared every frame.\n");
  }
  else if (strcasecmp ("-noclear", argv[i]) == 0)
  {
    do_clear = false;
    CsPrintf (MSG_INITIALIZATION, "Screen will not be cleared every frame.\n");
  }
  else if (strcasecmp ("-stats", argv[i]) == 0)
  {
    do_stats = true;
    CsPrintf (MSG_INITIALIZATION, "Statistics enabled.\n");
  }
  else if (strcasecmp ("-nostats", argv[i]) == 0)
  {
    do_stats = false;
    CsPrintf (MSG_INITIALIZATION, "Statistics disabled.\n");
  }
  else if (strcasecmp ("-fps", argv[i]) == 0)
  {
    do_fps = true;
    CsPrintf (MSG_INITIALIZATION, "Frame Per Second enabled.\n");
  }
  else if (strcasecmp ("-nofps", argv[i]) == 0)
  {
    do_fps = false;
    CsPrintf (MSG_INITIALIZATION, "Frame Per Second disabled.\n");
  }
  else if (strcasecmp ("-physics", argv[i]) == 0)
  {
    do_physics = true;
    CsPrintf (MSG_INITIALIZATION, "Enabled physics system.\n");
  }
  else if (strcasecmp ("-nophysics", argv[i]) == 0)
  {
    do_physics = false;
    CsPrintf (MSG_INITIALIZATION, "Disabled physics system.\n");
  }
  else if (strcasecmp ("-infinite", argv[i]) == 0)
  {
    extern bool do_infinite;
    do_infinite = true;
  }
  else if (strcasecmp ("-colldet", argv[i]) == 0)
  {
    do_cd = true;
    CsPrintf (MSG_INITIALIZATION, "Enabled collision detection system.\n");
  }
  else if (strcasecmp ("-nocolldet", argv[i]) == 0)
  {
    do_cd = false;
    CsPrintf (MSG_INITIALIZATION, "Disabled collision detection system.\n");
  }
  else
    SysSystemDriver::ParseArg (argc, argv, i);
}

void WalkTest::Help ()
{
  SysSystemDriver::Help ();
  CsPrintf (MSG_STDOUT, "  -clear/noclear     clear display every frame (default '%sclear')\n", do_clear ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -stats/nostats     statistics (default '%sstats')\n", do_stats ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -fps/nofps         frame rate printing (default '%sfps')\n", do_fps ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -physics/nophysics preliminary physics system (default '%sphysics')\n", do_physics ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -colldet/nocolldet collision detection system (default '%scolldet')\n", do_cd ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -infinite          special infinite level generation (ignores world file!)\n");
  CsPrintf (MSG_STDOUT, "  <filename>         load world file from <filename> (default '%s')\n", world_file);
}
  
/*------------------------------------------------------------------
 * The following handle_key_... routines are general movement
 * routines that are called by do_update() for the new movement
 * system and by do_keypress() for the old movement system (see
 * system.h for an explanation of the difference between the two
 * systems).
 *------------------------------------------------------------------*/

extern bool do_physics;
extern bool do_coord_check;
extern Vector2 coord_check_vector;
extern csCamera c;
extern WalkTest* Sys;

void WalkTest::handle_key_forward (float speed, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_up (shift, alt, ctrl); return; }
  if (do_physics)
  {
    if (speed==0) {}
    if (alt) {}
      /* cs->camera->get_physics ()->applyForce (0.0, speed, 0.0); */ // Up
    else {}
      /* cs->camera->get_physics ()->useFoot (0.0, 0.5, speed); */  // Forward
  }
  else
  {
    if (alt)
    {
      if (ctrl) cs->camera->move (.01*VEC_UP);
      else if (shift) cs->camera->move (.4*VEC_UP);
      else cs->camera->move (.2*VEC_UP);
    }
    else if (ctrl) cs->camera->move (.01*VEC_FORWARD);
    else if (shift) cs->camera->move (1.2*VEC_FORWARD);
    else cs->camera->move (.6*VEC_FORWARD);
  }
}

void WalkTest::handle_key_backwards (float speed, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_down (shift, alt, ctrl); return; }
  if (do_physics)
  {
    if (speed==0) {}
    if (alt) {}
    //  cs->camera->get_physics ()->applyForce (0.0, -speed, 0.0); // Down
    else {}
    //  cs->camera->get_physics ()->useFoot (0.0, 0.5, -speed); // Backwards
  }
  else
  {
    if (alt)
    {
      if (ctrl) cs->camera->move (.01*VEC_DOWN);
      else if (shift) cs->camera->move (.4*VEC_DOWN);
      else cs->camera->move (.2*VEC_DOWN);
    }
    else if (ctrl) cs->camera->move (.01*VEC_BACKWARD);
    else if (shift) cs->camera->move (1.2*VEC_BACKWARD);
    else cs->camera->move (.6*VEC_BACKWARD);
  }
}

void WalkTest::handle_key_left (float speed, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_left (shift, alt, ctrl); return; }
  if (do_physics)
  {
    if (speed==0) {}
    if (alt) {}
    //  cs->camera->get_physics ()->useFoot (-speed, 0.5, 0.0); // Left
    else if (ctrl) cs->camera->rotate (VEC_ROT_LEFT, .005);
    else if (shift) cs->camera->rotate (VEC_ROT_LEFT, .2);
    else cs->camera->rotate (VEC_ROT_LEFT, .1);
  }
  else
  {
    if (alt)
    {
      if (ctrl) cs->camera->move (.01*VEC_LEFT);
      else if (shift) cs->camera->move (.4*VEC_LEFT);
      else cs->camera->move (.2*VEC_LEFT);
    }
    else if (ctrl) cs->camera->rotate (VEC_ROT_LEFT, .005);
    else if (shift) cs->camera->rotate (VEC_ROT_LEFT, .2);
    else cs->camera->rotate (VEC_ROT_LEFT, .1);
  }
}

void WalkTest::handle_key_right (float speed, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_right (shift, alt, ctrl); return; }
  if (do_physics)
  {
    if (speed==0) {}
    if (alt) {}
    //  cs->camera->get_physics ()->useFoot (speed, 0.5, 0.0); // Right
    else if (ctrl) cs->camera->rotate (VEC_ROT_RIGHT, .005);
    else if (shift) cs->camera->rotate (VEC_ROT_RIGHT, .2);
    else cs->camera->rotate (VEC_ROT_RIGHT, .1);
  }
  else
  {
    if (alt)
    {
      if (ctrl) cs->camera->move (.01*VEC_RIGHT);
      else if (shift) cs->camera->move (.4*VEC_RIGHT);
      else cs->camera->move (.2*VEC_RIGHT);
    }
    else if (ctrl) cs->camera->rotate (VEC_ROT_RIGHT, .005);
    else if (shift) cs->camera->rotate (VEC_ROT_RIGHT, .2);
    else cs->camera->rotate (VEC_ROT_RIGHT, .1);
  }
}

void WalkTest::handle_key_pgup (float, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_pgup (shift, alt, ctrl); return; }
  if (alt)
  {
    if (ctrl) cs->camera->rotate (VEC_TILT_LEFT, .005);
    else cs->camera->rotate (VEC_TILT_LEFT, .1);
  }
  else if (ctrl) cs->camera->rotate (VEC_TILT_UP, .005);
  else if (shift) cs->camera->rotate (VEC_TILT_UP, .2);
  else cs->camera->rotate (VEC_TILT_UP, .1);
}

void WalkTest::handle_key_pgdn (float, bool shift, bool alt, bool ctrl)
{
  if (System->world->map_mode) { System->world->wf->key_pgdn (shift, alt, ctrl); return; }
  if (alt)
  {
    if (ctrl) cs->camera->rotate (VEC_TILT_RIGHT, .005);
    else cs->camera->rotate (VEC_TILT_RIGHT, .1);
  }
  else if (ctrl) cs->camera->rotate (VEC_TILT_DOWN, .005);
  else if (shift) cs->camera->rotate (VEC_TILT_DOWN, .2);
  else cs->camera->rotate (VEC_TILT_DOWN, .1);
}

void WalkTest::eatkeypress (int key, bool shift, bool alt, bool ctrl)
{
  int i;
  float speed = 1;
  if (ctrl) speed = .5;
  if (shift) speed = 2;

  if (System->Console->IsActive ())
    ((SimpleConsole *)System->Console)->AddChar (key);
  else switch (key)
  {
    case CSKEY_TAB:
      System->Console->Show ();
      break;

    case 'j': // Jump
    //  cs->camera->get_physics ()->useFoot (0.0, speed * 10.0, 0.0); // Jump up
      break;

    default:
      {
	csKeyMap* m = mapping;
        while (m)
        {
          if (key == m->key && shift == m->shift
           && alt == m->alt && ctrl == m->ctrl)
          {
            Command::perform_line (m->cmd);
            break;
          }
	  m = m->next;
        }
      }
      break;
  }
}

bool track_rmb = false;
int track_mousex;

void WalkTest::NextFrame (long elapsed_time, long current_time)
{
  SysSystemDriver::NextFrame (elapsed_time, current_time);

  if (!System->Console->IsActive ())
  {
    int alt,shift,ctrl;
    float speed = 1;

    alt = Keyboard->Key.alt;
    ctrl = Keyboard->Key.ctrl;
    shift = Keyboard->Key.shift;
    if (ctrl) speed = .5;
    if (shift) speed = 2;

    if (Keyboard->Key.up)    handle_key_forward (speed, shift, alt, ctrl);
    if (Keyboard->Key.down)  handle_key_backwards (speed, shift, alt, ctrl);
    if (Keyboard->Key.left)  handle_key_left (speed, shift, alt, ctrl);
    if (Keyboard->Key.right) handle_key_right (speed, shift, alt, ctrl);
    if (Keyboard->Key.pgup)  handle_key_pgup (speed, shift, alt, ctrl);
    if (Keyboard->Key.pgdn)  handle_key_pgdn (speed, shift, alt, ctrl);
  } /* endif */

  csEvent *Event;
  while ((Event = Sys->EventQueue->Get ()))
  {
    switch (Event->Type)
    {
      case csevKeyDown:
        eatkeypress (Event->Key.Code, Event->Key.ShiftKeys & CSMASK_SHIFT,
          Event->Key.ShiftKeys & CSMASK_ALT, Event->Key.ShiftKeys & CSMASK_CTRL);
        break;
      case csevBroadcast:
        if ((Event->Command.Code == cscmdFocusChanged)
         && (Event->Command.Info == NULL))
          memset (&Keyboard->Key, 0, sizeof (Keyboard->Key));
        break;
      case csevMouseDown:
        if (Event->Mouse.Button == 1)
        {
          unsigned long* zb;
	  Gfx3D->GetZBufPoint(Event->Mouse.x, Event->Mouse.y, &zb);

	  Vector3 v, vw;
	  v.z = 1. / (((float)*zb)/(256.*65536.));
	  v.x = (Event->Mouse.x-FRAME_WIDTH/2) * v.z / csCamera::aspect;
	  v.y = (FRAME_HEIGHT-1-Event->Mouse.y-FRAME_HEIGHT/2) * v.z / csCamera::aspect;
	  vw = cs->camera->cam2world (v);

          CsPrintf (MSG_CONSOLE, "LMB down : z_buf=%ld cam:(%f,%f,%f) world:(%f,%f,%f)\n", zb, v.x, v.y, v.z, vw.x, vw.y, vw.z);
          CsPrintf (MSG_DEBUG_0, "LMB down : z_buf=%ld cam:(%f,%f,%f) world:(%f,%f,%f)\n", zb, v.x, v.y, v.z, vw.x, vw.y, vw.z);

          coord_check_vector.x = Event->Mouse.x;
          coord_check_vector.y = FRAME_HEIGHT-Event->Mouse.y;
          do_coord_check = true;
          cs->Draw ();
          do_coord_check = false;
        } else if (Event->Mouse.Button == 2)
        {
          CsPrintf (MSG_CONSOLE, "RMB down : tracking left X with mouse\n");
          track_rmb = true;
          track_mousex = Event->Mouse.x;
        }
        break;
      case csevMouseMove:
        if (track_rmb)
        {
	  // DAN:
          //extern float global_deltaX;
          //global_deltaX = ((float)(Event->Mouse.x - track_mousex)) / 32;
        }
        break;
      case csevMouseUp:
        if (Event->Mouse.Button == 2)
        {
          CsPrintf (MSG_CONSOLE, "RMB up : tracking finished\n");
	  // DAN: this is commented out until I figure out how to do this.
//          extern float global_deltaX;
//          global_deltaX = 0;
          track_rmb = false;
        }
        break;
    }
    CHK (delete Event);
  }

  if (bot) bot->move (elapsed_time);

  PrepareFrame (elapsed_time, current_time);
  DrawFrame (elapsed_time, current_time);

  // Execute one line from the script.
  if (!busy_perf_test)
  {
    char buf[256];
    if (Command::get_script_line (buf, 255)) Command::perform_line (buf);
  }
}
