/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>

#include "system/sysprint.h"
#include "geometry/polyint.h"
#include "geometry/bsp.h"

//---------------------------------------------------------------------------

BspNode::BspNode ()
{
  front = back = NULL;
  polygons = NULL;
  num = max = 0;
}

BspNode::~BspNode ()
{
  if (polygons) CHKB (delete [] polygons);
}

void BspNode::add_polygon (PolygonInt* poly)
{
  if (!polygons)
  {
    max = 3;
    CHK (polygons = new PolygonInt* [max]);
  }

  if (num >= max)
  {
    CHK (PolygonInt** pp = new PolygonInt* [max+2]);
    memcpy (pp, polygons, sizeof (PolygonInt*)*max);
    max += 2;
    CHK (delete [] polygons);
    polygons = pp;
  }

  polygons[num] = poly;
  num++;
}

//---------------------------------------------------------------------------

BspTree::BspTree (PolygonParentInt* pset)
{
  BspTree::pset = pset;

  int i;
  int num = pset->get_num_polygon ();
  CHK (PolygonInt** polygons = new PolygonInt* [num]);
  for (i = 0 ; i < num ; i++) polygons[i] = pset->get_polygon (i);

  CHK (root = new BspNode);

  build (root, polygons, num);

  CHK (delete [] polygons);
}

BspTree::BspTree (PolygonParentInt* pset, PolygonInt** polygons, int num)
{
  BspTree::pset = pset;
  CHK (root = new BspNode);
  build (root, polygons, num);
}

BspTree::~BspTree ()
{
  clear (root);
}

void BspTree::clear (BspNode* node)
{
  if (!node) return;
  clear (node->front);
  clear (node->back);
  CHK (delete node);
}

// Set this to 1 to optimize the BSP tree based on the number of
// splits. Otherwise the splitter will be chosen which has the
// highest number of polygons on the same plane as the splitter.
#define BSP_MINIMAL_SPLIT 1

void BspTree::build (BspNode* node, PolygonInt** polygons, int num)
{
  int i, j, cnt, poly_idx;

#if BSP_MINIMAL_SPLIT
  // Choose the polygon which generates the least number of splits.
  int min_splits = 32767;
  poly_idx = -1;
  for (i = 0 ; i < num ; i++)
  {
    int cnt = 0;
    for (j = 0 ; j < num ; j++)
      if (polygons[j]->classify (polygons[i]) == POL_SPLIT_NEEDED) cnt++;
    if (cnt < min_splits) { min_splits = cnt; poly_idx = i; }
  }
#else
  // First choose a polygon which shares its plane with the highest
  // number of other polygons.
  int same_poly = 0;
  poly_idx = -1;
  for (i = 0 ; i < num ; i++)
  {
    cnt = 1;
    for (j = i+1 ; j < num ; j++)
      if (polygons[i]->same_plane (polygons[j])) cnt++;
    if (cnt > same_poly) { same_poly = cnt; poly_idx = i; }
  }
#endif //BSP_MINIMAL_SPLIT

  Plane* plane = polygons[poly_idx]->get_poly_plane ();

  // Now we split the node according to the plane of that polygon.
  CHK (PolygonInt** front_poly = new PolygonInt* [num]);
  CHK (PolygonInt** back_poly = new PolygonInt* [num]);
  int front_idx = 0, back_idx = 0;

  for (i = 0 ; i < num ; i++)
  {
    int c = polygons[i]->classify (polygons[poly_idx]);
    switch (c)
    {
      case POL_SAME_PLANE: node->add_polygon (polygons[i]); break;
      case POL_FRONT: front_poly[front_idx++] = polygons[i]; break;
      case POL_BACK: back_poly[back_idx++] = polygons[i]; break;
      case POL_SPLIT_NEEDED:
	{
	  Vector3* front_v, * back_v;
	  int front_n, back_n, j;
	  int n = polygons[i]->get_num_vertices ();
  	  CHK (front_v = new Vector3 [n*2]);
  	  CHK (back_v = new Vector3 [n*2]);
	  polygons[i]->split_with_plane (front_v, front_n, back_v, back_n, *plane);

	  PolygonInt* np1 = polygons[i]->clone ();
	  PolygonInt* np2 = polygons[i]->clone ();	// polygons[i]@@@
	  np1->reset ();
	  np2->reset ();

	  for (j = 0 ; j < front_n ; j++)
	    np2->add_vertex (front_v[j]);
	  for (j = 0 ; j < back_n ; j++)
	    np1->add_vertex (back_v[j]);
	  polygons[i]->get_parent ()->add_polygon (np1);
	  polygons[i]->get_parent ()->add_polygon (np2);

	  np1->finish ();
	  np2->finish ();

	  CHK (delete [] front_v);
	  CHK (delete [] back_v);

	  front_poly[front_idx++] = np2;
	  back_poly[back_idx++] = np1;
	}
	break;

    }
  }

  if (front_idx)
  {
    CHK (node->front = new BspNode);
    build (node->front, front_poly, front_idx);
  }
  if (back_idx)
  {
    CHK (node->back = new BspNode);
    build (node->back, back_poly, back_idx);
  }

  CHK (delete [] front_poly);
  CHK (delete [] back_poly);
}

void* BspTree::back2front (const Vector3& pos, BspVisitFunc* func, void* data)
{
  return back2front (root, pos, func, data);
}

void* BspTree::front2back (const Vector3& pos, BspVisitFunc* func, void* data)
{
  return front2back (root, pos, func, data);
}

void* BspTree::back2front (BspNode* node, const Vector3& pos, BspVisitFunc* func, void* data)
{
  if (!node) return NULL;
  void* rc;

  // Check if some polygon (just take the first) of the polygons array
  // is visible from the given point. If so, we are in front of this node.
  if (node->polygons[0]->get_poly_plane ()->visible_from_point (pos))
  {
    // Front.
    rc = back2front (node->back, pos, func, data);
    if (rc) return rc;
    rc = func (pset, node->polygons, node->num, data);
    if (rc) return rc;
    rc = back2front (node->front, pos, func, data);
    if (rc) return rc;
  }
  else
  {
    // Back.
    rc = back2front (node->front, pos, func, data);
    if (rc) return rc;
    rc = func (pset, node->polygons, node->num, data);
    if (rc) return rc;
    rc = back2front (node->back, pos, func, data);
    if (rc) return rc;
  }
  return NULL;
}

void* BspTree::front2back (BspNode* node, const Vector3& pos, BspVisitFunc* func, void* data)
{
  if (!node) return NULL;
  void* rc;

  // Check if some polygon (just take the first) of the polygons array
  // is visible from the given point. If so, we are in front of this node.
  if (node->polygons[0]->get_poly_plane ()->visible_from_point (pos))
  {
    // Front.
    rc = front2back (node->front, pos, func, data);
    if (rc) return rc;
    rc = func (pset, node->polygons, node->num, data);
    if (rc) return rc;
    rc = front2back (node->back, pos, func, data);
    if (rc) return rc;
  }
  else
  {
    // Back.
    rc = front2back (node->back, pos, func, data);
    if (rc) return rc;
    rc = func (pset, node->polygons, node->num, data);
    if (rc) return rc;
    rc = front2back (node->front, pos, func, data);
    if (rc) return rc;
  }
  return NULL;
}

void BspTree::dump (BspNode* node, int indent)
{
  if (!node) return;
  int i;
  char spaces[256];
  for (i = 0 ; i < indent ; i++) spaces[i] = ' ';
  spaces[indent] = 0;
  CsPrintf (MSG_DEBUG_0, "%sThere are %d polygons in this node.\n", spaces, node->num);
  CsPrintf (MSG_DEBUG_0, "%s Front:\n", spaces);
  dump (node->front, indent+2);
  CsPrintf (MSG_DEBUG_0, "%s Back:\n", spaces);
  dump (node->back, indent+2);
}

void BspTree::dump ()
{
  dump (root, 0);
}

//---------------------------------------------------------------------------
