/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef BSP_H
#define BSP_H

#include "geometry/math3d.h"

class PolygonInt;
class PolygonParentInt;
class BspTree;

/**
 * A BSP node.
 */
class BspNode
{
  ///
  friend class BspTree;
private:
  /// All the polygons in this node.
  PolygonInt** polygons;
  ///
  int num;
  ///
  int max;
  /// The front node.
  BspNode* front;
  /// The back node.
  BspNode* back;

private:
  ///
  BspNode ();
  ///
  ~BspNode ();

  /// Add a polygon to this BSP node.
  void add_polygon (PolygonInt* poly);
};

/**
 * Visit a node in the BSP tree. If this function returns non-NULL
 * the scanning will stop and the pointer will be returned.
 */
typedef void* (BspVisitFunc)(PolygonParentInt*, PolygonInt**, int num, void*);

/**
 * The BSP tree.
 */
class BspTree
{
private:
  /// The root of the tree.
  BspNode* root;

  /// The parent that this tree is made for.
  PolygonParentInt* pset;

  /// Build the tree from the given node and number of polygons.
  void build (BspNode* node, PolygonInt** polygons, int num);

  /// Clear the node.
  void clear (BspNode* node);

  /// Traverse the tree from back to front starting at 'node' and 'pos'.
  void* back2front (BspNode* node, const Vector3& pos, BspVisitFunc* func, void* data);
  /// Traverse the tree from front to back starting at 'node' and 'pos'.
  void* front2back (BspNode* node, const Vector3& pos, BspVisitFunc* func, void* data);

  /// Do a debugging dump of the BSP tree.
  void dump (BspNode* node, int indent);

public:
  /// Create the tree for the given parent.
  BspTree (PolygonParentInt* pset);
  /// Create the tree with a given set of polygons.
  BspTree (PolygonParentInt* pset, PolygonInt** polygons, int num);
  ///
  ~BspTree ();

  /// Traverse the tree from back to front starting at the root and 'pos'.
  void* back2front (const Vector3& pos, BspVisitFunc* func, void* data);
  /// Traverse the tree from front to back starting at the root and 'pos'.
  void* front2back (const Vector3& pos, BspVisitFunc* func, void* data);

  /// Get the polygonset for this BSP tree.
  PolygonParentInt* get_parent () { return pset; }

  /// Do a debugging dump for the BSP tree.
  void dump ();
};

#endif /*BSP_H*/

