/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Largely rewritten by Ivan Avramovic <iavramov@ibm.net>
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef MATH2D_H
#define MATH2D_H

#define POLY_IN 1
#define POLY_ON 0
#define POLY_OUT -1

//#include "util/def.h"

class Vector2;
class Box;

/**
 * A 2D vector.
 */
class Vector2
{
public:
  ///
  float x;
  ///
  float y;

  /// Make a new vector. No initialization is done.
  Vector2 () {}

  /// Make a new vector and initialize with the given values.
  Vector2 (float x, float y) { Vector2::x = x; Vector2::y = y; }

  /**
   * Return -1 if vector is left of 'v1-v2'. 1 is vector is right of 'v1-v2'
   * or 0 is vector is on 'v1-v2'.
   */
  int which_side_2d (Vector2& v1, Vector2& v2);

  /**
   * Return POLY_IN, POLY_OUT, or POLY_ON for this vector with respect
   * to the given polygon. The polygon is given as an array of 2D vectors
   * with a bounding box.
   */
  int in_poly_2d (Vector2* P, int n, Box* bounding_box);

  /// Set vector to given values
  inline void set (float ix, float iy)
  { x = ix; y = iy; }

  /// Return the norm (magnitude) of a 2D vector
  static float norm (const Vector2& v);

  /// Return the norm (magnitude) of this vector
  float norm() const;

  /// Rotate vector around the origin by a given angle in radians
  void Rotate(float angle);

  /**
   * Returns twice the signed area of the triangle determined by a,b,c,
   * positive if a,b,c are oriented ccw, and negative if cw.
   */
  static float Area2 (float ax, float ay,
                      float bx, float by, 
                      float cx, float cy);

  /**
   * Returns true iff c is strictly to the right of the directed
   * line through a to b.
   */
  static bool Right (float ax, float ay,
                     float bx, float by,
                     float cx, float cy);

  /**
   * Returns true iff c is strictly to the left of the directed
   * line through a to b.
   */
  static bool Left (float ax, float ay,
                    float bx, float by,
                    float cx, float cy);

  /// Add another vector to this vector.
  Vector2& operator+= (const Vector2& v) 
  { x += v.x;  y += v.y;  return *this; }
  /// Subtract another vector from this vector.
  Vector2& operator-= (const Vector2& v)
  { x -= v.x;  y -= v.y;  return *this; }
  /// Multiply this vector by a scalar.
  Vector2& operator*= (float f) { x *= f;  y *= f;  return *this; }
  /// Divide this vector by a scalar.
  Vector2& operator/= (float f) { x /= f;  y /= f;  return *this; }

  inline Vector2 operator+ () const { return *this; }
  inline Vector2 operator- () const { return Vector2(-x,-y); }

  /// add two vectors
  friend Vector2 operator+ (const Vector2& v1, const Vector2& v2);
  /// subtract two vectors
  friend Vector2 operator- (const Vector2& v1, const Vector2& v2);
  /// take the dot product of two vectors
  friend float operator* (const Vector2& v1, const Vector2& v2);
  /// multiply a vector and a scalar
  friend Vector2 operator* (const Vector2& v, float f);
  /// multiply a vector and a scalar
  friend Vector2 operator* (float f, const Vector2& v);
  /// divide a vector by a scalar
  friend Vector2 operator/ (const Vector2& v, float f);
  /// check if two vectors are equal
  friend bool operator== (const Vector2& v1, const Vector2& v2);
  /// check if two vectors are not equal
  friend bool operator!= (const Vector2& v1, const Vector2& v2);
};

/**
 * A bounding box in 2D space.
 */
class Box
{
public:
  /// The top-left x coordinate of the bounding box.
  float minx;
  /// The top-left y coordinate of the bounding box.
  float miny;
  /// The bottom-right x coordinate of the bounding box.
  float maxx;
  /// The bottom-right y coordinate of the bounding box.
  float maxy;

  /// Initialize this box to empty.
  Box() : minx(10000.), miny(10000.), maxx(-10000.), maxy(-10000.) {}

  /// Initialize this box with one point.
  Box(const Vector2& v) : minx(v.x), miny(v.y), maxx(v.x), maxy(v.y) {}

  /// Initialize this box with the given values
  Box(float x1,float y1,float x2,float y2) :
   minx(x1), miny(y1), maxx(x2), maxy(y2) {}

  /// Test if the given coordinate is in this box.
  bool in (float x, float y)
  {
    if (x < minx || x > maxx) return false;
    if (y < miny || y > maxy) return false;
    return true;
  }

  /// Test if this box overlaps with the given box.
  bool overlap (const Box& box)
  {
    if (maxx < box.minx || minx > box.maxx) return false;
    if (maxy < box.miny || miny > box.maxy) return false;
    return true;
  }

  /// Test if this box is empty
  bool empty() 
  {
    if (minx > maxx) return true;
    if (miny > maxy) return true;
    return false;
  }

  /// Initialize this box to empty.
  void start_bounding_box ()
  {
    minx =  10000.;  miny =  10000.;
    maxx = -10000.;  maxy = -10000.;
  }

  /// Initialize this box to one vertex.
  void start_bounding_box (Vector2& v)
  {
    minx = v.x;  miny = v.y;
    maxx = v.x;  maxy = v.y;
  }

  /// Add a new vertex and recalculate the bounding box.
  void add_bounding_vertex (float x, float y)
  {
    if (x < minx) minx = x;  if (x > maxx) maxx = x;
    if (y < miny) miny = y;  if (y > maxy) maxy = y;
  }

  /**
   * Add a new vertex and recalculate the bounding box.
   * This version is a little more optimal. It assumes however
   * that at least one point has been added to the bounding box.
   */
  void add_bounding_vertex_smart (float x, float y)
  {
    if (x < minx) minx = x; else if (x > maxx) maxx = x;
    if (y < miny) miny = y; else if (y > maxy) maxy = y;
  }

  /// Add a new vertex and recalculate the bounding box.
  void add_bounding_vertex (const Vector2& v)
  {
    if (v.x < minx) minx = v.x;  if (v.x > maxx) maxx = v.x;
    if (v.y < miny) miny = v.y;  if (v.y > maxy) maxy = v.y;
  }

  /// Add a new vertex and recalculate the bounding box.
  void add_bounding_vertex_smart (const Vector2& v)
  {
    if (v.x < minx) minx = v.x; else if (v.x > maxx) maxx = v.x;
    if (v.y < miny) miny = v.y; else if (v.y > maxy) maxy = v.y;
  }

  /// compute the union of two bounding boxes
  Box& operator+= (const Box& box);
  /// compute the union of a point with this bounding box
  Box& operator+= (const Vector2& point);
  /// compute the intersection of two bounding boxes
  Box& operator*= (const Box& box);

  /// compute the union of two bounding boxes
  friend Box operator+ (const Box& box1, const Box& box2);
  /// compute the union of a bounding box and a point
  friend Box operator+ (const Box& box, const Vector2& point);
  /// compute the intersection of two bounding boxes
  friend Box operator* (const Box& box1, const Box& box2);

  /// tests if two bounding boxes are equal
  friend bool operator== (const Box& box1, const Box& box2);
  /// tests if two bounding boxes are unequal
  friend bool operator!= (const Box& box1, const Box& box2);
  /// tests if box1 is a subset of box2
  friend bool operator< (const Box& box1, const Box& box2);
  /// tests if box1 is a superset of box2
  friend bool operator> (const Box& box1, const Box& box2);
  /// tests if a point is contained in a box
  friend bool operator< (const Vector2& point, const Box& box);
};

/** 
 * Some functions to perform various intersection calculations with 2D
 * line segments.  This is a static class and contains only static member
 * functions.
 */
class Intersect2
{
public:
  /**
   * Compute the intersection of the 2D segments.  Return true if they
   * intersect, with the intersection point returned in isect,  and the
   * distance from a1 of the intersection in dist.
   */
  static bool segments(
    const Vector2& a1, const Vector2& a2, // first segment
    const Vector2& b1, const Vector2& b2, // second segment
    Vector2& isect, float& dist);         // intersection point and distance

  /**
   * Compute the intersection of a 2D segment and a line.  Return true if they
   * intersect, with the intersection point returned in isect,  and the
   * distance from a1 of the intersection in dist.
   */
  static bool segment_line(
    const Vector2& a1, const Vector2& a2, // first segment
    const Vector2& b1, const Vector2& b2, // second line
    Vector2& isect, float& dist);         // intersection point and distance

  /**
   * Compute the intersection of 2D lines.  Return true if they
   * intersect, with the intersection point returned in isect.
   */
  static bool lines(
    const Vector2& a1, const Vector2& a2, // first line
    const Vector2& b1, const Vector2& b2, // second line
    Vector2& isect);                      // intersection point
};

#endif /*MATH_H*/
