/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Largely rewritten by Ivan Avramovic <iavramov@ibm.net>
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <time.h>

#include "geometry/math3d.h"
#include "util/util.h"

//---------------------------------------------------------------------------

Tables tables;

void Tables::initialize ()
{
  int i;
  for (i = -360 ; i <= 360 ; i++)
  {
    sin_tab[i+360] = sin (2.*M_PI*(float)i / 360.);
    cos_tab[i+360] = cos (2.*M_PI*(float)i / 360.);
  }
  for (i = -512 ; i <= 512 ; i++)
  {
    asin_tab[i+512] = QInt (360. * asin (((float)i) / 512.) / (2.*M_PI));
    acos_tab[i+512] = QInt (360. * acos (((float)i) / 512.) / (2.*M_PI));
    cos_asin_tab[i+512] = QInt (16.*256.*cos_tab [ asin_tab [ i+512 ] + 360 ]);
  }
  for (i = 0 ; i < 65535 ; i++)
  {
    float dist_x_density = ((float)i)/256.;
    exp_table[i] = (UByte)((1-exp (-dist_x_density))*255);
  }
}

//---------------------------------------------------------------------------

Matrix3::Matrix3 () 
{
  m12 = m13 = 0;
  m21 = m23 = 0;
  m31 = m32 = 0;
  m11 = m22 = m33 = 1;
}

Matrix3::Matrix3 (float m11, float m12, float m13,
  	    	  float m21, float m22, float m23,
  	   	  float m31, float m32, float m33)
{
  Matrix3::m11 = m11;
  Matrix3::m12 = m12;
  Matrix3::m13 = m13;
  Matrix3::m21 = m21;
  Matrix3::m22 = m22;
  Matrix3::m23 = m23;
  Matrix3::m31 = m31;
  Matrix3::m32 = m32;
  Matrix3::m33 = m33;
}

Matrix3& Matrix3::operator+= (const Matrix3& m)
{
  m11 += m.m11; m12 += m.m12; m13 += m.m13;
  m21 += m.m21; m22 += m.m22; m23 += m.m23;
  m31 += m.m31; m32 += m.m32; m33 += m.m33;
  return *this;
}

Matrix3& Matrix3::operator-= (const Matrix3& m)
{
  m11 -= m.m11; m12 -= m.m12; m13 -= m.m13;
  m21 -= m.m21; m22 -= m.m22; m23 -= m.m23;
  m31 -= m.m31; m32 -= m.m32; m33 -= m.m33;
  return *this;
}

Matrix3& Matrix3::operator*= (const Matrix3& m)
{
  Matrix3 r;
  r.m11 = m11*m.m11 + m12*m.m21 + m13*m.m31;
  r.m12 = m11*m.m12 + m12*m.m22 + m13*m.m32;
  r.m13 = m11*m.m13 + m12*m.m23 + m13*m.m33;
  r.m21 = m21*m.m11 + m22*m.m21 + m23*m.m31;
  r.m22 = m21*m.m12 + m22*m.m22 + m23*m.m32;
  r.m23 = m21*m.m13 + m22*m.m23 + m23*m.m33;
  r.m31 = m31*m.m11 + m32*m.m21 + m33*m.m31;
  r.m32 = m31*m.m12 + m32*m.m22 + m33*m.m32;
  r.m33 = m31*m.m13 + m32*m.m23 + m33*m.m33;
  *this = r;
  return *this;
}

Matrix3& Matrix3::operator*= (float s)
{
  m11 *= s; m12 *= s; m13 *= s;
  m21 *= s; m22 *= s; m23 *= s;
  m31 *= s; m32 *= s; m33 *= s;
  return *this;
}

void Matrix3::identity ()
{
  m12 = m13 = 0;
  m21 = m23 = 0;
  m31 = m32 = 0;
  m11 = m22 = m33 = 1;
}

void Matrix3::transpose ()
{
  float swap;
  swap = m12; m12 = m21; m21 = swap;
  swap = m13; m13 = m31; m31 = swap;
  swap = m23; m23 = m32; m32 = swap;
}

Matrix3 Matrix3::gettranspose () const
{
  Matrix3 t;
  t.m12 = m21; t.m21 = m12;
  t.m13 = m31; t.m31 = m13;
  t.m23 = m32; t.m32 = m23;
  t.m11 = m11; t.m22 = m22; t.m33 = m33;
  return t;
}

float Matrix3::determinant () const
{
  return
    m11 * (m22*m33 - m23*m32)
   -m12 * (m21*m33 - m23*m31)
   +m13 * (m21*m32 - m22*m31);
}


// this is the matrix for which we seek to find the eigen values.
// vout is the matrix of eigen vectors.
// dout is the vector of dominant eigen values.
// returns:  -1   - error failed to converge within 50 iterations.
//           0-50 - number of iterations.
int Matrix3::eigen(Matrix3* vout, Vector3* dout)
{
#define rotate(a1,a2) g=a1; h=a2; a1=g-s*(h+g*tau); a2=h+s*(g-h*tau);
#define rfabs(x) ((x < 0) ? -x : x)
  int i;
  float tresh,theta,tau,t,sm,s,h,g,c;
  int nrot;

  Matrix3 v;
  Vector3 b,z,d;

  // Load b and d with the diagonals of a.
  b.x = d.x = m11;
  b.y = d.y = m22;
  b.z = d.z = m33;

  nrot = 0;
  
  // Try up to 50 times.
  for(i=0; i<50; i++)
    {
      // See if bottom half of matrix a is non zero.
      sm=0.0; sm+=fabs(m12); sm+=fabs(m13); sm+=fabs(m23);
      // If it is 0 we are done.  Return the current vector v and d.
      if (sm == 0.0)
	{
	  *vout = v;
	  *dout = d;
	  return i;
	}
      
      if (i < 3) tresh=0.2*sm/(3*3); else tresh=0.0;
      
      // Try rotations in 1st dimension
      {
	g = 100.0*rfabs(m12);  
	// Does this make sense??
	// equiv to   if (i>3 && g == 0) 
	if (i>3 && rfabs(d.x)+g==rfabs(d.x) && rfabs(d.y)+g==rfabs(d.y))
	  m12 =0.0;
	else if (rfabs(m12)>tresh)
	  {
	    h = d.y-d.x;
	    if (rfabs(h)+g == rfabs(h)) t=(m12)/h;
	    else
	      {
		theta=0.5*h/(m12);
		t=1.0/(rfabs(theta)+sqrt(1.0+theta*theta));
		if (theta < 0.0) t = -t;
	      }
	    c=1.0/sqrt(1+t*t); s=t*c; tau=s/(1.0+c); h=t*m12;
	    z.x -= h; z.y += h; d.x -= h; d.y += h;
	    m12=0.0;
	    rotate(m13,m23); rotate(v.m11,v.m12);
            rotate(v.m21,v.m22); rotate(v.m31,v.m32); 
	    nrot++;
	  }
      }

      // Try rotations in the 2nd dimension.
      {
	g = 100.0*rfabs(m13);
	// See above, can be simplified.
	if (i>3 && rfabs(d.x)+g==rfabs(d.x) && rfabs(d.z)+g==rfabs(d.z))
	  m13=0.0;
	else if (rfabs(m13)>tresh)
	  {
	    h = d.z-d.x;
	    if (rfabs(h)+g == rfabs(h)) t=(m13)/h;
	    else
	      {
		theta=0.5*h/(m13);
		t=1.0/(rfabs(theta)+sqrt(1.0+theta*theta));
		if (theta < 0.0) t = -t;
	      }
	    c=1.0/sqrt(1+t*t); s=t*c; tau=s/(1.0+c); h=t*m13;
	    z.x -= h; z.z += h; d.x -= h; d.z += h;
	    m13=0.0;
	    rotate(m12,m23); rotate(v.m11,v.m13);
            rotate(v.m21,v.m23); rotate(v.m31,v.m33); 
	    nrot++;
	  }
      }


      // Try rotations in 3rd dimension.
      {
	g = 100.0*rfabs(m23);
	if (i>3 && rfabs(d.y)+g==rfabs(d.y) && rfabs(d.z)+g==rfabs(d.z))
	  m23=0.0;
	else if (rfabs(m23)>tresh)
	  {
	    h = d.z-d.y;
	    if (rfabs(h)+g == rfabs(h)) t=(m23)/h;
	    else
	      {
		theta=0.5*h/(m23);
		t=1.0/(rfabs(theta)+sqrt(1.0+theta*theta));
		if (theta < 0.0) t = -t;
	      }
	    c=1.0/sqrt(1+t*t); s=t*c; tau=s/(1.0+c); h=t*m23;
	    z.y -= h; z.z += h; d.y -= h; d.z += h;
	    m23=0.0;
	    rotate(m12,m13); rotate(v.m12,v.m13);
            rotate(v.m22,v.m23); rotate(v.m32,v.m33); 
	    nrot++;
	  }
      }

      b = b + z; d = b; z.set( 0, 0, 0);      
    }

  return -1;
#undef rotate
#undef rfabs
}

// Find eigen vectors of matrix and sort them to have the largest
// eigen vector in the first column.
int Matrix3::eigens1(Matrix3 *evecs)
{
#define swap(a,b) { float t = a; a = b; b = t; }
  int n;
  Vector3 evals;

  n = eigen(evecs, &evals);

  if (evals.z > evals.x)
    {
      if (evals.z > evals.y)
	{
	  // 3 is largest, swap with column 1
	  swap(evecs->m13,evecs->m11);
	  swap(evecs->m23,evecs->m21);
	  swap(evecs->m33,evecs->m31);
	}
      else
	{
	  // 2 is largest, swap with column 1
	  swap(evecs->m12,evecs->m11);
	  swap(evecs->m22,evecs->m21);
	  swap(evecs->m32,evecs->m31);
	}
    }
  else
    {
      if (evals.x > evals.y)
	{
	  // 1 is largest, do nothing
	}
      else
	{
  	  // 2 is largest
	  swap(evecs->m12,evecs->m11);
	  swap(evecs->m22,evecs->m21);
	  swap(evecs->m32,evecs->m31);
	}
    }
  // we are returning the number of iterations Meigen took.
  // too many iterations means our chosen orientation is bad.
  return n; 
#undef swap
}


Matrix3 operator+ (const Matrix3& m1, const Matrix3& m2) 
{
  return Matrix3 (m1.m11+m2.m11, m1.m12+m2.m12, m1.m13+m2.m13,
                  m1.m21+m2.m21, m1.m22+m2.m22, m1.m23+m2.m23,
                  m1.m31+m2.m31, m1.m32+m2.m32, m1.m33+m2.m33);
}
                  
Matrix3 operator- (const Matrix3& m1, const Matrix3& m2)
{
  return Matrix3 (m1.m11-m2.m11, m1.m12-m2.m12, m1.m13-m2.m13,
                  m1.m21-m2.m21, m1.m22-m2.m22, m1.m23-m2.m23,
                  m1.m31-m2.m31, m1.m32-m2.m32, m1.m33-m2.m33);
}
Matrix3 operator* (const Matrix3& m1, const Matrix3& m2)
{
  return Matrix3 (
   m1.m11*m2.m11 + m1.m12*m2.m21 + m1.m13*m2.m31,
   m1.m11*m2.m12 + m1.m12*m2.m22 + m1.m13*m2.m32,
   m1.m11*m2.m13 + m1.m12*m2.m23 + m1.m13*m2.m33,
   m1.m21*m2.m11 + m1.m22*m2.m21 + m1.m23*m2.m31,
   m1.m21*m2.m12 + m1.m22*m2.m22 + m1.m23*m2.m32,
   m1.m21*m2.m13 + m1.m22*m2.m23 + m1.m23*m2.m33,
   m1.m31*m2.m11 + m1.m32*m2.m21 + m1.m33*m2.m31,
   m1.m31*m2.m12 + m1.m32*m2.m22 + m1.m33*m2.m32,
   m1.m31*m2.m13 + m1.m32*m2.m23 + m1.m33*m2.m33 );
}

Matrix3 operator* (const Matrix3& m, float f)
{
  return Matrix3 (m.m11*f, m.m12*f, m.m13*f,
                  m.m21*f, m.m22*f, m.m23*f,
                  m.m31*f, m.m32*f, m.m33*f);
}

Matrix3 operator* (float f, const Matrix3& m)
{
  return Matrix3 (m.m11*f, m.m12*f, m.m13*f,
                  m.m21*f, m.m22*f, m.m23*f,
                  m.m31*f, m.m32*f, m.m33*f);
}

Matrix3 operator/ (const Matrix3& m, float f)
{
  return Matrix3 (m.m11/f, m.m12/f, m.m13/f,
                  m.m21/f, m.m22/f, m.m23/f,
                  m.m31/f, m.m32/f, m.m33/f);
}

bool operator== (const Matrix3& m1, const Matrix3& m2)
{ 
  if (m1.m11 != m2.m11 || m1.m12 != m2.m12 || m1.m13 != m2.m13) return false;
  if (m1.m21 != m2.m21 || m1.m22 != m2.m22 || m1.m23 != m2.m23) return false;
  if (m1.m31 != m2.m31 || m1.m32 != m2.m32 || m1.m33 != m2.m33) return false;
  return true;
}

bool operator!= (const Matrix3& m1, const Matrix3& m2)
{
  if (m1.m11 != m2.m11 || m1.m12 != m2.m12 || m1.m13 != m2.m13) return true;
  if (m1.m21 != m2.m21 || m1.m22 != m2.m22 || m1.m23 != m2.m23) return true;
  if (m1.m31 != m2.m31 || m1.m32 != m2.m32 || m1.m33 != m2.m33) return true;
  return false;
}

bool operator< (const Matrix3& m, float f)
{
  return ABS(m.m11)<f && ABS(m.m12)<f && ABS(m.m13)<f &&
         ABS(m.m21)<f && ABS(m.m22)<f && ABS(m.m23)<f &&
         ABS(m.m31)<f && ABS(m.m32)<f && ABS(m.m33)<f;
}

bool operator> (float f, const Matrix3& m)
{
  return ABS(m.m11)<f && ABS(m.m12)<f && ABS(m.m13)<f &&
         ABS(m.m21)<f && ABS(m.m22)<f && ABS(m.m23)<f &&
         ABS(m.m31)<f && ABS(m.m32)<f && ABS(m.m33)<f;
}

//---------------------------------------------------------------------------

void Vector3::between (Vector3& v1, Vector3& v2, Vector3& v, float pct, float wid)
{
  if (pct != -1)
    pct /= 100.;
  else
  {
    float dist = sqrt((v1-v2)*(v1-v2));
    pct = wid/dist;
  }
  v = v1 + pct*(v2-v1);
}

bool Vector3::visible (const Vector3& t1, const Vector3& t2, const Vector3& t3)
const
{
   float x1 = t1.x-x;
   float y1 = t1.y-y;
   float z1 = t1.z-z;
   float x2 = t2.x-x;
   float y2 = t2.y-y;
   float z2 = t2.z-z;
   float x3 = t3.x-x;
   float y3 = t3.y-y;
   float z3 = t3.z-z;
   float c = x3*((z1*y2)-(y1*z2))+
           y3*((x1*z2)-(z1*x2))+
           z3*((y1*x2)-(x1*y2));
   return c > 0;
}

int Vector3::which_side_3d (Vector3& v1, Vector3& v2)
{
  float s = (*this) * (v1%v2);
  if (s < 0) return 1;
  else if (s > 0) return -1;
  else return 0;
}
//---------------------------------------------------------------------------

Transform Plane::get_reflect()
{
  // Suppose that n is the plane normal in the direction of th reflection.
  // Suppose that u is the unit vector in the direction of the reflection
  // normal.  For any vector v, the component of v in the direction of
  // u is equal to (v * u) * u.  Thus, if v is reflected across a plane
  // through the origin with the given normal, the resulting vector is
  //  v' = v - 2 * [ (v * u) * u ] = v - 2 [ (v * n) * n ] / (n * n)
  //
  // x = <1,0,0>  =>  x' = <1,0,0> - 2 ( n.x * n ) / (n*n)
  // y = <0,1,0>  =>  y' = <0,1,0> - 2 ( n.y * n ) / (n*n)
  // z = <0,0,1>  =>  z' = <0,0,1> - 2 ( n.z * n ) / (n*n)
  // 
  // 3x3 transformation matrix = [x' y' z']

  float i_normsq = 1/(norm*norm);
  Vector3 xvec = (-2 * norm.x * i_normsq) * norm;
  Vector3 yvec = (-2 * norm.y * i_normsq) * norm;
  Vector3 zvec = (-2 * norm.z * i_normsq) * norm;
  xvec.x+=1;  yvec.y+=1;  zvec.z+=1;

  return Transform( Matrix3 ( xvec.x, yvec.x, zvec.x,
                              xvec.y, yvec.y, zvec.y,
                              xvec.z, yvec.z, zvec.z ), 
            /* neworig = */ (-2 * DD * i_normsq ) * norm );
}

//---------------------------------------------------------------------------

Plane Transform::world2this(const Plane& p) const
{
  Vector3 newnorm = m_w2t*p.norm;
// let N represent norm <A,B,C>, and X represent point <x,y,z>
//
// Old plane equation: N*X + D = 0
// There exists point X = <r*A,r*B,r*C> = r*N which satisfies the
// plane equation.  
//  => r*(N*N) + D = 0
//  => r = -D/(N*N)
//
// New plane equation: N'*X' + D' = 0
// If M is the transformation matrix, and V the transformation vector,
// N' = M*N, and X' = M*(X-V).  Assume that N' is already calculated.
//  => N'*(M*(X-V)) + D' = 0
//  => D' = -N'*(M*X) + N'*(M*V)
//        = -N'*(M*(r*N)) + N'*(M*V)
//        = -r*(N'*N') + N'*(M*V) = D*(N'*N')/(N*N) + N'*(M*V)
// Since N' is a rotation of N, (N'*N') = (N*N), thus
//  D' = D + N'*(M*V)
//
  return Plane(newnorm, p.DD + newnorm * (m_w2t*v_w2t) );
}

void Transform::world2this(const Plane& p, const Vector3& point,
                           Plane& result) const
{
  result.norm = m_w2t*p.norm;
  result.DD = - (result.norm * point);
}

Vector3 operator* (const Vector3& v, const Transform& t)
{ return t.world2this(v); }

Vector3 operator* (const Transform& t, const Vector3& v)
{ return t.world2this(v); }

Vector3& operator*= (Vector3& v, const Transform& t)
{ v = t.world2this(v); return v; }

Plane operator* (const Plane& p, const Transform& t)
{ return t.world2this(p); }

Plane operator* (const Transform& t, const Plane& p)
{ return t.world2this(p); }

Plane& operator*= (Plane& p, const Transform& t)
{
  p.norm = t.m_w2t * p.norm;
  p.DD += p.norm * (t.m_w2t*t.v_w2t);
  return p;
}

Matrix3 operator* (const Matrix3& m, const Transform& t)
{ return m*t.m_w2t; }

Matrix3 operator* (const Transform& t, const Matrix3& m) 
{ return t.m_w2t*m; }

Matrix3& operator*= (Matrix3& m, const Transform& t) 
{ return m *= t.m_w2t; }

//---------------------------------------------------------------------------

Plane ReversibleTransform::this2world(const Plane& p) const
{
  Vector3 newnorm = m_t2w * p.norm;
  return Plane(newnorm, p.DD - p.norm * (m_w2t*v_w2t) );
}

void ReversibleTransform::this2world(const Plane& p, const Vector3& point,
                           Plane& result) const
{
  result.norm = m_t2w*p.norm;
  result.DD = - (result.norm * point);
}

Vector3 operator/ (const Vector3& v, const ReversibleTransform& t) 
{ return t.this2world(v); }

Vector3& operator/= (Vector3& v, const ReversibleTransform& t) 
{ v = t.this2world(v); return v; }

Plane operator/ (const Plane& p, const ReversibleTransform& t)
{ return t.this2world(p); }

Plane& operator/= (Plane& p, const ReversibleTransform& t)
{
  p.DD -= p.norm * (t.m_w2t*t.v_w2t);
  p.norm = t.m_t2w * p.norm;
  return p;
}

ReversibleTransform& operator*= (ReversibleTransform& t1,
                                 const ReversibleTransform& t2)
{
  t1.v_w2t = t2.v_w2t + t2.m_t2w*t1.v_w2t;
  t1.m_w2t *= t2.m_w2t;
  t1.m_t2w = t2.m_t2w*t1.m_t2w;
  return t1;
}

ReversibleTransform operator* (const ReversibleTransform& t1,
                               const ReversibleTransform& t2)
{
  return ReversibleTransform(t1.m_w2t*t2.m_w2t, t2.m_t2w*t1.m_t2w, 
                             t2.v_w2t + t2.m_t2w*t1.v_w2t); 
}

Transform operator* (const Transform& t1, const ReversibleTransform& t2)
{
  return Transform(t1.m_w2t*t2.m_w2t, t2.v_w2t + t2.m_t2w*t1.v_w2t);
}

ReversibleTransform& operator/=(ReversibleTransform& t1,
                                const ReversibleTransform& t2)
{
 t1.v_w2t = t2.m_w2t*(t1.v_w2t - t2.v_w2t);
 t1.m_w2t *= t2.m_t2w;
 t1.m_t2w = t2.m_w2t*t1.m_t2w;
 return t1;
}

ReversibleTransform operator/ (const ReversibleTransform& t1,
                               const ReversibleTransform& t2)
{
  return ReversibleTransform(t1.m_w2t*t2.m_t2w, t2.m_w2t*t1.m_t2w, 
                   t2.m_w2t*(t1.v_w2t - t2.v_w2t)); 
}

//---------------------------------------------------------------------------

float SqDist::point_line (const Vector3& p, 
                           const Vector3& l1, const Vector3& l2)
{
  Vector3 W = l1-p;
  Vector3 L = l2-l1;
  Vector3 p2l = W - L * (W*L)/(L*L);
  return p2l * p2l;
}

float SqDist::point_poly (const Vector3& p, Vector3 *V, int n, 
                          const Plane& plane, float sqdist)
{
  Vector3 W, L;
  bool lflag = true, lflag0 = true;
  for (int i=0; i<n-1; i++)
  {
    W = V[i] - p;
    if (i==0)
    {
      if ( !(W*(V[n-1]-V[0]) > 0) ) lflag0 = false;
      else if (W*(V[1]-V[0]) > 0) return W*W;
      else lflag = false;
    }
    else if ( !(W*(L = V[i-1]-V[i]) > 0) )
    {
      if ( !lflag && W*(plane.norm % L) > 0 )
      { 
        L = W - L * (W*L)/(L*L);  
        return L*L; 
      }
      lflag = (W*(V[i+1]-V[i]) > 0);
    }
    else if (W*(V[i+1]-V[i]) > 0) return W*W;
    else lflag = false;
  }

  W = V[n-1] - p;
  if (!lflag)
  {
    lflag = W * (L = V[n-2]-V[n-1]) <= 0;
    if ( lflag && (W*(plane.norm % L) > 0) )
    {
      L = W - L * (W*L)/(L*L);
      return L*L;
    }
  }
  if (!lflag0)
  {
    lflag0 = W * (L = V[0]-V[n-1]) <= 0;
    if ( lflag0 && (W*(plane.norm % L) < 0) )
    {
      L = W - L * (W*L)/(L*L);
      return L*L;
    }
  }

  if (!lflag && !lflag0) return W*W;
  if (sqdist >= 0) return sqdist;
  return SqDist::point_plane(p,plane);
}

//---------------------------------------------------------------------------

void Intersect3::plane(const Vector3& u, const Vector3& v,
                       const Vector3& normal, const Vector3& a,
                       Vector3& isect)
{
  float counter = normal * (u - a);
  float divider = normal * (v - u);
  float dist;
  if (divider == 0) { isect = v; return; }
  dist = counter / divider;
  isect = u + dist*(u - v);
}

bool Intersect3::plane(const Vector3& u, const Vector3& v,
                       float A, float B, float C, float D,
                       Vector3& isect, float& dist)
{
  float x,y,z, denom;

  x = v.x-u.x;  y = v.y-u.y;  z = v.z-u.z;
  denom = A*x + B*y + C*z;
  if (ABS (denom) < SMALL_EPSILON) return false; // they are parallel

  dist = -(A*u.x + B*u.y + C*u.z + D) / denom;
  if (dist < -SMALL_EPSILON || dist > 1+SMALL_EPSILON) return false;

  isect.x = u.x + dist*x;  isect.y = u.y + dist*y;  isect.z = u.z + dist*z;
  return true;
}

bool Intersect3::plane(const Vector3& u, const Vector3& v,
                       const Plane& p,
                       Vector3& isect, float& dist)
{
  float x,y,z, denom;

  x = v.x-u.x;  y = v.y-u.y;  z = v.z-u.z;
  denom = p.norm.x*x + p.norm.y*y + p.norm.z*z;
  if (ABS (denom) < SMALL_EPSILON) return false; // they are parallel

  dist = -(p.norm*u + p.DD) / denom;
  if (dist < -SMALL_EPSILON || dist > 1+SMALL_EPSILON) return false;

  isect.x = u.x + dist*x;  isect.y = u.y + dist*y;  isect.z = u.z + dist*z;
  return true;
}

float Intersect3::z0_plane(
  const Vector3& u, const Vector3& v, Vector3& isect)
{
  float r = u.z / (u.z-v.z);
  isect.x = r * (v.x-u.x) + u.x;
  isect.y = r * (v.y-u.y) + u.y;
  isect.z = 0;
  return r;
} 

float Intersect3::z_plane(
  float zval, const Vector3& u, const Vector3& v, Vector3& isect)
{
  float r = (zval-u.z) / (v.z-u.z);
  isect.x = r * (v.x-u.x) + u.x;
  isect.y = r * (v.y-u.y) + u.y;
  isect.z = zval;
  return r;
}

float Intersect3::x_frustrum(
  float A, const Vector3& u, const Vector3& v, Vector3& isect)
{
  float r = (A*u.x+u.z) / ( A*(u.x-v.x) + u.z-v.z );
  isect.x = r * (v.x-u.x) + u.x;
  isect.y = r * (v.y-u.y) + u.y;
  isect.z = r * (v.z-u.z) + u.z;
  return r;
}

float Intersect3::y_frustrum(
  float B, const Vector3& u, const Vector3& v, Vector3& isect)
{
  float r = (B*u.y+u.z) / ( B*(u.y-v.y) + u.z-v.z );
  isect.x = r * (v.x-u.x) + u.x;
  isect.y = r * (v.y-u.y) + u.y;
  isect.z = r * (v.z-u.z) + u.z;
  return r;
}

//---------------------------------------------------------------------------
