/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Largely rewritten by Ivan Avramovic <iavramov@ibm.net>
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef MATH3D_H
#define MATH3D_H

#include "util/def.h"

class Vector3;
class Matrix3;
class Transform;
class ReversibleTransform;
class Plane;

/**
 * This class represents all precalculated tables that are needed.
 * Currently there are only some tables for FltLights.
 * Note: all angle measures in the tables are in degrees, NOT radians
 */
class Tables
{
public:
  /// Sinus table.
  float sin_tab[360*2+1];
  /// Cosinus table.
  float cos_tab[360*2+1];
  /// Arc-cosinus table.
  int acos_tab[512*2+1];
  /// Arc-sinus table.
  int asin_tab[512*2+1];
  /**
   * Combined cos(asin(x)) table,
   * which equals sin(acos(x)), and sqrt(1-x*x)
   */
  long cos_asin_tab[512*2+1];	// F20:12
  /**
   * Exponent table used for fogging.
   * Index is dist*density in 8:8 fixed point format.
   * Result is 1-exp(-dist*density) between 0 and 255.
   */
  UByte exp_table[256*256];

  /**
   * Compute all of the values of the tables.
   * Note! this cannot be done in a constructor because
   * the constructor is not called for global objects put
   * in a shared object.
   */
  void initialize ();
};

extern Tables tables;

/**
 * A 3D vector.
 */
class Vector3
{
public:
  ///
  float x;
  ///
  float y;
  ///
  float z;

  /// Make a new vector, initialized to the zero vector.
  Vector3 () { x = y = z = 0; }
  /// Make a new vector and initialize with the given values.
  Vector3 (float x, float y, float z = 0) 
   { Vector3::x = x; Vector3::y = y; Vector3::z = z; }

  /// Add another vector to this vector.
  inline Vector3& Vector3::operator+= (const Vector3& v)
  {
  #ifdef COMP_VC
    __asm
    {
	mov		eax, this
	mov		ebx, v
    
	fld		[eax]this.x
	fadd	[ebx]v.x
	fld		[eax]this.y
	fadd	[ebx]v.y
	fld		[eax]this.z
	fadd	[ebx]v.z

	fxch	st(2)

	fstp	[eax]this.x
	fstp	[eax]this.y
	fstp	[eax]this.z
    }
  #else
    x += v.x;
    y += v.y;
    z += v.z;
  #endif
  return *this;
  }

    /// Subtract another vector from this vector.
  inline Vector3& Vector3::operator-= (const Vector3& v)
  {
  #ifdef COMP_VC
    __asm
    {
	mov		eax, this
	mov		ebx, v
    
	fld		[eax]this.x
	fsub	[ebx]v.x
	fld		[eax]this.y
	fsub	[ebx]v.y
	fld		[eax]this.z
	fsub	[ebx]v.z

	fxch	st(2)

	fstp	[eax]this.x
	fstp	[eax]this.y
	fstp	[eax]this.z
    }
  #else
    x -= v.x;
    y -= v.y;
    z -= v.z;
  #endif
  return *this;
  }

  /// Multiply this vector by a scalar.
  inline Vector3& operator*= (float f) { x *= f; y *= f; z *= f; return *this; }
  /// Divide this vector by a scalar.
  inline Vector3& operator/= (float f) { x /= f; y /= f; z /= f; return *this; }

  ///
  inline Vector3 operator+ () const { return *this; }
  ///
  inline Vector3 operator- () const { return Vector3(-x,-y,-z); }

  /// Set the value of this vector.
  inline void set (float sx, float sy, float sz) { x = sx; y = sy; z = sz; }

  /// Return the squared distance between two vectors.
  static float sqdist (const Vector3& v1, const Vector3& v2)
  { return (v1-v2)*(v1-v2); }

  /// Returns the norm (magnitude) of a vector
  inline static float Vector3::norm (const Vector3& v) 
  { return sqrt(v*v); }

  /// Normalizes a vector to a unit vector
  static Vector3 unit (const Vector3& v) { return v/norm(v); }

  /// Returns the norm of this vector
  inline float Vector3::norm () const
  { return sqrt(x*x+y*y+z*z); }

  /// Returns the unit vector in the direction of this vector
  Vector3 unit() const { return (*this)/norm(); }

  /**
   * Return -1 if vector is left of plane '0-v1-v2'. 1 is vector is right of 
   * '0-v1-v2' or 0 is vector is on '0-v1-v2'. (0-v1-v2 is the plane through
   * 0, v1, and v2).
   */
  int which_side_3d (Vector3& v1, Vector3& v2);

  /**
   * Visibility test (backface culling) to see if the triangle
   * t1, t2, t3 s visible from this point.
   */
  bool visible (const Vector3& t1, const Vector3& t2, const Vector3& t3) const;

  /**
   * Compute the vector in v which is between v1 and v2 at the
   * given percentage or fixed width.
   */
  static void between (Vector3& v1, Vector3& v2, Vector3& v, float pct, float wid);

  /// add two vectors
  inline friend Vector3 operator+ (const Vector3& v1, const Vector3& v2)
  { return Vector3(v1.x+v2.x, v1.y+v2.y, v1.z+v2.z); }
  /// subtract two vectors
  inline friend Vector3 operator- (const Vector3& v1, const Vector3& v2)
  { return Vector3(v1.x-v2.x, v1.y-v2.y, v1.z-v2.z); }
  /// take the dot product of two vectors
  inline friend float operator* (const Vector3& v1, const Vector3& v2)
  { return v1.x*v2.x + v1.y*v2.y + v1.z*v2.z; }
  /// take the cross product of two vectors
  inline friend Vector3 operator% (const Vector3& v1, const Vector3& v2)
  { 
    return Vector3 (v1.y*v2.z-v1.z*v2.y,
                    v1.z*v2.x-v1.x*v2.z,
                    v1.x*v2.y-v1.y*v2.x);
  }
  /// multiply a vector and a scalar
  inline friend Vector3 operator* (const Vector3& v, float f)
  { return Vector3(v.x*f, v.y*f, v.z*f); }
  /// multiply a vector and a scalar
  inline friend Vector3 operator* (float f, const Vector3& v)
  { return Vector3(v.x*f, v.y*f, v.z*f); }
  /// divide a vector by a scalar
  inline friend Vector3 operator/ (const Vector3& v, float f)
  { f = 1.0/f; return Vector3(v.x*f, v.y*f, v.z*f); }
  /// check if two vectors are equal
  inline friend bool operator== (const Vector3& v1, const Vector3& v2)
  { return v1.x==v2.x && v1.y==v2.y && v1.z==v2.z; }
  /// check if two vectors are not equal
  inline friend bool operator!= (const Vector3& v1, const Vector3& v2)
  { return v1.x!=v2.x || v1.y!=v2.y || v1.z!=v2.z; }
  /// project one vector onto another
  inline friend Vector3 operator>> (const Vector3& v1, const Vector3& v2)
  { return v2*(v1*v2)/(v2*v2); }
  /// project one vector onto another
  inline friend Vector3 operator<< (const Vector3& v1, const Vector3& v2)
  { return v1*(v1*v2)/(v1*v1); }
  /// test if each component of a vector is less than a small epsilon value
  inline friend bool operator< (const Vector3& v, float f)
  { return ABS(v.x)<f && ABS(v.y)<f && ABS(v.z)<f; }
  /// test if each component of a vector is greater than a small epsilon value
  inline friend bool operator> (float f, const Vector3& v)
  { return ABS(v.x)<f && ABS(v.y)<f && ABS(v.z)<f; }
  // Set the min and max vector if this vector exceeds their current limits.
  void setminmax(Vector3 *min, Vector3 *max)
  {
  if (x > max->x) max->x = x; else if (x < min->x ) min->x = x;
  if (y > max->y) max->y = y; else if (y < min->y ) min->y = y;
  if (z > max->z) max->z = z; else if (z < min->z ) min->z = z;
  } 
};

/**
 * A 3x3 matrix.
 */
class Matrix3
{
public:
  float m11, m12, m13;
  float m21, m22, m23;
  float m31, m32, m33;

public:
  /// Construct a matrix, initialized to be the identity.
  Matrix3 ();

  /// Construct a matrix and initialize it.
  Matrix3 (float m11, float m12, float m13,
  	   float m21, float m22, float m23,
  	   float m31, float m32, float m33);

  inline Vector3 row1() const { return Vector3(m11,m12,m13); }
  inline Vector3 row2() const { return Vector3(m21,m22,m23); }
  inline Vector3 row3() const { return Vector3(m31,m32,m33); }
  inline Vector3 col1() const { return Vector3(m11,m21,m31); }
  inline Vector3 col2() const { return Vector3(m12,m22,m32); }
  inline Vector3 col3() const { return Vector3(m13,m23,m33); }

  /// Set matrix values
  inline void set (float m11, float m12, float m13,
                   float m21, float m22, float m23,
                   float m31, float m32, float m33)
  {
    Matrix3::m11 = m11; Matrix3::m12 = m12; Matrix3::m13 = m13;
    Matrix3::m21 = m21; Matrix3::m22 = m22; Matrix3::m23 = m23;
    Matrix3::m31 = m31; Matrix3::m32 = m32; Matrix3::m33 = m33;
  }

  /// Add another matrix to this matrix.
  Matrix3& operator+= (const Matrix3& m);

  /// Subtract another matrix from this matrix.
  Matrix3& operator-= (const Matrix3& m);

  /// Multiply another matrix with this matrix.
  Matrix3& operator*= (const Matrix3& m);

  /// Multiply this matrix with a scalar.
  Matrix3& operator*= (float s);

  /// Divide this matrix by a scalar.
  Matrix3& operator/= (float s);

  ///
  inline Matrix3 operator+ () const { return *this; }
  ///
  inline Matrix3 operator- () const
  { 
   return Matrix3(-m11,-m12,-m13,
                  -m21,-m22,-m23,
                  -m31,-m32,-m33);
  }

  /// Transpose this matrix.
  void transpose ();

  /// Return the transpose of this matrix
  Matrix3 gettranspose () const;

  /// Return the inverse of this matrix
  inline Matrix3 Matrix3::getinverse () const
  {
    float s;
    Matrix3 C(
             (m22*m33 - m23*m32),
            -(m12*m33 - m13*m32),
             (m12*m23 - m13*m22),
            -(m21*m33 - m23*m31),
             (m11*m33 - m13*m31),
            -(m11*m23 - m13*m21),
             (m21*m32 - m22*m31),
            -(m11*m32 - m12*m31),
             (m11*m22 - m12*m21) );
    s     =  (float)1./(m11*C.m11 + m12*C.m21 + m13*C.m31);

  #ifdef COMP_VC
    __asm
    {
	  fld	C.m11
	  fmul	s
	  fld	C.m12
	  fmul	s
	  fld	C.m13
	  fmul	s

	  fxch	st(2)

	  fstp	C.m11
	  fstp	C.m12
	  fstp	C.m13

	  fld	C.m21
	  fmul	s
	  fld	C.m22
	  fmul	s
	  fld	C.m23
	  fmul	s

	  fxch	st(2)

	  fstp	C.m21
	  fstp	C.m22
	  fstp	C.m23

	  fld	C.m31
	  fmul	s
	  fld	C.m32
	  fmul	s
	  fld	C.m33
	  fmul	s

	  fxch	st(2)

	  fstp	C.m31
	  fstp	C.m32
	  fstp	C.m33
    }
  #else
    C    *=  s;
  #endif
  return C;
}

  /// Invert this matrix.
  void invert() { *this = getinverse(); }

  /// Compute the determinant of this matrix.
  float determinant () const;

  /// Compute the eigen values and eigen vectors of this matrix.
  int eigen(Matrix3* evectors, Vector3* evalues);

  /// Compute the eigen vectors return largest one in 1st column.
  int eigens1(Matrix3 *evecs);

  /// Set this matrix to the identity matrix.
  void identity ();

  /// add two matricies
  friend Matrix3 operator+ (const Matrix3& m1, const Matrix3& m2);
  /// subtract two matricies
  friend Matrix3 operator- (const Matrix3& m1, const Matrix3& m2);
  /// multiply two matricies
  friend Matrix3 operator* (const Matrix3& m1, const Matrix3& m2);
  /// multiply a vector by a matrix (transform it)
//  friend Vector3 operator* (const Matrix3& m, const Vector3& v);
  inline friend Vector3 operator* (const Matrix3& m, const Vector3& v)
  {
   return Vector3(m.m11*v.x + m.m12*v.y + m.m13*v.z,
                  m.m21*v.x + m.m22*v.y + m.m23*v.z,
                  m.m31*v.x + m.m32*v.y + m.m33*v.z);
  }

  
  
  /// multiply a matrix and a scalar
  friend Matrix3 operator* (const Matrix3& m, float f);
  /// multiply a matrix and a scalar
  friend Matrix3 operator* (float f, const Matrix3& m);
  /// divide a matrix by a scalar
  friend Matrix3 operator/ (const Matrix3& m, float f);
  /// check if two matricies are equal
  friend bool operator== (const Matrix3& m1, const Matrix3& m2);
  /// check if two matricies are not equal
  friend bool operator!= (const Matrix3& m1, const Matrix3& m2);
  /// test if each component of a matrix is less than a small epsilon value
  friend bool operator< (const Matrix3& m, float f);
  /// test if each component of a matrix is greater than a small epsilon value
  friend bool operator> (float f, const Matrix3& m);

};

/**
 * A plane in 3D space.
 */
class Plane
{
public:
  /// The normal vector (or the (A,B,C) components).
  Vector3 norm;

  /// The D component of the plane.
  float DD;

public:

  /// Initializes to the xy plane
  Plane() : norm(0,0,1), DD(0) {}

  /// Initializes the plane
  Plane(const Vector3& plane_norm, float d=0) : norm(plane_norm), DD(d) {}

  /// Initializes the plane
  Plane(float a, float b, float c, float d=0) : norm(a,b,c), DD(d) {}

  /// Calculate a plane normal given three vectors.
  inline static Vector3 Plane::calc_plane_normal(const Vector3& v1, const Vector3& v2,
           const Vector3& v3) 
  {
  #ifdef COMP_VC
    float x1, y1, z1, x2, y2, z2;

/*
   operator% :

   v1.y*v2.z-v1.z*v2.y,
   v1.z*v2.x-v1.x*v2.z,
   v1.x*v2.y-v1.y*v2.x
*/
    __asm
    {
	mov		eax, v1
	mov		ebx, v2
	mov		ecx, v3

	; v1-v2
	fld		[eax]v1.x
	fsub	[ebx]v2.x
	fld		[eax]v1.y
	fsub	[ebx]v2.y
	fld		[eax]v1.z
	fsub	[ebx]v2.z

	; Save result
	fxch	st(2)
	fst		x1
	fxch	st(1)
	fst		y1
	fxch	st(2)
	fst		z1

    ; v1-v3
	fld		[eax]v1.z
	fsub	[ecx]v3.z
	fld		[eax]v1.y
	fsub	[ecx]v3.y
	fld		[eax]v1.x
	fsub	[ecx]v3.x
	fxch	st(2)
    
	fst		z2
	fmulp	st(5),st(0)
	fst		y2
	fmulp	st(3),st(0)
	fst		x2
	fmulp	st(1),st(0)

; Floating point stack:
;
; v1.z*v2.x
; v1.x*v2.y
; v1.y*v2.z

    fld		z1
	fmul	y2
	fld		y1
	fmul	x2
	fld		x1
	fmul	z2
	fxch	st(2)

	fsubp	st(5),st(0)
	fsubp	st(3),st(0)
	fsubp	st(1),st(0)
	fxch	st(2)

	fstp	x1
	fstp	z1
	fstp	y1
    }
    return Vector3( x1, y1, z1 );
  #else
    return (v1-v2)%(v1-v3);
  #endif
}

  /// Calculate the plane equation given three vectors.
  static void calc_plane (const Vector3& v1, const Vector3& v2, 
         const Vector3& v3, Vector3& normal, float& D)
  {
    normal = (v1-v2)%(v1-v3);
    D = - (normal * v1);
  }

  /// Classify a vector with regards to four plane components.
  static float classify (float A, float B, float C, float D, const Vector3& pt)
  { return A*pt.x + B*pt.y + C*pt.z + D; }

  /// Compute the normal given two (u,v) vectors.
  static void normal (const Vector3& v, const Vector3& u, Vector3& norm)
  { norm = u%v; /* NOT v%u - vertexes are defined clockwise */ }

  /// Classify the given vector with regards to this plane.
  inline float classify (const Vector3& pt) const { return norm*pt+DD; }

  /// Compute the distance from the given vector to this plane.
  inline float distance (const Vector3& pt) const 
  { return ABS (classify (pt)); }

  /// Reverses the direction of the plane while maintianing the plane itself.
  void invert() { norm = -norm;  DD = -DD; }
  
  /**
   * Check if the plane is visible from the given point.
   * To do this it will only do back-face culling.
   */
  bool visible_from_point (const Vector3& p) const
  { return classify(p) <= 0; }

  /// Create a transform which represents reflection across this plane.
  Transform get_reflect();

  /// Return the normal vector of this plane.
  inline Vector3& normal () { return norm; }

  /// Return the A component of this plane.
  inline float& A () { return norm.x; }
  /// Return the B component of this plane.
  inline float& B () { return norm.y; }
  /// Return the C component of this plane.
  inline float& C () { return norm.z; }
  /// Return the D component of this plane.
  inline float& D () { return DD; }

  /// Set the value of the four plane components.
  inline void set (float a, float b, float c, float d)
   { norm.x = a; norm.y = b; norm.z = c; DD = d; }

  /// Check if two planes are close together.
  bool close (const Plane& p) const
  { return ( (p.norm - norm) < (float).001 ) && ( ABS (p.DD-DD) < (float).001 ); }

  friend class Transform;
};

/**
 * A class which defines a transformation from one coordinate system to
 * another.
 */
class Transform
{
protected:
  /// Transformation matrix from world space to local space
  Matrix3 m_w2t;
  /// Location of the origin for the local transformation
  Vector3 v_w2t;

public:
  /// Initialize with the identity transformation
  Transform() : m_w2t(), v_w2t() {}
  /// Initialize with the given transformation
  Transform(const Matrix3& world2this, const Vector3& origin_pos) :
   m_w2t(world2this), v_w2t(origin_pos) {}

  /// Get world to local transformation matrix
  inline Matrix3 m_world2this() const { return m_w2t; }
  /// Get world to local translation
  inline Vector3 v_world2this() const { return v_w2t; }
  /// Get origin of transformed coordinate system
  inline Vector3 v_origin() const { return v_w2t; }

  /// Set world to local transformation matrix
  virtual void set_world2this(const Matrix3& m) { m_w2t = m; }
  /// Set world to local translation
  inline void set_origin(const Vector3& v) { v_w2t = v; }

  /// Move the world to local translation by a specified amount
  inline void translate(const Vector3& v) { v_w2t += v; }

  /// Convert worldspace vector v to local space
  inline Vector3 Transform::world2this(const Vector3& v) const
  {
  #ifdef COMP_VC
	float x,y,z;
	const Vector3 &v2 = v_w2t;
	const Matrix3 &m  = m_w2t;

    __asm
    {
	mov		eax, v
	mov		ecx, v2
	mov		ebx, m
	fld		[eax]v.x
	fsub	[ecx]v2.x
	fld		[eax]v.y
	fsub	[ecx]v2.y
	fld		[eax]v.z
	fsub	[ecx]v2.z
	fxch	st(2)

; Floating point stack:
;
; x
; y
; z
	fst		x
	fmul	[ebx]m.m21
	fld		x
	fmul	[ebx]m.m31
	fld		x
    fmul    [ebx]m.m11
	fxch    st(3)

; Floating point stack:
;
; y
; m31*x
; m21*x
; m11*x
; z
	fst		y
	fmul	[ebx]m.m12
	fld		y
	fmul	[ebx]m.m22
	fld		y
	fmul	[ebx]m.m32
	fxch	st(2)
	faddp	st(5),st(0)
	faddp	st(3),st(0)
	faddp	st(1),st(0)
	fxch	st(3)

; Floating point stack:
;
; z
; m21*x+m22*y
; m11*x+m12*y
; m31*x+m32*y
	fst		z
	fmul	[ebx]m.m33
	fld		z
	fmul	[ebx]m.m13
	fld		z
	fmul	[ebx]m.m23
	fxch	st(2)
	faddp	st(5),st(0)
	faddp	st(3),st(0)
	faddp	st(1),st(0)
	fxch	st(2)

	fstp	z
	fstp	x
	fstp	y
    }
    return Vector3(x, y, z);
  #else
    return m_w2t * (v - v_w2t);
  #endif
}

  /// Convert worldspace vector v to local space, using worldspace origin
  Vector3 world2this_rel(const Vector3& v) const
  { return m_w2t * v; }

  /// Convert a worldspace plane to local coordinates
  Plane world2this(const Plane& p) const;

  /**
   * Convert a plane to local coordinates.  This is an optimized version
   * for which a point on the new plane, 'point', is known.  The result
   * is stored in 'result'.
   */
  void world2this(const Plane& p, const Vector3& point, Plane& result) const;

  /// Apply a transformation to a 3D vector
  friend Vector3 operator* (const Vector3& v, const Transform& t);
  /// Apply a transformation to a 3D vector
  friend Vector3 operator* (const Transform& t, const Vector3& v);
  /// Apply a transformation to a 3D vector
  friend Vector3& operator*= (Vector3& v, const Transform& t);
  /// Apply a transformation to a Plane
  friend Plane operator* (const Plane& p, const Transform& t);
  /// Apply a transformation to a Plane
  friend Plane operator* (const Transform& t, const Plane& p);
  /// Apply a transformation to a Plane
  friend Plane& operator*= (Plane& p, const Transform& t);
  /// Multiply a matrix with the transformation matrix
  friend Matrix3 operator* (const Matrix3& m, const Transform& t);
  /// Multiply a matrix with the transformation matrix
  friend Matrix3 operator* (const Transform& t, const Matrix3& m);
  /// Multiply a matrix with the transformation matrix
  friend Matrix3& operator*= (Matrix3& m, const Transform& t);
  /// Combine two transforms, rightmost first
  friend Transform operator* (const Transform& t1,
                              const ReversibleTransform& t2);
};

/**
 * A class which defines a reversible transformation from one coordinate
 * system to another by maintaining an inverse transformation matrix.
 */
class ReversibleTransform : public Transform
{
protected:
  /// Inverse transformation matrix (local to world space).
  Matrix3 m_t2w;
  
  /// Initialize transform with both transform matrix and inverse tranform
  ReversibleTransform(const Matrix3& w2t, const Matrix3& t2w, 
    const Vector3& pos) : Transform(w2t,pos), m_t2w(t2w) {}
public:
  /// Initialize with the identity transformation
  ReversibleTransform() : Transform(), m_t2w() {}
  /// Initialize with the given transformation 
  ReversibleTransform(const Matrix3& w2t, const Vector3& pos) :
   Transform(w2t,pos) { m_t2w = m_w2t.getinverse(); }
  /// Initialize with the given transformation 
  ReversibleTransform(const Transform& t) :
   Transform(t) { m_t2w = m_w2t.getinverse(); }

  /// Get local to world transformation matrix
  inline Matrix3 m_this2world() const { return m_t2w; }
  /// Get local to world translation
  inline Vector3 v_this2world() const { return -m_w2t*v_w2t; }

  /// Get the inverse of this transform
  ReversibleTransform getinverse() const 
  { return ReversibleTransform(m_t2w, m_w2t, -m_w2t*v_w2t); } 

  /// Set world to local transformation matrix
  virtual void set_world2this(const Matrix3& m) 
  { m_w2t = m;  m_t2w = m_w2t.getinverse(); }
  /// Set local to world transformation matrix
  virtual void set_this2world(const Matrix3& m) 
  { m_t2w = m;  m_w2t = m_t2w.getinverse(); }

  /// Convert local vector v to worldspace
  Vector3 this2world(const Vector3& v) const
  { return v_w2t + m_t2w * v; }

  /// Convert local vector v to worldspace, relative to local origin
  inline Vector3 this2world_rel(const Vector3& v) const
  { return m_t2w * v; }

  /// Convert a local plane to worldspace
  Plane this2world(const Plane& p) const;

  /**
   * Convert a local plane to worldspace, given a point on the new plane
   * for which a point on the new plane, 'point', is known.  The result
   * is stored in 'result'.
   */
  void this2world(const Plane& p, const Vector3& point, Plane& result) const;

  /// Reverse a transformation on a 3D vector
  friend Vector3 operator/ (const Vector3& v, const ReversibleTransform& t); 
  /// Reverse a transformation on a 3D vector
  friend Vector3& operator/= (Vector3& v, const ReversibleTransform& t); 
  /// Reverse a transformation on a Plane
  friend Plane operator/ (const Plane& p, const ReversibleTransform& t);
  /// Reverse a transformation on a Plane
  friend Plane& operator/= (Plane& p, const ReversibleTransform& t);
  /// Combine two transforms, with the rightmost being applied first
  friend ReversibleTransform& operator*= (ReversibleTransform& t1,
                                          const ReversibleTransform& t2);
  /// Combine two transforms, with the rightmost being applied first
  friend ReversibleTransform operator* (const ReversibleTransform& t1,
                                        const ReversibleTransform& t2);
  /// Combine two transforms, with the rightmost being applied first
  friend Transform operator* (const Transform& t1, 
                              const ReversibleTransform& t2);
  /// Combine two transforms, reversing t2 then applying t1
  friend ReversibleTransform& operator/= (ReversibleTransform& t1,
                                          const ReversibleTransform& t2);
  /// Combine two transforms, reversing t2 then applying t1
  friend ReversibleTransform operator/ (const ReversibleTransform& t1,
                                        const ReversibleTransform& t2);
};

/**
 * A class which defines a reversible transformation from one coordinate
 * system to another by maintaining an inverse transformation matrix.
 * This is a variant which only works on orthonormal transformations (like
 * the camera transformation) and is consequently much more optimal.
 */
class OrthoTransform : public ReversibleTransform
{
public:
  /// Initialize with the identity transformation
  OrthoTransform() : ReversibleTransform() {}
  /// Initialize with the given transformation 
  OrthoTransform(const Matrix3& w2t, const Vector3& pos) :
   ReversibleTransform (w2t, w2t.gettranspose (), pos) { }
  /// Initialize with the given transformation 
  OrthoTransform(const Transform& t) :
   ReversibleTransform (t.m_world2this (), t.m_world2this ().gettranspose (), t.v_world2this ()) { }

  /// Set world to local transformation matrix
  virtual void set_world2this(const Matrix3& m) 
  { m_w2t = m;  m_t2w = m_w2t.gettranspose(); }
  /// Set local to world transformation matrix
  virtual void set_this2world(const Matrix3& m) 
  { m_t2w = m;  m_w2t = m_t2w.gettranspose(); }
};


/**
 * Some functions to perform squared distance calculations.
 * This is a static class and contains only static member functions.
 */
class SqDist
{
public:
  /// Returns the squared distance between two points.
  static float point_point (const Vector3& p1, const Vector3& p2)
  {  Vector3 v = p1-p2;  return v*v; }

  /// Returns the squared distance between a point and a line.
  static float point_line (const Vector3& p, 
                           const Vector3& l1, const Vector3& l2);

  /// Returns the squared distance between a point and a normalized plane.
  static float point_plane (const Vector3& p, const Plane& plane)
  { float r = plane.classify(p);  return r*r; }

  /**
   * Returns the squared distance between a point and a polygon.
   * If sqdist is >= 0, then it is used as the pre-calculated point to
   * plane distance.  V is an array of vertices, n is the number of
   * vertices, and plane is the polygon plane.
   */
  static float point_poly (const Vector3& p, Vector3 *V, int n, 
                           const Plane& plane, float sqdist = -1);
};

/**
 * Some functions to perform various intersection calculations with 3D
 * line segments.  This is a static class and contains only static member
 * functions.
 */
class Intersect3
{
public:
  /**
   * Intersect a 3D segment with a plane.  Returns true if there is an
   * interection, with the intersection point returned in isect.
   * The distance from u to the intersection point is returned in dist.
   */
  static void plane(
    const Vector3& u, const Vector3& v, // segment
    const Vector3& normal, const Vector3& a, // plane
    Vector3& isect);                    // intersection point

  /**
   * Intersect a 3D segment with a plane.  Returns true if there is an
   * interection, with the intersection point returned in isect.
   * The distance from u to the intersection point is returned in dist.
   */
  static bool plane(
    const Vector3& u, const Vector3& v, // segment
    float A, float B, float C, float D, // plane Ax+By+Cz+D=0
    Vector3& isect,                     // intersection point
    float& dist);                       // distance from u to isect

  /**
   * Intersect a 3D segment with a plane.  Returns true if there is an
   * interection, with the intersection point returned in isect.
   * The distance from u to the intersection point is returned in dist.
   */
  static bool plane(
    const Vector3& u, const Vector3& v, // segment
    const Plane& p,                     // plane Ax+By+Cz+D=0
    Vector3& isect,                     // intersection point
    float& dist);                       // distance from u to isect

  /**
   * Intersect a 3D segment with the z = 0 plane.  Assumes that there
   * is an intersection (fails if the segment is parallel to the plane),
   * and returns the distance from u to the intersection point.
   * The intersection point is returned in isect.
   */
  static float z0_plane(
    const Vector3& u, const Vector3& v, // segment
    Vector3& isect);                    // intersection point

  /**
   * Intersect a 3D segment with the plane z = zval. Assumes that there
   * is an intersection (fails if the segment is parallel to the plane),
   * and returns the distance from u to the intersection point.
   * The intersection point is returned in isect.
   */ 
  static float z_plane(float zval,      // plane z = zval
    const Vector3& u, const Vector3& v, // segment
    Vector3& isect);                    // intersection point
  
  /**
   * Intersect a 3D segment with the frustrum plane Ax + z = 0.
   * Assumes an intersection, and returns the intersection point in isect.
   */
  static float x_frustrum(
    float A, const Vector3& u, const Vector3& v, Vector3& isect);

  /**
   * Intersect a 3D segment with the frustrum plane By + z = 0.
   * Assumes an intersection, and returns the intersection point in isect.
   */
  static float y_frustrum(
    float B, const Vector3& u, const Vector3& v, Vector3& isect);

};


#endif /*MATH3D_H*/

