/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Andrew Zabolotny <bit@eltech.ru>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdarg.h>
#include "system/isystem.h"
#include "render/graph2d.h"
#include "render/igraph3d.h"
#include "engine/itexture.h"
#include "engine/iworld.h"
#include "general/stdassrt.h"

// Static csGraphics2D variables
int csGraphics2D::ClipX1, csGraphics2D::ClipX2;
int csGraphics2D::ClipY1, csGraphics2D::ClipY2;
int csGraphics2D::Font;
csPixelFormat csGraphics2D::pfmt;
unsigned char *csGraphics2D::Memory;
int *csGraphics2D::LineAddress;
bool csGraphics2D::FullScreen;
int csGraphics2D::Width, csGraphics2D::Height, csGraphics2D::Depth;
RGBpaletteEntry csGraphics2D::Palette[256];
bool csGraphics2D::PaletteAlloc[256];
IWorld* csGraphics2D::world;

// "Virtual functions" through pointers in csGraphics2D
void (*csGraphics2D::DrawPixel) (int x, int y, int color);
void (*csGraphics2D::WriteChar) (int x, int y, int fg, int bg, char c);
void (*csGraphics2D::DrawSprite) (hTexture hTex, int sx, int sy, int sw, int sh, int tx, int ty, int tw, int th);
unsigned char* (*csGraphics2D::GetPixelAt) (int x, int y);

#if defined(PROC_INTEL) && !defined(NO_ASSEMBLER)
#if defined(COMP_VC)
#  include "system/i386/drline.h"
#endif
#endif

BEGIN_INTERFACE_TABLE (csGraphics2D)
  IMPLEMENTS_COMPOSITE_INTERFACE_EX (IGraphics2D, XGraphics2D)
  IMPLEMENTS_COMPOSITE_INTERFACE_EX (IGraphicsInfo, XGraphicsInfo)
END_INTERFACE_TABLE ()

IMPLEMENT_UNKNOWN_NODELETE (csGraphics2D)

csGraphics2D::csGraphics2D ( ISystem* piSystem )
{
  ASSERT(piSystem != NULL);

  // Get the system parameters  
  piSystem->GetWidthSetting (Width);
  piSystem->GetHeightSetting (Height);
  piSystem->GetDepthSetting (Depth);
  piSystem->GetFullScreenSetting (FullScreen);

  Font = 0;
  SetClipRect (0, 0, Width, Height);
  pfmt.PalEntries = 256;
  pfmt.PixelBytes = 1;
  // Initialize pointers to default drawing methods
  DrawPixel = DrawPixel8;
  WriteChar = WriteChar8;
  GetPixelAt = GetPixelAt8;
  DrawSprite = DrawSprite8;
  // Mark all slots in palette as free
  for (int i = 0; i < 256; i++)
  {
    PaletteAlloc[i] = false;
    Palette[i].red = 0;
    Palette[i].green = 0;
    Palette[i].blue = 0;
  }

  system = piSystem;
}

csGraphics2D::~csGraphics2D ()
{
  // no operation
}

bool csGraphics2D::Open (char *Title)
{
  (void)Title;

  // Allocate buffer for address of each scan line to avoid multuplication
  CHK (LineAddress = new int [Height]);
  if (LineAddress == NULL) return false;

  // Initialize scanline address array
  int i,addr,bpl = Width * pfmt.PixelBytes;
  for (i = 0, addr = 0; i < Height; i++, addr += bpl)
    LineAddress[i] = addr;

  return true;
}

void csGraphics2D::Close ()
{
  if (LineAddress) { CHK (delete [] LineAddress); LineAddress = NULL; }
}

void csGraphics2D::complete_pixel_format ()
{
  long s;
  pfmt.RedShift = 0;   s = pfmt.RedMask;   while (s && !(s&1)) { pfmt.RedShift++; s >>= 1; }
  pfmt.GreenShift = 0; s = pfmt.GreenMask; while (s && !(s&1)) { pfmt.GreenShift++; s >>= 1; }
  pfmt.BlueShift = 0;  s = pfmt.BlueMask;  while (s && !(s&1)) { pfmt.BlueShift++; s >>= 1; }
  pfmt.RedBits = 0;    s = pfmt.RedMask >> pfmt.RedShift;     while (s) { pfmt.RedBits++; s >>= 1; }
  pfmt.GreenBits = 0;  s = pfmt.GreenMask >> pfmt.GreenShift; while (s) { pfmt.GreenBits++; s >>= 1; }
  pfmt.BlueBits = 0;   s = pfmt.BlueMask >> pfmt.BlueShift;   while (s) { pfmt.BlueBits++; s >>= 1; }
}

bool csGraphics2D::BeginDraw ()
{
  // no operation
  return true;
}

void csGraphics2D::FinishDraw ()
{
  // no operation
}

void csGraphics2D::Clear(int color)
{
  for (int y = 0; y < Height; y++)
    DrawHorizLine (0, Width - 1, y, color);
}

bool csGraphics2D::DoubleBuffer (bool Enable)
{
  return !Enable;
}

bool csGraphics2D::DoubleBuffer ()
{
  return false;
}

int csGraphics2D::GetPage ()
{
  return 0;
}

void csGraphics2D::ClearAll (int color)
{
  int CurPage = GetPage ();
  do
  {
    if (BeginDraw () != S_OK)
      break;
    Clear (color);
    FinishDraw ();
    Print ();
  } while (GetPage () != CurPage);
}

void csGraphics2D::DrawPixel8 (int x, int y, int color)
{
  if ((x >= ClipX1) && (x < ClipX2) && (y >= ClipY1) && (y < ClipY2))
    *(GetPixelAt (x, y)) = color;
}

void csGraphics2D::DrawPixel16 (int x, int y, int color)
{
  if ((x >= ClipX1) && (x < ClipX2) && (y >= ClipY1) && (y < ClipY2))
    *(short *)(GetPixelAt (x, y)) = color;
}

#ifndef NO_DRAWLINE
void csGraphics2D::DrawLine (int x1, int y1, int x2, int y2, int color)
{
#define abs(a) ((a) >= 0 ? (a) : -(a))

  if (ClipLine (x1, y1, x2, y2, ClipX1, ClipY1, ClipX2, ClipY2))
    return;

  if (abs (x2 - x1) > abs (y2 - y1))
  {
    if (x1 > x2)
    {
      int tmp = x1; x1 = x2; x2 = tmp;
      tmp = y1; y1 = y2; y2 = tmp;
    }
    int y = (y1 * 0x10000) + 0x7fff;	// use 1.15.16 fixed-point
    int deltay = ((y2 - y1) * 0x10000) / (x2 - x1);

    for (int x = x1; x <= x2; x++)
    {
      DrawPixel (x, y / 0x10000, color);
      y += deltay;
    }
  }
  else if ((x2 - x1) || (y2 - y1))
  {
    if (y1 > y2)
    {
      int tmp = y1; y1 = y2; y2 = tmp;
      tmp = x1; x1 = x2; x2 = tmp;
    }
    int x = (x1 * 0x10000) + 0x7fff;	// use 1.15.16 fixed-point
    int deltax = ((x2 - x1) * 0x10000) / (y2 - y1);

    for (int y = y1; y <= y2; y++)
    {
      DrawPixel (x / 0x10000, y, color);
      x += deltax;
    }
  }
  else
    DrawPixel (x1, y1, color);
}
#endif

void csGraphics2D::DrawHorizLine (int x1, int x2, int y, int color)
{
  if ((y >= ClipY1) && (y <= ClipY2))
  {
    if (x1 > x2) { int tmp = x1; x1 = x2; x2 = tmp; }
    if (x1 < ClipX1) x1 = ClipX1;
    if (x2 > ClipX2) x2 = ClipX2;
    if (x1 <= x2)
      switch (pfmt.PixelBytes)
      {
        case 1:
          memset (GetPixelAt (x1, y), color, x2 - x1 + 1);
          break;
        case 2:
        {
          UShort *dest = (UShort *)GetPixelAt (x1, y);
          int count = x2 - x1 + 1;
          while (count--) *dest++ = color;
          break;
        }
      } /* endswitch */
  }
}

void csGraphics2D::Write (int x, int y, int fg, int bg, char *str)
{
  for (; *str; ++str)
  {
    WriteChar(x, y, fg, bg, *str);
    if (FontList[Font].IndividualWidth)
      x += FontList[Font].IndividualWidth[(unsigned char)*str];
    else
      x += FontList[Font].Width;
  }
}

void csGraphics2D::WriteChar8 (int x, int y, int fg, int bg, char c)
{
  int charH = FontList [Font].Height;
  int charW;
  if (FontList[Font].IndividualWidth)
    charW = FontList[Font].IndividualWidth[(unsigned char)c];
  else
    charW = FontList[Font].Width;

  if ((x + charW <= ClipX1) || (x >= ClipX2) || (y + charH <= ClipY1) || (y >= ClipY2))
    return;

  register unsigned char *CharImage =
    &FontList[Font].FontBitmap[((unsigned char)c) * FontList[Font].BytesPerChar];

  if ((x < ClipX1) || (x + charW > ClipX2) || (y < ClipY1) || (y + charH > ClipY2))
  {
    // Perform full clipping
    int lX = x < ClipX1 ? ClipX1 - x : 0;
    int rX = x + charW >= ClipX2 ? ClipX2 - x : charW;
    x += lX;
    for (int i = 0; i < charH; i++, y++)
    {
      register char CharLine = (*CharImage++) << lX;
      if ((y < ClipY1) || (y >= ClipY2)) continue;
      register unsigned char *VRAM = GetPixelAt (x, y);
      for (int j = lX; j < rX; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else if (bg >= 0)
          *VRAM++ = bg;
        else
          VRAM++;
        CharLine <<= 1;
      } /* endfor */
    } /* endfor */
  } else
  for (int i = 0; i < charH; i++, y++)
  {
    register unsigned char *VRAM = GetPixelAt (x, y);
    register char CharLine = *CharImage++;
    if (bg < 0)
      for (int j = 0; j < charW; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else
          VRAM++;
        CharLine <<= 1;
      } else
      for (int j = 0; j < charW; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else
          *VRAM++ = bg;
        CharLine <<= 1;
      } /* endfor */
  } /* endfor */
}

void csGraphics2D::WriteChar16 (int x, int y, int fg, int bg, char c)
{
  int charH = FontList [Font].Height;
  int charW;
  if (FontList[Font].IndividualWidth)
    charW = FontList[Font].IndividualWidth[(unsigned char)c];
  else
    charW = FontList[Font].Width;

  if ((x + charW <= ClipX1) || (x >= ClipX2) || (y + charH <= ClipY1) || (y >= ClipY2))
    return;

  register unsigned char *CharImage =
    &FontList[Font].FontBitmap[((unsigned char)c) * FontList[Font].BytesPerChar];

  if ((x < ClipX1) || (x + charW > ClipX2) || (y < ClipY1) || (y + charH > ClipY2))
  {
    // Perform full clipping
    int lX = x < ClipX1 ? ClipX1 - x : 0;
    int rX = x + charW >= ClipX2 ? ClipX2 - x : charW;
    x += lX;
    for (int i = 0; i < charH; i++, y++)
    {
      register char CharLine = (*CharImage++) << lX;
      if ((y < ClipY1) || (y >= ClipY2)) continue;
      register unsigned short *VRAM = (unsigned short *)GetPixelAt (x, y);
      for (int j = lX; j < rX; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else if (bg >= 0)
          *VRAM++ = bg;
        else
          VRAM++;
        CharLine <<= 1;
      } /* endfor */
    } /* endfor */
  } else
  for (int i = 0; i < charH; i++, y++)
  {
    register unsigned short *VRAM = (unsigned short *)GetPixelAt (x, y);
    register char CharLine = *CharImage++;
    if (bg < 0)
      for (int j = 0; j < charW; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else
          VRAM++;
        CharLine <<= 1;
      } else
      for (int j = 0; j < charW; j++)
      {
        if (CharLine & 0x80)
          *VRAM++ = fg;
        else
          *VRAM++ = bg;
        CharLine <<= 1;
      } /* endfor */
  } /* endfor */
}

hTexture csGraphics2D::RegisterTexture (IMipMapContainer *Tex)
{
  return (hTexture)Tex;
}

bool csGraphics2D::DeregisterTexture (hTexture hTex)
{
  (void) hTex;
  return true;
}

void csGraphics2D::DrawSprite8 (hTexture hTex, int sx, int sy, int sw, int sh,
                                int tx, int ty, int tw, int th)
{
  int TranspColor; 
  ITextureMap *piTM = NULL; 

  ((IMipMapContainer *)hTex)->GetTransparent(TranspColor);
  ((IMipMapContainer *)hTex)->GetTexture(0, &piTM);

  unsigned char *bitmap; 
  int bw;

  piTM->GetBitmap( &bitmap );
  piTM->GetWidth( bw );
  FINAL_RELEASE(piTM);
  

  int dx = (tw << 16) / sw;
  int dy = (th << 16) / sh;

  // Clipping
  if ((sx >= ClipX2) || (sy >= ClipY2) || (sx + sw <= ClipX1) || (sy + sh <= ClipY1))
    return;                             // Sprite is totally invisible
  if (sx < ClipX1)                      // Left margin crossed?
  {
    int nw = sw - (ClipX1 - sx);        // New width
    tx += (ClipX1 - sx) * tw / sw;      // Adjust X coord on texture
    tw = (tw * nw) / sw;                // Adjust width on texture
    sw = nw; sx = ClipX1;
  } /* endif */
  if (sx + sw > ClipX2)                 // Right margin crossed?
  {
    int nw = ClipX2 - sx;               // New width
    tw = (tw * nw) / sw;                // Adjust width on texture
    sw = nw;
  } /* endif */
  if (sy < ClipY1)                      // Top margin crossed?
  {
    int nh = sh - (ClipY1 - sy);        // New height
    ty += (ClipY1 - sy) * th / sh;      // Adjust Y coord on texture
    th = (th * nh) / sh;                // Adjust height on texture
    sh = nh; sy = ClipY1;
  } /* endif */
  if (sy + sh > ClipY2)                 // Bottom margin crossed?
  {
    int nh = ClipY2 - sy;               // New height
    th = (th * nh) / sh;                // Adjust height on texture
    sh = nh;
  } /* endif */

  bitmap += ty * bw + tx;
  ty = 0;

  if (sw == tw)
  {
    for (; sh > 0; sh--, ty += dy, sy++)
    {
      unsigned char *VRAM = GetPixelAt (sx, sy);
      unsigned char *data = bitmap + (ty >> 16) * bw;
      if (TranspColor >= 0)
      {
        for (int w = sw; w; w--)
        {
          if (*data != (unsigned char)TranspColor)
            *VRAM = *data;
          VRAM++;
          data++;
        } /* endfor */
      } else
        memcpy (VRAM, data, sw);
    } /* endfor */
  } else
  {
    for (; sh > 0; sh--, ty += dy, sy++)
    {
      unsigned char *VRAM = GetPixelAt (sx, sy);
      unsigned char *data = bitmap + (ty >> 16) * bw;
      int tx = 0;
      if (TranspColor >= 0)
      {
        for (int w = sw; w; w--)
        {
          unsigned char pixel = data [tx >> 16];
          if (pixel != (unsigned char)TranspColor)
            *VRAM = pixel;
          VRAM++;
          tx += dx;
        } /* endfor */
      } else
      {
        for (int w = sw; w; w--)
        {
          *VRAM = data [tx >> 16];
          VRAM++;
          tx += dx;
        } /* endfor */
      } /* endif */
    } /* endfor */
  } /* endif */
}

void csGraphics2D::DrawSprite16 (hTexture hTex, int sx, int sy, int sw, int sh,
  int tx, int ty, int tw, int th)
{
  int TranspColor; 
  ITextureMap *piTM = NULL; 

  ((IMipMapContainer *)hTex)->GetTransparent(TranspColor);
  ((IMipMapContainer *)hTex)->GetTexture(0, &piTM);

  unsigned char *bitmap; 
  int bw;

  piTM->GetBitmap( &bitmap );
  piTM->GetWidth( bw );
  FINAL_RELEASE(piTM);

  int dx = (tw << 16) / sw;
  int dy = (th << 16) / sh;
  unsigned short *pal_table;
  
  ITextureContainer* pTC;
  TextureTablesPalette* pTexTablesPal;

  world->GetTextures( &pTC );
  
  pTC->GetPaletteLUT( &pTexTablesPal );
  pTC->Release();
  
  pal_table = pTexTablesPal->pal_to_true;

  // Clipping
  if ((sx >= ClipX2) || (sy >= ClipY2) || (sx + sw < ClipX1) || (sy + sh < ClipY1))
    return;                             // Sprite is totally invisible
  if (sx < ClipX1)                      // Left margin crossed?
  {
    int nw = sw - (ClipX1 - sx);        // New width
    tx += (ClipX1 - sx) * tw / sw;      // Adjust X coord on texture
    tw = (tw * nw) / sw;                // Adjust width on texture
    sw = nw; sx = ClipX1;
  } /* endif */
  if (sx + sw > ClipX2)                 // Right margin crossed?
  {
    int nw = ClipX2 - sx;               // New width
    tw = (tw * nw) / sw;                // Adjust width on texture
    sw = nw;
  } /* endif */
  if (sy < ClipY1)                      // Top margin crossed?
  {
    int nh = sh - (ClipY1 - sy);        // New height
    ty += (ClipY1 - sy) * th / sh;      // Adjust Y coord on texture
    th = (th * nh) / sh;                // Adjust height on texture
    sh = nh; sy = ClipY1;
  } /* endif */
  if (sy + sh > ClipY2)                 // Bottom margin crossed?
  {
    int nh = ClipY2 - sy;               // New height
    th = (th * nh) / sh;                // Adjust height on texture
    sh = nh;
  } /* endif */

  bitmap += ty * bw + tx;
  ty = 0;

  if (sw == tw)
  {
    for (; sh > 0; sh--, ty += dy, sy++)
    {
      unsigned short *VRAM = (unsigned short *)GetPixelAt (sx, sy);
      unsigned char *data = bitmap + (ty >> 16) * bw;
      if (TranspColor >= 0)
      {
        for (int w = sw; w; w--)
        {
          if (*data != (unsigned char)TranspColor)
            *VRAM = pal_table [*data];
          VRAM++;
          data++;
        } /* endfor */
      } else
        for (int w = sw; w; w--)
          *VRAM++ = pal_table [*data++];
    } /* endfor */
  } else
  {
    for (; sh > 0; sh--, ty += dy, sy++)
    {
      unsigned short *VRAM = (unsigned short *)GetPixelAt (sx, sy);
      unsigned char *data = bitmap + (ty >> 16) * bw;
      int tx = 0;
      if (TranspColor >= 0)
      {
        for (int w = sw; w; w--)
        {
          unsigned char pixel = data [tx >> 16];
          if (pixel != (unsigned char)TranspColor)
            *VRAM = pal_table [pixel];
          VRAM++;
          tx += dx;
        } /* endfor */
      } else
      {
        for (int w = sw; w; w--)
        {
          *VRAM = pal_table [data [tx >> 16]];
          VRAM++;
          tx += dx;
        } /* endfor */
      } /* endif */
    } /* endfor */
  } /* endif */
}

void csGraphics2D::SetClipRect (int xmin, int ymin, int xmax, int ymax)
{
  ClipX1 = xmin; ClipX2 = xmax;
  ClipY1 = ymin; ClipY2 = ymax;
}

void csGraphics2D::GetClipRect (int &xmin, int &ymin, int &xmax, int &ymax)
{
  xmin = ClipX1; xmax = ClipX2;
  ymin = ClipY1; ymax = ClipY2;
}

// This algorithm has been borrowed about eight years ago from a book on
// computer graphics, I believe its author was R.Wilson or something alike.
// It was first converted from ASM to Pascal, and now I converted it to C :-)
bool csGraphics2D::ClipLine (int &x1, int &y1, int &x2, int &y2,
  int xmin, int ymin, int xmax, int ymax)
{
  bool Inside,Outside;
  char ocu1,ocu2;

  // A line INCLUDES last pixel, so shrink clip area by one pixel
  xmax--; ymax--;

#define CLIP_LEFT   0x01
#define CLIP_TOP    0x02
#define CLIP_RIGHT  0x04
#define CLIP_BOTTOM 0x08
#define SetOutCodes(u, x, y) \
  if (x < xmin) u = CLIP_LEFT; else u = 0; \
  if (y < ymin) u |= CLIP_TOP; \
  if (x > xmax) u |= CLIP_RIGHT; \
  if (y > ymax) u |= CLIP_BOTTOM;

#define ISWAP(a,b) { int __tmp__ = a; a = b; b = __tmp__; }
#define CSWAP(a,b) { char __tmp__ = a; a = b; b = __tmp__; }

  SetOutCodes (ocu1, x1, y1);
  SetOutCodes (ocu2, x2, y2);

  Inside  = (ocu1 | ocu2) == 0;
  Outside = (ocu1 & ocu2) != 0;

  while ((!Outside) && (!Inside))
  {
    if (ocu1 == 0)                      // swap endpoints if necessary
    {                                   // so that (x1,y1) needs to be clipped
      ISWAP (x1, x2);
      ISWAP (y1,y2);
      CSWAP (ocu1, ocu2);
    }
    if (ocu1 & CLIP_LEFT)               // clip left
    {
      y1 = y1 + ((y2 - y1) * (xmin - x1)) / (x2 - x1);
      x1 = xmin;
    }
    else if (ocu1 & CLIP_TOP)           // clip above
    {
      x1 = x1 + ((x2 - x1) * (ymin - y1)) / (y2 - y1);
      y1 = ymin;
    }
    else if (ocu1 & CLIP_RIGHT)         // clip right
    {
      y1 = y1 + ((y2 - y1) * (xmax - x1)) / (x2 - x1);
      x1 = xmax;
    }
    else if (ocu1 & CLIP_BOTTOM)        // clip below
    {
      x1 = x1 + ((x2 - x1) * (ymax - y1)) / (y2 - y1);
      y1 = ymax;
    }
    SetOutCodes (ocu1, x1, y1);         // update for (x1,y1)
    Inside  = (ocu1 | ocu2) == 0;
    Outside = (ocu1 & ocu2) != 0;
   }
 return Outside;
}

bool csGraphics2D::SaveArea (ImageArea *&Area, int x, int y, int w, int h)
{
  if (x < 0)
  { w += x; x = 0; }
  if (x + w > Width)
    w = Width - x;
  if (y < 0)
  { h += y; y = 0; }
  if (y + h > Height)
    h = Height - y;
  if ((w <= 0) || (h <= 0))
  {
    Area = NULL;
    return true;
  } /* endif */

  CHK (Area = new ImageArea (x, y, w, h));
  if (!Area)
    return false;
  w *= pfmt.PixelBytes;
  CHK (char *dest = Area->data = new char [w * h]);
  if (!dest)
    return false;
  for (; h > 0; y++, h--)
  {
    unsigned char *VRAM = GetPixelAt (x, y);
    memcpy (dest, VRAM, w);
    dest += w;
  } /* endfor */
  return true;
}

void csGraphics2D::RestoreArea (ImageArea *Area, bool Free)
{
  if (Area)
  {
    char *dest = Area->data;
    int x = Area->x, y = Area->y, w = Area->w, h = Area->h;
    w *= pfmt.PixelBytes;
    for (; h; y++, h--)
    {
      unsigned char *VRAM = GetPixelAt (x, y);
      memcpy (VRAM, dest, w);
      dest += w;
    } /* endfor */
    if (Free)
      FreeArea (Area);
  } /* endif */
}

void csGraphics2D::FreeArea (ImageArea *Area)
{
  if (Area)
  {
    if (Area->data)
      CHKB (delete[] Area->data);
    CHK (delete Area);
  } /* endif */
}

void csGraphics2D::SetRGB (int i, int r, int g, int b)
{
  Palette[i].red = r;
  Palette[i].green = g;
  Palette[i].blue = b;
  PaletteAlloc[i] = true;
}

unsigned char *csGraphics2D::GetPixelAt8 (int x, int y)
{
  return (Memory + (x + LineAddress[y]));
}

unsigned char *csGraphics2D::GetPixelAt16 (int x, int y)
{
  return (Memory + (x + x + LineAddress[y]));
}

bool csGraphics2D::SetMouseCursor (int iShape, IMipMapContainer* iBitmap)
{
  (void)iBitmap;
  return (iShape == csmcArrow);
}

void csGraphics2D::GetStringError(HRESULT hRes, char* szError)
{
    strcpy(szError, "Internal Error.");
}

void csGraphics2D::SysPrintf(int mode, char* szMsg, ...)
{
	char buf[1024];
	va_list arg;
	
	va_start (arg, szMsg);
	vsprintf (buf, szMsg, arg);
	va_end (arg);
	
	system->Print(mode, buf);
}
