/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Andrew Zabolotny <bit@eltech.ru>
    Ported to COM by Dan Ogles <dogles@peachtree.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __IGRAPH2D_H__
#define __IGRAPH2D_H__

#include "com/com.h"

class csRect;
interface ISystem;
interface IMipMapContainer;

typedef void *hTexture;

/// This structure is used to define a external list of bitmap fonts
struct FontDef
{
  /// Width of font (if IndividualWidth == NULL)
  int Width;
  /// Height of font
  int Height;
  /// Bytes per character
  int BytesPerChar;
  /// An array of character widths or NULL
  unsigned char *IndividualWidth;
  /// A pointer to font bitmap
  unsigned char *FontBitmap;
};

/// The array containing the definitions of all fonts
extern FontDef FontList[];

enum FontType
{
  csFontParent = -1,
  csFontPolice,
  csFontPoliceFixed,
  csFontItalic,
  csFontItalicFixed,
  csFontCourier,
  csFontCourierFixed,
  csFontTiny,
  csFontTinyFixed
};

/// This structure is used for saving/restoring areas of screen
struct ImageArea
{
  int x, y, w, h;
  char *data;

  // @@@ not sure if this has binary compatibility in all cases - DAN
  inline ImageArea (int sx, int sy, int sw, int sh)
  { x = sx; y = sy; w = sw; h = sh; data = NULL; }
};

/// A RGB palette entry
struct RGBpaletteEntry
{
  UByte red;
  UByte green;
  UByte blue;
  UByte flags;
};

/**
 * Standard mouse cursor IDs
 */
enum csMouseCursorID
{
  /// No cursor
  csmcNone = -1,
  /// Arrow cursor
  csmcArrow = 0,
  /// Lens cursor
  csmcLens,
  /// Cross-hatch cursor
  csmcCross,
  /// Pen cursor
  csmcPen,
  /// Window move cursor
  csmcMove,
  /// Diagonal (\) resizing cursor
  csmcSizeNWSE,
  /// Diagonal (/) resizing cursor
  csmcSizeNESW,
  /// Vertical sizing cursor
  csmcSizeNS,
  /// Horizontal sizing cursor
  csmcSizeEW,
  /// Invalid operation cursor
  csmcStop,
  /// Wait (longplay operation) cursor
  csmcWait
};

/**
 * Structure describing the pixel format.
 */
struct csPixelFormat
{
  /**
   * The masks to extract the color information from a pixel (truecolor mode only).
   * Currently only masks for 16-bit/15-bit colors are supported.
   * Ignore the Mask and Shift fields of this structure if PalEntries != 0.
   */
  long RedMask, GreenMask, BlueMask;
  /// The shifts to extract the color information from a pixel (truecolor mode only).
  int RedShift, GreenShift, BlueShift;
  /// The number of significant bits for every color.
  int RedBits, GreenBits, BlueBits;

  /**
   * Number of palette entries. 0 for truecolor, else the number of palette
   * entries (this number should be equal to 256 even if not all of these 256
   * colors can be modified (like is the case in Windows)).
   * Currently only 0 and 256 are supported here.
   */
  int PalEntries;

  /**
   * Number of bytes for every pixel.
   * The only supported values currently are:<p>
   * <ul>
   * <li>1: for palette mode (256 palette entries)
   * <li>2: for truecolor mode (no palette entries)
   * </ul>
   */
  int PixelBytes;
};

extern const IID IID_IGraphics2D;
extern const IID IID_IGraphicsInfo;

/**
 * This is the interface for 2D renderer. System-dependent ports
 * should derive their own class from the standard implementation, 
 * csGraphics2D, and implement required (marked with an asterisk (*))
 * functions. Functions not marked with an asterisk are optional, 
 * but possibly slow since they are too general.</p>
 *
 */
interface IGraphics2D : public IUnknown
{
   /**
    * This routine should be called before any draw operations.
    * It should return S_OK if graphics context is ready.
    */
    COM_METHOD_DECL BeginDraw() = 0;
    /// This routine should be called when you finished drawing.
    COM_METHOD_DECL FinishDraw() = 0;
    
    /// (*) Flip video pages (or dump backbuffer into framebuffer).
    COM_METHOD_DECL Print(csRect* pArea) = 0;

    /// Clear backbuffer.
    COM_METHOD_DECL Clear(int color) = 0;
    /// Clear all video pages.
    COM_METHOD_DECL ClearAll(int color) = 0;

    /// Draw a horizontal line.
    COM_METHOD_DECL DrawHorizLine(int x1, int x2, int y, int color) = 0;
    /// Draw a line.
    COM_METHOD_DECL DrawLine(int x1, int y1, int x2, int y2, int color) = 0;
   /**
    * Clip a line against given rectangle.
    * Function returns S_OK if line is not visible.
    */
    COM_METHOD_DECL ClipLine(int& x1, int& y1, int& x2, int& y2, int xmin, int ymin, int xmax, int ymax) = 0;
    /// Draw a pixel.
    COM_METHOD_DECL DrawPixel(int x, int y, int color) = 0;
    /// Returns the address of the pixel at the specified (x, y) coordinates.
    COM_METHOD_DECL GetPixelAt (int x, int y, unsigned char** pPixel ) = 0;
    /// Draw a sprite using a rectangle from given texture
    COM_METHOD_DECL DrawSprite(hTexture hTex, int sx, int sy, int sw, int sh, int tx, int ty, int tw, int th) = 0;

   /**
    * Register a texture and return a handle.
    * Derived classes SHOULD allow multiple registrations of same texture!
    * If texture has already been registered, it should perform no operation
    * and return same handle as first time it was registered.
    */
    COM_METHOD_DECL RegisterTexture(IMipMapContainer* pTex, hTexture& hTex) = 0;
   
   /**
    * Deregister texture by handle.
    * Each handle can be deregistered multiple times, so deregistering
    * a handle multiple times should not be considered an error.
    * Texture should have a usage count and freed only when it reaches 0.
    */
    COM_METHOD_DECL DeregisterTexture(hTexture hTex) = 0;
   
    /// Enable or disable double buffering; returns S_OK or S_FALSE.
    COM_METHOD_DECL DoubleBuffer(bool Enable) = 0;
    /// Get the double buffer state
    COM_METHOD_DECL GetDoubleBufferState(bool& State) = 0;
 
   /**
    * Save a subarea of screen area into the variable Data.
    * Storage is allocated in this call, you should either FreeArea()
    * it after usage or RestoreArea() it.
    */
    COM_METHOD_DECL SaveArea(ImageArea** Area, int x, int y, int w, int h) = 0;
    /// Restore a subarea of screen saved with SaveArea()
    COM_METHOD_DECL RestoreArea(ImageArea* Area, bool bFree) = 0;
    /// Free storage allocated for a subarea of screen
    COM_METHOD_DECL FreeArea(ImageArea* Area) = 0;

    /// Retrieve clipping rectangle
    COM_METHOD_DECL GetClipRect(int& nMinX, int& nMinY, int& nMaxX, int& nMaxY) = 0;    
    /// Set clipping rectangle
    COM_METHOD_DECL SetClipRect(int nMinX, int nMinY, int nMaxX, int nMaxY) = 0;

   /**
    * Set mouse cursor to one of predefined shape classes
    * (see csmcXXX enum above). If a specific mouse cursor shape
    * is not supported, return 'false'; otherwise return 'true'.
    * If system supports it and iBitmap != NULL, shape should be
    * set to the bitmap passed as second argument; otherwise cursor
    * should be set to its nearest system equivalent depending on
    * iShape argument.
    */
    COM_METHOD_DECL SetMouseCursor(csMouseCursorID Shape, IMipMapContainer* iBitmap) = 0;
    /// (*) Set a color index to given R,G,B (0..255) values
    COM_METHOD_DECL SetRGB(int i, int r, int g, int b) = 0;
    /// Write a text string into the back buffer
    COM_METHOD_DECL Write(int x, int y, int fg, int bg, char* str) = 0;
    /// Write a single character.
    COM_METHOD_DECL WriteChar(int x, int y, int fg, int bg, char c) = 0;

    /// Gets the type of the font.
    COM_METHOD_DECL GetFontID(int& id) = 0;
    /// Sets the type of the font.
    COM_METHOD_DECL SetFontID(int id) = 0;

    /// Open the device.
    COM_METHOD_DECL Open (char *Title) = 0;
    /// Close the device.
    COM_METHOD_DECL Close() = 0;

    /**
     * Perform a system specific exension.
     * This is probably not the good way to do this but I see no other
     * way currently.
     */
    COM_METHOD_DECL PerformExtension(char* args) = 0;
};

/**
 * This interface is used to retrieve information
 * about the current pixel format and palette 
 * entries.
 */
interface IGraphicsInfo : public IUnknown
{
    /// Get active videopage number (starting from zero)
    COM_METHOD_DECL GetPage(int& nPage) = 0;

   /**
    * Return the number of bytes for every pixel.
    * This function is equivalent to the PixelBytes field that
    * you get from GetPixelFormat.
    */
    COM_METHOD_DECL GetPixelBytes(int& nPixelBytes) = 0;

   /**
    * Return information about about the pixel format.
    */
    COM_METHOD_DECL GetPixelFormat(csPixelFormat* PixelFormat) = 0;
 
    /// Get the current font ID
    COM_METHOD_DECL GetFontID(int& nID) = 0;

    /// Return the width of the framebuffer.
    COM_METHOD_DECL GetWidth(int& nWidth) = 0;
    /// Return the height of the framebuffer.
    COM_METHOD_DECL GetHeight(int& nHeight) = 0;
    /// Returns 'true' if the program is being run full-screen.
    COM_METHOD_DECL GetIsFullScreen(bool& bIsFullScreen) = 0;

   /**
    * Return the number of palette entries that can be modified.
    * This should return 0 if there is no palette (true color displays).
    * This function is equivalent to the PalEntries field that you
    * get from GetPixelFormat. It is just a little bit easier to obtain
    * this way.
    */
    COM_METHOD_DECL GetNumPalEntries(int& nEntries) = 0;
    /// Get the palette (if there is one)
    COM_METHOD_DECL GetPalette(RGBpaletteEntry** pPalette) = 0;

    /// Get the string equivilent of the HRESULT.
    COM_METHOD_DECL GetStringError( HRESULT hRes, char* szErrorString ) = 0;
};

extern const IID IID_IGraphics2DFactory;

interface IGraphics2DFactory : public IUnknown
{
    /// Create the graphics context
    COM_METHOD_DECL CreateInstance( REFIID riid, ISystem* piSystem, void** ppv ) = 0;

    /// Lock or unlock from memory.
    COM_METHOD_DECL LockServer(BOOL bLock) = 0;
};

#endif      // __IGRAPH2D_H__
