/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Jorrit Tyberghein, Dan Ogles, and Gary Clark.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __IGRAPH3D_H__
#define __IGRAPH3D_H__

#include "com/com.h"

#include "engine/basic/csid.h"
#include "engine/icamera.h"
#include "engine/basic/fog.h"

class Vector2;
class Matrix3;
class Vector3;
class csRect;

interface IWorld;
interface IPolygon3D;
interface IPolygonTexture;
interface IMipMapContainer;
interface IGraphics2D;
interface ISystem;

#define CS_FOG_FRONT 0
#define CS_FOG_BACK 1
#define CS_FOG_VIEW 2

///
class G3DVertex
{
public:
  /// Screen x space.
  float sx;
  /// Screen y space.
  float sy;
};

///
class G3DTexturePlane
{
public:
  /// Transformation from camera space to texture space.
  Matrix3* m_cam2tex;
  /// Transformation from camera space to texture space.
  Vector3* v_cam2tex;
};

///
class G3DPolyNormal
{
public:
  ///
  float A;
  ///
  float B;
  ///
  float C;
  ///
  float D;
};


///
struct G3DPolygon
{
  /// Vertices that form the polygon.
  G3DVertex vertices[200];
  /// Maximum number of vertices.
  int max;
  /// Current number of vertices.
  int num;
  /// The original polygon.
  IPolygon3D* polygon;
  /// Transformation matrices for the texture. @@@ BAD NAME
  G3DTexturePlane plane;
  /// The plane equation in camera space of this polygon. @@@ BAD NAME
  G3DPolyNormal normal;
  /// Unique ID number of the polygon.
  CS_ID id;

  /// The perspective-incorrect texture (used only if polygon == NULL)
  IMipMapContainer* pi_texture;
  
  ///
  struct poly_texture_def 
  { 
    float u, v, z; 
    float r, g, b;	// Used only with Gouroud shading (between 0 and 1).
  };

  /// Perspective-incorrect U & V for vertices; also z values
  poly_texture_def *pi_texcoords;
  /// The original triangle (that resulted in this polygon after clipping)
  Vector2 *pi_triangle;
  /// The original u,v,z
  poly_texture_def pi_tritexcoords [3];
};

///
class G3DFltLight
{
public:
    // all points are screen-projections.
    // this is a 2d circle (it's incorrect, but works well as an approx)
    float sx, sy, sz;
    float srad;
    float sqdist;       // camera space: radius squared.
    float r, g, b;
};

/// A Triangle used by the csSprite3D class.
class G3DTriangle
{
public:
  struct TrVertex
  {
    // Screen-projected x and y.
    float sx, sy;
    // Camera space z.
    float z;
    // Texture coordinates (0..1).
    float u, v;
  };
  TrVertex v[3];
  // The texture to use.
  IMipMapContainer* texture;
};

/// Don't test/write, write, test, and write/test, respectively.
enum ZBufMode { ZBuf_None = 0x00, ZBuf_Fill = 0x01, ZBuf_Test = 0x02, ZBuf_Use = 0x03};

//enumerations: DAN

///
enum G3D_RENDERSTATEOPTION
{
  G3DRENDERSTATE_NOTHING,
  G3DRENDERSTATE_ZBUFFERTESTENABLE,
  G3DRENDERSTATE_ZBUFFERFILLENABLE,
  G3DRENDERSTATE_DITHERENABLE,
  G3DRENDERSTATE_SPECULARENABLE,
  G3DRENDERSTATE_BILINEARMAPPINGENABLE,
  G3DRENDERSTATE_TRILINEARMAPPINGENABLE,
  G3DRENDERSTATE_TRANSPARENCYENABLE,
  G3DRENDERSTATE_MIPMAPENABLE,
  G3DRENDERSTATE_DEBUGENABLE,
  G3DRENDERSTATE_EDGESENABLE,
  G3DRENDERSTATE_LIGHTFRUSTRUMENABLE,
  G3DRENDERSTATE_TEXTUREMAPPINGENABLE,
  G3DRENDERSTATE_FILTERINGENABLE,
  G3DRENDERSTATE_PERFECTMAPPINGENABLE,
  G3DRENDERSTATE_LIGHTINGENABLE,
  G3DRENDERSTATE_INTERLACINGENABLE,
  G3DRENDERSTATE_MMXENABLE
};

/// Bit flags for IGraphics3D::BeginDraw ()
/// We're going to draw 2D graphics
#define CSDRAW_2DGRAPHICS   0x00000001
/// We're going to draw 3D graphics
#define CSDRAW_3DGRAPHICS   0x00000002
/// Clear Z-buffer ?
#define CSDRAW_CLEARZBUFFER 0x00000010
/// Clear frame buffer ?
#define CSDRAW_CLEARSCREEN  0x00000020

///
enum G3D_COLORMODEL
{
  G3DCOLORMODEL_MONO,
  G3DCOLORMODEL_RGB
};

///
enum G3D_RASTERCAPS
{
  G3DRASTERCAPS_DITHER = 0x01,
  G3DRASTERCAPS_SUBPIXEL = 0x02,
  G3DRASTERCAPS_ZBUFFERLESSHSR = 0x04,
};

///
enum G3D_SHADECAPS
{
  G3DRASTERCAPS_FLAT = 1,
  G3DRASTERCAPS_GOURAUD = 2,
  G3DRASTERCAPS_PHONG = 4,
  G3DRASTERCAPS_LIGHTMAP = 8
};

///
enum G3D_FILTERCAPS
{
  G3DFILTERCAPS_LINEAR = 0x01,
  G3DFILTERCAPS_LINEARMIPLINEAR = 0x02,
  G3DFILTERCAPS_LINEARMIPNEAREST = 0x04,
  G3DFILTERCAPS_MIPLINEAR = 0x08,
  G3DFILTERCAPS_MIPNEAREST = 0x10,
  G3DFILTERCAPS_NEAREST = 0x20
};

///
struct G3D_PRIMCAPS
{
  G3D_RASTERCAPS RasterCaps;
  bool canBlend;
  G3D_SHADECAPS ShadeCaps;
  bool PerspectiveCorrects;
  G3D_FILTERCAPS FilterCaps;
};

///
struct G3D_CAPS
{
  G3D_COLORMODEL ColorModel;
  bool CanClip;
  bool SupportsArbitraryMipMapping;
  int BitDepth;
  int ZBufBitDepth;
  int minTexHeight, minTexWidth;
  int maxTexHeight, maxTexWidth;
  G3D_PRIMCAPS PrimaryCaps;
};

enum G3D_COLORMAPFORMAT
{
    G3DCOLORFORMAT_ANY,             // can use either private color maps or global color maps
    G3DCOLORFORMAT_PRIVATE,         // only uses private color maps.
    G3DCOLORFORMAT_GLOBAL           // only uses global color maps.
};

extern const GUID IID_IGraphics3D;

/**
 * This is the standard 3D graphics interface.
 * All 3D graphics rasterizer servers for Crystal Space
 * should implement this interface, as well as the 
 * IGraphics2D and IGraphicsInfo interface.
 * The standard implementation is csGraphics3DSoftware.
 */
interface IGraphics3D : public IUnknown
{
public:

  /// Open the 3D graphics display.
  COM_METHOD_DECL Open( char *Title) = 0;
  /// Close the 3D graphics display.
  COM_METHOD_DECL Close() = 0;

  /// Change the dimensions of the display.
  COM_METHOD_DECL SetDimensions (int width, int height) = 0;

  /// Start a new frame (see CSDRAW_XXX bit flags)
  COM_METHOD_DECL BeginDraw (int DrawFlags) = 0;

  /// End the frame and do a page swap.
  COM_METHOD_DECL FinishDraw () = 0;

  /// Print the image in backbuffer
  COM_METHOD_DECL Print (csRect *area) = 0;

  /// Set the mode for the Z buffer used for drawing the next polygon.
  COM_METHOD_DECL SetZBufMode (ZBufMode mode) = 0;

  /// Draw the projected polygon with light and texture.
  COM_METHOD_DECL DrawPolygon ( G3DPolygon& poly) = 0;

  /**
   * Draw the projected polygon with light and texture.
   * Debugging version. This one does not actually draw anything
   * but it just prints debug information about what it would have
   * done.
   */
  COM_METHOD_DECL DrawPolygonDebug ( G3DPolygon& poly) = 0;

  /// Draw a line in camera space.
  COM_METHOD_DECL DrawLine (Vector3& v1, Vector3& v2, int color) = 0;

  /**
   * Draw a projected polygon.
   * This routine will draw the polygon using and filling
   * the z buffer if needed. The texture mapping need not
   * be perspective correct since this function is generally going
   * to be used for 3D sprites which are made up of small triangles.
   */
  COM_METHOD_DECL DrawPolygonQuick ( G3DPolygon& poly, bool gouroud) = 0;

  /**
   * Initiate a volumetric fog object. This function will be called
   * before front-facing and back-facing fog polygons are added to
   * the object. The fog object will be convex but not necesarily closed.
   * The given CS_ID can be used to identify multiple fog objects when
   * multiple objects are started.
   */
  COM_METHOD_DECL OpenFogObject (CS_ID id, csFog* fog) = 0;

  /**
   * Add a front or back-facing fog polygon in the current fog object.
   * Note that it is guaranteed that all back-facing fog polygons
   * will have been added before the first front-facing polygon.
   * fogtype can be:
   *	CS_FOG_FRONT:	a front-facing polygon
   *	CS_FOG_BACK:	a back-facing polygon
   *	CS_FOG_VIEW:	the view-plane
   */
  COM_METHOD_DECL AddFogPolygon (CS_ID id, G3DPolygon& poly, int fogtype) = 0;

  /**
   * Close a volumetric fog object. After the volumetric object is
   * closed it should be rendered on screen (whether you do it here
   * or in AddFogPolygon is not important).
   */
  COM_METHOD_DECL CloseFogObject (CS_ID id) = 0;

  /// Give a texture to csGraphics3D to cache it.
  COM_METHOD_DECL CacheTexture (IPolygonTexture* texture) = 0;

  /// Release a texture from the cache.
  COM_METHOD_DECL UncacheTexture (IPolygonTexture* texture) = 0;

  /// Set a renderstate value.
  COM_METHOD_DECL SetRenderState (G3D_RENDERSTATEOPTION op, long val) = 0;

  /// Get a renderstate value.
  COM_METHOD_DECL GetRenderState (G3D_RENDERSTATEOPTION op, long& val) = 0;

  /**
   * Get the current driver's capabilities. Each driver implements their
   * own function.
   */
  COM_METHOD_DECL GetCaps (G3D_CAPS* caps) = 0;

  /// Debugging: get the value of the Z buffer at some location.
  COM_METHOD_DECL GetZBufPoint (int x, int y, unsigned long** nResult) = 0;

  /// Dump the texture cache.
  COM_METHOD_DECL DumpCache (void) = 0;

  /// Clear the texture cache.
  COM_METHOD_DECL ClearCache (void) = 0;

  /// Get drawing buffer width
  COM_METHOD_DECL GetWidth (int& nResult) = 0;
  /// Get drawing buffer height
  COM_METHOD_DECL GetHeight (int& nResult) = 0;

  /// Set current World (for sprite textures)
  COM_METHOD_DECL SetWorld (IWorld* pWorld) = 0;
  
  /// Get the texture representation scheme.
  COM_METHOD_DECL GetColormapFormat( G3D_COLORMAPFORMAT& g3dFormat ) = 0;

  /// Returns S_OK if the driver needs PO2 lightmaps and texture maps, S_FALSE otherwise.
  COM_METHOD_DECL NeedsPO2Maps() = 0;

  /// Get the maximum aspect ratio of texture maps.
  COM_METHOD_DECL GetMaximumAspectRatio(int& retval) = 0;

  /// Set the camera object.
  COM_METHOD_DECL SetCamera( ICamera* pCamera ) = 0;

  /// Get the 2D Driver
  COM_METHOD_DECL Get2dDriver( IGraphics2D** piG2D ) = 0;
};

extern const IID IID_IGraphicsContextFactory;
/// 
interface IGraphicsContextFactory : public IUnknown
{
    /// Create the graphics context
    COM_METHOD_DECL CreateInstance( REFIID riid, ISystem* piSystem, void** ppv ) = 0;

    /// Lock or unlock from memory.
    COM_METHOD_DECL LockServer(BOOL bLock) = 0;
};

#endif      // __IGRAPH3D_H__
