/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <time.h>
#include <stdarg.h>

#include "com/com.h"
#include "system/isystem.h"
#include "geometry/math2d.h"
#include "geometry/math3d.h"
#include "render/graph3d.h"
#include "engine/polygon/ipolygon.h"
#include "engine/icamera.h"
#include "engine/itexture.h"
#include "engine/iworld.h"
#include "render/igraph2d.h"
#include "render/opengl/g3d_opengl.h"
#include "render/opengl/gl_texturecache.h"
#include <GL/gl.h>

#ifdef DO_MMX
#  include "system/i386/mmx.h"
#endif

// Static world object (containing textures)
IWorld *csGraphics3DOpenGL::world;
// Option variable: do texture lighting?
bool csGraphics3DOpenGL::do_lighting = true;
// Option variable: render transparent textures?
bool csGraphics3DOpenGL::do_transp = true;
// Option variable: render textures?
bool csGraphics3DOpenGL::do_textured = true;
// Option variable: do perfect texture mapping?
bool csGraphics3DOpenGL::do_perfect = false;
// Option variable: draw white borders for light frustrum?
bool csGraphics3DOpenGL::do_light_frust = false;
// @@todo: remove this
bool csGraphics3DOpenGL::do_debug = false;
// Interlacing.
int csGraphics3DOpenGL::cfg_interlacing = -1;
bool csGraphics3DOpenGL::ilace_fastmove = false;
// Texel filtering.
bool csGraphics3DOpenGL::do_texel_filt = false;

#if DO_MMX
/// True if 3D rendering should use MMX if available
bool csGraphics3DOpenGL::do_mmx = true;
#endif

//---------------------------------------------------------------------------

IMPLEMENT_UNKNOWN_NODELETE( csGraphics3DOpenGL )

BEGIN_INTERFACE_TABLE( csGraphics3DOpenGL )
    IMPLEMENTS_INTERFACE( IGraphics3D )
END_INTERFACE_TABLE()

csGraphics3DOpenGL::csGraphics3DOpenGL (ISystem* piSystem) : m_piG2D(NULL), m_piCamera(NULL)    
{
  HRESULT hRes;
  CLSID clsid2dDriver;
  char *sz2DDriver = "crystalspace.graphics2d.defaultgl";
  IGraphics2DFactory* piFactory = NULL;
  texture_cache = NULL;
  lightmap_cache = NULL;
  
  //ASSERT( piSystem );
  m_piSystem = piSystem;

  hRes = csCLSIDFromProgID( &sz2DDriver, &clsid2dDriver );

  if (FAILED(hRes))
  {	  
      SysPrintf(MSG_FATAL_ERROR, "Error! 2D Graphics DLL with ProgID \"%s\" not found on this system.", sz2DDriver);
      exit(0);
  }

  hRes = csCoGetClassObject( clsid2dDriver, CLSCTX_INPROC_SERVER, NULL, (REFIID)IID_IGraphics2DFactory, (void**)&piFactory );
  if (FAILED(hRes))
  {
      SysPrintf(MSG_FATAL_ERROR, "Error! Couldn't create 2D graphics driver instance.");
      exit(0);
  }

  hRes = piFactory->CreateInstance( (REFIID)IID_IGraphics2D, m_piSystem, (void**)&m_piG2D );
  if (FAILED(hRes))
  {
      SysPrintf(MSG_FATAL_ERROR, "Error! Couldn't create 2D graphics driver instance.");
      exit(0);
  }

  FINAL_RELEASE( piFactory );

  //z_buffer = NULL;
  z_buf_mode = ZBuf_None;

  width = height = -1;

  rstate_alphablend = true;
  rstate_flat = false;
  rstate_mipmap = 0;
  rstate_edges = false;
}

STDMETHODIMP csGraphics3DOpenGL::Open (char *Title)
{
  DrawMode = 0;
  //ASSERT( m_piG2D );

  IGraphicsInfo* piGI = NULL;

  VERIFY_SUCCESS( m_piG2D->QueryInterface((IID&)IID_IGraphicsInfo, (void**)&piGI) );

  if (FAILED(m_piG2D->Open (Title)))
  {
      SysPrintf (MSG_FATAL_ERROR, "Error opening Graphics2D context.");
      FINAL_RELEASE( piGI );
      // set "not opened" flag
      width = height = -1;

      return E_UNEXPECTED;
  }
  
  int nWidth, nHeight;
  bool bFullScreen;

  piGI->GetWidth(nWidth);
  piGI->GetHeight(nHeight);
  piGI->GetIsFullScreen(bFullScreen);

  width = nWidth;
  height = nHeight;
  width2 = nWidth/2;
  height2 = nHeight/2;
  SetDimensions (width, height);

  SysPrintf(MSG_INITIALIZATION, "Using %s mode at resolution %dx%d.\n",
            bFullScreen ? "full screen" : "windowed", nWidth, nHeight);
  
  int gi_pixelbytes;
  piGI->GetPixelBytes( gi_pixelbytes );
/*  
  if (gi_pixelbytes > 1)
  {
    csPixelFormat pfmt;
    piGI->GetPixelFormat( &pfmt );
      
    SysPrintf (MSG_INITIALIZATION, "Using truecolor mode with %d bytes per pixel and %d:%d:%d RGB mode.\n",
          gi_pixelbytes, pfmt.RedBits, pfmt.GreenBits, pfmt.BlueBits);

    CHK (tcache = new TextureCache16 ());
  }
  else
  {
    SysPrintf (MSG_INITIALIZATION, "Using palette mode with 1 byte per pixel (256 colors).\n");
    CHK (tcache = new TextureCache ());
  }
*/  
#   if defined (DO_MMX)
  char cpu_id[13];
  cpu_mmx = mmxDetect (cpu_id);
  if (cpu_mmx)
  {
    SysPrintf (MSG_INITIALIZATION, "%s MMX capable CPU detected\n", cpu_id);
  }
#   endif
  // need some good way of determining a good texture map size and
  // bit depth for the texture 'cache', since GL hides most of the details
  // from us.  Maybe we should just not worry about it... GJH
  CHK (texture_cache = new OpenGLTextureCache(1<<24,24));
  CHK (lightmap_cache = new OpenGLLightmapCache(1<<24,24));
  
  glClearColor(0.,0.,0.,0.);
  glClearDepth(-1.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
 
  FINAL_RELEASE( piGI );
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::Close()
{
  HRESULT rc = S_OK;
  if (width == height == -1)
    return rc;
  rc = m_piG2D->Close();
  width = height = -1;
  return rc;
}

csGraphics3DOpenGL::~csGraphics3DOpenGL ()
{
  //CHK (delete [] z_buffer);
 
  FINAL_RELEASE( m_piG2D );

  // note! this doesnt' actually destroy the world 
  // it only releases our reference to it.
  if (world) world->Release();

  CHK (delete texture_cache);
  CHK (delete lightmap_cache);
}

STDMETHODIMP csGraphics3DOpenGL::SetDimensions (int width, int height)
{
  csGraphics3DOpenGL::width = width;
  csGraphics3DOpenGL::height = height;
  csGraphics3DOpenGL::width2 = width/2;
  csGraphics3DOpenGL::height2 = height/2;
/*  if (z_buffer) CHKB (delete [] z_buffer);
  CHK (z_buffer = new unsigned long [width*height]);
  z_buf_size = sizeof (unsigned long)*width*height;*/

  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::BeginDraw (int DrawFlags)
{
  //ASSERT( m_piG2D );

  // if 2D graphics is not locked, lock it
  if ((DrawFlags & (CSDRAW_2DGRAPHICS | CSDRAW_3DGRAPHICS))
   && (!(DrawMode & (CSDRAW_2DGRAPHICS | CSDRAW_3DGRAPHICS))))
  {
    if (FAILED (m_piG2D->BeginDraw()))
      return E_UNEXPECTED;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glOrtho(0.,(GLdouble)width,0.,(GLdouble)height,0.001,1.0);
    glOrtho(0.,(GLdouble)width,0.,(GLdouble)height,-1.0,100.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor3f(1.,0.,0.);
    glClearColor(0.,1.,0.,0.);
  }

  if (DrawFlags & CSDRAW_CLEARZBUFFER)
  {
    if (DrawFlags & CSDRAW_CLEARSCREEN)
      glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    else
      glClear(GL_DEPTH_BUFFER_BIT);
  }
  else if (DrawFlags & CSDRAW_CLEARSCREEN)
    m_piG2D->Clear(0);
  
  DrawMode = DrawFlags;
  
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::FinishDraw ()
{
  //ASSERT( m_piG2D );

  if (DrawMode & (CSDRAW_2DGRAPHICS | CSDRAW_3DGRAPHICS))
  {
    m_piG2D->FinishDraw ();
  }
  DrawMode = 0;
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::Print (csRect *area)
{
  m_piG2D->Print (area);
  glClear(GL_COLOR_BUFFER_BIT);
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::SetZBufMode (ZBufMode mode)
{
  z_buf_mode = mode;
  return S_OK;
}

#define SMALL_D 0.01

float global_deltaX = 0;

/**
 * The engine often generates "empty" polygons, for example
 * (2, 2) - (317,2) - (317,2) - (2, 2)
 * To avoid too much computations, DrawPolygon detects such polygons by
 * counting the number of "real" vertices (i.e. the number of vertices,
 * distance between which is bigger that some amount). The "right" formula
 * for distance is sqrt(dX^2 + dY^2) but to avoid root and multiply
 * DrawPolygon checks abs(dX) + abs(dY). This is enough.
 */
#define VERTEX_NEAR_THRESHOLD   0.001

STDMETHODIMP csGraphics3DOpenGL::DrawPolygon (G3DPolygon& poly)
{
  int i;
  float P1, P2, P3, P4;
  float Q1, Q2, Q3, Q4;
  int max_i, min_i;
  float max_y, min_y;
  float min_z, max_z;
  unsigned char *d;
  
  IGraphicsInfo* piGI = NULL;
  int gi_pixelbytes;
  float inv_aspect;

  if (poly.num < 3) return S_FALSE;

  //ASSERT( m_piCamera );
  m_piCamera->GetInvAspect( inv_aspect );
  
  VERIFY_SUCCESS( m_piG2D->QueryInterface( (REFIID)IID_IGraphicsInfo, (void**)&piGI) );
  piGI->GetPixelBytes( gi_pixelbytes );
  FINAL_RELEASE( piGI );
  
  // Get the plane normal of the polygon. Using this we can calculate
  // '1/z' at every screen space point.
  float Ac, Bc, Cc, Dc, inv_Dc;
  Ac = poly.normal.A;
  Bc = poly.normal.B;
  Cc = poly.normal.C;
  Dc = poly.normal.D;

  float M, N, O;
  if (ABS (Dc) < SMALL_D)
  {

    ComVector3 vcam_0;
    poly.polygon->GetCameraVector(0, &vcam_0);

    // The Dc component of the plane normal is too small. This means that
    // the plane of the polygon is almost perpendicular to the eye of the
    // viewer. In this case, nothing much can be seen of the plane anyway
    // so we just take one value for the entire polygon.
    M = 0;
    N = 0;
    // For O choose the transformed z value of one vertex.
    // That way Z buffering should at least work.
    O = 1/vcam_0.z;
  }
  else
  {
    inv_Dc = 1/Dc;
    M = -Ac*inv_Dc*inv_aspect;
    N = -Bc*inv_Dc*inv_aspect;
    O = -Cc*inv_Dc;
  }
  // Compute the min_y and max_y for this polygon in screen space coordinates.
  // We are going to use these to scan the polygon from top to bottom.
  // Also compute the min_z/max_z in camera space coordinates. This is going to be
  // used for mipmapping.
  min_i = max_i = 0;
  min_y = max_y = poly.vertices[0].sy;
  max_z = min_z = M * (poly.vertices[0].sx - width2)
                + N * (poly.vertices[0].sy - height2) + O;
  // count 'real' number of vertices
  int num_vertices = 1;
  for (i = 1 ; i < poly.num ; i++)
  {
    if (poly.vertices[i].sy > max_y)
    {
      max_y = poly.vertices[i].sy;
      max_i = i;
    }
    else if (poly.vertices[i].sy < min_y)
    {
      min_y = poly.vertices[i].sy;
      min_i = i;
    }
    float inv_z = M * (poly.vertices[i].sx - width2)
                + N * (poly.vertices[i].sy - height2) + O;
    if (inv_z > min_z) min_z = inv_z;
    if (inv_z < max_z) max_z = inv_z;
    // theoretically we should do here sqrt(dx^2+dy^2), but
    // we can approximate it just by abs(dx)+abs(dy)
    if ((fabs (poly.vertices [i].sx - poly.vertices [i - 1].sx)
       + fabs (poly.vertices [i].sy - poly.vertices [i - 1].sy)) > VERTEX_NEAR_THRESHOLD)
      num_vertices++;
  }

  // if this is a 'degenerate' polygon, skip it
  if (num_vertices < 3) return S_FALSE;

  // Mipmapping.
  IPolygonTexture* tex;
  if (poly.polygon->UsesMipMaps () == S_FALSE  ||  rstate_mipmap == 1)
    poly.polygon->GetTexture (0, &tex);
  else if (rstate_mipmap == 0)
    poly.polygon->GetTextureByZ ((float)((min_z + max_z) / (2. * min_z * max_z)), &tex); // @@@
  else
    poly.polygon->GetTexture(rstate_mipmap - 1, &tex);

  // Initialize our static drawing information and cache
  // the texture in the texture cache (if this is not already the case).
  // If we are using the sub-texture optimization then we just allocate
  // the needed memory in the cache but don't do any calculations yet.
  int subtex_size;
  tex->GetSubtexSize (subtex_size);
/*  if (subtex_size) CacheInitTexture (tex);
  else*/ CacheTexture (tex);

  // @@@ The texture transform matrix is currently written as T = M*(C-V)
  // (with V being the transform vector, M the transform matrix, and C
  // the position in camera space coordinates. It would be better (more
  // suitable for the following calculations) if it would be written
  // as T = M*C - V.
  P1 = poly.plane.m_cam2tex->m11;
  P2 = poly.plane.m_cam2tex->m12;
  P3 = poly.plane.m_cam2tex->m13;
  P4 = - (P1 * poly.plane.v_cam2tex->x
        + P2 * poly.plane.v_cam2tex->y
        + P3 * poly.plane.v_cam2tex->z);
  Q1 = poly.plane.m_cam2tex->m21;
  Q2 = poly.plane.m_cam2tex->m22;
  Q3 = poly.plane.m_cam2tex->m23;
  Q4 = - (Q1 * poly.plane.v_cam2tex->x
        + Q2 * poly.plane.v_cam2tex->y
        + Q3 * poly.plane.v_cam2tex->z);

/*
  P1 *= Scan::tw; P2 *= Scan::tw; P3 *= Scan::tw; P4 *= Scan::tw;
  Q1 *= Scan::th; Q2 *= Scan::th; Q3 *= Scan::th; Q4 *= Scan::th;
  P4 -= Scan::fdu; Q4 -= Scan::fdv;
  */

  // Precompute everything so that we can calculate (u,v) (texture space
  // coordinates) for every (sx,sy) (screen space coordinates). We make
  // use of the fact that 1/z, u/z and v/z are linear in screen space.
  float J1, J2, J3, K1, K2, K3;
  if (ABS (Dc) < SMALL_D)
  {
    // The Dc component of the plane of the polygon is too small.
    J1 = J2 = J3 = 0;
    K1 = K2 = K3 = 0;
  } else
  {
    J1 = P1 * inv_aspect + P4 * M;
    J2 = P2 * inv_aspect + P4 * N;
    J3 = P3              + P4 * O;
    K1 = Q1 * inv_aspect + Q4 * M;
    K2 = Q2 * inv_aspect + Q4 * N;
    K3 = Q3              + Q4 * O;
  }
  
  
  // Select the right scanline drawing function.
//  ITextureContainer* txt;
//  world->GetTextures( &txt );
  
  int tex_transp, poly_fog, poly_alpha;
  
//  Scan::texture->GetTransparent( tex_transp );
  poly.polygon->GetAlpha( poly_alpha );

  HighColorCache_Data *cachedata;
  ITextureMap *thetexturemap;
  tex->GetTexture(&thetexturemap);
  IMipMapContainer *themipmap;
  thetexturemap->GetParent(&themipmap);
  themipmap->GetHighColorCache(&cachedata);
  themipmap->GetTransparent(tex_transp);
  GLuint texturehandle = *( (GLuint *) (cachedata->pData) );

  glBindTexture(GL_TEXTURE_2D,texturehandle);
  glShadeModel(GL_FLAT);
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

  if (z_buf_mode == ZBuf_None)
    glDisable(GL_DEPTH_TEST);
  else
  {
    glEnable(GL_DEPTH_TEST);
    if (z_buf_mode == ZBuf_Fill)
      glDepthFunc(GL_ALWAYS);
    else
      glDepthFunc(GL_GREATER);
  }

  if (poly_alpha > 0  )
  {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1.,1.,1.,poly_alpha/100.0);
  }
  else
  {
    glDisable(GL_BLEND);
    glColor4f(1.,1.,1.,0.);
  }

  glBegin(GL_TRIANGLE_FAN);
  float sx,sy,sz, one_over_sz, u_over_sz, v_over_sz;
  for (i=0; i<poly.num; i++)
  {
    sx = poly.vertices[i].sx - width2;
    sy = poly.vertices[i].sy - height2;
    one_over_sz = M*sx + N*sy + O; 
    sz = 1.0/one_over_sz;
    u_over_sz = (J1 * sx + J2 * sy + J3);
    v_over_sz = (K1 * sx + K2 * sy + K3);
    // we must communicate the perspective correction (1/z) for textures
    // by using homogenous coordinates in either texture space
    // or in object (vertex) space.  We do it in texture space.
    glTexCoord4f(u_over_sz,v_over_sz,one_over_sz,one_over_sz);
    glVertex3f(poly.vertices[i].sx,
		poly.vertices[i].sy, -one_over_sz);
  }
  glEnd();

finish:
  FINAL_RELEASE( tex );
  //FINAL_RELEASE( txt );
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::DrawPolygonDebug (G3DPolygon& poly)
{
#if 0
  int i;
  float P1, P2, P3, P4;
  float Q1, Q2, Q3, Q4;
  int max_i, min_i;
  float max_y, min_y;
  float min_z, max_z;

  IGraphicsInfo* piGI;
  int gi_pixelbytes;

  VERIFY_SUCCESS( m_piG2D->QueryInterface( (REFIID)IID_IGraphicsInfo, (void**)&piGI) );
  piGI->GetPixelBytes( gi_pixelbytes );
  FINAL_RELEASE( piGI );

  if (poly.num < 3) return S_FALSE;

  char* szName;
  poly.polygon->GetName(&szName);

  //ASSERT( m_piCamera );

  float inv_aspect, aspect;
  m_piCamera->GetInvAspect( inv_aspect );
  m_piCamera->GetAspect( aspect );

  SysPrintf (MSG_DEBUG_0, "------------------------------------------------------\n");
  SysPrintf (MSG_DEBUG_0, "Drawing polygon '%s' with %d vertices.\n", szName, poly.num);

  // Get the plane normal of the polygon. Using this we can calculate
  // '1/z' at every screen space point.
  float Ac, Bc, Cc, Dc, inv_Dc;
  Ac = poly.normal.A;
  Bc = poly.normal.B;
  Cc = poly.normal.C;
  Dc = poly.normal.D;

  SysPrintf (MSG_DEBUG_0, "Plane equation (camera space) of polygon: (%f,%f,%f,%f).\n",
             Ac, Bc, Cc, Dc);
  SysPrintf (MSG_DEBUG_0, "Camera aspect ratio: %f (and 1/aspect=%f)\n", aspect, inv_aspect);

  float M, N, O;
  if (ABS (Dc) < SMALL_D)
  {
    ComVector3 vcam_0;
    poly.polygon->GetCameraVector(0, &vcam_0);

    // The Dc component of the plane normal is too small. This means that
    // the plane of the polygon is almost perpendicular to the eye of the
    // viewer. In this case, nothing much can be seen of the plane anyway
    // so we just take one value for the entire polygon.
    M = 0;
    N = 0;
    // For O choose the transformed z value of one vertex.
    // That way Z buffering should at least work.
    O = 1/vcam_0.z;
    SysPrintf (MSG_DEBUG_0, "D component of plane is very small (< %f).\n", SMALL_D);
  } else
  {
    inv_Dc = 1/Dc;
    M = -Ac*inv_Dc*inv_aspect;
    N = -Bc*inv_Dc*inv_aspect;
    O = -Cc*inv_Dc;
  }
  SysPrintf (MSG_DEBUG_0, "M=%f, N=%f, O=%f\n", M, N, O);

  // Compute the min_y and max_y for this polygon in screen space coordinates.
  // We are going to use these to scan the polygon from top to bottom.
  // Also compute the min_z/max_z in camera space coordinates. This is going to be
  // used for mipmapping.
  min_i = max_i = 0;
  min_y = max_y = poly.vertices[0].sy;
  max_z = min_z = M * (poly.vertices[0].sx - width2)
                + N * (poly.vertices[0].sy - height2) + O;
  // count 'real' number of vertices
  int num_vertices = 1;
  for (i = 1 ; i < poly.num ; i++)
  {
    if (poly.vertices[i].sy > max_y)
    {
      max_y = poly.vertices[i].sy;
      max_i = i;
    }
    else if (poly.vertices[i].sy < min_y)
    {
      min_y = poly.vertices[i].sy;
      min_i = i;
    }
    float inv_z = M * (poly.vertices[i].sx - width2)
                + N * (poly.vertices[i].sy - height2) + O;
    if (inv_z > min_z) min_z = inv_z;
    if (inv_z < max_z) max_z = inv_z;
    // theoretically we should do here sqrt(dx^2+dy^2), but
    // we can approximate it just by abs(dx)+abs(dy)
    if ((fabs (poly.vertices [i].sx - poly.vertices [i - 1].sx)
       + fabs (poly.vertices [i].sy - poly.vertices [i - 1].sy)) > VERTEX_NEAR_THRESHOLD)
      num_vertices++;
  }
  SysPrintf (MSG_DEBUG_0, "Minimum y in screen space: %f\n", min_y);
  SysPrintf (MSG_DEBUG_0, "Maximum y in screen space: %f\n", max_y);
  SysPrintf (MSG_DEBUG_0, "Minimum 1/z: %f\n", max_z);
  SysPrintf (MSG_DEBUG_0, "Maximum 1/z: %f\n", min_z);

  // if this is a 'degenerate' polygon, skip it
  if (num_vertices < 3) return S_FALSE;

  // Mipmapping.
  IPolygonTexture* tex;
  IMipMapContainer* piMMC;
  ITextureMap* piTM;

  if (poly.polygon->UsesMipMaps () == S_FALSE  ||  rstate_mipmap == 1)
    poly.polygon->GetTexture (0, &tex);
  else if (rstate_mipmap == 0)
    poly.polygon->GetTextureByZ ((float)((min_z + max_z) / (2. * min_z * max_z)), &tex); // @@@
  else
    poly.polygon->GetTexture(rstate_mipmap - 1, &tex);

  // Initialize our static drawing information and cache
  // the texture in the texture cache (if this is not already the case).
  CacheTexture (tex);

  tex->GetTexture( &piTM );
  //ASSERT( piTM );

  piTM->GetParent( &piMMC );
  //ASSERT( piMMC );

  int nWidth, nHeight;

  piMMC->GetName(&szName);
  
  piTM->GetWidth(nWidth);
  piTM->GetHeight(nHeight);

  SysPrintf (MSG_DEBUG_0, "Drawing with texture '%s' (w=%d, h=%d)\n", szName, nWidth, nHeight);

  FINAL_RELEASE( piMMC );
  FINAL_RELEASE( piTM );

  Scan::init_draw (poly.polygon, tex);

  // @@@ The texture transform matrix is currently written as T = M*(C-V)
  // (with V being the transform vector, M the transform matrix, and C
  // the position in camera space coordinates. It would be better (more
  // suitable for the following calculations) if it would be written
  // as T = M*C - V.
  P1 = poly.plane.m_cam2tex->m11;
  P2 = poly.plane.m_cam2tex->m12;
  P3 = poly.plane.m_cam2tex->m13;
  P4 = - (P1 * poly.plane.v_cam2tex->x
        + P2 * poly.plane.v_cam2tex->y
        + P3 * poly.plane.v_cam2tex->z);
  Q1 = poly.plane.m_cam2tex->m21;
  Q2 = poly.plane.m_cam2tex->m22;
  Q3 = poly.plane.m_cam2tex->m23;
  Q4 = - (Q1 * poly.plane.v_cam2tex->x
        + Q2 * poly.plane.v_cam2tex->y
        + Q3 * poly.plane.v_cam2tex->z);

  P1 *= Scan::tw; P2 *= Scan::tw; P3 *= Scan::tw; P4 *= Scan::tw;
  Q1 *= Scan::th; Q2 *= Scan::th; Q3 *= Scan::th; Q4 *= Scan::th;
  P4 -= Scan::fdu; Q4 -= Scan::fdv;
  SysPrintf (MSG_DEBUG_0, "P1=%f P2=%f P3=%f P4=%f\n", P1, P2, P3, P4);
  SysPrintf (MSG_DEBUG_0, "Q1=%f Q2=%f Q3=%f Q4=%f\n", Q1, Q2, Q3, Q4);

  // Precompute everything so that we can calculate (u,v) (texture space
  // coordinates) for every (sx,sy) (screen space coordinates). We make
  // use of the fact that 1/z, u/z and v/z are linear in screen space.
  float J1, J2, J3, K1, K2, K3;
  if (ABS (Dc) < SMALL_D)
  {
    // The Dc component of the plane of the polygon is too small.
    J1 = 0;
    J2 = 0;
    J3 = 0;
    K1 = 0;
    K2 = 0;
    K3 = 0;
  } else
  {
    J1 = P1 * inv_aspect + P4 * M;
    J2 = P2 * inv_aspect + P4 * N;
    J3 = P3                      + P4 * O;
    K1 = Q1 * inv_aspect + Q4 * M;
    K2 = Q2 * inv_aspect + Q4 * N;
    K3 = Q3                      + Q4 * O;
  }
  SysPrintf (MSG_DEBUG_0, "J1=%f J2=%f J3=%f\n", J1, J2, J3);
  SysPrintf (MSG_DEBUG_0, "K1=%f K2=%f K3=%f\n", K1, K2, K3);

  // Steps for interpolating horizontally accross a scanline.
  Scan::M = M;
  Scan::J1 = J1;
  Scan::K1 = K1;
  Scan::dM = M*Scan::INTERPOL_STEP;
  Scan::dJ1 = J1*Scan::INTERPOL_STEP;
  Scan::dK1 = K1*Scan::INTERPOL_STEP;

  // Select the right scanline drawing function.
  SysPrintf (MSG_DEBUG_0, "Scanline drawer: ");
  ITextureContainer* txt;
  world->GetTextures(&txt);
  
  int tex_transp, poly_fog, poly_alpha;
  
  Scan::texture->GetTransparent( tex_transp );
  poly.polygon->GetAlpha( poly_alpha );
  poly.polygon->GetFog( poly_fog );

  if (gi_pixelbytes > 1)
  {
    // Truecolor mode
    if (Scan::tmap2)
    {
      if (z_buf_mode == ZBuf_Use)
        SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_z_buf_map");
      else if (do_transp && tex_transp != -1)
        SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_transp_map");
      else if (do_transp && poly_alpha)
      {
        if (poly_alpha == 25)
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_map_alpha25");
        else if (poly_alpha == 50)
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_map_alpha50");
        else
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_map_alpha75");
      }
      else
        SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_map");
    }
    else
    {
      int txtMode;
      txt->GetMode(txtMode);

      if (txtMode == TXT_PRIVATE)
      {
        if (z_buf_mode == ZBuf_Use)
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_z_buf_private");
        else
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_private");
      }
      else
      {
        if (z_buf_mode == ZBuf_Use)
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline_z_buf");
        else
          SysPrintf (MSG_DEBUG_0, "Scan16::draw_scanline");
      }
    }
  } else if (do_textured)
  {
    if (z_buf_mode == ZBuf_Use)
    {
      if (Scan::tmap2)
      {
        // DAN: commented this out until we get these working through COM.
/*        if (poly.polygon->theDynLight) // Use a dynamic light if it exists
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_z_buf_map_light");
        } else*/
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_z_buf_map");
      } else
      {
        int txtMode;
        txt->GetMode(txtMode);

        if (txtMode == TXT_PRIVATE)
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_z_buf_private");
        } else
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_z_buf");
      }
    } else if (do_transp && tex_transp != -1)
    {
      if (Scan::tmap2)
        SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_transp_map");
      else
      {
        int txtMode;
        txt->GetMode(txtMode);

        if (txtMode == TXT_PRIVATE)
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_transp_private");
        } else
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_transp");
      }
    } else if (Scan::tmap2)
    {
      // DAN: commented this out until these are supported through COM.
/*      if (poly.polygon->theDynLight) // Use a dynamic light if it exists
      {
        SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map_light");
      } else */
      if (do_transp && poly_alpha)
      {
        if (poly_alpha == 25)
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map_alpha1");
        } else if (poly_alpha == 50)
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map_alpha1");
        } else
        {
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map_alpha2");
        }
      } else
      {
        if (do_perfect)
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map_perfect");
        else
          SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_map");
      }
    } else
    {
      int txtMode;
      txt->GetMode(txtMode);

      if (txtMode == TXT_PRIVATE)
      {
        SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_private");
      } else
        SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline");
    }
  } else
  {
    if (z_buf_mode == ZBuf_Use)
      SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_z_buf_flat");
    else
      SysPrintf (MSG_DEBUG_0, "Scan::draw_scanline_flat");
  }

  // Scan both sides of the polygon at once.
  // We start with two pointers at the top (as seen in y-inverted
  // screen-space: bottom of display) and advance them until both
  // join together at the bottom. The way this algorithm works, this
  // should happen automatically; the left pointer is only advanced
  // when it is further away from the bottom than the right pointer
  // and vice versa.
  // Using this we effectively partition our polygon in trapezoids
  // with at most two triangles (one at the top and one at the bottom).

  int scanL1, scanL2, scanR1, scanR2;   // scan vertex left/right start/final
  float sxL, sxR, dxL, dxR;             // scanline X left/right and deltas
  int sy, fyL, fyR;                     // scanline Y, final Y left, final Y right
  int xL, xR;

  sxL = sxR = dxL = dxR = 0;            // avoid GCC warnings about "uninitialized variables"
  scanL2 = scanR2 = max_i;
  sy = fyL = fyR = QRound (poly.vertices [scanL2].sy);

  for ( ; ; )
  {
    //-----
    // We have reached the next segment. Recalculate the slopes.
    //-----
    bool leave;
    do
    {
      leave = true;
      if (sy <= fyR)
      {
        // Check first if polygon has been finished
        if (scanR2 == min_i)
          goto finish;
        scanR1 = scanR2;
        scanR2 = (scanR2 + 1) % poly.num;

        fyR = QRound (poly.vertices [scanR2].sy);
        float dyR = (poly.vertices [scanR1].sy - poly.vertices [scanR2].sy);
        if (dyR > 0)
        {
          sxR = poly.vertices [scanR1].sx;
          dxR = (poly.vertices [scanR2].sx - sxR) / dyR;
          // horizontal pixel correction
          sxR += dxR * (poly.vertices [scanR1].sy - ((float)sy - 0.5));
        } /* endif */
        leave = false;
      } /* endif */
      if (sy <= fyL)
      {
        scanL1 = scanL2;
        scanL2 = (scanL2 - 1 + poly.num) % poly.num;

        fyL = QRound (poly.vertices [scanL2].sy);
        float dyL = (poly.vertices [scanL1].sy - poly.vertices [scanL2].sy);
        if (dyL)
        {
          sxL = poly.vertices [scanL1].sx;
          dxL = (poly.vertices [scanL2].sx - sxL) / dyL;
          // horizontal pixel correction
          sxL += dxL * (poly.vertices [scanL1].sy - ((float)sy - 0.5));
        } /* endif */
        leave = false;
      } /* endif */
    } while (!leave); /* enddo */

    //@@debug
    sxL += global_deltaX;

    // Steps for interpolating vertically over scanlines.
    float vd_inv_z = - dxL * M + N;
    float vd_u_div_z = - dxL * J1 + J2;
    float vd_v_div_z = - dxL * K1 + K2;

    float cx = (float)(QRound (sxL) - width2);
    float cy = (float)((sy - 0.5) - height2);
    float inv_z = M * cx + N * cy + O;
    float u_div_z = J1 * cx + J2 * cy + J3;
    float v_div_z = K1 * cx + K2 * cy + K3;

    // Find the trapezoid top (or bottom in inverted Y coordinates)
    int fin_y;
    if (fyL > fyR)
      fin_y = fyL;
    else
      fin_y = fyR;

    SysPrintf (MSG_DEBUG_0, "Draw trapezoid:\n");
    SysPrintf (MSG_DEBUG_0, "  startY = %8.4f  leftX   = %8.4f  rightX  = %8.4f\n", (float)sy, sxL, sxR);
    SysPrintf (MSG_DEBUG_0, "  finalY - %8.4f  deltaLX = %8.4f  deltaRX = %8.4f\n", (float)fin_y, dxL, dxR);
    SysPrintf (MSG_DEBUG_0, "  1/Z    = %8.4f  U/Z     = %8.4f  V/Z     = %8.4f\n", inv_z, u_div_z, v_div_z);
    SysPrintf (MSG_DEBUG_0, "  deltaZ = %8.4f  deltaU  = %8.4f  deltaV  = %8.4f\n", vd_inv_z, vd_u_div_z, vd_v_div_z);

    int screenY = height - 1 - sy;
    while (sy > fin_y)
    {
      if ((sy & 1) != cfg_interlacing)
      {
        // Compute the rounded screen coordinates of horizontal strip
        xL = QRound (sxL);
        xR = QRound (sxR);

        // Sub-pixel U & V correction
        float deltaX = (float)xL - sxL;

        // do not draw the rightmost pixel - it will be covered
        // by neightbour polygon's left bound
        SysPrintf (MSG_DEBUG_0, "scanline (x = %d, y = %d, len = %d, z = %.2f, u = %.2f, v = %.2f)\n",
          xL, sy, xR - xL, inv_z + deltaX * M, u_div_z + deltaX * J1, v_div_z + deltaX * K1);
      }

      sxL += dxL;
      sxR += dxR;
      inv_z -= vd_inv_z;
      u_div_z -= vd_u_div_z;
      v_div_z -= vd_v_div_z;
      sy--;
      screenY++;
    } /* endwhile */
  } /* endfor */

finish:
  FINAL_RELEASE( tex );
  FINAL_RELEASE( txt );
#endif

  return S_OK;
}

// Calculate round (f) of a 16:16 fixed pointer number
// and return a long integer.
inline long round16 (long f)
{
  return (f + 0x8000) >> 16;
}

STDMETHODIMP csGraphics3DOpenGL::DrawPolygonQuick (G3DPolygon& poly, bool gouroud)
{
  texture_cache->Add(poly.pi_texture);

  HighColorCache_Data *cachedata;
  poly.pi_texture->GetHighColorCache(&cachedata);
  GLuint texturehandle = *( (GLuint *) (cachedata->pData) );

  glBindTexture(GL_TEXTURE_2D,texturehandle);
  glShadeModel(GL_FLAT);
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

  if (z_buf_mode == ZBuf_None)
    glDisable(GL_DEPTH_TEST);
  else
  {
    glEnable(GL_DEPTH_TEST);
    if (z_buf_mode == ZBuf_Fill)
      glDepthFunc(GL_ALWAYS);
    else
      glDepthFunc(GL_GREATER);
  }

  glBegin(GL_TRIANGLE_FAN);
  float sx,sy;
  for (int i=0; i<poly.num; i++)
  {
    sx = poly.vertices[i].sx - width2;
    sy = poly.vertices[i].sy - height2;
    glTexCoord2f(poly.pi_texcoords[i].u,poly.pi_texcoords[i].v);
    glVertex3f(poly.vertices[i].sx,poly.vertices[i].sy,-1./poly.pi_texcoords[i].z);
  }
  glEnd();

#if 0
  txt->GetBitmap(&bm);
  txt->GetWidth(itw);
  txt->GetHeight(ith);
  txt->GetWidthShift(shf_w);

  float tw = (float)itw; 
  float th = (float)ith; 
  int twfp = QInt16 (tw);
  int thfp = QInt16 (th);
  int i;

  IGraphicsInfo* piGI;
  int gi_pixelbytes;

  VERIFY_SUCCESS( m_piG2D->QueryInterface( (REFIID)IID_IGraphicsInfo, (void**)&piGI) );
  piGI->GetPixelBytes( gi_pixelbytes );
  FINAL_RELEASE( piGI );

  //-----
  // Get the values from the polygon for more conveniant local access.
  // Also look for the top-most and bottom-most vertices.
  //-----
  float uu[64], vv[64], iz[64];
  int top, bot;
  float top_y = -99999;
  float bot_y = 99999;
  top = bot = 0;                        // avoid GCC complains
  for (i = 0 ; i < poly.num ; i++)
  {
    uu[i] = tw * poly.pi_texcoords [i].u;
    vv[i] = th * poly.pi_texcoords [i].v;
    iz[i] = poly.pi_texcoords [i].z;
    if (poly.vertices [i].sy > top_y)
    {
      top_y = poly.vertices [i].sy;
      top = i;
    } /* endif */
    if (poly.vertices [i].sy < bot_y)
    {
      bot_y = poly.vertices [i].sy;
      bot = i;
    } /* endif */
  }

  //-----
  // Calculate constant du,dv,dz.
  //
  //          (At-Ct)*(By-Cy) - (Bt-Ct)*(Ay-Cy)
  //  dt/dx = ---------------------------------
  //          (Ax-Cx)*(By-Cy) - (Bx-Cx)*(Ay-Cy)
  //-----
  float dd = (poly.pi_triangle [0].x - poly.pi_triangle [2].x)
           * (poly.pi_triangle [1].y - poly.pi_triangle [2].y)
           - (poly.pi_triangle [1].x - poly.pi_triangle [2].x)
           * (poly.pi_triangle [0].y - poly.pi_triangle [2].y);
  float uu0 = tw * poly.pi_tritexcoords [0].u;
  float uu1 = tw * poly.pi_tritexcoords [1].u;
  float uu2 = tw * poly.pi_tritexcoords [2].u;
  int du = QInt16 (((uu0 - uu2) * (poly.pi_triangle [1].y - poly.pi_triangle [2].y)
                  - (uu1 - uu2) * (poly.pi_triangle [0].y - poly.pi_triangle [2].y)) / dd);
  float vv0 = th * poly.pi_tritexcoords [0].v;
  float vv1 = th * poly.pi_tritexcoords [1].v;
  float vv2 = th * poly.pi_tritexcoords [2].v;
  int dv = QInt16 (((vv0 - vv2) * (poly.pi_triangle [1].y - poly.pi_triangle [2].y)
                  - (vv1 - vv2) * (poly.pi_triangle [0].y - poly.pi_triangle [2].y)) / dd);
  float iz0 = poly.pi_tritexcoords [0].z;
  float iz1 = poly.pi_tritexcoords [1].z;
  float iz2 = poly.pi_tritexcoords [2].z;
  int dz = QInt24 (((iz0 - iz2) * (poly.pi_triangle [1].y - poly.pi_triangle [2].y)
                  - (iz1 - iz2) * (poly.pi_triangle [0].y - poly.pi_triangle [2].y)) / dd);


  // Select draw scanline routine
  void (*drawline) (void *dest, int len, long *zbuff, long u, long du, long v,
    long dv, long z, long dz, unsigned char *bitmap, int bitmap_log2w);

  
  if (gi_pixelbytes > 1)
  {
    ITextureContainer* piTC;
    TextureTablesPalette* lt_pal;
      
    world->GetTextures(&piTC);
    //ASSERT( piTC );

    piTC->GetPaletteLUT(&lt_pal);
    FINAL_RELEASE(piTC);
      
    Scan16::pal_table = lt_pal->pal_to_true;

#   ifdef DO_MMX
//    if (cpu_mmx && do_mmx)
//      drawline = Scan16::mmx_draw_pi_scanline;
//    else
#   endif
      drawline = Scan16::draw_pi_scanline;
  } else
#   ifdef DO_MMX
    if (cpu_mmx && do_mmx)
      drawline = Scan::mmx_draw_pi_scanline;
    else
#   endif
      drawline = Scan::draw_pi_scanline;

  //-----
  // Scan from top to bottom.
  // scanL1-scanL2 is the left segment to scan.
  // scanR1-scanR2 is the right segment to scan.
  //-----
  int scanL1, scanL2 = top;
  int scanR1, scanR2 = top;

  int xL = 0, xR = 0, dxdyL = 0, dxdyR = 0;
  int uL = 0, vL = 0, zL = 0;
  int dudyL = 0, dvdyL = 0, dzdyL = 0;

  int sy, fyL, fyR;
  sy = fyL = fyR = QRound (poly.vertices [top].sy);

  //-----
  // The loop.
  //-----
  for ( ; ; )
  {
    //-----
    // We have reached the next segment. Recalculate the slopes.
    //-----
    bool leave;
    do
    {
      leave = true;
      if (sy <= fyR)
      {
        //-----
        // Right side needs to be advanced.
        //-----
        // Check first if polygon has been finished
        if (scanR2 == bot)
          goto finish;
        scanR1 = scanR2;
        scanR2 = (scanR2 + 1) % poly.num;

        fyR = QRound (poly.vertices [scanR2].sy);
        float dyR = poly.vertices [scanR1].sy - poly.vertices [scanR2].sy;
        if (dyR > 0)
        {
          xR = QInt16 (poly.vertices [scanR1].sx);
          dxdyR = QInt16 ((poly.vertices [scanR2].sx - poly.vertices [scanR1].sx) / dyR);
          // horizontal pixel correction
          xR += QRound (dxdyR * (poly.vertices [scanR1].sy -
            ((float)QRound (poly.vertices [scanR1].sy) - 0.5)));
        } /* endif */
        leave = false;
      } /* endif */
      if (sy <= fyL)
      {
        //-----
        // Left side needs to be advanced.
        //-----
        scanL1 = scanL2;
        scanL2 = (scanL2 - 1 + poly.num) % poly.num;

        fyL = QRound (poly.vertices [scanL2].sy);
        float dyL = poly.vertices [scanL1].sy - poly.vertices [scanL2].sy;
        if (dyL > 0)
        {
          dxdyL = QInt16 ((poly.vertices [scanL2].sx - poly.vertices [scanL1].sx) / dyL);
          dudyL = QInt16 ((uu[scanL2] - uu[scanL1]) / dyL);
          dvdyL = QInt16 ((vv[scanL2] - vv[scanL1]) / dyL);
          dzdyL = QInt24 ((iz[scanL2] - iz[scanL1]) / dyL);
          xL = QInt16 (poly.vertices [scanL1].sx);

          // horizontal pixel correction
          float deltaY = poly.vertices [scanL1].sy - ((float)sy - 0.5);
          float deltaX = (dxdyL / 65536.) * deltaY;
          xL += QInt16 (deltaX);

          // apply sub-pixel accuracy factor
          float Factor;
          if (poly.vertices [scanL2].sx != poly.vertices [scanL1].sx)
            Factor = deltaX / (poly.vertices [scanL2].sx - poly.vertices [scanL1].sx);
          else
            Factor = 0;
          uL = QInt16 (uu [scanL1] + (uu [scanL2] - uu [scanL1]) * Factor);
          vL = QInt16 (vv [scanL1] + (vv [scanL2] - vv [scanL1]) * Factor);
          zL = QInt24 (iz [scanL1] + (iz [scanL2] - iz [scanL1]) * Factor);
        } /* endif */
        leave = false;
      } /* endif */
    } while (!leave); /* enddo */

    //@@debug
    xL += QInt16 (global_deltaX);

    //-----
    // Now draw a trapezoid.
    //-----
    int fin_y;
    if (fyL > fyR)
      fin_y = fyL;
    else
      fin_y = fyR;

    int screenY = height - 1 - sy;
    while (sy > fin_y)
    {
      if ((sy & 1) != cfg_interlacing)
      {
        //-----
        // Draw one scanline.
        //-----
        int xl = round16 (xL);
        int xr = round16 (xR);

        register long* zbuff = (long *)z_buffer + width * screenY + xl;

        // Check for texture overflows
        int uu = uL, vv = vL;
        int duu = du, dvv = dv;
        if (uu < 0) uu = 0; if (uu > twfp) uu = twfp;
        if (vv < 0) vv = 0; if (vv > thfp) vv = thfp;
        if (xr > xl)
        {
          int tmpu = uu + du * (xr - xl);
          int tmpv = vv + dv * (xr - xl);
          if (tmpu < 0 || tmpu > twfp)
          {
            if (tmpu < 0) tmpu = 0; if (tmpu > twfp) tmpu = twfp;
            duu = (tmpu - uu) / (xr - xl);
          } /* endif */
          if (tmpv < 0 || tmpv > thfp)
          {
            if (tmpv < 0) tmpv = 0; if (tmpv > thfp) tmpv = thfp;
            dvv = (tmpv - vv) / (xr - xl);
          } /* endif */
        } /* endif */

        unsigned char* pixel_at;
        m_piG2D->GetPixelAt(xl, screenY, &pixel_at);
        drawline (pixel_at, xr - xl, zbuff, uu, duu,
                  vv, dvv, zL, dz, bm, shf_w);
      } /* endif */

      xL += dxdyL;
      xR += dxdyR;
      uL += dudyL;
      vL += dvdyL;
      zL += dzdyL;
      sy--;
      screenY++;
    } /* endwhile */
  } /* endfor */

finish:
  if (rstate_edges)
  {
    ITextureContainer* piTC;
    int white;
    
    world->GetTextures(&piTC);
    piTC->GetWhite(white);
    FINAL_RELEASE( piTC );

    for (i = 0; i < poly.num; i++)
    {
      int next = (i == poly.num - 1 ? 0 : i + 1);
      m_piG2D->DrawLine (QRound (poly.vertices [i].sx),
                         height - 1 - QRound (poly.vertices [i].sy),
                         QRound (poly.vertices [next].sx),
                         height - 1 - QRound (poly.vertices [next].sy), white);
    } /* endfor */
  }
#endif
  return S_OK;
}

/**
 * Initiate a volumetric fog object. This function will be called
 * before front-facing and back-facing fog polygons are added to
 * the object. The fog object will be convex but not necesarily
 *         closed.
 * The given CS_ID can be used to identify multiple fog
 * objects when
 * multiple objects are started.
 */
STDMETHODIMP csGraphics3DOpenGL::OpenFogObject(CS_ID id, csFog *fog)
{
  return S_OK;
}

/**
 *   * Add a front or back-facing fog polygon in the current fog object.
 * Note that it is guaranteed that all back-facing fog polygons
 * will have been added before the first front-facing polygon.
 * fogtype can be:
 *    CS_FOG_FRONT:   a front-facing polygon
 *    CS_FOG_BACK:    a back-facing polygon
 *    CS_FOG_VIEW:    the view-plane
 */
STDMETHODIMP csGraphics3DOpenGL::AddFogPolygon(CS_ID id, G3DPolygon &poly, int fogtype)
{
  return S_OK;
}

/**
 * Close a volumetric fog object. After the volumetric object is
 * closed it should be rendered on screen (whether you do it here
 * or in DrawFrontFog/DrawBackFog is not important).
 */
STDMETHODIMP csGraphics3DOpenGL::CloseFogObject(CS_ID id)
{
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::CacheTexture (IPolygonTexture* texture)
{
  ITextureMap *thetexturemap;
  texture->GetTexture(&thetexturemap);
  IMipMapContainer *themipmap;
  thetexturemap->GetParent(&themipmap);
  texture_cache->Add(themipmap);
  lightmap_cache->Add(texture);
  return S_OK;
}

void csGraphics3DOpenGL::CacheInitTexture (IPolygonTexture* texture)
{
#if 0
  ITextureContainer* piTC;
  world->GetTextures (&piTC);
  tcache->init_texture (texture, piTC);
  piTC->Release ();
#endif
}

void csGraphics3DOpenGL::CacheSubTexture (IPolygonTexture* texture, int u, int v)
{
#if 0
  ITextureContainer* piTC;
  world->GetTextures (&piTC);
  tcache->use_sub_texture (texture, piTC, u, v);
  piTC->Release ();
#endif
}

void csGraphics3DOpenGL::CacheRectTexture (IPolygonTexture* tex,
	int minu, int minv, int maxu, int maxv)
{
#if 0
  ITextureContainer* piTC;
  world->GetTextures (&piTC);
  int subtex_size;
  tex->GetSubtexSize (subtex_size);

  int iu, iv;
  for (iu = minu ; iu < maxu ; iu += subtex_size)
  {
    for (iv = minv ; iv < maxv ; iv += subtex_size)
        tcache->use_sub_texture (tex, piTC, iu, iv);
    tcache->use_sub_texture (tex, piTC, iu, maxv);
  }
  for (iv = minv ; iv < maxv ; iv += subtex_size)
      tcache->use_sub_texture (tex, piTC, maxu, iv);
  tcache->use_sub_texture (tex, piTC, maxu, maxv);

  piTC->Release();
#endif
}

STDMETHODIMP csGraphics3DOpenGL::UncacheTexture (IPolygonTexture* texture)
{
  (void)texture;
  return E_NOTIMPL;
}

STDMETHODIMP csGraphics3DOpenGL::SetRenderState(G3D_RENDERSTATEOPTION op, long value)
{
  switch(op)
  {
    case G3DRENDERSTATE_NOTHING:
      return S_OK;

    case G3DRENDERSTATE_ZBUFFERTESTENABLE:
      if (value)
      {
         if (z_buf_mode == ZBuf_Test)
           return S_OK;
         if (z_buf_mode == ZBuf_None)
           z_buf_mode = ZBuf_Test;
         else if (z_buf_mode == ZBuf_Fill)
           z_buf_mode = ZBuf_Use;
      }
      else
      {
         if (z_buf_mode == ZBuf_Fill)
           return S_OK;
         if (z_buf_mode == ZBuf_Use)
           z_buf_mode = ZBuf_Fill;
         else if (z_buf_mode == ZBuf_Test)
           z_buf_mode = ZBuf_None;
      }
      break;
    case G3DRENDERSTATE_ZBUFFERFILLENABLE:
      if (value)
      {
        if (z_buf_mode == ZBuf_Fill)
          return S_OK;
        if (z_buf_mode == ZBuf_None)
          z_buf_mode = ZBuf_Fill;
        else if (z_buf_mode == ZBuf_Test)
          z_buf_mode = ZBuf_Use;
      }
      else
      {
        if (z_buf_mode == ZBuf_Test)
          return S_OK;
        if (z_buf_mode == ZBuf_Use)
          z_buf_mode = ZBuf_Test;
        else if (z_buf_mode == ZBuf_Fill)
          z_buf_mode = ZBuf_None;
      }
      break;
    case G3DRENDERSTATE_DITHERENABLE:
      rstate_dither = value;
      break;
    case G3DRENDERSTATE_SPECULARENABLE:
      rstate_specular = value;
      break;
    case G3DRENDERSTATE_BILINEARMAPPINGENABLE:
      rstate_bilinearmap = value;
      break;
    case G3DRENDERSTATE_TRILINEARMAPPINGENABLE:
      rstate_trilinearmap = value;
      break;
    case G3DRENDERSTATE_TRANSPARENCYENABLE:
      rstate_alphablend = value;
      break;
    case G3DRENDERSTATE_MIPMAPENABLE:
      rstate_mipmap = value;
      break;
    case G3DRENDERSTATE_TEXTUREMAPPINGENABLE:
      rstate_flat = !value;
      break;
    case G3DRENDERSTATE_EDGESENABLE:
      rstate_edges = value;
      break;
	// XAVIER: unhandled cases in enumerated switch (just to keep gcc happy)
    case G3DRENDERSTATE_DEBUGENABLE :
    case G3DRENDERSTATE_LIGHTFRUSTRUMENABLE :
    case G3DRENDERSTATE_FILTERINGENABLE :
    case G3DRENDERSTATE_PERFECTMAPPINGENABLE :
    case G3DRENDERSTATE_LIGHTINGENABLE :
    case G3DRENDERSTATE_INTERLACINGENABLE :
    case G3DRENDERSTATE_MMXENABLE :
      break;
    default:
      return E_INVALIDARG;
  }

  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::GetRenderState(G3D_RENDERSTATEOPTION op, long& retval)
{
  switch(op)
  {
    case G3DRENDERSTATE_NOTHING:
      retval = 0;
      break;
    case G3DRENDERSTATE_ZBUFFERTESTENABLE:
      retval = (bool)(z_buf_mode & ZBuf_Test);
      break;
    case G3DRENDERSTATE_ZBUFFERFILLENABLE:
      retval = (bool)(z_buf_mode & ZBuf_Fill);
      break;
    case G3DRENDERSTATE_DITHERENABLE:
      retval = rstate_dither;
      break;
    case G3DRENDERSTATE_SPECULARENABLE:
      retval = rstate_specular;
      break;
    case G3DRENDERSTATE_BILINEARMAPPINGENABLE:
      retval = rstate_bilinearmap;
      break;
    case G3DRENDERSTATE_TRILINEARMAPPINGENABLE:
      retval = rstate_trilinearmap;
      break;
    case G3DRENDERSTATE_TRANSPARENCYENABLE:
      retval = rstate_alphablend;
      break;
    case G3DRENDERSTATE_MIPMAPENABLE:
      retval = rstate_mipmap;
      break;
    case G3DRENDERSTATE_TEXTUREMAPPINGENABLE:
      retval = !rstate_flat;
      break;
    case G3DRENDERSTATE_EDGESENABLE:
      retval = rstate_edges;
      break;
    default:
      retval = 0;
      return E_INVALIDARG;
  }
  
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::GetCaps(G3D_CAPS *caps)
{
  if (!caps)
    return E_INVALIDARG;

  caps->ColorModel = G3DCOLORMODEL_RGB;
  caps->CanClip = false;
  caps->SupportsArbitraryMipMapping = true;
  caps->BitDepth = 24;
  caps->ZBufBitDepth = 16;
  caps->minTexHeight = 2;
  caps->minTexWidth = 2;
  caps->maxTexHeight = 1024;
  caps->maxTexWidth = 1024;
  caps->PrimaryCaps.RasterCaps = G3DRASTERCAPS_SUBPIXEL;
  caps->PrimaryCaps.canBlend = true;
  caps->PrimaryCaps.ShadeCaps = G3DRASTERCAPS_LIGHTMAP;
  caps->PrimaryCaps.PerspectiveCorrects = true;
  caps->PrimaryCaps.FilterCaps = G3D_FILTERCAPS((int)G3DFILTERCAPS_NEAREST | (int)G3DFILTERCAPS_MIPNEAREST);

  return 1;
}

STDMETHODIMP csGraphics3DOpenGL::ClearCache()
{
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::DumpCache()
{
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::DrawLine (Vector3& v1, Vector3& v2, int color)
{
#if 0
  //ASSERT( m_piCamera );
  if (v1.z < SMALL_Z && v2.z < SMALL_Z) return S_FALSE;

  float x1 = v1.x, y1 = v1.y, z1 = v1.z;
  float x2 = v2.x, y2 = v2.y, z2 = v2.z;

  if (z1 < SMALL_Z)
  {
    // x = t*(x2-x1)+x1;
    // y = t*(y2-y1)+y1;
    // z = t*(z2-z1)+z1;
    float t = (SMALL_Z-z1) / (z2-z1);
    x1 = t*(x2-x1)+x1;
    y1 = t*(y2-y1)+y1;
    z1 = SMALL_Z;
  }
  else if (z2 < SMALL_Z)
  {
    // x = t*(x2-x1)+x1;
    // y = t*(y2-y1)+y1;
    // z = t*(z2-z1)+z1;
    float t = (SMALL_Z-z1) / (z2-z1);
    x2 = t*(x2-x1)+x1;
    y2 = t*(y2-y1)+y1;
    z2 = SMALL_Z;
  }

  float aspect;
  m_piCamera->GetAspect( aspect );

  float iz1 = aspect/z1;
  int px1 = QInt (x1 * iz1 + (width/2));
  int py1 = height - 1 - QInt (y1 * iz1 + (height/2));
  float iz2 = aspect/z2;
  int px2 = QInt (x2 * iz2 + (width/2));
  int py2 = height - 1 - QInt (y2 * iz2 + (height/2));

  m_piG2D->DrawLine (px1, py1, px2, py2, color);
#endif
  return S_OK;
}

STDMETHODIMP csGraphics3DOpenGL::SetWorld(IWorld* piWorld)
{ 
	world = piWorld; 
#if 0	
	world->GetTextures(&Scan::piTC); 
	world->GetTextures(&Scan16::piTC);
#endif

	return S_OK; 
}

void csGraphics3DOpenGL::SysPrintf(int mode, char* szMsg, ...)
{
  char buf[1024];
  va_list arg;

  va_start (arg, szMsg);
  vsprintf (buf, szMsg, arg);
  va_end (arg);

  m_piSystem->Print(mode, buf);
}


//---------------------------------------------------------------------------

