/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef GRAPH3D_OPENGL_H
#define GRAPH3D_OPENGL_H

// GRAPH3D.H
// csGraphics3DOpenGL software rasterizer class.

// Concieved by Jorrit Tyberghein and Gary Clark
// Expanded by Dan Ogles

#include "com/com.h"
#include "render/igraph3d.h"

interface IPolygon3D;
interface IGraphics2D;
interface IMipMapContainer;
interface IWorld;
class OpenGLTextureCache;
class OpenGLLightmapCache;

///
class csGraphics3DOpenGL : public IGraphics3D
{
private:
//  long unsigned int *z_buffer;

protected:
  /// Z Buffer mode to use while rendering next polygon.
  ZBufMode z_buf_mode;

  /// Width of display.
  int width;
  /// Height of display.
  int height;
  /// Opt: width divided by 2.
  int width2;
  /// Opt: height divided by 2.
  int height2;

  /**
   * DAN: render-states
   * these override any other variable settings.
   */
  bool rstate_dither;
  bool rstate_specular;
  bool rstate_bilinearmap;
  bool rstate_trilinearmap;
  bool rstate_gouraud;
  bool rstate_flat;
  bool rstate_alphablend;
  int rstate_mipmap;
  bool rstate_edges;

  /// DrawFlags on last BeginDraw ()
  int DrawMode;

public:
#if defined (DO_MMX)
  /// True if CPU has MMX instructions
  bool cpu_mmx;
  /// True if 3D rendering should use MMX if available
  static bool do_mmx;
#endif

  /**
   * Low-level 2D graphics layer.
   * csGraphics3DOpenGL is in charge of creating and managing this.
   */
  IGraphics2D* m_piG2D;

  /// The texture cache.
  OpenGLTextureCache* texture_cache;
  OpenGLLightmapCache* lightmap_cache;

  /// The System interface. 
  ISystem* m_piSystem;

  /// The current camera
  ICamera* m_piCamera;
  
  /// A pointer to current selected world (for textures)
  static IWorld *world;

  /// Option variable: do texture lighting?
  static bool do_lighting;
  /// Option variable: render transparent textures?
  static bool do_transp;
  /// Option variable: render textures?
  static bool do_textured;
  /// Option variable: do expensive texel filtering?
  static bool do_texel_filt;
  /// Option variable: do perfect texture mapping?
  static bool do_perfect;

  /**
   * Option variable: draw white borders around the area of all polygons hit
   * by some light. The light is selected with selected_light.
   */
  static bool do_light_frust;

  /// @@todo: remove this
  static bool do_debug;
  /// For interlacing. 0 = draw odd lines, 1 = draw even lines, -1 = draw all lines.
  static int cfg_interlacing;
  /**
   * For interlacing. Temporary set to true if we moved quickly. This will decrease
   * the bluriness a little.
   */
  static bool ilace_fastmove;

  ///
  csGraphics3DOpenGL (ISystem* piSystem);
  ///
  virtual ~csGraphics3DOpenGL ();

  ///
  STDMETHODIMP Open(char *Title);
  ///
  STDMETHODIMP Close();

  /// Change the dimensions of the display.
  STDMETHODIMP SetDimensions (int width, int height);

  /// Start a new frame (see CSDRAW_XXX bit flags)
  STDMETHODIMP BeginDraw (int DrawFlags);

  /// End the frame and do a page swap.
  STDMETHODIMP FinishDraw ();

  /// Print the image in backbuffer
  STDMETHODIMP Print (csRect *area);

  /// Set the mode for the Z buffer used for drawing the next polygon.
  STDMETHODIMP SetZBufMode (ZBufMode mode);

  /// Draw the projected polygon with light and texture.
  STDMETHODIMP DrawPolygon (G3DPolygon& poly);

  /**
   * Draw the projected polygon with light and texture.
   * Debugging version. This one does not actually draw anything
   * but it just prints debug information about what it would have
   * done.
   */
  STDMETHODIMP DrawPolygonDebug (G3DPolygon& poly);

  /// Draw a line in camera space.
  STDMETHODIMP DrawLine (Vector3& v1, Vector3& v2, int color);

  /**
   * Draw a projected polygon.
   * This routine will draw the polygon using and filling
   * the z buffer if needed. The texture mapping need not
   * be perspective correct since this function is generally going
   * to be used for 3D sprites which are made up of small triangles.
   */
  STDMETHODIMP DrawPolygonQuick (G3DPolygon& poly, bool gouroud);

  /// Give a texture to csGraphics3DOpenGL to cache it.
  STDMETHODIMP CacheTexture (IPolygonTexture* texture);

  /**
   * Give a texture to csGraphics3DOpenGL to initialize the cache for it.
   * This is used together with the sub-texture optimization and is meant
   * to allocate the space in the cache but not do any actual calculations yet.
   */
  void CacheInitTexture (IPolygonTexture* texture);

  /// Give a sub-texture to csGraphics3DOpenGL to cache it.
  void CacheSubTexture (IPolygonTexture* texture, int u, int v);

  /**
   * Give a rectangle to csGraphics3DOpenGL so that all sub-textures
   * in this rectangle are cached.
   */
  void CacheRectTexture (IPolygonTexture* texture, int minu, int minv, int maxu, int maxv);

  /// Release a texture from the cache.
  STDMETHODIMP UncacheTexture (IPolygonTexture* texture);

  /// Set a renderstate boolean.
  STDMETHODIMP SetRenderState (G3D_RENDERSTATEOPTION op, long val);

  /// Get a renderstate value.
  STDMETHODIMP GetRenderState (G3D_RENDERSTATEOPTION op, long& val);

  /**
   * Get the current driver's capabilities. Each driver implements their
   * own function.
   */
  STDMETHODIMP GetCaps (G3D_CAPS *caps);

  /// Get address of Z-buffer at specific point
  STDMETHODIMP GetZBufPoint(int x, int y, unsigned long** retval) { /* *retval = z_buffer + x + y*width; */ return S_OK; }

  /// Dump the texture cache.
  STDMETHODIMP DumpCache (void);

  /// Clear the texture cache.
  STDMETHODIMP ClearCache (void);

  /// Get drawing buffer width
  STDMETHODIMP GetWidth(int& retval) { retval = width; return S_OK; }
  /// Get drawing buffer height
  STDMETHODIMP GetHeight(int& retval) { retval = height; return S_OK; }

  /// Set curren World (for sprite textures)
  STDMETHODIMP SetWorld (IWorld *piWorld);

  /// Get the IGraphics2D driver.
  STDMETHODIMP Get2dDriver(IGraphics2D** pi) { *pi = m_piG2D; return S_OK; }

  /// Returns S_OK if this driver requires all maps to be PO2.
  STDMETHODIMP NeedsPO2Maps() { return S_OK; }
  /// Returns the maximum aspect ratio of maps.
  STDMETHODIMP GetMaximumAspectRatio(int& retval) { retval = 32768; return S_OK; }

    /**
     *   * Initiate a volumetric fog object. This function will be called
     *      * before front-facing and back-facing fog polygons are added to
     *         * the object. The fog object will be convex but not necesarily
     *         closed.
     *            * The given CS_ID can be used to identify multiple fog
     *            objects when
     *               * multiple objects are started.
     *                  */
  STDMETHODIMP OpenFogObject (CS_ID id, csFog* fog);
    
      /**
       *   * Add a front or back-facing fog polygon in the current fog object.
       *      * Note that it is guaranteed that all back-facing fog polygons
       *         * will have been added before the first front-facing polygon.
       *            * fogtype can be:
       *               *    CS_FOG_FRONT:   a front-facing polygon
       *                  *    CS_FOG_BACK:    a back-facing polygon
       *                     *    CS_FOG_VIEW:    the view-plane
       *                        */
  STDMETHODIMP AddFogPolygon (CS_ID id, G3DPolygon& poly, int fogtype);
        
        /**
	 *   * Close a volumetric fog object. After the volumetric object is
	 *      * closed it should be rendered on screen (whether you do it
	 *      here
	 *         * or in DrawFrontFog/DrawBackFog is not important).
	 *            */
  STDMETHODIMP CloseFogObject (CS_ID id);
	  
	
  /// Set the active camera
  STDMETHODIMP SetCamera(ICamera* newval) { m_piCamera = newval; return S_OK; }

  /// Get the colorformat you want.
  STDMETHODIMP GetColormapFormat(G3D_COLORMAPFORMAT& retval) { retval = G3DCOLORFORMAT_ANY; return S_OK; }

  DECLARE_IUNKNOWN()
  DECLARE_INTERFACE_TABLE( csGraphics3DOpenGL )

  void SysPrintf (int mode, char* str, ...);
};

class csGraphics3DOpenGLFactory : public IGraphicsContextFactory
{
    /// Create the graphics context
    STDMETHODIMP CreateInstance( REFIID riid, ISystem* piSystem, void** ppv );

    /// Lock or unlock from memory.
    STDMETHODIMP LockServer(BOOL bLock);

    DECLARE_IUNKNOWN()
    DECLARE_INTERFACE_TABLE( csGraphics3DOpenGLFactory )
};

#endif

