/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Nathaniel Saint Martin <noote@bigfoot.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>

#include "sound/wave.h"
#include "sound/channel.h"
#include "sound/module.h"
#include "sound/cssound.h"

#define CLIP8(x) ((x<0)?0:(x>255)?255:x)
#define CLIP16(x) ((x<-0x8000)?-0x8000:(x>0x7FFF)?0x7FFF:x)

csSound::csSound(int nchannels)
{
  Memory = NULL;
  MemorySize = 0;
  Active = false;
  Frequency = 0;
  Bits16 = false;
  Stereo = false;
  BackGround = true;
  AllChannels = NULL;

  numberChannels = nchannels;
  CHK (Channels = new Channel* [numberChannels]);
  for(int i=0; i<numberChannels; i++)
    Channels[i]=NULL;
}

csSound::~csSound()
{
  Close();
}

bool csSound::Open(int frequency, bool bit16, bool stereo)
{
  Frequency = frequency;
  Bits16 = bit16;
  Stereo = stereo;

  MemorySize = 0;
  Memory = NULL;
  Active = true;
  
  return true;
}

bool csSound::Close()
{
  Active = false;
  if(Memory)
  {
    CHKB (delete [] Memory);
    Memory = NULL;
  }
  if(Channels)
  {
    CHKB (delete [] Channels);
    Channels = NULL;
  }

  return true;
}

int csSound::getFrequency()
{
  return Frequency;
}

bool csSound::isActive()
{
  return Active;
}

bool csSound::is16Bits()
{
  return Bits16;
}

bool csSound::isStereo()
{
  return Stereo;
}

bool csSound::nextMemory()
{
  return true;
}

void *csSound::getMemory()
{
  return (void *)Memory;
}

long csSound::getMemorySize()
{
  return MemorySize;
}

bool csSound::getBackGround()
{
  return BackGround;
}

void csSound::setVolume(float v)
{
  volume=v;
}

float csSound::getVolume()
{
  return volume;
}

Channel_ID csSound::add(Module *m, bool inLoop, int priority)
{
  if(m==NULL) return 0;

  Channel *c=m->makeChannel(Frequency, Bits16, Stereo, MemorySize);
  if(c)
  {
    c->setPriority(priority);
    c->setUserData(m);

    addChannel(c);

    return c->getID();
  }

  return 0;
}

Channel_ID csSound::add(Wave *w, bool inLoop, int priority)
{
  if(w==NULL) return 0;

  CHK (Channel *c=new Channel());

  c->setWave(w, inLoop);
  c->setPriority(priority);

  addChannel(c);

  return c->getID();
}

Channel_ID csSound::playEphemeral(Wave *w, int priority)
{
  if(w==NULL) return 0;

  CHK (Channel *c=new Channel());

  c->setWave(w, false);
  c->setEphemeral();
  c->setPriority(priority);
  c->setStarted();

  addChannel(c);

  return c->getID();
}

bool csSound::delChannel(Channel_ID id)
{
  if(AllChannels)
  {
    Channel *c=AllChannels;
    Channel *old=NULL;

    if(c && c->getID()==id)
    {
      if(c->next) AllChannels=c->next; 
      else AllChannels=NULL;

      CHK (delete c);
      return true;
    }
    else
    {
      while(c)
      {
        old=c;
        c=c->next;
        if(c && c->getID()==id)
        {
          if(c->next) old->next=c->next;
          else old->next=NULL;

          CHK (delete c);
          return true;
        }
      }
    }
  }

  return false;
}

void csSound::addChannel(Channel *c)
{
  if(c==NULL) return;

  c->next=NULL;
  if(AllChannels)
  {
    Channel *old=AllChannels;

    while(old->next) old=old->next;
    
    old->next=c;
  }
  else
  {
    AllChannels=c;
  }
}

void csSound::startChannel(Channel_ID id)
{
  Channel *c = getChannel(id);
  if(c)
    c->setStarted();
}

Channel *csSound::getChannel(Channel_ID cid)
{
  Channel *ret=NULL;

  if(AllChannels)
  {
    Channel *c=AllChannels;

    while(c)
    {
      if(c->getID()==cid)
      {
        ret=c;
        break;
      }

      c=c->next;
    }
  }

  return ret;
}

void csSound::update()
{
  if(!getBackGround())
    mixingFunction();
}

void csSound::setPriorityFunc(PRIORITY_FUNC mpf)
{
  PriorityFunc=mpf;
}

int csSound::getNumberOfChannels()
{
  return numberChannels;
}

void csSound::setChannels()
{
  Channel *c=AllChannels;
  int n=0, i;
  
  for(i=0; i<numberChannels; i++)
    Channels[i]=NULL;
  
  while(c!=NULL)
  {
    c->Used=false;
    if(c->isActive() && c->isStarted())
    {
      if(n>=numberChannels)
      {
        if(PriorityFunc == Priority_Up)
        {
          for(i=0; i<numberChannels; i++)
            if(Channels[i]->getPriority()<c->getPriority())
            {
              Channels[i]=c;
              c->Used=true;
              break;
            }
        }
        else if(PriorityFunc == Priority_Down)
        {
          for(i=0; i<numberChannels; i++)
            if(Channels[i]->getPriority()>c->getPriority())
            {
              Channels[i]=c;
              c->Used=true;
              break;
            }
        }
        else //if(PriorityFunc == Priority_None)
          break;
      }
      else
      {
        Channels[n]=c;
        c->Used=true;
        n++;
      }
    }
    c=c->next;
  }
}

void csSound::updateChannels()
{
  Channel *c=AllChannels;
  
  while(c!=NULL)
  {
    if(c->isActive() && !c->Used && !c->isFromModule())
      c->addStep((is16Bits())?getMemorySize()/2:getMemorySize());
    
    c=c->next;
  }
}

void csSound::killChannels()
{
  Channel *c=AllChannels;
  Channel *old;
  
  while(c!=NULL)
  {
    old=c;
    c=c->next;
    if(!old->isActive() && !old->isFromModule())
      delChannel(old->getID());
  }
}

void csSound::mixingFunction()
{
  if(Active)
  {
    if(!nextMemory()) return;
    
    int numberUsedChannels=0;
    setChannels();
    
    // Check if channel is from module
    for(int m=0; m<numberChannels; m++)
    {
      if(Channels[m]
        && Channels[m]->isFromModule()
        && Channels[m]->isStarted())
      {
        // update module output
        mixingModuleFunction(Channels[m]);
      }
    }
    
    if(!is16Bits())
    {
      unsigned char *ptr=(unsigned char *)getMemory();
      long ptr_size=getMemorySize();
      
      if(!isStereo())
      {
        // 8 bit mono
        for(int p=0; p<ptr_size; p++)
        {
          float sample = 0;
          numberUsedChannels=0;
          
          for(int i=0; i<numberChannels; i++)
          {
            if(Channels[i] && Channels[i]->isActive())
            {
              int smp=Channels[i]->getSample();
              if(Channels[i]->isStereo())
                smp=(smp+Channels[i]->getSample())/2;
              
              sample+=smp*Channels[i]->Volume;
              numberUsedChannels++;
            }
          }
          
          if(numberUsedChannels>0)
            *ptr++=sample/numberUsedChannels;
          //*ptr++=sample;
          else
            *ptr++=128;
        }
      }
      else // if(isStereo())
      {
        // 8 bit stereo
        for(int p=0; p<ptr_size; p+=2)
        {
          float left = 0, right = 0;
          
          numberUsedChannels=0;
          for(int i=0; i<numberChannels; i++)
          {
            if(Channels[i] && Channels[i]->isActive())
            {
              int samplel, sampler;
              samplel=sampler=Channels[i]->getSample();
              if(Channels[i]->isStereo())
                sampler=Channels[i]->getSample();
              
              left+=samplel*Channels[i]->Volume_left;
              right+=sampler*Channels[i]->Volume_right;
              numberUsedChannels++;
            }
          }
          
          if(numberUsedChannels>0)
          {
            *ptr++=left/numberUsedChannels;
            *ptr++=right/numberUsedChannels;
            //*ptr++=left;
            //*ptr++=right;
          }
          else
          {
            *ptr++=128;
            *ptr++=128;
          }
        }
      }
    }
    else // if(is16Bits())
    {
      short *ptr=(short *)getMemory();
      long ptr_size=getMemorySize()/2;
      
      if(!isStereo())
      {
        // 16 bit mono
        for(int p=0; p<ptr_size; p++)
        {
          float sample = 0;
          
          numberUsedChannels=0;
          for(int i=0; i<numberChannels; i++)
          {
            if(Channels[i] && Channels[i]->isActive())
            {
              int smp=Channels[i]->getSample();
              if(Channels[i]->isStereo())
                smp=(smp+Channels[i]->getSample())/2;
              
              sample+=smp*Channels[i]->Volume;
              numberUsedChannels++;
            }
          }
          
          if(numberUsedChannels>0)
            *ptr++=sample/numberUsedChannels;
          //*ptr++=sample;
          else
            *ptr++=0;
        }
      }
      else // if(isStereo())
      {
        // 16 bit stereo
        for(int p=0; p<ptr_size; p+=2)
        {
          float left = 0, right = 0;
          
          numberUsedChannels=0;
          for(int i=0; i<numberChannels; i++)
          {
            if(Channels[i] && Channels[i]->isActive())
            {
              int samplel, sampler;
              samplel=sampler=Channels[i]->getSample();
              if(Channels[i]->isStereo())
                sampler=Channels[i]->getSample();
              
              left+=samplel*Channels[i]->Volume_left;
              right+=sampler*Channels[i]->Volume_right;
              numberUsedChannels++;
            }
          }
          
          if(numberUsedChannels>0)
          {
            *ptr++=left/numberUsedChannels;
            *ptr++=right/numberUsedChannels;
            //*ptr++=left;
            //*ptr++=right;
          }
          else
          {
            *ptr++=0;
            *ptr++=0;
          }
        }
      }
    }
    updateChannels();
    killChannels();
  }
  else
  {
    if(!is16Bits())
    {
      unsigned char *ptr=(unsigned char *)getMemory();
      long ptr_size=getMemorySize();
      for(int p=0; p<ptr_size; p++)
        *ptr++=128;
    }
    else // if(is16Bits())
    {
      short *ptr=(short *)getMemory();
      long ptr_size=getMemorySize()/2;
      for(int p=0; p<ptr_size; p++)
        *ptr++=0;
    }
  }
}

void csSound::mixingModuleFunction(Channel * c)
{
  // hum, there is a big bug somewhere in mod->mixingfunction(), I don't find it yet,
  // Maybe I need to redo the mod player...
  Module *mod=(Module *)c->getUserData();
  if(mod) // calculate next buffer for this mod
    mod->mixingFunction();
}
