/*
    DOS support for Crystal Space 3D library
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by David N. Arnold <derek_arnold@fuse.net>
    Written by Andrew Zabolotny <bit@eltech.ru>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdarg.h>
#include "raw.h"
#include "system/djgpp/idjgpp.h"
#include "geometry/csrect.h"

#ifdef USE_ALLEGRO
#  include "allegro.h"
#else
#  include "djvidsys.h"
#endif

static VideoSystem VS;

#ifdef USE_ALLEGRO
BITMAP *_vdata;
#endif

//------------------------------------------------ csGraphics2DFactoryDOSRAW ---

IMPLEMENT_UNKNOWN_NODELETE (csGraphics2DFactoryDOSRAW)

BEGIN_INTERFACE_TABLE (csGraphics2DFactoryDOSRAW)
  IMPLEMENTS_INTERFACE (IGraphics2DFactory)
END_INTERFACE_TABLE ()

STDMETHODIMP csGraphics2DFactoryDOSRAW::CreateInstance (REFIID riid,
  ISystem* piSystem, void**ppv)
{
  if (!piSystem)
  {
    *ppv = 0;
    return E_INVALIDARG;
  }

  csGraphics2DDOSRAW *pNew = new csGraphics2DDOSRAW (piSystem);
  if (!pNew)
  {
    *ppv = 0;
    return E_OUTOFMEMORY;
  }

  return pNew->QueryInterface (riid, ppv);
}

STDMETHODIMP csGraphics2DFactoryDOSRAW::LockServer (BOOL bLock)
{
  (void)bLock;
  return S_OK;
}

//------------------------------------------------------- csGraphics2DDOSRAW ---

void __printf (bool error, char *format, ...)
{
  va_list arg;
  char buf[256];

  va_start (arg, format);
  vsprintf (buf, format, arg);
  va_end (arg);

  csGraphics2DDOSRAW::System->Print (error ? MSG_FATAL_ERROR : MSG_INITIALIZATION, buf);
}

BEGIN_INTERFACE_TABLE (csGraphics2DDOSRAW)
  IMPLEMENTS_COMPOSITE_INTERFACE_EX (IGraphics2D, XGraphics2D)
  IMPLEMENTS_COMPOSITE_INTERFACE_EX (IGraphicsInfo, XGraphicsInfo)
END_INTERFACE_TABLE ()

IMPLEMENT_UNKNOWN_NODELETE (csGraphics2DDOSRAW)

ISystem* csGraphics2DDOSRAW::System = NULL;
ISystemDriverDOS* csGraphics2DDOSRAW::DosSystem = NULL;

csGraphics2DDOSRAW::csGraphics2DDOSRAW (ISystem* piSystem) : csGraphics2D (piSystem)
{
  Memory = NULL;
  System = piSystem;

  if (FAILED (System->QueryInterface (IID_ISystemDriverDOS, (void**)&DosSystem)))
  {
    __printf (true, "The system driver does not support "
                    "the ISystemDriverDOS interface\n");
    exit (-1);
  }

  switch (Depth)
  {
    case 8:
      pfmt.RedMask = pfmt.GreenMask = pfmt.BlueMask = 0;
      pfmt.PalEntries = 256; pfmt.PixelBytes = 1;
      break;

    case 15:
      pfmt.RedMask   = 0x1f << 10;
      pfmt.GreenMask = 0x1f << 5;
      pfmt.BlueMask  = 0x1f;
      pfmt.PalEntries = 0;
      pfmt.PixelBytes = 2;

      DrawPixel = DrawPixel16;
      WriteChar = WriteChar16;
      GetPixelAt = GetPixelAt16;
      DrawSprite = DrawSprite16;
      break;

    case 16:
      pfmt.RedMask   = 0x1f << 11;
      pfmt.GreenMask = 0x3f << 5;
      pfmt.BlueMask  = 0x1f;
      pfmt.PalEntries = 0;
      pfmt.PixelBytes = 2;

      DrawPixel = DrawPixel16;
      WriteChar = WriteChar16;
      GetPixelAt = GetPixelAt16;
      DrawSprite = DrawSprite16;
      break;
  }

  complete_pixel_format();
}

csGraphics2DDOSRAW::~csGraphics2DDOSRAW ()
{
  Close();
  FINAL_RELEASE (DosSystem);
}

bool csGraphics2DDOSRAW::Open (char* Title)
{
  if (!csGraphics2D::Open (Title))
    return false;
  PaletteChanged = false;

#if USE_ALLEGRO
  allegro_init ();

  int rc;
  set_color_depth (Depth);
  if (Width == 320 && Height == 200 && Depth == 8)
    rc = set_gfx_mode (GFX_MODEX, 320, 200, 0, 600); // triple buffering
  else
  {
    rc = set_gfx_mode (GFX_AUTODETECT, Width, Height, 0, 0);
    if (rc)
    {
      rc = set_gfx_mode (GFX_VESA1, Width, Height, 0, 0);
      if (rc)
      {
        rc = set_gfx_mode (GFX_VESA2B, Width, Height, 0, 0);
        if (!rc)
          __printf (false, "VESA2 Banked mode selected.\n");
      }
      else
        __printf (false, "VESA1 mode selected.\n");
    }
  }

  if (rc)
  {
    __printf (true, "ERROR! Could not set graphics mode.\n");
    exit (-1);
  }

  _vdata = create_bitmap (Width, Height);
  Memory = (unsigned char *)_vdata->dat;

  if (Memory == NULL)
  {
    __printf (true, "Error initializing graphics subsystem: bad videomode!\n");
    exit (-1);
  }

#else // USE_ALLEGRO
  VS.WaitVRetrace (false);

  if (!VS.FindMode (Width, Height, Depth))
  {
    __printf (true, "Cannot find a suitable videomode match\n");
    exit (-1);
  }
  VS.Open ();

#endif // USE_ALLEGRO

  // Tell printf() to shut up
  DosSystem->EnablePrintf (false);

  // Clear all videopages
  ClearAll (0);
  return 1;
}

void csGraphics2DDOSRAW::Close ()
{
#if USE_ALLEGRO
  destroy_bitmap ((BITMAP *)_vdata);
#else
  VS.Close ();
#endif // USE_ALLEGRO
  csGraphics2D::Close ();
  // Tell printf() it can work now
  DosSystem->EnablePrintf (true);
}

void csGraphics2DDOSRAW::Print (csRect *area = NULL)
{
  (void)area;
#if USE_ALLEGRO
  if (area)
    blit (_vdata, screen, area->xmin, area->ymin, area->xmin, area->ymin,
      area->Width (), area->Height ());
  else
    blit (_vdata, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

   ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
            please check above "if" for correctness
                    then remove this text
#else
  if (area)
    VS.Flush (area->xmin, area->ymin, area->Width (), area->Height ());
  else
    VS.Flush (0, 0, Width, Height);
#endif // USE_ALLEGRO
}

bool csGraphics2DDOSRAW::SetMouseCursor (int iShape, TextureMM *iBitmap)
{
  (void)iShape; (void)iBitmap;
  return false;
}

void csGraphics2DDOSRAW::SetRGB (int i, int r, int g, int b)
{
  csGraphics2D::SetRGB (i, r, g, b);
  PaletteChanged = true;
}

bool csGraphics2DDOSRAW::BeginDraw ()
{
  if (PaletteChanged && pfmt.PalEntries)
  {
    outportb (0x3c8, 0);
    for (int i = 0; i <= 255; i++)
    {
      outportb (0x3c9, Palette[i].red >> 2);
      outportb (0x3c9, Palette[i].green >> 2);
      outportb (0x3c9, Palette[i].blue >> 2);
    } /* endfor */
    PaletteChanged = false;
  }
#if !USE_ALLEGRO
  Memory = VS.BackBuffer ();
  return (Memory != NULL);
#else
  return csGraphics2D::BeginDraw ();
#endif
}

#if !USE_ALLEGRO

void csGraphics2DDOSRAW::Clear (int color)
{
  VS.Clear (color);
}

void csGraphics2DDOSRAW::FinishDraw ()
{
  Memory = NULL;
}

int csGraphics2DDOSRAW::GetPage ()
{
  return VS.GetPage ();
}

bool csGraphics2DDOSRAW::DoubleBuffer ()
{
  return VS.DoubleBuffer ();
}

bool csGraphics2DDOSRAW::DoubleBuffer (bool Enable)
{
  return VS.DoubleBuffer (Enable);
}

#endif // USE_ALLEGRO
