/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define INITGUID

#include "com/com.h"
#include "system/isystem.h"
#include "system/libggi/driver2d/ggig2d.h"
#include "system/libggi/ggiinitf.h"
#include "geometry/csrect.h"

#include <ggi/ggi.h>

BEGIN_INTERFACE_TABLE(csGraphics2DGGI)
  IMPLEMENTS_COMPOSITE_INTERFACE_EX( IGraphics2D, XGraphics2D )
  IMPLEMENTS_COMPOSITE_INTERFACE_EX( IGraphicsInfo, XGraphicsInfo )
  IMPLEMENTS_COMPOSITE_INTERFACE_EX( IGGIGraphicsInfo, XGGIGraphicsInfo )
END_INTERFACE_TABLE()

IMPLEMENT_UNKNOWN_NODELETE(csGraphics2DGGI)


// csGraphics2DGGI functions

csGraphics2DGGI::csGraphics2DGGI(ISystem* piSystem, bool bUses3D) : csGraphics2D (piSystem),
	vis (NULL)
{
  HRESULT rc;
  IGGISystemDriver* piGGI = NULL;

  rc = piSystem->QueryInterface (IID_IGGISystemDriver, (void**)&piGGI);
  if (FAILED (rc))
  {
    //CsPrintf (MSG_FATAL_ERROR, "csGraphics2DGGI::Open() -- ISystem does not support IGGISystemDriver.");
    printf ("csGraphics2DGGI::Open() -- ISystem does not support IGGISystemDriver.");
    exit (1);
  }

  if (ggiInit() != 0) {
    //CsPrintf (MSG_FATAL_ERROR, "Couldn't initialize LibGGI\n");
    exit (1);
  }

  vis = ggiOpen(NULL);

  if (vis == NULL) {
    //CsPrintf (MSG_FATAL_ERROR, "Couldn't open GGI visual\n");
    exit (2);
  }
 
  ggiParseMode("", &vis_mode);

  vis_mode.visible.x = Width;
  vis_mode.visible.y = Height;
  vis_mode.graphtype = Depth;

  ggiCheckMode(vis, &vis_mode);

  if (ggiSetMode(vis, &vis_mode) < 0) {
    //CsPrintf (MSG_FATAL_ERROR, "Couldn't set mode\n");
    exit (3);
  }

  display_width  = vis_mode.virt.x;
  display_height = vis_mode.virt.y;

  // calculate pixel format
  
  const ggi_pixelformat *ggi_pf = ggiGetPixelFormat(vis);

  if (ggi_pf == NULL) {
    //CsPrintf (MSG_FATAL_ERROR, "Couldn't get pixel format\n");
    exit (4);
  }
  
  pfmt.RedMask    = 0;
  pfmt.GreenMask  = 0;
  pfmt.BlueMask   = 0;
  pfmt.PalEntries = 0;
  pfmt.PixelBytes = (GT_SIZE(vis_mode.graphtype)+7) / 8;

  switch (GT_SCHEME(vis_mode.graphtype)) {

    case GT_PALETTE:
      pfmt.PalEntries = 1 << GT_DEPTH(vis_mode.graphtype);
      break;

    case GT_TRUECOLOR:
      pfmt.RedMask    = ggi_pf->red_mask;
      pfmt.GreenMask  = ggi_pf->green_mask;
      pfmt.BlueMask   = ggi_pf->blue_mask;
      break;

    default:
      //CsPrintf (MSG_FATAL_ERROR, "Unsupported pixel format\n");
      exit (5);
  }

  complete_pixel_format ();

  //CsPrintf (MSG_INITIALIZATION, "Crystal Space GGI version");
  printf ("Crystal Space GGI version");
  fflush(stdout);

  Memory = NULL;

  // If in 16-bit mode, redirect drawing routines
  if (pfmt.PixelBytes == 2)
  {
    DrawPixel  = DrawPixel16;
    WriteChar  = WriteChar16;
    GetPixelAt = GetPixelAt16;
    DrawSprite = DrawSprite16;
  }
}

csGraphics2DGGI::~csGraphics2DGGI(void)
{
  // Destroy your graphic interface
  Close();

  if (Memory) {
  	delete [] Memory;
  	Memory = NULL;
  }
  
  printf("GGI2d driver terminating...\n");
  fflush(stdout);

  if (vis) {
    ggiClose(vis);
    vis = NULL;
  }

  ggiExit();
}

bool csGraphics2DGGI::Open(char *Title)
{
  // Open your graphic interface
  if (! csGraphics2D::Open (Title)) 
    return false;

  Memory = CHK( new unsigned char [Width*Height*pfmt.PixelBytes] );

  Clear (0);
  return true;
}

void csGraphics2DGGI::Close(void)
{
  // Close your graphic interface
  csGraphics2D::Close ();
}

void csGraphics2DGGI::Print (csRect *area)
{
  (void) area;

  ggiPutBox(vis, 0, 0, Width, Height, Memory);
  ggiFlush(vis);
}

void csGraphics2DGGI::SetRGB(int i, int r, int g, int b)
{
  ggi_color col;

  col.r = r << 8;
  col.g = g << 8;
  col.b = b << 8;

  ggiSetPalette(vis, i, 1, &col);
  ggiFlush(vis);

  csGraphics2D::SetRGB (i, r, g, b);
}

