//
//  GGI support for Crystal Space 3D library
//
//  Copyright (C) 1998 by Jorrit Tyberghein
//  Copyright (C) 1998 by Andrew Apted <andrew@ggi-project.org>
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Library General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Library General Public License for more details.
//
//  You should have received a copy of the GNU Library General Public
//  License along with this library; if not, write to the Free
//  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

//------------------------------------------------
//
//  For more information about GGI, see:
//
//      http://www.ggi-project.org/
//

#ifdef DISP_GGI

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <errno.h>

#include <ggi/ggi.h>

#include "util/def.h"
#include "system/system.h"
#include "util/inifile.h"
#include "system/libggi/libggi.h"
#include "system/libggi/driver2d/isysg2d.h"
// #include "render/graph3d.h"
#include "general/console.h"


static struct ggi_screen_info
{
  int mouse_x, mouse_y;

} ggi_screen = { 0, };


// The System driver ////////////////

BEGIN_INTERFACE_TABLE(SysSystemDriver)
  IMPLEMENTS_COMPOSITE_INTERFACE(GGISystemDriver)
  IMPLEMENTS_COMPOSITE_INTERFACE(System)
END_INTERFACE_TABLE()

IMPLEMENT_UNKNOWN_NODELETE(SysSystemDriver)


//-- static SysSystemDriver* sys; // Global pointer to system that can be used by event handler


//----------------------------------------------------------------------
//
//  Configuration
//

void SysSystemDriver::SetSystemDefaults ()
{
  csSystemDriver::SetSystemDefaults ();

  //FULL_SCREEN = true;
}

void SysSystemDriver::ParseArg (int argc, char* argv[], int &i)
{
  if (strcasecmp ("-foobar", argv[i]) == 0)
  {
      CsPrintf (MSG_FATAL_ERROR, "FOOBAR support not compiled into this version of Crystal Space!\n");
      //!! exit_clean (0);
  }
  else csSystemDriver::ParseArg (argc, argv, i);
}

void SysSystemDriver::Help ()
{
  csSystemDriver::Help ();
  CsPrintf (MSG_STDOUT, "  -foobar            enable foobar support (default=disabled)\n");
}


// COM Implementation

IMPLEMENT_COMPOSITE_UNKNOWN_AS_EMBEDDED( SysSystemDriver, GGISystemDriver )


//----------------------------------------------------------------------
//
//  Input Handling
//

SysKeyboardDriver::SysKeyboardDriver () : csKeyboardDriver ()
{
  // Create your keyboard interface
}

SysKeyboardDriver::~SysKeyboardDriver ()
{
  // Destroy your keyboard interface
}

bool SysKeyboardDriver::Open (csEventQueue* EvQueue)
{
  // Open your keyboard interface
  csKeyboardDriver::Open (EvQueue);
  return true;
}

void SysKeyboardDriver::Close ()
{
  // Close your keyboard interface
  csKeyboardDriver::Close ();
}

SysMouseDriver::SysMouseDriver () : csMouseDriver ()
{
  // Create your mouse interface
}

SysMouseDriver::~SysMouseDriver ()
{
  // Destroy your mouse interface
}

bool SysMouseDriver::Open(csEventQueue* EvQueue)
{
  // Open mouse system
  
  ggi_screen.mouse_x = 0;
  ggi_screen.mouse_y = 0;

  csMouseDriver::Open (EvQueue);
  return true;
}

void SysMouseDriver::Close()
{
  // Close mouse system
  csMouseDriver::Close();
}


#define GGI_EVENTS  (ggi_event_mask) (emKey | emPointer)

// System loop !
void SysSystemDriver::Loop ()
{
  ggi_visual_t vis = NULL;
  ggi_mode vis_mode;

  IGGIGraphicsInfo* piG2D = NULL;
  HRESULT hRes = Gfx2D->QueryInterface (IID_IGGIGraphicsInfo, (void**)&piG2D);

  if (SUCCEEDED(hRes))
  {
    hRes = piG2D->GetVisual (&vis);
    piG2D->Release ();
    piG2D = NULL;
  } else {
    fprintf(stderr, "No such interface: IGGIGraphicsInfo\n");
    exit(1);
  }

  ggiGetMode(vis, &vis_mode);

  while (!Shutdown && !ExitLoop)
  {
    static long prev_time = -1;
    long elapsed = (prev_time == -1) ? 0 : Time () - prev_time;
    prev_time = Time ();

    NextFrame (elapsed, Time());

    // handle input events

    for (;;)
    {
      ggi_event ev;
      struct timeval tv;

      tv.tv_sec  = 0;
      tv.tv_usec = 0;

      if (! ggiEventPoll(vis, GGI_EVENTS, &tv))
      { 
        // no events are available
        break;
      }

      ggiEventRead(vis, &ev, GGI_EVENTS);

      /* handle GGI key events */

      if ((ev.any.type == evKeyPress) ||
          (ev.any.type == evKeyRelease))
      {
        int state = (ev.any.type != evKeyRelease);
        int key=-1;
      
        // Normal latin-1 key ?
        if (GII_KTYP(ev.key.sym) == GII_KT_LATIN1)
          key = GII_KVAL(ev.key.sym);
        else
        {
          // Special key.
          //
          // The "label" field contains the key without being affected by
          // any modifiers (shift, alt or whatever).
          
          switch(ev.key.label)
          {
            case GIIK_Enter:   key = CSKEY_ENTER; break;
            case GIIK_Insert:  key = CSKEY_INS;   break;
            case GIIUC_Delete: key = CSKEY_DEL;   break;
            
            case GIIK_Up:    case GIIK_P5: key = CSKEY_UP;    break;
            case GIIK_Down:  case GIIK_P2: key = CSKEY_DOWN;  break;
            case GIIK_Right: case GIIK_P6: key = CSKEY_RIGHT; break;
            case GIIK_Left:  case GIIK_P4: key = CSKEY_LEFT;  break;
            case GIIK_PageUp:   case GIIK_P9: key = CSKEY_PGUP;  break;
            case GIIK_PageDown: case GIIK_P3: key = CSKEY_PGDN;  break;
      
            // Modifiers.  
            //
            // Currently we just set the Key.* variables and swallow them
            // (instead of passing them on as CSKEY_SHIFT/ALT/CTRL events).
      
            case GIIK_Shift:
              Keyboard->Key.shift = state; break;
      
            case GIIK_Ctrl:
              Keyboard->Key.ctrl = state; break;
      
            case GIIK_Alt:
              Keyboard->Key.alt = state; break;
          }
        }
      
        if (state && key >= 0)
          Keyboard->do_keypress(Time(), key);
        else
          Keyboard->do_keyrelease(Time(), key);
      }
      
      /* handle GGI pointer events */

      {
        if (ev.any.type == evPtrRelative)
        {
          ggi_screen.mouse_x += ev.pmove.x;
          ggi_screen.mouse_y -= ev.pmove.y;
        }
        else if (ev.any.type == evPtrAbsolute)
        {
          ggi_screen.mouse_x = ev.pmove.x;
          ggi_screen.mouse_y = ev.pmove.y;
        }
       
        // Limit pointer to screen
        
        if (ggi_screen.mouse_x < 0) 
          ggi_screen.mouse_x = 0;
          
        if (ggi_screen.mouse_y < 0) 
          ggi_screen.mouse_y = 0;
      
        if (ggi_screen.mouse_x >= vis_mode.visible.x)  
          ggi_screen.mouse_x = vis_mode.visible.x - 1;
      
        if (ggi_screen.mouse_y >= vis_mode.visible.y)  
          ggi_screen.mouse_y = vis_mode.visible.y - 1;
      
        Mouse->do_mousemotion(Time(), 
          ggi_screen.mouse_x, ggi_screen.mouse_y);
      }

      /* handle GGI button events */
      {
        if (ev.any.type == evPtrButtonPress)
          Mouse->do_buttonpress(Time(), ev.pbutton.button, 
          	 ggi_screen.mouse_x, ggi_screen.mouse_y, 
		 Keyboard->Key.shift, Keyboard->Key.alt, 
		 Keyboard->Key.ctrl);
      
        if (ev.any.type == evPtrButtonRelease)
          Mouse->do_buttonrelease(Time(), ev.pbutton.button,
               ggi_screen.mouse_x, ggi_screen.mouse_y);
      }
    }
  }
}

#endif // DISP_GGI
