/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Nathaniel Saint Martin <noote@bigfoot.com>
    Linux sound module by Gert Steenssens <gsteenss@eps.agfa.be>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
 
#include <stdio.h>  
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <signal.h>


#include "util/def.h"
#include "util/types.h"
#include "system/system.h"   


#include "sound/wave.h"
#include "sound/channel.h"
#include "sound/module.h"
#include "sound/cssnddrv.h"

#include "system/lxsound/syssound.h"


SoundBlock::SoundBlock(size_t block_size) 
{
  // we need a static counter here to get a
  // *unique* id for our shared memory block
  static size_t block_count=0;

  ptr=NULL;
  sndentry.mtype=1;
  sndentry.shmsize = block_size;
  sndentry.shmid = shmget(ftok(".",block_count++),sndentry.shmsize,IPC_CREAT);
  //fprintf(stderr,"err: %d\n",errno);

  ptr = (unsigned char*)shmat(sndentry.shmid,NULL,0);
    
  //probably need to clear to memory
  memset(ptr,0,sndentry.shmsize);

  //fprintf(stderr,"soundblock: %d,%d\n",sndentry.shmid,sndentry.shmsize);
}

SoundBlock::~SoundBlock()
{    
  if (ptr) shmdt(ptr);
}   

unsigned char* SoundBlock::Memory() 
{
  return ptr;
}

size_t SoundBlock::MemorySize()
{
  return sndentry.shmsize;
}
  
soundmsg SoundBlock::Message()
{
  return sndentry;
}
  
 


AudioDevice::AudioDevice()
{
  audio = -1;
  //setup message queue : 
  qid = msgget(ftok(".",666),IPC_CREAT);

  //fprintf(stderr,"using queue id: %d\n",qid);
}

         
void AudioDevice::Play(SoundBlock& sound)
{

  size_t sndsize=sizeof(soundmsg)-sizeof(long);
  soundmsg sndmsg=sound.Message();
    
  if (msgsnd(qid,&sndmsg,sndsize,0)==-1)
    fprintf(stderr,"could not send sound message: %d\n",errno);

  //fprintf(stderr,"msgsnd: %d,%d\n",sndmsg.shmid,sndmsg.shmsize);
        
}

void AudioDevice::startAudioLoop()
{
  // fork child process to do audio loop
  pid=fork();

  if (pid==0)
  { 
      
    fprintf(stderr,"Audio loop running\n");
      
    qid = msgget(ftok(".",666),IPC_CREAT);

    //fprintf(stderr,"using queue id: %d\n",qid);

    size_t size,sndsize=sizeof(soundmsg);
    soundmsg msg;
    unsigned char* ptr; 

    // loop forever: they'll kill us when they want us dead...
    while (1)
    {

      size=msgrcv(qid,&msg,sndsize,0,MSG_NOERROR);

      //fprintf(stderr,"sound msg received: %d\n",size);
      //fprintf(stderr,"msgrcv: %d,%d\n",msg.shmid,msg.shmsize);
      
      if (size>0)
      { 
        ptr = (unsigned char*)shmat(msg.shmid,NULL,0);

        write(audio,ptr,msg.shmsize);
      
        shmdt(ptr);

        shmctl(msg.shmid,IPC_RMID,NULL);
       }
      }
    }

  else // parent process just returns 
    return;
} 

size_t AudioDevice::Open(int frequency, bool bit16, bool stereo)
{
  int tmp,samplesize,dsp_stereo,dsp_speed;   

  dsp_speed=frequency;
  dsp_stereo=stereo;
  if (bit16) 
    samplesize=16;
  else
    samplesize=8;
   
  audio = open("/dev/dsp",O_WRONLY|O_NONBLOCK,0);
  //audio = open("/dev/dsp",O_WRONLY,0);

  if (audio == -1)
    {
      fprintf(stderr, "Error opening audio\n");
      exit (-1);
    }

  tmp = samplesize;
  ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &samplesize);
  if (tmp != samplesize)
  {
  	fprintf(stderr, "Unable to set the sample size\n");
  	exit(-1);
  }

  if (ioctl (audio, SNDCTL_DSP_STEREO, &dsp_stereo)==-1)
  {
      fprintf (stderr, "Unable to set mono/stereo\n");
      exit (-1);
  }

  if (ioctl (audio, SNDCTL_DSP_SPEED, &dsp_speed) == -1)
    {
      fprintf (stderr, "Unable to set audio speed\n");
      exit (-1);
    }
  
  ioctl (audio, SNDCTL_DSP_GETBLKSIZE, &block_size);
  if (block_size < 1024 || block_size > (2*65536))
    {
      if (block_size == -1)
	fprintf (stderr, "Error getting audio buffers size\n");
      else
	fprintf (stderr, "Invalid audio buffers size %d\n", block_size);
      exit (-1);
    }



  // set mixing buffer size
  //block_size = 16384; 
  //block_size = 8192;
  //block_size = 65536;


  startAudioLoop();

  return block_size;

}


void AudioDevice::Close()
{ 
  // stop audio loop (rather undelicate...)
  kill(pid,SIGTERM);

  // close audio device
  close(audio);
  audio=-1;

  // remove message queue
  msgctl(qid,IPC_RMID,NULL);
}

static void* soundptr;


void isTime(int)
{
  SysSoundDriver *mysound=(SysSoundDriver*)soundptr;
  mysound->update();
}
 
void SysSoundDriver::SetupTimer()
{
  struct itimerval itime;
  struct timeval val;
  val.tv_usec=250; 
  val.tv_sec=0;

  itime.it_value=val;
  itime.it_interval=val;

  // set static ptr because the timer handler needs it...
  soundptr=(void*)this;

  // set timer
  if (setitimer(ITIMER_VIRTUAL,&itime,NULL)==-1)
  {
    fprintf(stderr, "error setting timer\n");
  }
  
  // set signal handling function
  signal(SIGVTALRM,isTime);

}




SysSoundDriver::SysSoundDriver(int nchannels) : csSoundDriver(nchannels)
{
  // if background is set to true, we will use a timer,
  // else we rely on the main loop to call us...
  //BackGround = false;  
  BackGround = true;

  Active = false;
  block_size=0;
  
}

bool SysSoundDriver::Open(int frequency, bool bit16, bool stereo)
{

  Frequency = frequency;
  Bits16 = bit16;
  Stereo = stereo;
  
  block_size = device.Open(frequency,bit16,stereo);

  CsPrintf (MSG_INITIALIZATION, "Sound initialized to %d Hz %d bits %s (%d channels)\n", Frequency, (Bits16)?16:8, (Stereo)?"Stereo":"Mono", numberChannels);


  // setup our timer routine...
  if (BackGround) SetupTimer();

  Active = true;
  return Active;
}

bool SysSoundDriver::Close()
{  
  device.Close();
  return csSoundDriver::Close();
}



bool SysSoundDriver::ChannelsInUse()
{
  Channel *c=AllChannels;
  while(c!=NULL)
  {
    if (c->isActive()) return true;
    c=c->next;
  }
  return false;
}


// this a simple version of update(): it just copies the complete buffer 
// returned by mixingFunction() in one time, instead of writing it out in blocks of size/X
// not sure which is better,faster,...

void SysSoundDriver::update()
{
        
          // first check if we need to do this or not...
          if (!ChannelsInUse()) return;

          //fprintf(stderr,"SysSoundDriver doing update\n");

          // create new sound block
          SoundBlock block(block_size);

          Memory=block.Memory();
          MemorySize=block_size;
             	  
          // call mixing function
	  mixingFunction(); 

          // tell device to play this soundblock
          device.Play(block);
      
}



