/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include "util/def.h"
#include "util/types.h"
#include "geometry/math3d.h"
#include "geometry/csrect.h"
#include "render/igraph2d.h"
#include "system/system.h"
#include "util/archive.h"
#include "util/inifile.h"
#include "engine/camera.h"
#include "engine/texture.h"
#include "engine/polygon/polygon.h"
#include "engine/polygon/polytext.h"
#include "engine/sector.h"
#include "render/igraph3d.h"
#include "general/console.h"
#include "sound/cssnddrv.h"
#include "util/input/cseventq.h"
#include "util/input/csinput.h"
#include "system/sysdriv.h"

// The global system variable
csSystemDriver *System = NULL;

// Make Shutdown static so that even if System has not been initialized,
// application can tell system driver to exit immediately
bool csSystemDriver::Shutdown = false;
bool csSystemDriver::ExitLoop = false;
bool csSystemDriver::DemoReady = false;

// Global configuration file
csIniFile *config = NULL;

// Global debugging level.
int csSystemDriver::debug_level = 0;

// the CLSID of the rendersystem to use.
static CLSID clsidRenderSystem;

// IUnknown implementation for csSystemDriver
IMPLEMENT_UNKNOWN_NODELETE( csSystemDriver )

BEGIN_INTERFACE_TABLE( csSystemDriver )
    IMPLEMENTS_COMPOSITE_INTERFACE( System )
END_INTERFACE_TABLE()


void default_fatal_exit (int errorcode, bool canreturn)
{
  (void)canreturn;
  exit (errorcode);
}

void (*fatal_exit) (int errorcode, bool canreturn) = default_fatal_exit;

csSystemDriver::csSystemDriver()
{
  pprintf_init();
  System = this;
  FullScreen = false;
  Mouse = NULL;
  Keyboard = NULL;
  
  piG3D = NULL;
  piG2D = NULL;
  piGI = NULL;

  EventQueue = NULL;
  Console = NULL;
  Sound = NULL;
  IsFocused = true;

  csCoInitialize(0);
}

csSystemDriver::~csSystemDriver ()
{
  Close ();
  
  CHK (delete Mouse);
  CHK (delete Keyboard);
  CHK (delete EventQueue);
  CHK (delete Console);
  CHK (delete Sound);

  FINAL_RELEASE( piG3D );
  FINAL_RELEASE( piG2D );
  FINAL_RELEASE( piGI );

  csCoUninitialize();
}

#ifdef CS_STATIC_LINKED
extern "C" {
    extern HRESULT Software_DllInitialize ();
#ifdef DO_X11
    extern HRESULT XLib2D_DllInitialize ();
#endif
}
#endif

void csSystemDriver::Initialize (int argc, char *argv[])
{
  HRESULT hRes;

  SetSystemDefaults ();
  ParseCmdLine (argc, argv);
  
  // DAN 9.29.98 - the DRIVER variable is now the ProgID for the comclass to use to
  // retrieve the IGraphics3D, IGraphics2D, and IGraphicsInfo interfaces. 

  // for instance, "DRIVER=software" is now "DRIVER=crystalspace.graphics3d.software"
  char *p = config->GetStr ("VideoDriver", "DRIVER", "crystalspace.graphics3d.software");

#ifdef CS_STATIC_LINKED
  // Call here the DllInitialize functions of all static loaded DLL's that you require.
  Software_DllInitialize ();
# ifdef DO_X11
  XLib2D_DllInitialize ();
# endif
#endif

  hRes = csCLSIDFromProgID( &p, &clsidRenderSystem );
  if (FAILED(hRes))
  {
      CsPrintf(MSG_FATAL_ERROR, "Bad value '%s' for DRIVER in configuration file.", p);
      fatal_exit(0, false);
  }

#ifdef DO_MMX
 // DAN: This is commented out until I figure out a better way to do it.
 // csGraphics3D::do_mmx = config->GetYesNo ("VideoDriver", "MMX", true);
#endif

  CHK (EventQueue = new csEventQueue ());
  InitGraphics ();
  InitKeyboard ();
  InitMouse ();
  InitSound ();
}

bool csSystemDriver::Open (char *Title)
{
  // the initialization order is crucial! do not reorder!
  if (FAILED (piG3D->Open (Title)))
    return false;
  if (!Keyboard->Open (EventQueue))
    return false;
  if (!Mouse->Open (EventQueue))
    return false;
  if (config->GetYesNo ("SoundDriver", "ACTIVE", false))
  {
    if (!Sound->Open (config->GetInt ("SoundDriver", "FREQUENCY", 11025),
  		      config->GetYesNo ("SoundDriver", "16BITS", false),
		      config->GetYesNo ("SoundDriver", "STEREO", false)))
      return false;
    Sound->setVolume (config->GetFloat ("SoundDriver", "VOLUME", -1.0));
  }
  return true;
}

void csSystemDriver::Close(void)
{
  if (piG3D) piG3D->Close ();
  if (Keyboard) Keyboard->Close ();
  if (Mouse) Mouse->Close ();
  if (Sound) Sound->Close ();
}

bool csSystemDriver::InitGraphics ()
{
  HRESULT hRes;
  IGraphicsContextFactory* pFactory;
  
  hRes = csCoGetClassObject( clsidRenderSystem, CLSCTX_INPROC_SERVER, NULL, IID_IGraphicsContextFactory, (void**)&pFactory );
  if (FAILED(hRes)) goto OnError;

  hRes = pFactory->CreateInstance( (REFIID)IID_IGraphics3D, GetISystemFromSystem(this), (void**)&piG3D );
  if (FAILED(hRes)) goto OnError;

  hRes = piG3D->Get2dDriver( &piG2D );
  if (FAILED(hRes)) goto OnError;

  hRes = piG2D->QueryInterface( (REFIID)IID_IGraphicsInfo, (void**)&piGI );
  if (FAILED(hRes)) goto OnError;

OnError:
  
  if (FAILED (hRes))
  {
    CsPrintf (MSG_FATAL_ERROR, "Error loading graphics context server.");
    fatal_exit (0, false);
  }

  return (piG3D != NULL  &&  piG2D != NULL  &&  piGI != NULL);
}

bool csSystemDriver::InitKeyboard ()
{
  CHK (Keyboard = new SysKeyboardDriver());
  return (Keyboard != NULL);
}

bool csSystemDriver::InitMouse ()
{
  CHK (Mouse = new SysMouseDriver ());
  return (Mouse != NULL);
}

bool csSystemDriver::InitSound ()
{
  int nchannels = config->GetInt ("SoundDriver", "MAX_CHANNELS", 16);
#ifdef DO_SOUND
  CHK (Sound = new SysSoundDriver (nchannels));
#else
  CHK (Sound = new csSoundDriver (nchannels)); // Set default interface (dummy)
#endif
  return (Sound != NULL);
}

void csSystemDriver::NextFrame (long /*elapsed_time*/, long /*current_time*/)
{
  if (Sound) Sound->update ();
}

void csSystemDriver::SetSystemDefaults ()
{
  FrameWidth = config->GetInt ("VideoDriver", "WIDTH", 640);
  FrameHeight = config->GetInt ("VideoDriver", "HEIGHT", 480);
  Depth = config->GetInt ("VideoDriver", "DEPTH", 8);
}

void csSystemDriver::ParseCmdLine (int argc, char* argv[])
{
  int i;

  for (i = 1; i < argc; i++)
    ParseArg (argc, argv, i);
}

void csSystemDriver::ParseArg (int argc, char* argv[], int& i)
{
  if (strcasecmp ("-help", argv[i]) == 0)
  {
    Help ();
	 fatal_exit (0,false);
  }
  else if (strcasecmp ("-debug", argv[i]) == 0)
  {
    G3DSettings.do_debug = true;
    CsPrintf (MSG_INITIALIZATION, "Debugging info enabled.\n");
  }
  else if (strcasecmp ("-nodebug", argv[i]) == 0)
  {
    G3DSettings.do_debug = false;
    CsPrintf (MSG_INITIALIZATION, "Debugging info disabled.\n");
  }
  else if (strcasecmp ("-light", argv[i]) == 0)
  {
    PolyTexture::do_light = true;
    CsPrintf (MSG_INITIALIZATION, "Lighting system enabled.\n");
  }
  else if (strcasecmp ("-nolight", argv[i]) == 0)
  {
    PolyTexture::do_light = false;
    CsPrintf (MSG_INITIALIZATION, "Lighting system disabled.\n");
  }
  else if (strcasecmp ("-ilace", argv[i]) == 0)
  {
    G3DSettings.cfg_interlacing = 0;
    CsPrintf (MSG_INITIALIZATION, "Interlacing enabled (software version only).\n");
  }
  else if (strcasecmp ("-noilace", argv[i]) == 0)
  {
    G3DSettings.cfg_interlacing = -1;
    CsPrintf (MSG_INITIALIZATION, "Interlacing disabled.\n");
  }
  else if (strcasecmp ("-fov", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%d", &csCamera::aspect);
  }
  else if (strcasecmp ("-dmipmap1", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%f", &Polygon3D::zdist_mipmap1);
  }
  else if (strcasecmp ("-dmipmap2", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%f", &Polygon3D::zdist_mipmap2);
  }
  else if (strcasecmp ("-dmipmap3", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%f", &Polygon3D::zdist_mipmap3);
  }
  else if (strcasecmp ("-gamma", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%f", &Textures::Gamma);
  }
  else if (strcasecmp ("-cache", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%d", &G3DSettings.cache_size);
  }
  else if (strcasecmp ("-recalc", argv[i]) == 0)
  {
    Polygon3D::do_force_recalc = true;
    CsPrintf (MSG_INITIALIZATION, "Force recalculation of lightmaps.\n");
  }
  else if (strcasecmp ("-reflect", argv[i]) == 0)
  {
    i++;
    if (i < argc) sscanf (argv[i], "%d", &csSector::cfg_reflections);
    CsPrintf (MSG_INITIALIZATION, "Set number of reflections to %d.\n", csSector::cfg_reflections);
  }
  else if (strcasecmp ("-rad", argv[i]) == 0)
  {
    csSector::do_radiosity = true;
    CsPrintf (MSG_INITIALIZATION, "Pseudo-radiosity enabled.\n");
  }
  else if (strcasecmp ("-norad", argv[i]) == 0)
  {
    csSector::do_radiosity = false;
    CsPrintf (MSG_INITIALIZATION, "Pseudo-radiosity disabled.\n");
  }
  else if (strcasecmp ("-accthg", argv[i]) == 0)
  {
    PolyTexture::do_accurate_things = true;
    CsPrintf (MSG_INITIALIZATION, "Accurate shadow calculation of Things enabled.\n");
  }
  else if (strcasecmp ("-noaccthg", argv[i]) == 0)
  {
    PolyTexture::do_accurate_things = false;
    CsPrintf (MSG_INITIALIZATION, "Accurate shadow calculation of Things disabled.\n");
  }
  else if (strcasecmp ("-cosfact", argv[i]) == 0)
  {
    i++;
    if (i < argc)
    {
      sscanf (argv[i], "%f", &PolyTexture::cfg_cosinus_factor);
      CsPrintf (MSG_INITIALIZATION, "Cosinus factor set to %f.\n", PolyTexture::cfg_cosinus_factor);
    }
  }
  else if (strcasecmp ("-mixing", argv[i]) == 0)
  {
    i++;
    if (i < argc) if (!Textures::force_mixing (argv[i])) fatal_exit (0,false);
  }
  else if (strcasecmp ("-txtmode", argv[i]) == 0)
  {
    i++;
    if (i < argc) if (!Textures::force_txtmode (argv[i])) fatal_exit (0,false);
  }
  else if (strcasecmp ("-fastrgb", argv[i]) == 0)
  {
    if (!Textures::force_accurate_rgb_mapper (false)) fatal_exit (0,false);
  }
  else if (strcasecmp ("-slowrgb", argv[i]) == 0)
  {
	 if (!Textures::force_accurate_rgb_mapper (true)) fatal_exit (0,false);
  }
  else if (strcasecmp ("-mode", argv[i]) == 0)
  {
    if (++i < argc) SetMode (argv[i]);
  }
  else if (strcasecmp ("-depth", argv[i]) == 0)
  {
    if (++i < argc) Depth = atoi (argv[i]);
  }
  else if (argv[i][0] == '-')
  {
    CsPrintf (MSG_FATAL_ERROR, "Unknown commandline parameter '%s'!\n\
      (Use '-help' to get a list of options)\n", argv[i]);
    fatal_exit (0,false);
  }
}

void csSystemDriver::Help ()
{
  CsPrintf (MSG_STDOUT, "Commandline options:\n");
  CsPrintf (MSG_STDOUT, "  -exec <filename>   execute this script in 'data'\n");
  CsPrintf (MSG_STDOUT, "  -help              this help\n");
  CsPrintf (MSG_STDOUT, "  -debug/nodebug     keep more debugging info for crashes (default '%sdebug')\n", G3DSettings.do_debug ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -light/nolight     lighting system (default 'light')\n");
  CsPrintf (MSG_STDOUT, "  -ilace/noilace     interlacing (default '%silace')\n", G3DSettings.cfg_interlacing == -1 ? "no" : "");
  CsPrintf (MSG_STDOUT, "  -gamma <value>     set the gamma for the display (default %f)\n", Textures::Gamma);
  CsPrintf (MSG_STDOUT, "  -fov <aspect>      set the FOV (default=display height)\n");
  CsPrintf (MSG_STDOUT, "  -dmipmap1 <dist>   set the mipmap distance 1 (default=%f)\n", Polygon3D::zdist_mipmap1);
  CsPrintf (MSG_STDOUT, "  -dmipmap2 <dist>   set the mipmap distance 2 (default=%f)\n", Polygon3D::zdist_mipmap2);
  CsPrintf (MSG_STDOUT, "  -dmipmap3 <dist>   set the mipmap distance 3 (default=%f)\n", Polygon3D::zdist_mipmap3);
  CsPrintf (MSG_STDOUT, "  -cache <size>      set the texture cache size (default=%d)\n", G3DSettings.cache_size);
  CsPrintf (MSG_STDOUT, "  -line <lines>      set the number of console lines (default 6)\n");
  CsPrintf (MSG_STDOUT, "  -confg <r,g,b>     set r,g,b value for console text (default 255,255,255)\n");
  CsPrintf (MSG_STDOUT, "  -recalc            force recalculation of lightmaps (VERY slow)\n");
  CsPrintf (MSG_STDOUT, "  -reflect <num>     max number of reflections (with -recalc) (default %d)\n", csSector::cfg_reflections);
  CsPrintf (MSG_STDOUT, "  -rad/norad         pseudo-radiosity system (default '%srad')\n", csSector::do_radiosity ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -accthg/noaccthg   accurate shadow calculation of Things (default '%saccthg')\n", PolyTexture::do_accurate_things ? "" : "no");
  CsPrintf (MSG_STDOUT, "  -cosfact <float>   set the cosinus factor used for lighting (default %f)\n", PolyTexture::cfg_cosinus_factor);
  CsPrintf (MSG_STDOUT, "  -mixing <mode>     set the mode used for mixing\n");
  CsPrintf (MSG_STDOUT, "  -txtmode <mode>    set the texture mode\n");
  CsPrintf (MSG_STDOUT, "  -fastrgb           use fast (less accurate) RGB mapper\n");
  CsPrintf (MSG_STDOUT, "  -slowrgb           use slow (accurate) RGB mapper\n");
  CsPrintf (MSG_STDOUT, "  -mode <w>x<y>      set resolution (default=%dx%d)\n", FRAME_WIDTH, FRAME_HEIGHT);
  CsPrintf (MSG_STDOUT, "  -depth <d>         set depth (default=%d bpp)\n", Depth);
}

void csSystemDriver::SetMode (char* mode)
{
  int wres, hres;
  if (sscanf(mode, "%dx%d", &wres, &hres) != 2)
  {
    CsPrintf (MSG_INITIALIZATION, "Mode %s unknown : assuming '-mode %dx%d'\n", mode,
      FRAME_WIDTH, FRAME_HEIGHT);
  }
  else
  {
    FRAME_WIDTH = wres;
    FRAME_HEIGHT = hres;
  }
}

void csSystemDriver::Alert (char* msg)
{
  pprintf (msg);
  debugprintf (true, msg);
}

void csSystemDriver::Warn (char* msg)
{
  pprintf (msg);
  debugprintf (true, msg);
}

void csSystemDriver::Printf (int mode, char* str, ...)
{
  char buf[1024];
  va_list arg;

  va_start (arg, str);
  vsprintf (buf, str, arg);
  va_end (arg);

  switch (mode)
  {
    case MSG_INTERNAL_ERROR:
    case MSG_FATAL_ERROR:
      if (System)
        System->Alert (buf);
      else
      {
        pprintf (buf);
        debugprintf (true, buf);
      }
      break;

    case MSG_WARNING:
      if (System)
        System->Warn (buf);
      else
      {
        pprintf (buf);
        debugprintf (true, buf);
      }
      break;

    case MSG_INITIALIZATION:
      pprintf (buf);
      debugprintf (true, buf);
      if (System->DemoReady)
        System->DemoWrite (buf);
      break;

    case MSG_CONSOLE:
      if (System && System->Console)
        System->Console->PutText (buf);
      else
        pprintf (buf);
      break;

    case MSG_STDOUT:
      pprintf (buf);
      break;

    case MSG_DEBUG_0:
      debugprintf (false, buf);
      break;

    case MSG_DEBUG_1:
      if (debug_level >= 1) debugprintf (false, buf);
      break;

    case MSG_DEBUG_2:
      if (debug_level >= 2) debugprintf (false, buf);
      break;

    case MSG_DEBUG_0F:
      debugprintf (true, buf);
      break;

    case MSG_DEBUG_1F:
      if (debug_level >= 1) debugprintf (true, buf);
      break;

    case MSG_DEBUG_2F:
      if (debug_level >= 2) debugprintf (true, buf);
      break;
  } /* endswitch */
}

void csSystemDriver::DemoWrite (char* buf)
{
  if (Console)
  {
    
    if (piG2D)
    {
        piG2D->Clear (0);
    }
    Console->PutText (buf);
    Console->Print ();
    
    if (piG2D)
    {
      piG2D->FinishDraw ();
      piG2D->Print (NULL);
    }
  
  }
}

/*
* Print a message on the console or in stdout/debug.txt.
*/
void csSystemDriver::debugprintf (bool flush, char *str, ...)
{
  char buf[1024];
  va_list arg;

  va_start (arg, str);
  vsprintf (buf, str, arg);
  va_end (arg);

  static FILE* fp = NULL;
  if (!fp) fp = fopen ("debug.txt", "a+");
  if (fp)
  {
    fprintf (fp, "%s", buf);
    if (flush) fflush (fp);
  }
}

void csSystemDriver::do_focus (int enable)
{
  if (enable == (int)IsFocused)
    return;
  IsFocused = enable;

  CHK (EventQueue->Put (new csEvent (csevBroadcast, cscmdFocusChanged, (void *)enable)));
  if (!enable)
  {
    Keyboard->Reset ();
    Mouse->Reset ();
  }
}

// COM implementation

IMPLEMENT_COMPOSITE_UNKNOWN_AS_EMBEDDED( csSystemDriver, System )

STDMETHODIMP csSystemDriver::XSystem::GetDepthSetting(int& retval)
{
    METHOD_PROLOGUE( csSystemDriver, System );
    retval = pThis->Depth;
    return S_OK;
}

STDMETHODIMP csSystemDriver::XSystem::GetFullScreenSetting(bool& retval)
{
    METHOD_PROLOGUE( csSystemDriver, System );
    retval = pThis->FullScreen;
    return S_OK;
}

STDMETHODIMP csSystemDriver::XSystem::GetHeightSetting(int& retval)
{
    METHOD_PROLOGUE( csSystemDriver, System );
    retval = pThis->FrameHeight;
    return S_OK;
}

STDMETHODIMP csSystemDriver::XSystem::GetWidthSetting(int& retval)
{
    METHOD_PROLOGUE( csSystemDriver, System );
    retval = pThis->FrameWidth;
    return S_OK;
}

STDMETHODIMP csSystemDriver::XSystem::Print(int mode, char* string)
{
    METHOD_PROLOGUE( csSystemDriver, System );
    pThis->Printf(mode, string);
    return S_OK;
}

STDMETHODIMP csSystemDriver::XSystem::GetShutdown (bool &Shutdown)
{
  METHOD_PROLOGUE( csSystemDriver, System );
  Shutdown = pThis->Shutdown;
  return S_OK;
}
