/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SYSTEM_H
#define SYSTEM_H

#include <stdio.h>

#include "com/com.h"
#include "util/def.h"
#include "system/isystem.h"

class csEventQueue;
class csKeyboardDriver;
class csMouseDriver;
class csConsole;
class csSoundDriver;
class csWorld;

interface IGraphics3D;
interface IGraphics2D;
interface IGraphicsInfo;

/**
 * This is the interface to operating system.<p>
 * This driver takes care of all system-dependent parts such as
 * video hardware and input hardware. Note that system-dependent
 * source code should NOT write implementations for methods of
 * csSystemDriver (they are already implemented in system.cpp),
 * but inherit a new class from csSystemDriver, overriding desired
 * methods. Note that some methods it is required to override,
 * otherwise program simply will not compile (they are marked
 * as abstract).
 */
class csSystemDriver
{
public:
  /// Sound driver
  csSoundDriver* Sound;
  /// Program event queue
  csEventQueue *EventQueue;
  /// 3D Graphics context
  IGraphics3D* piG3D;
  /// 2D Graphics context
  IGraphics2D* piG2D;
  /// Graphics info context
  IGraphicsInfo* piGI;
  /// the width of this frame
  int FrameWidth;
  /// the height of this frame
  int FrameHeight;
  /// whether this device is full-screen or not.
  bool FullScreen;
  /// the bits-per-pixel of the display.
  int Depth;
  /// a pointer to the world object.
  csWorld* world;
  /// Keyboard driver
  csKeyboardDriver *Keyboard;
  /// Mouse driver
  csMouseDriver *Mouse;
  /// Set to non-zero to exit csSystemDriver::Loop()
  static bool Shutdown;
  /// Same as Shutdown but set manually by windowing system
  static bool ExitLoop;
  /// System console
  csConsole *Console;
  /// Debugging level (0 = no debug, 1 = normal debug, 2 = verbose debug)
  static int debug_level;
  /// true if demo console is ready
  static bool DemoReady;
  /// true if CrystalSpace visual is active (focused)
  bool IsFocused;

  /// Initialize system-dependent data
  csSystemDriver ();
  /// Deinitialize system-dependent parts
  virtual ~csSystemDriver ();

  /// This is usually called right after object creation
  virtual void Initialize (int argc, char *argv[]);

  /// Initialize 3D Graphics context object
  virtual bool InitGraphics ();
  /// Initialize Keyboard object
  virtual bool InitKeyboard ();
  /// Initialize Mouse object
  virtual bool InitMouse ();
  /// Initialize Sound system
  virtual bool InitSound ();

  /// this stores all of the cached command-line settings that are passed to IGraphics3D on creation.
  struct 
  {
    bool do_debug;
    int cfg_interlacing;
    int cache_size;
  } G3DSettings;

  /**
   * Open the graphics context (with optional title on titlebar),
   * mouse and keyboard.
   */
  virtual bool Open (char *Title);
  /// Close the system
  virtual void Close ();

  /**
   * System loop. This should be called last since it returns
   * only on program exit
   */
  virtual void Loop () = 0;

  /**
   * SysSystemDriver::Loop should call this method once per frame.
   * 'elapsed_time' is the time elapsed since the previous call to
   * NextFrame. 'current_time' is a global time counter.
   * The time is expressed in milliseconds.
   */
  virtual void NextFrame (long elapsed_time, long current_time);

  /// Sleep for given number of 1/1000 seconds (very inacurate)
  virtual void Sleep (int /*SleepTime*/) {}

  /**
   * System dependent function to set all system defaults (like
   * the default resolution, if SHM is used or not (only for X),
   * and so on...).<p>
   * This routine can also make use of the config file to get
   * the defaults from there (so instead of having to specify
   * -mode 640x480 everytime you can set this in the config file
   * and this routine should read that).
   */
  virtual void SetSystemDefaults ();

  /**
   * This is a system-independent function used to parse a
   * commandline. It will iterate all command-line arguments
   * through system-dependent ParsArg() method.
   */
  void ParseCmdLine (int argc, char* argv[]);

  /**
   * This is a function that prints the commandline help text.
   * If system has system-dependent switches, it should override
   * this method and type its own text (possibly invoking
   * csSystemDriver::Help() first).
   */
  virtual void Help ();

  /**
   * Return time in miliseconds (if not supported by a system
   * just return the time in seconds multiplied by 1000).
   * Default implementation is in 'def_sys'.
   */
  static long Time ();

  /**
   * General printf routine. This routine should eventually
   * replace all other variants of printf that exist in
   * Crystal Space. 'mode' is any of the MSG_... fields
   * as defined above.
   */
  static void Printf (int mode, char* str, ...);

  /// Print a string into debug.txt
  static void debugprintf (bool flush, char* str, ...);

  /**
   * Printf version that works on all systems.
   * Default implementation is in 'def_sys'.
   */
  static void printf_init ();
  static void printf_close ();
  static int printf (char *str, ...);

  /**
   * Open a file. This function first translates the directory
   * separator ('/' is used here) to whatever the system uses
   * and after that passes control to 'fopen'.
   * Default implementation of first function is in 'def_sys'.
   */
  static FILE* fopen (char* filename, char* mode);

protected:
  /**
   * This is a system-dependent function which eats a single
   * command-line option (like -help, ...). If system-dependent
   * part does not recognize the option, it should pass it to
   * its parent class method.
   */
  virtual void ParseArg (int argc, char* argv[], int& i);

  /**
   * Show an alert. This function is called by csPrintf and
   * should show an alert in case of an error.
   * The default implementation writes the message in debug.txt
   * and on standard output.
   */
  virtual void Alert (char* msg);

  /**
   * Show a warning. This function is called by csPrintf and
   * should show an alert in case of a warning.
   * The default implementation writes a message in debug.txt
   * and on standard output.
   */
  virtual void Warn (char* msg);

  /**
   * Write a message on the display in demo mode (used by
   * Printf with mode MSG_INITIALIZATION).
   * The default implementation just uses the console.
   */
  virtual void DemoWrite (char* msg);

  /**
   * This is a system independent function that just initilizes
   * FRAME_WIDTH and FRAME_HEIGHT from the given mode string.
   */
  void SetMode (char* mode);

  /**
   * System-dependent code should call this to emit 'focus changed' events.
   * Initially system assumes that CrystalSpace window is active.
   */
  void do_focus (int enable);

  // the COM ISystem interface implementation.
  class XSystem : public ISystem
  {
      DECLARE_IUNKNOWN()
      STDMETHODIMP GetDepthSetting(int& retval);
      STDMETHODIMP GetFullScreenSetting(bool& retval);
      STDMETHODIMP GetHeightSetting(int& retval);
      STDMETHODIMP GetWidthSetting(int& retval);
      STDMETHODIMP Print(int mode, char* string);
      STDMETHODIMP GetShutdown (bool &Shutdown);
  };
  
  DECLARE_IUNKNOWN()
  DECLARE_INTERFACE_TABLE(csSystemDriver)
  DECLARE_COMPOSITE_INTERFACE_EMBEDDED (System);
};

#define GetISystemFromSystem(a) (&a->m_xSystem)

// Shortcuts for compatibility
#define pprintf_init csSystemDriver::printf_init
#define pprintf_close csSystemDriver::printf_close
#define pprintf csSystemDriver::printf
#define SystemFopen csSystemDriver::fopen
#define SystemGetTime csSystemDriver::Time

// CrystalSpace global variables
extern csSystemDriver *System;
extern csIniFile *config;
#define Gfx3D System->piG3D
#define Gfx2D System->piG2D
#define GfxInfo System->piGI
#define FRAME_WIDTH System->FrameWidth
#define FRAME_HEIGHT System->FrameHeight

// Fatal exit routine (which can be replaced if neccessary)
extern void (*fatal_exit) (int errorcode, bool canreturn);

#endif // SYSTEM_H
