/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Nathaniel Saint Martin <noote@bigfoot.com>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <windows.h>
#include <mmsystem.h>

#include "util/def.h"
#include "util/types.h"
#include "system/system.h"
#include "util/inifile.h"

#include "sound/wave.h"
#include "sound/channel.h"
#include "sound/module.h"
#include "sound/cssnddrv.h"

#include "system/win32/SysSound.h"

#define MYMMSYSERR_NOCREATETHREAD 12

void sys_fatalerror(char *s);

typedef struct
{
	SysSoundDriver *driver;
	HGLOBAL hHeader;
  LPBYTE lpSoundData;
} SoundBlock;

#define NUM_ThreadPriority 3
struct
{
	DWORD priority;
  char *name;
} ThreadPriority[NUM_ThreadPriority]=
{
  { THREAD_PRIORITY_LOWEST, "lowest"},
  { THREAD_PRIORITY_NORMAL, "normal"},
  { THREAD_PRIORITY_HIGHEST, "highest"}
};

SysSoundDriver::SysSoundDriver(int nchannels) : csSoundDriver(nchannels)
{
  BackGround = true;
  hwo = NULL;
  hThread = NULL;
  dwThread = 0;
}

bool SysSoundDriver::Open(int frequency, bool bit16, bool stereo)
{
  Active = false;
  
  MMRESULT res;
  WAVEFORMATEX format;
  memset(&format, 0, sizeof(PCMWAVEFORMAT));
	format.wFormatTag = WAVE_FORMAT_PCM;
	if(stereo)
    format.nChannels = 2;
  else
    format.nChannels = 1;
	format.nSamplesPerSec = frequency;
  if(bit16)
	  format.wBitsPerSample = 16;
  else
    format.wBitsPerSample = 8;
	format.nBlockAlign = format.nChannels * format.wBitsPerSample / 8;
  format.nAvgBytesPerSec = format.nBlockAlign * format.nSamplesPerSec;
	format.cbSize = 0;

  Frequency = frequency;
  Bits16 = bit16;
  Stereo = stereo;

  char *callback_func=config->GetStr("SoundDriver.waveOut", "CALLBACK", "function");
  char *thread_func=config->GetStr("SoundDriver.waveOut", "THREAD_PRIORITY", "normal");
  bool threading = false;

  if(stricmp(callback_func, "thread")==NULL)
    threading = true;

  int refresh=config->GetInt("SoundDriver.waveOut", "REFRESH", 5);

  if(refresh<1) refresh=1;
  else if(refresh>format.nAvgBytesPerSec) refresh=format.nAvgBytesPerSec;

  MemorySize = format.nAvgBytesPerSec/refresh;
  MemorySize -=(MemorySize%format.nBlockAlign);
  Memory = NULL;

	// Open the playback device
	if(threading)
  {
    CsPrintf (MSG_INITIALIZATION, "SoundDriver use thread method\n");
    hThread = CreateThread(NULL, 0, &waveOutThreadProc, this, 0, &dwThread);
    if(hThread)
    {
      for(int p=0; p<NUM_ThreadPriority; p++)
        if(stricmp(ThreadPriority[p].name, thread_func) == 0)
        {
          if(SetThreadPriority(hThread, ThreadPriority[p].priority)!=0)
            CsPrintf (MSG_INITIALIZATION, "SoundDriver thread set to %s priority\n", ThreadPriority[p].name);
          break;
        }
      res = waveOutOpen(&hwo, WAVE_MAPPER, &format, (LONG)dwThread, 0L, CALLBACK_THREAD);
    }
    else
      res = MYMMSYSERR_NOCREATETHREAD;
  }
  else
  {
    CsPrintf (MSG_INITIALIZATION, "SoundDriver use function callback method\n");
    res=waveOutOpen(&hwo, WAVE_MAPPER, &format, (LONG)&waveOutProc, 0L, CALLBACK_FUNCTION);
  }
  if(res!=MMSYSERR_NOERROR)
  {
    switch(res)
    {
    case MYMMSYSERR_NOCREATETHREAD:
      sys_fatalerror("WaveOut error : Cannot create thread !");
      break;

    case MMSYSERR_ALLOCATED:
      sys_fatalerror("WaveOut error : Ressource already allocated by an other program !");
      break;

    case MMSYSERR_BADDEVICEID:
      sys_fatalerror("WaveOut error : Bad device !");
      break;

    case MMSYSERR_NODRIVER:
      sys_fatalerror("WaveOut error : there is no device !");
      break;

    case MMSYSERR_NOMEM:
      sys_fatalerror("WaveOut error : Unable to allocate memory !");
      break;

    case WAVERR_BADFORMAT:
      sys_fatalerror("WaveOut error : Unsupported audio format !");
      break;

    default:
      sys_fatalerror("WaveOut error : Inknown Error !");
      break;
    }
    return false;
  }

  waveOutGetVolume(hwo, &old_Volume);
  int old_v2=old_Volume&0xffff;
  csSoundDriver::setVolume((float)old_v2/65535.0);
	
  if (MixChunk())
	{
		if (MixChunk())
		{
      CsPrintf (MSG_INITIALIZATION, "Sound initialized to %d Hz %d bits %s (%d channels)\n",
        Frequency, (Bits16)?16:8, (Stereo)?"Stereo":"Mono", numberChannels);
      Active = true;
      return true;
		}
	}

  sys_fatalerror("WaveOut error : Error then prepare chunk");
  return false;
}

bool SysSoundDriver::Close()
{
  Active = false;
  if (hwo)
	{
    waveOutSetVolume(hwo, old_Volume);
    waveOutReset(hwo);
		waveOutClose(hwo);
		hwo=NULL;
	}

  Memory = NULL;
  csSoundDriver::Close();

  return true;
}

void SysSoundDriver::setVolume(float v)
{
  if(v<0 || v>1) return;

  csSoundDriver::setVolume(v);
  if(hwo)
  {
    unsigned long av=v*65535.0;
    av=(av<<16)|av;
    waveOutSetVolume(hwo, av);
  }
}

static bool are_you_playing = false;

DWORD WINAPI SysSoundDriver::waveOutThreadProc( LPVOID dwParam)
{
  MSG msg;
  SysSoundDriver *snd = (SysSoundDriver *)dwParam;

  while(1)
  {
    WaitMessage();
    if(GetMessage( &msg, NULL, 0, 0 ) )
    {
      if(msg.message==MM_WOM_DONE && !are_you_playing)
      {
        if(snd==NULL || snd->hwo==NULL) continue;
        
        are_you_playing = true;
        
        CHK(SoundBlock * block = new SoundBlock);
        
        // Allocate memory for both the header and the actual sound data itself
        block->driver = snd;
        block->hHeader = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, sizeof(WAVEHDR)+snd->MemorySize);
        block->lpSoundData = (LPBYTE)GlobalLock(block->hHeader);
        
        // Set up various pointers
        LPWAVEHDR lpWaveHdr = (LPWAVEHDR)block->lpSoundData;
        
        snd->Memory = (unsigned char *)&block->lpSoundData[sizeof(WAVEHDR)];
        
        if(snd->Memory == NULL) continue;
        
        snd->mixingFunction();
        
        // Set up and prepare header. 
        lpWaveHdr->lpData = (char *)snd->Memory;
        lpWaveHdr->dwBufferLength = snd->MemorySize;
        lpWaveHdr->dwFlags = 0L;
        lpWaveHdr->dwLoops = 0L;
        lpWaveHdr->dwUser = (DWORD)block;
        waveOutPrepareHeader(snd->hwo, lpWaveHdr, sizeof(WAVEHDR));
        
        // Now the data block can be sent to the output device. The 
        // waveOutWrite function returns immediately and waveform 
        // data is sent to the output device in the background.
        MMRESULT result = waveOutWrite(snd->hwo, lpWaveHdr, sizeof(WAVEHDR));
        
        // Free this block up
        waveOutUnprepareHeader(snd->hwo, lpWaveHdr, sizeof(WAVEHDR));
        GlobalUnlock(block->hHeader);
        GlobalFree(block->hHeader);
        delete block;
        
        are_you_playing = false;
      }
      
      TranslateMessage(&msg); 
      DispatchMessage(&msg);
    }
    else return msg.wParam;
  }
    
  return -1;
}

void CALLBACK SysSoundDriver::waveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
  if (uMsg==MM_WOM_DONE)
  {
    if(are_you_playing) return;

    are_you_playing = true;

    LPWAVEHDR lpWaveHdr = (LPWAVEHDR)dwParam1;
    // void data ?
    if(lpWaveHdr->dwUser==NULL
      || lpWaveHdr->dwUser==0xcdcdcdcd) return;

    SoundBlock * block = (SoundBlock *)lpWaveHdr->dwUser;
    SysSoundDriver *me = block->driver;

    if(me==NULL) return;

    // Free this block up
    waveOutUnprepareHeader(me->hwo, lpWaveHdr, sizeof(WAVEHDR));
    GlobalUnlock(block->hHeader);
    GlobalFree(block->hHeader);
    delete block;

    // If we're still playing then mix another chunk
		me->MixChunk();
    
    are_you_playing = false;
  } 
}

bool SysSoundDriver::MixChunk()
{
  if(hwo==NULL) return false;

	CHK(SoundBlock * block = new SoundBlock);

	// Allocate memory for both the header and the actual sound data itself
	block->driver = this;
	block->hHeader = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, sizeof(WAVEHDR)+MemorySize);
	block->lpSoundData = (LPBYTE)GlobalLock(block->hHeader);

	// Set up various pointers
	LPWAVEHDR lpWaveHdr = (LPWAVEHDR)block->lpSoundData;

	Memory = (unsigned char *)&block->lpSoundData[sizeof(WAVEHDR)];

  if(Memory == NULL) return false;

  mixingFunction();

	// Set up and prepare header. 
	lpWaveHdr->lpData = (char *)Memory;
	lpWaveHdr->dwBufferLength = MemorySize;
	lpWaveHdr->dwFlags = 0L;
	lpWaveHdr->dwLoops = 0L;
	lpWaveHdr->dwUser = (DWORD)block;
	waveOutPrepareHeader(hwo, lpWaveHdr, sizeof(WAVEHDR));
 
	// Now the data block can be sent to the output device. The 
	// waveOutWrite function returns immediately and waveform 
	// data is sent to the output device in the background.
	MMRESULT result = waveOutWrite(hwo, lpWaveHdr, sizeof(WAVEHDR)); 
	if (result != 0) 
	{ 
		waveOutUnprepareHeader(hwo, lpWaveHdr, sizeof(WAVEHDR)); 
		GlobalUnlock(block->hHeader);
		GlobalFree(block->hHeader);
    return false;
	}

  return true;
}
