/*
    Copyright (C) 1998 by Jorrit Tyberghein
    Contributions made by Ivan Avramovic <iavramov@ibm.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef IMAGE_H
#define IMAGE_H

#include <stdio.h>
#include "util/types.h"

/// An RGB color.
struct RGBcolor
{
  unsigned char red, green, blue;

  ///
  RGBcolor() {}
  ///
  RGBcolor(unsigned char r, unsigned char g, unsigned char b) : 
   red(r), green(g), blue(b) {}
  ///
  bool operator== (const RGBcolor& c) 
  { return c.red == red && c.green == green && c.blue == blue; }
  ///
  bool operator!= (const RGBcolor& c) { return !(*this == c); }
  ///
  RGBcolor operator+ (const RGBcolor& c)
  { return RGBcolor(c.red+red, c.green+green, c.blue+blue); }
};

/// An RGB pixel.
struct RGBPixel : public RGBcolor
{
// r,g, and b values inherited from RGBcolor
  unsigned char pad;

  ///
  RGBPixel() {}
  ///
  RGBPixel(const RGBcolor& c) : RGBcolor(c), pad(0) {}
};

/// An RGB palette entry with statistics information.
struct RGBPalEntry : public RGBcolor
{
  long count;

  ///
  RGBPalEntry() {}
  ///
  RGBPalEntry(const RGBcolor& c) : RGBcolor(c), count(0) {}
};

/// Status flag indicating that the image has loaded without problem.
#define IFE_OK 0
/// Status flag indicating that the image data is in the wrong format.
#define IFE_BadFormat 1
/// Status flag indicating that the image data is corrupt.
#define IFE_Corrupt 2


/**
 * An abstract class implementing an image loader. For every image
 * type supported, a subclass should be created for loading that image
 * type.
 */
class ImageFile
{
private:
  /// Width of image.
  int width;
  /// Height of image.
  int height;

  /// The image data.
  RGBPixel* image;

protected:
  /**
   * ImageFile constructor.
   * This object can only be created by an appropriate loader, which is why
   * the constructor is protected.
   */
  ImageFile ();

  /**
   * Status of the loaded image.  
   * (status == IFE_OK) if the image loaded correctly.
   * (status & IFE_BadFormat) indicates that the image is in the wrong format.
   * (status & IFE_Corrupt) indicates that the image is in the correct format,
   * but the data is corrupt and unreadable.
   */
  int status;

  /**
   * Set the width and height. 
   * This will also allocate the 'image' buffer to hold the bitmap.
   */
  void set_dimensions (int w, int h);

  /// Get the buffer in which to write image data.
  RGBPixel* get_buffer() { return image; }

public:
  ///
  virtual ~ImageFile ();

  /// Returns the image width.
  int get_width () const { return width; }
  /// Returns the image height.
  int get_height () const { return height; }
  /// Returns the image size (in pixels)
  long get_size () const { return width * (long)height; }
  /// 
  const RGBPixel* get_image () const { return image; }

  /// Returns the error status of the loaded image.
  int get_status() const { return status; }
  /// Returns a text message explaining the image status.
  virtual const char* get_status_mesg() const;
};

///
class ImageColorInfo
{
private:
  /**
   * Calculate the color table from the image stored in 'buf', length 'size'. 
   * If the number of colors specified in 'ncol' is greater than zero, that
   * value is used.
   */
  void calc_table(const RGBPixel* buf, long size);

  ///
  int num_colors;
  ///
  RGBPalEntry* color_table;

public:
  /**
   * Construct the table using an image stored in 'buf' of length 'size'.
   * If the number of colors (ncol) is given and is greater than 0, that
   * value is used.
   */ 
  ImageColorInfo(const RGBPixel* buf, long size) { calc_table(buf,size); }
  /// Construct the color info table using the given image.
  ImageColorInfo(const ImageFile& img) { calc_table(img.get_image(), img.get_size() ); }

  ///
  ~ImageColorInfo();

  /**
   * Return a table with all colors used in the image.
   * This function returns a table with all colors that are used in the
   * image, together with a count of the number of times each color has
   * been used.  The table is sorted, with the most frequently occuring
   * colors first.
   */
  const RGBPalEntry* get_color_table () const { return color_table; }
  
  /// Returns the number of colors in the image.
  int get_num_colors() const { return num_colors; }
};


/// Forward declaration
class csVector;


/**
 * Extend this class to support a particular type of image loading.
 */
class ImageLoader
{
protected:
  /**
   * Load an image from the given buffer.
   * Attempts to read an image from the buffer 'buf' of length 'size'.
   * If successful, returns a pointer to the resulting ImageFile.  Otherwise
   * returns NULL.
   */
  virtual ImageFile* loadimage(UByte* buf, ULong size) = 0;

  /**
   * Register a loader for a given image type.
   * Adds 'loader' to the list of image formats to be checked during an
   * ImageLoader::load(...) call.
   */
  static void reg_loader(ImageLoader* loader);

  ///
  virtual ~ImageLoader() {}

public:
  /// Return the name of the image type supported by this loader.
  virtual const char* get_name() const = 0;

  /// Return a descriptive line about this image format.
  virtual const char* get_desc() const = 0;

  /**
   * Load an image given the filename.
   * This routine will open the file named filename, try to recognize its
   * type, and return an ImageFile of the appropriate type.  Returns a
   * pointer to the ImageFile on success, or NULL on failure.
   */
  static ImageFile* load (char* filename);

  /**
   * Load an image given a file pointer.
   * This routine will read from the given file, try to recognize its
   * type, and return an ImageFile of the appropriate type.  Returns a
   * pointer to the ImageFile on success, or NULL on failure.
   */
  static ImageFile* load (FILE* fp);

  /**
   * Load an image from a buffer.
   * This routine will read from the buffer buf of length size, try to 
   * recognize the type of image contained within, and return an ImageFile
   * of the appropriate type.  Returns a pointer to the ImageFile on 
   * success, or NULL on failure.
   */
  static ImageFile* load (UByte* buf, ULong size);

private:
  /// A static, NULL-terminated list of registered loaders.
  //static ImageLoader**& loaderlist();
  static csVector& loaderlist();
};

#endif

