/*
    Crystal Space input library
    Copyright (C) 1998 by Jorrit Tyberghein
    Written by Andrew Zabolotny <bit@eltech.ru>
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef __CSINPUT_H__
#define __CSINPUT_H__

/**
 * These are the low-level interfaces to input devices like keyboard
 * and mouse (possibly joystick will be added in the future).
 * System-dependent code should inherit system-specific classes from
 * those defined below, implementing as much functionality as possible.
 */

#include "util/csbase.h"
#include "util/input/csevent.h"
#include "util/input/cseventq.h"

/**
 * Keyboard driver maintains a structure containing the current state of
 * major control keys (see below). It is NOT (unlike in releases before 008)
 * the responsability of the keyboard driver to maintain this structure,
 * it is maintained by do_keypress and do_keyrelease methods. Instead, it
 * is highly desirable (read: required) for driver to call do_keypress and
 * do_keyrelease with appropiate key codes (see CSKEY_XXX constants below).
 */
struct csKey
{
  bool up;
  bool down;
  bool left;
  bool right;
  bool pgup;
  bool pgdn;
  bool esc;
  bool alt;
  bool ctrl;
  bool shift;
  bool home;
  bool end;
  bool ins;
  bool del;
  bool enter;
  bool tab;
  bool backspace;
};

/**
 * Control key codes
 */
/// ESCape key
#define CSKEY_ESC		27
/// Enter key
#define CSKEY_ENTER		'\n'
/// Tab key
#define CSKEY_TAB		'\t'
/// Back-space key
#define CSKEY_BACKSPACE	'\b'
/// Up arrow key
#define CSKEY_UP		1000
/// Down arrow key
#define CSKEY_DOWN		1001
/// Left arrow key
#define CSKEY_LEFT		1002
/// Right arrow key
#define CSKEY_RIGHT		1003
/// PageUp key
#define CSKEY_PGUP		1004
/// PageDown key
#define CSKEY_PGDN		1005
/// Home key
#define CSKEY_HOME		1006
/// End key
#define CSKEY_END		1007
/// Insert key
#define CSKEY_INS		1008
/// Delete key
#define CSKEY_DEL		1009
/// Control key
#define CSKEY_CTRL		1010
/// Alternative shift key
#define CSKEY_ALT		1011
/// Shift key
#define CSKEY_SHIFT		1012
/// Function keys
#define CSKEY_F1		1013
#define CSKEY_F2		1014
#define CSKEY_F3		1015
#define CSKEY_F4		1016
#define CSKEY_F5		1017
#define CSKEY_F6		1018
#define CSKEY_F7		1019
#define CSKEY_F8		1020
#define CSKEY_F9		1021
#define CSKEY_F10		1022
#define CSKEY_F11		1023
#define CSKEY_F12		1024
/// The "center" key ("5" on numeric keypad)
#define CSKEY_CENTER		1025

/**
 * This is the lowest-level interface to keyboard.<p>
 * Keyboard driver should generate events and put them into the
 * event queue which is passed to object on Open().
 */
class csKeyboardDriver
{
protected:
  /// Event queue which keyboard driver is bound to
  csEventQueue *EventQueue;

public:
  /// Partial key state: arrows, pgup/pgdn, ins/del, home/end, esc etc.
  csKey Key;
  /// Alphanumeric key state
  bool aKey[256];

  /// Initialize keyboard interface
  csKeyboardDriver ();
  /// Deinitialize keyboard interface
  virtual ~csKeyboardDriver ();

  /// Start receiving keyboard events
  virtual bool Open (csEventQueue *EvQueue);
  /// Finish receiving keyboard events
  virtual void Close ();

  /// Check if keyboard driver is ready
  virtual bool Ready ()
  { return (EventQueue != NULL); }

  /**
   * Call to release all key down flags
   * (when focus switches from application window, for example)
   */
  virtual void Reset ();

  /// Call this routine to add a key down event to queue
  void do_keypress (long current_time, int key);
  /// Call this routine to add a key up event to queue
  void do_keyrelease (long current_time, int key);

protected:
  /**
   * Set key state. For example SetKey (CSKEY_UP, true). Called
   * automatically by do_keypress and do_keyrelease.
   */
  void SetKey (int key, bool state);
};

/**
 * This is the lowest-level interface to mouse.<p>
 * Mouse driver should generate events and put them into the
 * event queue which is passed to object on Open().
 */
class csMouseDriver
{
protected:
  /// Event queue which keyboard driver is bound to
  csEventQueue *EventQueue;
  /// Last "mouse down" event time
  long LastClickTime;
  /// Last "mouse down" event button
  int LastClickButton;
  /// Last "mouse down" event position
  int LastClickX, LastClickY;
  /// Last mouse position
  int LastMouseX, LastMouseY;

public:
  /// Mouse buttons state (only 9 supported :-))
  bool Button[10];
  /// Mouse double click max interval in 1/1000 seconds
  static long DoubleClickTime;
  /// Mouse double click max distance
  static long DoubleClickDist;

  /// Initialize mouse interface
  csMouseDriver ();
  /// Deinitialize mouse interface
  virtual ~csMouseDriver ();

  /// Start receiving mouse events
  virtual bool Open(csEventQueue *EvQueue);
  /// Finish receiving mouse events
  virtual void Close ();

  /// Check if mouse driver is ready
  virtual bool Ready ()
  { return (EventQueue != NULL); }

  /**
   * Call to release all mouse buttons
   * (when focus switches from application window, for example)
   */
  virtual void Reset ();

  /// Call this to add a 'mouse button down' event to queue
  void do_buttonpress (long current_time, int button, int x, int y, bool shift, bool alt, bool ctrl);
  /// Call this to add a 'mouse button up' event to queue
  void do_buttonrelease (long current_time, int button, int x, int y);
  /// Call this to add a 'mouse moved' event to queue
  void do_mousemotion (long current_time, int x, int y);
};

#endif // __CSINPUT_H__
