/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef MEM_CHECK

#ifdef MEM_CHECK_EXTENSIVE
#  undef MEM_CHECK_FILL
#  define MEM_CHECK_FILL
#endif
#ifdef MEM_CHECK_FILL
#  define FILL_MEMORY(addr, count)					\
   {									\
     char *fill = addr;							\
     size_t garbage_count = count;					\
     while (garbage_count)						\
     {									\
       const char *garbage_str = "* this is garbage for filling memory *";\
       int how_many = MIN (garbage_count, strlen (garbage_str));	\
       memmove (fill, garbage_str, how_many);				\
       fill += how_many;						\
       garbage_count -= how_many;					\
     } /* endwhile */							\
   }
#else
#  define FILL_MEMORY(addr, count)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <ctype.h>
#include "memory.h"

volatile char MemFile[255];
volatile int MemLine = -1;
MemEntry* entries = NULL;
int num_entries = 0;
// rudimentary thread-safe lock
static int mem_lock = 0;

void* operator new (size_t s)
{
  while (mem_lock) ;
  mem_lock++;

  char* p = (char*)malloc (s+WALL_SIZE*2);
  memcpy (p, "SabcSdefSghiSjkl", WALL_SIZE);
  memcpy (p+s+WALL_SIZE, "EabcEdefEghiEjkl", WALL_SIZE);

  FILL_MEMORY (p + WALL_SIZE, s);

  MemEntry* me = (MemEntry*)malloc (sizeof (MemEntry));
  me->p = (void*)(p+WALL_SIZE);
  me->size = s;
  me->next = entries;
  me->prev = NULL;
  me->line = MemLine;
  me->file = NULL;
  me->freed = false;
  if (MemLine != -1)
  {
    me->file = (char*)malloc (strlen ((char *)MemFile)+1);
    strcpy (me->file, (char *)MemFile);
  }
  else
  {
    pprintf ("WARNING! Memory was allocated without CHK (%d bytes)!\n", s);
    me->file = NULL;
  }
  if (entries) entries->prev = me;
  entries = me;
  num_entries++;

  mem_lock--;
  return (void*)(p+WALL_SIZE);
}

void dump_memory (char* p, size_t len)
{
  int i;
  char* op = p;
  for (i = 0 ; i < (int)len ; i++)
  {
    pprintf ("%c", isprint (*p) ? *p : '.');
    p++;
  }
  p = op;
  pprintf ("    ");
  for (i = 0 ; i < (int)len ; i++)
  {
    pprintf ("%02x ", *p);
    p++;
  }
  pprintf ("\n");
}

void operator delete (void* p)
{
  while (mem_lock) ;
  mem_lock++;

  char* pp;

  MemEntry* e = entries;
  while (e)
  {
    if (e->p == p) break;
    e = e->next;
  }

#ifdef MEM_CHECK_EXTENSIVE
  if (e && e->freed)
  {
    pprintf ("ERROR! Block was already freed earlier!\n");
    pprintf ("       Error discovered while freeing memory at %s:%d.\n",
    	MemLine != -1 ? MemFile : "", MemLine);
    pprintf ("       Block was originally allocated at %s:%d.\n",
    	e->line != -1 ? e->file : "", e->line);
    pprintf ("       Block was first freed at %s:%d.\n",
    	e->free_line != -1 ? e->free_file : "", e->free_line);
    mem_lock--;
    return;
  }
#endif

  if (e)
  {
    pp = ((char*)p)-WALL_SIZE;
    if (strncmp (pp, "SabcSdefSghiSjkl", WALL_SIZE))
    {
      pprintf ("ERROR! Memory was overwritten before block allocated at %s:%d with size %d!\n",
      	e->file ? e->file : "", e->line, e->size);
      pprintf ("       Overwritten with: ");
      dump_memory (pp, WALL_SIZE);
      pprintf ("       Should be       : ");
      dump_memory ("SabcSdefSghiSjkl", WALL_SIZE);
      pprintf ("       Error discovered while freeing memory at %s:%d.\n",
    	MemLine != -1 ? MemFile : "", MemLine);
    }
    if (strncmp (pp+e->size+WALL_SIZE, "EabcEdefEghiEjkl", WALL_SIZE))
    {
      pprintf ("ERROR! Memory was overwritten after block allocated at %s:%d with size %d!\n",
      	e->file ? e->file : "", e->line, e->size);
      pprintf ("       Overwritten with: ");
      dump_memory (pp+e->size+WALL_SIZE, WALL_SIZE);
      pprintf ("       Should be       : ");
      dump_memory ("EabcEdefEghiEjkl", WALL_SIZE);
      pprintf ("       Error discovered while freeing memory at %s:%d.\n",
    	MemLine != -1 ? MemFile : "", MemLine);
    }

#ifdef MEM_CHECK_FILL
    FILL_MEMORY (((char *)e->p) - WALL_SIZE, e->size + WALL_SIZE*2);
#endif

#ifndef MEM_CHECK_EXTENSIVE
    free (pp);
    if (e->next) e->next->prev = e->prev;
    if (e->prev) e->prev->next = e->next;
    else entries = e->next;
    if (e->file) free (e->file);
    free (e);
    num_entries--;
#else
    int i;
    e->freed = true;
    e->free_line = MemLine;
    if (MemLine != -1)
    {
      e->free_file = (char*)malloc (strlen ((char *)MemFile)+1);
      strcpy (e->free_file, (char *)MemFile);
    }
    else e->free_file = NULL;
#endif
  }
  else
  {
    pprintf ("ERROR! Trying to free something that is not allocated!\n");
    pprintf ("       Error discovered while freeing memory at %s:%d.\n",
    	MemLine != -1 ? MemFile : "", MemLine);
  }
  mem_lock--;
}

#ifdef MEM_CHECK_EXTENSIVE

void check_mem ()
{
  int num = 0;
  MemEntry* e = entries;
  while (e)
  {
    if (e->freed)
    {
      char* p = ((char*)(e->p))-WALL_SIZE;
      int i;
      for (i = 0 ; i < (int)(e->size+WALL_SIZE*2) ; i++)
      {
        if (*p != 'X')
	{
          pprintf ("ERROR! Memory was overwritten after block allocated at %s:%d with size %d was freed!\n",
      	    e->file ? e->file : "", e->line, e->size);
          pprintf ("       Overwritten with: ");
          dump_memory (p-8, 16);
          pprintf ("       Should have been: ");
          dump_memory ("XXXXXXXXXXXXXXXX", 16);
          pprintf ("       Block was first freed at %s:%d.\n",
    	    e->free_line != -1 ? e->free_file : "", e->free_line);
	  break;
	}
	p++;
      }
    }
    else num++;
    e = e->next;
  }
}

class X { public: ~X () { check_mem (); } } x;

#endif // MEM_CHECK_EXTENSIVE

void dump_mem_list ()
{
  pprintf ("------------------- memory block list -------------------\n");
  MemEntry *cur = entries;
  while (cur)
  {
    if (!cur->freed)
      pprintf ("%08lX (%ld)  %s:%d\n", (unsigned long)cur->p, cur->size, cur->file, cur->line);
    cur = cur->next;
  } /* endwhile */
  pprintf ("---------------- end of memory block list ---------------\n");
}

#endif // MEM_CHECK
