/*
    Copyright (C) 1998 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "system/system.h"
#include "system/sysprint.h"
#include "system/sysdriv.h"
#include "util/parser.h"

const char *kWhiteSpace = " ";

/*----------------------------------------------------------------
  Pass in a pointer to a buffer of text.  This pointer is modified
  to point to the text after the object description.
  The token id for the object is returned.  The name pointer
  will point to the optional name string in the buffer.
  The data pointer will point to the optional data for
  the object.  The text buffer will get modified so BEWARE.
  eg text: ROOM "test room" ( 1, 2, 3 )
  returning -1 is an error.
  returning -2 is EOF.
----------------------------------------------------------------*/
long csGetObject (char **buf, tokenDesc * tokens, char **name, char **data)
{
  SkipCharacters (buf, kWhiteSpace);

  // skip comment lines.
  while (**buf == ';')
  {
    *buf = strchr (*buf, '\n');
    SkipCharacters (buf, kWhiteSpace);
  }

  if (!**buf)                   // at end of file
    return -2;

  // find the token.
  // for now just use brute force.  Improve later.
  tokenDesc *ptokens = tokens;

  while (tokens->id != 0)
  {
    if (!strncasecmp (tokens->token, *buf, strlen (tokens->token)))
    {
      // here we could be matching PART of a string
      // we could detect this here or the token list
      // could just have the longer tokens first in the list
      break;
    }
    ++tokens;
  }
  if (tokens->id == 0)
  {
    char *p = strchr (*buf, '\n');

    if (p)
      *p = 0;
    CsPrintf (MSG_FATAL_ERROR, "Found token '%s' which was not one of the list.\n", *buf);
    while (ptokens->id != 0)
    {
      CsPrintf (MSG_FATAL_ERROR, "    %s\n", ptokens->token);
      ptokens++;
    }
    fatal_exit (0, false);	// to help debugging
    return -1;                  // token not found !error!
  }
  // skip the token
  *buf += strlen (tokens->token);
  SkipCharacters (buf, kWhiteSpace);

  // get optional name
  *name = GetSubText (buf, 0x27, 0x27); // single quotes
  SkipCharacters (buf, kWhiteSpace);

  // get optional data
  if (**buf == '=')             // hmm, this is an assignment and not a command/object
    *data = GetAssignmentText (buf);
  else
    *data = GetSubText (buf, '(', ')');

  return tokens->id;
}

/*----------------------------------------------------------------
  Pass in a pointer to a buffer of text.  This pointer is modified
  to point to the text after the command description.
  The token id for the command is returned.
  The params pointer will point to the optional data for
  the object.  The text buffer will get modified so BEWARE.
  eg text: TEXTURE( 1, 2, 3 )
  returning -1 is an error.
  returning -2 is EOF.
  Note: Should be modified to accept assignments like:
        LIGHT=1
----------------------------------------------------------------*/
long csGetCommand (char **buf, tokenDesc * tokens, char **params)
{
  char *name;

  return csGetObject (buf, tokens, &name, params);
}

/*----------------------------------------------------------------
  Skips all control characters and any characters in the toSkip
  string.  Changes the buf pointer.
----------------------------------------------------------------*/
void SkipCharacters (char **buf, const char *toSkip)
{
  char theChar;
  char *skip;

  while ((theChar = **buf) != 0)
  {
    if (theChar >= 0x20)        // NOT control character, tab cr, lf, etc
    {
      skip = (char *) toSkip;
      while (*skip)
      {
        if (theChar == *skip)
          break;
        ++skip;
      }
      if (*skip == 0)           // exhausted the skip string
        return;                 // we are now at first NON skipped character
    }
    ++*buf;                     // skip this character
  }
  // we must be at the end of the buffer if we get here
}

/*----------------------------------------------------------------
  Returns the string of text between the open and close
  characters.  Modifies the buffer.  Moves the buffer pointer
  to after the last delimiter.  Can return nil;
  buffer MUST already be at the opening delimiter.
  Skips nested delimiters too.
  NOTE: Should skip quoted text, does not at this time.
----------------------------------------------------------------*/
char *GetSubText (char **buf, char open, char close)
{
  if (**buf == 0 || **buf != open)
    return 0;
  ++*buf;                       // skip opening delimiter
  char *result = *buf;

  // now find the closing delimiter
  char theChar;
  long skip = 1;

  if (open == close)            // we cant nest identical delimiters
    open = 0;
  while ((theChar = **buf) != 0)
  {
    if (theChar == open)
      ++skip;
    if (theChar == close)
    {
      if (--skip == 0)
      {
        **buf = 0;              // null terminate the result string
        ++*buf;                 // move past the closing delimiter
        break;
      }
    }
    ++*buf;                     // try next character
  }
  return result;
}

/*----------------------------------------------------------------
  Returns the string of text after a = up to the next
  whitespace.  Terminates the string and moves the buf pointer.
----------------------------------------------------------------*/
char *GetAssignmentText (char **buf)
{
  ++*buf;                       // skip the '='
  SkipCharacters (buf, kWhiteSpace);
  char *result = *buf;

  // now, we need to find the next whitespace to end the data
  char theChar;

  while ((theChar = **buf) != 0)
  {
    if (theChar <= 0x20)        // space and control chars
      break;
    ++*buf;
  }
  **buf = 0;                    // null terminate it
  if (theChar)                  // skip the terminator
    ++*buf;
  return result;
}
