#ifndef __SESSIONDATA_H

#define __SESSIONDATA_H


#define SESSIONSFILE "ktelnet.sessions"


#include <qobject.h>
#include <qlist.h>
#include <qstring.h>
#include <qcolor.h>
#include <qstrlist.h>



/**
 *  This is all data needed for a session. The ID for the toolbar is
 *  only necessary to get the right position on the toolbar and to 
 *  find the button again.
 */
struct SessionData
  {
    QString commandLine;

    QString terminal;
    QString font;
    QString connectionProgram;

    QString profileName;          
    QString host;
    int port;
    QString account;

    QColor fgcolor;
    QColor bgcolor;

    bool addToToolbar;
    QString toolbarIcon;

    // the session ID (needed for toolbar entries)
    int ID;
  };



/**
 *  This is the main data managing class of this application. Handles 
 *  saving and loading of configuration files. Starts sessions. constructs
 *  kdelnks from sessions.
 */
class SessionManager : public QObject
{

 Q_OBJECT

 public:

  SessionManager ();
  ~SessionManager ();

  /// return the current foreground color as QColor
  const QColor& fgColor ();
  const QColor& bgColor ();

  /// return current terminal
  const char* terminal ();
  const char* connector ();
  const char* account ();
  const char* host ();
  const char* cmdLine ();
  int port ();
  const char* textsize ();

  const char* icon ();

  /// return the icon at index i
  const char* icon (int i);

  /**
   * return toolbar ID for current session
   */
  int ID ();
  int ID (int i);

  /**
   * is session on toolbar?
   */
  bool addToToolbar ();
  bool addToToolbar (int i);

  /**
   * return the name of the current session
   */
  const char* session ();
  
  /**
   * return the session with index i (from 0 on), if index is out of range, 
   * a NULL pointer will be returned
   */
  const char* session (int i);

  /**
   * return the index of the current session
   */
  int current ();

  /**
   * return the number of elements
   */
  int count ();

  /** returns the terminal list. This is needed by mainwidget to build the
   *  Combobox.
   */
  QStrList& theTerminals ();

  /**
   *  If reject = TRUE all ...HasChanged... slots will not change the internal
   *  data structure. This is important when the current session should
   *  changes because all widgets emit signals to change data.
   */
  void rejectChangingSignals (bool reject);

  /**
   * returns the status of session saving
   */
  bool needSessionSave ();


 public slots:

   /**
    * Save the sessions
    */
  void saveSessions ();
 
  /** 
   * Deletes current session
   */
  void deleteSession ();

  /**
   *  append a session to the list of sessions. A new ID will be calculated.
   *  The appended session will be the current one at the end.
   *  Returns TRUE if everything worked all right, FALSE if session name is
   *  already in use. The theNewSession will be deleted on an error.
   */
  bool appendSession (SessionData* theNewSession);

  /**
   *  Set the current session. Should be connected to Cursor up and down and
   *  to the session combobox. This slot just sets the new session. Nothing
   *  more. Be careful, that all Widgets of mainWidget send signals on 
   *  contents change. So use rejectChangingSignals before this slot. Then
   *  change the Widgets contents manually and then call rejectChangingSignals
   *  again.
   */
  void newSessionSelected (const char* newSession);

  /// Start current session -> forwards to startSession (int) which is private.
  void startSession ();

  /**
   * For toolbar purpose:
   * start session with toolbarID ID.
   * Also forwards to startSession (int) which is private.
   */
  void startSession_ID (int ID);

  /** 
   * creates a kdelnk file from the current session
   */
  void createKDElnk ();

  /** 
   * should be called if the port number has changed in the LineEdit
   */
  void sessionNameHasChanged (const char* newName);
  void iconHasChanged (const char* newIcon);
  void portHasChanged (const char* newPort);
  void hostHasChanged (const char* newText);
  void accountHasChanged (const char* newText);
  void terminalHasChanged (const char* newTerminal);
  void textsizeHasChanged (const char* newTextsize);
  void connectorHasChanged (const char* newConnector);
  void fgcolorHasChanged (const QColor& newColor);
  void bgcolorHasChanged (const QColor& newColor);
  void cmdlineHasChanged (const char* newText);
  void shownOnToolbarHasChanged (bool show);


 signals:

  /**
   * always emitted if something need to be saved. Other slots which change
   * e.g. buttonstate can be connected.
   */
  void needSessionSaving (bool);

  /**
   * is emitted when a new session is really appended, so other widgets 
   * like the session combobox can update themselves.
   */
  void appendedNewSession (int index);
  
  /**
   * Other widgets like the session combobox can update themselves.
   */
  void deletedSession (const char* sessionName);


 private:

  /// Start session with index number.
  void startSession (int number);

  /**
   * Writes the commandline defined by the session with index i.
   * This commandline is "executable"! Uses commandLineList ().
   */
  QString commandLine (int i);
  
  /**
   * Calculates a list of arguments for starting the new session.
   * The first argument is the program name which should be started, of course.
   * Be shure to delete the List. This is your job.
   */
  QList<QString>& commandLineList (int i);

  /**
   * loads the sessions, at least one dummy session will be loaded if no
   * configuration file is found.
   */
  void loadSessions ();

  /**
   * Helper function to convert...
   */
  const char* QColor2string (QColor col);
  unsigned int getNewSessionID ();

  /**
   * internal data: the SessionList
   */
  QList<SessionData> theList;

  // the current entry
  int cur;

  bool needSaving;
  
  /**
   * This variable is set when a change of session is prepared by mainwidget.
   * No change signals (see above) will change state of needSaving.
   */
  bool rejectChangeSignals;

  // 
  // a list of terminals read out of the resource-file and their prefered
  // commandline options that differ from terminal to terminal
  //
  QStrList terminals;
  QStrList termFgColor;
  QStrList termBgColor;
  QStrList termFontSize;
};


#endif
