#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "diadef.h"
#include "dialog.h"
#include "dialog.m"
#include "../diajava/proto.h"

PUBLIC FIELD::FIELD(const char *_prompt)
{
	readonly = false;
	prompt = strdup(_prompt);
	box.width = 30;		// Default visible width
	maybeempty = true;
	may_select = false;
	is_head = false;
	vsize = 1;
}


PUBLIC VIRTUAL FIELD::~FIELD()
{
	free (prompt);
}

/*
	Return the notepad level. 0 for no notepad.
	Normal edit field return 0. FIELD_TITLE return their level.
	This allows DIALOG::editgui() to tell where are the FIELD_TITLE field
	and create a special form at the beginning of the dialog.
*/
PUBLIC VIRTUAL int FIELD::getnotepadlevel() const
{
	return 0;
}


/*
	Indicate that this field may not be empty.
	This does not apply to all field. For checkbox, it is meaningless.
*/
PUBLIC VIRTUAL void FIELD::set_noempty()
{
	maybeempty = false;
}

/*
	Indicate that this field may be selected or not.
*/
PUBLIC VIRTUAL void FIELD::set_selectable(bool _may_select)
{
	may_select = _may_select;
}
/*
	Return true if this field is selectable
*/
PUBLIC VIRTUAL bool FIELD::is_selectable()
{
	return may_select;
}

/*
	Return true if this field is a passthrough
*/
PUBLIC VIRTUAL bool FIELD::is_passthrough()
{
	return false;
}



/*
	This function is called when the user quits the dialog (with accept)
	each field may return -1 if something is not appropriate.

	The field is responsible to signal the error and it will become
	the current field immediatly.
*/
PUBLIC VIRTUAL int FIELD::post_validate()
{
	return 0;
}
/*
	Is the field editable or is it write protect.
	Title field do use this feature.
*/
PUBLIC bool FIELD::is_readonly()
{
	return readonly;
}
/*
	Set the readonly status of a field.
*/
PUBLIC void FIELD::set_readonly()
{
	readonly = true;
}

/*
	Process conditionnally a message and conditionnally redraw
	yourself if needed. See radio.c for the intended application.
*/
PROTECTED VIRTUAL void FIELD::processmsg(WINDOW *, FIELD_MSG &, int)
{
}
/*
	Get the width of the columns of a field.
	This is generally used by menu item and title field.
*/	  
PROTECTED VIRTUAL int FIELD::getwidths (
	int [],		// Will contain the width
	int &)		// Will contain the 1 if this field is starting
				// a new disposition set
				// It already contain 0 so no need to touch it
{
	return 0;
}
/*
	Set the width of the first column of a field.
	This is only meaningful for a Menu item.
*/	  
PROTECTED VIRTUAL void FIELD::setwidths (int, int [])
{
}
/*
	Return the second string of a menu item
*/
PUBLIC VIRTUAL const char *FIELD::getmenustr() const
{
	return NULL;
}

/*
	Build a key that uniquely identify this field in the dialog
*/
PUBLIC VIRTUAL void FIELD::format_htmlkey(char *key, int nof)
{
	html_formatkey (key,"%s-%d",prompt,nof);
}


PRIVATE void FIELD_STRING_BASE::init (int maxsiz)
{
	x.input = x.scroll = 0;
	password_mode = 0;
	size = maxsiz;
	buf = (char*)malloc_err (size+1);
	maybeempty = 1;
}

PROTECTED FIELD_STRING_BASE::FIELD_STRING_BASE(
	const char *_prompt,
	const char *_str,
	int maxsiz)
	: FIELD (_prompt)
{
	init(maxsiz);
	strncpy (buf,_str,maxsiz);
	buf[maxsiz] = '\0';
}

PROTECTED FIELD_STRING_BASE::FIELD_STRING_BASE(
	const char *_prompt,
	int maxsiz)
	: FIELD (_prompt)
{
	init(maxsiz);
	buf[0] = '\0';
}
PUBLIC FIELD_STRING_BASE::~FIELD_STRING_BASE()
{
	free (buf);
}

/*
	Return true if the field is empty
*/
PUBLIC bool FIELD_STRING_BASE::is_empty()
{
	bool ret = true;
	char *pt = buf;
	while (*pt != '\0'){
		if (*pt != ' '){
			ret = false;
			break;
		}
		pt++;
	}
	return ret;
}

PUBLIC int FIELD_STRING_BASE::post_validate()
{
	int ret = 0;
	if (!maybeempty && is_empty()){
		if (prompt[0] != '\0'){
			xconf_error (MSG_U(E_NOEMPTYNAME
				,"The field `%s`\nmay not be empty")
				,prompt);
		}else{
			xconf_error (MSG_U(E_NOEMPTY,"This field may not be empty"));
		}
		ret = -1;
	}
	return ret;
}
PUBLIC FIELD_STRING::FIELD_STRING(
	const char *_prompt,
	char *_str,
	int maxsiz)
	: FIELD_STRING_BASE (_prompt, _str, maxsiz)
{
	str = _str;
	backup = strdup(_str);
}

PUBLIC FIELD_STRING::~FIELD_STRING()
{
	free (backup);
}

PUBLIC FIELD_PASSWORD::FIELD_PASSWORD(
	const char *_prompt,
	SSTRING &_str)
	: FIELD_SSTRING (_prompt,_str)
{
	buf[0] = '\0';
	password_mode = 1;
}

PUBLIC void FIELD_STRING::save()
{
	strcpy (str,buf);
	strip_end (str);
}

/*
	Restore the original value of the field from the backup
*/
PUBLIC void FIELD_STRING::restore()
{
	strcpy (str,backup);
}
/*
	Reload the input buffer from the client variable value.
*/
PUBLIC void FIELD_STRING::reload()
{
	strcpy (buf,str);
}

/*
	Draw only the input part of a field
*/
PUBLIC void FIELD_STRING_BASE::drawtxt (WINDOW *dialog)
{
	int blank_start;
	wattrset(dialog, inputbox_attr);
	wmove(dialog, box.y,box.x);
	if (password_mode){
		blank_start = 0;
	}else{
		// Make sure the string is not too long
		char *instr = buf + x.scroll;
		char *last_visible = instr + box.width;
		char tmp = *last_visible;

		*last_visible = '\0';
		waddstr (dialog,instr);
		blank_start = strlen(instr);
		*last_visible = tmp;
	}
	const int box_width = box.width;
	for (int i=blank_start; i<box_width; i++) waddch (dialog,' ');
}

/*
	Draw the field with the prompt
*/
PUBLIC void FIELD_STRING_BASE::html_draw(int nof)
{
	html_printf ("<tr><td>%s<td>",prompt);
	if (readonly){
		html_printf ("%s\n",buf);
	}else{
		char key[100];
		format_htmlkey (key,nof);
		html_defvar (password_mode ? "password" : "text",key,buf
			,"size=30 maxlength=256");
		html_defvarcur (key,buf);
	}
}

PUBLIC void FIELD_STRING_BASE::gui_draw(int nof, int &)
{
	if (prompt[0] != '\0'){
		diagui_send_Label (prompt);
	}else{
		diagui_sendcmd (P_Skip,"1\n");
	}
	if (readonly){
		diagui_send_Label (buf);
	}else if (password_mode){
		char tmp[1000];
		diagui_sendcmd (P_Password,"S%d 30 %s\n",nof,diagui_quote(buf,tmp));
	}else{
		char tmp[1000];
		diagui_sendcmd (P_String,"S%d 30 %s\n",nof,diagui_quote(buf,tmp));
	}
}

PUBLIC void FIELD_STRING_BASE::gui_get(int nof)
{
	strcpy_cut (buf,diagui_getval('S',nof),size);
}
/*
	Get the letter use to define the ID of a field in GUI mode.
*/
PUBLIC char FIELD_STRING_BASE::getidprefix ()
{
	return 'S';
}

/*
	Grab the input of the user received from the www browser.
	Validate that the Draw the field with the prompt
	and compare the old value with the content of the buffer. If there
	is a mismatch, it means this www form was too old and the input
	is refused.
*/
PUBLIC int FIELD_STRING_BASE::html_validate(int nof)
{
	int ret = -1;
	char key[100];
	format_htmlkey (key,nof);
	const char *old_val = html_getoldval (key);
	const char *new_val = html_getval (key);
	fprintf (stderr,"validate %s val :%s: old :%s: buf :%s:\n",key,new_val,old_val,buf);
	if (strcmp(buf,old_val)==0
		&& (int)strlen(new_val)<size){
		strcpy (buf,new_val);
		ret = 0;
	}
	return ret;
}



/*
	Draw the field with the prompt and the surrounding box
*/
PUBLIC void FIELD::draw (WINDOW *dialog)
{
	int posy = box.y;
	wmove(dialog, posy,1);
	wattrset(dialog, dialog_attr);
	waddstr (dialog,prompt);
	int len = strlen(prompt);
	int lastx = box.x-2;
	for ( ; len < lastx; len++) waddch (dialog,' ');
	drawtxt(dialog);
}

/*
	Position the cursor in the field
*/
PUBLIC VIRTUAL void FIELD::setcursor(WINDOW *dialog)
{
	wmove (dialog,box.y,box.x);
}
/*
	Called when the field loose selection. This was
	done mostly for menu entries which is a very special case.
	The default behavior does nothing.
*/
PUBLIC VIRTUAL void FIELD::unselect(WINDOW *)
{
}
PUBLIC void FIELD_STRING_BASE::setcursor(WINDOW *dialog)
{
	wmove (dialog,box.y,box.x+x.input-x.scroll);
}

PUBLIC void FIELD_STRING_BASE::dokey (
	WINDOW *dialog,
	int key,
	FIELD_MSG &)
{
	if (readonly) return;
	int input_x = x.input;
	int scroll = x.scroll;
	const int box_width = box.width;
	const int box_y = box.y;
	const int box_x = box.x;
	char *instr = buf;
	int curlen = strlen(instr);
	int last_scroll = scroll;
	int mustdraw = 0;
	switch (key){
	case 1:		/* ^A like emacs */
	case KEY_HOME:
		input_x = 0;
		scroll = 0;
		break;
	case 5:		/* ^E like Emacs */
	case KEY_END:
		input_x = curlen;
		if (input_x > box_width){
			scroll = input_x - box_width + 1;
		}else{
			scroll = 0;
		}
		break;
	case KEY_LEFT:
		if (input_x > 0) input_x--;
		break;
	case KEY_RIGHT:
		if (input_x < curlen) input_x++;
		break;
	case KEY_BACKSPACE:
	case DEL:
		if (input_x > 0) {
			int i;
			input_x--;
			for (i=input_x; i<curlen; i++) instr[i] = instr[i+1];
			mustdraw = 1;
		}
		break;
	case 4:		/* ^d like Emacs */
	case KEY_DC:
		if (input_x < curlen){
			int i;
			for (i=input_x; i<curlen; i++) instr[i] = instr[i+1];
			mustdraw = 1;
		}
		break;
	case 11:	// ^k like Emacs: delete every char from actual position until the end
		instr[input_x] = '\0';
		mustdraw = 1;
		break;
	default:
		if (isprint(key)) {
			if (input_x < size) {
				wattrset(dialog, inputbox_attr);
				if (input_x < curlen){
					int i;
					for (i=size; i>input_x; i--) instr[i] = instr[i-1];
					mustdraw = 1;
				}
				instr[input_x] = key;
				int last = curlen + 1;
				if (last > size) last = size;
				instr[last] = '\0';
				input_x++;
				if (input_x - scroll == box_width) {
					scroll++;
				}else{
					wmove(dialog, box_y, input_x-1 + box_x);
					if (!password_mode) waddch(dialog, key);
				}
			}
		}
	}
	if (input_x < scroll){
		scroll--;
	}else if (input_x - scroll == box_width){
		scroll++;
	}
	x.input = input_x;
	x.scroll = scroll;
	if (mustdraw || last_scroll != scroll) drawtxt (dialog);
}


PUBLIC FIELD_SSTRING::FIELD_SSTRING(
	const char *_prompt,
	SSTRING &_str)
	: FIELD_STRING_BASE (_prompt, _str.get(), _str.getmaxsiz()), str(_str)
{
	backup.setfrom (_str);
}

PUBLIC void FIELD_SSTRING::save()
{
	str.setfrom(buf);
}

PUBLIC void FIELD_SSTRING::restore ()
{
	str.setfrom(backup);
}
PUBLIC void FIELD_SSTRING::reload ()
{
	strcpy (buf,str.get());
}
/*
	Draw the field with the prompt
*/
PUBLIC void FIELD_SSTRING::html_draw(int nof)
{
	html_printf ("<tr><td>%s<td>",prompt);
	if (readonly){
		html_printf ("%s\n",buf);
	}else{
		char key[100];
		format_htmlkey (key,nof);
		html_defvar (password_mode ? "password" : "text",key,buf
			,"size=30 maxlength=256");
		html_defvarcur (key,backup.get());
	}
}

/*
	Add a string field to the dialog.
*/
PUBLIC FIELD_STRING *DIALOG::newf_str(
	const char *prompt,
	char *str,
	int maxsiz)
{
	FIELD_STRING *s = new FIELD_STRING(prompt,str,maxsiz);
	add (s);
	return s;
}

/*
	Add an information field (write protect)
*/
PUBLIC void DIALOG::newf_info(
	const char *prompt,
	const char *str)
{
	FIELD_STRING *s = new FIELD_STRING(prompt,(char*)str,strlen(str));
	add (s);
	set_lastreadonly();
}

/*
	Add a SSTRING field to the dialog.
*/
PUBLIC FIELD_SSTRING *DIALOG::newf_str(
	const char *prompt,
	SSTRING &str)
{
	FIELD_SSTRING *s = new FIELD_SSTRING(prompt,str);
	add (s);
	return s;
}
/*
	Add a password field to the dialog.
*/
PUBLIC FIELD_PASSWORD *DIALOG::newf_pass(
	const char *prompt,
	SSTRING &str)
{
	FIELD_PASSWORD *s = new FIELD_PASSWORD(prompt,str);
	add (s);
	return s;
}
