#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>
#include "internal.h"
#include "dnsconf.h"
#include "dnsconf.m"
#include <netconf.h>
#include <dialog.h>

static DNSCONF_HELP_FILE help_secondary("secondary");


PUBLIC SECONDARY::SECONDARY(
	const char *_domain,
	const char *_backup,
	const char **_ip_addr,
	int nb_ip)
{
	domain.setfrom (_domain);
	backup.setfrom (_backup);
	for (int i=0; i<nb_ip; i++) addr[i].setfrom(_ip_addr[i]);
}

PUBLIC SECONDARY::SECONDARY()
{
}

PUBLIC void SECONDARY::write(bool bind8, FILE *fout, const char *named_dir) const
{
	/* #Specification: secondary / backup file / location
		dnsconf set the backup file in the subdirectory
		"sec" of the DNS configuration directory.
		It simply use the domain as the name of the file.

		You can override this by editing /etc/named.boot by
		hand. dnsconf will preserve the new path.

		The "sec" subdirectory will be created automaticly.
	*/
	SSTRING file;
	if (backup.is_empty()){
		char path[PATH_MAX];
		sprintf (path,"%s/sec",named_dir);
		context_mkdir (path,"root","root",0755);
		file.setfrom ("sec/");
		file.append (domain.get());
	}else{
		file.setfrom (backup);
	}
	if (bind8){
		fprintf (fout,"zone \"%s\"{\n",domain.get());
		fprintf (fout,"\ttype slave;\n");
		fprintf (fout,"\tfile \"%s\";\n",file.get());
		fputs ("\tmasters{\n",fout);
		for (int i=0; i<4; i++){
			const char *s = addr[i].get();
			if (s[0] != '\0') fprintf (fout,"\t\t%s;\n",s);
		}
		fputs ("\t};\n",fout);
		fputs ("};\n",fout);
	}else{
		fprintf (fout,"secondary\t%s",domain.get());
		for (int i=0; i<4; i++){
			fprintf (fout," %s",addr[i].get());
		}
		fprintf (fout,"\t%s\n",file.get());
	}
}


/*
	Edit the minimal spec so this dns will act as a secondary for another
	Return -1 if the user escaped away.
	Return  0 if the changes are accepted
	Return  1 if the user wish to delete this entry
*/
PUBLIC int SECONDARY::edit(DNS &dns)
{
	DIALOG dia;
	dia.newf_str (MSG_R(F_DOMAIN),domain);
	dia.newf_str (MSG_U(F_IPSERV,"IP address of the server"),addr[0]);
	dia.newf_str ("",addr[1]);
	dia.newf_str ("",addr[2]);
	dia.newf_str ("",addr[3]);
	int ret = -1;
	int nof = 0;
	while (1){
		MENU_STATUS status = dia.edit (
			 MSG_U(T_SECSPEC,"Secondary specification")
			,MSG_U(I_SECSPEC
			 ,"You must enter a domain and at least IP address\n"
			  "of one DNS server of that domain\n")
			,help_secondary
			,nof
			,MENUBUT_CANCEL|MENUBUT_ACCEPT|MENUBUT_DEL);
		if (status == MENU_CANCEL || status == MENU_ESCAPE){
			break;
		}else if (status == MENU_DEL){
			if (xconf_areyousure(MSG_U(Q_DELSECONDARY
				,"Confirm deletion of a secondary spec"))){
				ret = 1;
				break;
			}
		}else{
			const char *domname = domain.get();
			SECONDARY *sec = dns.findsecond (domname);
			PRIMARY  *dom = dns.finddomain (domname);
			if (domain.is_empty()
				|| addr[0].is_empty()){
				xconf_error(MSG_U(E_FILLDOM
					,"Fill at least the domain\n"
					 "and the first IP address"));
			}else if (sec != NULL && sec != this){
				xconf_error (MSG_R(E_DOMEXIST));
			}else if (dom != NULL){
				xconf_error (MSG_U(E_ISAPRIMARY
					,"This DNS is already a primary for this domain.\n"
					 "It can't be both at the same time"));
			}else{
				ret = 0;
				break;
			}
		}
	}
	if (ret != 0) dia.restore();
	return ret;
}

PUBLIC SECONDARY *SECONDARYS::getitem (int no) const
{
	return (SECONDARY*)ARRAY::getitem(no);
}
PUBLIC SECONDARY *SECONDARYS::getitem (const char *name) const
{
	SECONDARY *ret = NULL;
	int n = getnb();
	for (int i=0; i<n; i++){
		SECONDARY *s = getitem(i);
		if (s->domain.cmp(name)==0){
			ret = s;
			break;
		}
	}
	return ret;
}

PUBLIC void SECONDARYS::write (bool bind8, FILE *fout, const char *named_dir) const
{
	for (int i=0; i<getnb(); i++) getitem(i)->write(bind8,fout,named_dir);
}

PUBLIC void SECONDARYS::edit(DNS &dns)
{
	int choice=0;
	while (1){
		int nb = getnb();
		DIALOG dia;
		dia.newf_head ("",MSG_U(H_SECONDARIES,"domain\tprimary\tother"));
		for (int i=0; i<nb; i++){
			SECONDARY *s = getitem(i);
			char buf[4*LEN_IP_ASC];
			char *pt = buf;
			const char *ctl = "%s";
			for (int a=0; a<4; a++){
				if (!s->addr[a].is_empty()){
					pt += sprintf (pt,ctl,s->addr[a].get());
					ctl = "\t%s";
				}
			}
			*pt = '\0';
			dia.new_menuitem (s->domain.get(),buf);
		}
		dia.addwhat (MSG_U(I_ADDSEC,"Select [Add] to add one secondary spec"));
		MENU_STATUS code = dia.editmenu (MSG_U(T_SECONDARYS,"Secondaries")
			,MSG_U(I_SECONDARYS
				,"You are allowed to edit/add/remove secondaries\n")
			,help_secondary
			,choice,0);
		if (code == MENU_QUIT || code == MENU_ESCAPE){
			break;
		}else if (code == MENU_ADD){
			SECONDARY *sec = new SECONDARY;
			if (sec->edit(dns)==0){
				add (sec);
				dns.write();
			}else{
				delete sec;
			}
		}else{
			SECONDARY *sec = getitem(choice);
			if (sec != NULL){
				int ok = sec->edit(dns);
				if (ok != -1){
					if (ok == 1) remove_del (sec);
					dns.write();
				}
			}
		}
	}
}

