/* *** wav_interface ***
   *** Filename: wav_interface.c
   *** History: 19980121 Ver 1.00 | Michiel Bronswijk
   ***          19980225 Release version 1.00
   *** Description: Simple interface for creating Microsoft .WAV files
   *** Remarks: This is a very simple interface, just one WAV file
   *** at a time can be handled.
   ***          WAV structure is according to Microsoft specifications,
   ***		so if there's a bug, it's not mine!
   *** Changes:
*/

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "wav_interface.h"

#define	WORD		unsigned short
#define	DWORD		unsigned long
	
struct	fmtChunk_t {
	char		fmtSignature[4];
	DWORD		fmtSize;
	
	WORD		wFormatTag;
	WORD		wChannels;
	DWORD		dwSamplesPerSec;
	DWORD		dwAvgBytesPerSec;
	WORD		wBlockAlign;
	WORD		wBitsPerSample;
	};
	
struct dataChunk_t {	
	char		dataSignature[4];
	DWORD		dataSize;
	};


struct	WAVHeader_t {
	char			RIFFSignature[4];
	DWORD			RIFFSize;
	char			WAVESignature[4];
	struct fmtChunk_t	fmtChunk;
	struct dataChunk_t	dataChunk;
	};
	
/* Our .WAV filedescriptor */
int	WAV_fd;
int	WAV_size;


int	isWAVFile(const char *pathname, long *speed)
{
	int	my_fd;
	struct WAVHeader_t	WAVHeader;

	if (strlen(pathname) < 4) {
		printf("ERROR Filename too short\n");
		return(-1);
		}
	if (strcasecmp(&pathname[strlen(pathname)-4], ".wav") != 0) {
		printf("ERROR Wrong extension\n");
		return(-1);
		}
	if ((my_fd = open(pathname, O_RDONLY))==-1) {
		printf ("ERROR Couldn't open file\n");
		return(-1);
		}
	if (read(my_fd, &WAVHeader, sizeof(WAVHeader)) !=sizeof(WAVHeader)) {
		printf("ERROR: could not read file\n");
		close(my_fd);
		return(-1);
		};
	close(my_fd);
	if ( (strncmp(WAVHeader.RIFFSignature, "RIFF", 4) == 0) &
	     (strncmp(WAVHeader.WAVESignature, "WAVE", 4) == 0) &
	     (strncmp(WAVHeader.fmtChunk.fmtSignature, "fmt ", 4) == 0)) {
	        if (speed != NULL) {
	          *speed = WAVHeader.fmtChunk.dwAvgBytesPerSec;
	        }
	    	return(0); /* could be a valid .WAV */
	} else {
		printf("ERROR Signature not found in file\n");
		return(-1); /* definitely not a valid .WAV */
	}
}

/*	creatWAV:
	Creates a WAV file */
int	creatWAV(const char *pathname, int width, int channels, int rate)
{
	struct WAVHeader_t	WAVHeader;
	
	WAV_size = 0;
	if ((WAV_fd = open(pathname, O_CREAT | O_TRUNC | O_RDWR, S_IRUSR | S_IWUSR)) == -1)
		return(-1);
	strncpy(WAVHeader.RIFFSignature, "RIFF", 4);
	WAVHeader.RIFFSize	= sizeof(WAVHeader)-8;
	strncpy(WAVHeader.WAVESignature, "WAVE", 4);
	strncpy(WAVHeader.fmtChunk.fmtSignature, "fmt ", 4);
	WAVHeader.fmtChunk.fmtSize = sizeof(WAVHeader.fmtChunk)-8;
	WAVHeader.fmtChunk.wFormatTag = 0x0001;
	WAVHeader.fmtChunk.wChannels = channels;
	WAVHeader.fmtChunk.dwSamplesPerSec = rate;
	WAVHeader.fmtChunk.dwAvgBytesPerSec = rate*channels*(width/8);		
	WAVHeader.fmtChunk.wBlockAlign = channels*(width/8);
	WAVHeader.fmtChunk.wBitsPerSample = width;
	strncpy(WAVHeader.dataChunk.dataSignature, "data", 4);
	WAVHeader.dataChunk.dataSize = 0;	
	if (write(WAV_fd, &WAVHeader, sizeof(WAVHeader)) < sizeof(WAVHeader))
		return(-1);
	else
		return(0);
}

/*	writeWAV:
	Writes to the previously created WAV file */
int	writeWAV(const void *buf, size_t count)
{
	if (write(WAV_fd, buf, count) < count) 
		return(-1);
	WAV_size += count;
	return(count);
}

/*	closeWAV:
	Closes the previously created and written WAV file */
int	closeWAV(void)
{
	struct WAVHeader_t	WAVHeader;
	
	lseek(WAV_fd, 0, SEEK_SET);
	read(WAV_fd, &WAVHeader, sizeof(WAVHeader));
	lseek(WAV_fd, 0, SEEK_SET);
	WAVHeader.RIFFSize += WAV_size;
	WAVHeader.dataChunk.dataSize += WAV_size;
	write(WAV_fd, &WAVHeader, sizeof(WAVHeader));
	return(close(WAV_fd));
}
