/*
 *  dialog.c:		Info or warning dialog windows, option menu	
 * 
 *  Written by:		Ullrich Hafner
 *		
 *  Copyright (C) 1998 Ullrich Hafner <hafner@bigfoot.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

/*
 *  $Date: 1999/01/20 22:22:53 $
 *  $Author: hafner $
 *  $Revision: 1.17 $
 *  $State: Exp $
 */

#include "config.h"

#include <gtk/gtk.h>

#include "dialog.h"
#include "misc.h"
#include "icons.h"

#include "error.h"

/*******************************************************************************

				public code
  
*******************************************************************************/

void
dialog_popup (dialog_e type, const char *text,
	      void (*ok_function) (GtkWidget *, gpointer),
	      gpointer dataptr_ok_function)
/*
 *  Popup dialog to display an info, question, warning or error message
 *  with printf style syntax of 'format'.
 *  Message 'type' must be one of DIALOG_INFO, DIALOG_QUESTION,
 *  DIALOG_WARNING, DIALOG_ERROR.
 *  If 'ok_function' is not NULL then 'ok_function' is called in case
 *    user activates 'OK' button (which is labeled by 'ok_text').
 *    'dataptr_ok_function' argument is passed to 'ok_function'.
 *    A 'Cancel' button (which is labeled by 'close_text') is also shown.
 *  Otherwise only the 'Cancel' button is shown.
 *
 *  No return value.
 */
{
   if (ok_function)
      dialog (type, text, "No", "Yes", ok_function, dataptr_ok_function);
   else
      dialog (type, text, "Close", NULL, ok_function, dataptr_ok_function);
}

void
dialog (dialog_e type, const char *text, const char *close_text,
	const char *ok_text, void (*ok_function) (GtkWidget *, gpointer),
	gpointer dataptr_ok_function)
/*
 *  Popup dialog to display an info, question, warning or error message
 *  with printf style syntax of 'format'.
 *  Message 'type' must be one of DIALOG_INFO, DIALOG_QUESTION,
 *  DIALOG_WARNING, DIALOG_ERROR.
 *  If 'ok_function' is not NULL then 'ok_function' is called in case
 *    user activates 'OK' button (which is labeled by 'ok_text').
 *    'dataptr_ok_function' argument is passed to 'ok_function'.
 *    A 'Cancel' button (which is labeled by 'close_text') is also shown.
 *  Otherwise only the 'Cancel' button is shown.
 *
 *  No return value.
 */
{
   GtkWidget *hbox;
   GtkWidget *window, *button;
   char	     *window_title[] = {"Info", "Question", "Warning", "Error"};
   xpms_e    window_icon[] = {P_INFO, P_QUESTION, P_WARNING, P_ERROR};
   
   window = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (window), window_title [type]);
   gtk_widget_realize (window);

   gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		       GTK_SIGNAL_FUNC (gtk_false), NULL);
   gtk_signal_connect_object (GTK_OBJECT (window), "destroy",
			      GTK_SIGNAL_FUNC (gtk_widget_destroy),
			      GTK_OBJECT (window));

   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->vbox), hbox,
		       FALSE, FALSE, 0);
   gtk_container_border_width (GTK_CONTAINER (hbox), 10);
   
   gtk_box_pack_start (GTK_BOX (hbox),
		       gtk_pixmap_new (p_array [window_icon [type]].pixmap,
				       p_array [window_icon [type]].mask),
		       TRUE, FALSE, 5);
   
   gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new (text), TRUE, TRUE, 5);
   
   button = gtk_button_new_with_label (ok_function ? ok_text : close_text);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), button,
		       TRUE, TRUE, 10);
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
   gtk_widget_grab_default (button);
   if (ok_function)
   {
      gtk_signal_connect (GTK_OBJECT (button), "clicked",
			  GTK_SIGNAL_FUNC (ok_function), dataptr_ok_function);
      gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
				 (GtkSignalFunc) gtk_widget_destroy,
				 GTK_OBJECT (window));
      
      button = gtk_button_new_with_label (close_text);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (window)->action_area), button,
			  TRUE, TRUE, 10);
   }
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			      (GtkSignalFunc) gtk_widget_destroy,
			      GTK_OBJECT (window));

   gtk_widget_show_all (window);
}

GtkWidget *
generate_option_menu (const char *description,
		      GtkTooltips *tooltips, const char *info,
		      const char **entry, const char *current,
		      void (*callback) (GtkWidget *, gpointer),
		      gpointer callback_data)
/*
 *  Generate an option menu of all available text strings given in NULL
 *  terminated array 'entry'. Optional 'description' is prepended if not NULL.
 *  'current' is the string of the default option.
 *  'callback' function is called if user changes options,
 *  'callback_data' argument is passed to 'callback'.
 *
 *  Return value:
 *	container widget containing the option menu
 */
{
   GtkWidget	*hbox = NULL;
   GtkWidget	*menu;
   GtkWidget	*menu_item;
   GtkWidget	*option_menu;
   const char	**ptr;
   
   option_menu = gtk_option_menu_new ();
   menu = gtk_menu_new ();

   if (description)
   {
      hbox = gtk_hbox_new (FALSE, 5);
      gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new (description),
			  FALSE, FALSE, 5);
      gtk_widget_show_all (hbox);
   }

   for (ptr = entry; *ptr; ptr++)
   {
      menu_item = gtk_menu_item_new_with_label (*ptr);
      if (callback)
	 gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			     GTK_SIGNAL_FUNC (callback), callback_data);
      gtk_object_set_user_data (GTK_OBJECT (menu_item), (gpointer) *ptr);
      gtk_menu_append (GTK_MENU (menu), menu_item);
      gtk_widget_show (menu_item);
   }
   gtk_option_menu_set_menu (GTK_OPTION_MENU (option_menu), menu);
   set_option_menu_default (GTK_OPTION_MENU (option_menu), entry, current);
   
   gtk_widget_show (option_menu);
   if (info)
      gtk_tooltips_set_tip (tooltips, option_menu, info, NULL);  

   if (description)
   {
      gtk_box_pack_start (GTK_BOX (hbox), option_menu, FALSE, FALSE, 5);
      return hbox;
   }
   else
      return option_menu;
}

void
set_option_menu_default (GtkOptionMenu *menu, const char **list,
			 const char *value)
{
   const char **ptr;
   
   for (ptr = list; *ptr; ptr++)
      if (strcaseeq (value, *ptr))
      {
	 gtk_option_menu_set_history (GTK_OPTION_MENU (menu), ptr - list);
	 break;
      }
}
