/*
 *  texture.c:		Texture widgets, color widgets	
 *
 *  Written by:		Ullrich Hafner
 *  
 *  Copyright (C) 1998 Ullrich Hafner <hafner@bigfoot.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */

/*
 *  $Date: 1999/01/20 22:22:55 $
 *  $Author: hafner $
 *  $Revision: 1.44 $
 *  $State: Exp $
 */

#include "config.h"

#include <gtk/gtk.h>
#include <proplist.h>
#include <ctype.h>

#if HAVE_STRING_H
#	include <string.h>
#else /* not HAVE_STRING_H */
#	include <strings.h>
#endif /* not HAVE_STRING_H */

#ifdef TILE_OF_DAY      
#include "tileofday.h"
#endif /* TILE_OF_DAY */

#include "misc.h"
#include "icons.h"
#include "window.h"
#include "dialog.h"
#include "previews.h"
#include "texture.h"

#include "error.h"

/*******************************************************************************

		    global variables (imported from window.c)
  
*******************************************************************************/

extern proplist_t	*windowmaker;
extern bool_t		changed;
extern proplist_t	*pixmap_path;
extern proplist_t	*plist_changed;
extern proplist_t	*pl_yes;
extern GtkTooltips	*tooltips;

/*******************************************************************************

			    local variables
  
*******************************************************************************/

static const char *wtexture_types[]  = {"Solid color", "Gradient", "Pixmap",
					"Textured gradient", "None",
					"Default", NULL};
static const char *wwtexture_types[]  = {"Solid color", "Gradient", "Pixmap",
					 "Textured gradient", "None",
					 NULL};
static const char *texture_types[]  = {"Solid color", "Gradient", "Pixmap",
				       "Textured gradient", NULL};
static const char *gradient_types[] = {"Horizontal", "Vertical",
				       "Diagonal", NULL};
static const char *pixmap_types[]   = {"Scaled", "Tiled", "Centered", NULL};
static const char *wpixmap_types[]   = {"Scaled", "Tiled", "Centered",
					"Maxaspect", NULL};
static const char **workspace_names = NULL;

#define KEY_OF_TEXTURE	(array == PLGetDictionaryEntry (windowmaker, wtexture->key) ? wtexture->key: wtexture->wssback)

/*******************************************************************************

				prototypes
  
*******************************************************************************/

static gint
list_button_press (GtkWidget *widget, GdkEventButton *event);
static void
list_change (GtkWidget *widget, GtkWidget *list);
static void
list_remove (GtkWidget *widget, GtkWidget *list);
static void
list_append (GtkWidget *widget, GtkWidget *list);
static void
list_insert (GtkWidget *widget, GtkWidget *list);
static GtkWidget *
new_list_item (wtexture_t *wtexture, proplist_t *color);
static void
set_pixmap (GtkWidget *widget, gpointer filesel);
static void
set_pixmap_name (GtkWidget *widget, gpointer ptr);
static void
change_path (GtkWidget *widget, gpointer filesel);
static void
file_selection (GtkWidget *widget, wtexture_t *wtexture);
static void
color_selection_ok (GtkWidget *w, GtkColorSelection *colorsel);
static void
update_pixmap (GtkWidget *widget, GtkFileSelection *fs);
static GtkWidget *
generate_pixmap_dialog (wtexture_t *wtexture, bool_t preview_browser);
static GtkWidget *
generate_simple_gradient (wtexture_t *wtexture);
static GtkWidget *
generate_gradient_dialog (wtexture_t *wtexture);
static GtkWidget *
generate_default_color (wtexture_t *wtexture);
static void
show_texture_dialog (wtexture_t *wtexture);
static void
gradient_type_toggle (GtkWidget *widget, gpointer ptr);
static void
pixmap_type_toggle (GtkWidget *widget, gpointer ptr);
static void
texture_type_toggle (GtkWidget *widget, gpointer ptr);
static void
color_selection_dialog (GtkWidget *widget, proplist_t *key);
static void
set_opacity (GtkWidget *widget, gpointer ptr);
static char **
get_workspace_names (void);
static void
workspace_toggle (GtkWidget *widget, gpointer ptr);
static proplist_t *
get_texture_array (wtexture_t *wtexture);

/*******************************************************************************

				public code
  
*******************************************************************************/

void
texture_dialog (GtkWidget *box, proplist_t *key, bool_t preview_browser)
/*
 *  Generate a new texture dialog widget and pack widget into given 'box'.
 *
 *  No return value.
 */ 
{
   wtexture_t *wtexture;
   GtkWidget  *hbox;

   wtexture             = Calloc (1, sizeof (wtexture_t));
   wtexture->wpixmap    = Calloc (PIXMAP_LAST, sizeof (GtkWidget *));
   wtexture->wgradient  = Calloc (GRADIENT_LAST, sizeof (GtkWidget *));
   wtexture->wsgradient = Calloc (SGRADIENT_LAST, sizeof (GtkWidget *));
   wtexture->key	= key;
   wtexture->wssback	= PLMakeString ("WorkspaceSpecificBack");
   wtexture->wsnumber	= 0;
   
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 0);
   gtk_widget_show (hbox);

   /*
    *  Texture type
    */
   if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
   {
      GtkWidget *frame = gtk_frame_new ("Workspace");
      GtkWidget *vbox  = gtk_vbox_new (FALSE, 5);

      workspace_names = (const char **) get_workspace_names ();
	 
      gtk_box_pack_start (GTK_BOX (vbox),
			  generate_option_menu (NULL, tooltips,
						"For each workspace a different "
						"texture can be specified. "
						"The texture of the default "
						"workspace is used if the "
						"texture type is set"
						" to 'default' for a specific "
						"workspace.", workspace_names,
						"Default", workspace_toggle,
						wtexture),
			  FALSE, FALSE, 0);
	 
      gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
      gtk_container_border_width (GTK_CONTAINER (frame), 5);
      gtk_container_border_width (GTK_CONTAINER (vbox), 5);
      gtk_container_add (GTK_CONTAINER (frame), vbox);

      gtk_widget_show_all (frame);
      gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, TRUE, 5);
   }
   {
      GtkWidget *frame = gtk_frame_new ("Texture type");
      GtkWidget *vbox  = gtk_vbox_new (FALSE, 5);
      
      gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
      gtk_container_border_width (GTK_CONTAINER (frame), 5);
      gtk_container_border_width (GTK_CONTAINER (vbox), 5);
      gtk_container_add (GTK_CONTAINER (frame), vbox);

      if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
	 wtexture->wttype = generate_option_menu (NULL, tooltips,
						  "A solid color, a gradient,"
						  " a pixmap, or a gradient "
						  "combined with a pixmap. "
						  "If 'None' is selected, "
						  "then the root background is "
						  "not changed by Window Maker. "
						  "If 'Default' is selected, "
						  "then the texture of the "
						   "default workspace is used",
						  wtexture_types,
						  wtexture_types [0],
						  texture_type_toggle, wtexture);
      else
	 wtexture->wttype = generate_option_menu (NULL, tooltips,
						  "A solid color, a gradient,"
						  " a pixmap, or a gradient "
						  "combined with a pixmap.",
						  texture_types,
						  texture_types [0],
						  texture_type_toggle, wtexture);
      gtk_box_pack_start (GTK_BOX (vbox), wtexture->wttype, FALSE, FALSE, 0);

      if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
      {
	 wtexture->wwttype = generate_option_menu (NULL, tooltips,
						  "A solid color, a gradient,"
						  " a pixmap, or a gradient "
						  "combined with a pixmap. "
						  "If 'None' is selected, "
						  "then the root background is "
						  "not changed by Window Maker.",
						   wwtexture_types,
						   wwtexture_types [0],
						   texture_type_toggle, wtexture);
	 gtk_box_pack_start (GTK_BOX (vbox), wtexture->wwttype, FALSE, FALSE, 0);
      }

      gtk_widget_show_all (frame);
      gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, TRUE, 5);
   }
   
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (box), hbox, FALSE, FALSE, 0);
   gtk_widget_show (hbox);

   gtk_box_pack_start (GTK_BOX (hbox),
		       generate_pixmap_dialog (wtexture, preview_browser),
		       FALSE, TRUE, 5);
   gtk_box_pack_start (GTK_BOX (hbox), generate_simple_gradient (wtexture),
		       FALSE, TRUE, 5);
   gtk_box_pack_start (GTK_BOX (hbox), generate_gradient_dialog (wtexture),
		       FALSE, TRUE, 5);
   gtk_box_pack_start (GTK_BOX (hbox), generate_default_color (wtexture),
		       FALSE, TRUE, 5);

   show_texture_dialog (wtexture);
   changed = NO;
}

GtkWidget *
color_button (wtexture_t *wtexture, proplist_t *key, proplist_t *color, 
	      GtkTooltips *tooltips, const char *info)
/*
 *  Generate color button.
 *  The color value (in proplist format) is assigned to preview.
 */
{
   GtkWidget *frame   = gtk_frame_new (NULL);
   GtkWidget *preview = gtk_preview_new (GTK_PREVIEW_COLOR);
   GtkWidget *button  = gtk_button_new ();

   gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
   gtk_container_border_width (GTK_CONTAINER (frame), 3);
		  
   gtk_preview_size (GTK_PREVIEW (preview), 30, 15);
   gtk_container_add (GTK_CONTAINER (frame), preview);
   gtk_object_set_user_data (GTK_OBJECT (preview), (gpointer) color);
   gtk_object_set_data (GTK_OBJECT (preview), "wtexture", wtexture);
   fill_preview (GTK_PREVIEW (preview), color ? PLGetString (color) : "tan");

   if (info)
      gtk_tooltips_set_tip (tooltips, button, info, NULL); 

   gtk_container_add (GTK_CONTAINER (button), frame);
   gtk_container_border_width (GTK_CONTAINER (button), 0);
   gtk_object_set_user_data (GTK_OBJECT (button), preview);
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (color_selection_dialog),
		       (gpointer) key);

   gtk_widget_show_all (button);

   return button;
}

void
fill_preview (GtkPreview *preview, const char *col_spec)
/*
 *  Fill a 'preview' with color 'col_spec'.
 *
 *  No return value.
 */
{
   unsigned	 y;
   GdkColor	 color;
   unsigned	 width   = GTK_WIDGET (preview)->requisition.width;
   unsigned	 height  = GTK_WIDGET (preview)->requisition.height;
   unsigned char *buffer = Calloc (width * 3, sizeof (unsigned char));
   
   if (!gdk_color_parse (col_spec, &color))
      color.red = color.green = color.blue = 65535;

   for (y = 0; y < height; y++)
   {
      unsigned i;
      unsigned char *ptr = buffer;
			   
      for (i = 0; i < width; i++)
      {
	 *ptr++ = color.red / 256;
	 *ptr++ = color.green / 256;
	 *ptr++ = color.blue / 256;
      }

      gtk_preview_draw_row (preview, buffer, 0, y, width);
   }
   Free (buffer);
   gtk_widget_draw (GTK_WIDGET (preview), NULL);
}

GtkWidget *
show_pixmap (proplist_t *key, const char *name, GtkWidget *pixmap)
{
   const char *keyname = PLGetString (key);

   if (strcaseeq (keyname, "iconback"))
      return make_pixmap (name, 64, 64, YES, pixmap);
   else if (strcaseeq (keyname, "workspaceback") ||
	    strcaseeq (keyname, "workspacespecificback"))
      return make_pixmap (name, 200, 150, YES, pixmap);
   else if (strcaseeq (keyname + 1, "titleback")
	    || strcaseeq (keyname, "menutitleback")
	    || strcaseeq (keyname, "menutextback"))
      return make_pixmap (name, 200, 18, YES, pixmap);
   else
      return make_pixmap (name, 128, 128, NO, pixmap);
}

void
verify_pixmap_name (GtkWidget *widget, wtexture_t *wtexture)
/*
 *  Verify if typed pixmap name is valid.
 *
 *  No return value.
 *
 *  Side effects:
 *	pixmap is changed 
 */
{
   char		*filename = gtk_entry_get_text (GTK_ENTRY (widget));
   proplist_t	*array    = get_texture_array (wtexture);

   show_pixmap (KEY_OF_TEXTURE, filename, wtexture->wpixmap [PIXMAP_PIXMAP]);
#ifdef TILE_OF_DAY
   check_valid_tileofday (wtexture);
#endif /* TILE_OF_DAY */
}

/*******************************************************************************

				private code
  
*******************************************************************************/

static GtkWidget *
generate_pixmap_dialog (wtexture_t *wtexture, bool_t preview_browser)
/*
 *  Generate all widgets for the pixmap dialog.
 *
 *  Return value:
 *	Container widget of pixmap dialog.
 */
{
   GtkWidget	*frame;
   const char	*name = PLGetString (wtexture->key);
   
   wtexture->wpixmap [PIXMAP_FRAME] = frame = gtk_frame_new ("Pixmap");
   gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
   gtk_container_border_width (GTK_CONTAINER (frame), 5);

   if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
      wtexture->wpixmap [PIXMAP_TYPE]
	 = generate_option_menu (NULL,
				 tooltips,
				 "Scaled: The pixmap will be scaled to "
				 "adapt to the destination's size.\n"
				 "Centered: "
				 "The pixmap will be shown centered.\n"
				 "Tiled: The pixmap will be tiled in the "
				 "destination.\n"
				 "Maxaspect: The pixmap will be scaled to "
				 "adapt to the destination's size but"
				 " will maintain aspect ratio.",   
				 wpixmap_types, wpixmap_types [0],
				 pixmap_type_toggle, wtexture);
   else
      wtexture->wpixmap [PIXMAP_TYPE]
	 = generate_option_menu (NULL, tooltips,
				 "Scaled: The pixmap will be scaled to "
				 "adapt to the destination's size.\n"
				 "Centered: "
				 "The pixmap will be shown centered.\n"
				 "Tiled: The pixmap will be tiled in the "
				 "destination.",   
				 pixmap_types, pixmap_types [0],
				 pixmap_type_toggle, wtexture);

   if (strcaseeq (name + 1, "titleback")
       || strcaseeq (name, "menutitleback")
       || strcaseeq (name, "menutextback"))
   {
      GtkWidget *vbox = gtk_vbox_new (FALSE, 5);
      gtk_container_border_width (GTK_CONTAINER (vbox), 5);

      gtk_box_pack_start (GTK_BOX (vbox), wtexture->wpixmap [PIXMAP_TYPE],
			  FALSE, TRUE, 5);

      /*
       *  Frame for pixmap
       */
      {
	 GtkWidget	 *frame	   = gtk_frame_new (NULL);
      
	 wtexture->wpixmap [PIXMAP_PIXMAP]
	    = show_pixmap (wtexture->key, PKGDATADIR "/black.xpm", NULL);
	 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	 gtk_container_add (GTK_CONTAINER (frame),
			    wtexture->wpixmap [PIXMAP_PIXMAP]);

	 gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 0);
      }
      /*
       *  Filename entry
       */
      {
	 GtkWidget	*entry;

	 wtexture->wpixmap [PIXMAP_NAME] = entry = gtk_entry_new ();
	 gtk_widget_set_usize (entry, 200, 0);
	 gtk_signal_connect (GTK_OBJECT (entry), "changed",
			     GTK_SIGNAL_FUNC (set_pixmap_name), wtexture);
	 gtk_signal_connect (GTK_OBJECT (entry), "activate",
			     GTK_SIGNAL_FUNC (verify_pixmap_name),  wtexture);
	 gtk_tooltips_set_tip (tooltips, entry,
			       "Filename of the pixmap. <RETURN> checks if "
			       "the pixmap filename is valid.", NULL);

	 gtk_box_pack_start (GTK_BOX (vbox), entry, FALSE, FALSE, 5);
      }
      /*
       *  Buttons
       */
      {
	 GtkWidget *hbox, *button;

	 hbox = gtk_hbox_new (TRUE, 0);
	 
	 gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

	 button = gtk_button_new_with_label ("Browser");
	 gtk_widget_set_usize (button, 90, 30);
	 gtk_signal_connect (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (file_selection), wtexture);
	 gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
	 gtk_tooltips_set_tip (tooltips, button,
			       "Use file browser to select pixmap.", NULL);

	 if (preview_browser)
	 {
	    button = gtk_button_new_with_label ("Previews");
	    gtk_widget_set_usize (button, 90, 30);
	    gtk_signal_connect (GTK_OBJECT (button), "clicked",
				GTK_SIGNAL_FUNC (image_browser), wtexture);
	    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
	    gtk_tooltips_set_tip (tooltips, button,
				  "Use preview browser to select pixmap.", NULL);
	 }
      
#ifdef TILE_OF_DAY
	 if (preview_browser)
	 {
	    hbox = gtk_hbox_new (TRUE, 5);
	    gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
	 }

	 wtexture->wpixmap [PIXMAP_TILEOFDAY] = button = gtk_button_new ();
	 gtk_container_border_width (GTK_CONTAINER (button), 0);
      
	 {
	    GtkWidget *frame = gtk_frame_new (NULL);
	    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	    gtk_container_add (GTK_CONTAINER (button), frame);
	    gtk_container_add (GTK_CONTAINER (frame),
			       make_pixmap (PKGDATADIR "/t.o.mini.xpm",
					    0, 0, NO, NULL));
	 }
	 gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
	 gtk_signal_connect (GTK_OBJECT (button), "pressed",
			     GTK_SIGNAL_FUNC (tileofday_dialog), wtexture);
	 gtk_tooltips_set_tip (tooltips, button,
			       "Select graphics from http://themes.org.", NULL);
#endif /* TILE_OF_DAY */     
      }

      gtk_container_add (GTK_CONTAINER (frame), vbox);
   }
   else
   {
      GtkWidget	*table = gtk_table_new (3, 2, FALSE);
      
      gtk_table_set_row_spacings (GTK_TABLE (table), 5);
      gtk_table_set_col_spacings (GTK_TABLE (table), 5);
      gtk_container_border_width (GTK_CONTAINER (table), 5);
      gtk_container_add (GTK_CONTAINER (frame), table);
      
      gtk_table_attach (GTK_TABLE (table), wtexture->wpixmap [PIXMAP_TYPE],
			0, 2, 0, 1, GTK_FILL | GTK_SHRINK | GTK_EXPAND,
			GTK_FILL | GTK_SHRINK | GTK_EXPAND, 0, 0);

      /*
       *  Scrolled window for pixmap
       */
      if (strcaseeq (name, "iconback"))
      {
	 GtkWidget	 *frame	   = gtk_frame_new (NULL);
	 GtkWidget	 *vbox     = gtk_vbox_new (FALSE, 0);
      
	 wtexture->wpixmap [PIXMAP_PIXMAP]
	    = show_pixmap (wtexture->key, PKGDATADIR "/black.xpm", NULL);
	 gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	 gtk_container_border_width (GTK_CONTAINER (frame), 0);
	 gtk_container_add (GTK_CONTAINER (frame),
			    wtexture->wpixmap [PIXMAP_PIXMAP]);
	 gtk_container_add (GTK_CONTAINER (vbox),
			    frame);
	 gtk_table_attach (GTK_TABLE (table), vbox, 0, 1, 1, 2,
			   GTK_SHRINK, GTK_SHRINK, 0, 0);
      }
      else
      {
	 GtkWidget *scrolled = gtk_scrolled_window_new (NULL, NULL);
      
	 gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
					 GTK_POLICY_ALWAYS, GTK_POLICY_ALWAYS);
	 gtk_widget_set_usize (scrolled, 100, 100);
	 gtk_container_border_width (GTK_CONTAINER (scrolled), 5);
	 gtk_table_attach (GTK_TABLE (table), scrolled, 0, 1, 1, 2,
			   GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);

	 wtexture->wpixmap [PIXMAP_PIXMAP]
	    = show_pixmap (wtexture->key, PKGDATADIR "/black.xpm", NULL);
	 gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled),
						wtexture->wpixmap [PIXMAP_PIXMAP]); 
      }
      /*
       *  Filename entry
       */
      {
	 GtkWidget	*entry;

	 wtexture->wpixmap [PIXMAP_NAME] = entry = gtk_entry_new ();
	 gtk_widget_set_usize (entry, 120, 0);
	 gtk_signal_connect (GTK_OBJECT (entry), "changed",
			     GTK_SIGNAL_FUNC (set_pixmap_name), wtexture);
	 gtk_signal_connect (GTK_OBJECT (entry), "activate",
			     GTK_SIGNAL_FUNC (verify_pixmap_name),  wtexture);
	 gtk_table_attach (GTK_TABLE (table), entry, 0, 2, 2, 3,
			   GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);
	 gtk_tooltips_set_tip (tooltips, entry,
			       "Filename of the pixmap. <RETURN> checks if "
			       "the pixmap filename is valid.", NULL);
      }
   
      /*
       *  Browse button
       */
      {
	 GtkWidget	*vbox, *button;

	 vbox = gtk_vbox_new (FALSE, 5);
	 gtk_table_attach (GTK_TABLE (table), vbox, 1, 2, 1, 2,
			   GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);

	 button = gtk_button_new_with_label ("Browser");
	 gtk_widget_set_usize (button, 90, 30);
	 gtk_signal_connect (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (file_selection), wtexture);
	 gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, TRUE, 0);
	 gtk_tooltips_set_tip (tooltips, button,
			       "Use file browser to select pixmap.", NULL);

	 if (preview_browser)
	 {
	    button = gtk_button_new_with_label ("Previews");
	    gtk_widget_set_usize (button, 90, 30);
	    gtk_signal_connect (GTK_OBJECT (button), "clicked",
				GTK_SIGNAL_FUNC (image_browser), wtexture);
	    gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, TRUE, 0);
	    gtk_tooltips_set_tip (tooltips, button,
				  "Use preview browser to select pixmap.", NULL);
	 }
      
#ifdef TILE_OF_DAY
	 wtexture->wpixmap [PIXMAP_TILEOFDAY] = button = gtk_button_new ();
	 gtk_container_border_width (GTK_CONTAINER (button), 0);
      
	 {
	    GtkWidget *frame = gtk_frame_new (NULL);
	    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	    gtk_container_add (GTK_CONTAINER (button), frame);
	    gtk_container_add (GTK_CONTAINER (frame),
			       make_pixmap (PKGDATADIR "/t.o.mini.xpm",
					    0, 0, NO, NULL));
	 }
	 {
	    GtkWidget *hhbox = gtk_hbox_new (FALSE, 0);
	    gtk_box_pack_start (GTK_BOX (vbox), hhbox, FALSE, FALSE, 0);
	    gtk_box_pack_start (GTK_BOX (hhbox), button, TRUE, FALSE, 0);
	 }
	 
	 gtk_signal_connect (GTK_OBJECT (button), "pressed",
			     GTK_SIGNAL_FUNC (tileofday_dialog), wtexture);
	 gtk_tooltips_set_tip (tooltips, button,
			       "Select graphics from http://themes.org.", NULL);
#endif /* TILE_OF_DAY */     
      }
   }
   
   gtk_widget_show_all (frame);

   return frame;
}

static GtkWidget *
generate_simple_gradient (wtexture_t *wtexture)
/*
 *  Generate all widgets for the simple gradient dialog.
 *
 *  Return value:
 *	Container widget of simple gradient dialog.
 */
{
   GtkWidget	*frame = gtk_frame_new ("Gradient");
   GtkWidget	*vbox  = gtk_vbox_new (FALSE, 0);
   GtkWidget	*hbox;

   wtexture->wsgradient [SGRADIENT_FRAME] = frame;
   gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
   gtk_container_border_width (GTK_CONTAINER (frame), 5);
   gtk_container_add (GTK_CONTAINER (frame), vbox);
   gtk_container_border_width (GTK_CONTAINER (vbox), 5);

   wtexture->wsgradient [SGRADIENT_TYPE]
      = generate_option_menu (NULL, tooltips,
			      "Horizontal: "
			      "A gradient going from left "
			      "to right, from the first to the last color.\n"
			      "Vertical: A gradient going from top "
			      "to bottom, from the first to the last color.\n"
			      "Diagonal: A gradient going from the "
			      "top-left corner to the bottom-right corner, ",
			      gradient_types, gradient_types [0],
			      gradient_type_toggle, wtexture);

   gtk_box_pack_start (GTK_BOX (vbox), wtexture->wsgradient [SGRADIENT_TYPE],
		       FALSE, TRUE, 5);
   
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
   gtk_widget_show (hbox);

   gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new ("Color 1: "),
		       FALSE, TRUE, 0);
   wtexture->wsgradient [SGRADIENT_COLOR1]
      = color_button (wtexture, wtexture->key, NULL, NULL, NULL);
   gtk_box_pack_start (GTK_BOX (hbox), wtexture->wsgradient [SGRADIENT_COLOR1],
		       FALSE, TRUE, 0);

   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
   gtk_widget_show (hbox);

   gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new ("Color 2: "),
		       FALSE, TRUE, 0);
   wtexture->wsgradient [SGRADIENT_COLOR2]
      = color_button (wtexture, wtexture->key, NULL, NULL, NULL);
   gtk_box_pack_start (GTK_BOX (hbox), wtexture->wsgradient [SGRADIENT_COLOR2],
		       FALSE, TRUE, 0);
   
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 5);
   gtk_widget_show (hbox);

   gtk_box_pack_start (GTK_BOX (hbox), gtk_label_new ("Opacity:"),
		       FALSE, TRUE, 0);
   {
      GtkObject	*adj = gtk_adjustment_new (0, 0, 255, 1.0, 10.0, 0.0);
      GtkWidget	*spinner;
      
      spinner = wtexture->wsgradient [SGRADIENT_SPINNER]
	      = gtk_spin_button_new (GTK_ADJUSTMENT (adj), 0, 0);
      gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spinner), TRUE);
      gtk_spin_button_set_wrap (GTK_SPIN_BUTTON (spinner), TRUE);
      gtk_widget_show_all (spinner);
      gtk_object_set_user_data (GTK_OBJECT (adj), (gpointer) spinner);
      gtk_box_pack_start (GTK_BOX (hbox), spinner, TRUE, TRUE, 0);
      gtk_tooltips_set_tip (tooltips, spinner, "Opacity of the pixmap. If set "
			    "to zero, only the pixmap is visible "
			    "(and vice versa).", NULL);  

      gtk_signal_connect (GTK_OBJECT (&GTK_SPIN_BUTTON (spinner)->entry),
			  "changed", GTK_SIGNAL_FUNC (set_opacity), wtexture);
   }
   
   gtk_widget_show_all (frame);

   return frame;
}

static GtkWidget *
generate_default_color (wtexture_t *wtexture)
/*
 *  Generate default color dialog.
 *
 *  Return value:
 *	Container widget of default color dialog.
 */
{
   GtkWidget	*frame = gtk_frame_new ("Default color");
   GtkWidget	*box   = gtk_hbox_new (FALSE, 0);
   GtkWidget	*vbox  = gtk_vbox_new (FALSE, 0);

   wtexture->wcolorbox = frame;
   gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
   gtk_container_border_width (GTK_CONTAINER (frame), 5);
   gtk_container_add (GTK_CONTAINER (frame), box);
   gtk_container_border_width (GTK_CONTAINER (box), 5);

   wtexture->wcolor = color_button (wtexture, wtexture->key, NULL,
				    tooltips, "This color is used as solid color,"
				    " if a gradient can't be drawn, or during "
				    "the pixmap scaling/tiling process.");  
   gtk_container_border_width (GTK_CONTAINER (box), 5);
   gtk_box_pack_start (GTK_BOX (box), wtexture->wcolor, FALSE, TRUE, 15);

   gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, TRUE, 0);
   gtk_widget_show_all (vbox);

   return vbox;
}

static GtkWidget *
generate_gradient_dialog (wtexture_t *wtexture)
/*
 *  Generate all widgets for the gradient dialog.
 *
 *  Return value:
 *	Container widget of gradient dialog.
 */
{
   GtkWidget  *hbox;
   GtkWidget  *frame;
   GtkWidget  *scrolled;
   GtkWidget  *table;
   GtkWidget  *list;
   
   frame = wtexture->wgradient [GRADIENT_FRAME]
	 = gtk_frame_new ("Gradient");
   gtk_frame_set_label_align (GTK_FRAME (frame), 0.5, 0.5);
   gtk_container_border_width (GTK_CONTAINER (frame), 5);
   
   wtexture->wgradient [GRADIENT_TYPE]
      = generate_option_menu (NULL, tooltips,
			      "Horizontal: "
			      "A gradient going from left "
			      "to right, from the first to the last color.\n"
			      "Vertical: A gradient going from top "
			      "to bottom, from the first to the last color.\n"
			      "Diagonal: A gradient going from the "
			      "top-left corner to the bottom-right corner, ",
			      gradient_types, gradient_types [0],
			      gradient_type_toggle, wtexture);

   table = gtk_table_new (2, 1, FALSE);
   gtk_container_border_width (GTK_CONTAINER (table), 5);
   gtk_table_set_row_spacings (GTK_TABLE (table), 5);
   gtk_table_set_col_spacings (GTK_TABLE (table), 5);
   gtk_container_add (GTK_CONTAINER (frame), table);
   
   gtk_table_attach (GTK_TABLE (table), wtexture->wgradient [GRADIENT_TYPE],
		     0, 2, 0, 1,
		     GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);
   
   hbox = gtk_hbox_new (FALSE, 5);
   gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, 1, 2,
		     GTK_FILL | GTK_SHRINK, GTK_FILL | GTK_SHRINK, 0, 0);

   /*
    *  Scrolled window for list
    */
   scrolled = gtk_scrolled_window_new (NULL, NULL);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
   gtk_widget_set_usize (scrolled, 73, 100);
   gtk_box_pack_start (GTK_BOX (hbox), scrolled, FALSE, FALSE, 0);
   
   /*
    *  Color list
    */
   wtexture->wgradient [GRADIENT_LIST] = list = gtk_list_new ();
   gtk_list_set_selection_mode (GTK_LIST (list), GTK_SELECTION_BROWSE);
   gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scrolled), list);
   gtk_object_set_user_data (GTK_OBJECT (list), wtexture->key);
         
   gtk_signal_connect (GTK_OBJECT (list), "button_press_event",
		       GTK_SIGNAL_FUNC (list_button_press), NULL);


   /*
    *  List manipulating buttons
    */
   {
      GtkWidget  *bbox; 
      const char *text [] = {"Remove", "Insert", "Append", "Change"};
      const char *info [] = {"Remove selected color from gradient color list.",
			     "Insert a new color before current color.",
			     "Append a new color after the last color.",
			     "Change the current color."};
      void (*fct []) (GtkWidget *, GtkWidget *list) = {list_remove, list_insert,
						       list_append, list_change};
      unsigned	i;
      
      bbox = gtk_vbutton_box_new ();
      gtk_box_pack_start (GTK_BOX (hbox), bbox, FALSE, TRUE, 0);

      for (i = 0; i < 4; i++)
      {
	 GtkWidget *button = gtk_button_new_with_label (text [i]);
	 if (strcaseeq ("remove", text [i]))
	    wtexture->wgradient [GRADIENT_REMOVE] = button;
	 gtk_box_pack_start (GTK_BOX (bbox), button, FALSE, TRUE, 0);
	 gtk_object_set_user_data (GTK_OBJECT (button), wtexture);
	 gtk_signal_connect (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (fct [i]), list);
	 gtk_tooltips_set_tip (tooltips, button, info [i], NULL);  
      }
   }

   gtk_widget_show_all (frame);

   return frame;
}   

static void
show_texture_dialog (wtexture_t *wtexture)
/*
 *  Generate set of visible/hidden widgets for the current texture attribute
 */
{
   unsigned	n;
   char		*type   = NULL;
   texture_e	texture = TEXTURE_LAST;
   proplist_t	*array	= get_texture_array (wtexture);

   if (!PLIsArray (array))		/* uups */
   {
      if (strcaseeq ("workspaceback", PLGetString (wtexture->key))
	  && wtexture->wsnumber > 0)
	 PLInsertDictionaryEntry (windowmaker, wtexture->wssback,
				  array = PLMakeArrayFromElements (NULL));
      else
	 PLInsertDictionaryEntry (windowmaker, wtexture->key,
				  array = PLMakeArrayFromElements (NULL));
   }
   /*
    *  Check which type of texture is used
    */
   if ((n = PLGetNumberOfElements (array)) != 0)
   {
      type = PLGetString (PLGetArrayElement (array, 0));

      if (strcaseeq (type + 1, "gradient")) /* convert to m[hvd]gradient */
      {
	 if (n == 3)
	 {
	    GdkColor	color1, color2;
	    char	*c1name = PLGetString (PLGetArrayElement (array, 1));
	    char	*c2name = PLGetString (PLGetArrayElement (array, 2));
	    /*
	     *  Compute default color = (color1 + color2) / 2
	     */
	    if (!gdk_color_parse (c1name, &color1))
	       color1.red = color1.green = color1.blue = 65535; /* white */
	    if (!gdk_color_parse (c2name, &color2))
	       color2.red = color2.green = color2.blue = 65535; /* white */

	    color1.red   = (color1.red + color2.red) >> 1;
	    color1.green = (color1.green + color2.green) >> 1;
	    color1.blue  = (color1.blue + color2.blue) >> 1;

	    {
	       char colstr [MAXSTRLEN];

	       sprintf (colstr, "#%02x%02x%02x", /* new color */
			(unsigned) (color1.red   / 256 + 0.5),
			(unsigned) (color1.green / 256 + 0.5),
			(unsigned) (color1.blue  / 256 + 0.5));
	       PLInsertArrayElement (array, PLMakeString (colstr), 1);
	    }

	    /*
	     *  Replace gradient by mgradient
	     */
	    {
	       char	*tmp;
	       
	       tmp = Calloc (strlen (type) + 2, sizeof (char));
	       strcpy (tmp, "m");
	       strcat (tmp, type);
	       PLRemoveArrayElement (array, 0);
	       PLInsertArrayElement (array, PLMakeString (tmp), 0);
	       Free (tmp);
	       type = PLGetString (PLGetArrayElement (array, 0));
	    }
	    n++;
	 }
      }
      if (strcaseeq (type, "solid"))
	 texture = n == 2 ? TEXTURE_SOLID : TEXTURE_LAST;
      else if (strcaseeq (type, "none"))
	 texture = TEXTURE_NONE;
      else if (strcaseeq (type + 2, "gradient"))
      {
	 if (*type != 'm' && *type != 't')
	    texture = TEXTURE_LAST;
	 else
	 {
	    if (*type == 't')
	       texture = n == 5 ? TEXTURE_TGRADIENT : TEXTURE_LAST;
	    else 
	       texture = n > 3 ? TEXTURE_GRADIENT : TEXTURE_LAST;
	    if (*(type + 1) != 'v' && *(type + 1) != 'h' && *(type + 1) != 'd')
	       texture = TEXTURE_LAST;
	 }
      }
      else if (strcaseeq (type + 1, "pixmap"))
      {
	 if ((*type != 's' && *type != 't' && *type != 'c' && *type != 'm')
	     || (*type == 'm'
		 && !strcaseeq ("workspaceback", PLGetString (wtexture->key))))
	    texture = TEXTURE_LAST;
	 else
	    texture = n == 3 ? TEXTURE_PIXMAP : TEXTURE_LAST;
      }
   }
   else if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
      texture = TEXTURE_DEFAULT;

   /*
    *  Convert unknown texture types to (solid, black)
    */
   if (texture == TEXTURE_LAST)
   {
      unsigned i;
      
      warning ("Texture `%s' not yet supported or wrong number of arguments.\n"
	       "Converting to (solid, black) ...", type ? type : "()");
      for (i = n; i; i--)
	 PLRemoveArrayElement (array, 0);
      PLInsertArrayElement (array, PLMakeString ("solid"), 0);
      PLInsertArrayElement (array, PLMakeString ("black"), 1);
      texture = TEXTURE_SOLID;
   }
   /*
    *  Set texture type option menu to given value
    */
   if (strcaseeq ("workspaceback", PLGetString (wtexture->key)))
   {
      if (wtexture->wsnumber == 0)
      {
	 gtk_widget_show (wtexture->wwttype);
	 gtk_widget_hide (wtexture->wttype);
	 set_option_menu_default (GTK_OPTION_MENU (wtexture->wwttype),
				  wtexture_types, wtexture_types [texture]);
      }
      else
      {
	 gtk_widget_show (wtexture->wttype);
	 gtk_widget_hide (wtexture->wwttype);
	 set_option_menu_default (GTK_OPTION_MENU (wtexture->wttype),
				  wtexture_types, wtexture_types [texture]);
      }
   }
   else
   {
      gtk_widget_show (wtexture->wttype);
      set_option_menu_default (GTK_OPTION_MENU (wtexture->wttype),
			       wwtexture_types, wwtexture_types [texture]);
   }
   if (texture == TEXTURE_SOLID)
   {
      gtk_widget_hide (wtexture->wpixmap [PIXMAP_FRAME]);
      gtk_widget_hide (wtexture->wgradient [GRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wsgradient [SGRADIENT_FRAME]);
      gtk_widget_show (wtexture->wcolorbox);

      /*
       *  Set default color
       */
      {
	 GtkWidget *preview;

	 preview = gtk_object_get_user_data (GTK_OBJECT (wtexture->wcolor));
	 fill_preview (GTK_PREVIEW (preview),
		       PLGetString (PLGetArrayElement (array, 1)));
	 gtk_object_set_user_data (GTK_OBJECT (preview),
				   PLGetArrayElement (array, 1));
      }
   }
   else if (texture == TEXTURE_NONE || texture == TEXTURE_DEFAULT)
   {
      gtk_widget_hide (wtexture->wpixmap [PIXMAP_FRAME]);
      gtk_widget_hide (wtexture->wgradient [GRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wsgradient [SGRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wcolorbox);
   }
   else if (texture == TEXTURE_PIXMAP)
   {
      gtk_widget_show (wtexture->wpixmap [PIXMAP_FRAME]);
      gtk_widget_show (wtexture->wpixmap [PIXMAP_TYPE]);
      gtk_widget_hide (wtexture->wgradient [GRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wsgradient [SGRADIENT_FRAME]);
      gtk_widget_show (wtexture->wcolorbox);
#ifdef TILE_OF_DAY
      gtk_widget_show (wtexture->wpixmap [PIXMAP_TILEOFDAY]);
#endif /* TILE_OF_DAY */
      /*
       *  Set default color
       */
      {
	 GtkWidget *preview;

	 preview = gtk_object_get_user_data (GTK_OBJECT (wtexture->wcolor));
	 fill_preview (GTK_PREVIEW (preview),
		       PLGetString (PLGetArrayElement (array, 2)));
	 gtk_object_set_user_data (GTK_OBJECT (preview),
				   PLGetArrayElement (array, 2));
      }
      /*
       *  Set pixmap and filename
       */
      gtk_entry_set_text (GTK_ENTRY (wtexture->wpixmap [PIXMAP_NAME]),
			  PLGetString (PLGetArrayElement (array, 1)));
      show_pixmap (KEY_OF_TEXTURE, PLGetString (PLGetArrayElement (array, 1)),
		   wtexture->wpixmap [PIXMAP_PIXMAP]);
      /*
       *  scaled, tiled, centered, maxaspect pixmap
       */
      switch (*type)
      {
	 case 't':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wpixmap [PIXMAP_TYPE]), wpixmap_types, wpixmap_types [PIXMAP_TILED - PIXMAP_SCALED]);
	    break;
	 case 'c':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wpixmap [PIXMAP_TYPE]), wpixmap_types, wpixmap_types [PIXMAP_CENTERED - PIXMAP_SCALED]);
	    break;
	 case 'm':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wpixmap [PIXMAP_TYPE]), wpixmap_types, wpixmap_types [PIXMAP_MAXASPECT - PIXMAP_SCALED]);
	    break;
	 default:
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wpixmap [PIXMAP_TYPE]), wpixmap_types, wpixmap_types [PIXMAP_SCALED - PIXMAP_SCALED]);
	    break;
      }
   }
   else if (texture == TEXTURE_GRADIENT)
   {
      gtk_widget_hide (wtexture->wpixmap [PIXMAP_FRAME]);
      gtk_widget_show (wtexture->wgradient [GRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wsgradient [SGRADIENT_FRAME]);
      gtk_widget_show (wtexture->wcolorbox);

      /*
       *  Set default color
       */
      {
	 GtkWidget *preview;

	 preview = gtk_object_get_user_data (GTK_OBJECT (wtexture->wcolor));
	 fill_preview (GTK_PREVIEW (preview),
		       PLGetString (PLGetArrayElement (array, 1)));
	 gtk_object_set_user_data (GTK_OBJECT (preview),
				   PLGetArrayElement (array, 1));
      }
      /*
       *  horizontal, vertical, or diagonal gradient
       */
      switch (*(type + 1))
      {
	 case 'h':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wgradient [GRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_HOR - GRADIENT_HOR]);
	    break;
	 case 'v':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wgradient [GRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_VER - GRADIENT_HOR]);
	    break;
	 default:
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wgradient [GRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_DIAG - GRADIENT_HOR]);
	    break;
      }
      /*
       *  gradient colors
       */
      {
	 unsigned  i;
	 GtkWidget *list = wtexture->wgradient [GRADIENT_LIST];
	 
	 while (g_list_length (GTK_LIST (list)->children))
	    list_remove (wtexture->wgradient [GRADIENT_REMOVE], list);
	 for (i = 2; i < n; i++)
	    gtk_container_add (GTK_CONTAINER (list),
			       new_list_item (wtexture,
					      PLGetArrayElement (array, i)));
	 gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], n > 4);
	 gtk_list_select_item (GTK_LIST (list), 0);
      }
   }
   else if (texture == TEXTURE_TGRADIENT)
   {
      gtk_widget_show (wtexture->wpixmap [PIXMAP_FRAME]);
      gtk_widget_hide (wtexture->wgradient [GRADIENT_FRAME]);
      gtk_widget_show (wtexture->wsgradient [SGRADIENT_FRAME]);
      gtk_widget_hide (wtexture->wcolorbox);
      gtk_widget_hide (wtexture->wpixmap [PIXMAP_TYPE]);

      /*
       *  horizontal, vertical, or diagonal gradient
       */
      switch (*(type + 1))
      {
	 case 'h':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wsgradient [SGRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_HOR - GRADIENT_HOR]);
	    break;
	 case 'v':
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wsgradient [SGRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_VER - GRADIENT_HOR]);
	    break;
	 default:
	    set_option_menu_default (GTK_OPTION_MENU (wtexture->wsgradient [SGRADIENT_TYPE]), gradient_types, gradient_types [GRADIENT_DIAG - GRADIENT_HOR]);
	    break;
      }
      
      /*
       *  Set pixmap and filename
       */
      gtk_entry_set_text (GTK_ENTRY (wtexture->wpixmap [PIXMAP_NAME]),
			  PLGetString (PLGetArrayElement (array, 1)));
      show_pixmap (KEY_OF_TEXTURE, PLGetString (PLGetArrayElement (array, 1)),
		   wtexture->wpixmap [PIXMAP_PIXMAP]);
#ifdef TILE_OF_DAY
      gtk_widget_hide (wtexture->wpixmap [PIXMAP_TILEOFDAY]);
#endif /* TILE_OF_DAY */
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (wtexture->wsgradient [SGRADIENT_SPINNER]), strtod (PLGetString (PLGetArrayElement (array, 2)), NULL));
      {
	 GtkWidget *preview;

	 preview = gtk_object_get_user_data (GTK_OBJECT (wtexture->wsgradient [SGRADIENT_COLOR1]));
	 fill_preview (GTK_PREVIEW (preview),
		       PLGetString (PLGetArrayElement (array, 3)));
	 gtk_object_set_user_data (GTK_OBJECT (preview),
				   PLGetArrayElement (array, 3));
	 preview = gtk_object_get_user_data (GTK_OBJECT (wtexture->wsgradient [SGRADIENT_COLOR2]));
	 fill_preview (GTK_PREVIEW (preview),
		       PLGetString (PLGetArrayElement (array, 4)));
	 gtk_object_set_user_data (GTK_OBJECT (preview),
				   PLGetArrayElement (array, 4));
      }
   }
#ifdef TILE_OF_DAY
   check_valid_tileofday (wtexture);
#endif /* TILE_OF_DAY */

}

static void
color_selection_dialog (GtkWidget *widget, proplist_t *key)
/*
 *  Generate a color selection popup window.
 *  associated user data of 'widget' must be a pointer to corresponding preview.
 *  'key' is the associated wmaker attribute name
 *
 *  No return value.
 */
{
   static GtkWidget *window = NULL;

   if (!window)
   {
      GdkColor	 color;
      GtkWidget  *preview = gtk_object_get_user_data (GTK_OBJECT (widget));
      proplist_t *value   = gtk_object_get_user_data (GTK_OBJECT (preview));
      gdouble	 rgb [4];
      GtkColorSelection		*cs;
      GtkColorSelectionDialog	*dialog;
      
      gtk_preview_set_install_cmap (TRUE);
      gtk_widget_push_visual (gtk_preview_get_visual ());
      gtk_widget_push_colormap (gtk_preview_get_cmap ());

      window = gtk_color_selection_dialog_new ("Color selection");
      dialog = GTK_COLOR_SELECTION_DIALOG (window);
      cs     = GTK_COLOR_SELECTION (dialog->colorsel);

      gtk_widget_hide (dialog->help_button);
      gtk_object_set_data (GTK_OBJECT (cs), "key", key);
      gtk_object_set_user_data (GTK_OBJECT (cs), preview);
      
      if (!gdk_color_parse (PLGetString (value), &color))
	 color.red = color.green = color.blue = 65530;

      rgb [0] = color.red   / 65535.0;
      rgb [1] = color.green / 65535.0;
      rgb [2] = color.blue  / 65535.0;

      gtk_color_selection_set_color (cs, rgb);
      gtk_color_selection_get_color (cs, rgb);
      gtk_color_selection_set_color (cs, rgb);
      gtk_color_selection_set_opacity (cs, FALSE);
      gtk_color_selection_set_update_policy (cs, GTK_UPDATE_CONTINUOUS);
      gtk_window_position (GTK_WINDOW (window), GTK_WIN_POS_MOUSE);

      gtk_signal_connect (GTK_OBJECT (window), "destroy",
                          GTK_SIGNAL_FUNC (gtk_widget_destroyed),
                          &window);

      gtk_signal_connect (GTK_OBJECT (dialog->ok_button),
			  "clicked",
			  GTK_SIGNAL_FUNC (color_selection_ok), cs);
      gtk_signal_connect_object (GTK_OBJECT (dialog->ok_button),
				 "clicked",
				 GTK_SIGNAL_FUNC (gtk_widget_destroy),
				 GTK_OBJECT (window));

      gtk_signal_connect_object (GTK_OBJECT (dialog->cancel_button),
				 "clicked",
				 GTK_SIGNAL_FUNC (gtk_widget_destroy),
				 GTK_OBJECT (window));
      gtk_widget_pop_colormap ();
      gtk_widget_pop_visual ();
   }

   if (!GTK_WIDGET_VISIBLE (window))
      gtk_widget_show (window);
   else
      gtk_widget_destroy (window);
}

static proplist_t *
get_texture_array (wtexture_t *wtexture)
{
   proplist_t *array;

   if (strcaseeq ("workspaceback", PLGetString (wtexture->key))
       && wtexture->wsnumber > 0)
   {
      array = PLGetDictionaryEntry (windowmaker, wtexture->wssback);
      if (!array || !PLIsArray (array))
      {
	 if (array)
	    PLRelease (array);
	 array = PLMakeArrayFromElements (NULL);
	 PLInsertDictionaryEntry (windowmaker, wtexture->wssback, array);
      }
      while (PLGetNumberOfElements (array) < wtexture->wsnumber)
      {
	 PLAppendArrayElement (array, PLMakeArrayFromElements (NULL));
	 changed = YES;
	 PLInsertDictionaryEntry (plist_changed, wtexture->wssback, pl_yes);
      }
      array = PLGetArrayElement (array, wtexture->wsnumber - 1);
   }
   else
      array = PLGetDictionaryEntry (windowmaker, wtexture->key);

   return array;
}

static void
gradient_type_toggle (GtkWidget *widget, gpointer ptr)
/*
 *  Toggle gradient type (horizontal, vertical, diagonal).
 *
 *  Side effects:
 *	Window Maker attribute is updated
 */
{
   const char *text     = gtk_object_get_user_data (GTK_OBJECT (widget));
   wtexture_t *wtexture = (wtexture_t *) ptr;
   proplist_t *array    = get_texture_array (wtexture);
   const char *type     = PLGetString (PLGetArrayElement (array, 0));
   
   if (strcaseeq (type + 2, "gradient")
       && tolower (*(type + 1)) != tolower (*text))
   {
      char *tmp = g_strdup (type);

      *(tmp + 1) = tolower (*text);
      PLRemoveArrayElement (array, 0);
      PLInsertArrayElement (array, PLMakeString (tmp), 0);
      Free (tmp);
      PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
      changed = YES;			
   }
}

static void
pixmap_type_toggle (GtkWidget *widget, gpointer ptr)
/*
 *  Toggle pixmap type (scaled, tiled, centered, maxaspect).
 *
 *  Side effects:
 *	Window Maker attribute is updated
 */
{
   const char *text     = gtk_object_get_user_data (GTK_OBJECT (widget));
   wtexture_t *wtexture = (wtexture_t *) ptr;
   proplist_t *array    = get_texture_array (wtexture);
   const char *type     = PLGetString (PLGetArrayElement (array, 0));
   
   if (strcaseeq (type + 1, "pixmap")
       && tolower (*type) != tolower (*text))
   {
      char *tmp = g_strdup (type);

      *tmp = tolower (*text);
      PLRemoveArrayElement (array, 0);
      PLInsertArrayElement (array, PLMakeString (tmp), 0);
      Free (tmp);
      PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
      changed = YES;			
   }
}

static void
texture_type_toggle (GtkWidget *widget, gpointer ptr)
/*
 *  Toggle texture type (solid, pixmap, gradient, textured gradient).
 *
 *  Side effects:
 *	Window Maker attribute is updated
 */
{
   const char *text     = gtk_object_get_user_data (GTK_OBJECT (widget));
   wtexture_t *wtexture = (wtexture_t *) ptr;
   proplist_t *array    = get_texture_array (wtexture);
   const char *type     = PLGetNumberOfElements (array) > 0
			  ? PLGetString (PLGetArrayElement (array, 0)) : NULL;
   
   if ((type == NULL && tolower (*text) != 'd')
       || (type
	   && (!((strcaseeq (type + 1, "pixmap") && tolower (*text) == 'p')
		 || (strcaseeq (type, "solid") && tolower (*text) == 's')
		 || (strcaseeq (type + 2, "gradient") && tolower (*type) == 'm'
		     && tolower (*text) == 'g')
		 || (strcaseeq (type + 2, "gradient") && tolower (*type) == 't'
		     && tolower (*text) == 't')
		 || (strcaseeq (type, "none") && tolower (*text) == 'n')))))
   {
      unsigned n;

      for (n = PLGetNumberOfElements (array); n; n--)
	 PLRemoveArrayElement (array, 0); /* clear array */

      if (tolower (*text) == 's')	/* solid */
      {
	 PLInsertArrayElement (array, PLMakeString ("solid"), 0);
	 PLInsertArrayElement (array, PLMakeString ("white"), 1);
      }
      else if (tolower (*text) == 'n') /* none */
      {
	 PLInsertArrayElement (array, PLMakeString ("none"), 0);
      }
      else if (tolower (*text) == 'p') /* pixmap */
      {
	 PLInsertArrayElement (array, PLMakeString ("spixmap"), 0);
	 PLInsertArrayElement (array, PLMakeString (PKGDATADIR "/black.xpm"),
			       1);
	 PLInsertArrayElement (array, PLMakeString ("white"), 2);
      }
      else if (tolower (*text) == 'g') /* gradient */
      {
	 PLInsertArrayElement (array, PLMakeString ("mhgradient"), 0);
	 PLInsertArrayElement (array, PLMakeString ("blue"), 1);
	 PLInsertArrayElement (array, PLMakeString ("yellow"), 2);
	 PLInsertArrayElement (array, PLMakeString ("blue"), 3);
      }
      else if (tolower (*text) == 't') /* textured gradient */
      {
	 PLInsertArrayElement (array, PLMakeString ("thgradient"), 0);
	 PLInsertArrayElement (array, PLMakeString (PKGDATADIR "/black.xpm"),
			       1);
	 PLInsertArrayElement (array, PLMakeString ("128"), 2);
	 PLInsertArrayElement (array, PLMakeString ("yellow"), 3);
	 PLInsertArrayElement (array, PLMakeString ("blue"), 4);
      } /* nothing to do for () */
	 
      show_texture_dialog (wtexture);
      PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
      changed = YES;			
   }
}

static void
workspace_toggle (GtkWidget *widget, gpointer ptr)
/*
 *  Toggle workspace name.
 */
{
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   char		*text     = gtk_object_get_user_data (GTK_OBJECT (widget));
   
   for (wtexture->wsnumber = 0;
	workspace_names [wtexture->wsnumber];
	wtexture->wsnumber++)
      if (streq (workspace_names [wtexture->wsnumber], text))
	 break;
   show_texture_dialog (wtexture);
}

static void
set_opacity (GtkWidget *widget, gpointer ptr)
/*
 *  Set opacity.
 *
 *  Side effects:
 *	Window Maker attribute is updated
 */
{
   char		*value;
   wtexture_t	*wtexture = (wtexture_t *) ptr;
   proplist_t	*array 	  = get_texture_array (wtexture);

   value = gtk_entry_get_text (GTK_ENTRY (widget));
   PLRemoveArrayElement (array, 2);
   PLInsertArrayElement (array, PLMakeString (value), 2);
   PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
   changed = YES;			
}

static void
color_selection_ok (GtkWidget *w, GtkColorSelection *colorsel)
/*
 *  Replace old color with new one.
 *
 *  No return value.
 *
 *  Side effects:
 *	value of wmaker attribute w->key is modified.
 *	corresponding preview color update
 */
{
   gdouble	color [4];
   proplist_t	*key     = gtk_object_get_data (GTK_OBJECT (colorsel), "key");
   GtkPreview	*preview = gtk_object_get_user_data (GTK_OBJECT (colorsel));
   proplist_t	*value   = gtk_object_get_user_data (GTK_OBJECT (preview));
   
   gtk_color_selection_get_color (colorsel, color);
   gtk_color_selection_set_color (colorsel, color);
 
   {
      char tmp [MAXSTRLEN];

      sprintf (tmp, "#%02x%02x%02x",	/* new color */
	       (unsigned) (color [0] * 255 + 0.5),
	       (unsigned) (color [1] * 255 + 0.5),
	       (unsigned) (color [2] * 255 + 0.5));

      {
	 wtexture_t *wtexture = gtk_object_get_data (GTK_OBJECT (preview),
						     "wtexture");
	 if (wtexture)			/* texture */
	 {
	    unsigned	i;
	    proplist_t  *array = get_texture_array (wtexture);
	    
	    for (i = 1; i < PLGetNumberOfElements (array); i++)
	       if (PLGetArrayElement (array, i) == value)
	       {
		  PLRemoveArrayElement (array, i);
		  PLInsertArrayElement (array, PLMakeString (tmp), i);
		  gtk_object_set_user_data (GTK_OBJECT (preview),
					    PLGetArrayElement (array, i));
		  fill_preview (preview, tmp);
		  PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
		  break;
  	       }
	 }
	 else				/* color attribute */
	 {
	    proplist_t *plval = PLMakeString (tmp);

	    gtk_object_set_user_data (GTK_OBJECT (preview), plval);
	    fill_preview (preview, tmp);
	    PLInsertDictionaryEntry (windowmaker, key, plval);
	    PLInsertDictionaryEntry (plist_changed, key, pl_yes);
	 }
      }
      changed = YES;			
   }
}

static void
file_selection (GtkWidget *widget, wtexture_t *wtexture)
/*
 *  Show pixmap file selection popup window.
 *
 *  No return value.
 */
{
   static GtkWidget	*filesel = NULL;

   if (!filesel)
   {
      GtkFileSelection *fs;

      /*
       *  Fileselection dialog window
       */
      filesel = gtk_file_selection_new ("Pixmap file browser");
      fs      = GTK_FILE_SELECTION (filesel);
      gtk_file_selection_hide_fileop_buttons (fs);
      gtk_window_position (GTK_WINDOW (filesel), GTK_WIN_POS_MOUSE);
      gtk_object_set_user_data (GTK_OBJECT (fs), wtexture);
      gtk_signal_connect (GTK_OBJECT (filesel), "destroy",
                          GTK_SIGNAL_FUNC (gtk_widget_destroyed),
                          &filesel);
      gtk_signal_connect (GTK_OBJECT (fs->ok_button), "clicked",
			  GTK_SIGNAL_FUNC (set_pixmap),
			  (gpointer) filesel);
      gtk_signal_connect_object (GTK_OBJECT (fs->cancel_button), "clicked",
				 GTK_SIGNAL_FUNC (gtk_widget_destroy),
				 GTK_OBJECT (filesel));
      gtk_signal_connect_object (GTK_OBJECT (fs->ok_button), "clicked",
				 GTK_SIGNAL_FUNC (gtk_widget_destroy),
				 GTK_OBJECT (filesel));
      
      /*
       *  Initial filename and pixmap display window
       */
      {
	 proplist_t	*array = get_texture_array (wtexture);
	 char		*filename;
	 
	 filename = get_pixmap_path (PLGetString (PLGetArrayElement (array, 1)));
	 if (filename)
	    gtk_file_selection_set_filename (fs, filename);
	 /*
	  *  Display of pixmap 
	  */
	 {
	    GtkWidget *frame  = gtk_frame_new (NULL);
	    GtkWidget *hbox   = gtk_hbox_new (FALSE, 0);
	    GtkWidget *pixmap = show_pixmap (KEY_OF_TEXTURE,
					     filename
					     ? filename
					     : PKGDATADIR "/black.xpm", NULL);

	    gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	    gtk_container_border_width (GTK_CONTAINER (frame), 0);
	    gtk_container_add (GTK_CONTAINER (frame), pixmap);
	    
	    gtk_box_pack_start (GTK_BOX (fs->main_vbox), hbox, FALSE, FALSE, 10);
	    gtk_box_pack_start (GTK_BOX (hbox), frame, TRUE, FALSE, 10);

	    gtk_signal_connect (GTK_OBJECT (fs->selection_entry), "changed",
				GTK_SIGNAL_FUNC (update_pixmap), fs);
	    gtk_object_set_data (GTK_OBJECT (fs), "preview", pixmap);
	    gtk_widget_show_all (hbox);
	 }
	 if (filename)
	    Free (filename);
      }
      /*
       *  Optionmenu pixmap path
       */
      {
	 const char	**pathlist;
	 unsigned	n;
	 GtkWidget	*box;
	 GtkWidget	**wpixmap;
	 
	 wpixmap = gtk_object_get_user_data (GTK_OBJECT (widget));
	 pathlist = Calloc (PLGetNumberOfElements (pixmap_path) + 1,
			    sizeof (char *));

	 for (n = 0; n < PLGetNumberOfElements (pixmap_path); n++)
	    pathlist [n] = PLGetString (PLGetArrayElement (pixmap_path, n));
	 pathlist [n] = NULL;

	 box = generate_option_menu ("Pixmap path:", tooltips,
				     "Window Maker's searchpath for pixmaps",
				     pathlist, pathlist [0],
				     change_path, filesel);
	 gtk_box_pack_start (GTK_BOX (fs->main_vbox), box, FALSE, TRUE, 5);
      }
   }
   
   if (!GTK_WIDGET_VISIBLE (filesel))
      gtk_widget_show (filesel);
   else
      gtk_widget_destroy (filesel);
}

static void
update_pixmap (GtkWidget *widget, GtkFileSelection *fs)
{
   char       *path     = gtk_entry_get_text (GTK_ENTRY (widget));
   wtexture_t *wtexture = gtk_object_get_user_data (GTK_OBJECT (fs));
   GtkWidget  *pixmap	= gtk_object_get_data (GTK_OBJECT (fs), "preview");
   char	      *filename;

   if (!strlen (path) || (strlen (path) && path [strlen (path) - 1] == '/'))
      return;

   filename = get_pixmap_path (path);
   if (filename)
   {
      proplist_t *array = get_texture_array (wtexture);

      show_pixmap (KEY_OF_TEXTURE, filename, pixmap);
      Free (filename);
   }
}

static void
change_path (GtkWidget *widget, gpointer filesel)
/*
 *  Change current path of fileselection dialog 'filesel'.
 *  'path' is given by user data of the option menu 'widget'.
 *
 *  No return value.
 */
{
   const char *path = gtk_object_get_user_data (GTK_OBJECT (widget));

   if (strlen (path) && path [strlen (path) - 1] != '/') /* append a slash */
   {
      char *tmp = Calloc (strlen (path) + 2, sizeof (char *));

      strcpy (tmp, path);
      strcat (tmp, "/");
      gtk_file_selection_set_filename (GTK_FILE_SELECTION (filesel), tmp);
      Free (tmp);
   }
   else
      gtk_file_selection_set_filename (GTK_FILE_SELECTION (filesel), path);
}

static void
set_pixmap (GtkWidget *widget, gpointer filesel)
/*
 *  Set pixmap name of texture.
 *
 *  No return value.
 *
 *  Side effects:
 *	value of wmaker array element 1 of attribute wtexture->key is modified.
 *	update of filename entry and pixmap
 *	
 */
{
   const char		*filename;
   GtkFileSelection	*fs       = GTK_FILE_SELECTION (filesel);
   wtexture_t		*wtexture = gtk_object_get_user_data (GTK_OBJECT (fs));
   char			*path     = gtk_file_selection_get_filename (fs);

   if ((filename = strrchr (path, '/'))) /* remove path components */
      filename++;
   else
      filename = path;
   /*
    *  Check if pixmap is in PixmapPath
    */
   {
      char *tmp = get_pixmap_path (filename);

      if (!tmp)				/* pixmap not found in PixmapPath */
	 filename = path;
      else
	 Free (tmp);
   }
   
   {
      proplist_t *array = get_texture_array (wtexture);
      
      PLRemoveArrayElement (array, 1);
      PLInsertArrayElement (array, PLMakeString ((char *) filename), 1);

      show_pixmap (KEY_OF_TEXTURE, filename, wtexture->wpixmap [PIXMAP_PIXMAP]);
      gtk_entry_set_text (GTK_ENTRY (wtexture->wpixmap [PIXMAP_NAME]), filename);

#ifdef TILE_OF_DAY
      check_valid_tileofday (wtexture);
#endif /* TILE_OF_DAY */

      PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
      changed = YES;
   }
   
}

static GtkWidget *
new_list_item (wtexture_t *wtexture, proplist_t *color)
/*
 *  Allocate new color cell in color list.
 *
 *  Return value:
 *	list item
 */
{
   GtkWidget *list_item;
   GtkWidget *preview;
   GtkWidget *lhbox;
      
   list_item = gtk_list_item_new ();
   gtk_widget_show (list_item);
   gtk_container_border_width (GTK_CONTAINER (list_item), 2);
      
   lhbox = gtk_hbox_new (FALSE, 2);
   gtk_widget_show (lhbox);
   gtk_container_add (GTK_CONTAINER (list_item), lhbox);
			
   preview = gtk_preview_new (GTK_PREVIEW_COLOR);
   gtk_object_set_data (GTK_OBJECT (preview), "wtexture", wtexture);
   gtk_preview_size (GTK_PREVIEW (preview), 30, 15);
   gtk_box_pack_start (GTK_BOX (lhbox), preview, FALSE, TRUE, 5);
   gtk_widget_show (preview);
   fill_preview (GTK_PREVIEW (preview), PLGetString (color));
   gtk_object_set_user_data (GTK_OBJECT (list_item), preview);
   gtk_object_set_user_data (GTK_OBJECT (preview), color);

   return list_item;
}

static char *last_color = NULL;		/* save last color name */

static gint
list_button_press (GtkWidget *widget, GdkEventButton *event)
{
   if (event->type == GDK_2BUTTON_PRESS)
      list_change (widget, widget);
}

static void
list_insert (GtkWidget *widget, GtkWidget *list)
/*
 *  Insert new color cell before current position.
 *
 *  No return value.
 *
 *  Side effects:
 *	new element is inserted in wmaker array attribute wtexture->key
 */
{
   GtkWidget  *list_item;
   int	      position;
   proplist_t *key      = gtk_object_get_user_data (GTK_OBJECT (list));
   wtexture_t *wtexture = gtk_object_get_user_data (GTK_OBJECT (widget));
   proplist_t *array    = get_texture_array (wtexture);
   proplist_t *element  = PLMakeString (last_color?last_color:g_strdup ("black"));

   position = gtk_list_child_position (GTK_LIST (list),
				       (GTK_LIST (list)->selection)->data);
   
   PLInsertArrayElement (array, element, position + 2);
   list_item = new_list_item (wtexture, element);
   /*
    *  Insert list item at selection
    */
   {
      GList *gl = Calloc (1, sizeof (GList));
      
      gl->next = gl->prev = NULL;
      gl->data = list_item;

      gtk_list_insert_items (GTK_LIST (list), gl, position);
   }

   if (g_list_length (GTK_LIST (list)->children) > 2)
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], TRUE);
   else
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], FALSE);

   gtk_list_select_item (GTK_LIST (list), position);
   list_change (widget, list);
   PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
   changed = YES;
}

static void
list_append (GtkWidget *widget, GtkWidget *list)
/*
 *  Append new color cell after last position.
 *
 *  No return value.
 *
 *  Side effects:
 *	new element is inserted in wmaker array attribute wtexture->key
 */
{
   proplist_t *key      = gtk_object_get_user_data (GTK_OBJECT (list));
   wtexture_t *wtexture = gtk_object_get_user_data (GTK_OBJECT (widget));
   proplist_t *array    = get_texture_array (wtexture);
   proplist_t *element  = PLMakeString (last_color?last_color:g_strdup("black"));

   PLAppendArrayElement (array, element);
   
   gtk_container_add (GTK_CONTAINER (list), new_list_item (wtexture, element));
   if (g_list_length (GTK_LIST (list)->children) > 2)
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], TRUE);
   else
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], FALSE);

   gtk_list_select_item (GTK_LIST (list),
			 g_list_length (GTK_LIST (list)->children) - 1);
   list_change (widget, list);

   PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
   changed = YES;
}

static void
list_remove (GtkWidget *widget, GtkWidget *list)
/*
 *  Remove color cell at current position.
 *
 *  No return value.
 *
 *  Side effects:
 *	one element of wmaker array attribute wtexture->key is removed.
 */
{
   GList      *tmp_list;
   GList      *clear_list;
   unsigned   i;
   proplist_t *key      = gtk_object_get_user_data (GTK_OBJECT (list));
   wtexture_t *wtexture = gtk_object_get_user_data (GTK_OBJECT (widget));
   proplist_t *array    = get_texture_array (wtexture);
   int	      position;

   position = gtk_list_child_position (GTK_LIST (list),
				       (GTK_LIST (list)->selection)->data);
   
   tmp_list   = GTK_LIST (list)->selection;
   clear_list = NULL;

   /*
    *  Remove color from proplist array
    */
   for (i = 2; i < PLGetNumberOfElements (array); i++)
   {
      GtkPreview *p = gtk_object_get_user_data (GTK_OBJECT (tmp_list->data));
      
      if (PLGetArrayElement (array, i)
	  == gtk_object_get_user_data (GTK_OBJECT (p)))
      {
	 if (last_color)
	    Free (last_color);
	 last_color = g_strdup (PLGetString (PLGetArrayElement (array, i)));
	 PLRemoveArrayElement (array, i);
	 break;
      }
   }
   
   /*
    *  Remove list item
    */
   while (tmp_list)
   {
      clear_list = g_list_prepend (clear_list, tmp_list->data);
      tmp_list = tmp_list->next;
   }

   clear_list = g_list_reverse (clear_list);
   gtk_list_remove_items (GTK_LIST (list), clear_list);
   g_list_free (clear_list);

   if (g_list_length (GTK_LIST (list)->children))
   {
      if (position < g_list_length (GTK_LIST (list)->children))
	 gtk_list_select_item (GTK_LIST (list), position);
      else
	 gtk_list_select_item (GTK_LIST (list), position - 1);
   }

   if (g_list_length (GTK_LIST (list)->children) > 2)
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], TRUE);
   else
      gtk_widget_set_sensitive (wtexture->wgradient [GRADIENT_REMOVE], FALSE);

   PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
   changed = YES;
}

static void
list_change (GtkWidget *widget, GtkWidget *list)
/*
 *  Popup a colorselection dialog to change select color cell.
 *
 *  No return value.
 *
 *  Side effects:
 *	value of wmaker array attribute wtexture->key is modified.
 */
{
   GtkWidget	*list_item;

   list_item = (GTK_LIST (list)->selection)->data;

   color_selection_dialog (list_item,
			   gtk_object_get_user_data (GTK_OBJECT (list)));
}

static void
set_pixmap_name (GtkWidget *widget, gpointer ptr)
/*
 *  Update pixmap string value.
 *
 *  No return value.
 */
{
   wtexture_t *wtexture  = (wtexture_t *) ptr;
   proplist_t *attribute = wtexture->key;
   proplist_t *element;
   proplist_t *array = get_texture_array (wtexture);

   element = PLMakeString ((char *) gtk_entry_get_text (GTK_ENTRY (widget)));

   PLRemoveArrayElement (array, 1);	/* filename */
   PLInsertArrayElement (array, element, 1);
   
#ifdef TILE_OF_DAY
   check_valid_tileofday (wtexture);
#endif /* TILE_OF_DAY */
   PLInsertDictionaryEntry (plist_changed, KEY_OF_TEXTURE, pl_yes);
   changed = YES;
}

static char **
get_workspace_names (void)
{
   char		*filename = get_gnustep_path ("Defaults/WMState");
   proplist_t	*state    = PLGetProplistWithPath (filename);
   proplist_t	*ws       = PLMakeString ("Workspaces");
   proplist_t	*name	  = PLMakeString ("Name");
   char		**array   = NULL;
   unsigned	pos       = 0;

   if (state)
   {
      proplist_t *workspaces = PLGetDictionaryEntry (state, ws);
      unsigned	 i, n;
      
      if (workspaces && (n = PLGetNumberOfElements (workspaces)) > 0)
      {
	 array         = Calloc (n + 2, sizeof (char *));
	 array [pos++] = g_strdup ("Default");
	 for (i = 0; i < n; i++)
	 {
	    proplist_t *wspace = PLGetArrayElement (workspaces, i);

	    if (wspace)
	    {
	       proplist_t *wname = PLGetDictionaryEntry (wspace, name);

	       array [pos++] = wname ? PLGetString (wname) : NULL;
	    }
	    else
	       array [pos++] = NULL;
	 }
      }
   }

   if (!array)
   {
      array         = Calloc (2, sizeof (char *));
      array [pos++] = g_strdup ("Default");
      array [pos++] = NULL;
   }
   else
      array [pos++] = NULL;

   Free (filename);
   PLRelease (ws);
   PLRelease (name);
   
   return array;
}

