#include "mfindscrbar.h"
#include <qpainter.h>
#include <qbitmap.h>

enum ScrollControl { ADD_LINE = 0x1, SUB_LINE = 0x2, ADD_PAGE = 0x4,
		     SUB_PAGE = 0x8, FIRST = 0x10, LAST = 0x20,
		     SLIDER = 0x40, NONE = 0x80 };

class MfindScrBar_Private : public MfindScrBar
{
public:
	void sliderMinMax(int *, int *) const;
	void metrics(int *, int *, int *) const;

	ScrollControl pointOver(const QPoint &p) const;

	int rangeValueToSliderPos(int val) const;
	int sliderPosToRangeValue(int  val) const;

	void action(ScrollControl control);

	void drawControls(uint controls, uint activeControl) const;
	void drawControls(uint controls, uint activeControl, QPainter *p) const;
};

#undef PRIV
#define PRIV ((MfindScrBar_Private *)this)

static const int thresholdTime = 500;
static const int repeatTime	= 10;

#define HORIZONTAL (orientation() == Horizontal)
#define VERTICAL	!HORIZONTAL
#define MOTIF_BORDER	2
#define SLIDER_MIN 9 // ### motif says 6 but that's too small

MfindScrBar::MfindScrBar(QWidget *parent, const char *name)
    			:QWidget(parent, name)
{
	QColor x = QColor(160, 164, 160);
	QColorGroup CGnorm = QColorGroup(black, x, x.light(), x.dark(), x.dark(120), black, x);
	setPalette(QPalette(CGnorm, CGnorm, CGnorm));

	orient = Vertical;
	init();
}

void MfindScrBar::init()
{
	track = TRUE;
	sliderPos = 0;
	pressedControl = NONE;
	clickedAt = FALSE;
	setFocusPolicy(NoFocus);

	setBackgroundMode(PaletteBackground);
}

void MfindScrBar::valueChange()
{
	int tmp = sliderPos;
	positionSliderFromValue();
	if(tmp != sliderPos)
		PRIV->drawControls(ADD_PAGE | SLIDER | SUB_PAGE, pressedControl);
	emit valueChanged(value());
}

void MfindScrBar::stepChange()
{
    rangeChange();
}

void MfindScrBar::rangeChange()
{
	positionSliderFromValue();
	PRIV->drawControls(ADD_LINE | ADD_PAGE | SLIDER | SUB_PAGE | SUB_LINE, pressedControl);
}

void MfindScrBar::timerEvent(QTimerEvent *)
{
	if(!isTiming)
		return;
	if(!thresholdReached) {
		thresholdReached = TRUE;
		killTimers();
		startTimer(repeatTime);
	}
	if(clickedAt)
		PRIV->action((ScrollControl) pressedControl);
}

void MfindScrBar::resizeEvent(QResizeEvent *)
{
	positionSliderFromValue();
}

void MfindScrBar::paintEvent(QPaintEvent *e)
{
	QPainter p(this);
	if(e)
		p.setClipRect(e->rect());

	qDrawShadePanel(&p, rect(), colorGroup(), TRUE);

	PRIV->drawControls(ADD_LINE | SUB_LINE | ADD_PAGE | SUB_PAGE | SLIDER, pressedControl, &p);
}

static QCOORD sliderStartPos = 0;

void MfindScrBar::mousePressEvent(QMouseEvent *e)
{
	if(e->button() != LeftButton)
		return;

	if(maxValue() == minValue())
		return;

	clickedAt = TRUE;
	pressedControl = PRIV->pointOver(e->pos());

	if(pressedControl == SLIDER) {
		clickOffset = (QCOORD)(e->pos().y() - sliderPos);
		slidePrevVal = value();
		sliderStartPos = sliderPos;
		emit sliderPressed();
	} else if (pressedControl != NONE) {
		PRIV->drawControls(pressedControl, pressedControl);
		PRIV->action((ScrollControl) pressedControl);
		thresholdReached = FALSE;
		startTimer(thresholdTime);
		isTiming = TRUE;
	}
}

void MfindScrBar::mouseReleaseEvent(QMouseEvent *e)
{
	if(!clickedAt || !(e->button() == LeftButton))
		return;

	ScrollControl tmp = (ScrollControl) pressedControl;
	clickedAt = FALSE;
	if(isTiming)
		killTimers();
	mouseMoveEvent(e);  // Might have moved since last mouse move event.
	pressedControl = NONE;

	switch(tmp) {
		case SLIDER: // Set value directly, we know we don't have to redraw.
			directSetValue(calculateValueFromSlider());
	    	emit sliderReleased();
	    	if(value() != prevValue())
				emit valueChanged(value());
			break;
		case ADD_LINE:
		case SUB_LINE:
	    	PRIV->drawControls(tmp, pressedControl);
	    	break;
		default:
	    	break;
	}
}

void MfindScrBar::mouseMoveEvent(QMouseEvent *e)
{
	if(!isVisible()) {
		clickedAt = FALSE;
		return;
	}
 	if(!clickedAt || !(e->state() & LeftButton))
		return;

	int newSliderPos;
	if(pressedControl == SLIDER) {
		int sliderMin, sliderMax;
		PRIV->sliderMinMax(&sliderMin, &sliderMax);
		QRect r = rect();
		r.setRect(r.x() - 30, r.y() - 20, r.width() + 60, r.height() + 40);

		if(!r.contains(e->pos()))
	    	newSliderPos = sliderStartPos;
		else
	    	newSliderPos = e->pos().y() - clickOffset;

		if(newSliderPos < sliderMin)
	    	newSliderPos = sliderMin;
		else if(newSliderPos > sliderMax)
	    	newSliderPos = sliderMax;
		if(newSliderPos == sliderPos)
	    	return;
	
		int newVal = PRIV->sliderPosToRangeValue(newSliderPos);
		if(newVal != slidePrevVal)
	    	emit sliderMoved(newVal);
		if(track && newVal != value()) {
	    	directSetValue(newVal); // Set directly, painting done below
	    	emit valueChanged(value());
		}
		slidePrevVal = newVal;
		sliderPos = (QCOORD)newSliderPos;
		PRIV->drawControls(ADD_PAGE | SLIDER | SUB_PAGE, pressedControl);
	}
}

QRect MfindScrBar::sliderRect() const
{
	int sliderMin, sliderMax, sliderLength;
	PRIV->metrics(&sliderMin, &sliderMax, &sliderLength);
	int b = style() == MotifStyle ? MOTIF_BORDER : 0;

	return QRect(b, sliderStart(), width() - b*2, sliderLength);
}

void MfindScrBar::positionSliderFromValue()
{
	sliderPos = (QCOORD)PRIV->rangeValueToSliderPos(value());
}

int MfindScrBar::calculateValueFromSlider() const
{
	return PRIV->sliderPosToRangeValue(sliderPos);
}


/*****************************************************************************
  MfindScrBar_Private member functions
 *****************************************************************************/

void MfindScrBar_Private::sliderMinMax(int *sliderMin, int *sliderMax) const
{
	int dummy;
	metrics(sliderMin, sliderMax, &dummy);
}


void MfindScrBar_Private::metrics(int *sliderMin, int *sliderMax, int *sliderLength) const
{
	int buttonDim, maxLength;

	if(height() > (width() - 1)*2 + SLIDER_MIN)
		buttonDim = width();
	else
		buttonDim = (height() - SLIDER_MIN)/2 - 1;

	*sliderMin = 0;     // era   buttonDim
	
	maxLength = height() - buttonDim*2;

	if(maxValue() == minValue()) {
		*sliderLength = maxLength;
	} else {
		*sliderLength = (pageStep()*maxLength)/(maxValue()-minValue()+pageStep());
		if(*sliderLength < SLIDER_MIN)
	    	*sliderLength = SLIDER_MIN;
		if(*sliderLength > maxLength)
	    	*sliderLength = maxLength;
	}
	*sliderMax = *sliderMin + maxLength - *sliderLength;
}

ScrollControl MfindScrBar_Private::pointOver(const QPoint &p) const
{
	if(!rect().contains(p))
		return NONE;

	int sliderMin, sliderMax, sliderLength, pos, dimB;
	metrics(&sliderMin, &sliderMax, &sliderLength);
	pos = p.y();

	if(height() > (width() - 1)*2 + SLIDER_MIN)
		dimB = width();
	else
		dimB = (height() - SLIDER_MIN)/2 - 1;

	if(pos < sliderStart())
		return SUB_PAGE;
	if(pos < sliderStart() + sliderLength)
		return SLIDER;
	if(pos < sliderMax + sliderLength)
		return ADD_PAGE;
	if(pos < height() - dimB)
		return SUB_LINE;
	return ADD_LINE;
}

int MfindScrBar_Private::rangeValueToSliderPos(int v) const
{
	int smin, smax;
	sliderMinMax(&smin, &smax);
	if(maxValue() == minValue())
		return smin;
	int sliderMin=smin, sliderMax=smax;
	return((sliderMax-sliderMin)*2*(v-minValue())+1)/
	   ((maxValue()-minValue())*2) + sliderMin;
}

int MfindScrBar_Private::sliderPosToRangeValue(int pos) const
{
    int sliderMin, sliderMax;
    sliderMinMax( &sliderMin, &sliderMax );
    if ( pos <= sliderMin || sliderMax == sliderMin )
	return minValue();
    if ( pos >= sliderMax )
	return maxValue();
    return (maxValue() - minValue() + 1)*(pos - sliderMin)/
	   (sliderMax - sliderMin) + minValue();
}

void MfindScrBar_Private::action(ScrollControl control)
{
	switch(control) {
		case ADD_LINE:
	    	emit nextLine();
	    	addLine();
	    	break;
		case SUB_LINE:
	    	emit prevLine();
	    	subtractLine();
	    	break;
		case ADD_PAGE:
	    	emit nextPage();
	    	addPage();
	    	break;
		case SUB_PAGE:
			emit prevPage();
	    	subtractPage();
	    	break;
	   default:
	   	break;
	}
}


void MfindScrBar_Private::drawControls(uint controls, uint activeControl) const
{
	QPainter p(this);
	drawControls(controls, activeControl, &p);
}


void MfindScrBar_Private::drawControls(uint controls, uint activeControl, QPainter *p) const
{
#define ADD_LINE_ACTIVE (activeControl == ADD_LINE)
#define SUB_LINE_ACTIVE (activeControl == SUB_LINE)
	QColorGroup g = colorGroup();
	int dimB;
	int sliderMin, sliderMax, sliderLength;
	
	metrics(&sliderMin, &sliderMax, &sliderLength);

	if(height() > (width() - 1)*2 + SLIDER_MIN)
		dimB = width();
	else
		dimB = (height() - SLIDER_MIN)/2 - 1;
	QRect addB;
	QRect subB;
	QRect addPageR;
	QRect subPageR;
	QRect sliderR;
	int addX, addY, subX, subY;

	subX = addX = (width() - dimB) / 2;
	subY = height() - dimB*2;
	addY = height() - dimB;   

	subB.setRect(subX,subY,dimB,dimB);
	addB.setRect(addX,addY,dimB,dimB);

	int sliderEnd = sliderStart() + sliderLength;
	int sliderW = width();
	subPageR.setRect(0, 0, sliderW, sliderStart());
			  			  
	addPageR.setRect(0, sliderEnd, sliderW, subY - sliderEnd);
	
	sliderR.setRect(0, sliderStart(), sliderW, sliderLength);

	bool maxedOut = (maxValue() == minValue());
	if(controls & SUB_LINE) {
	   qDrawWinPanel(p, subB.x(), subB.y(),
			subB.width(), subB.height(), g, SUB_LINE_ACTIVE);
	   qDrawArrow(p, UpArrow, WindowsStyle, SUB_LINE_ACTIVE, 
	   	subB.x()+2, subB.y()+2,
				subB.width()-4, subB.height()-4, g, !maxedOut);
	}
	if(controls & ADD_LINE) {
		qDrawWinPanel(p, addB.x(), addB.y(),
			 addB.width(), addB.height(), g, ADD_LINE_ACTIVE);
	   qDrawArrow(p, DownArrow, WindowsStyle, ADD_LINE_ACTIVE, 
			addB.x()+2, addB.y()+2,
				addB.width()-4, addB.height()-4, g, !maxedOut);
	}
	p->setBrush(QBrush(QColor(140, 140, 140)));
	p->setPen(NoPen);
	p->setBackgroundMode(OpaqueMode);
	if(maxedOut) {
	   p->drawRect(sliderR);
	} else {
	   if(controls & SUB_PAGE)
			p->drawRect(subPageR);
	   if(controls & ADD_PAGE)
			p->drawRect(addPageR);
	   if(controls & SLIDER) {
			if (!maxedOut) {
		    	QBrush fill(g.background());
		    	qDrawWinPanel(p, sliderR.x(), sliderR.y(),
				   sliderR.width(), sliderR.height(), g,
				   	FALSE, &fill);
			}
		}
	}
	if(hasFocus() && (controls & SLIDER))
		 p->drawWinFocusRect(sliderR.x()+2, sliderR.y()+2,
				 sliderR.width()-5, sliderR.height()-5, 
				 backgroundColor());
}

#undef ADD_LINE_ACTIVE
#undef SUB_LINE_ACTIVE

void qDrawArrow(QPainter *p, ArrowType type, GUIStyle style, bool down,
		 int x, int y, int w, int h, const QColorGroup & g)
{
    qDrawArrow(p, type, style, down, x, y, w, h, g, TRUE);
}
