/*
    Copyright (C) 1998 by Jorrit Tyberghein
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef POLYSET_H
#define POLYSET_H

#include "com/com.h"
#include "geometry/math3d.h"
#include "geometry/math2d.h" // texel coords
#include "geometry/polyint.h"
#include "engine/polygon/ipolygon.h"
#include "engine/basic/fog.h"
#include "colldet/collp.h"

class PolygonInt;
class csSector;
class csWorld;
class csCamera;
class Textures;
class TextureMM;
class CLights;
class Polygon3D;
class Collider;
class BspTree;
class Dumper;
class RenderView;
interface IPolygonSet;
class csCurve;


/**
 * A PolygonSet class is a set of polygons (amazing, isn't it :-)
 * A PolygonSet describes a set of polygons that form a convex and
 * (probably) closed hull. All polygons in a set share vertices
 * from the same pool.<p>
 *
 * A recent extension also allows PolygonSets that are not convex
 * by adding an optional BSP tree. This BSP tree is only useful
 * if the PolygonSet is concave. Otherwise the tree is wasted.<p>
 *
 * Every polygon in the set has a visible and an invisible face;
 * if the vertices of the polygon are ordered clockwise then the
 * polygon is visible. Using this feature it is possible to define
 * two kinds of PolygonSets: in one kind the polygons are oriented
 * such that they are visible from within the hull. In other words,
 * the polygons form a sort of container or room where the camera
 * can be located. We call this kind of PolygonSet a csSector (a
 * subclass of PolygonSet). In another kind the polygons are
 * oriented such that they are visible from the outside. We call
 * this kind of PolygonSet a Thing (another subclass of PolygonSet).<p>
 *
 * Things and csSectors have many similarities. That's why the
 * PolygonSet class was created: to exploit these similarities.
 * However, there are some important differences between Things and
 * csSectors:<p>
 * <ul>
 * <li> Currently, only things can move. This means that the object
 *      space coordinates of a csSector are ALWAYS equal to the world
 *      space coordinates. It would be possible to allow moveable
 *      csSectors but I don't how this should be integrated into an
 *      easy model of the world.
 * <li> Things do not require portals but can use them. 
 * </ul>
 */
class PolygonSet : public PolygonParentInt
{
  friend class Dumper;
  friend class Collider;
  friend class csSector;

protected:
  /**
   * PolygonSets are linked either in a csWorld object or in another
   * PolygonSet (Thing in csSector for example).
   */
  PolygonSet* next;

  int num_vertices;
  int max_vertices;
  
  /// Vertices in world space.
  Vector3* wor_verts;
  /// Vertices in object space.
  Vector3* obj_verts;
  /// Vertices in camera space.
  Vector3* cam_verts;

  /// Table of ptr to polygons forming the outside of the set
  PolygonInt** polygons;
  /// Number of polygons used.
  int num_polygon;
  /// Maximum number of polygons in 'polygon' array.
  int max_polygon;

  /// Table of ptr to curves forming the outside of the set
  csCurve** curves;
  /// Number of curves used.
  int num_curves;
  /// Maximum number of curves in 'curves' array.
  int max_curves;

  /// csSector where this polyset belongs (pointer to 'this' if it is a sector).
  csSector* sector;

  /// Optional bsp tree.
  BspTree* bsp;

  /**
   * Get a new transformation set of vertices (camera space) and return
   * the old set. This function is required when transforming to camera
   * space because this polyset may already have been transformed to
   * camera space in some other way. It is important to restore the
   * transformation set if you are done with it.
   * The new transformation set is automatically put in the standard
   * 'cam_verts' array and accessible using 'vcam'.
   * Do not use the result from this function except for giving it
   * to restore_transformation. The result is not meaningfull.
   * This function also depends on the correct value of 'draw_busy'.
   * 'draw_busy' should be equal to 1 for the first time you call
   * this 'new_transformation' and should also be equal to 1 when
   * you restore the set for the last time.
   */
  Vector3* new_transformation ();

  /**
   * After calling 'new_transformation' you MUST call this
   * function with the result from 'new_transformation'!
   */
  void restore_transformation (Vector3* old);

  /**
   * Draw the given array of polygons from this PolygonSet. This
   * function is called by subclasses of PolygonSet (csSector and
   * Thing currently).
   */
  void draw_polygon_array (PolygonInt** polygon, int num, RenderView* rview, bool use_z_buf);

private:
  /**
   * Fog information.
   */
  csFog fog;

public:
  /**
   * How many times are we busy drawing this polyset (recursive).
   * This is an important variable as it indicates to
   * 'new_transformation' which set of camera vertices it should
   * use.
   */
  int draw_busy;


  /// Tesselation parameters
  /// Center of thing to determine distance from
  Vector3 curves_center;
  /// scale param (the larger this param it, the more the curves are tesselated)
  float   curves_scale;  

  /**
   * curve vertices. 
   */
  /// Vertices to be used by curves
  Vector3* curve_vertices;
  /// Texture coords of curve vertices
  Vector2* curve_texels;

  /// Number of vertices.
  int num_curve_vertices;
  /// Maximum number of vertices.
  int max_curve_vertices;


  /**
   * Construct a PolygonSet. 'type' is the csObject type and should
   * be one of CS_THING or CS_SECTOR.
   */
  PolygonSet (char* name, int type);

  ///
  virtual ~PolygonSet ();

  /**
   * Just add a new vertex to the polygonset.
   */
  int add_vertex (Vector3& v) { return add_vertex (v.x, v.y, v.z); }

  /**
   * Just add a new vertex to the polygonset.
   */
  int add_vertex (float x, float y, float z);

  /**
   * Only add a new vertex if it doesn't already exist.
   */
  int add_vertex_smart (Vector3& v) { return add_vertex_smart (v.x, v.y, v.z); }

  /**
   * Only add a new vertex if it doesn't already exist.
   */
  int add_vertex_smart (float x, float y, float z);

  /**
   * Return the world space vector for the vertex.
   */
  Vector3& vwor (int idx) { return wor_verts[idx]; }

  /**
   * Return the object space vector for the vertex.
   */
  Vector3& vobj (int idx) { return obj_verts[idx]; }

  /**
   * Return the camera space vector for the vertex.
   */
  Vector3& vcam (int idx) { return cam_verts[idx]; }

  /**
   * Return the number of vertices.
   */
  int get_num_vertices () { return num_vertices; }

  /**
   * Add a polygon to this polygonset.
   */
  void add_polygon (PolygonInt* spoly);

  /**
   * Create a new polygon in this polygonset.
   */
  Polygon3D* new_polygon (char* name, TextureMM* texture);

  /**
   * Get the number of polygons in this polygonset.
   */
  int get_num_polygon () { return num_polygon; }

  /**
   * Get the specified polygon from this set.
   */
  PolygonInt* get_polygon (int idx) { return polygons[idx]; }

  /**
   * Get the named polygon from this set.
   */
  Polygon3D* get_polygon (char* name);

  /**
   * Add a curve to this polygonset.
   */
  void add_curve (csCurve* curve);

  /**
   * Get the number of curves in this polygonset.
   */
  int get_num_curves () { return num_curves; }

  /**
   * Get the specified polygon from this set.
   */
  csCurve* get_curve (int idx) { return curves[idx]; }

  /**
   * Get the named polygon from this set.
   */
  csCurve* get_curve (char* name);

  int get_num_curve_vertices () { return num_curve_vertices; }
  ///
  Vector3& curve_vertex (int i) { return curve_vertices[i]; }
  ///
  Vector2& curve_texel (int i) { return curve_texels[i]; }
  ///
  void add_curve_vertex (Vector3& v, Vector2& t);
  ///


  /**
   * Enable the optional BSP tree. This should be done AFTER
   * all polygons and vertices have been added and all polygons
   * are correctly initialized (textures and planes).
   */
  void use_bsp ();

  /// Return true if there is a BSP tree used in this sector.
  bool is_bsp () { return bsp != NULL; }

  /**
   * Intersect world-space segment with polygons of this set. Return
   * polygon it intersects with (or NULL) and the intersection point
   * in world coordinates.<p>
   *
   * If 'pr' != NULL it will also return a value between 0 and 1
   * indicating where on the 'start'-'end' vector the intersection
   * happened.
   */
  virtual Polygon3D* intersect_segment (const Vector3& start, 
                                        const Vector3& end, Vector3& isect,
				        float* pr = NULL);

  /**
   * Return false if none of the vertices of this sector is in front
   * of the camera.
   */
  bool transform_world2cam (csCamera& c);

  /**
   * Translate the vertices of this polygonset so that the given
   * vector is at the origin. The resulting transformation is
   * put in the 'camera' section.
   */
  void translate_vector (Vector3& trans);

  /**
   * Get the next polygonset in its linked list.
   */
  PolygonSet* get_next () { return next; }

  /**
   * Set the next polygonset in its linked list.
   */
  void set_next (PolygonSet* next) { PolygonSet::next = next; }

  /**
   * Set the sector that this polygonset belongs to.
   * This is either the polygonset itself if it is a sector
   * or else the sector that this thing is in.
   */
  void set_sector (csSector* sector) { PolygonSet::sector = sector; }

  /// Return the sector that this polygonset belongs to.
  csSector* get_sector() { return sector; }

  /// Return true if this has fog.
  bool has_fog () { return fog.enabled; }

  /// Return fog structure.
  csFog& get_fog () { return fog; }

  /*
   * Collision Detection data.
   */
  ColliderP collider;

  DECLARE_INTERFACE_TABLE( PolygonSet )
  DECLARE_IUNKNOWN()

  DECLARE_COMPOSITE_INTERFACE( PolygonSet )
};

#define GetIPolygonSetFromPolygonSet(a)  &a->m_xPolygonSet;
#define GetPolygonSetFromIPolygonSet(a)  ((PolygonSet*)((size_t)a - offsetof(PolygonSet, m_xPolygonSet)))

#endif /*POLYSET_H*/
